/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access;

import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Resource;

/**
 * Finds objects within a particular repository.
 * <p>
 * Provides various utilities for finding NDR objects based upon different
 * criteria. Is likely to be expanded as need dictates.
 * </p>
 * 
 * @author birkland
 * 
 * @since 1.0
 */
public interface Finder {

    /**
     * Find and return a resource given an identifier.
     * <p>
     * Convenience method for finding resources by their resource identifier
     * (typically a URL).
     * </p>
     * 
     * @param idType
     *            the "type" of identifier. This is the predicate of the
     *            property that contains the identifier. See
     *            {@link org.nsdl.repository.model.content.ResourceModel.Identifiers}.
     * @param identifier
     *            the identifier value.
     * @return instance of the matching resource, null if none are found.
     */
    public Resource findResource(Property.Name idType, String identifier);

    /**
     * Find all objects that match a given object template
     * <p>
     * Finds all objects that
     * <ul>
     * <li>have the same type as the template oject</li>
     * <li>have the same state as the template object</li>
     * <li>have every property + value found in the template object</li>
     * <li>have every relationship + related object found in the template
     * object
     * <li>
     * <li>have every datastream (though not necessarily the same datastream
     * content)</li>
     * <li>have every disseminator (though not necessarily the same bmech)</li>
     * </ul>
     * </p>
     * 
     * @param template
     *            object containing every value to match
     * @return Iterable result set containing the identity of every matching
     *         object
     */
    public Results<NDRObjectInfo> matchObject(NDRObject template);

    /**
     * Find all relationships that point 'in' to an object
     * 
     * @param relatedTo
     *            object to which we are finding relationships that point
     * @return Results containing every inward relationship
     */
    public Results<Relationship> getRelatedTo(NDRObject relatedTo);

    /**
     * Find instances of a given relationship that point 'in' to an object
     * 
     * @param relatedTo
     *            object to which we are finding relatiomships that point
     * @param relationship
     *            the relationship that is desired
     * @return Results containing every inward relationship
     */
    public Results<Relationship> getRelatedTo(NDRObject relatedTo,
            Relationship.Name relationship);
}
