/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access;

import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.types.Type;

/**
 * Main access point for interacting with an NDR repository
 * <p>
 * Defines the Create, Read, Update, and Delete operations on individual
 * repository objects.
 * </p>
 * 
 * @author birkland
 * 
 * @since 1.0
 */
public interface NDRAccess {

    public static final String PROP_REPOSITORY_BASE = "param.repository.baseURL";

    public NDRObject getObject(String identity) throws AccessException;

    public <T extends NDRObject> T getObject(String identity, Type<T> type)
            throws ExistenceException;

    /**
     * Retrieve an NDR object for viewing.
     * <p>
     * Returns an {@link NDRObject} representing the current state of the object
     * in the repository. Note: its content is only guaranteed to be current as
     * of time of invocation. If the object subsequently modified in the
     * repository, its state change will not necessarily be reflected in this
     * object instance.
     * </p>
     * 
     * @param identity
     *            The identity of the object to get.
     * @return Object instance from the repository.
     * @throws AccessException
     *             if there is a problem retrieving the object (if it doesn't
     *             exist, for example).
     */
    public NDRObject getObject(NDRObjectInfo identity) throws AccessException;

    /**
     * Get a concrete object instance of the given type.
     * 
     * @param <T>
     *            Concrete type (e.g. from
     *            <code>org.nsdl.repository.model.types</code>).
     * @param identity
     *            Identity of the object to get.
     * @param type
     *            Type of object expected.
     * @return Object instance from the repository.
     * @throws ExistenceException
     *             will be thrown if the object does not exist, or if an object
     *             does exist with the given identity, but it does not match the
     *             given type.
     */
    public <T extends NDRObject> T getObject(NDRObjectInfo identity,
            Type<T> type) throws ExistenceException;

    /**
     * Creates a new object of a given type for later insertion into the
     * repository.
     * <p>
     * Upon creation, the object will populated with some initial data,
     * including (but not limited to)
     * <ul>
     * <li>The given object type</li>
     * <li>An initial state of 'Active'</li>
     * </ul>
     * </p>
     * 
     * <p>
     * Creating a new object does not cause any change to the repository state.
     * After the new object is externally manipulated, it may be added to the
     * repository by two means: it's own
     * {@link org.nsdl.repository.model.NDRObject#commit() commit} method or the
     * Repository instance's {@link #addObject(NDRObject)} method. Both will
     * have exactly the same end effect.
     * </p>
     * 
     * @param type
     *            Type of object being created.
     * @return New mutable object to be populated with data.
     */
    public <T extends NDRObject> T newObject(Type<T> type);

    /**
     * Delete an object from the NDR.
     * <p>
     * Deleting an NDR object does not actually remove it from the repository,
     * It marks the object's state as 'Deleted', and possibly performs other
     * actions
     * </p>
     * 
     * @see org.nsdl.repository.model.NDRObject.State
     * 
     * @param object
     *            The object to delete.
     * @throws AccessException
     *             if there is a problem deleting.
     */
    public void deleteObject(NDRObjectInfo object) throws AccessException;

    /**
     * Removes an object completely from the NDR.
     * <p>
     * Unlike <code>deleteObject</code>, leaves no trace of the object in the
     * repository once complete. Use with caution.
     * </p>
     * 
     * @param object
     *            Identity of the object to purge.
     * @throws AccessException
     *             If there is some problem purging.
     */
    public void purgeObject(NDRObjectInfo object) throws AccessException;

    /**
     * Get an instance of an {@link Finder } to search for objects in the NDR.
     * <p>
     * The finder will perform solely on the contents of the repository from
     * which it was invoked.
     * </p>
     * 
     * @return The finder for its repository.
     */
    public Finder getObjectFinder();

    /**
     * Determines if this NDRAccess is read only
     * 
     * @return true if this NDRAccess instance cannot be written to;
     */
    public boolean isReadOnly();
}
