/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access;

import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.ConfigurationFactory;

/**
 * Creates NDRAccess instances based upon supplied configuration.
 * <p>
 * Reads the supplied configuration, and generates an {@link NDRAccess} instance
 * based on that specification. If the specified <code>NDRAccess</code> class
 * implements
 * {@link org.nsdl.repository.util.Configuration.Listener Configuration.Listener},
 * that class will be configured and initialized before returning.
 * <p>
 * 
 * <h2>Configuration properties</h2>
 * <p>
 * <dl>
 * <dt><code>param.repository.access</code></dt>
 * </dd>
 * Name of the {@link NDRAccess} implementing class that will be created by the
 * factory</dd>
 * </dl>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class NDRAccessFactory {

    /** NDRAccess impl class property */
    private static final String PROP_CLASS = "param.repository.access";

    private static final ConfigurationFactory<NDRAccess> factory =
            ConfigurationFactory.getInstance(NDRAccess.class);

    /** No need to create an instance */
    private NDRAccessFactory() {
    }

    /**
     * Build an NDRAccess as specified in the supplied configuration.
     * 
     * @param config
     *            Configuration properties.
     * @return Initialized <code>NDRAccess</code> instance.
     */
    public static NDRAccess getAccess(Configuration config) {
        return factory.buildFromProperty(PROP_CLASS, config);
    }

    /**
     * Build a specific NDRAccess impl in the supplied configuration.
     * 
     * @param ndrAccessClass
     *            Class of the desired NDRAccess implementation
     * @param config
     *            Configuration properties.
     * @return Initialized <code>NDRAccess</code> instance.
     */
    public static NDRAccess getAccess(String ndrAccessClass,
            Configuration config) {
        return factory.buildFromClassName(ndrAccessClass, config);
    }
}
