/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access;

/**
 * Allows iteration over result sets.
 * 
 * <p>
 * Result sets may be arbitrarily large, so access to them is provided as an
 * iterable stream. To conserve resources and prevent leaks, these streams
 * are closeable via <code>close</code>, which is a method present on the
 * iterable <code>Results</code>, but not present on any individual
 * <code>Iterator</code> created by it. It was done this way to make use
 * of the <code>close()</code> with the enhanced for loop feasible (which
 * essentially "hides" an Iterator instance behind the scenes which is not
 * accessible through code). For example:
 * <pre>
 * Results<NDRObjectInfo> matches = finder.matchObject(template);
 * try {
 *     for (NDRObjectInfo match : matches) {
 *        process(match);
 *     }
 * } finally {
 *     matches.close();   
 * }
 * </pre>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 * @param <T>
 *            Type returned in the results.
 */
public interface Results<T> extends Iterable<T> {

    /**
     * Close any open result iterators.
     * 
     * <p>
     * It is important to close any result iterators once they are no longer
     * needed in order to prevent resource leaks. While most implementations
     * of <code>Results</code> will auto-close individual iterators once
     * all elements have been accessed, it is safest to always
     * {@link #close()} whenever results are no longer needed, especially in
     * situations where an iterator may not have been completely exhausted
     * (such as in a loop where some exception is thrown in mid-processing).
     * Failure to close results may result in hard-to-debug problems when
     * there is high resource contention!
     * </p>
     * 
     * <p>
     * <code>close()</code> may be called any number of times with no ill
     * effect. It will close all active iterators, but will not prevent new
     * iterators from being created upon invoking <code>iterator()</code>.
     * </p>
     */
    public void close();
}
