/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access.filters;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.content.ContentModel;
import org.nsdl.repository.model.content.ContentModelException;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.ConfigurationFactory;

import static org.nsdl.repository.access.filters.FilterState.COMMIT;

/**
 * Applies content model checks to NDR objects.
 * <p>
 * For write operations (Adds, Updates, and Deletes), this filter applies any
 * applicable {@link ContentModel} checks on the object. If the operation winds
 * up violating any of the configured models, it will throw a
 * {@link ContentModelException} and terminate the operation.
 * </p>
 * 
 * <h2>Configuration Properties</h2>
 * 
 * <p>
 * <dl>
 * <dt><code>repository.access.filter.ContentModelFilter.model.*</code></dt>
 * <dd>Classes of any content models that this filter is to apply. Will execute
 * model checks in alphabetical order based on property name</dd>
 * </dl>
 * </p>
 * 
 * <p>
 * <h3>Example:</h3>
 * 
 * <pre>
 * repository.access.filter.ContentModelFilter.model.1 = org.nsdl.repository.model.content.NSDLContentModel
 * repository.access.filter.ContentModelFilter.model.2 = org.example.models.OtherContentModel
 * </pre>
 * 
 * </p>
 * 
 * @author birkland
 * 
 * @since 1.1
 * 
 */
public class ContentModelFilter implements NDRAccessFilter,
        Configuration.Listener {

    private static final String PROP_MODEL_BASE =
            "repository.access.filter.ContentModelFilter.model";

    private static final Logger logger =
            Logger.getLogger(ContentModelFilter.class);

    private List<ContentModel> models = new ArrayList<ContentModel>();

    /**
     * Apply all applicable {@link ContentModel} checks
     * <p>
     * Actually invokes all content model checks upon a given object. The
     * individual content models are expected to quickly pass any unrelated or
     * non-applicable objects
     * </p>
     * 
     * @param object
     *            The object to check for model compliance
     * @param context
     *            The context of the current operation
     * @throws ContentModelException
     *             if the object is in violation
     */
    public void doFilter(NDRObject object, FilterContext context)
            throws ContentModelException {

        if (context.state().equals(COMMIT)) {

            logger.trace("BEGIN applying content models...\n");
            for (ContentModel m : models) {
                m.checkCompliance(object);
            }
            logger.trace("END applying content models...\n");
        }
    }

    public void updateConfig(Configuration config) {
        ConfigurationFactory<ContentModel> factory =
                ConfigurationFactory.getInstance(ContentModel.class);

        for (String modelClass : config.getProperties(PROP_MODEL_BASE)) {
            models.add(factory.buildFromClassName(modelClass, config));
        }
    }

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        for (ContentModel m : models) {
            m.init(ndr);
        }
    }
}
