/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access.filters;

import java.util.Arrays;
import java.util.Collection;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.AggregatorModel.Components.AGGREGATOR_FOR;
import static org.nsdl.repository.model.content.AuthorizationModel.Grouping.Components.AUTHORIZED;
import static org.nsdl.repository.model.content.MetadataProviderModel.Components.METADTA_PROVIDER_FOR;

/**
 * For objects that should have authorization info, specifies reasonable
 * defaults.
 * <p>
 * Some objects, such as Aggregators or MetadataProviders, may have
 * authorization relationships associated with them, in order to specify which
 * agent or group of agents may modify them. If unspecified, this filter
 * explicitly authorizes the 'owning' Agent for any of its own Aggregators or
 * MetadataProviders by creating the necessary relationship.
 * </p>
 * 
 * @author birkland
 * 
 * @see org.nsdl.repository.model.content.AuthorizationModel
 * @since 1.1
 */
public class DefaultAuthorizationFilter implements NDRAccessFilter {

    private static final Logger logger =
            Logger.getLogger(DefaultAuthorizationFilter.class);

    /**
     * Add a default authorization relationship, if applicable
     * <p>
     * For Aggregator or MetadataProvider objects, if no
     * <code>authorizedToChange</code> relationships have been specified,
     * create a default relationship to its owning agent.
     * <p>
     * 
     * @param object
     *            The object to filter.
     * @param context
     *            The context of the current operation.
     */
    public void doFilter(NDRObject object, FilterContext context) {

        if (!context.state().equals(FilterState.BEFORE_COMMIT)) {
            return;
        }

        /* Exit if our object isn't an Agg. or MP */
        if (!(Type.Aggregator.equals(object.type()) || Type.MetadataProvider
                .equals(object.type()))) {
            return;
        }

        Collection<Relationship> authorized = object.getRelationships(AUTHORIZED);
        if (authorized.isEmpty()) {
            object.addRelationships(Arrays.asList(getDefaultAuthRel(object)));
        }
    }

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        /* not needed */
    }

    public static Relationship getDefaultAuthRel(NDRObject o)
            throws AccessException {
        Relationship toAgent;

        if (Type.Aggregator.equals(o.type())) {
            toAgent = o.getRelationship(AGGREGATOR_FOR);
            if (toAgent == null)
                throw new RepositoryException(
                        "Aggregator must have an aggregatorFor relationship");
        } else {
            toAgent = o.getRelationship(METADTA_PROVIDER_FOR);
            if (toAgent == null)
                throw new RepositoryException(
                        "MetadataProvider must have a metadataProviderFor "
                                + "relationship");
        }

        NDRObjectInfo agent = toAgent.related();

        logger.debug("adding default authorization to " + agent.handle());

        return new Relationship(o, AUTHORIZED, agent);
    }
}
