/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access.filters;

/**
 * Represents the current filtering context of an NDR object during a repository
 * operation.
 * <p>
 * Throughout the invocation, execution, and possible completion of an operation
 * on an object within the NDR (such as creating, adding, deleting), various
 * {@link NDRAccessFilter} may be invoked. The FilterContext persists state
 * throughout the life cycle of a single operation while also indicating the
 * current {@link FilterState} of the operation on an NDR object (e.g. upon
 * creation, during commit, etc.).
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public interface FilterContext {

    /**
     * Get a particular value that has been persisted in the context.
     * <p>
     * The invoking filter presumably knows the class of object it is seeking in
     * this specific location. If no value has been previously set in the
     * operation by this or another filter, it will return null;
     * </p>
     * 
     * @param key
     *            Name of the attribute being retrieved.
     * @return Object that has been stored at that location.
     */
    public Object getAttribute(String key);

    /**
     * Set a value to be persisted in the current filter context.
     * <p>
     * A filter may store anything here for retrieval later in the object's NDR
     * operation life cycle.
     * </p>
     * 
     * @param key
     *            Name of the attribute to store.
     * @param value
     *            Object to be stored at that location.
     */
    public void setAttribute(String key, Object value);

    /**
     * Get the current state of the NDR operation on the object
     * <p>
     * A filter will be invoked at multiple points during the execution of an
     * NDR operation. The current state at the time of each invocation is
     * observable through this method.
     * </p>
     * 
     * @return The current state of execution of an object in an NDR operation
     * 
     * @see FilterState
     */
    public FilterState state();
}
