/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access.filters;

/**
 * Indicates a particular state within an NDR object operation's life cycle.
 * 
 * @author birkland
 * @since 1.1
 */
public enum FilterState {
    /**
     * After <code>commit</code> is called, but before the data is written to
     * the underlying repository.
     * <p>
     * Filters that are to run upon commit, but do <em>not</em> need to be
     * part of an atomic commit operation should run here, especially if they
     * take a long time to complete.
     * </p>
     */
    BEFORE_COMMIT,

    /**
     * In the exclusive critical section immediately prior to data being written
     * to repository
     * <p>
     * In this state, filters are guaranteed to have exclusive access to the
     * repository. No other objects can be added, removed, or altered until the
     * NDR operation sicceeds. This is useful for filter operations that need to
     * implement referential integrity or other critical functions that make
     * assumptions about the repository state at time of commit.
     * </p>
     */
    COMMIT,

    /**
     * Immediately after the data has been committed to the underlying
     * repository
     */
    AFTER_COMMIT,

    /** Upon creation of a new object */
    NEW_OBJECT,

    /** Upon opening an existing object */
    OPEN,
    
    /** Upon purging of the object from the repository */
    PURGE,

    /**
     * At the termination of filter events without a <code>commit</code>.
     * <p>
     * Occurs under any of the following conditions:
     * <ul>
     * <li>object is open but never committed before garbage collection</li>
     * <li>An exception is thrown somewhere in processing</li>
     * </ul>
     * </p>
     */
    ROLLBACK
}
