/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.access.filters;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.locks.ReentrantLock;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.NDRAccessFactory;
import org.nsdl.repository.access.Finder;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.types.TypeFactory;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.content.ContentModelException;
import org.nsdl.repository.model.impl.MutableObjectWrapper;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.ConfigurationFactory;

/**
 * Wraps an {@link NDRAccess} instance and invokes {@link NDRAccessFilter
 * NDRAccessFilters} for each operation on it.
 * 
 * <p>
 * Throughout the life cycle of an object in the NDR that is accessed through
 * this class (including creation, updating, committing, deleting), this class
 * will invoke a set of {@link NDRAccessFilter NDRAccessFilters} at each
 * significant milestone, and persist state between these invocations.
 * </p>
 * 
 * <p>
 * For each NDR operation, this class will invoke every applicable filter before
 * passing along the operation to the wrapped <code>NDRAccess</code>
 * implementation to execute. Filters may change aspects of the operation (such
 * as modify the object being operated upon) or terminate it altogther by
 * throwing an exception.
 * </p>
 * 
 * <p>
 * This class may be instantiated in two ways. If instantiated through its
 * single argument constructor, it will establish a new, instantly-useable
 * FilteredNDRAccess. Filters may be added manually through
 * {@link #addFilter(NDRAccessFilter)} or through configuration using
 * {@link #updateConfig(Configuration)}. If instantiated with the no-argument
 * constructor, it <em>must</em> be instantiated with a valid configuration
 * (containing a delegate class) through {@link #updateConfig(Configuration)},
 * before it can be used.
 * <p>
 * 
 * <h2>Configuration Parameters</h2>
 * <p>
 * <dl>
 * <dt><code>repository.access.filter.delegate</code></dt>
 * <dd>Class of the delegate {@link NDRAccess} to which this class will wrap and
 * filter access.</dd>
 * <dt><code>repository.access.filter.def.*</code></dt>
 * <dd>{@link NDRAccessFilter} classes that will be applied to all operations.
 * Filters will be executed in alphabetical order of their property name.</dd>
 * </p>
 * 
 * <p>
 * <h3>Example</h3>
 * 
 * <pre>
 * repository.access.filter.delegate = org.nsdl.repository.impl.fedora.BasicFedoraNDRAccess
 * 
 * repository.access.filter.def.1 = org.nsdl.repository.access.filters.ContentModelFilter
 * repository.access.filter.def.2 = org.nsdl.repository.access.filters.HandleManagementFilter
 * </pre>
 * 
 * </p>
 * 
 * @author birkland
 * 
 * @since 1.1
 */
public class FilteredNDRAccess implements NDRAccess, Configuration.Listener {

    private static final String PROP_ACCESS_DELEGATE = "repository.access.filter.delegate";
    private static final String PROP_FILTER_BASE = "repository.access.filter.def.";

    private boolean configIsRead = false;

    private static final Logger logger = Logger
            .getLogger(FilteredNDRAccess.class);

    private NDRAccess delegate;

    private TypeFactory factory = new TypeFactory(this);

    private final FilterManager filters = new FilterManager();

    private final ReentrantLock m_commitLock = new ReentrantLock(true);

    /**
     * Create a new, uninitialized FilteredNDRAccess.
     * <p>
     * Cannot be used until initialized with
     * {@link #updateConfig(Configuration)}, where it will expect a delegate
     * NDRAccess class to be specified.
     * </p>
     */
    public FilteredNDRAccess() {
    }

    /**
     * Create a new FilteredNDRAccess over the supplied delegate implementation.
     * <p>
     * Initialize with the given NDRAccess implementation delegate, at which
     * point FilteredNDRAccess will be ready to use. Filters may be added
     * manually through {@link #addFilter(NDRAccessFilter)}, or through
     * configuration properties in {@link #updateConfig(Configuration)};
     * 
     * </p>
     * 
     * @param delegate
     */
    public FilteredNDRAccess(NDRAccess delegate) {
        this.delegate = delegate;
    }

    /** {@inheritDoc} */
    public void deleteObject(NDRObjectInfo object) throws ExistenceException,
            ContentModelException {
        NDRObject toBeDeleted = new FilteredObjectWrapper(delegate
                .getObject(object));
        toBeDeleted.setState(NDRObject.State.Deleted);
        toBeDeleted.commit();
    }

    /** {@inheritDoc} */
    public NDRObject getObject(NDRObjectInfo identity)
            throws ExistenceException {
        FilteredObjectWrapper o = new FilteredObjectWrapper(delegate
                .getObject(identity));
        o.runFilters(FilterState.OPEN);
        return o;
    }

    /** {@inheritDoc} */
    public NDRObject getObject(String identity) throws ExistenceException {
        FilteredObjectWrapper o = new FilteredObjectWrapper(delegate
                .getObject(identity));
        o.runFilters(FilterState.OPEN);
        return o;
    }

    /** {@inheritDoc} */
    public Finder getObjectFinder() {
        return delegate.getObjectFinder();
    }

    /** {@inheritDoc} */
    public <T extends NDRObject> T newObject(Type<T> type) {
        FilteredObjectWrapper o = new FilteredObjectWrapper(delegate
                .newObject(type));
        o.runFilters(FilterState.NEW_OBJECT);
        return factory.instantiate(o, type);
    }

    public <T extends NDRObject> T getObject(NDRObjectInfo object, Type<T> type) {
        FilteredObjectWrapper o = new FilteredObjectWrapper(delegate.getObject(
                object, type));
        o.runFilters(FilterState.OPEN);
        return factory.instantiate(o, type);
    }

    public <T extends NDRObject> T getObject(String object, Type<T> type) {
        FilteredObjectWrapper o = new FilteredObjectWrapper(delegate.getObject(
                object, type));
        o.runFilters(FilterState.OPEN);
        return factory.instantiate(o, type);
    }

    /** {@inheritDoc} */
    public void purgeObject(NDRObjectInfo identity) throws ExistenceException,
            ContentModelException {
        FilteredObjectWrapper o = new FilteredObjectWrapper(delegate
                .getObject(identity));
        o.runFilters(FilterState.PURGE);
        delegate.purgeObject(o);
    }

    public boolean isReadOnly() {
        return delegate.isReadOnly();
    }

    /**
     * Add a filter to be invoked at each operation hence.
     * 
     * @param filter
     *            The filter to be applied
     */
    public void addFilter(NDRAccessFilter filter) {
        filters.addFilter(filter);
    }

    public void updateConfig(Configuration config) {

        /*
         * NB: We are not subscribing to the configuration, since we can't
         * handle dynamic changes yet. To do so: config.addListener(this);
         */

        if (configIsRead) {
            logger.warn("Changing a running configuration is not supported.  "
                    + "Ignoring");
            return;
        } else {
            logger.info("config: initializing delegate");
            delegate = NDRAccessFactory.getAccess(config
                    .getProperty(PROP_ACCESS_DELEGATE), config);

            initializeFilters(config);
            configIsRead = true;
        }
    }

    private void initializeFilters(Configuration config) {

        ConfigurationFactory<NDRAccessFilter> factory = ConfigurationFactory
                .getInstance(NDRAccessFilter.class);

        for (String filterClass : config.getProperties(PROP_FILTER_BASE)) {
            logger.info("initializing filter: " + filterClass);
            NDRAccessFilter filter = factory.buildFromClassName(filterClass,
                    config);
            filter.init(this);
            filters.addFilter(filter);
        }
    }

    private static class FilterContextImpl implements FilterContext {
        private Map<String, Object> attributes = new HashMap<String, Object>();
        private FilterState sequence;

        public FilterContextImpl() {
        }

        public Object getAttribute(String key) {
            return attributes.get(key);
        }

        public void setAttribute(String key, Object value) {
            this.attributes.put(key, value);
        }

        public void setState(FilterState currentSequence) {
            this.sequence = currentSequence;
        }

        public FilterState state() {
            return sequence;
        }
    }

    /** Performs filter operations upon commit() */
    class FilteredObjectWrapper extends MutableObjectWrapper {
        private final NDRObject objectDelegate;
        public final FilterContextImpl context = new FilterContextImpl();

        public FilteredObjectWrapper(NDRObject delegate) {
            super(delegate);
            this.objectDelegate = delegate;
        }

        @Override
        /*
         * note: we arenot performing a ROLLBACK if anything goes wrong her. A
         * thrown exception can result in two actions: (a) calling code does
         * nothing, object passes out of scope, and is ROLLBACK upon
         * finalization or (b) calling code can do something smart.
         */
        public void commit() {
            runFilters(FilterState.BEFORE_COMMIT);

            m_commitLock.lock();
            try {
                runFilters(FilterState.COMMIT);
                objectDelegate.commit();
                this.isModified = false;
            } finally {
                m_commitLock.unlock();
            }

            runFilters(FilterState.AFTER_COMMIT);

        }

        NDRObject getDelegate() {
            return objectDelegate;
        }

        /**
         * If this object has uncommitted changes at GC, notify rollback filters
         */
        @Override
        protected void finalize() throws Throwable {
            try {
                if (isModified) {
                    runFilters(FilterState.ROLLBACK);
                }
            } finally {
                super.finalize();
            }
        }

        private void runFilters(FilterState state) {
            logger.trace("Running filters for state " + state);
            this.context.setState(state);
            filters.invoke(this, context);
            logger.trace("Done running filters for state " + state);
        }
    }

    private class FilterManager {

        private List<NDRAccessFilter> filters = new ArrayList<NDRAccessFilter>();

        public void addFilter(NDRAccessFilter filter) {
            filters.add(filter);
        }

        public void invoke(NDRObject o, FilterContext context) {
            for (NDRAccessFilter filter : filters) {
                filter.doFilter(o, context);
            }
        }
    }
}
