/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.access.filters;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.ConfigurationFactory;

import static org.nsdl.repository.model.content.CommonModel.Components.HAS_HANDLE;

/**
 * Creates, assigns, and manages CNRI handles for NDR objects
 * <p>
 * Will register and assign a handle for all newly created objects. If any new
 * objects are not committed, these handles will eventually be un-registered and
 * destroyed. Likewise for purged objects, their handle will be destroyed from
 * the registry upon purge.
 * </p>
 * 
 * <h2>Configuration Parameters</h2>
 * 
 * <p>
 * <dl>
 * <dt>
 * <code>repository.access.filter.HandleManagementFilter.HandleFactory</code></dt>
 * </dd> Class name of the {@link HandleFactory} to use to generate/manage
 * handles. ex. <code>org.nsdl.repository.handle.CNRIHandleFactory</code>.
 * </dl>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class HandleManagementFilter implements NDRAccessFilter,
        Configuration.Listener {

    private static final Logger logger = Logger
            .getLogger(HandleManagementFilter.class);

    private static final String NEW_HANDLE = HandleManagementFilter.class
            .toString()
            + ":newHandle";
    private static final String HANDLE_FACTORY_CLASS = "repository.access.filter.HandleManagementFilter.HandleFactory";

    private HandleFactory m_handleFactory;

    /**
     * Register and assign a CNRI handle to the given object if it is new.
     * 
     * @param object
     *            Object to assign a handle to (if new).
     * @param context
     *            The context of the current operation.
     */
    public void doFilter(NDRObject object, FilterContext context) {
        switch (context.state()) {
        case NEW_OBJECT: /* Assign a new handle */
            logger.trace("Getting handle for new object...");

            assignHandle(object);

            logger.info("Assigned handle " + object.handle());

            context.setAttribute(NEW_HANDLE, object.handle());
            break;
        case ROLLBACK: /* Delete a handle iff uncommitted */

            String newHandle = (String) context.getAttribute(NEW_HANDLE);

            if (newHandle != null) {
                logger.trace("Rolling back handle " + newHandle);
                dropHandle(newHandle);
                context.setAttribute(NEW_HANDLE, null);
            }
            break;
        case PURGE: /* Delete handle */
            logger.trace("Dropping handle due to purge");

            dropHandle(object.handle());
            break;
        case AFTER_COMMIT: /* Mark new handles as committed */

            if (context.getAttribute(NEW_HANDLE) != null) {
                logger.trace("Marking handle as committed");
            }

            context.setAttribute(NEW_HANDLE, null);
            break;
        default:
            logger.trace("Nothing to do");
            return;
        }
        logger.trace("done");
    }

    private void assignHandle(NDRObject o) {
        try {
            o.setHandle(m_handleFactory.getHandle(o));

            /* Make sure the hasHandle property is set correctly */
            if (o.getProperty(HAS_HANDLE) == null) {
                o.addProperty(HAS_HANDLE, o.handle());
            } else if (o.handle() != null
                    && !o.handle().equals(o.getProperty(HAS_HANDLE).value())) {
                o.getProperty(HAS_HANDLE).setValue(o.handle());
            }
        } catch (HandleException e) {
            throw new RepositoryException("Error assigning handle: "
                    + e.getMessage(), e);
        }
    }

    private void dropHandle(String handle) {
        try {
            m_handleFactory.deleteHandle(handle);
        } catch (HandleException e) {
            /* We can't really stop anything from proceeding, so just log */
            logger.error("Could not drop handle " + handle, e);
        }
    }

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        /* not needed */
    }

    public void updateConfig(Configuration config) {
        m_handleFactory = ConfigurationFactory.getInstance(HandleFactory.class)
                .buildFromProperty(HANDLE_FACTORY_CLASS, config);
    }

    protected void setHandleFactory(HandleFactory i) {
        this.m_handleFactory = i;
    }

    /**
     * Simple interface to handle services
     * <p>
     * Should probably do something a bit more formal/configurable, but for now,
     * the primary motivation for this interface is for testability. If we
     * actually wanted to *do* anything with handles in the future, this would
     * need to be expanded or abandoned
     * </p>
     */
    public interface HandleFactory {

        /** Get a new handle with some handle values populated */
        public String getHandle(NDRObject o) throws HandleException;

        /** Delete an existing handle */
        public void deleteHandle(String handle) throws HandleException;
    }

    public static class HandleException extends Exception {

        private static final long serialVersionUID = 1L;

        public HandleException(String message, Throwable t) {
            super(message, t);
        }
    }
}
