/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.access.filters;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.content.OAIModel;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.util.NDRObjectTemplate;

import static org.nsdl.repository.model.content.OAIModel.Set.Components.VISIBILITY;
import static org.nsdl.repository.model.content.MetadataModel.Components.METADATA_PROVIDED_BY;

/**
 * If the visibility of an OAI set changes, this filter touches all its members.
 * <p>
 * The {@link org.nsdl.repository.model.content.OAIModel OAI model} defines a
 * property that determines the
 * {@link org.nsdl.repository.model.content.OAIModel.Common#VISIBILITY
 * visibility} or a given OAI set or item. Some tools that implement OAI
 * services are not able to deduce that a change to set visibility should
 * propagate to all the set's items, and are only aware of changes to the items
 * themselves. Thus, this filter will iterate through a set's members if the
 * visibility is changed, modifying each one (but not actually making any
 * changes)
 * </p>
 * 
 * @author birkland
 * @author Josh Endries (jce54@cornell.edu)
 * @since 1.1
 */
public class OAIVisibilityFilter implements NDRAccessFilter {
	private static final Logger logger = Logger.getLogger(OAIVisibilityFilter.class);
	private static final String VISIBILITY_KEY = OAIVisibilityFilter.class.getName() + ".initialState";
	private static final String DEFAULT_VISIBILITY = OAIModel.Visibility.PUBLIC.toString();
	private NDRAccess ndr;

	public void doFilter(NDRObject object, FilterContext context) {

		if (!Type.MetadataProvider.equals(object.type())) {
			/* Only care about MetadataProviders */
			logger.trace(object.handle()+": Object is not a MetadataProvider; skipping filter.");
			return;
		} else {
			logger.debug(object.handle()+": Object is a MetadataProvider; processing filter.");
		}

		switch (context.state()) {
			case OPEN:
				/*
				 * This needs to COPY the current property, not reference it,
				 * otherwise it will never appear to have changed.
				 */
				Property pReal = object.getProperty(VISIBILITY);
				Property pCopy = null;
				if (pReal != null) {
					logger.debug(object.handle()+": Caught OPEN context. Setting context visibility attribute to " + object.getProperty(VISIBILITY));
					pCopy = new Property(pReal.subject(), pReal.predicate(), pReal.value());
				} else {
					logger.info(object.handle()+": Caught OPEN context but object contains no visibility property.");
				}
				context.setAttribute(VISIBILITY_KEY, pCopy);
				break;
			case AFTER_COMMIT:
				logger.trace("Caught AFTER COMMIT context. Checking if updating children is necessary...");
				updateChildrenIfNecessary(object, context);
				break;
			default:
				return;
		}
	}

	public void init(NDRAccess ndr) {
		this.ndr = ndr;
	}

	private void updateChildrenIfNecessary(NDRObject o, FilterContext context) {
		String currentVisibility = null;
		String previousVisibility = null;

		Property currentVisibilityProperty = o.getProperty(VISIBILITY);
		Property previousVisibilityProperty = (Property) context.getAttribute(VISIBILITY_KEY);
		if (currentVisibilityProperty != null) {
			currentVisibility = currentVisibilityProperty.value();
			logger.trace("Current visibility is " + currentVisibilityProperty);
		} else {
			logger.trace("Current visibility is null");
		}

		if (previousVisibilityProperty != null) {
			previousVisibility = previousVisibilityProperty.value();
			logger.trace("Previous visibility is " + previousVisibilityProperty);
		} else {
			logger.trace("Previous visibility is null");
		}

		if (currentVisibility == null && previousVisibility == null) {
			/* Nothing to do (probably most common case */
			logger.trace("Both previous and current visibility are null. Nothing to do.");
			return;
		}

		if (previousVisibility == null && currentVisibility.equals(DEFAULT_VISIBILITY)) {
			logger.trace("Changing implicit default visibility to explicit. No changes needed.");
			return;
		} else if (currentVisibility == null && previousVisibility.equals(DEFAULT_VISIBILITY)) {
			logger.trace("Changing explicit defaut visibility to implicit. No changes needed.");
		}

		if (currentVisibility == null && previousVisibility.equals(OAIModel.Visibility.PUBLIC.toString()))
			;

		if ((previousVisibility != null && !previousVisibility.equals(currentVisibility)) || !currentVisibility.equals(previousVisibility)) {
			logger.trace("Visibilities are different. Updating...");
			MetadataUpdater updater = new MetadataUpdater(o, currentVisibility);
			new Thread(updater, this.getClass().getSimpleName() + ":updater-" + o.handle()).start();
			logger.trace("Exiting filter control");
		} else {
			logger.trace("Previous and current visibilities are the same. Nothing to do.");
		}

	}

	private class MetadataUpdater implements Runnable {
		private final NDRObject provider;
		private final String visibility;

		public MetadataUpdater(NDRObject parent, String currentVisibility) {
			this.provider = parent;
			this.visibility = currentVisibility;
		}

		public void run() {
			NDRObjectTemplate template = new NDRObjectTemplate(Type.Metadata);
			template.setState(State.Active);
			template.addRelationship(METADATA_PROVIDED_BY, provider);

			logger.info("OAI visibility of parent " + provider.handle() + " changed to " + visibility);

			Results<NDRObjectInfo> matches = ndr.getObjectFinder().matchObject(template);

			try {
				for (NDRObjectInfo match : matches) {
					String handle = null;
					try {
						handle = match.handle();
						NDRObject metadata = ndr.getObject(match);
						logger.info("Touching member " + handle);
						metadata.commit();
					} catch (Throwable t) {
						logger.error("Error touching member " + handle + " from " + provider.handle(), t);
					}
				}
			} finally {
				matches.close();
			}
		}
	}
}
