/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.access.filters;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.content.ModelUtil;

public class ReadOnlyPropertyFilter implements NDRAccessFilter {

    private static final Logger log = Logger.getLogger(ReadOnlyPropertyFilter.class);

    private static final String READ_ONLY_PROPS = ReadOnlyPropertyFilter.class
            .getName()
            + "#readOnly";

    public void doFilter(NDRObject object, FilterContext context) {
        switch (context.state()) {
        case NEW_OBJECT:
        case OPEN:
            context.setAttribute(READ_ONLY_PROPS, getReadOnlyProps(object));
            break;
        case BEFORE_COMMIT:
            compareReadOnlyProps(object, context.getAttribute(READ_ONLY_PROPS));
            break;
        default:
        }
    }

    @SuppressWarnings("unchecked")
    private void compareReadOnlyProps(NDRObject o, Object formerReadOnlyProps) {
        Map<Property.Name, String> oldProps = (Map) formerReadOnlyProps;
        Map<Property.Name, String> newProps = getReadOnlyProps(o);

        /* FIXME:  oldProps is null whenever ndr.deleteObject() is called */
        if (oldProps == null) {
            log.warn("filter never ran OPEN state, not comparing old properties");
            return;
        }
        for (Entry<Property.Name, String> entry : newProps.entrySet()) {
            if (oldProps.containsKey(entry.getKey())
                    && !oldProps.get(entry.getKey()).equals(entry.getValue())) {
                throw new AccessException("Cannot modify read-only property "
                        + entry.getKey());
            }
        }
    }

    private Map<Property.Name, String> getReadOnlyProps(NDRObject o) {
        HashMap<Property.Name, String> props = new HashMap<Property.Name, String>();
        for (Property p : o.getProperties()) {
            if (ModelUtil.isReadOnlyProperty(p.predicate())) {
                props.put(p.predicate(), p.value());
            }
        }
        return props;
    }

    public void init(NDRAccess ndr) {
        /* Do Nothing */
    }

}
