/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.access.filters;

import java.net.URL;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.types.Resource;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.Configuration.Listener;

/**
 * Manages <code>identifier</code> property and <code>content</code> datastream
 * for Resources.
 * <p>
 * Resources in the NDR are guaranteed to have an identifier and a content
 * datastream at a minimum. There is a relationship between the two: If a URL
 * defined, then the content datastream contains a redirect to that URL. This
 * contains implications for the request processing:
 * <ul>
 * <li>If an identifier is given but a datastream is not, then there's an
 * implicit operation on the corresponding datastream. For example, replacing
 * the identifier would imply replacing the content datastream with the redirect
 * to the new identifier.</li>
 * <li>If a datastream is given but an identifier is not, then the identifier is
 * taken to be the <tt>get</tt> command url pointing to the content datastream</li>
 * <li>If the identifier is not a URL, and there is no content datastream, then
 * an empty content datastream shall be created</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.2
 */
public class ResourceContentFilter implements NDRAccessFilter, Listener {

    private static final Logger logger = Logger
            .getLogger(ResourceContentFilter.class);

    private static final String ATTRIBUTE_URL = ResourceContentFilter.class
            .getName()
            + ".url";

    private final URIFactory m_uriFactory = new URIFactory();

    public void doFilter(NDRObject object, FilterContext context) {
        if (Type.Resource.equals(object.type())) {
            switch (context.state()) {
            case OPEN:
                Property url = object.getProperty(Resource.IDENTIFIER_URL);
                if (url != null) {
                    context.setAttribute(ATTRIBUTE_URL, url.value());
                }
                break;
            case BEFORE_COMMIT:
                syncContent(object, context);
                break;
            default:
                return;
            }
        }
    }

    public void init(NDRAccess ndr) {
        /* Not needed */
    }

    public void updateConfig(Configuration config) {
        m_uriFactory.updateConfig(config);
    }

    private void syncContent(NDRObject object, FilterContext context) {
        Property url = object.getProperty(Resource.IDENTIFIER_URL);
        Datastream content = object.getDatastream(Resource.CONTENT);
        String previous_url = (String) context.getAttribute(ATTRIBUTE_URL);

        if (url == null) {
            /* No identifier. Add one from the content stream */
            if (content != null) {
                /* Content stream, but no identifier URI */
                object.addProperty(Resource.IDENTIFIER_URL, m_uriFactory
                        .getURL(object, content));
            } else {
                /* If identifier and content are both null, that's bad! */
                Property other = object.getProperty(Resource.IDENTIFIER_OTHER);
                if (other == null) {
                    throw new RepositoryException(
                            "Resource does not have an identifier or content stream!");
                } else {
                    logger.info("Encountered useless resource without a URL!");
                }
            }
        } else {
            if (!url.value().equals(previous_url)) {
                if (object.getDatastream(Resource.CONTENT) != null) {
                    object.removeDatastream(Resource.CONTENT);
                }
                Datastream new_content = object.addDatastream(Resource.CONTENT);
                try {
                    new_content.setURL(new URL(url.value()));
                } catch (Exception e) {
                    throw new AccessException("Could not interpret url '"
                            + url.value() + "'", e);
                }
            }
        }
    }

    private class URIFactory implements Listener {
        String m_baseURL;

        public void updateConfig(Configuration config) {
            m_baseURL = config.getProperty("param.repository.baseURL");
            if (!m_baseURL.endsWith("/")) {
                m_baseURL += "/";
            }
        }

        public String getURL(NDRObject object, Datastream d) {
            if (d.getURL() == null || d.getURL().toString().contains(m_baseURL)) {
                return m_baseURL + "get/" + object.handle() + "/"
                        + Resource.CONTENT.getID();
            } else {
                return d.getURL().toString();
            }
        }
    }
}
