/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.access.filters;

import java.net.MalformedURLException;
import java.net.URL;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.types.Type;
import org.nsdl.util.uri.NormalizedURL;

import static org.nsdl.repository.access.filters.FilterState.BEFORE_COMMIT;
import static org.nsdl.repository.model.content.ResourceModel.Components.CONTENT;
import static org.nsdl.repository.model.content.ResourceModel.Components.IDENTIFIER_URL;

/**
 * Normalizes resource identifiers and content datastreams before commit.
 * 
 * @author birkland
 * @since 1.3
 */
public class ResourceNormalizationFilter implements NDRAccessFilter {

    private static final Logger logger = Logger
            .getLogger(ResourceNormalizationFilter.class);

    public void doFilter(NDRObject object, FilterContext context) {
        if (Type.Resource.equals(object.type())
                && context.state().equals(BEFORE_COMMIT)) {
            fixURL(object);
        }
    }

    private static void fixURL(NDRObject object) {
        Datastream content = object.getDatastream(CONTENT);
        Property identiferProperty = object.getProperty(IDENTIFIER_URL);

        if (content != null && !content.isCanonical()) {
            try {
                String contentURL = content.getURL().toExternalForm();
                if (needsNormalization(contentURL)) {
                    content.setURL(new URL(normalize(contentURL)));
                    logger.debug("Normalized content datastream from "
                            + contentURL + " to " + normalize(contentURL));
                }
            } catch (MalformedURLException e) {
                throw new RepositoryException("malformed URL "
                        + content.getURL(), e);
            }
        }

        if (identiferProperty != null
                && needsNormalization(identiferProperty.value())) {
            identiferProperty.setValue(normalize(identiferProperty.value()));
        }
    }

    private static boolean needsNormalization(String url) {
        String normalized = normalize(url);
        return !normalized.equals(url);
    }

    private static String normalize(String input) {
        try {
            NormalizedURL normalized = new NormalizedURL(input);
            normalized.normalize();
            return normalized.toString();
        } catch (Exception e) {
            throw new RepositoryException("Could not normalize URL " + input, e);
        }
    }

    public void init(NDRAccess ndr) {
        /* Nothing to do */
    }
}
