/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access.filters;

import java.io.ByteArrayInputStream;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.content.MetadataModel;
import org.nsdl.repository.model.types.Metadata;
import org.nsdl.util.DateUtility;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import static org.nsdl.repository.model.content.MetadataModel.Components.DC;

/**
 * Updates the DC datastream every time Metadata is changed
 * <p>
 * Some Metadata disseminators are bound to the DC datastream. In other words,
 * the update date of the disseminator changes when the DC datastream changes,
 * <em>not</em> when any other datastreams change. This filter assures that the
 * DC field is always modifed whenever Metadata is updated. <b>Note: This is not
 * true in Fedora 3.0, at which point this filter becomes unnecessary</b>
 * </p>
 * <p>
 * <b>Important!</b> Currently, this filter will <em>replace</em> the DC
 * contents upon modify. Clients should not be putting anything in there.
 * </p>
 * 
 * @author birkland
 * @author cornwell
 * 
 * @since 1.1
 */
public class TouchDCFilter implements NDRAccessFilter {

    private static final Logger logger = Logger.getLogger(TouchDCFilter.class);

    private static final String DC_URI = "http://purl.org/dc/elements/1.1/";

    /**
     * Check and update the DC field if necessary
     * <p>
     * Only updates the DC field if the object is of type Metadata. If DC is not
     * defined, it will add it, otherwise it will seek out, and replace (or add,
     * if not defined) a "date" element with ne containing the current date
     * </p>
     * 
     * @param object
     *            Mutable object in which to check and modify DC stream
     * @param context
     *            Current filter context
     */
    public void doFilter(NDRObject object, FilterContext context) {

        if (!Type.Metadata.equals(object.type())
                || !context.state().equals(FilterState.BEFORE_COMMIT)) {
            /* If this is not a metadata object, leave now */
            return;
        }

        reviseDCDatastream(object);
    }

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        /* No need for NDR access */
    }

    private static void reviseDCDatastream(NDRObject o) {
        Datastream dc = o.getDatastream(MetadataModel.Components.DC);

        if (dc == null) {
            createNewDCDatastream(o);
        } else {
            updateDCDatastream(o);
        }
    }

    /** Create entirely new DC datastream content */
    private static void createNewDCDatastream(NDRObject o) {
        String description = "<oai_dc:dc xmlns:oai_dc=\"http://www.openarchives.org/OAI/2.0/oai_dc/\" "
                + "xmlns:dc=\"http://purl.org/dc/elements/1.1/\">\n"
                + " <dc:title>NSDL Repository Metadata</dc:title>\n"
                + " <dc:date>"
                + DateUtility.getCurrTimeInDashedISO8601Plus()
                + "</dc:date>\n"
                + " <dc:identifier>"
                + o.handle()
                + "</dc:identifier>\n" + "</oai_dc:dc>";

        Datastream dc = o.getDatastream(Metadata.DC);
        if (dc == null) {
            dc = o.addDatastream(Metadata.DC);
        }
        dc.setMimeType("text/xml");
        dc.setContent(new ByteArrayInputStream(description.getBytes()));
    }

    /** Update the DC datastream: if there's a date, replace. If not, add */
    private static void updateDCDatastream(NDRObject o) {

        Document dcDoc = null;

        Datastream dc = o.getDatastream(DC);

        try {
            if (dc == null) {
                createNewDCDatastream(o);
                return;
            }
            dcDoc = NDRXMLutil.getDocument(dc.getContent(), true, false);
        } catch (Exception e) {
            logger.warn("Could not read DC datastream for update.  "
                    + "Replacing instead", e);
            createNewDCDatastream(o);
        }
        Node existingDate = NDRXMLutil.getNodeByName(dcDoc, DC_URI, "date");
        Node newDate = dcDoc.createElementNS(DC_URI, "dc:date");
        newDate.setTextContent(DateUtility.getCurrTimeInDashedISO8601Plus());

        if (existingDate == null) {
            dcDoc.getDocumentElement().appendChild(newDate);
        } else {
            dcDoc.getDocumentElement().replaceChild(newDate, existingDate);
        }

        try {
            dc.setContent(new ByteArrayInputStream(NDRXMLutil.serializeDOMdoc(
                    dcDoc).getBytes()));
        } catch (Exception e) {
            throw new RuntimeException("Could not set DC content");
        }
        dc.setMimeType("text/xml");
    }
}
