/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.api;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;
import java.util.jar.Attributes;
import java.util.jar.Manifest;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;

import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.apache.struts.action.ActionServlet;
import org.apache.struts.action.PlugIn;
import org.apache.struts.config.ModuleConfig;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.NDRAccessFactory;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.impl.fedora.BasicFedoraAccess;
import org.nsdl.repository.impl.fedora.FedoraAccess;
import org.nsdl.repository.impl.fedora.FedoraNDRAccess;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.ConfigurationFactory;
import org.nsdl.repository.util.StreamUtil;

import fedora.client.HttpInputStream;

import static fedora.common.Constants.FOXML1_1;

/**
 * Initializes access to a Fedora repository.
 * 
 * @author birkland
 * 
 */
public class FedoraPlugIn implements PlugIn {

    private static boolean ACTIVE = true;

    private final Logger logger = Logger.getLogger(FedoraPlugIn.class);

    /** Attribute name for a FedoraAccess object stored in the servlet context */

    /**
     * System property containing the location of the configuration
     * file/directory
     */
    public static final String PROP_CONFIG_LOCATION = "repository.config";

    public static final String LOG4J_LOCATION = "repository.config.log4j";

    public static final String FEDORA_ACCESS = FedoraPlugIn.class.getName()
            + ".access";

    /** TODO: Find a good home for this */
    public static final String CONFIG = FedoraPlugIn.class.getName()
            + ".config";

    private String m_propertyDirectory;

    private static final Pattern PID_PATTERN = Pattern
            .compile("PID\\s*=\\s*\"(.+?)\"");

    public void destroy() {
        /* no cleanup necessary */
    }

    /**
     * Create an instance of a Fedora APi wrapper, test that it can establish a
     * connection to Fedora, and place that wrapper as an accessible attribute
     * of the servlet context.
     * 
     * The wrapper may be retrieved from the context by the folowing: wrapper =
     * (FedoraAPIWrapper) servletContext.getAttribute(FedoraPlugIn.ACCESS);
     */
    public synchronized void init(ActionServlet servlet, ModuleConfig modConfig)
            throws ServletException {

        Configuration config = new Configuration();
        
        /*
         * Load build properties that are stored in the manifest file.
         */
        try {
           String[] keys = { "Build-Host", "Build-Timestamp", "Build-User", "Version" };
      	  InputStream is = servlet.getServletContext().getResourceAsStream("/META-INF/MANIFEST.MF");
      	  Manifest mf = new Manifest(is);
      	  Attributes ma = mf.getMainAttributes();
      	  for (String k: keys) config.putProperty(k, ma.getValue(k));
        } catch (IOException e1) {
      	  e1.printStackTrace();
        }

        try {
            Properties defaultProps = new Properties();
            defaultProps.load(getClass().getResourceAsStream(
                    "/resources/default.properties"));
            config.putProperties(defaultProps);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

        ServletContext context = servlet.getServletContext();

        /* Initialize log4j if we have an external location */
        String externaLog4J = System.getProperty(LOG4J_LOCATION);
        if (externaLog4J == null) {
            externaLog4J = context.getInitParameter(LOG4J_LOCATION);
        }

        if (externaLog4J != null) {
            logger.info("Using external log4j location: " + externaLog4J);
            PropertyConfigurator.configure(externaLog4J);
        } else {
            logger.info("Using default log4j config");
        }

        /* If we've already initialized, don't do it again */
        if (context.getAttribute(FEDORA_ACCESS) != null) {
            logger.warn("Refusing to re-init");
            return;
        }

        if (m_propertyDirectory == null) {
            m_propertyDirectory = System.getProperty(PROP_CONFIG_LOCATION,
                    context.getInitParameter(PROP_CONFIG_LOCATION));
        } else {
            logger.info("Config property directory already set to "
                    + m_propertyDirectory);
        }

        /*
         * If we have an external properties directory, load these properties.
         * They will overwrite the defaults
         */
        if (m_propertyDirectory != null) {
            logger.info("loading configuration properties from "
                    + m_propertyDirectory);

            File configFile = new File(m_propertyDirectory);
            config.putProperties(configFile);
        } else {
            logger.warn("No configuration properties set.  Using defaults");
        }

        /* Bootstrap the NDR */
        Bootstrap.exec(config);

        context.setAttribute(CONFIG, config);

        NDRAccess ndr = NDRAccessFactory.getAccess(config);
        FedoraAccess fedora = ConfigurationFactory.getInstance(
                FedoraAccess.class).buildFromClass(BasicFedoraAccess.class,
                config);
        /*
         * For compatibility with components that still use direct Fedora access
         */
        context.setAttribute(FEDORA_ACCESS, fedora);
        context.setAttribute(Constants.NDR_ACCESS, ndr);
    }

    /**
     * Configures the FedoraPlugIn to use a specific properties file or
     * directory.
     * 
     * 
     * @param propertiesFile
     *            path to the properties file or directory
     */
    public void setProperties(String propertiesFile) {
        m_propertyDirectory = propertiesFile;
    }

    public static boolean isActive() {
        return ACTIVE;
    }

    private static class Bootstrap extends Thread {

        private Logger logger = Logger.getLogger(Bootstrap.class);
        private final Configuration m_config = new Configuration();
        private final int MAX_TRIES = 7;

        public static void exec(Configuration config) {
            Bootstrap boot = new Bootstrap(config);
            boot.start();
        }

        private Bootstrap(Configuration config) {
            super("NDR-boot");
            for (Map.Entry<String, String> prop : config.getPropertyMap("")
                    .entrySet()) {
                m_config.putProperty(prop.getKey(), prop.getValue());
            }
        }

        @Override
        public void run() {
            ACTIVE = false;
            String rootPID = m_config
                    .getProperty(FedoraNDRAccess.PROP_ROOT_PID);

            if (rootPID == null) {
                throw new RepositoryException("PID of root object is not set");
            }

            m_config.putProperty("param.repository.impl.fedora.timeout", "5");

            FedoraAccess fedora = null;

            fedora = ConfigurationFactory.getInstance(FedoraAccess.class)
                    .buildFromClass(BasicFedoraAccess.class, m_config);

            File bootstrapObjectsDir = new File(this.getClass().getResource(
                    "/resources/objects").getFile());

            for (File foxmlFile : bootstrapObjectsDir
                    .listFiles(new FoxmlFilter())) {

                String foxmlString = "";
                try {
                    foxmlString = StreamUtil
                            .streamToString(new FileInputStream(foxmlFile));
                } catch (FileNotFoundException e) {
                    /* never happen */
                    throw new RuntimeException(e);
                }
                Matcher match = PID_PATTERN.matcher(foxmlString);

                if (!match.find()) {
                    logger.warn("Could not find a PID in " + foxmlFile
                            + ".  Skipping");
                    continue;
                }

                String pid = match.group(1);
                Exception lastException = null;

                for (int tries = 0; tries < MAX_TRIES; tries++) {

                    HttpInputStream generalResponse = null;
                    HttpInputStream objectResponse = null;
                    try {
                        objectResponse = fedora.getAPI().getClient().get(
                                "/objects/" + pid, false);
                        generalResponse = fedora.getAPI().getClient().get(
                                "/objects", false);

                        if (objectResponse.getStatusCode() == 200) {
                            logger.info("bootstrap object " + pid
                                    + " is present");
                            break;
                        } else {
                            if (objectResponse.getStatusCode() == 404
                                    && generalResponse.getStatusCode() == 200) {
                                /* Object is simply not present. Ingest it */
                                logger
                                        .info("Ingesting bootstrap object "
                                                + pid);
                                try {
                                    fedora.getAPI().ingest(
                                            foxmlString.getBytes(),
                                            FOXML1_1.uri,
                                            "Object added as part of the "
                                                    + "bootstrap process");
                                } catch (Exception ingestException) {
                                    logger
                                            .warn("Could not ingest bootstrap object, "
                                                    + "trying again in 30 seconds");
                                    lastException = ingestException;
                                }
                                break;
                            } else if (tries + 1 == MAX_TRIES) {
                                throw new RepositoryException(
                                        "Could not finish bootstrap process "
                                                + "after several attemts.  "
                                                + "Not trying again",
                                        lastException);
                            } else {
                                logger.warn("Could not connect to Fedora  "
                                        + "Trying again in 30 seconds... ");
                                if (logger.isDebugEnabled()) {
                                    logger.debug("Recieved code "
                                            + objectResponse.getStatusCode()
                                            + " for " + objectResponse.getURL()
                                            + ", "
                                            + generalResponse.getStatusCode()
                                            + " for "
                                            + generalResponse.getURL());
                                }
                                try {
                                    Thread.sleep(30000);
                                } catch (InterruptedException i) {
                                    Thread.currentThread().interrupt();
                                }
                            }
                        }
                    } catch (Exception e) {
                        logger.warn("Could not connect to Fedora  "
                                + "Trying again in 30 seconds... ");
                        try {
                            Thread.sleep(30000);
                        } catch (InterruptedException i) {
                            Thread.currentThread().interrupt();
                        }
                    } finally {
                        try {
                            if (generalResponse != null) {
                                generalResponse.close();
                            }

                            if (objectResponse != null) {
                                objectResponse.close();
                            }
                        } catch (IOException e) {

                        }
                    }
                }
            }
            
            /* If we reach this point, we're OK */
            logger.info("Repository " + m_config.getProperty("Version") + " initialized OK");
            logger.info("Built by " + m_config.getProperty("Build-User") + " on " + m_config.getProperty("Build-Host") + " at " + m_config.getProperty("Build-Timestamp"));
            ACTIVE = true;
        }
    }

    private static class FoxmlFilter implements FilenameFilter {
        public boolean accept(File dir, String name) {
            return name.contains(".xml");
        }
    }
}
