/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.action;

import java.util.Collection;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.apache.struts.action.ActionErrors;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;
import org.apache.struts.util.MessageResources;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.actionForm.Data;
import org.nsdl.repository.api.auth.AgentAuthInfo;
import org.nsdl.repository.api.auth.AuthorizationException;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.api.request.Request.Commands;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.content.ContentModelException;
import org.nsdl.repository.model.content.ModelUtil;
import org.nsdl.repository.model.content.ResourceModel;
import org.nsdl.repository.model.content.Scope;
import org.nsdl.repository.model.types.Type;

/**
 * Struts action to add an object to the NDR
 * <p>
 * Generic action to add a Resource, Metadata, MetadataProvider, Agent, etc.
 * </p>
 * 
 * 
 */
public class Add extends GenericAction {
    private static final Logger logger = Logger.getLogger(Add.class);

    @SuppressWarnings("unchecked")
    public ActionForward execute(ActionMapping mapping, ActionForm form,
            HttpServletRequest request, HttpServletResponse response)
            throws Exception {

        /* Set up basics */
        initDataFormBean(form, request);

        /* Get message resources, errors */
        MessageResources resources = this.getResources(request);
        ActionErrors errors = new ActionErrors();

        Request addRequest = null;
        AgentAuthInfo auth = null;
        NDRAccess ndr = null;

        logger.trace("Executing Add request");
        /* Check auth, populate, and add */
        try {

            addRequest = getApiRequest(request);
            auth = getAuth(request);
            ndr = getNDR(request);
            checkAuth(auth, addRequest);

            NDRObject toAdd = null;

            /* For resources, see if one exists. If so, just open it */
            if (Type.Resource.equals(addRequest.getObjectInfo().type())) {
                logger.debug("Object to add is a resource.  "
                        + "Determining if ia already exists");
                toAdd = findExistingResource(addRequest, ndr);

                if (toAdd == null) {
                    logger.trace("resource does not exist");
                } else {
                    logger.trace("resource exists as " + toAdd.handle());
                }
            }

            if (toAdd == null) {
                toAdd = ndr.newObject((Type<NDRObject>) addRequest.getObjectInfo()
                        .type());
            }

            populate(toAdd, addRequest);

            toAdd.commit();

            ((Data) form).setHandle(toAdd.handle());

        } catch (AuthorizationException e) {
            ((Data) form).setCode(resources.getMessage("error.security.code"));
            errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(
                    "error.security.authorization", "Add :" + e.getMessage()));
            saveMessages(request, errors);

            logger.info("Auth error from " + auth.agent().handle() + ":"
                    + e.getMessage());
            logger.debug("Fail trace: \n",  e);
            response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
            return (mapping.findForward(ERROR_FORWARD));
        } catch (Exception e) {

            /* Cut down on logging pollution */
            if (e instanceof ContentModelException) {
                logger.info("Content model failed: " + e.getMessage());
                logger.trace("Failure trace: \n", e);
            } else {
                logger.warn("Error adding object", e);
            }
            ((Data) form).setCode(resources.getMessage("error.object.code"));
            errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(
                    "error.application.api", "Add: " + e.getMessage()));
            saveMessages(request, errors);

            response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
            return (mapping.findForward(ERROR_FORWARD));
        }

        logger.trace("Success");
        return (mapping.findForward(SUCCESS_FORWARD));
    }

    private NDRObject findExistingResource(Request r, NDRAccess ndr) {

        Collection<Property.Name> identifiers =
                ModelUtil.getPropertyNames(ResourceModel.Identifiers.class,
                        Scope.All);

        for (Property.Name name : identifiers) {
            logger.trace("looking for identifier " + name);
        }

        for (Property id : r.getProperties(Commands.ADD)) {
            if (identifiers.contains(id.predicate())) {

                NDRObjectInfo found =
                        ndr.getObjectFinder().findResource(id.predicate(),
                                id.value());

                if (found != null) {
                    return ndr.getObject(found);
                }
            } else {
                logger.debug("it's not an identifier..");
            }
        }

        return null;
    }

    /*
     * Populate the new object with stuff in the request. In reality, nobody in
     * their right mind would use any command other than 'add', so we just make
     * sure there are no other commands, and just do the adds.
     */
    private static void populate(NDRObject o, Request r) {

        /* Just for sanity */
        for (QName command : r.getCommands()) {
            if (!Commands.ADD.equals(command)) {
                throw new IllegalArgumentException(
                        "Add requests cannot tolerate "
                                + command.getLocalPart() + " commands.  Use "
                                + Commands.ADD.getLocalPart() + " instead");
            }
        }

        logger.trace("Registering property changes");
        o.addProperties(r.getProperties(Commands.ADD));

        /* Apply relationship changes */
        logger.trace("Registering relationship changes");
        o.addRelationships(r.getRelationships(Commands.ADD));

        /* Apply datastream changes */
        logger.trace("Registering datastream changes");
        for (Datastream d : r.getDatastreams(Commands.ADD)) {
            logger.trace("Request to add datastream " + d.getName().getID());
        }
        o.addDatastreams(r.getDatastreams(Commands.ADD));
    }
}
