/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.action;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.apache.struts.action.ActionErrors;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;
import org.apache.struts.util.MessageResources;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.actionForm.Data;
import org.nsdl.repository.api.auth.AgentAuthInfo;
import org.nsdl.repository.api.auth.AuthorizationException;
import org.nsdl.repository.api.request.MutableRequest;
import org.nsdl.repository.api.request.MutableRequestWrapper;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.content.ContentModelException;
import org.nsdl.repository.model.types.Type;

public class Delete extends GenericAction {
    private static final Logger logger =
            Logger.getLogger(Delete.class.getName());

    public ActionForward execute(ActionMapping mapping, ActionForm form,
            HttpServletRequest request, HttpServletResponse response)
            throws Exception {

        /* Set up basics */
        initDataFormBean(form, request);

        /* Get message resources, errors */
        MessageResources resources = this.getResources(request);
        ActionErrors errors = new ActionErrors();

        /* Do the delete! */
        try {
            /* Get necessary information from the request */
            Request req = getApiRequest(request);
            AgentAuthInfo auth = getAuth(request);
            NDRAccess ndr = getNDR(request);

            /* Check permissions */
            checkAuth(auth, req);

            doDelete(req, ndr);

            ((Data) form).setHandle(req.getObjectInfo().handle());

        } catch (AuthorizationException e) {
            AgentAuthInfo auth = getAuth(request);
            Request req = getApiRequest(request);
            logger.info("Error authorizing delete of "
                    + req.getObjectInfo().handle() + " by "
                    + auth.agent().handle() + ": " + e.getMessage());
            logger.debug("Fail trace: \n", e);

            ((Data) form).setCode(resources.getMessage("error.security.code"));
            errors.add(ActionMessages.GLOBAL_MESSAGE,
                    new ActionMessage("error.security.authorization",
                            "Delete: " + e.getMessage()));
            saveMessages(request, errors);

            response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
            return (mapping.findForward(ERROR_FORWARD));
        } catch (Exception e) {

            if (e instanceof ExistenceException) {
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
                return (mapping.findForward(ERROR_FORWARD));
            } else {
                response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
            }
            logger.warn("Error deleting object ", e);
            ((Data) form).setCode(resources.getMessage("error.object.code"));
            errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(
                    "error.application.api", "delete: " + e.getMessage(), e
                            .getMessage()));
            saveMessages(request, errors);

            return (mapping.findForward(ERROR_FORWARD));
        }

        logger.trace("Success");
        return mapping.findForward(SUCCESS_FORWARD);
    }

    private static void doDelete(Request r, NDRAccess ndr)
            throws ContentModelException {

        logger.debug("cascade: " + r.getParameter("cascade"));
        if (r.getParameter("cascade").size() > 0) { /*
                                                     * Cascade XXX: see note on
                                                     * cascadeDelete()
                                                     */
            if ("true".equals(r.getParameter("cascade").get(0)))
                cascadeDelete(r, ndr);
        }

        /* If there's anything we need to do before deleting */
        if (r.getCommands().size() > 0) {
            logger.trace("Applying pre-delete changes");
            NDRObject lastMinute = ndr.getObject(r.getObjectInfo());
            Modify.propagateChanges(r, lastMinute);
            lastMinute.commit();
        }

        logger.trace("Deleting object " + r.getObjectInfo().handle());

        NDRObject obj =  ndr.getObject(r.getObjectInfo());
        obj.setState(State.Deleted);
        obj.commit();
        
        //ndr.deleteObject(r.getObjectInfo());
    }

    /*
     * FIXME: This general functionality should probably not be here - should
     * ideally be exposed by NDRAccess. TODO: expose it.
     */
    private static void cascadeDelete(Request r, NDRAccess ndr)
            throws ContentModelException {
        Type<? extends NDRObject> type = ndr.getObject(r.getObjectInfo()).type();
        if (Type.MetadataProvider.equals(type)) {

            logger.info("Cascading delete to children of MetadataProvider "
                    + r.getObjectInfo().handle());
            /* Delete underlying metdata items */
            for (NDRObjectInfo mdMember : ListMembers.list(r.getObjectInfo(),
                    ndr)) {
                MutableRequest deleteRequest = new DeleteRequest(mdMember);

                logger.info("Deleting child " + mdMember.handle());
                doDelete(deleteRequest, ndr);
            }
        } else {
            throw new RepositoryException(
                    "'cascade' parameter currently only works "
                            + "with MetadataProviders, cascading deletes "
                            + "to its Metadata children. ");
        }
    }

    /* FIXME: We probably dont' want to do this using requests at this layer */
    private static class DeleteRequest extends MutableRequestWrapper {
        public DeleteRequest(NDRObjectInfo toBeDeleted) {
            setRequestType(Request.Type.Delete);
            setObjectInfo(toBeDeleted);
        }
    }
}
