/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.api.action;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.apache.struts.action.ActionErrors;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;
import org.apache.struts.util.MessageResources;

import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.api.actionForm.Data;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.content.AggregatorModel;
import org.nsdl.repository.model.content.CollectionsModel;
import org.nsdl.repository.model.content.MetadataModel;
import org.nsdl.repository.model.content.MetadataProviderModel;
import org.nsdl.util.DateUtility;

public class Describe extends GenericAction {
    private static final Logger logger = Logger.getLogger(Describe.class
            .getName());

    private static String repositoryBase;

    private static Collection<Relationship.Name> relsForInverse = Arrays
            .asList(CollectionsModel.Components.AGGREGATED_BY,
                    AggregatorModel.Components.AGGREGATOR_FOR,
                    MetadataModel.Components.METADATA_FOR,
                    MetadataProviderModel.Components.METADTA_PROVIDER_FOR,
                    AggregatorModel.Components.ASSOCIATED_WITH);

    private MessageResources resources;

    public ActionForward execute(ActionMapping mapping, ActionForm form,
            HttpServletRequest request, HttpServletResponse response)
            throws Exception {

        /* Set up basics */
        initDataFormBean(form, request);

        /* Get message resources, errors */
        ActionErrors errors = new ActionErrors();

        /* Initialize the message resources, if not initialized already */
        if (resources == null)
            resources = this.getResources(request);

        if (repositoryBase == null) {
            repositoryBase = resources
                    .getMessage("param.repository.handle.url");
        }

        /* Will be from request. View = "html", for example */
        String view = null;

        try {
            NDRAccess ndr = getNDR(request);
            Request describeRequest = getApiRequest(request);
            NDRObject item = ndr.getObject(describeRequest.getObjectInfo());

            populateForm(item, ndr, (Data) form);
            ((Data) form).setHandle(item.handle());

            List<String> viewParams = describeRequest.getParameter("view");
            if (viewParams != null && viewParams.size() == 1) {
                view = describeRequest.getParameter("view").get(0);
            }
        } catch (Exception e) {
            if (e instanceof ExistenceException) {
                if (logger.isDebugEnabled()) {
                    logger.debug("Could not process Get call for "
                            + request.getPathInfo(), e);
                }
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
            } else {
                logger.error("Could not process Get call for "
                        + request.getPathInfo(), e);
                response
                        .setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
            }
            ((Data) form).setCode(resources.getMessage("error.object.code"));
            errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(
                    "error.object.fedora", "get: ", e.getMessage()));
            saveMessages(request, errors);
            return (mapping.findForward(ERROR_FORWARD));
        }

        if ("html".equals(view)) {
            return (mapping.findForward("successHTML"));
        } else {
            return (mapping.findForward(SUCCESS_FORWARD));
        }
    }

    private static void populateForm(NDRObject object, NDRAccess ndr,
            Data outputForm) {

        /* Populate Describe's 'properties' (in the Fedora sense) */
        outputForm.setCreatedDate(DateUtility
                .getISO8601MilisecondFromDate(object.getCreatedDate()));
        outputForm.setLastModifiedDate(DateUtility
                .getISO8601MilisecondFromDate(object.getLastModifiedDate()));
        outputForm.setState(object.state().toString());
        outputForm.setLabel(object.type().toString());

        /* Populate the output's properties and relationships */
        outputForm.setRelationships(getRels(object, ndr, true));

        /* Get the "inverse relationships */
        outputForm.setRelatedObjects(getInverseRels(object, ndr, true));

        /* Get the datastreams */
        populateDSInfo(object, outputForm);

        /* Get the metadata for display */
        populateMetadata(object, ndr, outputForm);
    }

    private static ArrayList<Map<String, String>> getRels(NDRObject object,
            NDRAccess ndr, boolean activeOnly) {
        ArrayList<Map<String, String>> rels = new ArrayList<Map<String, String>>();

        /* Get object's own properties */
        for (Property p : object.getProperties()) {
            Map<String, String> prop = new HashMap<String, String>();
            prop.put("rel", p.predicate().getLocalPart());
            prop.put("target", p.value());
            rels.add(prop);
        }

        /* Get object's own relationships */
        for (Relationship r : object.getRelationships()) {

            /* Skip deleted objects, if requested */
            if (activeOnly
                    && !NDRObject.State.Active.equals(ndr
                            .getObject(r.related()).state())) {
                continue;
            }

            Map<String, String> prop = new HashMap<String, String>();
            prop.put("rel", r.predicate().getLocalPart());
            prop.put("target", r.related().handle());
            rels.add(prop);
        }

        return rels;
    }

    private static ArrayList<Map<String, String>> getInverseRels(
            NDRObject object, NDRAccess ndr, boolean activeOnly) {
        ArrayList<Map<String, String>> rels = new ArrayList<Map<String, String>>();

        /* Get the 'inverse' relationships pointing in */

        for (Relationship.Name desiredRel : relsForInverse) {

            Results<Relationship> related = ndr.getObjectFinder().getRelatedTo(
                    object, desiredRel);
            try {
                for (Relationship r : related) {

                    logger.debug("Inverse relationship: " + r.predicate());
                    /* Skip non-actice objects, if requested */
                    if (activeOnly
                            && !NDRObject.State.Active.equals(ndr.getObject(
                                    r.subject()).state())) {
                        logger.debug("Is not active...");
                        continue;
                    }
                    Map<String, String> rel = new HashMap<String, String>();
                    rel.put("relationship", r.predicate().getLocalPart());
                    rel.put("relatedHandle", r.subject().handle());

                    rels.add(rel);
                }
            } finally {
                related.close();
            }
        }

        return rels;
    }

    /* This is basically just ripped from RepositoryObject.java */
    private static void populateDSInfo(NDRObject object, Data mfrmData) {
        String lstrContentIDList = "";
        String lstrFormatIDList = "";
        String lstrServiceDescriptionIDList = "";
        String lstrHarvestRequestIDList = "";

        for (Datastream d : object.getDatastreams()) {
            String lstrStreamID = d.getName().getID();

            if (lstrStreamID.startsWith(Constants.INPUT_DATA_CONTENT_DS)) {
                // Node is content datastream, make a list.
                if (lstrContentIDList != "") {
                    lstrContentIDList = lstrContentIDList + "," + lstrStreamID;
                } else {
                    lstrContentIDList = lstrStreamID;
                }
                mfrmData.setContentDSIDList(lstrContentIDList);

            } else if (lstrStreamID.startsWith(Constants.INPUT_DATA_FORMAT_DS)) {
                // Node is format (metadata) datastream
                if (lstrFormatIDList != "") {
                    lstrFormatIDList = lstrFormatIDList + "," + lstrStreamID;
                } else {
                    lstrFormatIDList = lstrStreamID;
                }
                mfrmData.setFormatDSIDList(lstrFormatIDList);

                // If it's not an info stream, get the data too.
                if (!lstrStreamID.endsWith("_info")) {
                    try {
                        mfrmData.setMetadata(streamToString(d.getContent()));
                    } catch (AccessException e) {
                        throw new RepositoryException(e);
                    }
                }

            } else if (lstrStreamID.startsWith(Constants.INPUT_DATA_SERVICE_DS)) {

                // Node is a service description
                if (lstrServiceDescriptionIDList != "") {
                    lstrServiceDescriptionIDList = lstrServiceDescriptionIDList
                            + "," + lstrStreamID;
                } else {
                    lstrServiceDescriptionIDList = lstrStreamID;
                }
                mfrmData
                        .setServiceDescriptionDSIDList(lstrServiceDescriptionIDList);
            } else if (lstrStreamID
                    .startsWith(Constants.INPUT_DATA_HARVEST_REQUEST_DS)) {

                // Node is a harvestRequest
                if (lstrHarvestRequestIDList != "") {
                    lstrHarvestRequestIDList = lstrHarvestRequestIDList + ","
                            + lstrStreamID;
                } else {
                    lstrHarvestRequestIDList = lstrStreamID;
                }
                mfrmData.setDataSourceDSIDList(lstrHarvestRequestIDList);
            }
        }
    }

    private static void populateMetadata(NDRObject target, NDRAccess ndr,
            Data output) {
        for (Relationship r : ndr.getObjectFinder().getRelatedTo(target,
                MetadataModel.Components.METADATA_FOR)) {

            /* NSDL we only seal with nsdl_dc here... */
            Datastream md = ndr.getObject(r.subject()).getDatastream(
                    MetadataModel.Components.NSDL_DC);
            if (md != null) {
                try {
                    output.setMetadata(streamToString(md.getContent()));
                } catch (Exception e) {
                    throw new RepositoryException(
                            "Could not load related metadata", e);
                }
            }
        }
    }

    private static String streamToString(InputStream stream) {
        StringBuilder contents = new StringBuilder();
        String line;
        BufferedReader reader = new BufferedReader(
                new InputStreamReader(stream));
        try {
            while ((line = reader.readLine()) != null) {
                contents.append(line + "\n");
            }
        } catch (IOException e) {
            throw new RepositoryException(e);
        } finally {
            try {
                reader.close();
            } catch (IOException e) {
                logger.warn("Could not close reader (not that anyone cares..)");
            }
        }
        return contents.toString();
    }
}
