/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.api.action;

import java.io.IOException;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.log4j.Logger;
import org.apache.struts.action.ActionErrors;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;
import org.apache.struts.util.MessageResources;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.actionForm.Data;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.api.request.RequestUtils;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.util.Configuration;
import org.nsdl.util.DateUtility;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class Get extends GenericAction {

    private static final String RESPONSE_NS = "http://ns.nsdl.org/ndr/response_v1.00/";

    private static final Logger logger = Logger.getLogger(Get.class.getName());

    private static String repositoryBase = "";

    private MessageResources resources;

    public ActionForward execute(ActionMapping mapping, ActionForm form,
            HttpServletRequest request, HttpServletResponse response)
            throws Exception {

        /* Set up basics */
        initDataFormBean(form, request);

        /* Get message resources, errors */
        ActionErrors errors = new ActionErrors();

        /* Initialize the message resources, if not initialized already */
        if (resources == null)
            resources = this.getResources(request);

        Configuration config = getConfig(request);

        if (repositoryBase == "") {
            repositoryBase = config.getProperty("param.repository.baseURL", "");
        }

        try {
            NDRAccess ndr = getNDR(request);
            Request getRequest = getApiRequest(request);

            /*
             * Hack for pid-type handles. Struts will send them here instead of
             * getDatastream. Make sure we don't have any additional path info.
             * if we do, this is really a datastream request.
             */
            if (RequestUtils.getPath(request).hasPath()) {
                return new GetDatastream().execute(mapping, form, request,
                        response);
            }

            NDRObject item = ndr.getObject(getRequest.getObjectInfo());

            ((Data) form).setResponseXML(getAllContents(item, ndr, true));
            ((Data) form).setHandle(item.handle());
        } catch (Exception e) {

            if (e instanceof ExistenceException) {
                if (logger.isDebugEnabled()) {
                    logger.debug("Could not process Get call for "
                            + request.getPathInfo(), e);
                } else {
                    logger.info(e.getMessage());
                }
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
            } else {
                logger.error("Could not process Get call for "
                        + request.getPathInfo(), e);
                response
                        .setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
            }
            ((Data) form).setCode(resources.getMessage("error.object.code"));
            errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(
                    "error.object.fedora", "get: ", e.getMessage()));
            saveMessages(request, errors);

            return (mapping.findForward(ERROR_FORWARD));
        }

        return (mapping.findForward(SUCCESS_FORWARD));
    }

    protected static String getAllContents(NDRObject item, NDRAccess ndr,
            boolean activeOnly) throws Exception {
        /* Setup response container */
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        Document response = factory.newDocumentBuilder().newDocument();

        /* TODO: put ns declarations in document */
        Element objectInfo = response.createElementNS(RESPONSE_NS, "NDRObject");

        if (item.getProperties().size() > 0) {
            Element properties = response.createElementNS(RESPONSE_NS,
                    "properties");

            /* First, add the 'system' properties */
            Element prop;

            /* createdDate */
            prop = response.createElementNS(
                    "info:fedora/fedora-system:def/view#",
                    "fedora-view:createdDate");
            NDRXMLutil.setElementValue(prop, DateUtility
                    .getISO8601MilisecondFromDate(item.getCreatedDate()));
            NDRXMLutil.declareNamespace(objectInfo,
                    "info:fedora/fedora-system:def/view#", "fedora-view");
            properties.appendChild(prop);

            /* lastModifiedDate */
            prop = response.createElementNS(
                    "info:fedora/fedora-system:def/view#",
                    "fedora-view:lastModifiedDate");
            NDRXMLutil.setElementValue(prop, DateUtility
                    .getISO8601MilisecondFromDate(item.getLastModifiedDate()));
            properties.appendChild(prop);

            /* state */
            prop = response.createElementNS(
                    "info:fedora/fedora-system:def/model#",
                    "fedora-model:state");
            NDRXMLutil.declareNamespace(objectInfo,
                    "info:fedora/fedora-system:def/model#", "fedora-model");
            NDRXMLutil.setElementValue(prop, item.state().toString());
            properties.appendChild(prop);

            /* Next, all the 'user' properties */
            for (Property p : item.getProperties()) {
                String qualifiedName = "";
                boolean localDeclaration = false;
                if (p.predicate().getPrefix() != "") {
                    qualifiedName = p.predicate().getPrefix() + ":"
                            + p.predicate().getLocalPart();
                    NDRXMLutil.declareNamespace(objectInfo, p.predicate()
                            .getNamespaceURI(), p.predicate().getPrefix());
                } else {
                    localDeclaration = true;
                    qualifiedName = p.predicate().getLocalPart();
                }
                prop = response.createElementNS(
                        p.predicate().getNamespaceURI(), qualifiedName);
                if (localDeclaration) {
                    prop.setAttribute("xmlns", p.predicate().getNamespaceURI());
                }
                NDRXMLutil.setElementValue(prop, p.value());
                properties.appendChild(prop);
            }

            objectInfo.appendChild(properties);
        }

        Element ds = null;
        Element datastreams = response.createElementNS(RESPONSE_NS, "data");
        for (Datastream d : item.getDatastreams()) {
            if ("RELS-EXT".equals(d.getName().getID()))
                continue;
            /* Special case for metadata */
            if (d.getName().getID().startsWith("format_")) {
                ds = response.createElementNS(RESPONSE_NS, "format");

                /* Get the type (could also use a regex here */
                String type = d.getName().getID();
                type = type.replaceAll("^format_", "");
                type = type.replaceAll("^native_", "");
                type = type.replaceAll("_info$", "");
                type = type.replaceAll("_v\\d+\\.*\\d*$", "");
                ds.setAttribute("type", type);
            } else {
                /* TODO: handle datastream namespaces when it's available */
                ds = response.createElementNS(RESPONSE_NS, d.getName().getID());
            }

            ds.setAttribute("ID", d.getName().getID());

            if (d.getLabel() != null && d.getLabel() != "") {
                ds.setAttribute("LABEL", d.getLabel());
            }

            /*
             * Get the datastream info from foxml for determining its size and
             * type
             */
            ds.setAttribute("MIMETYPE", d.getMimeType());

            if (d.isCanonical()) {
                ds.setAttribute("canonical", "true");
            } else {
                ds.setAttribute("canonical", "false");
            }

            /*
             * NOTE: getting the datastream size is expensive, since it requires
             * a trip to Fedora. For now, then, we disable it
             */
            // ds.setAttribute("SIZE", Long.toString(d.getSize()));
            NDRXMLutil.setElementValue(ds, repositoryBase + "get/"
                    + item.handle() + "/" + d.getName().getID());
            datastreams.appendChild(ds);
        }

        objectInfo.appendChild(datastreams);

        if (item.getRelationships().size() > 0) {
            Element relationships = response.createElementNS(RESPONSE_NS,
                    "relationships");
            for (Relationship r : item.getRelationships()) {

                boolean localDeclaration = false;
                /* Skip non-active relationships, if requested */
                if (activeOnly
                        && !NDRObject.State.Active.equals(ndr.getObject(
                                r.related()).state())) {
                    continue;
                }

                String qualifiedName = "";
                if (r.predicate().getPrefix() != "") {
                    qualifiedName = r.predicate().getPrefix() + ":"
                            + r.predicate().getLocalPart();
                    NDRXMLutil.declareNamespace(objectInfo, r.predicate()
                            .getNamespaceURI(), r.predicate().getPrefix());
                } else {
                    localDeclaration = true;
                    qualifiedName = r.predicate().getLocalPart();
                }

                Element prop = response.createElementNS(r.predicate()
                        .getNamespaceURI(), qualifiedName);

                if (localDeclaration) {
                    prop.setAttribute("xmlns", r.predicate().getNamespaceURI());
                }
                NDRXMLutil.setElementValue(prop, r.related().handle());
                relationships.appendChild(prop);
            }
            objectInfo.appendChild(relationships);
        }

        try {
        	String responseXML = NDRXMLutil.serializeDOMElement(objectInfo).replaceAll(
                    "<\\?.+\\?>", "")
                    + "\n"; 
            return responseXML;
        } catch (IOException e) {
            throw new RepositoryException("Error serializing object content", e);
        }
    }
}
