/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.api.action;

import java.io.ByteArrayOutputStream;
import java.util.*;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.apache.struts.action.ActionErrors;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;
import org.apache.struts.util.MessageResources;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.actionForm.Data;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.oai.OAIMetadata;
import org.nsdl.repository.oai.OAITransformer;
import org.nsdl.repository.api.actionForm.OAIResponse;
import org.nsdl.repository.api.request.Request;

/**
 * The NDR OAI Provider calls this NAPI method to prepare the OAI server data
 * cache
 * 
 * @author Edwin Shin
 */
public class GetMetadata extends GenericAction {
    private static final String ME = "getMetadata";

    private static final Logger logger = Logger.getLogger(GetMetadata.class);

    // Message properties file access is through this:
    private MessageResources mrscResources = null;

    private static final HashMap<String, OAITransformer> mmapOAITransformers = new HashMap<String, OAITransformer>();

    public ActionForward execute(ActionMapping mapping, ActionForm form,
            HttpServletRequest request, HttpServletResponse response)
            throws Exception {

        // Cast incoming form so we can reference it
        Data frmData = (Data) form;

        // Set up basics
        initDataFormBean(frmData, request);

        // Message properties file access is through this:
        if (mrscResources == null)
            mrscResources = this.getResources(request);

        // Set up error mechanism
        ActionErrors errors = new ActionErrors();

        NDRAccess ndr = getNDR(request);
        Request ndrRequest = getApiRequest(request);

        try {
            String xslObjectID = frmData.getXsl();
            String metadataFormat = frmData.getFormat();

            ByteArrayOutputStream out = new ByteArrayOutputStream();

            OAITransformer oait = getTransformer(xslObjectID, ndr);

            OAIMetadata metadata = new OAIMetadata(ndrRequest.getObjectInfo(),
                    oait, ndr);

            if (request.getPathInfo().startsWith("/getMetadataAbout")) {
                metadata.writeOAIAbout(metadataFormat, out);
            } else {
                metadata.writeMetadata(metadataFormat, out);
            }

            OAIResponse lobjResponse = new OAIResponse(new String(out
                    .toByteArray(), "UTF-8"));
            out.close();

            request.setAttribute("Metadata", lobjResponse);

            return (mapping.findForward(SUCCESS_FORWARD));
        } catch (Exception e) {
            if (e instanceof ExistenceException) {
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
            } else {
                response
                        .setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
            }
            String lstrErMsg = e.getMessage();
            frmData.setCode(mrscResources.getMessage("error.application.code"));
            errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(
                    "error.application.api", ME + " " + lstrErMsg, ME + " "
                            + lstrErMsg));

            if (e instanceof AccessException && !logger.isDebugEnabled()) {
                logger.info("Error accessing metadata: " + e.getMessage());
            } else {
                logger.error("Could not process metadata for "
                        + request.getPathInfo(), e);
            }
            saveMessages(request, errors);
            return (mapping.findForward(ERROR_FORWARD));
        }
    }

    private synchronized OAITransformer getTransformer(String xslObject,
            NDRAccess ndr) {
        if (mmapOAITransformers.containsKey(xslObject)) {
            return mmapOAITransformers.get(xslObject);
        } else {
            // Need to generate it
            OAITransformer oait = new OAITransformer(xslObject, ndr);
            mmapOAITransformers.put(xslObject, oait);
            return oait;
        }
    }
}
