/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.action;

import java.io.IOException;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.struts.action.ActionErrors;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;
import org.apache.struts.util.MessageResources;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.api.actionForm.Data;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Relationship;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Lists all relationships inside and around an object.
 * 
 * @author birkland
 * @since 1.2
 */
public class ListRelationships extends GenericAction {

    private static final String RESPONSE_NS = "http://ns.nsdl.org/ndr/response_v1.00/";
    private MessageResources resources;

    public ActionForward execute(ActionMapping mapping, ActionForm form,
            HttpServletRequest request, HttpServletResponse response)
            throws Exception {
        /* Set up basics */
        initDataFormBean(form, request);

        /* Get message resources, errors */
        ActionErrors errors = new ActionErrors();

        /* Initialize the message resources, if not initialized already */
        if (resources == null)
            resources = this.getResources(request);

        try {
            NDRAccess ndr = getNDR(request);
            Request getRequest = getApiRequest(request);
            NDRObject item = ndr.getObject(getRequest.getObjectInfo());

            ((Data) form).setResponseXML(getRelationshipSummary(item, ndr));
            ((Data) form).setHandle(item.handle());
        } catch (Exception e) {
            if (e instanceof ExistenceException) {
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
            } else {
                response
                        .setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
            }
            ((Data) form).setCode(resources.getMessage("error.object.code"));
            errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(
                    "error.object.fedora", "get: ", e.getMessage()));
            saveMessages(request, errors);
            return (mapping.findForward(ERROR_FORWARD));
        }

        return (mapping.findForward(SUCCESS_FORWARD));
    }

    private static String getRelationshipSummary(NDRObject o, NDRAccess ndr)
            throws Exception {
        /* Setup response container */
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        Document response = factory.newDocumentBuilder().newDocument();

        Element resultData = response
                .createElementNS(RESPONSE_NS, "resultData");

        Element summary = response.createElementNS(RESPONSE_NS,
                "relationshipSummary");
        resultData.appendChild(summary);

        /* Summarize all internal relationships */
        Element internalRels = response
                .createElementNS(RESPONSE_NS, "internal");
        appendRels(o.getRelationships(), internalRels, response, ndr);
        summary.appendChild(internalRels);

        /* Use a finder and summarize all external (pointing-in) rels */
        Element externalRels = response
                .createElementNS(RESPONSE_NS, "external");
        Results<Relationship> extRels = ndr.getObjectFinder().getRelatedTo(o);

        try {
            appendRels(extRels, externalRels, response, ndr);
        } finally {
            extRels.close();
        }

        summary.appendChild(externalRels);

        try {
            return NDRXMLutil.serializeDOMElement(resultData).replaceAll(
                    "<\\?.+\\?>", "")
                    + "\n";
        } catch (IOException e) {
            throw new RepositoryException("Error serializing object content", e);
        }
    }

    private static void appendRels(Iterable<Relationship> rels,
            Element element, Document parentDoc, NDRAccess ndr) {
        for (Relationship r : rels) {

            NDRObject subject = ndr.getObject(r.subject());
            NDRObject related = ndr.getObject(r.related());

            String qualifiedName = "";
            if (r.predicate().getPrefix() != "") {
                qualifiedName = r.predicate().getPrefix() + ":"
                        + r.predicate().getLocalPart();
                NDRXMLutil.declareNamespace(element, r.predicate()
                        .getNamespaceURI(), r.predicate().getPrefix());
            } else {
                qualifiedName = r.predicate().getLocalPart();
            }

            Element prop = parentDoc.createElementNS(r.predicate()
                    .getNamespaceURI(), qualifiedName);
            prop.setAttribute("subject", subject.handle().toString());
            prop.setAttribute("subject.type", subject.type().toString());
            prop.setAttribute("subject.state", subject.state().toString());
            prop.setAttribute("type", related.type().toString());
            prop.setAttribute("state", related.state().toString());
            NDRXMLutil.setElementValue(prop, r.related().handle());

            element.appendChild(prop);
        }
    }
}
