/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.api.action;

import java.util.HashSet;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.apache.struts.action.ActionErrors;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;
import org.apache.struts.util.MessageResources;

import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.actionForm.Data;
import org.nsdl.repository.api.auth.AgentAuthInfo;
import org.nsdl.repository.api.auth.AuthorizationException;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.api.request.Request.Commands;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;

/**
 * Modify an NDR object
 * 
 */
public class Modify extends GenericAction {
    private static final Logger logger = Logger.getLogger(Modify.class);

    public ActionForward execute(ActionMapping mapping, ActionForm form,
            HttpServletRequest request, HttpServletResponse response)
            throws Exception {

        logger.trace("Entering execute");
        /* Set up basics */
        initDataFormBean(form, request);

        /* Get message resources, errors */
        MessageResources resources = this.getResources(request);
        ActionErrors errors = new ActionErrors();

        Request modifyRequest = null;
        AgentAuthInfo auth = null;
        NDRAccess ndr = null;

        NDRObject currentObject;

        logger.trace("Executing Modify request");
        /* Check authorization, and open object for editing */
        try {

            /* Get necessary information from the request */
            modifyRequest = getApiRequest(request);
            auth = getAuth(request);
            ndr = getNDR(request);

            logger.trace("Checking auth");
            checkAuth(auth, modifyRequest);

            logger.trace("Opening object for modification");
            currentObject = ndr.getObject(modifyRequest.getObjectInfo());

            /* Make all requested modifications and commit */

            propagateChanges(modifyRequest, currentObject);
            logger.trace("Committing changes");

            currentObject.commit();

            ((Data) form).setHandle(currentObject.handle());

        } catch (AuthorizationException e) {
            logger.info("Error authorizing modification of "
                    + modifyRequest.getObjectInfo().handle() + " by "
                    + auth.agent().handle() + ": " + e.getMessage());
            logger.debug("Fail trace: \n", e);

            ((Data) form).setCode(resources.getMessage("error.security.code"));
            errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(
                    "error.security.authorization", " Modify: "
                            + e.getMessage()));
            saveMessages(request, errors);

            response.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
            return (mapping.findForward(ERROR_FORWARD));
        } catch (Exception e) {
            if (e instanceof ExistenceException) {
                response.setStatus(HttpServletResponse.SC_NOT_FOUND);
            } else {
                response
                        .setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
            }
            logger.warn("Could not apply changes: ", e);
            String message = e.getMessage();

            ((Data) form).setCode(resources
                    .getMessage("error.application.code"));
            errors.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(
                    "error.application.api", " Modify: " + message));
            saveMessages(request, errors);

            return (mapping.findForward(ERROR_FORWARD));
        }

        logger.trace("Success");
        return mapping.findForward(SUCCESS_FORWARD);
    }

    protected static void propagateChanges(Request r, NDRObject o) {

        /*
         * XXX: State changes are intended to exist in other calls (i.e. delete,
         * deactivate, etc). Until that's done, this exists here for backwards
         * compatability. We screen the request for the 'special case'
         * properties, remove them, and perform the requested action here. We
         * know that ingest uses the default command for these, so we only have
         * one payload to filter.
         */
        HashSet<Property> replaceProperties = new HashSet<Property>(r
                .getProperties(Commands.REPLACE));

        logger.trace("checking state modification");
        for (Property p : replaceProperties) {
            if (p.predicate().getLocalPart().equalsIgnoreCase("state")) {
                replaceProperties.remove(p);

                logger.info("Backdoor state change to " + p.value() + " for "
                        + o.handle());
                if (p.value().equals("A") || p.value().equals("Active")) {
                    o.setState(NDRObject.State.Active);
                } else if (p.value().equals("D") || p.value().equals("Deleted")) {
                    o.setState(NDRObject.State.Deleted);
                } else if (p.value().equals("I")
                        || p.value().equals("Inactive")) {
                    o.setState(NDRObject.State.Inactive);
                } else {
                    logger.warn("Unknown state '" + p.value() + "'");
                }
            } else if (p.predicate().getLocalPart().equalsIgnoreCase("label")) {
                replaceProperties.remove(p);
                logger.warn("Discarded label " + p.value());
            }
        }

        /* Apply property changes */
        logger.trace("Registering property changes");
        o.addProperties(r.getProperties(Commands.ADD));
        o.removeProperties(r.getProperties(Commands.DELETE));
        for (Property p : replaceProperties) {
            int size = o.getProperties(p.predicate()).size();
            if (size > 1) {
                throw new AccessException("More than one property named "
                        + p.predicate()
                        + ", cannot determine which to modify.  "
                        + "Please explicitly delete the desired relationship, "
                        + "and add its replacement");
            } else if (size == 0) {
                o.addProperty(p.predicate(), p.value());
            } else if (size == 1) {
                o.getProperty(p.predicate()).setValue(p.value());
            }
        }

        /* Apply relationship changes */
        logger.trace("Registering relationship changes");
        o.addRelationships(r.getRelationships(Commands.ADD));
        o.removeRelationships(r.getRelationships(Commands.DELETE));
        for (Relationship rel : r.getRelationships(Commands.REPLACE)) {
            int size = o.getRelationships(rel.predicate()).size();
            if (size > 1) {
                throw new AccessException("More than one relationship named "
                        + rel.predicate()
                        + ", cannot determine which to modify.  "
                        + "Please explicitly delete the desired relationship, "
                        + "and add its replacement");
            } else if (size == 0) {
                o.addRelationship(rel.predicate(), rel.related());
            } else if (size == 1) {
                o.getRelationship(rel.predicate()).setRelated(rel.related());
            }
        }

        /* Apply datastream changes */
        Set<Datastream> toAdd = r.getDatastreams(Commands.ADD);
        logger.debug("Registering " + toAdd.size() + " datastream adds");
        for (Datastream d : r.getDatastreams(Commands.ADD)) {
            logger.debug("Will add datastream: " + d.getName());
        }
        o.addDatastreams(toAdd);

        Set<Datastream> toReplace = r.getDatastreams(Commands.REPLACE);
        logger.debug("Registering " + toReplace.size()
                + " datastream replacements");
        for (Datastream d : toReplace) {
            logger.debug("Will replace datastream: " + d.getName());

            Datastream toUpdate = o.getDatastream(d.getName());
            if (toUpdate == null) {
                toUpdate = o.addDatastream(d.getName());
                toUpdate.setMimeType(d.getMimeType());
            }

            if (!d.isCanonical()) {
                toUpdate.setURL(d.getURL());
            } else {
                toUpdate.setContent(d.getContent());
            }
        }

        Set<Datastream> toDelete = r.getDatastreams(Commands.DELETE);
        logger.debug("Registering " + toDelete.size() + " datastream deletes");
        for (Datastream d : toDelete) {
            logger.debug("Will delete datastream: " + d.getName());
        }
        o.removeDatastreams(toDelete);
    }
}
