/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth;

import java.util.Set;

import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;

/** Agent Authorization/Verification Information
 * <p>
 * Provides in interface into determining if a given NSDL agent has
 * been successfully authenticated and verified
 * (isVerified()), as well as determining if the that agent is allowed
 * to modify a given object or carry out a given request 
 * </p>
 * @author birkland
 *
 */
public interface AgentAuthInfo {
	
	/** Return the identity of this particular agent
	 * 
	 * @return NDRObjectInfo describing this particular agent
	 */
	public NDRObject agent();
	
	/** Determine if this instance contains information about a 
	 * an agent whose authenticity has been verified
	 * 
	 * @return true if the agent's identity has been verified
	 */
	public boolean isVerified();
	
	/** Determine if an authorized agent is a member of any of the given
	 * aggregations
	 * 
	 * @param aggregations The set of aggregations to test membership in
	 * @return	true if the authorized agent is a member of at least one of the
	 * given aggregations
	 */
	public boolean memberOf(Set<NDRObject> aggregations);
	
	/** Determine of an authorized agent is a member of a particular
	 * aggregation
	 * 
	 * @param aggregation The aggregation to test the Agents membership in
	 * @return true if the authorized agent is in the given aggregation
	 */
	public boolean memberOf(NDRObjectInfo aggregation);
	
	
	/** Determine if this particular agent is authorized to modify or delete the
	 * current object.
	 * <p>
	 * Note: By 'modify' we mean not only modifying content in the object, but
	 * also modifying information that can be 'logically' considered to be part 
	 * of it. For example, adding a resource to an aggregation is actually modifying
	 * the resource object (since that is where the relationship is defined), but 
	 * logically it can be viewed as modifying the contents of the aggregation.
	 * </p>
	 * @param object The object we are wondering if this agent is authorized to change
	 * @return true if this agent can modify the given object
	 */
	public boolean isAuthorizedFor(NDRObjectInfo object);
	
	/** Determine if this particular agent is authorized to perform every
	 * action specified in the given request 
	 * 
	 * @param request
	 * @return true if thhe Agent is authorized for the given request.
	 */
    
	public boolean isAuthorizedFor(Request request);
    
    /** Add a message related to authorization or authentication 
     * <p>
     * The intent this message is to convey any useful information related to 
     * authentaction or authorization to the end user. 
     * </p>
     * @param message The message to give this AgentAuthInfo instance
     */
    public void addMessage(String message);
    
    /** Read all messages passed to this object
     * <p>
     * Messages are likely to contain text that is useful to the end user.
     * For example, suppose an isAuthorizedFor() request fails.  The message
     * might contain some words as to why.  Messages are not compulsory, so
     * be prepared for empty strings, but not nulls.
     * </p>
     */
    public String getMessage();
	
}
