/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.api.auth;

import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.Defaults;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.auth.rules.RuleManager;
import org.nsdl.repository.api.auth.rules.RuleSet;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.content.AuthorizationModel;
import org.nsdl.repository.model.content.MetadataModel;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.StaticObjectInfo;

/**
 * The current NSDL Agent authorization implementation
 * 
 * @author birkland
 * 
 */
public class AgentAuthorization implements AgentAuthInfo {

    private final StringBuilder messages = new StringBuilder();

    private static final Logger logger = Logger
            .getLogger(AgentAuthorization.class.getName());

    private final NDRObject authorizedAgent;

    private final NDRAccess ndr;

    private final boolean authorized;

    private final RuleManager m_rules;

    private final String ROOT_AGENT_HANDLE;
    private final NDRObjectInfo TRUSTED_AGENTS;

    public AgentAuthorization(NDRObject agent, NDRAccess ndr,
            Configuration config, boolean authorized) {
        if (!agent.type().equals(Type.Agent)) {
            throw new IllegalArgumentException("AgentAuthInfo must be "
                    + "given an agent upon construction, but instead "
                    + "was given a " + agent.type().toString());
        }
        this.authorizedAgent = agent;
        this.authorized = authorized;
        this.ndr = ndr;
        m_rules = new RuleManager();
        m_rules.updateConfig(config);
        ROOT_AGENT_HANDLE = config.getProperty(
                AuthGateway.PARAM_ROOT_AGENT_HANDLE, Defaults.ROOT_AGENT
                        .handle());
        TRUSTED_AGENTS = new StaticObjectInfo(config.getProperty(
                AuthGateway.PARAM_TRUSTED_AGENTS_AGGREGATOR,
                Defaults.TRUSTED_AGENTS.handle()), Type.Aggregator);

    }

    public NDRObject agent() {
        return this.authorizedAgent;
    }

    public boolean isVerified() {
        return this.authorized;
    }

    public boolean memberOf(Set<NDRObject> aggs) {
        logger.warn("memberOf(Set<NDRObjectInfo> aggs) is not implementrd");
        /* TODO: Insert logic here */
        return false;
    }

    public boolean memberOf(NDRObjectInfo agg) {
        /* XXX: probavbly more efficient way to do this */
        logger.trace("Checking aggregation membership in " + agg.handle());
        for (Relationship r : authorizedAgent
                .getRelationships(Components.MEMBER_OF)) {
            ;
            if (r.related().handle().equals(agg.handle())) {
                logger.trace("Yes, is member of aggregation " + agg.handle());
                return true;
            }
        }
        logger.trace("no, is not a member of aggregation " + agg.handle());
        return false;
    }

    public boolean isAuthorizedFor(NDRObjectInfo identity) {

        NDRObject object = ndr.getObject(identity);

        logger.trace("BEGIN checking auth for OBJECT " + object.handle());

        /* Root can do anything.. */
        if (this.authorizedAgent.handle().equals(ROOT_AGENT_HANDLE)) {
            logger.trace("SUCCESS checking auth for OBJECT " + object.handle()
                    + ", Root can do anything");
            return true;
        }

        /* An agent is always authorized for itself */
        if (this.authorizedAgent.handle().equals(object.handle())) {
            logger.trace("SUCCESS checking auth for OBJECT " + object.handle()
                    + ", agent is authorized for itself");
            return true;
        }

        if (object.type().equals(Type.Agent) && memberOf(TRUSTED_AGENTS)) {
            logger.trace("SUCCESS checking auth for OBJECT " + object.handle()
                    + ", Trusted agents can modify other agents");
            return true;
        }

        /*
         * If the object is an aggregator or MetadataProvider, then check for
         * the presence of the Authorized relationship
         */
        if (object.type().equals(Type.MetadataProvider)
                || object.type().equals(Type.Aggregator)) {

            logger.trace("This is a MP or Aggregator, "
                    + "looking for auth relationship..");
            for (Relationship rel : object
                    .getRelationships(AuthorizationModel.Grouping.Components.AUTHORIZED)) {
                if (rel.related().handle().equals(authorizedAgent.handle())) {
                    /* The agent itself has been explicitly authorized */
                    logger.trace("SUCCESS checking auth for OBJECT "
                            + object.handle()
                            + " Agent has been explicitly authorized");
                    return true;
                } else if (rel.related().type().equals(Type.Aggregator)) {

                    /*
                     * An aggregation has been 'authorized'. Check to see if our
                     * agent is in this aggregation
                     */
                    try {
                        if (this.memberOf(ndr.getObject(rel.related()))) {
                            logger.trace("SUCCESS checking auth for OBJECT "
                                    + object.handle()
                                    + " Agent is in an authorized aggregation");
                            return true;
                        }
                    } catch (Exception e) {
                        throw new RepositoryException("Could not determine "
                                + "authorization " + e.getMessage(), e);
                    }
                }
            }
        }

        /*
         * For Metadata objects, check whether this agent is authorized for its
         * provider
         */
        if (object.type().equals(Type.Metadata)) {
            logger.trace(object.handle()
                    + "is Metadata, checking auth for metadata provider");
            NDRObject provider = null;
            try {
                provider = ndr.getObject(object.getRelationship(
                        MetadataModel.Components.METADATA_PROVIDED_BY)
                        .related());
            } catch (Exception e) {
                /* Not much we can do */
                throw new RepositoryException(e);
            }

            logger
                    .trace("FINISHED checking auth for OBJECT "
                            + object.handle());
            return this.isAuthorizedFor(provider);
        }

        /* Resources don't have authorization yet.. */
        if (object.type().equals(Type.Resource)) {
            logger.warn("Not checking auth for resource " + object.handle());
            logger
                    .trace("FINISHED checking auth for OBJECT "
                            + object.handle());
            return true;
        }

        logger.trace("FAIL, not authorized for OBJECT " + object.handle());
        return false;
    }

    public boolean isAuthorizedFor(Request request) {

        logger.debug("BEGIN checking auth for REQUEST " + request.getRequest());
        /* Root can do anything.. */
        if (this.authorizedAgent.handle().equals(ROOT_AGENT_HANDLE)) {
            logger.trace("SUCCESS checking auth for REQUEST "
                    + request.getRequest() + ", Root can do anything");
            return true;
        } 

        /*
         * We don't check adds, since we know new objects don't exist yet, and
         * thus don't have an established authorization yet
         */
        if (!request.getRequest().startsWith("add")) {
            logger.trace("This is not an add request, "
                    + "so checking object...");
            try {
                if (!isAuthorizedFor(ndr.getObject(request.getObjectInfo()))) {
                    logger.trace("FAIL, not authorized for REQUEST "
                            + request.getRequest());
                    return false;
                }
            } catch (ExistenceException e) {
                /* Should never happen! */
                throw new RepositoryException(e.getMessage(), e);
            }
        }

        /* Check the relationship rules */
        logger.trace("Getting rules for request " + request.getRequest());
        RuleSet rules = m_rules.getRules(request, ndr);

        logger.trace("Checking each request component...");

        for (Component r : request.getComponents()) {

            /*
             * XXX hack: relationships may not have had their types set, if they
             * are straight from a request. There's some redundancy in looking
             * this up here.
             */

            if (r instanceof Relationship) {
                logger.trace("finding relationship target");
                Relationship rel = (Relationship) r;
                r = new Relationship(rel.subject(), rel.predicate(), ndr
                        .getObject(rel.related()));
            } else {
                logger.trace("not a relationship, so not finding target");
            }
            if (!rules.allows(r, request, this)) {
                logger.trace("FAIL component check for REQUEST "
                        + request.getRequest());
                return false;
            }
        }

        logger
                .trace("SUCCESS, authorized for  REQUEST "
                        + request.getRequest());
        return true;
    }

    public void addMessage(String message) {
        messages.append(message.trim() + "\n");
    }

    public String getMessage() {
        return messages.toString();
    }

}
