/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth;

import java.security.KeyManagementException;
import java.security.PublicKey;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.apache.struts.action.ActionServlet;
import org.apache.struts.action.PlugIn;
import org.apache.struts.config.ModuleConfig;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.api.FedoraPlugIn;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.NDRObjectTemplate;
import org.nsdl.repository.util.StaticObjectInfo;

import javax.servlet.ServletException;
import javax.servlet.http.*;

/**
 * Struts validator class for handling authentication/authorization.
 * 
 * <p>
 * Note: The functionality in this struts plugin may be moved into the general
 * NDRAccess API. At that point, this class will become unnecessary, and be
 * removed.
 * </p>
 * 
 * <h2>Configuration Parameters</h2>
 * <p>
 * <dl>
 * <dt><code>param.repository.api.auth.canonicalBasePath</code></dt>
 * <dd>Path used in signed header canonicalization. Needed to handle cases where
 * the repository is running behind a proxy - this value should contain the
 * <em>public</em> path as used by clients when signing requests using APi
 * version 1.0. For example, if the repository is publicly visible as
 * <code>http://ndr.nsdl.org/repository/api/...</code>, the canonical base path
 * here would be <code>/repository/api</code>.</dd>
 * 
 * <dt>param.repository.api.auth.alternatePath.*</dt>
 * <dd>Other paths that may have been used in canonicalization</dd>
 * 
 * <dt>param.repository.api.auth.agent.root.handle</dt>
 * <dd>Handle of the root Agent in the repository</dd>
 * 
 * <dt>param.repository.api.auth.trusted.aggregator.handle</dt>
 * <dd>Handle of the trusted Agents aggregator in the repository</dd>
 * 
 * </dl>
 * </p>
 * 
 * 
 * @author birkland
 */
public class AuthGateway implements PlugIn {

    public static final String PARAM_BASE_PATH = "param.repository.api.auth.canonicalBasePath";
    public static final String PARAM_ALTERTATE_PATH = "param.repository.api.auth.alternatePath.";
    public static final String PARAM_ROOT_AGENT_HANDLE = "param.repository.api.auth.agent.root.handle";
    public static final String PARAM_TRUSTED_AGENTS_AGGREGATOR = "param.repository.api.auth.trusted.aggregator.handle";

    private boolean authenticate;

    private boolean authorize;

    private static final Logger logger = Logger.getLogger(AuthGateway.class
            .getName());

    private NDRAccess ndr;

    private KeyManager manager;

    public static final String GATEWAY_KEY = AuthGateway.class.getName()
            + ".gateway";

    public static final String AUTH_INFO_KEY = AuthGateway.class.getName()
            + ".info";

    public void destroy() {
        ndr = null;
    }

    private String serverPath = "/repository/api"; /* Default */

    private final List<String> redirPaths = new ArrayList<String>();

    private Configuration m_config = new Configuration();

    public void init(ActionServlet servlet, ModuleConfig modConfig)
            throws ServletException {

        ndr = (NDRAccess) servlet.getServletContext().getAttribute(
                Constants.NDR_ACCESS);

        if (ndr == null) {
            throw new ServletException("Repository not initialized!");
        }

        this.manager = new KeyManagerCache(new NDRPubKeyManager(ndr));

        /* Place this gateway in the servlet context */
        servlet.getServletContext().setAttribute(GATEWAY_KEY, this);

        m_config = (Configuration) servlet.getServletContext().getAttribute(
                FedoraPlugIn.CONFIG);

        if (m_config != null) {
            serverPath = m_config.getProperty(PARAM_BASE_PATH, serverPath);
            logger.info("Using canonical path " + serverPath);

            for (String path : m_config.getProperties(PARAM_ALTERTATE_PATH)) {
                redirPaths.add(path);
                logger.info("Adding alternate path " + path);
            }
        } else {
            logger.info("Canonical paths not defined. Using default");
            m_config = new Configuration();
        }

    }

    public AgentAuthInfo getAuthInfo(HttpServletRequest request)
            throws AuthenticationException {

        /* Grab the entire header as a string */

        logger.debug("Getting auth info");
        HttpHeader header = new HttpHeader(request);
        MessageAuthInfo auth;

        /*
         * Because of the proxy re-writes, we need to enumerate the valid http
         * paths in order to choose the correct one to canonicalize. Right now
         * they are baked in
         */
        String host;
        String path;
        if (header.getValue("x-forwarded-host") != null) {
            host = header.getValue("x-forwarded-host");
            if (!redirPaths.isEmpty()) {
                path = redirPaths.get(0);
            } else {
                path = serverPath;
            }
        } else {
            host = request.getServerName() + ":" + request.getServerPort();
            path = serverPath;
        }

        auth = getAuthHeader(request, host, path);

        if (auth.isSigned()) {
            logger.debug("Header is signed by " + auth.getAgent());

            String agentHandle = auth.getAgent();

            if (authorize == false) {
                logger.warn("Authorization is set to off, "
                        + "returning promiscuous info");
                return new PromiscuousAuthInfo();
            }

            NDRObject agentInfo = null;
            try {
                agentInfo = ndr.getObject(new StaticObjectInfo(agentHandle,
                        Type.Agent));
            } catch (Exception e) {
                throw new AuthenticationException(
                        "Error retrieving Agent information " + e.getMessage(),
                        e);
            }

            /* Check to see if the header has a valid signiture */
            try {
                PublicKey pkey = manager.getPubKey(agentHandle);

                if (!auth.isValid(pkey) && !checkAlternatePaths(request, pkey)) {

                    /* Fail. Try re-loading the key, it may have changed */
                    pkey = manager.forcePubKey(agentHandle);

                    if (!auth.isValid(pkey)
                            && !checkAlternatePaths(request, pkey)) {
                        /*
                         * Pass along auth info stating that this agent is not
                         * verified.
                         */
                        logger.debug("Agent's request fails authorization");
                        return new AgentAuthorization(agentInfo, ndr, m_config,
                                false);
                    } else {
                        logger
                                .debug("Successfully loaded changed public key for "
                                        + agentHandle);
                    }
                }
            } catch (KeyManagementException e) {
                throw new AuthenticationException(
                        "Error retrieving public key", e);
            }

            logger.debug("Returning complete authorization credentials");
            return new AgentAuthorization(agentInfo, ndr, m_config, true);
        }

        /* Request is not signed */
        logger.debug("Request is not signed");
        if (authenticate == true) {
            logger.debug("Authenticate is " + authenticate
                    + ", so returning unauthorized credentials");

            return new AgentAuthorization(new NDRObjectTemplate(Type.Agent,
                    "unknown"), ndr, m_config, false);
        } else if (authorize == false) {
            logger.warn("Authorization is set to off, "
                    + "so returning promiscuous credentials");
            return new PromiscuousAuthInfo();
        } else {
            logger.warn("Cannot authorize without authenticating, "
                    + "so returning unauthorized credentials");
            return new AgentAuthorization(new NDRObjectTemplate(Type.Agent,
                    "unknown"), ndr, m_config, false);
        }
    }

    public void setAuthorization(String authorization) {
        if (authorization.equals("on")) {
            logger.debug("Authorization on");
            this.authorize = true;
        } else if (authorization.equals("off")) {
            logger.debug("Authorization off");
            this.authorize = false;
        } else {
            throw new RuntimeException(
                    "Expecting 'on' or 'off', but was given " + authorization);
        }
    }

    public void setAuthentication(String authentication) {
        if (authentication.equals("on")) {
            logger.debug("Authentication on");
            this.authenticate = true;
        } else if (authentication.equals("off")) {
            logger.debug("Authentication off");
            this.authenticate = false;
        } else {
            throw new RuntimeException(
                    "Expecting 'on' or 'off', but was given " + authentication);
        }
    }

    public String getAuthorization() {
        if (authorize) {
            return "on";
        } else {
            return "off";
        }
    }

    public String getAuthentication() {
        if (authenticate) {
            return "on";
        } else {
            return "off";
        }
    }

    private boolean checkAlternatePaths(HttpServletRequest request,
            PublicKey publicKey) throws AuthenticationException {
        HttpHeader header = new HttpHeader(request);
        String host;
        if (header.getValue("x-forwarded-host") != null) {
            host = header.getValue("x-forwarded-host");
        } else {
            host = request.getServerName() + ":" + request.getServerPort();
        }

        for (String redir : redirPaths) {
            if (getAuthHeader(request, host, redir).isValid(publicKey)) {
                return true;
            }
        }

        return false;
    }

    private HttpAuthHeader getAuthHeader(HttpServletRequest request,
            String host, String pathOverride) throws AuthenticationException {

        String path = request.getRequestURI().replace(serverPath, pathOverride);

        logger.debug("replacing " + serverPath + " with " + pathOverride
                + " in " + request.getRequestURI() + " producing " + path);

        return new HttpAuthHeader(new HttpHeader(request), request.getMethod()
                + " " + request.getScheme() + "://" + host + path);
    }
}
