/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.api.auth;

import java.io.IOException;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.log4j.Logger;
import org.apache.struts.action.ActionServlet;
import org.apache.struts.action.RequestProcessor;
import org.apache.struts.config.ModuleConfig;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.api.FedoraPlugIn;
import org.nsdl.repository.api.request.APIRequestProcessor;
import org.nsdl.repository.api.request.DefaultRequestProcessor;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.util.ThreadExecutionTimeAppender;

/**
 * Extension of Struts RequestProcessor for authentication processing
 * 
 * Overrides the processPreProcess method to add agent authentication as a
 * preprocess step of request processing.
 * 
 * Builds an AgentAuthInfo containing validation/authorization information, and
 * places it into an attribute in the HttpServletRequest named
 * AuthenticationGateway.AUTH_INFO_KEY
 * 
 * @author birkland
 * 
 */
public class AuthRequestProcessor extends RequestProcessor {

    private final Logger logger = Logger.getLogger(AuthRequestProcessor.class
            .getName());

    /**
     * Filter requests through the AuthenticationGateway.
     * 
     * Uses an instance of the Authentication Gateway to retrieve any agent
     * authentication information that may be present in the request
     * 
     * @param request
     *            HttpServletRequest
     * @param response
     *            HttpServletResponse
     */
    protected boolean processPreprocess(HttpServletRequest request,
            HttpServletResponse response) {

        if (!FedoraPlugIn.isActive()) {
            throw new RepositoryException("Repository is not initialized");
        }

        /* First, get the authentication info */
        AuthGateway gateway = (AuthGateway) request.getSession()
                .getServletContext().getAttribute(AuthGateway.GATEWAY_KEY);
        AgentAuthInfo authInfo = null;
        try {
            authInfo = gateway.getAuthInfo(request);
        } catch (AuthenticationException e) {
            try {
                logger.warn("Problem authenticatating request", e);
                response.sendError(HttpServletResponse.SC_UNAUTHORIZED, e
                        .getMessage());
            } catch (IOException ioe) {
                throw new RuntimeException(ioe);
            }
            return false;
        }

        request.setAttribute(AuthGateway.AUTH_INFO_KEY, authInfo);

        return true;
    }

    public void process(HttpServletRequest request, HttpServletResponse response)
            throws IOException, ServletException {
        /* Log execution time of requests */
        ThreadExecutionTimeAppender.start();
        logger.debug("Processing request" + request.getPathInfo());
        try {
            super.process(request, response);
        } finally {
            ThreadExecutionTimeAppender.finish();
        }
    }

    /**
     * Initialize the Struts and NDR request processors
     * <p>
     * Right now, this is hard-wired to instantiate the NDR
     * {@link DefaultRequestProcessor}. Configuration issues need to be
     * addressed..
     * </p>
     */
    public void init(ActionServlet servlet, ModuleConfig config)
            throws ServletException {
        super.init(servlet, config);
        APIRequestProcessor apiFilter = new DefaultRequestProcessor();

        apiFilter.init((NDRAccess) servlet.getServletContext().getAttribute(
                Constants.NDR_ACCESS));

        /* TODO: this should probably live somewhere else? */
        servlet.getServletContext().setAttribute(
                Constants.API_REQUEST_PROCESSOR, apiFilter);

    }
}
