/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth;

import java.util.Date;
import java.util.Locale;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.io.UnsupportedEncodingException;
import java.security.InvalidKeyException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.NoSuchAlgorithmException;
import java.security.SignatureException;
import java.security.Signature;

import org.apache.axis.encoding.Base64;
import org.apache.log4j.Logger;

/**
 * Signs and validates HTTP headers according to NSDL signing protocols
 * 
 * 
 * @author birkland
 * 
 */
public final class HttpAuthHeader implements MessageAuthInfo {

    private final Logger logger = Logger.getLogger(HttpAuthHeader.class
            .getName());

    private final String canonicalHeaderContent;
    private final String agent;
    private final byte[] signature;
    private final Date headerDate;

    private static final String[] CANONICAL_HEADERS = { "content-md5", "date" };
    private static final String AUTH_PROTOCOL = "nsdl-1.0";
    private static final String NSDL_HEADER_PREFIX = "x-nsdl";
    private static final String AUTH_HEADER = "x-nsdl-auth";
    private static final String SIGNATURE_ALGORITHM = "SHA1withRSA";
    private static final String CHARSET = "UTF-8";

    private static final long MAX_TIME_DEVIATION = 1000 * 60 * 10; /*
                                                                    * 10 minutes
                                                                    */
    private ThreadLocal<DateFormat> df = new ThreadLocal<DateFormat>() {
        protected DateFormat initialValue() {
            return new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss Z");
        }
    };

    /**
     * Read in header contents and signatures, if present.
     * <p>
     * Upon construction, all signature data in the given HttpHeader is parsed
     * and stored for subsequent validation.
     * </p>
     * 
     * @param header
     *            HttpHeader object representing the header to validate
     * @param request
     *            HTTP request URL string (e.g.
     *            http://ndr.nsdl.org/repositor/api/describe/)
     * @throws AuthenticationException. This
     *             exception is thrown only in the case where the header
     *             contains an NSDL signature, but some required piece of data
     *             is missing or unreadable.
     */
    public HttpAuthHeader(HttpHeader header, String request)
            throws AuthenticationException {

        /* Get the agent and signature, if present */
        String authEntry = header.getValue(AUTH_HEADER);
        if (authEntry == null) {
            this.agent = null;
            this.signature = null;
            this.canonicalHeaderContent = null;
            this.headerDate = null;
        } else {

            /* Generate the canonical header content */
            this.canonicalHeaderContent = generateCanonicalString(header,
                    request);
            /* Hack up the authentication line into the relevant bits */
            String[] authEntryParts = authEntry.split(" ");
            if (authEntryParts[0].equalsIgnoreCase(AUTH_PROTOCOL)
                    && authEntryParts[1] != null) {
                String[] sig = authEntryParts[1].split(":");

                if (sig.length == 3) {
                    this.agent = sig[0] + ":" + sig[1];
                    this.signature = Base64.decode(authEntryParts[1]
                            .replaceFirst(this.agent + ":", ""));
                } else if (sig.length == 2) {
                    this.agent = sig[0];
                    this.signature = Base64.decode(authEntryParts[1]
                            .replaceFirst(this.agent + ":", ""));
                } else {
                    throw new AuthenticationException(
                            "Malformed authorization string: '" + authEntry
                                    + "'");
                }
            } else {
                throw new AuthenticationException(
                        "Malformed authorization string: '" + authEntry + "'");
            }

            /* Get the request date */
            String dateString = null;
            try {
                if (header.getValue("x-nsdl-date") != null) {
                    dateString = header.getValue("x-nsdl-date");
                } else {
                    dateString = header.getValue("date");
                }
                this.headerDate = df.get().parse(dateString);
            } catch (Exception e) {
                throw new AuthenticationException(
                        "Could not parse header date " + dateString, e);
            }
        }
    }

    /** Returns the canonical header content as a single string */
    public String toString() {
        return this.canonicalHeaderContent;
    }

    /**
     * Returns the identity of the NSDL Agent that signed the message, or null
     * if there is none
     */
    public final String getAgent() {
        return this.agent;
    }

    /** Returns true if this header has been assigned, false otherwise */
    public final boolean isSigned() {
        return (this.agent != null && this.signature != null);
    }

    /**
     * Determines if this header has a valid signature
     * 
     * @param key
     *            Public Key that will be used to verify the message signature
     * @throws AuthenticationException
     */
    public final boolean isValid(PublicKey key) throws AuthenticationException {
        if (!isSigned()) {
            /* An unsigned message is never valid ! */
            logger.warn("Header sig is missing or invalid!");
            return false;
        }

        if (this.headerDate == null) {
            /* A header without a date is never valid */
            logger.warn("Header has no date");
            return false;
        }

        if (Math.abs(this.headerDate.getTime() - new Date().getTime()) > MAX_TIME_DEVIATION) {
            logger
                    .warn("Header date is too far from system clock, "
                            + this.headerDate.getTime() + " vs "
                            + new Date().getTime());
            return false;
        }

        try {
            Signature sig = Signature.getInstance(SIGNATURE_ALGORITHM);
            sig.initVerify(key);
            sig.update(this.canonicalHeaderContent.getBytes(CHARSET));
            boolean verification = (sig.verify(this.signature));
            if (!verification) {
                logger.info("Header failed to match signature");
                logger.trace("Failed header: \n" + canonicalHeaderContent);
            }
            return verification;
        } catch (SignatureException e) {
            throw new AuthenticationException(e);
        } catch (NoSuchAlgorithmException e) {
            throw new AuthenticationException(e);
        } catch (InvalidKeyException e) {
            throw new AuthenticationException(e);
        } catch (UnsupportedEncodingException e) {
            throw new AuthenticationException(e);
        }
    }

    /**
     * Given a HttpHeader, creates a signed version containing any required
     * header elements
     * 
     * @param header
     *            The header to sign
     * @param request
     *            The request URL as a string
     * @param agent
     *            The handle of an NSDL agent
     * @param key
     *            Private key of the NSDL agent
     * @return A signed HttpHeader
     * @throws AuthenticationException
     */
    public static HttpHeader sign(HttpHeader header, String request,
            String agent, PrivateKey key) throws AuthenticationException {
        String headerString = generateSignature(header, request, agent, key);

        return new HttpHeader(headerString + "\n" + header.toString());
    }

    /**
     * Given a HttpHeader, generates a header element string containing a valid
     * NSDL signature
     * 
     * @param header
     *            The header to sign
     * @param request
     *            The request URL as a string
     * @param agent
     *            The handle of an NSDL agent
     * @param key
     *            Private key of the NSDL agent
     * @return A string containing the an NSDL signiture
     * @throws AuthenticationException
     */
    public static String generateSignature(HttpHeader header, String request,
            String agent, PrivateKey key) throws AuthenticationException {
        String canonicalHeader = generateCanonicalString(header, request);
        Signature signature;
        try {
            signature = Signature.getInstance(SIGNATURE_ALGORITHM);
            signature.initSign(key);
            signature.update(canonicalHeader.getBytes(CHARSET));
            return AUTH_HEADER + ": " + AUTH_PROTOCOL + " " + agent + ":"
                    + Base64.encode(signature.sign());

        } catch (NoSuchAlgorithmException e) {
            throw new AuthenticationException(e);
        } catch (InvalidKeyException e) {
            throw new AuthenticationException(e);
        } catch (SignatureException e) {
            throw new AuthenticationException(e);
        } catch (UnsupportedEncodingException e) {
            throw new AuthenticationException(e);
        }
    }

    /**
     * Generates a canonical string from header content
     * 
     * @param header
     *            The HTTP header contents
     * @param request
     *            THe request URL as a string
     * @return String containing the canonical header content
     * @throws AuthenticationException
     */
    private static String generateCanonicalString(HttpHeader header,
            String request) throws AuthenticationException {

        /* First, we require either a Date or x-nsdl-date field */
        if (header.getValue("Date") == null
                && header.getValue("x-nsdl-date") == null) {
            throw new AuthenticationException("Header must contain a date");
        }
        /* Build the canonical header string */

        StringBuilder constructedHeaderContent = new StringBuilder(request
                + "\n");

        /* First, put in all the known canonical headers */
        for (int i = 0; i < CANONICAL_HEADERS.length; i++) {
            String value = header.getValue(CANONICAL_HEADERS[i]);
            if (value == null) {
                /* If it isn't defined, then just put in a newline */
                constructedHeaderContent.append("\n");
            } else {
                /* We want to trim the whitespace for consistency */
                constructedHeaderContent.append(CANONICAL_HEADERS[i] + ": "
                        + value.trim() + "\n");
            }
        }

        /* Now, get all the nsdl headers, sort them, and place them in */
        java.util.Iterator<String> keys = header.getKeys().iterator();
        java.util.ArrayList<String> nsdlKeys = new java.util.ArrayList<String>();
        while (keys.hasNext()) {
            String key = keys.next().toLowerCase(Locale.ENGLISH);
            if (key.indexOf(NSDL_HEADER_PREFIX) == 0
                    && !(key.equalsIgnoreCase(AUTH_HEADER))) {
                nsdlKeys.add(key);
            }

        }

        java.util.Collections.sort(nsdlKeys);

        /*
         * When we place nsdl values in, concatenate the key and value with a
         * ':'
         */
        for (int i = 0; i < nsdlKeys.size(); i++) {
            String key = nsdlKeys.get(i);
            if (key != null) {
                constructedHeaderContent.append(key.trim() + ": "
                        + header.getValue(key).trim() + "\n");
            }
        }

        return constructedHeaderContent.toString();
    }
}
