/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth;

import java.util.List;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Locale;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;



/** Representation of an HTTP header.
 * 
 * <p>
 * Allows access to the individual elements of an http header in a map-like fashion.
 * For example, in the following header:
 * <pre>
 * date: 2005-06-04
 * content-type: text/htnl
 * </pre>
 * <tt>date</tt> would be a key and <tt>2005-06-04</tt> would be its mapped value
 * </p>
 * @author birkland
 *
 */
public class HttpHeader {
	
	private final LinkedHashMap<String, String> header;
	
	/** Generates a HTTPHeader based on the values found in a specific
	 * servlet request
	 * 
	 * @param request HTTPservletRequest from which the header will be extracted
	 */
	
	public HttpHeader(HttpServletRequest request) {
	
	    @SuppressWarnings("unchecked")
		Enumeration e = request.getHeaderNames();
		this.header = new LinkedHashMap<String, String> ();
		
		while(e.hasMoreElements()) {
			String name = (String) e.nextElement();
			String value = request.getHeader(name);
			header.put(name.toLowerCase(Locale.ENGLISH), value);
		}
	}
	
	/** Construct an HTTPHeader from text
	 * 
	 * @param headerString String containing header text
	 */
	public HttpHeader(String headerString) {
		this.header = new LinkedHashMap<String, String>();
		String[] headerParts = headerString.split("\n");
		
		for (int i = 0; i < headerParts.length; i++) {
			String[] headerElement = headerParts[i].split(": ");
			String key = headerElement[0].toLowerCase(Locale.ENGLISH);
			StringBuilder value = new StringBuilder();
			
			/* Don't know if this is entrely necessary per spec, just being paranoid.. */
			if (headerElement.length > 2) {
				for (int j = 1; j < headerElement.length - 1; j++) {
					value.append(headerElement[j] + ": ");
				}
				value.append(headerElement[headerElement.length-1] + ": ");
			} else {
				value.append(headerElement[1]);
			}
			
			/* Handle duplicate keys (usually, it's a mistake to have duplicate 
			 * keys in the first place).  Again, just being overly cautious */
			if (this.header.get(key) != null) {
				List<String> v = java.util.Arrays.asList(this.header.get(key).split(","));
				ArrayList<String> values = new ArrayList<String>();
				for (int j=0; j < v.size(); j++) {
					values.add(v.get(j));
				}
				values.add(value.toString());
				java.util.Collections.sort(values);
				String newValue = "";
				
				for (int j = 0; j < values.size() -1; j++) {
					newValue += values.get(j) + ",";
				} newValue += values.get(values.size()-1);
				
				this.header.put(key, newValue);
			} else {
				this.header.put(key, value.toString());
			}
		}
	}
	
	public String toString() {
		String headerString = "";		
		Iterator<String> keys = header.keySet().iterator();
		
		while (keys.hasNext()) {
			String key = keys.next();
			headerString += key + ": " + header.get(key) + "\n";
		}
		return headerString;
	}
	
	/** Returns the value of a given HTTP header key.
	 * 
	 * For example, in the following header:
	 * <pre>
	 * date: 2005-06-04
	 * content-type: text/htnl
	 * </pre>
	 * <tt>date</tt> would be a key and <tt>2005-06-04</tt> would be its mapped value, 
	 * so header.getValue("date") would return "2005-06-04";
	 * @param key Text value of an HTTP header key
	 * @return it's value
	 */
	public String getValue(String key) {
		return this.header.get(key.toLowerCase(Locale.ENGLISH));
	}
	
	/** Return all the keys found in the HTTP header
	 * 
	 * @return Set of strings found
	 */
	public Set<String> getKeys() {
		return header.keySet();
	}
}
