/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth;

import java.security.KeyManagementException;
import java.security.PublicKey;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.StreamUtil;
import org.nsdl.util.security.KeyTranslator;
import org.nsdl.util.security.KeyTranslatorFactory;

import static org.nsdl.repository.model.content.AuthorizationModel.Agent.Components.PUBLIC_KEY;

/**
 * Retrieves public keys from NDR Agent object datastreams
 * 
 * @author birkland
 * 
 */
public class NDRPubKeyManager implements KeyManager {

    private final NDRAccess ndr;
    private final KeyTranslator translator;

    public NDRPubKeyManager(NDRAccess repository) {
        this.ndr = repository;
        try {
            /*
             * Right now, all keys in the NDR are stored as W3C XML. This may
             * change at some point, at which time, we need to do something more
             * intelligent here
             */
            this.translator = KeyTranslatorFactory.getInstance("RSA", "W3C");
        } catch (KeyManagementException e) {
            throw new RuntimeException("Error getting key translator", e);
        }
    }

    public PublicKey forcePubKey(String agentHandle)
            throws KeyManagementException {
        return getPubKey(agentHandle);
    }

    public PublicKey getPubKey(String agentHandle)
            throws KeyManagementException {

        NDRObject agent;
        try {
            agent = ndr.getObject(agentHandle, Type.Agent);
        } catch (ExistenceException e) {
            throw new KeyManagementException("Agent " + agentHandle
                    + " does not exist in the repository!", e);
        }

        Datastream pubKey = agent.getDatastream(PUBLIC_KEY);

        if (pubKey == null) {
            throw new KeyManagementException("Agent " + agent.handle()
                    + " does not have a public key");
        }

        try {
            return translator.decodePublicKey(StreamUtil.streamToString(pubKey
                    .getContent()));
        } catch (Throwable e) {
            throw new KeyManagementException("Could not read public key from "
                    + agent.handle(), e);
        }
    }
}
