/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth.rules;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.Defaults;
import org.nsdl.repository.api.auth.AgentAuthInfo;
import org.nsdl.repository.api.auth.AuthGateway;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.Component.Pattern;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.content.AggregatorModel;
import org.nsdl.repository.model.content.MetadataProviderModel;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.StaticObjectInfo;
import org.nsdl.repository.util.Configuration.Listener;

/**
 * Patterns:
 * <ul>
 * <li>MetadataProvider metadataProviderFor Agent</li>
 * <li>Aggregator aggregatorFor Agent</li>
 * </ul>
 * 
 * Rule:
 * <ul>
 * <li>Add</li>
 * <ul>
 * <li>Agent may add one to itself</li>
 * <li>Trusted agent may add one to another</li>
 * </ul>
 * <li>Modify</li>
 * <ul>
 * <li>Must be authorized for both subject and object, or be a trusted agent anf
 * authorized for the subject.</li>
 * </ul>
 * </ul>
 * 
 * @author birkland
 */
public class AgentRelationships implements ComponentRule, Listener {

    private static final Logger logger = Logger
            .getLogger(AgentRelationships.class.getName());

    private static final HashSet<Relationship.Pattern> patterns = new HashSet<Relationship.Pattern>(
            Arrays.asList(new Relationship.Pattern(Type.MetadataProvider,
                    MetadataProviderModel.Components.METADTA_PROVIDER_FOR,
                    Type.Agent),

            new Relationship.Pattern(Type.Aggregator,
                    AggregatorModel.Components.AGGREGATOR_FOR, Type.Agent)));

    private NDRObjectInfo m_trusted = Defaults.TRUSTED_AGENTS;

    /*
     * Add: - Agent may add one to itself - Trusted agent may add one to another
     * 
     * Modify - Must be authorized for both subject and object, or be a trusted
     * agent anf authorized for the subject.
     */
    public boolean checkCompliance(Component c, Request rq, AgentAuthInfo a) {
        Relationship r;

        logger.trace("Checking compliance");
        if (!patterns.contains(Relationship.Pattern.getPattern(c))) {
            /* We do not assert anythinng about non-matches */
            return true;
        } else {
            /* We know from the pattern that this component is a relationship */
            r = (Relationship) c;
        }

        if (rq.getRequestType().equals(Request.Type.Create)) {
            /*
             * If this is an add of a relationshup, we allow the agent itself OR
             * a trusted application to use this relationship
             */

            logger
                    .debug(("Checking permission for agent "
                            + a.agent().handle() + " to touch relationship ")
                            + r.predicate()
                            + " to agent "
                            + r.related().handle());

            boolean authorized = false;
            try {
                authorized = a.isAuthorizedFor(r.related())
                        || a.memberOf(m_trusted);
            } catch (ExistenceException e) {
                throw new RepositoryException(e.getMessage(), e);
            }

            if (!authorized) {
                String message = "Your agent " + a.agent().handle()
                        + " is not a trusted agent, nor " + "authorized for "
                        + r.related().type() + " " + r.related().handle();
                a.addMessage(message);
                logger.warn("Rule does not pass check: " + message);
            } else {
                logger.debug("passes check");
            }

            return authorized;

        } else if (rq.getRequestType().equals(Request.Type.Update)) {

            /*
             * Must be authorized for subject and object, or be a trusted agent
             * that is authorized for the subject
             */

            logger
                    .debug(("Checking permission for agent "
                            + a.agent().handle() + " to touch relationship ")
                            + r.predicate()
                            + " to existing agent "
                            + r.related().handle()
                            + " from "
                            + r.subject().type() + " " + r.subject().handle());

            boolean authorized;
            try {
                authorized = a.isAuthorizedFor(r.subject())
                        && (a.isAuthorizedFor(r.related()) || a
                                .memberOf(m_trusted));

                if (!authorized) {

                    String message = "Your agent is not a trusted agent, nor ";
                    if (!a.isAuthorizedFor(r.subject())) {
                        message += "authorized for " + r.subject().type() + " "
                                + r.subject().handle();
                    } else {
                        message += "authorized for " + r.related().type() + " "
                                + r.related().handle();
                    }

                    a.addMessage(message);
                    logger.debug("does not pass check: " + message);
                } else {
                    logger.debug("passes check");
                }
            } catch (ExistenceException e) {
                authorized = false;
                a.addMessage(r.subject().handle()
                        + " does not exist in the repository");
            }

            return authorized;
        }

        /* This relationship is OK in other api calls */
        return true;
    }

    public Set<Pattern> matchingPatterns() {
        return new HashSet<Pattern>(patterns);
    }

    public void updateConfig(Configuration config) {
        m_trusted = new StaticObjectInfo(config.getProperty(
                AuthGateway.PARAM_TRUSTED_AGENTS_AGGREGATOR,
                Defaults.TRUSTED_AGENTS.handle()), Type.Aggregator);
    }
}
