/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth.rules;

import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.auth.AgentAuthInfo;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.Component.Pattern;
import org.nsdl.repository.model.content.AggregatorModel;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;

/**
 * Patterns:
 * <ul>
 * <li>Aggregation associatedWith Resource</li>
 * </ul>
 * 
 * Rule:
 * <ul>
 * <li>UserAgent = Aggregation->AuthorizedAgent</li>
 * </ul>
 * 
 * @author birkland
 */
public class AssociationRelationships implements ComponentRule {

    private static final Logger logger =
            Logger.getLogger(AssociationRelationships.class.getName());

    private final NDRAccess ndr;

    /* We are hard-coding the relationships this class handles here */
    private Relationship.Pattern pattern =
            new Relationship.Pattern(Type.Aggregator,
                    AggregatorModel.Components.ASSOCIATED_WITH,
                    Type.Resource);

    public AssociationRelationships(NDRAccess ndr) {
        this.ndr = ndr;
    }

    public boolean checkCompliance(Component r, Request rq, AgentAuthInfo a) {
        logger.trace("Checking compliance");
        if (!pattern.matches(r)) {
            return true;
        }

        /* We only check for 'modify' operations */
        try {
            if (rq == null || Request.Type.Update.equals(rq.getRequestType())) {
                boolean auth = a.isAuthorizedFor(ndr.getObject(r.subject()));
                logger.trace("Result is " + auth);
                return auth;
            }
        } catch (ExistenceException e) {
            a.addMessage(r.subject() + "does not exist");
            logger.warn(e);
            return false;
        }
        logger.trace("Not checking, since not an update");
        return true;

    }

    public Set<Pattern> matchingPatterns() {
        HashSet<Pattern> mp = new HashSet<Pattern>();
        mp.add(pattern);
        return mp;
    }
}
