/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth.rules;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.Defaults;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.auth.AgentAuthInfo;
import org.nsdl.repository.api.auth.AuthGateway;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Component.Pattern;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.StaticObjectInfo;
import org.nsdl.repository.util.Configuration.Listener;

import static org.nsdl.repository.model.content.AuthorizationModel.Grouping.Components.AUTHORIZED;

/**
 * Rules that govern adding and modifying authorizedToChange relationships
 * 
 * 
 * @author birkland
 * 
 */
public class AuthorizationRelationships implements ComponentRule, Listener {

    private static final Logger logger = Logger
            .getLogger(AuthorizationRelationships.class.getName());

    private static final HashSet<Pattern> patterns = new HashSet<Pattern>(
            Arrays.asList(new Relationship.Pattern(Type.MetadataProvider,
                    AUTHORIZED, Type.Agent), new Relationship.Pattern(
                    Type.Aggregator, AUTHORIZED, Type.Agent)));

    private NDRObjectInfo m_trusted = Defaults.TRUSTED_AGENTS;

    public AuthorizationRelationships(NDRAccess ndr) {
    }

    public Set<Pattern> matchingPatterns() {
        return patterns;
    }

    /**
     * Determine if the given agent may touch an 'authorizedToChange'
     * relationship.
     * 
     * We allow an agent to add an 'authorizedToChange' relationship only if
     * <ul>
     * <li>It is a trusted application
     * <li>The request is an <tt> add </tt> operation,
     * <li>The trusted application makes itself or one of its aggregations the
     * object of the 'authorizedToChange' relationship
     * </ul>
     * 
     * TODO: Allow 'modify' operations where the trusted application substitutes
     * itself for one of its aggregations, and vice versa
     * 
     * @param component
     *            Presumably an 'authorizedToChange' relationship
     * @param requestContext
     *            The request from which the component originates
     * @param agentAuth
     *            Agent authorization
     */
    public boolean checkCompliance(Component component, Request requestContext,
            AgentAuthInfo agentAuth) {

        logger.trace("Checking compliance");
        Relationship r;
        if (!patterns.contains(Relationship.Pattern.getPattern(component))) {
            /* If we don't match */
            return true;
        } else {
            r = (Relationship) component;
        }

        logger.debug("Checking permission of agent "
                + agentAuth.agent().handle() + " to touch relationship "
                + r.subject().handle() + " " + r.predicate() + " "
                + r.related().handle());

        /* First, see if we're a member of trusted agents */
        if (agentAuth.memberOf(m_trusted)) {

            return true;

        } else {

            logger.warn(agentAuth.agent().handle()
                    + " is not a trusted app, but "
                    + "tried to meddle with an authorizedToChange "
                    + "relationship");
            agentAuth.addMessage("Only trusted applications may add or modify "
                    + "authorizedToChange relationships");
        }

        return false;
    }

    public void updateConfig(Configuration config) {
        m_trusted = new StaticObjectInfo(config.getProperty(
                AuthGateway.PARAM_TRUSTED_AGENTS_AGGREGATOR,
                Defaults.TRUSTED_AGENTS.handle()), Type.Aggregator);
    }

}
