/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth.rules;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.auth.AgentAuthInfo;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.Component.Pattern;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.Configuration.Listener;

/**
 * Default Set of rules applied to every API request
 * 
 * Collecton of all the core NSDL request rules that are applied to every
 * request
 * 
 * @author birkland
 * 
 */
public class DefaultRuleSet implements RuleSet {

    private static final Logger logger = Logger.getLogger(DefaultRuleSet.class
            .getName());
    /* The master rule collection */
    private final HashMap<Pattern, Set<ComponentRule>> rules;

    /*
     * We currently pick all of our rule classes here, and load each one into
     * the master rule collection in the constructor;
     */

    /** Creates a rule set populated with a default set of rules. */
    public DefaultRuleSet(NDRAccess ndr, Configuration config) {
        this.rules = new HashMap<Pattern, Set<ComponentRule>>();

        /*
         * We currently pick all of our rule classes here, and load each one
         * into the master rule collection in the constructor;
         */
        ComponentRule[] initialRules = { new AgentRelationships(),
                new AssociationRelationships(ndr), new MemberOfRelationships(),
                new MetadataProvisionRelationships(ndr),
                new AuthorizationRelationships(ndr), };
        for (int i = 0; i < initialRules.length; i++) {
            if (initialRules[i] instanceof Listener) {
                ((Listener) initialRules[i]).updateConfig(config);
            }
            addRule(initialRules[i]);
        }
    }

    public boolean matches(Component c) {

        /*
         * XXX: Right now, we just interate through all the patterns this set
         * contains. There's probably a smarter way, but we'll get to that when
         * it becomes a problem
         */
        for (Pattern p : rules.keySet()) {
            if (p.matches(c)) {
                return true;
            }
        }
        return false;
    }

    public boolean matches(Pattern r) {
        return rules.containsKey(r);
    }

    public boolean allows(Component c, Request rq, AgentAuthInfo a) {
        /*
         * XXX: Again, there's probably a smarter way to so this than just
         * iterating through all patterms, but this works for the time being
         */
        logger.trace("Checking allowance of component " + c);
        for (Pattern p : rules.keySet()) {
            if (p.matches(c)) {
                for (ComponentRule r : this.rules.get(p)) {
                    logger.trace("Checking rule: " + r.getClass().getName());
                    if (r.checkCompliance(c, rq, a) == false) {
                        logger.debug("Failed on " + c + "\n with pattern " + p
                                + "\n");
                        return false;
                    }
                }
            }
        }
        /*
         * If all rules passed, or no rules matched, this rule set allows the
         * given rule
         */
        logger.trace("PASS");
        return true;
    }

    /**
     * Adds a rule to this RuleSet.
     * 
     * Note: This method does <b>not</b> check for redundancy. Therefore, adding
     * a duplicate rule will ultimately end in that rule being evaluated twice.
     * 
     * @param rule
     *            The rule to add to this set
     */
    public void addRule(ComponentRule rule) {
        for (Pattern p : rule.matchingPatterns()) {
            addRule(p, rule);
        }
    }

    /**
     * Adds a rule to this RuleSet, given a specific matching pattern.
     * 
     * Explicitely states that a given tule is to be inserted into the RuleSet,
     * and marked as having a particular pattern. This pattern will be used for
     * matching in the matches() methods. Any other patterns that this rule
     * matches will be ignored unless explicitly given.
     * 
     * @param pattern
     *            The pattern the rule is to match
     * @param rule
     *            The rule to add to this set
     */
    public void addRule(Pattern pattern, ComponentRule rule) {

        if (pattern == null || rule == null) {
            throw new NullPointerException(
                    "A pattern and a rule must both be given");
        }

        if (!this.rules.containsKey(pattern)) {
            this.rules.put(pattern, new HashSet<ComponentRule>());
        }

        this.rules.get(pattern).add(rule);
    }
}
