/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth.rules;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.api.auth.AgentAuthInfo;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.Component.Pattern;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.CommonModel.Components.MEMBER_OF;

/**
 * Patterns:
 * <ul>
 * <li>Resource memberOf Aggregation</li>
 * <li>Aggregation memberOf Aggregation</li>
 * <li>Agent memberOf Aggregation</li>
 * </ul>
 * 
 * Rule:
 * <ul>
 * <li>UserAgent = Aggregation->AuthorizedAgent</li>
 * </ul>
 * 
 * @author birkland
 */
public class MemberOfRelationships implements ComponentRule {

    private static final Logger logger =
            Logger.getLogger(MemberOfRelationships.class.getName());

    private static final HashSet<Relationship.Pattern> patterns =
            new HashSet<Relationship.Pattern>(Arrays.asList(

            new Relationship.Pattern(Type.Resource, MEMBER_OF,
                    Type.Aggregator),

            new Relationship.Pattern(Type.Aggregator, MEMBER_OF,
                    Type.Aggregator),

            new Relationship.Pattern(Type.Agent, MEMBER_OF,
                    Type.Aggregator)

            ));

    public boolean checkCompliance(Component c, Request rq, AgentAuthInfo a) {

        logger.trace("Checking compliance");
        Relationship r;
        if (!patterns.contains(Relationship.Pattern.getPattern(c))) {
            return true;
        } else {
            r = (Relationship) c;
        }

        logger.debug("Checking permision of " + a.agent().handle() + " to add"
                + " memberOf to " + r.related().handle());
        if (rq.getRequestType().equals(Request.Type.Create)
                || rq.getRequestType().equals(Request.Type.Update)) {

            boolean authorized = false;
            try {
                authorized = a.isAuthorizedFor(r.related());
            } catch (ExistenceException e) {
                throw new RepositoryException(e);
            }

            if (authorized) {
                logger.debug("passes check");
            } else {
                String message =
                        "Agent " + a.agent().handle() + " is not "
                                + "authorized for " + r.related().type() + " "
                                + r.related().handle();
                a.addMessage(message);
                logger.warn(message);
            }
            return authorized;
        }
        return true;
    }

    public Set<Pattern> matchingPatterns() {
        return new HashSet<Pattern>(patterns);
    }
}
