/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth.rules;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.auth.AgentAuthInfo;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.Component.Pattern;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.MetadataModel.Components.METADATA_FOR;
import static org.nsdl.repository.model.content.MetadataModel.Components.METADATA_PROVIDED_BY;

/**
 * Rules for dealing with Metadata relationships
 * 
 * <p>
 * Patterns:
 * <ul>
 * <li>Metadata metadataFor Resource</li>
 * <li>Metadata metadataFor Aggregation</li>
 * <li>Metadata metadataFor Agent</li>
 * <li>Metadata metadataProvidedBy MetadataProvider</li>
 * </ul>
 * </p>
 * 
 * <p>
 * Rule:
 * <ul>
 * <li>UserAgent = Metadata->MetadataProvider->AuthorizedAgent</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 */
public class MetadataProvisionRelationships implements ComponentRule {
    private static final Logger logger =
            Logger.getLogger(MetadataProvisionRelationships.class.getName());

    private static final Relationship.Pattern metadataProvidedBy =
            new Relationship.Pattern(Type.Metadata, METADATA_PROVIDED_BY,
                    Type.MetadataProvider);

    private final NDRAccess ndr;

    public MetadataProvisionRelationships(NDRAccess ndr) {
        this.ndr = ndr;
    }

    /*
     * We currently just accept all relatitonships that match these patterns.
     * Translation: We allow anybody to add metadata about anything
     */
    private static final HashSet<Relationship.Pattern> patterns =
            new HashSet<Relationship.Pattern>(Arrays.asList(
                    new Relationship.Pattern(Type.Metadata, METADATA_FOR,
                            Type.Resource),

                    new Relationship.Pattern(Type.Metadata, METADATA_FOR,
                            Type.Aggregator),

                    new Relationship.Pattern(Type.Metadata, METADATA_FOR,
                            Type.Agent),

                    metadataProvidedBy

            ));

    public boolean checkCompliance(Component c, Request rq, AgentAuthInfo a) {
        logger.trace("BEGIN checking compliance");
        Relationship r;
        if (!metadataProvidedBy.matches(c)) {
            /*
             * All relationships that match 'metadataFor' are automatically
             * compliant, so no need to check explicitly. We just lump them with
             * all the non-matching relationships
             */
            return true;
        } else {
            r = (Relationship) c;
        }

        /* We only protect add and modify commands */
        if (!(Request.Type.Create.equals(rq.getRequestType())
                || Request.Type.Update.equals(rq.getRequestType()) || Request.Type.Delete
                .equals(rq.getRequestType()))) {
            return true;
        }

        logger.trace("Checking rule for " + r.predicate() + " "
                + r.related().handle());

        /*
         * If this object already exists in the repository, we need to check its
         * *original* contents to see if we can modify it
         */
        if (rq.getRequestType().equals(Request.Type.Update)) {

            Relationship metadataProvByRel;
            try {
                metadataProvByRel =
                        ndr.getObject(c.subject()).getRelationship(
                                metadataProvidedBy.predicate());
                if (metadataProvByRel != null) { /* This object exists in NDR */
                    if (!a.isAuthorizedFor(ndr.getObject(metadataProvByRel
                            .related()))) {
                        String message =
                                "Agent " + a.agent().handle() + " is not "
                                        + "authorized to change "
                                        + metadataProvByRel.subject().type()
                                        + " "
                                        + metadataProvByRel.subject().handle()
                                        + " because it is "
                                        + " not authorized for "
                                        + metadataProvByRel.related().type()
                                        + " "
                                        + metadataProvByRel.related().handle();

                        logger.warn(message);
                        a.addMessage(message);
                        return false;
                    }
                }

            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        /*
         * Now, make sure that we are authorized to add stuff to the
         * metadataProvider specified in this relationship
         */
        boolean authorized = false;
        try {
            logger.trace("Checking to see if agent is authorized to"
                    + " modify the contents of " + r.related().handle());
            authorized = a.isAuthorizedFor(r.related());
        } catch (ExistenceException e) {
            throw new RuntimeException(e.getMessage(), e);
        }

        if (!authorized) {
            String message =
                    "Agent " + a.agent().handle() + " is not "
                            + "authorized to change " + r.related().type() + " "
                            + r.related().handle();

            logger.warn(message);
            a.addMessage(message);
        } else {
            logger.debug("passes check");
        }

        return authorized;
    }

    public Set<Pattern> matchingPatterns() {
        return new HashSet<Pattern>(patterns);
    }

}
