/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.request;

import java.util.HashSet;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.request.filters.MetadataReferenceFilter;
import org.nsdl.repository.api.request.filters.URINormalizationFilter;

/**
 * Instantiates and filters {@link Request}s based on HTTP submitted data
 * <p>
 * Right now, the default request processor is hardwired to dispatch
 * {@link XMLRequest} requests (i.e. NDR API v1.0), and apply the standard NSDL
 * filters. This can/should be easily changed to have its behaviour defined by
 * configuration files, when the time is necessary
 * </p>
 * 
 * @author birkland
 * 
 * @since 1.0
 */
public class DefaultRequestProcessor implements APIRequestProcessor {

    private static final Logger logger =
            Logger.getLogger(DefaultRequestProcessor.class);

    private Class<?> requestClass;

    private Set<RequestFilter> filters = new HashSet<RequestFilter>();

    private NDRAccess ndr;

    /**
     * Set up initial filters and prepare for use
     * 
     * @param ndrAccess
     *            access to the repository
     */
    public void init(NDRAccess ndrAccess) {
        ndr = ndrAccess;
        try {

            /* For now, we just bake this in (but it's ripe for a changin') */
            requestClass =
                    Class.forName("org.nsdl.repository.api.request.XMLRequest");
            logger.debug("Initializing with request class "
                    + requestClass.getName());
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

        initFilters();
    }

    /**
     * Get a filtered XMLRequest from the given HTTP req.
     * 
     * @param httpRequest
     *            Should contain raw HTTP request and associated data
     */
    public Request getRequest(HttpServletRequest httpRequest) {
        try {

            /* Instantiate and build the request from the given HTTP source */
            InitializableRequest rq =
                    (InitializableRequest) requestClass.newInstance();
            rq.init(httpRequest);

            return filterRequest(rq);
        } catch (Exception e) {
            logger.warn("Could not instantate request:" + e.getMessage());
            throw new RuntimeException(e);
        }
    }

    private Request filterRequest(Request r) {
        /* Wrap the request in a mutator and apply filters */
        MutableRequest processRequest = new MutableRequestWrapper(r);
        for (RequestFilter f : filters) {
            f.process(processRequest);
        }

        return processRequest;
    }

    /**
     * Create and initialize a baked list of NSDL filters
     * <p>
     * TODO: make this configurable
     * </p>
     */
    private void initFilters() {
        filters.add(new MetadataReferenceFilter());
        filters.add(new URINormalizationFilter());

        for (RequestFilter f : filters) {
            f.init(ndr);
        }
    }
}
