/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.request;

import javax.xml.namespace.QName;

import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;

/**
 * Allows the modification of requests
 * <p>
 * Typically employed in pre-processing stages, modified by
 * {@link RequestFilter}s. Mutable requests allow the request to be modified in
 * the following ways:
 * <ul>
 * <li>Add or remove a component mentioned under a particular command (for
 * example, add a DC datastream under a 'replace' command for modified Metadata
 * objects)</li>
 * <li>Change the object referenced in the request</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 * 
 */
public interface MutableRequest extends Request {

    /**
     * Set the object referenced in the request
     * <p>
     * Replaces any existing object info reference in the query with the given
     * value.
     * </p>
     * 
     * @param info
     *            Object identity that will replace whatever is present
     */
    public void setObjectInfo(NDRObjectInfo info);

    /**
     * Add a property to the request under a given command
     * 
     * @param command
     *            Command that applies to the added property
     * @param p
     *            Property to add
     */
    public void putProperty(QName command, Property p);

    /**
     * Remove a property from the request under a given command
     * 
     * @param command
     *            Command under which the removed property is found
     * @param p
     *            Particular poperty to remove
     */
    public void removeProperty(QName command, Property p);

    /**
     * Add a relationship to the request under a given command
     * 
     * @param command
     *            Command that applies to the added relationship
     * @param r
     *            Relationship to add
     */
    public void putRelationship(QName command, Relationship r);

    /**
     * Remove a relationship from the request under a given command
     * 
     * @param command
     *            Command under which the removed property is found
     * @param r
     *            Particular relationship to remove
     */
    public void removeRelationship(QName command, Relationship r);

    /**
     * Add a datastream to the request under a given command
     * 
     * @param command
     *            Command that applies to the added datastream
     * @param d
     *            Datastream to add
     */
    public void putDatastream(QName command, Datastream d);

    /**
     * Remove a datastream from the request under a given command
     * 
     * @param command
     *            Command under which the remoed datastream is found
     * @param d
     *            Particular datastream to remove
     */
    public void removeDatastream(QName command, Datastream d);

}
