/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.request;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;

public class MutableRequestWrapper implements MutableRequest {

    private static final Logger logger =
            Logger.getLogger(MutableRequestWrapper.class);
    private HashMap<QName, Set<Relationship>> relationships =
            new HashMap<QName, Set<Relationship>>();
    private HashMap<QName, Set<Property>> properties =
            new HashMap<QName, Set<Property>>();
    private HashMap<QName, Set<Datastream>> datastreams =
            new HashMap<QName, Set<Datastream>>();

    private String request;
    private NDRObjectInfo object;
    private Request.Type requestType;

    private Request origRequest;

    public MutableRequestWrapper(Request r) {
        init(r);
    }

    protected MutableRequestWrapper() {
        /* So subclasses can initialize in whatever way they choose */
    }

    protected final void init(Request r) {
        this.request = r.getRequest();
        this.object = r.getObjectInfo();
        this.requestType = r.getRequestType();
        this.origRequest = r;

        /* Populate from original request */
        for (QName command : Request.Commands.values()) {
            Set<Property> origProperties = origRequest.getProperties(command);
            Set<Relationship> origRelationships = origRequest.getRelationships(command);
            Set<Datastream> origDatastreams = origRequest.getDatastreams(command);
            
            if (!origProperties.isEmpty()) {
                logger.trace("Found initial properties for command " + command);
                properties.put(command, origProperties);
            }
            if (!origDatastreams.isEmpty()) {
                logger
                        .trace("Found initial datastreams for command "
                                + command);
                datastreams.put(command, origDatastreams);
            }
            if (!origRelationships.isEmpty()) {
                logger.trace("Found initial relationships for command "
                        + command);
                relationships.put(command, origRelationships);
            }
        }
    }

    protected final void setRequest(String requestString) {
        this.request = requestString;
    }

    protected final void setRequestType(Request.Type type) {
        this.requestType = type;
    }

    public final void setObjectInfo(NDRObjectInfo o) {
        this.object = o;
    }

    public final void putDatastream(QName command, Datastream d) {
        Set<Datastream> existing = datastreams.get(command);
        if (existing == null) {
            datastreams.put(command, new HashSet<Datastream>());
            existing = datastreams.get(command);
        }

        for (Datastream ex : existing) {
            if (d.getName().equals(ex.getName())) {
                throw new IllegalStateException("Cannot add duplicate "
                        + "datastream " + d.getName() + " to request");
            }
        }

        logger.trace("Putting datastream " + command + " " + d.getName());
        existing.add(d);
    }

    public final void putProperty(QName command, Property p) {
        Set<Property> existing = properties.get(command);
        if (existing == null) {
            properties.put(command, new HashSet<Property>());
            existing = properties.get(command);
        }

        logger.trace("Putting property " + command + " " + p.predicate());
        existing.add(p);
    }

    public final void putRelationship(QName command, Relationship r) {
        Set<Relationship> existing = relationships.get(command);
        if (existing == null) {
            relationships.put(command, new HashSet<Relationship>());
            existing = relationships.get(command);
        }

        logger.trace("Putting relationship " + command + " " + r.predicate());
        existing.add(r);
    }

    public final void removeDatastream(QName command, Datastream d) {
        Set<Datastream> existing = datastreams.get(command);
        if (existing == null) {
            return;
        }

        for (Datastream ex : existing) {
            if (ex.getName().equals(d.getName())) {
                logger
                        .trace("Removing datastream " + command + " "
                                + d.getName());
                existing.remove(d);
                return;
            }
        }
    }

    public final void removeProperty(QName command, Property p) {
        Set<Property> existing = properties.get(command);
        if (existing == null) {
            return;
        }

        logger.trace("Removing property " + command + " " + p.predicate());
        existing.remove(p);
    }

    public final void removeRelationship(QName command, Relationship r) {
        Set<Relationship> existing = relationships.get(command);
        if (existing == null) {
            return;
        }

        logger.trace("Removing relationship " + command + " " + r.predicate());
        existing.remove(r);
    }

    public final Set<QName> getCommands() {
        HashSet<QName> commands = new HashSet<QName>();
        commands.addAll(relationships.keySet());
        commands.addAll(properties.keySet());
        commands.addAll(datastreams.keySet());

        return commands;
    }

    public final Set<Component> getComponents() {
        HashSet<Component> components = new HashSet<Component>();

        /* Add all properties */
        for (Set<Property> pset : properties.values()) {
            for (Property p : pset) {
                components.add(p);
            }
        }

        /* Add all relationships */
        for (Set<Relationship> rset : relationships.values()) {
            for (Relationship r : rset) {
                components.add(r);
            }
        }

        /* XXX: Need to make datastreams components! */
        return components;
    }

    public final Set<Datastream> getDatastreams(QName command) {
        HashSet<Datastream> d = new HashSet<Datastream>();
        if (datastreams.get(command) != null) {
            d.addAll(datastreams.get(command));
        }
        return d;
    }

    public final NDRObjectInfo getObjectInfo() {
        return this.object;
    }

    public final Set<Property> getProperties(QName command) {
        HashSet<Property> p = new HashSet<Property>();
        if (properties.get(command) != null) {
            p.addAll(properties.get(command));
        }
        return p;
    }

    public final Set<Relationship> getRelationships(QName command) {
        HashSet<Relationship> r = new HashSet<Relationship>();
        if (relationships.get(command) != null) {
            r.addAll(relationships.get(command));
        }
        return r;
    }

    public final String getRequest() {
        return this.request;
    }

    public final Type getRequestType() {
        return this.requestType;
    }

    /* XXX We do this for now... */
    public List<String> getParameter(String key) {
        List<String> values = new ArrayList<String>();
        if (this.origRequest != null) {
            values.addAll(origRequest.getParameter(key));
        }

        return values;
    }

}
