/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.request;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.xml.namespace.QName;

import org.nsdl.repository.api.Constants;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.w3c.dom.Node;

/**
 * Representation of an NDR API request
 * 
 * INCOMPLETE TODO: finish this interface
 * 
 * @author birkland
 * 
 */
public interface Request {

    /**
     * Get the name of the API request. (e.g. "addAggregator", "get",
     * "describe")
     */
    String getRequest();

    /**
     * Get the type of request.
     * <p>
     * A given request may be classified as a Create, Read, Update, or Delete.
     * This method returns the classification of the request as determined by
     * this Requests implementation
     * </p>
     * 
     * @return Request type.
     */
    Type getRequestType();

    /**
     * Return information on the object being acted upon by this request
     * <p>
     * Some requests (such as Reads, Updates, or Deletes) act upon a single
     * object. If such an object is specified in the request, this method
     * returns NDRObjectInfo describing that object. Otherwise, if no single
     * object is specified in the request, this method will return null;
     * </p>
     * 
     */
    NDRObjectInfo getObjectInfo();

    /**
     * Get all properties under a specific command.
     * <p>
     * Due to legacy reasons, some properties (resource identifiers, in
     * particular) in the request are not directly analogous to the actual RDF
     * triples they represent. For example, resource identifiers in the request
     * are identified as
     * <tt>&lt;identifier type=XYZ&gt;id&lt;/identifier&gt;</tt>, yet exist
     * in RDF as <tt>&lt;hasResourceXYZ&gt;id&lt;/hasResourceXYZ&gt;</tt>.
     * Thus, we specify that app Properties returned by this method have already
     * been translated into their actual Triple/RDF forms.
     * </p>
     * <p>
     * In addition to possibly translating the form of properties, all returned
     * Properties will be presented in their native storage namespace,
     * regardless of what is specified in the incoming request. For example,
     * relationships that appear in the request in
     * <tt>http://ns.nsdl.org/ndr/request_v1.00/</tt> will be returned as
     * existing in <tt>http://ns.nsdl.org/api/relationships#</tt>
     * </p>
     * 
     * @param command
     *            Qualified name of a command in the request
     * @return properties that match
     */
    Set<Property> getProperties(QName command);

    /**
     * Get all relationships under a specific command
     * <p>
     * All returned are in their native namespace for storage, regardless of
     * what has been specified int the request. For example, relationships that
     * appear in the request in <tt>http://ns.nsdl.org/ndr/request_v1.00/</tt>
     * will be returned as existing in
     * <tt>http://ns.nsdl.org/api/relationships#</tt>
     * </p>
     * 
     * @param command
     *            Qualified name of a command in the request
     * @return relationships that match
     */
    Set<Relationship> getRelationships(QName command);

    /**
     * Get all datastreams under a specific command
     * <p>
     * Nonstandard datastream identifiers in the request will have been
     * translated to their native form. For example, metadata is specified in
     * the request as <tt>&lt;format type=XYZ&gt;</tt>, which actually
     * corresponds to a datastream with the ID <tt>format_XYZ</tt>.
     * <p>
     * 
     * @param command
     * @return Set of matching datastreams.
     */
    Set<Datastream> getDatastreams(QName command);

    Set<QName> getCommands();

    /** Get all components from a request */
    Set<Component> getComponents();

    /**
     * Get a specific request parameter
     * 
     * @param key
     *            Name of the parameter
     * @return The parameter value, null if not defined
     */
    public List<String> getParameter(String key);

    /**
     * Default NSDL API request commands
     * 
     * These commands represent actions to be performed on particular request
     * components (Relationships, Properties, Datastreams).
     * 
     * TODO: make command not a QName
     * 
     * @author birkland
     * 
     */
    public static class Commands {

        private static final HashMap<String, QName> commands =
                new HashMap<String, QName>();
        static {
            commands.put("add", new QName(Constants.REQUEST_NAMESPACE, "add"));
            commands.put("delete", new QName(Constants.REQUEST_NAMESPACE,
                    "delete"));
            commands.put("modify", new QName(Constants.REQUEST_NAMESPACE,
                    "modify"));
            commands.put("match", new QName(Constants.REQUEST_NAMESPACE,
                    "match"));
        }

        /** Qualified name of the "add" command */
        public static final QName ADD = commands.get("add");

        /** Qualifled name of the "delete" command */
        public static final QName DELETE = commands.get("delete");

        /** Qualified name of the "modify" command */
        public static final QName REPLACE = commands.get("modify");

        /** Qualified name of the "match" command */
        public static final QName MATCH = commands.get("match");

        public static boolean isCommand(QName candidate) {
            return commands.containsValue(candidate);
        }

        public static boolean isCommand(Node candidate) {
            QName candidateName =
                    new QName(candidate.getNamespaceURI(), candidate
                            .getLocalName());
            return commands.containsValue(candidateName);
        }

        public static Set<QName> values() {
            return new HashSet<QName>(commands.values());
        }

    }

    /**
     * Basic request type or class
     * <p>
     * Every API request can be classified as performing one of the following
     * classes of action:
     * <ul>
     * <li>Create: Adds a new object or objects to the repository</li>
     * <li>Read: Reads, queries, or reports on an object</li>
     * <li>Update: Modifies the contents of one or more objects</li>
     * <li>Delete: Deletes/purges an object/objects entirely</li>
     * </ul>
     * This enum, therefore, is an enumeration of these four types
     * </p>
     */
    public static enum Type {
        Create, Read, Update, Delete
    }
}
