/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.api.request;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.xml.namespace.QName;

import org.nsdl.repository.api.Constants;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Collection of convenience methods for dealing with API input
 * 
 * Some of the methods were moved from FedoreUtils and modified to support
 * commands
 * 
 * @author birkland, cornwell, cramer
 * 
 */
public abstract class RequestUtils {

    /**
     * Represents the three components of request XML: properties,
     * relationships, and datastreams
     */
    private static enum Component {

        /** Contains all relationships to other objects */
        relationships("relationships"),

        /** Contains all properties (literals) */
        properties("properties"),

        /** Contains all datastream declarations */
        datastreams("data");

        private final QName elementName;

        private Component(String tagName) {
            this.elementName = new QName(Constants.REQUEST_NAMESPACE, tagName);
        }

        /**
         * Get the qualified name for the element representing this component in
         * api input XML
         * 
         * @return Qualified element name
         */
        public QName xmlName() {
            return elementName;
        }
    }

    /**
     * Get relationships from api input XML
     * 
     * Searches the given input XML node for the the 'relationships' components
     * (As defined in the core API model) When that is found, it returns results
     * as follows:
     * <ul>
     * <li>If a command element is given, all its children are included in the
     * returned result
     * <li>If the given command is designated by a true 'isDefault' value, all
     * elements that are relationships but <em>not</em> wrapped in a command are
     * given
     * <li>If no command is given, all elements are returned.
     * </ul>
     * 
     * @param input
     *            input Node that contains the given relationships element
     *            somewhere
     * @param command
     *            element (representing a command) that contains desired
     *            relationships
     * @param isDefault
     *            True if the given command is a 'default' command
     * @return List of matching nodes
     */
    public static List<Node> getInputRelationships(Node input, QName command,
            boolean isDefault) {
        return getInputComponents(input, Component.relationships.xmlName(),
                command, isDefault);
    }

    /**
     * Get properties from api input XML
     * 
     * Searches the given input XML node for the the 'properties' components (As
     * defined in the core API model) When that is found, it returns results as
     * follows:
     * <ul>
     * <li>If a command element is given, all its children are included in the
     * returned result
     * <li>If the given command is designated by a true 'isDefault' value, all
     * elements that are properties but <em>not</em> wrapped in a command are
     * given
     * <li>If no command is given, all elements are returned.
     * </ul>
     * 
     * @param input
     *            input Node that contains the given properties element
     *            somewhere
     * @param command
     *            element (representing a command) that contains desired
     *            properties
     * @param isDefault
     *            True if the given command is a 'default' command
     * @return List of matching nodes
     */
    public static List<Node> getInputProperties(Node input, QName command,
            boolean isDefault) {
        return getInputComponents(input, Component.properties.xmlName(),
                command, isDefault);
    }

    /**
     * Get datastreams from api input XML
     * 
     * Searches the given input XML node for the the 'datastreams' components
     * (As defined in the core API model) When that is found, it returns results
     * as follows:
     * <ul>
     * <li>If a command element is given, all its children are included in the
     * returned result
     * <li>If the given command is designated by a true 'isDefault' value, all
     * elements that are datastreams but <em>not</em> wrapped in a command are
     * given
     * <li>If no command is given, all elements are returned.
     * </ul>
     * 
     * @param input
     *            input Node that contains the given datastreams element
     *            somewhere
     * @param command
     *            element (representing a command) that contains desired
     *            datastreams
     * @param isDefault
     *            True if the given command is a 'default' command
     * @return List of matching nodes
     */
    public static List<Node> getInputDatastreams(Node input, QName command,
            boolean isDefault) {
        return getInputComponents(input, Component.datastreams.xmlName(),
                command, isDefault);
    }

    /**
     * Get members of a particular input XML component
     * 
     * Searches the given input XML node for the specified component element.
     * When that is found, it returns results as follows:
     * <ul>
     * <li>If a command element is given, all its children are included in the
     * returned result
     * <li>If the given command is designated by a true 'isDefault' value, all
     * elements that are in the component but <em>not</em> wrapped in a command
     * are given
     * <li>If no command is given, all elements are returned.
     * </ul>
     * 
     * @param inputXML
     *            input Node that contains the given 'component' element
     *            somewhere
     * @param component
     *            element that contains the the components we are fetching (In
     *            the core model, these are properties, datastreams, or
     *            relationships)
     * @param command
     *            element (representing a command) that contains desired
     *            components
     * @param isDefault
     *            True if the given command is a 'default' command
     * @return List of matching nodes
     */
    public static List<Node> getInputComponents(Node inputXML, QName component,
            QName command, boolean isDefault) {
        ArrayList<Node> results = new ArrayList<Node>();

        if (inputXML == null) {
            return results;
        }
        Node componentContainer = NDRXMLutil.getNodeByName(inputXML, component);
        if (componentContainer == null) {
            return results;
        }

        NodeList components = componentContainer.getChildNodes();
        for (int i = 0; i < components.getLength(); i++) {
            if (components.item(i).getNodeType() != Node.ELEMENT_NODE)
                continue;

            Element child = (Element) components.item(i);
            if (Request.Commands.isCommand(child)) {
                /* This element is a command wrapper */

                QName elementName = new QName(child.getNamespaceURI(), child
                        .getLocalName());

                if (command == null || elementName.equals(command)) {
                    /*
                     * We want all elements if the command matches or is
                     * unspecified
                     */
                    results.addAll(getInputComponents(child, elementName, null,
                            false));
                }

            } else if (command == null || isDefault) {
                /*
                 * Default or unspecified commands pick up all non-wrapped
                 * elements
                 */
                results.add(components.item(i));
            }
        }
        return results;
    }

    /**
     * Translate relationships in inputXML to NDR-normal form
     * <p>
     * Translates inputXML nomenclature ('request' namespace into the correct
     * internal NDR namespace, as well as relationship names that are different
     * in the inputXML than the internal representation
     * </p>
     * 
     * @param relationships
     *            Collection of relationship elements gathered from request XML
     * @return List of translated relationships.
     * @throws Exception
     */
    public static List<Node> translateRelationships(
            Collection<Node> relationships) throws Exception {
        ArrayList<Node> rels = new ArrayList<Node>();
        for (Node relationship : relationships) {
            String value = NDRXMLutil.getElementValue((Element) relationship);

            String namespace = relationship.getNamespaceURI();
            String prefix = relationship.getPrefix();

            /* We will translate request namespace */
            if (namespace.equals(Constants.REQUEST_NAMESPACE)) {
                namespace = Constants.REPOS_NAMESPACE;
                prefix = Constants.REPOS_NAMESPACE_PREFIX;
            }

            /* We're creating a new node here so we don't muck with the input */
            Node relsExtRelationship = relationship.getOwnerDocument()
                    .createElementNS(namespace, relationship.getLocalName());
            relsExtRelationship.setPrefix(prefix);

            NDRXMLutil.setElementValue((Element) relsExtRelationship, value);

            rels.add(relsExtRelationship);
        }
        return rels;
    }

    public static ApiRequestPath getPath(HttpServletRequest request) {
        return new ParsedRequestPath(request);
    }

    public interface ApiRequestPath {
        public boolean hasHandle();

        public String getHandle();

        public String getApiCall();

        public boolean hasPath();

        public String[] getPath();
    }

    public static class ParsedRequestPath implements ApiRequestPath {
        private final String m_apiCall;
        private final String m_handle;
        private final String[] m_path;

        public ParsedRequestPath(HttpServletRequest request) {
            String[] parts = request.getPathInfo().replaceFirst("^\\/", "")
                    .split("/");
            m_apiCall = parts[0];

            if (parts.length == 2) {
                m_handle = parts[1];
                m_path = new String[0];
            } else if (parts.length > 2 && parts[1].contains(":")) {
                String handle = parts[1];
                /* XXX Hack for slash encoding madness */
                if (handle.contains("hdl:") && !handle.contains("%2F")) {
                    handle += "%2F" + parts[2];
                    m_path = slurp(parts, 3);
                } else {
                    m_path = slurp(parts, 2);
                }

                m_handle = handle;
            } else if (parts.length >= 3) {
                String trialHandle = parts[1] + "/" + parts[2];
                if (trialHandle.matches("^[0-9]+\\/.+?")) {
                    m_handle = trialHandle;
                    m_path = slurp(parts, 3);
                } else {
                    throw new RuntimeException(
                            "Expecting to find a handle or PID in "
                                    + request.getPathInfo());
                }
            } else {
                m_handle = null;
                m_path = new String[0];
            }
        }

        public boolean hasHandle() {
            return m_handle != null;
        }

        public String getHandle() {
            return m_handle;
        }

        public String getApiCall() {
            return m_apiCall;
        }

        public boolean hasPath() {
            return m_path.length > 0;
        }

        public String[] getPath() {
            return m_path;
        }

        private final String[] slurp(String[] input, int startIndex) {
            ArrayList<String> slurped = new ArrayList<String>();

            for (int i = startIndex; i < input.length; i++) {
                slurped.add(input[i]);
            }

            return slurped.toArray(new String[0]);
            // Java 6 only!
            // return Arrays.copyOfRange(input, startIndex, input.length);
        }
    }

}
