/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.api.request;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import javax.servlet.http.HttpServletRequest;
import javax.xml.namespace.QName;

import org.apache.commons.fileupload.FileItem;
import org.apache.commons.fileupload.disk.DiskFileItemFactory;
import org.apache.commons.fileupload.servlet.ServletFileUpload;
import org.apache.log4j.Logger;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.api.request.RequestUtils.ApiRequestPath;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.impl.BasicDatastream;
import org.nsdl.repository.util.NDRObjectTemplate;
import org.nsdl.repository.util.StaticObjectInfo;
import org.nsdl.repository.util.StreamUtil;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.traversal.DocumentTraversal;
import org.w3c.dom.traversal.NodeIterator;

/**
 * NDR API request - inputXML format 1.0
 * <p>
 * Responsible for interpreting NDR API requests and expressing them in the NDR
 * model.
 * </p>
 * 
 * @author birkland
 * 
 */
public class XMLRequest implements Request, InitializableRequest {

    private static final Logger logger = Logger.getLogger(XMLRequest.class
            .getName());

    private static final List<String> addRequests = new ArrayList<String>();

    private static final List<String> modifyRequests = new ArrayList<String>();

    private Map<String, List<String>> params = new HashMap<String, List<String>>();

    private static final ServletFileUpload m_file_upload = new ServletFileUpload(
            new DiskFileItemFactory());

    /* Instead of listing all add/modify request types, generate them */
    static {
        addRequests.add("add");
        addRequests.add("modify");
        for (org.nsdl.repository.model.types.Type<?> type : org.nsdl.repository.model.types.Type
                .values()) {
            addRequests.add("add" + type);
            modifyRequests.add("modify" + type);
        }
    }

    private Map<String, FileItem> m_content;

    private Node requestNode;

    private NDRObjectInfo object;

    private String apiCall;

    private HttpServletRequest m_request;

    /**
     * Create an interpreted NDR APi request.
     * 
     * @param request
     *            HTTPServletRequest containing an NDR API request with an
     *            'inputXML' xml parameter.
     */
    public XMLRequest(HttpServletRequest request) {
        this.init(request);
    }

    /**
     * Create an unitialized XMLRequest.
     * <p>
     * Request must be initialized with {@link #init(HttpServletRequest)} before
     * being used.
     * </p>
     */
    public XMLRequest() {

    }

    /**
     * Create an API request from information present in a HttpServletRequest
     * 
     * <p>
     * If there is a specific object acted upon in the request (e.g.
     * /get/2200/12345T), it will be used as the requests object. If one does
     * not exist yet, such as in the case of Adds, an empty object will be
     * created.
     * </p>
     * 
     * @param request
     *            HTTP servlet request to extract info from.
     * @throws ExistenceException
     *             if either the request object or an object declared in a
     *             relationship does not exist
     */
    @SuppressWarnings("unchecked")
    public void init(HttpServletRequest request) throws ExistenceException {

        m_request = request;
        m_content = getAttachedContent(request);

        try {
            this.requestNode = getRequestNode(request);
        } catch (Exception e) {
            /* Just wrap in a runtime exception */
            throw new RepositoryException(e);
        }

        logger.debug("Got request path: " + request.getPathInfo());

        ApiRequestPath requestPath = RequestUtils.getPath(request);
        this.apiCall = requestPath.getApiCall();

        if (requestPath.hasHandle()) {
            this.object = new StaticObjectInfo(requestPath.getHandle());
        } else {
            if (requestNode != null) {
                NDRObjectTemplate e = new NDRObjectTemplate(
                        org.nsdl.repository.model.types.Type
                                .getInstance(requestNode.getLocalName()));
                logger.debug("No specific request object.  Using anonymous "
                        + e.type());
                this.object = e;
            } else {
                logger.debug("No request object or type given");
            }
        }

        params = new HashMap<String, List<String>>();

        /* Ugg.. httpservletrequest is not generic */
        /* XXX: consequences of binary attachments? */
        for (Map.Entry<String, String[]> e : (Set<Entry<String, String[]>>) request
                .getParameterMap().entrySet()) {

            /* This bit just for debugging.. */
            String paramValues = "\n";
            for (String value : e.getValue()) {
                paramValues += value + "\n";
            }
            logger.debug("Got parameter '" + e.getKey() + "' with values "
                    + paramValues);

            params.put(e.getKey(), Arrays.asList(e.getValue()));
        }
    }

    /** {@inheritDoc} */
    public NDRObjectInfo getObjectInfo() {
        return this.object;
    }

    /** {@inheritDoc} */
    public String getRequest() {
        return this.apiCall;
    }

    /*
     * (non-Javadoc)
     * 
     * Current implementation of this is pretty simplistic: determining requst
     * type by the request name. This is a bit fragile, so at some point, will
     * move on to some sore of API call registration/config where each API call
     * is registered as belonging to a particular type. This works for now,
     * thoigh
     * 
     * @see org.nsdl.repository.api.Request#getRequestType()
     */
    public Request.Type getRequestType() {
        if (this.apiCall.startsWith("add")) {
            return Request.Type.Create;
        } else if (this.apiCall.startsWith("modify")) {
            return Request.Type.Update;
        } else if (this.apiCall.startsWith("delete")) {
            return Request.Type.Delete;
        }
        return Request.Type.Read;
    }

    public Set<Relationship> getRelationships(QName command) {
        HashSet<Relationship> rels = new HashSet<Relationship>();
        List<Node> requestRels = null;

        logger.debug("Getting relationships.  Command: " + command
                + " default: " + isDefault(command));
        try {
            requestRels = RequestUtils.translateRelationships(RequestUtils
                    .getInputRelationships(requestNode, command,
                            isDefault(command)));

        } catch (Exception e) {
            e.printStackTrace();
            throw new RuntimeException(
                    "Error retrieving request relationships", e);
        }

        for (Node r : requestRels) {
            Relationship.Name predicate = new Relationship.Name(r
                    .getNamespaceURI(), r.getLocalName(), r.getPrefix());

            NDRObjectInfo relatedObject = new StaticObjectInfo(NDRXMLutil
                    .getElementValue((Element) r));

            rels.add(new Relationship(this.object, predicate, relatedObject));
        }

        return rels;
    }

    public Set<Property> getProperties(QName command) {
        HashSet<Property> properties = new HashSet<Property>();

        logger.debug("Finding properties.  Command: " + command + " default: "
                + isDefault(command));
        for (Node p : RequestUtils.getInputProperties(requestNode, command,
                isDefault(command))) {
            String ns = p.getNamespaceURI();
            String localName = p.getLocalName();
            String prefix = p.getPrefix();

            /* Translate namespace */
            if (ns.equals(Constants.REQUEST_NAMESPACE)) {
                ns = Constants.REPOS_NAMESPACE;
                prefix = "nsdl";
            }

            /* handle nonstandard 'identifier' syntax */
            if (ns.equals(Constants.REPOS_NAMESPACE)
                    && localName.equals("identifier")) {
                String type = NDRXMLutil.getAttributeValue((Element) p, null,
                        "type").toUpperCase(Locale.ENGLISH);
                localName = "hasResource" + type;
            }

            logger.debug("  Found property: " + ns + localName + " "
                    + p.getTextContent());
            /* Build the Property */
            Property.Name predicate = new Property.Name(ns, localName, prefix);
            properties.add(new Property(getObjectInfo(), predicate, p
                    .getTextContent()));
        }

        return properties;
    }

    public Set<Datastream> getDatastreams(QName command) {
        HashSet<Datastream> datastreams = new HashSet<Datastream>();

        logger.debug("Finding datastreams.  Command: " + command + " default "
                + isDefault(command));
        for (Node p : RequestUtils.getInputDatastreams(requestNode, command,
                isDefault(command))) {

            /* First, take care of any "special case" processing */
            if (isSpecialCaseDatastreamDef(p)) {
                logger.trace("Found special case datastream def");
                for (Datastream d : getSpecialCaseDatastreams(p)) {
                    logger.trace("Adding special case datastream "
                            + d.getName().getID());
                    datastreams.add(d);
                }
                continue;
            }

            /*
             * Get the ID of the datastream. If the ID attribute is explicitly
             * set, use this. Otherwise, use the local enclosing element name.
             */
            Datastream.Name id = null;
            String xmlid = NDRXMLutil
                    .getAttributeValue((Element) p, null, "ID");
            if (xmlid == null) {
                id = new Datastream.Name(p.getLocalName());
            } else {
                id = new Datastream.Name(xmlid);
            }

            /*
             * Mimetype may not be specified in the request, (since it is not
             * required) so expect a null here from time to time
             */
            String mimeType = NDRXMLutil.getAttributeValue((Element) p, null,
                    "MIMETYPE");

            /*
             * The enclosing tag defines the datastream. Its content in the API
             * is either: 1) More XML elements - That XML is the datastream
             * content 2) A URL - The URL contains the datastream content 3)
             * Text - The enclosed text is the datastream content 3) Nothing -
             * The datastream has zero content, or it resides as a "file" in an
             * attached multi-part form
             */

            if (NDRXMLutil.getFirstChildElement((Element) p) != null) {
                declareNS((Element) p);
                logger.trace("Adding xml datastream " + id);

                Datastream toAdd = new XMLDatastream(id, p, mimeType);
                String mime = ((Element) p).getAttribute("MIMETYPE");
                if (mime != null) {
                    toAdd.setMimeType(mime);
                }

                String label = ((Element) p).getAttribute("LABEL");
                if (label != null) {
                    toAdd.setLabel(label);
                }
                datastreams.add(toAdd);
            } else if (NDRXMLutil.getElementValue((Element) p) != null) {
                logger.trace("Adding redirect datastream " + id);
                String value = NDRXMLutil.getElementValue((Element) p);

                try {
                    URL url = new URL(value);
                    Datastream toAdd = new BasicDatastream(id);
                    toAdd.setURL(url);
                    datastreams.add(toAdd);

                    String mime = ((Element) p).getAttribute("MIMETYPE");
                    if (mime != null) {
                        toAdd.setMimeType(mime);
                    }

                    String label = ((Element) p).getAttribute("LABEL");
                    if (label != null) {
                        toAdd.setLabel(label);
                    }
                } catch (MalformedURLException e) {
                    throw new IllegalArgumentException("Could not parse url "
                            + value);
                }
            } else {
                logger.trace("Adding attached datastream");

                FileItem fp = m_content.get(id.getID().toLowerCase());

                if (fp != null) {
                    datastreams.add(new FileItemDatastream(id, fp, mimeType));
                } else {

                    /* Last ditch, use any given parameters */;
                    Datastream ds = new BasicDatastream(id);
                    ds.setContent(StreamUtil.stringToStream(m_request
                            .getParameter(id.getID())));

                    String mime = ((Element) p).getAttribute("MIMETYPE");
                    if (mime != null) {
                        ds.setMimeType(mime);
                    }

                    String label = ((Element) p).getAttribute("LABEL");
                    if (label != null) {
                        ds.setLabel(label);
                    }

                    datastreams.add(ds);
                }
            }
        }

        return datastreams;
    }

    /** {@inheritDoc} */
    public Set<Component> getComponents() {
        Set<Component> ret = new HashSet<Component>();

        for (QName command : Request.Commands.values()) {
            ret.addAll(getRelationships(command));
        }

        return ret;
    }

    public Set<QName> getCommands() {
        HashSet<QName> commands = new HashSet<QName>();

        for (QName command : Request.Commands.values()) {
            Node commandExample = NDRXMLutil.getNodeByName(requestNode, command
                    .getNamespaceURI(), command.getLocalPart());
            if (commandExample != null) {
                commands.add(command);
            }
        }

        return commands;
    }

    public List<String> getParameter(String key) {
        List<String> values = new ArrayList<String>();
        if (params.get(key) != null) {
            values.addAll(params.get(key));
        }
        return values;
    }

    private boolean isDefault(QName command) {
        if (addRequests.contains(this.apiCall)) {
            /* For 'add' requests, the default command is 'add' */
            return Request.Commands.ADD.equals(command);
        } else if (this.apiCall.startsWith("find")) {
            return Request.Commands.MATCH.equals(command);
        } else {
            /* For every other request, the default command is 'modify' */
            return Request.Commands.REPLACE.equals(command);
        }
    }

    private Node getRequestNode(HttpServletRequest request) throws Exception {
        String inputXML = request.getParameter("inputXML");
        if (inputXML == null) {
            if (m_content.containsKey("inputxml")) {
                inputXML = m_content.get("inputxml").getString();
            } else {
                inputXML = "<inputXML xmlns=\"http://ns.nsdl.org/ndr/request_v1.00/\" />";
            }
        }
        return NDRXMLutil.getFirstChildElement(NDRXMLutil.getDocument(inputXML,
                true, false).getDocumentElement());
    }

    /* Right now, 'format' metadata datastreams are the only special case */
    private boolean isSpecialCaseDatastreamDef(Node def) {
        if (def.getLocalName().equals("format")
                && NDRXMLutil.getAttributeValue((Element) def, null, "type") != null) {

            logger.trace("Special case found in " + def.getLocalName());
            return true;
        } else {
            return false;
        }
    }

    private static Set<Datastream> getSpecialCaseDatastreams(Node def) {
        if (def.getLocalName().startsWith("format_")) {
            return getSingleStyleMetadataDs(def);
        } else if (def.getLocalName().startsWith("format")) {
            return getBlockStyleDatastreams(def);
        } else {
            throw new RuntimeException("Could not identify special case for "
                    + def.getLocalName());
        }
    }

    @SuppressWarnings("unchecked")
    private static Map<String, FileItem> getAttachedContent(HttpServletRequest r) {
        HashMap<String, FileItem> content = new HashMap<String, FileItem>();

        if (ServletFileUpload.isMultipartContent(r)) {
            try {
                List<FileItem> attached = m_file_upload.parseRequest(r);
                for (FileItem f : attached) {
                    content.put(f.getFieldName().toLowerCase(), f);
                }

                /* For the benefit of struts */
                Enumeration<String> names = r.getAttributeNames();
                while (names.hasMoreElements()) {
                    Object attr = r.getAttribute(names.nextElement());
                    if (attr instanceof FileItem) {
                        FileItem f = (FileItem) attr;
                        content.put(f.getFieldName().toLowerCase(), f);
                    }
                }
            } catch (Exception e) {
                throw new RuntimeException("Could not parse content", e);
            }
        }
        return content;
    }

    private static Set<Datastream> getSingleStyleMetadataDs(Node def) {
        Set<Datastream> specialCases = new HashSet<Datastream>();
        Datastream.Name id = new Datastream.Name(def.getLocalName());
        String mime = "application/xml";
        specialCases.add(new XMLDatastream(id, NDRXMLutil
                .getFirstChildElement((Element) def), mime));
        return specialCases;
    }

    private static Set<Datastream> getBlockStyleDatastreams(Node def) {
        Set<Datastream> specialCases = new HashSet<Datastream>();

        String type = NDRXMLutil.getAttributeValue((Element) def, null, "type");

        /* Just being paranoid */
        if (type == null) {
            throw new NullPointerException("'type' is null");
        }

        Datastream.Name id = new Datastream.Name("format_" + type);
        String mime = "application/xml";

        Element meta = NDRXMLutil.getChildElement((Element) def,
                Constants.REQUEST_NAMESPACE, "meta");
        Element info = NDRXMLutil.getChildElement((Element) def,
                Constants.REQUEST_NAMESPACE, "info");

        declareNS(NDRXMLutil.getFirstChildElement(meta));
        XMLDatastream metaDS = new XMLDatastream(id, NDRXMLutil
                .getFirstChildElement(meta), mime);
        specialCases.add(metaDS);

        if (info != null) {
            declareNS(info);
            Datastream.Name info_id = new Datastream.Name(id.getID() + "_info");
            XMLDatastream infoDs = new XMLDatastream(info_id, info, mime);
            specialCases.add(infoDs);
        }

        return specialCases;
    }

    private static void declareNS(Element element) {
        if (element == null)
            return;

        NodeIterator dsNodes = ((DocumentTraversal) element.getOwnerDocument())
                .createNodeIterator(element, Node.ELEMENT_NODE, null, false);
        Map<String, String> namespacesToAdd = new HashMap<String, String>();
        Map<String, String> declaredNamespaces = new HashMap<String, String>();

        /* Go through all elements in this XML datastream */
        Node next = dsNodes.nextNode();
        while (next != null) {

            /* Remember any namespaces that have been explicitly declared */
            NamedNodeMap attributes = next.getAttributes();
            for (int i = 0; i < attributes.getLength(); i++) {
                Node attribute = attributes.item(i);
                if (attribute.getNodeName().contains("xmlns:")) {
                    String prefix = attribute.getNodeName().substring(6);
                    String ns = attribute.getNodeValue();
                    declaredNamespaces.put(prefix, ns);
                } else if (attribute.getNamespaceURI() != null
                        && attribute.getPrefix() != null) {

                    /* Check and declare the attribute's ns, if not declared */
                    if (!declaredNamespaces.values().contains(
                            attribute.getNamespaceURI())) {
                        namespacesToAdd.put(attribute.getPrefix(), attribute
                                .getNamespaceURI());
                        declaredNamespaces.put(attribute.getPrefix(), attribute
                                .getNamespaceURI());
                    }
                }
            }

            /*
             * If this element has a ns that hasnot been declared, remember it
             */
            if (next.getNamespaceURI() != null && next.getPrefix() != null) {
                if (!declaredNamespaces.values().contains(
                        next.getNamespaceURI())) {
                    namespacesToAdd.put(next.getPrefix(), next
                            .getNamespaceURI());
                    declaredNamespaces.put(next.getPrefix(), next
                            .getNamespaceURI());
                }
            }
            next = dsNodes.nextNode();
        }

        /* Add all undeclared namespaces */
        for (Map.Entry<String, String> e : namespacesToAdd.entrySet()) {
            NDRXMLutil
                    .declareNamespace(element, e.getValue(), e.getKey(), true);
        }
    }

    static class XMLDatastream extends BasicDatastream {
        private final String m_xml;

        public XMLDatastream(Name id, Node content, String mime) {
            super(id);
            if (mime == null) {
                setMimeType("application/xml");
            } else {
                setMimeType(mime);
            }

            try {
                if (content == null) {
                    m_xml = "";
                    logger.warn("XML datastream " + id + " has no content");
                } else {
                    m_xml = NDRXMLutil.serializeDOMElement((Element) content)
                            .replaceAll("<\\?.+\\?>\n", "");
                }
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }

        @Override
        public InputStream getContent() {
            return StreamUtil.stringToStream(m_xml);
        }

        @Override
        public void setContent(InputStream i) {
            throw new UnsupportedOperationException("Datastream is read-only");
        }

        @Override
        public boolean isCanonical() {
            return true;
        }
    }

    static class FileItemDatastream implements Datastream {
        private final FileItem m_file;
        private final Datastream.Name m_name;
        private String m_mime;
        private URL m_url;

        public FileItemDatastream(Datastream.Name name, FileItem f,
                String mimeType) {
            m_name = name;
            if (f == null) {
                throw new NullPointerException(
                        "File content was null for datastream " + name.getID());
            }
            m_file = f;
            m_mime = mimeType;
        }

        public Datastream.Name getName() {
            return m_name;
        }

        public String getMimeType() {
            if (m_mime != null) {
                return m_mime;
            } else if (m_file != null) {
                String mime = m_file.getContentType();
                if (mime.indexOf(';') > 0) {
                    return mime.substring(0, mime.indexOf(';'));
                } else {
                    return mime;
                }
            }

            return null;
        }

        public long getSize() {
            if (m_file != null) {
                return m_file.getSize();
            } else {
                return 0;
            }
        }

        public URL getURL() {
            return m_url;
        }

        public void setURL(URL u) {
            m_url = u;
        }

        public InputStream getContent() {
            if (m_file != null) {
                try {
                    return m_file.getInputStream();
                } catch (IOException e) {
                    throw new RuntimeException("Could not read ds content", e);
                }
            } else {
                logger.warn("Attached datastream content for " + m_name
                        + " is empty");
                return new ByteArrayInputStream(new byte[0]);
            }
        }

        public InputStream getContent(String x) {
            return getContent();
        }

        public void setContent(InputStream content) {
            throw new UnsupportedOperationException("Content is read only");
        }

        public void setMimeType(String mime) {
            m_mime = mime;
        }

        public boolean isCanonical() {
            return true;
        }

        public String getLabel() {
            return m_file.getName();
        }

        public void setLabel(String name) {
            throw new UnsupportedOperationException("File name read-only");
        }
    }
}
