/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.request.filters;

import java.util.Collection;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.request.MutableRequest;
import org.nsdl.repository.api.request.RequestFilter;
import org.nsdl.repository.api.request.Request.Commands;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.content.ModelUtil;
import org.nsdl.repository.model.content.ResourceModel;
import org.nsdl.repository.model.content.Scope;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.api.request.Request.Type.Create;

/**
 * Handles adding of duplicate resources.
 * 
 * <p>
 * Looks for 'create' requests for new resorces. If the resource has an
 * identifier defined, it checks to see if that resource identifier already
 * exists. If so, then the object info of the request will be modified to point
 * to the existing resource
 * </p>
 * 
 * @author birkland
 * @since 1.0
 */
public class AddResourceFilter implements RequestFilter {

    private static final Logger logger =
            Logger.getLogger(AddResourceFilter.class);
    private NDRAccess ndr;

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        this.ndr = ndr;
    }

    /**
     * Will determine if a request for a new resource maps to an existing
     * resource.
     * <p>
     * Will search the NDR for a resource that has the same identifier as the
     * one given. If one exists, it will insert that resource's data into the
     * request's objectInfo.
     * </p>
     * 
     * @param r
     *            The request for a new resource
     */
    public void process(MutableRequest r) {

        /* If this is not a create on a resource, leave */
        if (!Create.equals(r.getRequestType())
                || !Type.Resource.equals(r.getObjectInfo().type())) {
            return;
        }

        String identifier = "";
        Property.Name idType =
                new Property.Name("none_defined_yet", "none", "no");

        Collection<Property.Name> resourceIdentifiers =
                ModelUtil.getPropertyNames(ResourceModel.Identifiers.class,
                        Scope.All);

        for (QName command : r.getCommands()) {

            /* We only care about add or replace properties */
            if (!Commands.ADD.equals(command)
                    || Commands.REPLACE.equals(command)) {
                continue;
            }

            /* Find the identifier! */
            for (Property p : r.getProperties(command)) {
                if (resourceIdentifiers.contains(p.predicate())) {

                    identifier = p.value();
                    idType = p.predicate();
                    continue;
                }
            }
        }

        NDRObject existing =
                ndr.getObjectFinder().findResource(idType, identifier);

        if (existing != null) {
            logger.info("Request to add redundant resource.  Using "
                    + existing.handle() + " instead");

            /* XXX: Should we be setting the request type to 'update'? */
            r.setObjectInfo(existing);
        }
    }
}
