/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.api.request.filters;

import java.util.HashSet;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.request.MutableRequest;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.api.request.RequestFilter;
import org.nsdl.repository.api.request.Request.Commands;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.StreamUtil;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import static org.nsdl.repository.model.content.MetadataModel.Components.*;
import static org.nsdl.repository.model.content.ResourceModel.Components.IDENTIFIER_URL;

/**
 * Derives 'references' relationships from metadata content.
 * <p>
 * Looks for <tt>nsdl_dc</tt> formatted datastreams in a Metadata request
 * containing <code>&lt;dct:references&gt;</code> fields. If present, and if it
 * contains a URL of an object in the NDR, it will create a 'references'
 * relationship in the Metadata object which points to that resource.
 * </p>
 * <p>
 * If the Metadata object in the request is is being modified to remove or alter
 * any <tt>&lt;dct:references&gt;</tt> fields, this filter removes existing or
 * adds new 'references' relationships as necessary.
 * </p>
 * 
 * @author birkland
 * @author cornwell
 * 
 */
public class MetadataReferenceFilter implements RequestFilter {
    private NDRAccess ndr;

    private static final String DCT = "http://purl.org/dc/terms/";

    private static final Logger logger = Logger
            .getLogger(MetadataReferenceFilter.class);

    /**
     * Create/remove references relationships based on nsdl_dc content.
     * <p>
     * Requests that aren't adds or updates of Metadata objects are ignored. For
     * metadata requests containing nsdl_dc payload, 'references' relationships
     * will be created or deleted based on the presence (or lack, thereof) of a
     * dct:references element
     * </p>
     * 
     * @param r
     *            Request to analyze for metadata content Request to be analyzed
     *            for referenced content
     */
    public void process(MutableRequest r) {

        if (r.getObjectInfo() == null) {
            return;
        }
        /* If the request object isn't Metadata, we're done */
        if (!Type.Metadata.equals(r.getObjectInfo().type()))
            return;

        if (Request.Type.Create.equals(r.getRequestType())) {
            addDerived(r);
        } else if (Request.Type.Update.equals(r.getRequestType())) {
            updateDerived(r);
        }
    }

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        this.ndr = ndr;
    }

    /**
     * Find a format_nsdl_dc datastream and return its container element.
     * 
     * @param datastreams
     *            Datastreams to look through.
     * @return Node containing nsdl_dc.
     */
    private static Node getNsdlDC(Iterable<Datastream> datastreams) {
        Datastream ds = null;
        for (Datastream d : datastreams) {
            if (d.getName().equals(NSDL_DC))
                ds = d;
        }

        if (ds == null)
            return null;

        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory
                    .newInstance();
            factory.setNamespaceAware(true);
            factory.setValidating(false);

            DocumentBuilder builder = factory.newDocumentBuilder();

            return builder.parse(ds.getContent()).getDocumentElement();
        } catch (Exception e) {
            throw new RuntimeException("Error parsing content of "
                    + ds.getName() + " (" + ds.getClass() + ") :\n"
                    + StreamUtil.streamToString(ds.getContent()), e);
        }
    }

    /**
     * Read through nsdl_dc and find elements that look like 'references'.
     * 
     * @param metadata
     *            Node containing nsdl_dc elements within.
     * @return Set of URLs that could possibly be references to NDR objects.
     */
    private static Set<String> getImpliedReferences(Node metadata) {
        HashSet<String> implied = new HashSet<String>();

        NodeList references = ((Element) metadata).getElementsByTagNameNS(DCT,
                "references");

        for (int i = 0; i < references.getLength(); i++) {
            Node reference = references.item(i);

            if (reference != null) {
                String val = NDRXMLutil.getElementValue((Element) reference);
                if (val != null && !"".equals(val)) {
                    implied
                            .add(NDRXMLutil
                                    .getElementValue((Element) reference));
                }
            }
        }
        return implied;
    }

    private void addDerived(MutableRequest r) {
        /* Get the DC metadata */
        Node dc = getNsdlDC(r.getDatastreams(Request.Commands.ADD));
        if (dc == null) {
            /* Last ditch, see if user did 'replace' for some reason */
            dc = getNsdlDC(r.getDatastreams(Request.Commands.REPLACE));
        }

        /* If we don't have any nsdl_dc, we can't continue */
        if (dc == null) {
            logger.info("No nsdl_dc metadata found in Metadata ");
            return;
        }

        /* Add all annotations, if they can be added */
        for (String id : getImpliedReferences(dc)) {
            addReference(id, r, null);
        }
    }

    private void updateDerived(MutableRequest r) {
        NDRObject thisMetadata = ndr.getObject(r.getObjectInfo());
        /* We need to be a bit more careful with updates.. */
        Node dc;

        /* If we're adding nsdl_dc (unlikely), then it's easy */
        dc = getNsdlDC(r.getDatastreams(Request.Commands.ADD));
        if (dc != null) {
            for (String id : getImpliedReferences(dc)) {
                addReference(id, r, thisMetadata);
            }
            return;
        }

        /* If we're removing nsdl_dc (unlikely), that's easy too */
        dc = getNsdlDC(r.getDatastreams(Request.Commands.DELETE));
        if (dc != null) {
            for (String id : getImpliedReferences(dc)) {
                removeReference(id, r);
            }
            return;
        }

        /*
         * If we're modifying, we need to remove no-longer-present annotations,
         * and add new ones (this is the most common case)
         */
        dc = getNsdlDC(r.getDatastreams(Request.Commands.REPLACE));
        if (dc != null) {
            NDRObject existing = ndr.getObject(r.getObjectInfo());

            Set<String> givenReferenceURLs = getImpliedReferences(dc);
            Set<String> existingReferenceURLS = new HashSet<String>();

            /* Get all existing annotation URLs */
            for (Relationship rel : existing.getRelationships(REFERENCES)) {
                try {
                    existingReferenceURLS.add(ndr.getObject(rel.related())
                            .getProperty(IDENTIFIER_URL).value());
                } catch (AccessException e) {
                    /* Should never really happen */
                    throw new RuntimeException("Too many Identifier URLs");
                }
            }

            /* Remove annotations in the object but implied by nsdl_dc */
            for (String url : existingReferenceURLS) {
                if (!givenReferenceURLs.contains(url)) {
                    removeReference(url, r);
                }
            }

            /* Add annotations implied by nsdl_dc, but not in object */
            for (String url : givenReferenceURLs) {
                if (!existingReferenceURLS.contains(url)) {
                    addReference(url, r, thisMetadata);
                }
            }
        }
    }

    private void addReference(String identifier, MutableRequest r,
            NDRObject existing) {
        NDRObject referencedResource = ndr.getObjectFinder().findResource(
                IDENTIFIER_URL, identifier);
        if (referencedResource == null) {
            logger.info("Not referencing resource '" + identifier
                    + "'.  It does not exist in the NDR!");
        } else {

            Relationship toAdd = new Relationship(r.getObjectInfo(),
                    REFERENCES, referencedResource);

            if (existing != null
                    && existing.getRelationships(REFERENCES).contains(toAdd)) {
                /* do nothing, the relationship is already there */
                return;
            } else {
                logger.trace("Adding 'references' to "
                        + referencedResource.handle());
                r.putRelationship(Commands.ADD, toAdd);
            }
        }
    }

    private void removeReference(String identifier, MutableRequest r) {
        NDRObject referencedResource = ndr.getObjectFinder().findResource(
                IDENTIFIER_URL, identifier);
        if (referencedResource == null) {
            logger.warn("Cannot de-reference resource '" + identifier
                    + "'.  It does not exist in the NDR!");
        } else {
            logger.trace("Commissioning delete of  'references' to  "
                    + referencedResource.handle());
            r.putRelationship(Commands.DELETE, new Relationship(r
                    .getObjectInfo(), REFERENCES, referencedResource));
        }
    }
}
