/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.request.filters;

import javax.xml.namespace.QName;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.request.MutableRequest;
import org.nsdl.repository.api.request.RequestFilter;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Property;
import org.nsdl.util.uri.NormalizedURL;

import static org.nsdl.repository.model.content.ResourceModel.Components.IDENTIFIER_URL;

/**
 * Normalize any resource URLs in the request.
 * <p>
 * This filter will normalize all URLs given in requests involving resources.
 * Since there may be many equivalent URLs based on encoding rules, trailing
 * slashes, etc, this will make sure that for every set of equivalent URLs, the
 * repository will only deal with a single 'canonical' version.
 * </p>
 * 
 * @author birkland
 * @since 1.0
 */
public class URINormalizationFilter implements RequestFilter {

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        /* Not needed */
    }

    /**
     * Canonicalize resource URLs, if necessary.
     * <p>
     * Will look for a URL identifier in the request. If it is not in canonical
     * form, this filter will replace it with the canonical version.
     * 
     * @param r
     *            Request that may contain URLs to normalize.
     */
    public void process(MutableRequest r) {
        for (QName command : r.getCommands()) {
            for (Property p : r.getProperties(command)) {
                if (IDENTIFIER_URL.equals(p.predicate())) {
                    fixURL(r, command, p);
                }
            }
        }

        /* XXX: Should we worry about URL parameters? */
    }

    private static void fixURL(MutableRequest r, QName command, Property p) {
        if (p.value() == null)
            return;
        try {
            NormalizedURL url = new NormalizedURL(p.value());
            url.normalize();
            Property normalizedURL =
                    new Property(p.subject(), p.predicate(), url.toString());
            r.removeProperty(command, p);
            r.putProperty(command, normalizedURL);
        } catch (Exception e) {
            throw new RepositoryException("Could not normalize URL "
                    + p.value(), e);
        }
    }

}
