/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.handle;

import java.util.HashMap;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.filters.HandleManagementFilter.HandleException;
import org.nsdl.repository.access.filters.HandleManagementFilter.HandleFactory;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.Configuration.Listener;

/**
 * HandleFactory that manages handles in the CNRI handle system. 
 * <h2>Configuration Parameters</h2>
 * 
 * <p>
 * <dl>
 * <dt><code>param.repository.handle.url</code></dt>
 * </dd> Default URL base that will be registered with CNRI when creating a new
 * handle. For example, if the url base is <code>http://nsdl.org/object/</code>
 * and a new handle is <code>2200/200710151301345T</code>, then the URL
 * registered with the handle service is
 * <code>http://nsdl.org/object/2200/200710151301345T</code>.
 * </dl>
 * </p>
 * 
 * <p>
 * <dl>
 * <dt><code>param.repository.handle.url.&lt;type&gt;</code></dt>
 * </dd> URL base that will be registered with CNRI when creating a new handle
 * for an object of the given type. For example, if the configuration is
 * <code>param.repository.handle.url.resource = http://nsdl.org/resource/</code>
 * and a new handle is <code>2200/200710151301345T</code>, then the URL
 * registered with the handle service is
 * <code>http://nsdl.org/resource/2200/200710151301345T</code>.
 * </dl>
 * </p>
 * 
 * @author birkland
 * @since 1.2
 */
public class CNRIHandleFactory implements HandleFactory, Listener {

    private static final Logger logger = Logger
            .getLogger(CNRIHandleFactory.class);

    private static final String PROP_HANDLE_URL = "param.repository.handle.url";

    private final int MAX_HANDLE_ATTEMPTS = 3;

    private HandleManager m_handleManager;
    private HashMap<Type<NDRObject>, String> m_baseUrls;

    public CNRIHandleFactory() {
    }

    /** Simply try (at most) three times from server */
    public String getHandle(NDRObject o) throws HandleException {
        Exception he = null;
        for (int current_try = 0; current_try < MAX_HANDLE_ATTEMPTS; current_try++) {
            try {
                String hdl = m_handleManager.generateHandle();
                String url = m_baseUrls.get(o.type()) + hdl;
                m_handleManager.createHandleWithURL(hdl, url);
                return hdl;
            } catch (net.handle.hdllib.HandleException e) {
                he = e;
                logger.debug("Attempt to assign handle failed, code "
                        + e.getCode() + " (" + e.getMessage()
                        + "), trying again");
            }
        }

        throw new HandleException("Could not assign handle after "
                + MAX_HANDLE_ATTEMPTS + " attempts", he);
    }

    public void deleteHandle(String handle) throws HandleException {
        if (handle != null) {
            try {
                m_handleManager.deleteHandle(handle);
            } catch (Exception e) {
                throw new HandleException("Could not delete handle", e);
            }
        }
    }

    public void updateConfig(Configuration config) {
        m_handleManager = new HandleManager(config);
        String base = config.getProperty(PROP_HANDLE_URL);
        if (base == null) {
            throw new RuntimeException("No handle url base found. "
                    + "Looking for property " + PROP_HANDLE_URL);
        } else {
            base = base.trim();
        }

        m_baseUrls = new HashMap<Type<NDRObject>, String>();

        for (Type<NDRObject> t : Type.values()) {
            String url = config.getProperty(PROP_HANDLE_URL + "."
                    + t.toString().toLowerCase());
            if (url == null) {
                url = base;
            }

            m_baseUrls.put(t, url);
        }
    }
}
