/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.handle;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import org.nsdl.repository.error.RepositoryException;

/**
 * Should work with any JDBC 3.0 compliant driver that supports
 * Statement.RETURN_GENERATED_KEYS
 * 
 * Expects a table such as the following:
 * 
 	CREATE TABLE handleGenerator (
 		localName INT NOT NULL AUTO_INCREMENT, 
		namingAuthority VARCHAR(64), PRIMARY KEY (localName))
 * 
 * @author Edwin Shin
 */

public class DBHandleGenerator implements HandleGenerator {
    private final static String DB_TABLE = "handleGenerator";
    private final static String DB_NA = "namingAuthority";
    //private final static String DB_ID = "localName";

    private Connection conn;
    private PreparedStatement insertStmt;
    private ResultSet rs;

    /**
     * 
     * @param driverClass e.g. "com.mysql.jdbc.Driver"
     * @param dbURL e.g. "jdbc:mysql://localhost/test?user=monty&password=greatsqldb"
     */
    public DBHandleGenerator(String driverClass, String dbURL) {
    	long start = System.currentTimeMillis();
    	conn = getConnection(driverClass, dbURL);
    	long end = System.currentTimeMillis();
    	System.out.println("getConnection: " + (end - start)/1000);
    	try {
			insertStmt = conn.prepareStatement("INSERT INTO " + DB_TABLE + 
					                           " (" + DB_NA + ") " + "values (?)",
										       Statement.RETURN_GENERATED_KEYS);
		} catch (SQLException e) {
			throw new RepositoryException(e.getMessage(), e);
		}
		long end2 = System.currentTimeMillis();
    	System.out.println("prepare: " + (end2 - end)/1000);
    }
    
    /* (non-Javadoc)
     * @see org.nsdl.repository.handle.HandleGenerator#generateHandle(java.lang.String)
     */
    public String generateHandle(String namingAuthority) {
    	long start = System.currentTimeMillis();
    	
        int id = getNextID(namingAuthority);
        long end = System.currentTimeMillis();
        System.out.println("generateHandle: " + (end - start)/1000);
		return namingAuthority + "/" + id;
    }
    
    private int getNextID(String namingAuthority) {
		rs = null;

		try {
			insertStmt.setString(1, namingAuthority);
			insertStmt.executeUpdate();
			
			int id = -1;
			rs = insertStmt.getGeneratedKeys();

			if (rs.next()) {
				id = rs.getInt(1);
			} else {
				throw new RepositoryException("No generated key returned for inserted row");
			}
			rs.close();
			rs = null;

			return id;
		} catch (SQLException e) {
			throw new RepositoryException(e.getMessage(), e);
		} finally {
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException e) {
					throw new RepositoryException(e.getMessage(), e);
				}
			}
			if (insertStmt != null) {
				try {
					insertStmt.close();
				} catch (SQLException e) {
					throw new RepositoryException(e.getMessage(), e);
				}
			}
		}
    }
    
    private Connection getConnection(String driverClass, String dbURL) {
    	try {
			Class.forName(driverClass);
			return DriverManager.getConnection(dbURL);
		} catch (ClassNotFoundException e) {
			throw new RepositoryException(e.getMessage(), e);
		} catch (SQLException e) {
			throw new RepositoryException(e.getMessage(), e);
		}
    }

    /* (non-Javadoc)
     * @see org.nsdl.repository.handle.HandleGenerator#neverGenerate(java.lang.String)
     */
    public void neverGenerate(String handle) {
        // TODO Auto-generated method stub
        
    }

}
