/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.handle;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.security.PrivateKey;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import net.handle.hdllib.AbstractMessage;
import net.handle.hdllib.AbstractResponse;
import net.handle.hdllib.AddValueRequest;
import net.handle.hdllib.AdminRecord;
import net.handle.hdllib.ClientSessionTracker;
import net.handle.hdllib.Common;
import net.handle.hdllib.CreateHandleRequest;
import net.handle.hdllib.DeleteHandleRequest;
import net.handle.hdllib.Encoder;
import net.handle.hdllib.HandleException;
import net.handle.hdllib.HandleResolver;
import net.handle.hdllib.HandleValue;
import net.handle.hdllib.Interface;
import net.handle.hdllib.ListHandlesRequest;
import net.handle.hdllib.ModifyValueRequest;
import net.handle.hdllib.PublicKeyAuthenticationInfo;
import net.handle.hdllib.RemoveValueRequest;
import net.handle.hdllib.ResolutionRequest;
import net.handle.hdllib.ServerInfo;
import net.handle.hdllib.SessionSetupInfo;
import net.handle.hdllib.SiteInfo;
import net.handle.hdllib.Util;

import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.util.Configuration;

/**
 * Handle client
 * 
 * @author Edwin Shin
 */
public class HandleManager {
    public static final String PROP_NA = "handle.na";
    public static final String PROP_HANDLE_PREFIX = "handle.generator.prefix";
    public static final String PROP_ADMIN_INDEX = "handle.admin.index";
    public static final String PROP_ADMIN_GROUP_INDEX = "handle.admin.group.index";
    public static final String PROP_ADMIN_PRIVKEY = "handle.admin.privateKey";
    public static final String PROP_ADMIN_PASSPHRASE = "handle.admin.passphrase";
    public static final String PROP_GENERATOR = "handle.generator";
    public static final String PROP_GENERATOR_ARGS = "handle.generator.args";
    /**
     * The ResourceBundle that provides access to configuration settings from
     * org/nsdl/repository/resources/Handle.properties.
     */
    // private static ResourceBundle cfg =
    // ResourceBundle.getBundle("org.nsdl.repository.resources.Handle");
    /**
     * The Naming Authority Handle (e.g. 0.NA/2200)
     */
    private final String adminHandle; // e.g. 0.NA/2200

    /**
     * The Naming Authority Public Key index
     */
    private final int adminHandleIndex; // e.g. 300

    private final int adminHandleGroupIndex; // e.g. 200
    private final String adminHandlePrivateKey;
    private final String adminHandlePassphrase;
    private final PublicKeyAuthenticationInfo pubkeyAuthInfo;
    private final String handlePrefix;

    private final String namingAuthority; // e.g. 2200

    /**
     * Handle Administrator. Every handle must have at least one administrator.
     * Every administrator value is made up of a handle and handle value index.
     */
    private final HandleValue adminHandleValue;

    private final HandleResolver resolver = new HandleResolver();

    private final HandleGenerator generator;

    public HandleManager(Properties props) {
        generator = getHandleGenerator(getRequired(props, PROP_GENERATOR),
                props.getProperty(PROP_GENERATOR_ARGS));

        handlePrefix = props.getProperty(PROP_HANDLE_PREFIX);

        namingAuthority = getRequired(props, PROP_NA);
        adminHandle = Util.decodeString(Common.NA_HANDLE_PREFIX)
                + namingAuthority;
        adminHandleIndex = Integer
                .parseInt(getRequired(props, PROP_ADMIN_INDEX)); // e.g. 300
        adminHandleGroupIndex = Integer.parseInt(getRequired(props,
                PROP_ADMIN_GROUP_INDEX)); // e.g. 200
        adminHandlePrivateKey = getRequired(props, PROP_ADMIN_PRIVKEY);
        adminHandlePassphrase = getRequired(props, PROP_ADMIN_PASSPHRASE);

        adminHandleValue = getAdminHandleValue();
        pubkeyAuthInfo = getPubKeyAuthInfo();

        /* Create a new session */
        checkSession();
    }

    public HandleManager(Configuration props) {
        generator = getHandleGenerator(getRequired(props, PROP_GENERATOR),
                props.getProperty(PROP_GENERATOR_ARGS));

        handlePrefix = props.getProperty(PROP_HANDLE_PREFIX);

        namingAuthority = getRequired(props, PROP_NA);
        adminHandle = Util.decodeString(Common.NA_HANDLE_PREFIX)
                + namingAuthority;
        adminHandleIndex = Integer
                .parseInt(getRequired(props, PROP_ADMIN_INDEX)); // e.g. 300
        adminHandleGroupIndex = Integer.parseInt(getRequired(props,
                PROP_ADMIN_GROUP_INDEX)); // e.g. 200
        adminHandlePrivateKey = getRequired(props, PROP_ADMIN_PRIVKEY);
        adminHandlePassphrase = getRequired(props, PROP_ADMIN_PASSPHRASE);

        adminHandleValue = getAdminHandleValue();
        pubkeyAuthInfo = getPubKeyAuthInfo();

        /* Create a new session */
        checkSession();
    }

    /**
     * Convenience method for generating a handle string, but *not* registering
     * it.
     * 
     * @return String that could be used for creating a handle
     */
    public String generateHandle() {
        if (handlePrefix != null) {
            String[] handleParts = generator.generateHandle(namingAuthority)
                    .split("/");
            return handleParts[0] + "/" + handlePrefix + "." + handleParts[1];
        }
        return generator.generateHandle(namingAuthority);
    }

    /**
     * Convenience method to create a handle
     * 
     * @return a new Handle
     */
    public Handle createHandle() throws HandleException {
        String hdl = generateHandle();
        return createHandle(hdl, new ArrayList<HandleValue>());
    }

    /**
     * Convenience method to create a handle with an URL data type
     * 
     * @param url
     *            the url to store
     * @return a Handle
     */
    public Handle createHandleWithURL(String url) throws HandleException {
        String hdl = generateHandle();
        return createHandleWithURL(hdl, url);
    }

    /**
     * Convenience method to create the specified handle with an URL data type
     * 
     * @param hdl
     *            the handle to create
     * @param url
     *            the url to store
     * @return a Handle
     */
    public Handle createHandleWithURL(String hdl, String url)
            throws HandleException {
        HandleValue val = null;
        val = createHandleValue(1, Common.STD_TYPE_URL, Util.encodeString(url));

        List<HandleValue> values = new ArrayList<HandleValue>();
        values.add(val);
        return createHandle(hdl, values);
    }

    /**
     * Create a handle with the specified handle and values
     * 
     * @param hdl
     *            the desired handle
     * @param values
     *            a List of HandleValues
     * @return a Handle that includes the default admin handle value
     */
    public Handle createHandle(String hdl, List<HandleValue> values)
            throws HandleException {
        // add the default admin handle value
        if (values == null) {
            values = new ArrayList<HandleValue>();
        }
        values.add(0, adminHandleValue);

        HandleValue[] hvals = new HandleValue[values.size()];
        CreateHandleRequest req = null;
        req = new CreateHandleRequest(Util.encodeString(hdl), values
                .toArray(hvals), pubkeyAuthInfo);
        resolver.traceMessages = false;
        checkSession();
        AbstractResponse response = resolver.processRequest(req);
        if (response.responseCode != AbstractMessage.RC_SUCCESS) {
            throw new HandleException(HandleException.INTERNAL_ERROR,
                    "createHandle(" + hdl + ", <values>) "
                            + "request failed with error: " + response);
        }
        return new Handle(hdl, hvals);
    }

    /**
     * 
     * @param hdl
     * @return the requested handle
     */
    public Handle getHandle(String hdl) throws HandleException {
        HandleValue[] values;
        checkSession();
        values = resolver.resolveHandle(hdl);
        return new Handle(hdl, values);
    }

    public void modifyHandleValue(Handle hdl, HandleValue val)
            throws HandleException {
        ModifyValueRequest req = new ModifyValueRequest(Util.encodeString(hdl
                .toString()), val, pubkeyAuthInfo);
        checkSession();
        AbstractResponse response = resolver.processRequest(req);
        if (response.responseCode != AbstractMessage.RC_SUCCESS) {
            throw new HandleException(HandleException.INTERNAL_ERROR,
                    "modifyHandleValue failed: " + response);
        }
    }

    /**
     * Add a new HandleValue for the given handle
     * 
     * @param hdl
     *            the handle to add a new value to
     * @param val
     *            the HandleValue to add
     * @throws HandleException
     */
    public void addHandleValue(Handle hdl, HandleValue val)
            throws HandleException {
        AddValueRequest req = new AddValueRequest(Util.encodeString(hdl
                .toString()), val, pubkeyAuthInfo);
        checkSession();
        AbstractResponse response = resolver.processRequest(req);
        if (response.responseCode != AbstractMessage.RC_SUCCESS) {
            throw new HandleException(HandleException.INTERNAL_ERROR,
                    "addHandleValue for " + hdl + " failed: " + response);
        }
    }

    public void removeHandleValue(Handle hdl, HandleValue val)
            throws HandleException {
        RemoveValueRequest req = new RemoveValueRequest(Util.encodeString(hdl
                .toString()), val.getIndex(), pubkeyAuthInfo);
        checkSession();
        AbstractResponse response = resolver.processRequest(req);
        if (response.responseCode != AbstractMessage.RC_SUCCESS) {
            throw new HandleException(HandleException.INTERNAL_ERROR,
                    "removeHandleValue for " + hdl + " failed: " + response);
        }
    }

    public void deleteHandle(String hdl) throws HandleException {
        DeleteHandleRequest req = new DeleteHandleRequest(Util
                .encodeString(hdl), pubkeyAuthInfo);

        checkSession();
        AbstractResponse response = resolver.processRequest(req);
        if (response.responseCode != AbstractMessage.RC_SUCCESS) {
            throw new HandleException(HandleException.INTERNAL_ERROR,
                    "deleteHandle failed for " + hdl + ": " + response);
        }
    }

    public Iterator<String> listHandles() {
        ListHandlesRequest req = new ListHandlesRequest(Util
                .encodeString(adminHandle), pubkeyAuthInfo);

        // dummy request in order to fetch local sites for the naming authority
        ResolutionRequest resReq = new ResolutionRequest(Util
                .encodeString(namingAuthority + "/test"), null, null, null);
        SiteInfo sites[] = new SiteInfo[0];
        try {
            sites = resolver.findLocalSites(resReq);
        } catch (HandleException e) {
            e.printStackTrace();
        }

        // send a list-handles request to each server in the site
        HandleIterator hit = new HandleIterator();

        if (sites.length > 0 && sites[0].servers != null) {
            for (int i = 0; i < sites[0].servers.length; i++) {
                try {
                    ServerInfo server = sites[0].servers[i];
                    AbstractResponse response = resolver.sendRequestToServer(
                            req, server, hit);
                    if (response.responseCode != AbstractMessage.RC_SUCCESS) {
                        // TODO do sth
                    }
                } catch (Throwable t) {
                    System.err.println("\nHDLError: " + t);
                    // TODO t.printStackTrace();
                }
            }
        }
        return hit;
    }

    public HandleGenerator getHandleGenerator() {
        return generator;
    }

    /**
     * 
     * @param className
     *            Name of the HandleGenerator class (e.g.
     *            "org.nsdl.repository.handle.RandomHandleGenerator")
     * @param parameters
     *            Space-separated list of typed constructor parameters for
     *            className. Separate the parameter type from the value with a
     *            colon. (e.g."String:com.mysql.jdbc.Driver String:jdbc:mysql://localhost/test?user=monty&password=greatsqldb"
     *            )
     * @return an instance of the specified HandleGenerator
     */
    private HandleGenerator getHandleGenerator(String className,
            String parameters) {
        try {
            Class<?> hgClass = Class.forName(className);
            if (parameters == null || parameters.equals("")) {
                return (HandleGenerator) hgClass.newInstance();
            }

            String[] typedArgs = parameters.split(" ");
            String[] args = new String[typedArgs.length];
            Class<?>[] parms = new Class[typedArgs.length];
            for (int i = 0; i < typedArgs.length; i++) {
                parms[i] = Class.forName(typedArgs[i].split(":")[0]);
                args[i] = typedArgs[i].split(":", 2)[1];
            }
            Constructor<?> con = hgClass.getConstructor(parms);
            return (HandleGenerator) con.newInstance((Object[]) args);
        } catch (ClassNotFoundException e) {
            throw new RepositoryException(e.getMessage(), e);
        } catch (IllegalAccessException e) {
            throw new RepositoryException(e.getMessage(), e);
        } catch (InstantiationException e) {
            throw new RepositoryException(e.getMessage(), e);
        } catch (SecurityException e) {
            throw new RepositoryException(e.getMessage(), e);
        } catch (NoSuchMethodException e) {
            throw new RepositoryException(e.getMessage(), e);
        } catch (IllegalArgumentException e) {
            throw new RepositoryException(e.getMessage(), e);
        } catch (InvocationTargetException e) {
            throw new RepositoryException(e.getMessage(), e);
        }
    }

    private HandleValue getAdminHandleValue() {
        AdminRecord adminRecord = getAdminRecord();
        return createHandleValue(100, Common.STD_TYPE_HSADMIN, Encoder
                .encodeAdminRecord(adminRecord));
    }

    private AdminRecord getAdminRecord() {
        return new AdminRecord(Util.encodeString(adminHandle),
                adminHandleGroupIndex, true, true, true, true, true, true,
                true, true, true, true, true, true);
    }

    private PrivateKey getPrivateKey() {
        // Read the private key from disk
        byte[] key = null;
        FileInputStream fs = null;
        try {
            File f = new File(adminHandlePrivateKey);
            fs = new FileInputStream(f);
            key = new byte[(int) f.length()];
            int n = 0;
            while (n < key.length)
                key[n++] = (byte) fs.read();
            fs.read(key);
        } catch (Throwable t) {
            throw new RuntimeException("Cannot read private key "
                    + adminHandlePrivateKey + ": " + t);
        } finally {
            if (fs != null) {
                try {
                    fs.close();
                } catch (IOException e) {

                }
            }
        }

        // Check to see if the private key is encrypted. If so, read in the
        // user's passphrase and decrypt. Finally, convert the byte[]
        // representation of the private key into a PrivateKey object.
        PrivateKey privkey = null;
        byte secKey[] = null;
        try {
            if (Util.requiresSecretKey(key)) {
                secKey = Util.encodeString(adminHandlePassphrase);
            }
            key = Util.decrypt(key, secKey);
            privkey = Util.getPrivateKeyFromBytes(key, 0);
        } catch (Throwable t) {
            System.err.println("Can't load private key in "
                    + adminHandlePrivateKey + ": " + t);
        }
        return privkey;
    }

    /**
     * Create a PublicKeyAuthenticationInfo object to pass to HandleResolver.
     * This is constructed with the admin handle, index, and PrivateKey as
     * arguments.
     * 
     * @return PublicKeyAuthenticationInfo
     */
    private PublicKeyAuthenticationInfo getPubKeyAuthInfo() {
        PrivateKey privkey = getPrivateKey();
        PublicKeyAuthenticationInfo auth = null;
        try {
            auth = new PublicKeyAuthenticationInfo(Util
                    .encodeString(adminHandle), adminHandleIndex, privkey);
        } catch (Throwable t) {
            System.err.println("\nError: " + t);
        }
        return auth;
    }

    private static int getTimestamp() {
        return (int) (System.currentTimeMillis() / 1000);
    }

    protected static String getRequired(Properties props, String key) {
        String val = props.getProperty(key);
        if (val == null) {
            throw new RepositoryException("Required property is not set: "
                    + key);
        }
        return val.trim();
    }

    protected static String getRequired(Configuration props, String key) {
        String val = props.getProperty(key);
        if (val == null) {
            throw new RepositoryException("Required property is not set: "
                    + key);
        }
        return val.trim();
    }

    /**
     * Convenience method for creating a HandleValue for a string
     * 
     * @param index
     * @param name
     * @param value
     * @return HandleValue
     */
    public HandleValue createHandleValue(int index, String name, String value) {
        return createHandleValue(index, Util.encodeString(name), Util
                .encodeString(value));
    }

    public HandleValue createHandleValue(int index, byte[] name, byte[] value) {
        return new HandleValue(index, name, value,
                HandleValue.TTL_TYPE_RELATIVE, 86400, getTimestamp(), null,
                true, true, true, false);
    }

    private void checkSession() {
        if (resolver.getSessionTracker() != null) {
            try {
                net.handle.hdllib.ClientSideSessionInfo[] si = resolver
                        .getSessionTracker().getAllSessions();
                if (si.length == 1) {
                    /* leave if the session is OK and not expired */
                    if (!(si[0].hasExpired())) {
                        return;
                    }
                }
            } catch (Exception e) {/*
                                    * If we have a problem, just create a new
                                    * session
                                    */
            }
        }

        /* If we reached this point, then we need to try to create a new session */
        SessionSetupInfo ss = null;
        try {
            ss = new SessionSetupInfo(pubkeyAuthInfo);
            resolver.setSessionTracker(new ClientSessionTracker(ss));
            resolver.setPreferredProtocols(new int[] { Interface.SP_HDL_TCP });
        } catch (Exception e) {
            System.err
                    .println("Could not set up handle session.  Defaulting to non-session mode");
        }
    }
}
