/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.commons.httpclient.UsernamePasswordCredentials;
import org.apache.commons.httpclient.auth.AuthScope;
import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.impl.fedora.ri.RIQueryEngine;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.ConfigurationFactory;

import fedora.common.Constants;
import fedora.common.PID;

/**
 * Basic implementation of FedoraAccess.
 * 
 * <h2>Configuration Parameters</h2>
 * <p>
 * <dl>
 * <dt><code>param.repository.impl.fedora.ri.engine</code></dt>
 * <dd>Implementing class of
 * {@link org.nsdl.repository.impl.fedora.ri.RIQueryEngine RIQueryEngine}.
 * Default is <code>org.nsdl.repository.impl.fedora.ri.MPTQueryEngine</code></dd>
 * 
 * <dt><code>param.repository.impl.fedora.timeout</code></dt>
 * <dd>Timeout (in seconds) when accessing Fedora content. Default is
 * <code>60</code></dd>
 * 
 * <dt><code>param.repository.impl.fedora.connections.max</code></dt>
 * <dd>Maximum number of http connecitons to Fedora at a given time. Default is
 * <code>1024</code></dd>
 * </dl>
 * </p>
 * 
 * @author birkland
 * @since 1.0
 */
public class BasicFedoraAccess implements FedoraAccess, Configuration.Listener {

    private static final String PROP_RI_ENGINE = "param.repository.impl.fedora.ri.engine";
    public static final String PROP_TIMEOUT = "param.repository.impl.fedora.timeout";
    private static final String PROP_MAX_HTTP_CONNECTIONS = "param.repository.impl.fedora.connections.max";

    public static final String TIMEOUT_DEFAULT = "60";
    private static final String MAX_HTTP_CONNECTIONS_DEFAULT = "1024";

    private static final String CREATED_DATE_PREDICATE = " <"
            + Constants.MODEL.CREATED_DATE.uri + "> ";

    private FedoraAPIWrapper wrapper;
    private RIQueryEngine ri;
    private HttpClient client;

    public BasicFedoraAccess(FedoraAPIWrapper wrapper, RIQueryEngine ri,
            HttpClient client) {
        this.wrapper = wrapper;
        this.ri = ri;
        this.client = client;
    }

    public BasicFedoraAccess() {
    }

    public FedoraAPIWrapper getAPI() {
        return wrapper;
    }

    public RIQueryEngine getRI() {
        return ri;
    }

    public HttpClient getHttpClient() {
        return client;
    }

    public void updateConfig(Configuration config) {

        /* Read the RI specified in the config */
        String riClass = config.getProperty(PROP_RI_ENGINE,
                "org.nsdl.repository.impl.fedora.ri.MPTQueryEngine");

        this.ri = ConfigurationFactory.getInstance(RIQueryEngine.class)
                .buildFromClassName(riClass, config);

        this.wrapper = ConfigurationFactory.getInstance(FedoraAPIWrapper.class)
                .buildFromClass(FedoraAPIWrapper.class, config);

        String timeout = config.getProperty(PROP_TIMEOUT, TIMEOUT_DEFAULT);
        String maxConnections = config.getProperty(PROP_MAX_HTTP_CONNECTIONS,
                MAX_HTTP_CONNECTIONS_DEFAULT);

        /*
         * Now, set up an HttpClient with the proper credentials. We could ask
         * FedoraClient for its HTTPClient, but we've seen the need to adjust
         * the timeout value
         */
        MultiThreadedHttpConnectionManager connManager = new MultiThreadedHttpConnectionManager();
        connManager.getParams().setConnectionTimeout(
                Integer.parseInt(timeout) * 1000);
        connManager.getParams().setSoTimeout(Integer.parseInt(timeout) * 1000);
        connManager.getParams().setMaxTotalConnections(
                Integer.parseInt(maxConnections));
        connManager.getParams().setDefaultMaxConnectionsPerHost(
                Integer.parseInt(maxConnections));

        UsernamePasswordCredentials credentials = new UsernamePasswordCredentials(
                config.getProperty(FedoraAPIWrapper.PROP_FEDORA_USER), config
                        .getProperty(FedoraAPIWrapper.PROP_FEDORA_PASSWORD));

        client = new HttpClient(connManager);
        client.getState().setCredentials(
                new AuthScope(wrapper.HOST, AuthScope.ANY_PORT), credentials);
        client.getParams().setAuthenticationPreemptive(true);
    }

    public boolean objectExists(PID pid) {
        String subjectPart = "<" + pid.toURI() + ">";
        QueryResults results = null;
        try {
            results = getRI().execute(
                    subjectPart + CREATED_DATE_PREDICATE + "*", RIQueryEngine.Language.SPO);
            if (results.hasNext()) {
                return true;
            } else {
                return false;
            }
        } catch (QueryException e) {
            throw new RepositoryException("Could not poll for pid", e);
        } finally {
            if (results != null) {
                results.close();
            }
        }
    }
}
