/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.rmi.RemoteException;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.content.ContentModelException;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.types.TypeFactory;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.ConfigurationFactory;
import org.nsdl.repository.util.StaticObjectInfo;

import fedora.common.PID;

/**
 * Base implementation of NDRAccess to an underlying Fedora repository.
 * 
 * <h2>Configuration Parameters</h2>
 * <p>
 * <dl>
 * <dt><code>repository.impl.fedora.BasicFedoraNDRAccess.connector</dt></dt>
 * <dd>Class implementing {@link FedoraAccess} that allows connectivity to the
 * underlying Fedora repository.</dd>
 * </dl>
 * </p>
 * 
 * @author birkland
 * @since 1.0
 */
public class BasicFedoraNDRAccess implements FedoraNDRAccess, NDRAccess,
        Configuration.Listener {

    private static final Logger logger = Logger
            .getLogger(BasicFedoraNDRAccess.class);

    private static final String PROP_FEDORA_ACCESS = "repository.impl.fedora.BasicFedoraNDRAccess.connector";
    private FedoraAccess repository;
    private final FedoraNDRAccess ndr;
    private FedoraNDRObjectFinder finder;
    private boolean configIsRead = false;
    private TypeFactory factory = new TypeFactory(this);
    private FedoraDatastreamFactory m_dsFactory = new FedoraDatastreamFactory();

    public BasicFedoraNDRAccess(FedoraAccess repo) {
        this.repository = repo;
        this.finder = new FedoraNDRObjectFinder(this);
        this.ndr = this;
    }

    public BasicFedoraNDRAccess() {
        this.ndr = this;
    }

    /** {@inheritDoc} */
    public FedoraNDRObject getObject(NDRObjectInfo identifier) {
        if (FedoraNDRObject.class.isAssignableFrom(identifier.getClass())) {
            return (FedoraNDRObject) identifier;
        } else {
            return new MutableAPIMFedoraObject(this, identifier);
        }
    }

    public <T extends NDRObject> T getObject(NDRObjectInfo identity,
            Type<T> type) {
        return factory.instantiate(getObject(identity), type);
    }

    public <T extends NDRObject> T getObject(String identity, Type<T> type) {
        return factory.instantiate(getObject(identity), type);
    }

    public FedoraNDRObject getObject(PID identifier) {
        return new MutableAPIMFedoraObject(this, identifier);
    }

    public FedoraNDRObject getObject(String identifier) {
        if (FedoraUtils.isFedoraIdentifier(identifier)) {
            return new MutableAPIMFedoraObject(this, PID
                    .getInstance(identifier));
        } else {
            return new MutableAPIMFedoraObject(this, new StaticObjectInfo(
                    identifier));
        }
    }

    /** {@inheritDoc} */
    public <T extends NDRObject> T newObject(Type<T> type) {
        /*
         * XXX: It would be nice to have a way of specifing a pid generator
         * here. For now, though, we just assign a null pid (commit() will
         * generate a pid based on the handle if pid is still null at that time)
         */
        FedoraNDRObject f = new NewObjectWrapper(type, null);

        return factory.instantiate(f, type);
    }

    /** {@inheritDoc} */
    public void deleteObject(NDRObjectInfo object) throws ExistenceException,
            ContentModelException {

        NDRObject toBeDeleted = this.getObject(object);

        toBeDeleted.setState(NDRObject.State.Deleted);
        toBeDeleted.commit();
    }

    /** {@inheritDoc} */
    public FedoraNDRObject openObject(NDRObjectInfo object)
            throws ExistenceException {
        return new MutableAPIMFedoraObject(this, object);
    }

    /** {@inheritDoc} */
    public void purgeObject(NDRObjectInfo object) throws RepositoryException,
            ContentModelException {
        FedoraNDRObject o = ndr.getObject(object);
        try {
            repository.getAPI().purgeObject(o.pid().toString(),
                    "Purged by NDR API", false);
        } catch (RemoteException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    /** {@inheritDoc} */
    public FedoraNDRObjectFinder getObjectFinder() {
        return finder;
    }

    public FedoraAccess getRepository() {
        return repository;
    }

    public void updateConfig(Configuration config) {

        /*
         * NB: We are not subscribing to the configuration, since we can't
         * handle dynamic changes yet. To do so: config.addListener(this);
         */

        if (configIsRead) {
            logger.warn("Changing a running configuration is not supported.  "
                    + "Ignoring");
            return;
        } else {
            logger.info("config: initializing local Fedora Access");
            repository = ConfigurationFactory.getInstance(FedoraAccess.class)
                    .buildFromProperty(PROP_FEDORA_ACCESS, config);
            this.finder = new FedoraNDRObjectFinder(this);

            m_dsFactory.updateConfig(config);
            configIsRead = true;
        }
    }

    public FedoraDatastreamFactory getDatastreamFactory() {
        return m_dsFactory;
    }

    public boolean isReadOnly() {
        return false;
    }

    private class NewObjectWrapper extends FoxmlObjectBuilder {
        public NewObjectWrapper(Type<? extends NDRObject> type, PID pid) {
            super(type, pid, ndr);
        }

        public void commit() {
            if (pid() == null) {
                PID pid = null;
                if (handle().contains("/")) {
                    pid = PID.getInstance("hdl:"
                            + this.handle().replace("/", "%2F"));
                    logger
                            .debug("PID was not explicictly defined for object.  "
                                    + "Using default value from handle: "
                                    + pid.toString());
                } else {
                    pid = PID.getInstance(handle());
                }

                this.setPID(pid);
                super.commit();
            }
        }
    }
}
