/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.HashMap;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.log4j.Logger;
import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.impl.fedora.ri.RIQueryEngine;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.util.StreamUtil;

import fedora.common.Constants;
import fedora.common.PID;

/**
 * Datastream resident in Fedora, accessed through API-A and the RI
 * <p>
 * Provides access to datastream definition and content as defined in and ripped
 * from a live Fedora object. When advantageous to do so, it will defer reading
 * data from Fedora until it is absolutely necessary to do so, instead
 * attempting to retrieve datastream properties from the Resource Index
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class DeferredAPIADatastream implements FedoraDatastream {

    private fedora.server.types.gen.Datastream fedoraDS;

    private final FedoraAccess repo;

    private final PID pid;

    private final Datastream.Name dsID;

    private HashMap<String, String> dsInfo;

    private String m_label;

    private static final Logger logger = Logger
            .getLogger(DeferredAPIADatastream.class.getName());

    private byte[] m_newContent;
    private String m_newMime;
    private URL m_newURL;

    /**
     * Create as a wrapper to the Fedora generated datastream type
     * <p>
     * Most of the datastream definition is provided in the Fedora datastream
     * representation object. This class is primarily a wrapper for these
     * definitions (i.e. mimetype, ID, label, etc), and uses these definitions
     * to access the fedora repository when datastream contents are requested.
     * </p>
     * 
     * @param ds
     *            Datastream, as returned by fedora client methods
     * @param pid
     *            persistent ID (PID) of the Fedora object
     * @param repository
     *            Accessor to the specific fedora repository
     */
    public DeferredAPIADatastream(fedora.server.types.gen.Datastream ds,
            PID pid, FedoraAccess repository) {
        this.fedoraDS = ds;

        /* XXX: this is making a lot of assumptions with namespaces */
        this.dsID = new Datastream.Name(ds.getID());
        this.repo = repository;
        this.pid = pid;
    }

    /**
     * Create a somewhat lighter-weight access to a Fedora datastream
     * <p>
     * Constructed in this way, this class tries to access datastream attributes
     * in a lightweight way (using the RI) if possible. If it is not, it will
     * poll Fedora for datastream info (which is somewhat expensive).
     * </p>
     * 
     * @param dsID
     *            ID of the datastream
     * @param pid
     *            persistent ID (PID) of the Fedora object
     * @param repository
     *            Accessor to the specific fedora repository
     */
    public DeferredAPIADatastream(String dsID, PID pid, FedoraAccess repository) {
        /* XXX: this is making a lot of assumptions with namespaces */
        this.dsID = new Datastream.Name(dsID);
        this.repo = repository;
        this.pid = pid;
    }

    /** {@inheritDoc} */
    public ControlGroup controlGroup() {
        getDSInfo();
        return ControlGroup.getInstance(fedoraDS.getControlGroup().toString());
    }

    /** {@inheritDoc} */
    public String[] getAltIDs() {
        return fedoraDS.getAltIDs();
    }

    public void setContent(InputStream content) {
        if (isCanonical()) {
            m_newContent = StreamUtil.streamToByteArray(content);
        } else {
            throw new AccessException(
                    "Cannot change the content of non-canonical datastream '"
                            + dsID + "'");
        }
    }

    /** {@inheritDoc} */
    public InputStream getContent() throws AccessException {
        if (m_newContent == null) {
            return (getDsStream(getDsURL()));
        } else {
            return new ByteArrayInputStream(m_newContent);
        }
    }

    /** {@inheritDoc} */
    public InputStream getContent(String version) throws AccessException {
        if (m_newContent == null) {
            String path = getDsURL() + "/" + version;
            return getDsStream(path);
        } else {
            return new ByteArrayInputStream(m_newContent);
        }
    }

    /** {@inheritDoc} */
    public String getFormatURI() {
        getDSInfo();
        return fedoraDS.getFormatURI();
    }

    /** {@inheritDoc} */
    public Datastream.Name getName() {
        return dsID;
    }

    /** {@inheritDoc} */
    public String getLabel() {
        if (m_label == null) {
            getDSInfo();
            m_label = fedoraDS.getLabel();
        }

        return m_label;
    }

    public void setMimeType(String mime) {
        m_newMime = mime;
    }

    /** {@inheritDoc} */
    public String getMimeType() {
        if (m_newMime == null) {
            getDSInfoLite();
            return dsInfo.get(Constants.VIEW.MIME_TYPE.uri);
        } else {
            return m_newMime;
        }
    }

    /** {@inheritDoc} */
    public URL getURL() {
        if (m_newURL == null) {
            if (controlGroup().equals(FedoraDatastream.ControlGroup.Redirect)) {

                /*
                 * I don't believe there's any way to determine the URL of the
                 * redirect other than to give it a try and look at the address
                 * that is returned back in the redirect header
                 */

                GetMethod dsGet = new GetMethod(getDsURL());
                dsGet.setFollowRedirects(false);
                try {
                    repo.getHttpClient().executeMethod(dsGet);
                    Header resp = dsGet.getResponseHeader("location");
                    if (resp != null) {
                        return new URL(resp.getValue());
                    } else {
                        throw new AccessException(
                                "Did not get a redirect from server");
                    }

                } catch (Exception e) {
                    logger.error("Could not determine redirect for "
                            + getDsURL());
                    throw new RepositoryException("Could not determine "
                            + "redirect for " + getDsURL());
                }
            } else {
                try {
                    return new URL(getDsURL());
                } catch (Exception e) {
                    throw new RuntimeException(e);
                }

            }
        } else {
            return m_newURL;
        }
    }

    public void setURL(URL url) {
        if (isCanonical()) {
            throw new AccessException(
                    "Cannot change the URL of canonical datastream '" + dsID
                            + "'");
        }
        m_newURL = url;
    }

    /** {@inheritDoc} */
    public long getSize() {
        getDSInfo();
        return fedoraDS.getSize();
    }

    /** {@inheritDoc} */
    public State getState() {
        getDSInfoLite();
        return FedoraDatastream.State.fromURI(dsInfo
                .get(Constants.MODEL.STATE.uri));
    }

    /** {@inheritDoc} */
    public boolean isVersionable() {
        getDSInfo();
        return fedoraDS.isVersionable();
    }

    public boolean isCanonical() {
        return (controlGroup().equals(ControlGroup.Managed) || controlGroup()
                .equals(ControlGroup.InlineXML));
    }

    /** Get the APIA-Lite datastream access URL */
    private String getDsURL() {
        return repo.getAPI().PROTOCOL + "://" + repo.getAPI().HOST + ":"
                + repo.getAPI().PORT + "/fedora/get/" + pid.toString() + "/"
                + dsID;
    }

    public void setLabel(String name) {
        m_label = name;
    }

    /** Open a stream to read datastream content */
    private InputStream getDsStream(String path) throws AccessException {
        /* Is is okay not to explicitly close the connection? probably */
        logger.trace("Opening datastream stream from " + path);
        GetMethod dsGet = new GetMethod(path);
        dsGet.getParams().setParameter("Connection", "Keep-Alive");
        try {
            repo.getHttpClient().executeMethod(dsGet);
            return dsGet.getResponseBodyAsStream();
        } catch (Exception e) {
            logger.warn("Error reading from " + path);
            throw new AccessException("Could not access datastream ", e);
        }
    }

    /** Get datastream description from Fedora */
    private void getDSInfo() {
        logger.trace("Getting fedora DS info...");
        if (fedoraDS == null) {
            try {
                this.fedoraDS = repo.getAPI().getDatastream(pid.toString(),
                        dsID.getID(), null);
                logger.trace("Done getting DS info");
            } catch (RemoteException e) {
                throw new RepositoryException(e);
            }
        }
    }

    /** Get some datastream attributes from the RI, if possible */
    private void getDSInfoLite() {
        if (dsInfo == null) {
            dsInfo = new HashMap<String, String>();
            logger.trace("Getting ds Info Lite..");
            try {
                dsInfo.put(Constants.MODEL.STATE.uri,
                        getInfoValue(Constants.MODEL.STATE.uri));
                dsInfo.put(Constants.VIEW.MIME_TYPE.uri,
                        getInfoValue(Constants.VIEW.MIME_TYPE.uri));
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
            logger.trace("Done getting ds Info Lite");
        }
    }

    /** Qery the RI for a specific attribute */
    private String getInfoValue(String infoType) throws QueryException {
        String dsName = "<" + pid.toURI() + "/" + dsID + "> ";
        QueryResults r = repo.getRI().execute(dsName + " <" + infoType + "> *",
                RIQueryEngine.Language.SPO);
        if (r.hasNext()) {
            String value = r.next().get(2).getValue();
            r.close();
            return value;
        } else {
            return null;
        }
    }

}
