/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.MalformedURLException;
import java.rmi.RemoteException;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamReader;

import org.apache.axis.types.NonNegativeInteger;
import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.util.Configuration;
import org.nsdl.util.LogUtility;
import org.nsdl.util.xml.XMLStreamComparator;

import fedora.client.FedoraClient;
import fedora.server.access.FedoraAPIA;
import fedora.server.management.FedoraAPIM;
import fedora.server.types.gen.Datastream;
import fedora.server.types.gen.DatastreamDef;
import fedora.server.types.gen.FieldSearchQuery;
import fedora.server.types.gen.FieldSearchResult;
import fedora.server.types.gen.MIMETypedStream;
import fedora.server.types.gen.ObjectMethodsDef;
import fedora.server.types.gen.ObjectProfile;
import fedora.server.types.gen.Property;
import fedora.server.types.gen.RepositoryInfo;

/**
 * Wraps the fedora access (APIA and APIM) calls for use by the NSDL API
 * servlets.
 * 
 * <h2>Configuration Parameters</h2>
 * <p>
 * <dl>
 * 
 * <dt><code>param.fedora.use_ssl</code></dt>
 * <dd>Determines whether to use a SSL connection to Fedora or not.
 * <code>true</code> will force ssl, <code>false</code> will disable ssl.
 * Default is <code>true</code></dd>
 * 
 * <dt><code>param.protocol.nonssl</code></dt>
 * <dd>Name of the default non-ssl protocol. Default is <code>http</code></dd>
 * 
 * <dt><code>param.protocol.ssl</code></dt>
 * <dd>Name of the default ssl protocol. Default is <code>https</code></dd>
 * 
 * <dt><code>param.host.fedora</code></dt>
 * <dd>Fedora hostname. Default is <code>localhost</code></dd>
 * 
 * <dt><code>param.port.fedora.nonssl</code></dt>
 * <dd>Non-SSL fedora port. Default is <code>8080</code></dd>
 * 
 * <dt><code>param.port.fedora.ssl</code></dt>
 * <dd>SSL fedora port. Default is <code>8443</code></dd>
 * 
 * <dt><code>param.usr.fedora</code></dt>
 * <dd>Fedora user to connect as. Default is <code>fedoraAdmin</code></dd>
 * 
 * <dt><code>param.pwd.fedora</code></dt>
 * <dd>Fedora user password. Default is <code>fedoraAdmin</code>, though it is
 * very wise to change this</dd>
 * 
 * <dt><code>param.truststore.loc.fedora</code></dt>
 * <dd>Trust store file location. Default is
 * <code>/usr/local/fedora/server/truststore</code></dd>
 * 
 * <dt><code>param.truststore.pwd.fedora</code></dt>
 * <dd>Trust store password. Default is <code>tomcat</code>, though is is very
 * wise to create one's own keystore with a different identity and password</dd>
 * </dl>
 * </p>
 * 
 * @author Tim & Elly
 * 
 */
public class FedoraAPIWrapper implements Configuration.Listener {
    private static final Logger logger = Logger
            .getLogger(FedoraAPIWrapper.class.getName());

    private static final String DEFAULT_USER = "fedoraAdmin";
    private static final String DEFAULT_PASSWD = "fedoraAdmin";
    private static final String DEFAULT_TRUSTSTORE_LOC = "${FEDORA_HOME}/server/truststore";
    private static final String DEFAULT_TRUSTSTORE_PASSWD = "tomcat";

    /** property that determines if ssl is used by default */
    public static final String PROP_USE_SECURE = "param.fedora.use_ssl";

    // protocol ports
    /** name of property for non-secure transport protocol for NDR requests */
    public static final String PROP_PROTOCOL_NONSSL = "param.protocol.nonssl";
    /** name of property for secure transport protocol for NDR requests */
    public static final String PROP_PROTOCOL_SSL = "param.protocol.ssl";

    // fedora host and ports
    /** name of property for the fedora host machine */
    public static final String PROP_FEDORA_HOST = "param.host.fedora";
    /** name of property for the fedora nonssl port */
    public static final String PROP_FEDORA_NONSSL_PORT = "param.port.fedora.nonssl";
    /** name of property for the fedora ssl port */
    public static final String PROP_FEDORA_SSL_PORT = "param.port.fedora.ssl";

    // fedora account
    /** name of property for the fedora user (account) */
    public static final String PROP_FEDORA_USER = "param.usr.fedora";
    /** name of property for the fedora user account password */
    public static final String PROP_FEDORA_PASSWORD = "param.pwd.fedora";

    // fedora truststore file location/password
    /**
     * name of property for the fedora truststore (for fedora client to talk to
     * fedora server) file location
     */
    public static final String PROP_FEDORA_TRUSTSTORE_LOC = "param.truststore.loc.fedora";
    /** name of property for the fedora truststore password */
    public static final String PROP_FEDORA_TRUSTSTORE_PWD = "param.truststore.pwd.fedora";

    // java truststore system property names
    /** name of property for javax ssl truststore */
    public static final String PROP_JAVAX_SSL_TRUSTSTORE = "javax.net.ssl.trustStore";
    /** name of property for javax ssl truststore */
    public static final String PROP_JAVAX_SSL_TRUSTSTORE_PWD = "javax.net.ssl.trustStorePassword";

    public static boolean FAIL_ON_ERROR = true;

    public static boolean FOLLOW_REDIRECTS = false;

    public String PROTOCOL = "http";
    public String PROTOCOL_SECURE = "https";

    public String HOST = "localhost";
    public String PORT = "8080";
    public String PORT_SECURE = "8443";

    private XMLInputFactory mclsXMLInputFactory = XMLInputFactory.newInstance();

    private FedoraClient client;

    // URL for reading apia access params - query url.
    private URL murlQueryURL;

    private static boolean TRACE_SYSTEM_OUT = false;

    private static final String ME = "FedoraAPIWrapper";

    /** No argument constructor. */
    public FedoraAPIWrapper() {
    }

    /**
     * Create a configured API wrapper.
     * 
     * @param lstrRepositoryParams
     *            Connection parameters. Should be an array of
     *            [param.protocol.nonssl, param.host.fedora, param.port.nonssl,
     *            param.user.fedora, param.pwd.fedora,
     *            param.truststore.loc.fedora,param.truststore.pwd.fedora,
     *            param.protocol.ssl, param.port.ssl]
     * @param lblnSecure
     *            Force secure SSL connections.
     * @throws MalformedURLException
     *             if there is some problem forming the fedora URL from the
     *             given protocol, host, and port.
     */
    public FedoraAPIWrapper(String[] lstrRepositoryParams, boolean lblnSecure)
            throws MalformedURLException {
        PROTOCOL = lstrRepositoryParams[0];
        HOST = lstrRepositoryParams[1];
        PORT = lstrRepositoryParams[2];
        // user
        // pwd
        // truststore loc
        // truststore password
        PROTOCOL_SECURE = lstrRepositoryParams[7];
        PORT_SECURE = lstrRepositoryParams[8];

        if (lblnSecure) {
            initFedoraAPIWrapper(PROTOCOL_SECURE, HOST, PORT_SECURE,
                    lstrRepositoryParams[3], lstrRepositoryParams[4],
                    lstrRepositoryParams[5], lstrRepositoryParams[6]);
        } else {
            initFedoraAPIWrapper(PROTOCOL, HOST, PORT, lstrRepositoryParams[3],
                    lstrRepositoryParams[4], lstrRepositoryParams[5],
                    lstrRepositoryParams[6]);
        }

    }

    /** {@inheritDoc} */
    public void updateConfig(Configuration config) {
        PROTOCOL = config.getProperty(FedoraAPIWrapper.PROP_PROTOCOL_NONSSL,
                PROTOCOL);
        HOST = config.getProperty(FedoraAPIWrapper.PROP_FEDORA_HOST, HOST);
        PORT = config.getProperty(FedoraAPIWrapper.PROP_FEDORA_NONSSL_PORT,
                PORT);
        String user = config.getProperty(FedoraAPIWrapper.PROP_FEDORA_USER,
                DEFAULT_USER);
        String passwd = config.getProperty(
                FedoraAPIWrapper.PROP_FEDORA_PASSWORD, DEFAULT_PASSWD);
        String trustStoreLoc = config.getProperty(
                FedoraAPIWrapper.PROP_FEDORA_TRUSTSTORE_LOC,
                DEFAULT_TRUSTSTORE_LOC).replace("${FEDORA_HOME}",
                System.getenv("FEDORA_HOME")).replace("/", File.separator);
        String trustStorePassword = config.getProperty(
                FedoraAPIWrapper.PROP_FEDORA_TRUSTSTORE_PWD,
                DEFAULT_TRUSTSTORE_PASSWD);
        PROTOCOL_SECURE = config.getProperty(
                FedoraAPIWrapper.PROP_PROTOCOL_SSL, PROTOCOL_SECURE);
        PORT_SECURE = config.getProperty(FedoraAPIWrapper.PROP_FEDORA_SSL_PORT,
                PORT_SECURE);

        String secure = config.getProperty(PROP_USE_SECURE);
        try {
            if (secure != null && secure.equals("false")) {

                initFedoraAPIWrapper(PROTOCOL, HOST, PORT, user, passwd,
                        trustStoreLoc, trustStorePassword);

            } else {
                initFedoraAPIWrapper(PROTOCOL_SECURE, HOST, PORT_SECURE, user,
                        passwd, trustStoreLoc, trustStorePassword);
            }
            client.TIMEOUT_SECONDS = new Integer(config.getProperty(
                    BasicFedoraAccess.PROP_TIMEOUT,
                    BasicFedoraAccess.TIMEOUT_DEFAULT)).intValue();
            client.SOCKET_TIMEOUT_SECONDS = client.TIMEOUT_SECONDS;
        } catch (MalformedURLException e) {
            throw new AccessException(
                    "Could not instantiate Fedora APi Wrapper", e);
        }
    }

    private void initFedoraAPIWrapper(String lstrprotocol, String lstrhost,
            String lstrsPort, String lstruser, String lstrpass,
            String trustStoreLoc, String trustStorePassword)
            throws MalformedURLException {
        int lintport = new Integer(lstrsPort).intValue();

        // set up java ssl trustStore properties, if they haven't yet be set
        String sysPropSetting = System.getProperty(PROP_JAVAX_SSL_TRUSTSTORE);
        if (sysPropSetting == null || sysPropSetting.length() < 1)
            System.setProperty(PROP_JAVAX_SSL_TRUSTSTORE, trustStoreLoc);
        sysPropSetting = System.getProperty(PROP_JAVAX_SSL_TRUSTSTORE_PWD);
        if (sysPropSetting == null || sysPropSetting.length() < 1)
            System.setProperty(PROP_JAVAX_SSL_TRUSTSTORE_PWD,
                    trustStorePassword);

        String baseURL = lstrprotocol + "://" + lstrhost + ":" + lintport
                + "/fedora/";

        client = new FedoraClient(baseURL, lstruser, lstrpass);

        client.MAX_CONNECTIONS_PER_HOST = 50; // Should cover future
        // (multi-threaded) client use
        client.MAX_TOTAL_CONNECTIONS = 60; // 10 extras for connecting to
        // something else.

        // Create url for use in queries
        this.murlQueryURL = new URL(this.PROTOCOL, this.HOST, new Integer(
                this.PORT).intValue(), "/fedora/risearch");
    }

    public String addDatastream(String pid, String dsID, String[] altIDs,
            String dsLabel, boolean versionable, String dsMIME,
            String formatURI, String dsLocation, String dsControlGroup,
            String dsState, String logMessage) throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "entered .addDatastream", pid + " ds: " + dsID);
        }

        // make the SOAP call on API-M using the connection stub
        /* Note: Assumes no checksumming */
        String datastreamID = myAPIM.addDatastream(pid, dsID, altIDs, dsLabel,
                versionable, dsMIME, formatURI, dsLocation, dsControlGroup,
                dsState, null, null, logMessage);

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "leaving .addDatastream", pid + " ds: " + dsID);
        }

        return datastreamID;
    }

    public RepositoryInfo describeRepository() throws RemoteException {
        FedoraAPIA myAPIA;
        try {
            myAPIA = client.getAPIA();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        return myAPIA.describeRepository();
    }

    public byte[] export(String pid, String format, String exportContext)
            throws RemoteException, IOException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        // make the SOAP call on API-M
        byte[] objectXML = myAPIM.export(pid, format, exportContext);

        return objectXML;
    }

    public FieldSearchResult findObjects(String[] resultFields,
            String maxResults, FieldSearchQuery query) throws RemoteException {
        FedoraAPIA myAPIA;
        try {
            myAPIA = client.getAPIA();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }
        NonNegativeInteger lMaxResults = new NonNegativeInteger(maxResults
                .getBytes());
        return myAPIA.findObjects(resultFields, lMaxResults, query);
    }

    public FedoraClient getClient() {
        return client;
    }

    public Datastream getDatastream(String pid, String dsID, String asOfDateTime)
            throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        return myAPIM.getDatastream(pid, dsID, asOfDateTime);
    }

    public Datastream[] getDatastreamHistory(String pid, String dsID)
            throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        return myAPIM.getDatastreamHistory(pid, dsID);
    }

    public Datastream[] getDatastreams(String pid, String asOfDateTime,
            String dsState) throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        return myAPIM.getDatastreams(pid, asOfDateTime, dsState);
    }

    public MIMETypedStream getDatastreamDissemination(String pid, String dsID,
            String asOfDateTime) throws RemoteException {
        FedoraAPIA myAPIA;
        try {
            myAPIA = client.getAPIA();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }
        return myAPIA.getDatastreamDissemination(pid, dsID, asOfDateTime);
    }

    /**
     * Method to use the fedora client's get-current-DS-as-string method.
     * 
     * @param lstrPID
     *            pid to fetch
     * @param lstrStreamName
     *            datastream to fetch
     * @return The Datastream as a string.
     * @throws IOException
     */
    public String getDataStreamAsString(String lstrPID, String lstrStreamName)
            throws IOException {

        String lstrNull = "";

        return this.getDatastreamAsString(lstrPID, lstrStreamName, lstrNull);

        // Defaults to:
        // fail on error - first boolean arg
        // follow redirects - second boolean arg
        // String lstrDS = null;
        //    
        // String lstrDSFetchURL = FedoraClient.FEDORA_URI_PREFIX + lstrPID +
        // "/" +
        // lstrStreamName;
        //    
        // try
        // {
        // lstrDS = client.getResponseAsString(
        // lstrDSFetchURL,
        // true,true);
        // } catch (Exception e)
        // {
        // }
        //    
        // return lstrDS;

    }

    public String getDatastreamAsString(String lstrPID, String dsID,
            String asOfDateTime) throws RemoteException, IOException {
        String lstrValue = "";
        // Call Fedora Access SOAP service to request dissemination.
        MIMETypedStream dsDissemination = null;
        dsDissemination = getDatastreamDissemination(lstrPID, dsID,
                asOfDateTime);
        int lintLength = dsDissemination.getStream().length;
        ByteArrayInputStream dissemResult = new ByteArrayInputStream(
                dsDissemination.getStream());
        byte[] buffer = new byte[lintLength];
        dissemResult.read(buffer);
        lstrValue = new String(buffer);
        return lstrValue;
    }

    public MIMETypedStream getDissemination(String pid, String bDefPid,
            String methodName, Property[] parameters, String asOfDateTime)
            throws RemoteException {
        FedoraAPIA myAPIA;
        try {
            myAPIA = client.getAPIA();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }
        return myAPIA.getDissemination(pid, bDefPid, methodName, parameters,
                asOfDateTime);
    }

    public FedoraClient getFedoraClient() {
        return client;
    }

    public String getNextPID() throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        String lstrVal[] = myAPIM.getNextPID(new NonNegativeInteger("1"),
                "nsdl");

        return lstrVal[0];
    }

    public String[] getNextPID(String sNumPids, String pidNamespace)
            throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        NonNegativeInteger lnumPids = new NonNegativeInteger(sNumPids);
        return myAPIM.getNextPID(lnumPids, pidNamespace);
    }

    public String[] getObjectHistory(String pid) throws RemoteException {
        FedoraAPIA myAPIA;
        try {
            myAPIA = client.getAPIA();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }
        return myAPIA.getObjectHistory(pid);
    }

    public ObjectProfile getObjectProfile(String pid, String asOfDateTime)
            throws RemoteException {
        FedoraAPIA myAPIA;
        try {
            myAPIA = client.getAPIA();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }
        return myAPIA.getObjectProfile(pid, asOfDateTime);
    }

    public byte[] getObjectXML(String pid) throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        // make the SOAP call on API-M
        byte[] objectXML = myAPIM.getObjectXML(pid);

        // print results

        return objectXML;
    }

    public URL getQueryURL() {
        return murlQueryURL;
    }

    public String ingest(InputStream ingestStream, String ingestFormat,
            String logMessage) throws RemoteException, IOException {

        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "entered .ingest", "");
        }

        // prep
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        pipeStream(ingestStream, out, 4096);

        // make the SOAP call on API-M using the connection stub
        String pid = myAPIM.ingest(out.toByteArray(), ingestFormat, logMessage);

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "leaving .ingest", pid);
        }

        return pid;
    }

    /**
     * ingests a byte array of xml into a fedora repository object. Less
     * overhead than InputStream method. Useful for small object ingest.
     * 
     * @param ingestArray
     *            byte array
     * @param ingestFormat
     * @param logMessage
     * @return PID of new object
     * @throws RemoteException
     * @throws IOException
     */
    public String ingest(byte[] ingestArray, String ingestFormat,
            String logMessage) throws RemoteException, IOException {

        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "entered .ingest", "");
        }

        // make the SOAP call on API-M using the connection stub
        String pid = myAPIM.ingest(ingestArray, ingestFormat, logMessage);

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "leaving .ingest", pid);
        }

        return pid;
    }

    public DatastreamDef[] listDatastreams(String pid, String asOfDateTime)
            throws RemoteException {
        FedoraAPIA myAPIA;
        try {
            myAPIA = client.getAPIA();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        return myAPIA.listDatastreams(pid, asOfDateTime);
    }

    public ObjectMethodsDef[] listMethods(String pid, String asOfDateTime)
            throws RemoteException {
        FedoraAPIA myAPIA;
        try {
            myAPIA = client.getAPIA();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        return myAPIA.listMethods(pid, asOfDateTime);
    }

    public String modifyDatastreamByReference(String pid, String dsID,
            String[] altIDs, String dsLabel, boolean versionable,
            String dsMIME, String formatURI, String dsLocation, String dsState,
            String logMessage, boolean force) throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "entered .modifyDatastreamByReference", pid
                    + " ds: " + dsID);
        }

        // make the SOAP call on API-M using the connection stub
        /*
         * Note: assumes no checksumming. XXX: State, versionability are
         * discarded here
         */
        String datastreamID = myAPIM.modifyDatastreamByReference(pid, dsID,
                altIDs, dsLabel, dsMIME, formatURI, dsLocation, null, null,
                logMessage, force);

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "leaving .modifyDatastreamByReference", pid
                    + " ds: " + dsID);
        }

        return datastreamID;
    }

    public String modifyDatastreamByValue(String pid, String dsID,
            String[] altIDs, String dsLabel, boolean versionable,
            String MIMEType, String formatURI, String dsContent,
            String dsState, String logMessage, boolean force)
            throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();

        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        String lstrResult = null;

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "entered .modifyDatastreamByValue", pid
                    + " ds: " + dsID);
        }

        try {
            /* Note: discarding versionability, state */
            lstrResult = myAPIM.modifyDatastreamByValue(pid, dsID, altIDs,
                    dsLabel, MIMEType, formatURI, dsContent.getBytes("UTF-8"),
                    null, null, logMessage, force);

            if (TRACE_SYSTEM_OUT) {
                LogUtility.log(ME, "leaving .modifyDatastreamByValue", pid
                        + " ds: " + dsID);
            }
        } catch (UnsupportedEncodingException e) {
            throw new RemoteException(e.getMessage(), e);
        }
        return lstrResult;

    }

    public synchronized String modifyDatastreamByValueIfDifferent(String pid,
            String dsID, String[] altIDs, String dsLabel, boolean versionable,
            String MIMEType, String formatURI, String dsContent,
            String dsState, String logMessage, boolean force)
            throws RemoteException {
        String lstrCurrentVal = "";

        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();

        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        String lstrResult;

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "entered .modifyDatastreamByValue...", pid
                    + " ds: " + dsID);
        }

        try {
            lstrCurrentVal = this.getDatastreamAsString(pid, dsID, null);
        } catch (IOException e) {
            throw new RemoteException(e.getMessage(), e);
        }

        // synchronized(mclsXMLInputFactory)
        {
            try {
                /* Creating XMLStreamReaders from strings */
                XMLStreamReader xmlReader1 = mclsXMLInputFactory
                        .createXMLStreamReader(new StringReader(lstrCurrentVal));
                XMLStreamReader xmlReader2 = mclsXMLInputFactory
                        .createXMLStreamReader(new StringReader(dsContent));
                if (XMLStreamComparator.equals(xmlReader1, xmlReader2)) {
                    if (TRACE_SYSTEM_OUT) {
                        LogUtility.log(ME,
                                "leaving .modifyDatastreamByValue...", pid
                                        + " ds: " + dsID + " (no change)");
                    }

                    logger.debug("datastreams are equal.  No modify necessary");
                    return "new value = current value.  Update not required";
                }
            } catch (Exception e) {
                throw new RemoteException(e.getMessage(), e);
            }
        }

        try {
            logger.debug("Content is different.  Actually modifying by value");

            /* XXX: discarding versionability, state */
            lstrResult = myAPIM.modifyDatastreamByValue(pid, dsID, altIDs,
                    dsLabel, MIMEType, formatURI, dsContent.getBytes("UTF-8"),
                    null, null, logMessage, force);
        } catch (UnsupportedEncodingException e) {
            throw new RemoteException(e.getMessage(), e);
        }

        if (TRACE_SYSTEM_OUT) {
            LogUtility.log(ME, "leaving .modifyDatastreamByValue...", pid
                    + " ds: " + dsID + " (changed)");
        }

        return lstrResult;

    }

    public String modifyObject(String pid, String state, String label,
            String logMessage) throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        /* XXX: using null ownerID by default */
        return myAPIM.modifyObject(pid, state, label, null, logMessage);
    }

    public String[] purgeDatastream(String pid, String dsID, String endDT,
            String logMessage, boolean force) throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        return myAPIM
                .purgeDatastream(pid, dsID, null, endDT, logMessage, force);
    }

    public String purgeObject(String pid, String logMessage, boolean force)
            throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        return myAPIM.purgeObject(pid, logMessage, force);
    }

    public FieldSearchResult resumeFindObjects(String sessionToken)
            throws RemoteException {
        FedoraAPIA myAPIA;
        try {
            myAPIA = client.getAPIA();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        return myAPIA.resumeFindObjects(sessionToken);
    }

    public void riFlush() throws Exception {
        try {
            URL lurlFlusher = new URL(murlQueryURL.getProtocol(), murlQueryURL
                    .getHost(), murlQueryURL.getPort(), murlQueryURL.getPath()
                    + "?flush=true");

            InputStreamReader lrdrContent = new InputStreamReader(lurlFlusher
                    .openStream());
            // Not sure how to tell if this failed....
            // BufferedReader content = new BufferedReader(lrdrContent);
            // StringBuffer lstbContent = new StringBuffer();
            // for (String line = content.readLine(); line != null; line =
            // content.readLine())
            // {
            // lstbContent.append(line);
            // }
            // content.close();
            lrdrContent.close();

        } catch (Exception e) {
            String lstrErMsg = e.getMessage();
            e.printStackTrace();
            throw new Exception("ERROR: Could not flush ri.  \n" + lstrErMsg, e);
        }
    }

    public String setDatastreamState(String pid, String dsID, String dsState,
            String logMessage) throws RemoteException {
        FedoraAPIM myAPIM;
        try {
            myAPIM = client.getAPIM();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage(), e);
        }

        return myAPIM.setDatastreamState(pid, dsID, dsState, logMessage);
    }


    // //////////////////////////////////////////////////////////////////////////
    // C L A S S M E T H O D S (other than fedora api calls)
    // //////////////////////////////////////////////////////////////////////////
    /**
     * Copies the contents of an InputStream to an OutputStream, then closes
     * both.
     * 
     * @param in
     *            The source stream.
     * @param out
     *            The target stram.
     * @param bufSize
     *            Number of bytes to attempt to copy at a time.
     * @throws IOException
     *             If any sort of read/write error occurs on either stream.
     */
    private static void pipeStream(InputStream in, OutputStream out, int bufSize)
            throws IOException {
        try {
            byte[] buf = new byte[bufSize];
            int len;
            while ((len = in.read(buf)) > 0) {
                out.write(buf, 0, len);
            }
        } finally {
            try {
                in.close();
                out.close();
            } catch (IOException e) {
                System.err.println("WARNING: Could not close stream.");
            }
        }

    }
}
