/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.util.HashSet;
import java.util.Set;

import org.nsdl.repository.model.Datastream;

import fedora.common.Constants;

public interface FedoraDatastream extends Datastream {
    /* Accessors for datastream properties */

    /** Datastream {@link State} */
    public State getState();

    /** Content {@link ControlGroup} (storage model) */
    public ControlGroup controlGroup();

    /**
     * Datastream label.
     * <p>
     * May be null id not defined. This is a Fedora detail that we may not want
     * in the NDR spec.
     * </p>
     * 
     */
    public String getLabel();

    /**
     * Alternate IDs .
     * <p>
     * Fedora detail that we may not want in the NDR spec.
     * </p>
     * 
     */
    public String[] getAltIDs();

    /**
     * Format URI.
     * <p>
     * May be null if not defined (which is usually the case). Fedora detail
     * that we may not want in the NDR spec.
     * </p>
     * 
     */
    public String getFormatURI();

    /**
     * True if the datastream is versionable .
     * <p>
     * Fedora detail that we may not want in the NDR spec..
     * </p>
     */
    public boolean isVersionable();

    /** Indicates the physical nature the Datastream storage. */
    public static enum ControlGroup {

        /**
         * Externally managed content - distributed by repository, but content
         * may not actually be hosted there
         */
        External("E"),

        /** Redirected content - data does not physically exist in repository */
        Redirect("R"),

        /** Managed content - stored in the repository */
        Managed("M"),

        /** Inline XML */
        InlineXML("X");

        private static final Set<String> vals = new HashSet<String>();

        /* Get the set of string values for checks in getInstance() */
        static {
            for (ControlGroup c : ControlGroup.values()) {
                vals.add(c.toString());
            }
        }

        private final String foxmlName;

        private ControlGroup(String representation) {
            this.foxmlName = representation;
        }

        public String toString() {
            return foxmlName;
        }

        /** Get a control group instance by its name */
        public static ControlGroup getInstance(String name) {
            if (!vals.contains(name)) {
                throw new IllegalArgumentException("Unknown control group "
                        + name);
            }

            for (ControlGroup c : ControlGroup.values()) {
                if (c.toString().equals(name))
                    return c;
            }

            /* Should never get here */return null;
        }
    };

    /** Indicates the datastream state. */
    public static enum State {
        Inactive("I"), Active("A"), Deleted("D");

        private final String foxmlName;

        private static final Set<String> vals = new HashSet<String>();
        static {
            for (State s : State.values()) {
                vals.add(s.toString());
            }
        }

        private State(String representation) {
            this.foxmlName = representation;
        }

        public String toString() {
            return foxmlName;
        }

        public static State getInstance(String name) {
            if (!vals.contains(name)) {
                throw new IllegalArgumentException("Unknown state " + name);
            }

            for (State s : State.values()) {
                if (s.toString().equals(name))
                    return s;
            }

            /* Should never get here */return null;
        }

        public static State fromURI(String uri) {
            if (Constants.MODEL.ACTIVE.uri.equals(uri))
                return Active;
            else if (Constants.MODEL.INACTIVE.uri.equals(uri))
                return Inactive;
            else if (Constants.MODEL.DELETED.uri.equals(uri))
                return Deleted;

            throw new IllegalArgumentException("Unknown state URI " + uri);
        }
    }
}
