/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.util.HashMap;
import java.util.Map;

import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.impl.fedora.FedoraDatastream.ControlGroup;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.Configuration.Listener;

/**
 * Creates datastreams that have a Fedora storage control group.
 * <p>
 * Fedora datastreams are created under the following rules, in the following
 * order
 * <ul>
 * <li>If the input datastream is already a Fedora-aware datastream (i.e.
 * existing datastream that originates from Fedora), leave it alone</li>
 * <li>If the ID of the input datastream matches a table of pre-determined
 * control groups, use the control group from that table.</li>
 * <li>If the URL of the input datastream is not null, give it a 'redirect'
 * control group.</li>
 * <li>Otherwise, by default use a 'Managed' control group.
 * </ul>
 * 
 * The pre-defined table of datastream IDs to control groups is determined by
 * properties in the ndr config files: *
 * <dt><code>param.repository.impl.fedora.datastream.controlgroup.{ID}</code></dt>
 * <dd>Determines the default controlgroup for a datastream named
 * <code>ID</code>. Acceptable values are: {Managed, InlineXML, Redirect}</dd>
 * </p>
 * 
 * <p>
 * Example:
 * 
 * <pre>
 * param.repository.impl.fedora.datastream.controlgroup.format_nsdl_dc = InlineXML
 * param.repository.impl.fedora.datastream.controlgroup.format_oai_dc = InlineXML
 * param.repository.impl.fedora.datastream.controlgroup.IMAGE = Managed
 * param.repository.impl.fedora.datastream.controlgroup.URL = Redirect
 * </pre>
 * 
 * </p>
 * 
 * @author birkland
 * @since 1.3
 */
public class FedoraDatastreamFactory implements Listener {
    private static final String CONTROL_GROUP_PREFIX = "param.repository.impl.fedora.datastream.controlgroup.";

    private Map<String, ControlGroup> m_controlGroups = new HashMap<String, ControlGroup>();

    public FedoraDatastream toFedoraDatastream(Datastream d, ControlGroup cg) {
        switch (cg) {
        case Managed:
            return new ManagedDatastream(d.getName(), d.getContent(), d.getLabel(), d
                    .getMimeType());
        case Redirect:
            return new RedirectDatastream(d.getName(), d.getURL(), d.getLabel(), d
                    .getMimeType());
        case InlineXML:
            return new XMLDatastream(d.getName(), d.getContent(), d.getLabel(), d
                    .getMimeType());
        case External:
            throw new AccessException("External control group not supported");
        default:
            throw new AccessException("Unknown control group " + cg);
        }
    }

    /**
     * Create a fedora datastream from a given source
     * 
     * @param d
     *            Source datastream
     * @return configured FedoraDatastream.
     */
    public FedoraDatastream toFedoraDatastream(Datastream d) {
        /*
         * If we already have a fedora datastream (e.g. from an existing fedora
         * object, etc), just cast it as so
         */
        if (d instanceof FedoraDatastream) {
            return (FedoraDatastream) d;
        }

        /* Default control group is managed */
        ControlGroup controlGroup = ControlGroup.Managed;

        if (m_controlGroups.containsKey(d.getName().getID())) {
            /*
             * If we know the desired control group based on a pre-configured
             * mapping, use the configuration to determine the correct class
             */
            controlGroup = m_controlGroups.get(d.getName().getID());
        } else if (!d.isCanonical()) {
            controlGroup = ControlGroup.Redirect;
        }

        return toFedoraDatastream(d, controlGroup);
    }

    public void updateConfig(Configuration config) {
        Map<String, ControlGroup> groups = new HashMap<String, ControlGroup>();
        for (Map.Entry<String, String> property : config.getPropertyMap(
                CONTROL_GROUP_PREFIX).entrySet()) {
            String dsID = property.getKey().replaceFirst(CONTROL_GROUP_PREFIX,
                    "");
            ControlGroup group = ControlGroup.valueOf(property.getValue());
            groups.put(dsID, group);
        }
        m_controlGroups = groups;
    }
}
