/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.util.Iterator;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.Finder;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Resource;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.NDRObjectTemplate;

/**
 * Finds NDR objects in a local Fedora instance
 * <p>
 * Uses access to the local triple store, etc to implement findidng objects.
 * Instances of this class are configured to search a single running Fedora
 * instance.
 * </p>
 * 
 * @author birkland
 * 
 */
public class FedoraNDRObjectFinder implements Finder {

    private static final Logger logger = Logger
            .getLogger(FedoraNDRObjectFinder.class);

    private final FedoraNDRAccess fedora;

    public FedoraNDRObjectFinder(FedoraNDRAccess ndr) {
        this.fedora = ndr;
    }

    /**
     * Find and return a resource given a URL
     * 
     * @param url
     *            The URL of the resource to find
     * @return instance of the matching resource, null if none are found
     */
    public Resource findResource(Property.Name idType, String url) {
        NDRObjectTemplate resourceTemplate = new NDRObjectTemplate(
                Type.Resource);
        resourceTemplate.addProperty(idType, url);

        /* There will be zero or one results */

        logger.trace("Finding resource " + idType.getLocalPart() + ": '" + url
                + ";");
        Results<NDRObjectInfo> results = matchObject(resourceTemplate);

        try {
            Iterator<NDRObjectInfo> r = results.iterator();
            Resource found;
            if (r.hasNext()) {
                found = fedora.getObject(r.next(), Type.Resource);
                if (r.hasNext()) {
                    throw new RepositoryException(
                            "More than one resource result found for "
                                    + idType.getLocalPart() + ": '" + url + "'");
                }

                return found;
            }
        } finally {
            results.close();
        }

        return null;
    }

    public Results<NDRObjectInfo> matchObject(NDRObject template) {
        return new FedoraObjectMatcher(fedora.getRepository(), template);
    }

    public Results<Relationship> getRelatedTo(NDRObject relatedTo) {
        return new InboundRelationshipFinder(relatedTo, fedora);
    }

    public Results<Relationship> getRelatedTo(NDRObject relatedTo,
            Relationship.Name relType) {
        return new InboundRelationshipFinder(relatedTo, relType, fedora);
    }
}
