/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.mptstore.query.component.BasicNodeFilter;
import org.nsdl.mptstore.query.component.BasicNodePattern;
import org.nsdl.mptstore.query.component.GraphPattern;
import org.nsdl.mptstore.query.component.GraphQuery;
import org.nsdl.mptstore.rdf.Node;
import org.nsdl.mptstore.util.NTriplesUtil;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.NDRObjectTemplate;
import org.nsdl.repository.util.StaticObjectInfo;

import fedora.common.Constants;
import fedora.common.PID;

import static org.nsdl.repository.impl.fedora.ri.TripleUtils.getPattern;
import static org.nsdl.repository.impl.fedora.ri.TripleUtils.nTripleEncodeLiteral;
import static org.nsdl.repository.impl.fedora.ri.TripleUtils.nTripleEncodeURI;

/**
 * Matches objects based on a template.
 * 
 * <p>
 * Given an object to use as a template (another object, or an
 * {@link NDRObjectTemplate} will work), will attempt to find all objects in the
 * repository that match it, where match is defined as the following:
 * <ul>
 * <li>Any {@link Property properties} in the template must be present in, and
 * have the same values as, those in a matching object.</li>
 * <li>Any {@link Relationship relationships} in the template must be present
 * in, and have the same values as, those in a matching object.</li>
 * <li>If defined, the {@link org.nsdl.repository.model.NDRObject.State state}
 * of the template, if defined, must be equal to that of a matching object.</li>
 * <li>If defined, the {@link org.nsdl.repository.model.types.Type object type}
 * of the template must be equal to that of a matching object.</li>
 * </ul>
 * Soon, we will be adding:
 * <ul>
 * <li>Any datastream in the template must be present in the matching object,
 * but their contents need not be the same.
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class FedoraObjectMatcher implements Results<NDRObjectInfo> {
    private final FedoraAccess fedora;

    private final Set<ResultIterator> usedResults = new HashSet<ResultIterator>();
    private final GraphQuery query;

    private static final Logger logger = Logger
            .getLogger(FedoraObjectMatcher.class);

    /**
     * Create a FedoraObject Matcher with a given match template.
     * 
     * @param fedora
     *            Repository to search within.
     * @param template
     *            Template object to match.
     */
    public FedoraObjectMatcher(FedoraAccess fedora, NDRObject template) {
        this.fedora = fedora;
        try {
            this.query = getQuery(template);
        } catch (ParseException e) {
            throw new RuntimeException(e);
        }
    }

    public Iterator<NDRObjectInfo> iterator() {
        try {
            ResultIterator results = new ResultIterator(query);
            usedResults.add(results);
            return results;
        } catch (QueryException e) {
            throw new RuntimeException(e);
        }
    }

    public void close() {
        for (ResultIterator i : new ArrayList<ResultIterator>(usedResults)) {
            i.close();
        }
    }

    private class ResultIterator implements Iterator<NDRObjectInfo> {
        private final QueryResults results;
        private NDRObjectInfo next;

        public ResultIterator(GraphQuery q) throws QueryException {
            results = fedora.getRI().execute(query,
                    Arrays.asList("$handle", "$type"));
            advance();
        }

        public void close() {
            results.close();
            usedResults.remove(this);
        }

        public void remove() {
            throw new UnsupportedOperationException("remove() not implemented");
        }

        public NDRObjectInfo next() {
            NDRObjectInfo toReturn = next;
            if (next != null) {
                advance();
            }
            return toReturn;
        }

        public boolean hasNext() {
            return next != null;
        }

        private void advance() {
            next = null;
            if (results.hasNext()) {
                List<Node> result = results.next();
                String declaredType = result.get(1).getValue();
                String declaredHandle = result.get(0).getValue();
                try {
                    next = new StaticObjectInfo(declaredHandle, Type
                            .getInstance(declaredType));

                } catch (Type.UnknownTypeException e) {
                    logger.debug("Skipping declaration of type " + declaredType
                            + " in " + declaredHandle);
                    advance();
                }
            } else {
                close();
            }
        }
    }

    private GraphQuery getQuery(NDRObject template) throws ParseException {
        GraphQuery query = new GraphQuery();
        GraphPattern pattern = new GraphPattern();
        logger.trace("Getting request query");

        /* Make sure to select the handle of the object (we'll be returning it) */
        pattern.addTriplePattern(getPattern("$object",
                nTripleEncodeURI(Components.HAS_HANDLE.getURI()), "$handle"));

        /*
         * Make sure to select the type of the object (we'll be returning it
         * too)
         */
        pattern.addTriplePattern(getPattern("$object",
                nTripleEncodeURI(Components.OBJECT_TYPE.getURI()), "$type"));

        /* Restrict results to the desired type, if specified in the request */
        if (template.type() != null) {
            logger.debug("Explicit type given: " + template.type());
            pattern.addFilter(new BasicNodeFilter<Node>(
                    new BasicNodePattern<Node>("$type"), "=",
                    new BasicNodePattern<Node>(NTriplesUtil.parseLiteral("\""
                            + template.type() + "\""))));
        } else {
            logger.debug("No state constraint given");
        }

        /*
         * Restrict to desired state, if specified.
         */
        if (template.state() != null) {
            logger.debug("Explicit state given: " + template.state());
            pattern.addTriplePattern(getPattern("$object",
                    nTripleEncodeURI(Constants.MODEL.STATE.uri),
                    nTripleEncodeURI(template.state().getURI())));
        }

        /*
         * For every property EXCEPT objectType and Handle, add a triple
         * pattern: $object <propertyURI> "value"
         * 
         * NB: All objects of properties are treated as literals
         */
        for (Property p : template.getProperties()) {
            if (!Components.OBJECT_TYPE.equals(p.predicate())
                    && !Components.HAS_HANDLE.equals(p.predicate())) {
                logger.debug("Will match property: " + p);
                pattern.addTriplePattern(getPattern("$object",
                        nTripleEncodeURI(p.predicate().getURI()),
                        nTripleEncodeLiteral(p.value())));
            } else if (Components.HAS_HANDLE.equals(p.predicate())) {
                logger.debug("Will match handle " + p.value());
                /* Filter by handle */

                pattern
                        .addFilter(new BasicNodeFilter<Node>(
                                new BasicNodePattern<Node>("$handle"), "=",
                                new BasicNodePattern<Node>(NTriplesUtil
                                        .parseLiteral(nTripleEncodeLiteral(p
                                                .value())))));
            }
        }

        for (Relationship r : template.getRelationships()) {

            if ((r.subject() != null && r.subject().handle() != null)
                    && (r.related() == null || r.related().handle() == null)) {
                /* We have <related_object> <predicate> <this_object> */
                logger.debug("Will match external relationship: " + r);
                PID relatedPID = FedoraUtils.getPidFromHandle(fedora, r
                        .subject().handle());
                if (relatedPID == null) {
                    return new GraphQuery();
                }

                pattern.addTriplePattern(getPattern(nTripleEncodeURI(relatedPID
                        .toURI()), nTripleEncodeURI(r.predicate().getURI()),
                        "$object"));

            } else {
                /*
                 * We have <this_object> <predicate> <related_object> (most
                 * common case)
                 */

                logger.debug("Will match internal relationship: " + r);
                PID relatedPID = FedoraUtils.getPidFromHandle(fedora, r
                        .related().handle());
                if (relatedPID == null) {
                    return new GraphQuery();
                }
                pattern.addTriplePattern(getPattern("$object",
                        nTripleEncodeURI(r.predicate().getURI()),
                        nTripleEncodeURI(relatedPID.toURI())));
            }
        }

        query.addRequired(pattern);

        logger.trace("done getting query");

        return query;
    }
}
