/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora;

import java.io.IOException;
import java.net.URL;
import java.text.ParseException;
import java.util.ArrayList;

import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.multipart.FilePart;
import org.apache.commons.httpclient.methods.multipart.MultipartRequestEntity;
import org.apache.commons.httpclient.methods.multipart.Part;
import org.apache.log4j.Logger;

import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.mptstore.query.component.BasicNodePattern;
import org.nsdl.mptstore.query.component.BasicTriplePattern;
import org.nsdl.mptstore.query.component.GraphPattern;
import org.nsdl.mptstore.query.component.GraphQuery;
import org.nsdl.mptstore.rdf.ObjectNode;
import org.nsdl.mptstore.rdf.PredicateNode;
import org.nsdl.mptstore.rdf.SubjectNode;
import org.nsdl.mptstore.util.NTriplesUtil;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.DatastreamPartSource;
import org.nsdl.repository.util.StreamUtil;

import fedora.common.PID;

public abstract class FedoraUtils {
    private static final Logger logger = Logger.getLogger(FedoraUtils.class
            .getName());

    /**
     * Finds a specific PID given a handle.
     * 
     * @param repository
     *            Access to local Fedora instance.
     * @param handle
     *            Complete handle as a string.
     * 
     * @return PID of matching object.
     * 
     * @throws ParseException
     *             if the handle string is malformed
     */
    public static PID getPidFromHandle(FedoraAccess repository, String handle)
            throws ParseException {
        return getPIDFromHandle(repository, handle, null);
    }

    /**
     * Finds a specific PID given a handle and specified object type
     * 
     * @param repository
     *            FedoraAccess of repository to seach for pid
     * @param handle
     *            Complete handle as a string
     * @param objectType
     *            Type constraint on object (e.g. Aggregator, MetadataProvider)
     * @return PID of matching object
     * 
     * @throws ParseException
     *             if the handle string is malformed
     */
    public static PID getPIDFromHandle(FedoraAccess repository, String handle,
            Type<? extends NDRObject> objectType) throws ParseException {
        GraphQuery query = new GraphQuery();
        GraphPattern g = new GraphPattern();

        /* Add the handle constraint */
        g.addTriplePattern(new BasicTriplePattern(
                new BasicNodePattern<SubjectNode>("$pid"),
                new BasicNodePattern<PredicateNode>(NTriplesUtil
                        .parsePredicate("<" + Constants.REPOS_NAMESPACE
                                + "hasHandle>")),
                new BasicNodePattern<ObjectNode>(NTriplesUtil.parseObject("\""
                        + handle + "\""))));

        /* Add the objecType constraint */
        if (objectType != null) {
            g.addTriplePattern(new BasicTriplePattern(
                    new BasicNodePattern<SubjectNode>("$pid"),
                    new BasicNodePattern<PredicateNode>(NTriplesUtil
                            .parsePredicate("<" + Constants.REPOS_NAMESPACE
                                    + "objectType>")),
                    new BasicNodePattern<ObjectNode>(NTriplesUtil
                            .parseObject("\"" + objectType + "\""))));
        }

        query.addRequired(g);
        ArrayList<String> targets = new ArrayList<String>();
        targets.add("$pid");

        QueryResults results;
        try {
            results = repository.getRI().execute(query, targets);
        } catch (QueryException e) {
            throw new RuntimeException("Could not execute RI query", e);
        }

        if (results.hasNext()) {
            PID result = PID.getInstance(results.next().get(0).getValue());

            /* Make sure we have exactly one */
            if (results.hasNext()) {
                throw new RepositoryException("Found too many objects with "
                        + " handle " + handle);
            }
            results.close();
            return result;
        } else {
            /* No matches found.. */
            results.close();
            return null;
        }

    }

    /**
     * Finds and returns a handle for a given PID
     * 
     * @param repository
     *            Access to the repository containing the object
     * @param pid
     *            PID of the object to find a handle for
     * @return handle as a string
     */
    public static String getHandleFromPID(FedoraAccess repository, PID pid) {
        GraphQuery query = new GraphQuery();
        GraphPattern g = new GraphPattern();

        if (pid == null) {
            return null;
        }

        try {
            g.addTriplePattern(new BasicTriplePattern(
                    new BasicNodePattern<SubjectNode>(NTriplesUtil
                            .parseSubject("<" + pid.toURI() + ">")),
                    new BasicNodePattern<PredicateNode>(NTriplesUtil
                            .parsePredicate("<" + Constants.REPOS_NAMESPACE
                                    + "hasHandle>")),
                    new BasicNodePattern<ObjectNode>("$handle")));
        } catch (ParseException e) {
            /* Should never get here */
            throw new RuntimeException("Parse failed", e);
        }

        query.addRequired(g);
        ArrayList<String> targets = new ArrayList<String>();
        targets.add("$handle");

        QueryResults results = null;
        try {
            results = repository.getRI().execute(query, targets);

            /* If there are multiple handles, a random one will be returned */
            if (results.hasNext()) {
                String handle = results.next().get(0).getValue();
                return handle;
            }
        } catch (QueryException e) {
            throw new RepositoryException("Could not execute query", e);
        } finally {
            if (results != null) {
                results.close();
            }
        }
        return null;
    }

    /**
     * Finds and returns a handle for a given object PID
     */
    public static String getHandleFromPID(URL lurlQuery, String PID)
            throws Exception {

        /* XXX Would be nice to have generic risearch functionality */
        logger.info("SPO: getting handle from " + PID);

        /*
         * XXX once we move to using PID objects exclusively, this check won't
         * be necessary
         */
        if (PID.startsWith(Constants.FEDORA_INFO_URI)) {
            PID = PID.replaceFirst(Constants.FEDORA_INFO_URI, "");
        }

        String query;
        java.net.URL url;
        java.io.BufferedReader content = null;
        try {
            /* hack to protect agaist PID URI argumens through.. */
            PID = PID.replaceFirst(Constants.FEDORA_INFO_URI, "");
            query = java.net.URLEncoder
                    .encode(
                            "<info:fedora/"
                                    + PID
                                    + "> "
                                    + "<http://ns.nsdl.org/api/relationships#hasHandle> "
                                    + "*", "US-ASCII");
        } catch (java.io.UnsupportedEncodingException e) {
            throw new Exception("Could not encode handle query", e);
        }
        try {
            url = new java.net.URL(
                    lurlQuery.getProtocol(),
                    lurlQuery.getHost(),
                    lurlQuery.getPort(),
                    Constants.RESOURCE_INDEX_URL
                            + "type=triples&lang=spo&format=N-Triples&stream=on&query="
                            + query);

            content = new java.io.BufferedReader(new java.io.InputStreamReader(
                    url.openStream()));
        } catch (java.net.MalformedURLException e) {
            if (content != null)
                content.close();
            throw new Exception("Could formulate query", e);
        } catch (java.io.IOException e) {
            if (content != null)
                content.close();
            throw new Exception("Could not connect Fedora.  Try again", e);
        }

        String result = content.readLine();
        content.close();

        logger.debug("SPO: Got result " + result);

        if (result == null) {
            logger
                    .warn("SPO:Recieved null respose from server when lookung up pid "
                            + PID + " at " + url);
            throw new RepositoryException("Cannot find handle for given PID "
                    + PID);
        }

        if (result.contains("\"")) {
            String[] parts = result.split("\"");
            return parts[1];
        } else {
            if (content != null)
                content.close();
            throw new Exception("Cannot find handle for given pid" + result);
        }
    }

    /**
     * Add a datastream
     * <p>
     * </p>
     * 
     * @param repo
     *            Wrapper to repository
     * @param pid
     *            PID of object to have datastream added
     * @param ds
     *            Datastream definition
     * @param message
     *            Message to log upon add
     */
    public static void addDatastream(FedoraAccess repo, PID pid,
            FedoraDatastream ds, String message) {
        logger.trace("Adding datastream " + ds.getName());
        String loc = null;
        try {

            /*
             * If content is an actual payload (and not a URL), we need to add
             * it by reference only XXX: Check this!
             */
            if (ds.controlGroup()
                    .equals(FedoraDatastream.ControlGroup.Redirect)
                    || ds.controlGroup().equals(
                            FedoraDatastream.ControlGroup.External)) {
                logger.trace("control group is " + ds.controlGroup()
                        + ", uploading by value");
                loc = StreamUtil.streamToString(ds.getContent());
            } else {
                logger.trace("control group is " + ds.controlGroup()
                        + ", uploading by reference");
                loc = uploadContent(repo, ds);
            }

            logger.trace("Adding datastream");
            repo.getAPI().addDatastream(pid.toString(), ds.getName().getID(),
                    ds.getAltIDs(), ds.getLabel(), ds.isVersionable(),
                    ds.getMimeType(), ds.getFormatURI(), loc,
                    ds.controlGroup().toString(), ds.getState().toString(),
                    message);
        } catch (Exception e) {
            logger.trace("Fail, " + e.getMessage());
            throw new RepositoryException("Could not add datastream "
                    + ds.getName(), e);
        }
        logger.trace("Success");
    }

    /** modify a datastream */
    public static void modifyDatastream(FedoraAccess repo, PID pid,
            FedoraDatastream ds, String message) {

        try {

            switch (ds.controlGroup()) {

            /*
             * These three are all modified by value. In all three cases,
             * datastream content will be a text string. In the case of
             * InlineXML, that text string contains XML. In the case of external
             * or redirect, the text string contains a URL
             * 
             * XXX: Don't know if we can do Redirect/external this way
             */
            case InlineXML:
            case External:
            case Redirect: {
                modifyDatastreamByValue(repo, pid, ds, message);
            }
                break;

            /* Managed datastreams are always passed by reference */
            case Managed: {
                modifyDatastreamByReference(repo, pid, ds, message);
            }

            }
        } catch (Exception e) {
            throw new RepositoryException("Could not modify datastream", e);
        }
    }

    /** Delete a datastream */
    public static void deleteDatastream(FedoraAccess repo, PID pid,
            FedoraDatastream ds, String message) {
        try {
            repo.getAPI().purgeDatastream(pid.toString(), ds.getName().getID(),
                    null, message, false);
        } catch (Exception e) {
            throw new RepositoryException("Could not delete datastream", e);
        }
    }

    /* InlineXML is modified by value, and is a String */
    private static void modifyDatastreamByValue(FedoraAccess repo, PID pid,
            FedoraDatastream ds, String message) throws Exception {

        repo.getAPI().modifyDatastreamByValue(pid.toString(),
                ds.getName().getID(), ds.getAltIDs(), ds.getLabel(),
                ds.isVersionable(), ds.getMimeType(), ds.getFormatURI(),
                StreamUtil.streamToString(ds.getContent()),
                ds.getState().toString(), message, true);
    }

    /* loc is either a URL or an opaque identifier */
    private static void modifyDatastreamByReference(FedoraAccess repo, PID pid,
            FedoraDatastream ds, String message) throws Exception {

        String location = uploadContent(repo, ds);

        logger.trace("Modifying datastream by reference");
        repo.getAPI().modifyDatastreamByReference(pid.toString(),
                ds.getName().getID(), ds.getAltIDs(), ds.getLabel(),
                ds.isVersionable(), ds.getMimeType(), ds.getFormatURI(),
                location, ds.getState().toString(), message, true);
        logger.trace("SUCCESS");
    }

    /** Uploads datastream content to Fedora and returns opaque identifier */
    private static String uploadContent(FedoraAccess repository, Datastream ds)
            throws IOException {
        String loc = null;

        String URL = repository.getAPI().getClient().getUploadURL();

        logger.trace("Uploading content to " + URL);

        PostMethod dsPost = new PostMethod(URL);
        dsPost.getParams().setParameter("Connection", "Keep-Alive");

        /* In case datastream content is huge */
        dsPost.setContentChunked(true);

        Part[] parts = { new FilePart("file", new DatastreamPartSource(ds)) };

        dsPost.setRequestEntity(new MultipartRequestEntity(parts, dsPost
                .getParams()));

        try {
            int result = repository.getHttpClient().executeMethod(dsPost);

            if (result != 201) {
                throw new RepositoryException(""
                        + "Uploading datastream content failed, code " + result);
            } else {
                loc = dsPost.getResponseBodyAsString().trim();
                logger.debug("successfully uploaded ds contents of "
                        + ds.getName());
            }
        } finally {
            if (dsPost != null) {
                dsPost.releaseConnection();
            }
        }

        logger.trace("Content is stored at location " + loc);
        return loc;
    }

    /* XXX: Bad thing to do, probably */
    public static boolean isFedoraIdentifier(String id) {
        try {
            PID.getInstance(id);
            return true;
        } catch (Exception e) {
            return false;
        }
    }
}