/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import org.apache.log4j.Logger;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.impl.fedora.FedoraDatastream.ControlGroup;

import static org.nsdl.repository.util.Encoder.xmlEncode;

/**
 * Serialized a datastream into foxml form
 * <p>
 * Presents a foxml representation of an inline XML, Refirect, or External
 * datastream. Used in creating foxml appropriate for ingest. As such, does not
 * represent multiple versions (always assume the datastream is the first
 * revision) , nor may is represent "out of band" datastreams such as those with
 * Managed content
 * </p>
 * 
 * @author birkland
 * 
 */
public class FoxmlDatastreamWriter implements DOPartWriter {

    private final FedoraDatastream datastream;

    private static final Logger log = Logger
            .getLogger(FoxmlDatastreamWriter.class);

    public FoxmlDatastreamWriter(FedoraDatastream d) {
        /* Right now, we don't handle managed datastreams */
        if (d.controlGroup().equals(ControlGroup.Managed)) {
            throw new IllegalArgumentException(
                    "Cannot serialize managed datastream");
        }

        this.datastream = d;
    }

    /** {@inheritDoc} */
    public String getRepresentation() {
        StringBuilder rep = new StringBuilder();
        rep.append("<foxml:datastream CONTROL_GROUP=\"");
        rep.append(datastream.controlGroup().toString());
        rep.append("\" ID=\"");
        rep.append(datastream.getName());
        rep.append("\" STATE=\"");
        rep.append(datastream.getState().toString());
        rep.append("\" VERSIONABLE=\"");
        rep.append(datastream.isVersionable());
        rep.append("\">\n");
        rep.append(" <foxml:datastreamVersion ID=\"");
        rep.append(datastream.getName());
        rep.append(".0\" ");
        if (datastream.getLabel() != null) {
            rep.append("LABEL=\"");
            rep.append(datastream.getLabel());
            rep.append("\" ");
        }
        rep.append("MIMETYPE=\"");
        rep.append(datastream.getMimeType());
        rep.append("\" SIZE=\"");
        rep.append(datastream.getSize());
        rep.append("\">\n");

        if (ControlGroup.InlineXML.equals(datastream.controlGroup())) {
            rep.append("  <foxml:xmlContent>\n");
            String line;
            BufferedReader in = null;
            try {
                in = new BufferedReader(new InputStreamReader(datastream
                        .getContent()));
                while ((line = in.readLine()) != null) {
                    rep.append(line + "\n");
                }
            } catch (Exception e) {
                throw new RepositoryException(
                        "Could not serialize XML datastream "
                                + datastream.getName());
            } finally {
                if (in != null) {
                    try {
                        in.close();
                    } catch (IOException e) {
                        log.warn("Could not datastream reader", e);
                    }
                }
            }
            rep.append("  </foxml:xmlContent>\n");
        } else if (ControlGroup.Redirect.equals(datastream.controlGroup())
                || ControlGroup.External.equals(datastream.controlGroup())) {
            rep.append("  <foxml:contentLocation TYPE=\"URL\" REF=\"");
            rep.append(xmlEncode(datastream.getURL().toString()));
            rep.append("\" />\n");
        } else {
            /* Should not get here, unless something major changes */
            throw new RuntimeException("Unknown datastream type "
                    + datastream.controlGroup());
        }

        rep.append(" </foxml:datastreamVersion>\n</foxml:datastream>\n");
        return rep.toString();
    }

}
