/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora;

import java.net.URI;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.impl.fedora.FedoraDatastream.ControlGroup;

import fedora.common.Constants;
import fedora.common.Models;
import fedora.common.PID;

/**
 * Build a Foxml 1.1 object from scratch for adding to a repository
 * <p>
 * Upon commit, ingests into the NDR, and uploads any datastreams that cannot be
 * defined in the initial foxml (e.g. managed binary content);
 * </p>
 * 
 * @author birkland
 * 
 */
public class FoxmlObjectBuilder extends FedoraMutableObjectBase implements
        DOPartWriter {

    private State state = State.Active;

    /* XXX: Hardcoded for now */
    public static final Map<Type<? extends NDRObject>, URI> m_modelMap = new HashMap<Type<? extends NDRObject>, URI>();
    static {
        m_modelMap
                .put(Type.Agent, URI.create("info:fedora/ndr:cmodel_Agent-1"));
        m_modelMap.put(Type.Aggregator, URI
                .create("info:fedora/ndr:cmodel_Aggregator-1"));
        m_modelMap.put(Type.Metadata, URI
                .create("info:fedora/ndr:cmodel_Metadata-1"));
        m_modelMap.put(Type.MetadataProvider, URI
                .create("info:fedora/ndr:cmodel_MetadataProvider-1"));
        m_modelMap.put(Type.Resource, URI
                .create("info:fedora/ndr:cmodel_Resource-1"));
    }

    private final FedoraNDRAccess ndr;

    public FoxmlObjectBuilder(Type<? extends NDRObject> type, PID pid,
            FedoraNDRAccess ndr) {

        if (ndr == null) {
            throw new NullPointerException("NDR is null!");
        }

        if (ndr == null) {
            throw new NullPointerException("NDR is null!");
        }

        this.ndr = ndr;

        /** Just use the pid as the handle, if one is specified */
        if (pid != null) {
            initNew(type, pid.toString(), pid);
        } else {
            initNew(type, null, null);
        }
    }

    private static final Logger logger = Logger
            .getLogger(FoxmlObjectBuilder.class.getName());

    public Date getCreatedDate() {
        return new Date();
    }

    public Date getLastModifiedDate() {
        return new Date();
    }

    public void commit() {
        HashMap<ControlGroup, Set<FedoraDatastream>> sortedDS = sortDatastreamsByControlGroup();

        if (this.pid() == null) {
            generatePID();
        }

        logger.trace("Building FOXML");
        String foxml = getRepresentation();
        try {
            logger.trace("Ingesting");
            ndr.getRepository().getAPI().ingest(foxml.getBytes("UTF-8"),
                    Constants.FOXML1_1.uri, "Object added by NSDL API");
        } catch (Exception e) {
            logger.debug("Foxml from failed load: " + foxml);
            throw new RepositoryException("Could not ingest object", e);
        }

        /*
         * Add managed datasreams XXX: This would perhaps be better served by a
         * concurrent uploader service rathern than serially here
         */
        logger.trace("BEGIN uploading managed datastreams");
        try {
            for (FedoraDatastream d : sortedDS.get(ControlGroup.Managed)) {
                FedoraUtils.addDatastream(ndr.getRepository(), pid(), d,
                        "Datastream added by NSDL API");
            }
        } catch (Exception e) {
            ndr.purgeObject(this);
            throw new AccessException(
                    "Failed adding object: error storing managed content");
        }
        logger.trace("DONE uploading managed datastreams");
    }

    /** Sort datastreams by control group */
    private HashMap<ControlGroup, Set<FedoraDatastream>> sortDatastreamsByControlGroup() {
        HashMap<ControlGroup, Set<FedoraDatastream>> sortedDS = new HashMap<ControlGroup, Set<FedoraDatastream>>();

        /* Create empty buckets for datastreams by controlgroup */
        for (ControlGroup g : ControlGroup.values()) {
            sortedDS.put(g, new HashSet<FedoraDatastream>());
        }

        /* Sort 'n add */
        for (Datastream d : getDatastreams()) {
            FedoraDatastream fd = getFedoraDatastream(d.getName().getID());
            addTo(sortedDS, fd.controlGroup(), fd);
        }
        return sortedDS;
    }

    /* XXX: Somewhat NSDL-specific. Need to make this configurable */
    private void generatePID() {
        if (this.handle() != null) {
            PID pid = PID.getInstance("hdl:"
                    + this.handle().replace("/", "%2F"));
            setPID(pid);
        } else {
            try {
                PID pid = PID.getInstance(ndr.getRepository().getAPI()
                        .getNextPID());
                this.setPID(pid);
            } catch (RemoteException e) {
                throw new RepositoryException("Could not get new pid: "
                        + e.getMessage(), e);
            }
        }
    }

    public String getRepresentation() {

        StringBuilder foxmlBuilder = new StringBuilder();

        HashMap<ControlGroup, Set<FedoraDatastream>> sortedDS = sortDatastreamsByControlGroup();

        generatePreamble(foxmlBuilder, this.pid());
        generateFoxmlProperties(foxmlBuilder, type(), state);
        generateRelsExt(pid(), foxmlBuilder, type(), getProperties(),
                getRelationships());
        generateFoxmlDatastreams(foxmlBuilder, sortedDS);
        close(foxmlBuilder);

        return foxmlBuilder.toString();
    }

    private static <K, V> void addTo(HashMap<K, Set<V>> container, K key,
            V value) {
        if (container.get(key) == null) {
            container.put(key, new HashSet<V>());
        }
        container.get(key).add(value);
    }

    /** Initial foxml tags */
    private static void generatePreamble(StringBuilder preamble, PID pid) {

        /* Open up the initial foxml tag + ns, schema declatations */
        preamble.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        preamble.append("<foxml:digitalObject VERSION=\"1.1\" PID=\"");
        preamble.append(pid.toString());
        preamble
                .append("\" xmlns:foxml=\"info:fedora/fedora-system:def/foxml#\" ");
        preamble
                .append("xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" ");
        preamble
                .append("xsi:schemaLocation=\"info:fedora/fedora-system:def/foxml# "
                        + "http://www.fedora.info/definitions/1/0/foxml1-1.xsd\">\n");
    }

    /** Create foxml-level object properties */
    private static void generateFoxmlProperties(StringBuilder foxml,
            Type<? extends NDRObject> type, State state) {

        /* Declare that this is an object with the given state */
        foxml.append(" <foxml:objectProperties>\n");

        foxml.append("  <foxml:property NAME="
                + "\"info:fedora/fedora-system:def/model#state\" " + "VALUE=\""
                + state.getFoxmlName() + "\" />\n");
        foxml.append("  <foxml:property NAME="
                + "\"info:fedora/fedora-system:def/model#label\" " + "VALUE=\""
                + type + "\" />\n");

        foxml.append(" </foxml:objectProperties>\n");
    }

    /** Generate the RELS-EXT datastream from NDR properties, relationships */
    private void generateRelsExt(PID pid, StringBuilder foxml,
            Type<? extends NDRObject> type, Collection<Property> props,
            Collection<Relationship> rels) {

        /* Create a new RELS-EXT writer */
        RelsExtBuilder builder = new RelsExtBuilder(pid, ndr);

        /* Add all properties and relationships */
        builder.addProperties(props);
        builder.addRelationships(rels);
        builder.addModel(Models.FEDORA_OBJECT_3_0.getURI());
        builder.addModel(m_modelMap.get(type));

        /* Write to foxml */
        foxml.append(builder.getRepresentation());
    }

    /** Insert any in-foxml datastreams (inlineXML, redirect, external) */
    private static void generateFoxmlDatastreams(StringBuilder foxml,
            HashMap<ControlGroup, Set<FedoraDatastream>> dsMap) {

        /* First, pick out datastreams to write */
        List<FedoraDatastream> dsToWrite = new ArrayList<FedoraDatastream>();
        dsToWrite.addAll(dsMap.get(ControlGroup.InlineXML));
        dsToWrite.addAll(dsMap.get(ControlGroup.Redirect));
        dsToWrite.addAll(dsMap.get(ControlGroup.External));

        /* Now write them to the foxml */
        for (FedoraDatastream inline : dsToWrite) {
            FoxmlDatastreamWriter dsWriter = new FoxmlDatastreamWriter(inline);
            foxml.append(dsWriter.getRepresentation());
        }
    }

    public FedoraDatastream getFedoraDatastream(String id) {
        return ndr.getDatastreamFactory().toFedoraDatastream(getDatastream(id));
    }

    private static void close(StringBuilder foxml) {
        foxml.append("</foxml:digitalObject>");
    }
}
