/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.log4j.Logger;
import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.mptstore.rdf.Node;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.content.AuthorizationModel;
import org.nsdl.repository.model.content.CommonModel;

import fedora.common.PID;

import static org.nsdl.repository.impl.fedora.ri.RIQueryEngine.Language.SPO;

/**
 * Finds relationships pointing "in" to a given Fedora object.
 * 
 * Inward relationships are not present in within the object the point to, so
 * they need to be discovered via the ResourceIndex. This class makes the
 * appropriate queries to the ResourceIndex and returns all discovered
 * inward-pointing relationships.
 * 
 * @author birkland
 * @since 1.1
 */
public class InboundRelationshipFinder implements Results<Relationship> {

    private static final Logger logger =
            Logger.getLogger(InboundRelationshipFinder.class);

    private final Set<RelationshipIterator> usedResults =
            new HashSet<RelationshipIterator>();
    
    private final Map<String, Relationship.Name> predicates =
        new ConcurrentHashMap<String, Relationship.Name>();
    
    private final FedoraNDRObject target;

    private final FedoraNDRAccess fedora;

    private final Relationship.Name relationship;

    /**
     * Create a finder that will search for all inward relationships to an
     * object.
     * 
     * @param target
     *            NDRObject that matching relationships will point to.
     * @param fedora
     *            FedoraNDRAccess to search within.
     */
    public InboundRelationshipFinder(NDRObject target, FedoraNDRAccess fedora) {
        logger.debug("NOT USING object parameter");
        this.target = fedora.getObject(target);
        this.fedora = fedora;
        this.relationship = null;
    }

    /**
     * Create a finder that will search for all instances of a specific
     * relationship to an object.
     * 
     * @param target
     *            NDRObject that matching relationships will point to.
     * @param relationship
     *            Specific relationship to look for.
     * @param fedora
     *            FedoraNDRAccess to search within.
     */
    public InboundRelationshipFinder(NDRObject target,
            Relationship.Name relationship, FedoraNDRAccess fedora) {
        logger.trace("USING relationship parameter: " + relationship);
        this.target = fedora.getObject(target);
        this.fedora = fedora;
        this.relationship = relationship;
    }

    /** {@inheritDoc} */
    public void close() {
        for (RelationshipIterator i : new ArrayList<RelationshipIterator>(
                usedResults)) {
            i.close();
        }
    }

    /** {@inheritDoc} */
    public Iterator<Relationship> iterator() {
        String type = "";
        if (relationship != null) {
            type = " of type " + relationship.toString();
        }
        logger.trace("BEGIN finding relationships related to "
                + target.handle() + type);
        try {
            RelationshipIterator results = new RelationshipIterator();
            usedResults.add(results);
            return results;
        } catch (QueryException e) {
            throw new RepositoryException("Could not execute query", e);
        }
    }

    private class RelationshipIterator implements Iterator<Relationship> {
        private final QueryResults contents;

        public RelationshipIterator() throws QueryException {
            String query;
            if (relationship == null) {
                query = "* * " + "<" + target.pid().toURI() + ">";
            } else {
                query =
                        "* " + "<" + relationship.getURI() + "> " + "<"
                                + target.pid().toURI() + ">";
            }
            this.contents = fedora.getRepository().getRI().execute(query, SPO);
        }

        public void close() {
            contents.close();
            usedResults.remove(this);
        }

        public void remove() {
            throw new UnsupportedOperationException("remove() not supported");
        }

        public boolean hasNext() {
            if (contents.hasNext()) {
                return true;
            } else {
                logger.trace("DONE finding relationships");
                close();
                return false;
            }
        }

        public Relationship next() {
            List<Node> nextResult = contents.next();

            FedoraNDRObject subject =
                    fedora.getObject(PID.getInstance(nextResult.get(0)
                            .getValue()));

            Relationship.Name predicate =
                    parseQName(nextResult.get(1).getValue(), subject);
            return new Relationship(subject, predicate, target);
        }

        private Relationship.Name parseQName(String uri, NDRObject subject) {

            logger.debug("Finding qname for " + uri);
            if (predicates.containsKey(uri)) {
                return predicates.get(uri);
            }

            Relationship.Name newPredicate = null;
            if (uri.contains(CommonModel.NS)) {
                String localPart = uri.replace(CommonModel.NS, "");
                newPredicate =
                        new Relationship.Name(CommonModel.NS,
                                localPart, CommonModel.PREFIX);

            } else if (uri.contains(AuthorizationModel.NS)) {
                String localPart = uri.replace(AuthorizationModel.NS, "");
                newPredicate =
                        new Relationship.Name(AuthorizationModel.NS,
                                localPart, AuthorizationModel.PREFIX);

            } else {
                logger.info("Unknown predicate " + uri + "looking up...");
                for (Relationship r : subject.getRelationships()) {
                    if (r.predicate().getURI().equals(uri)) {
                        newPredicate = r.predicate();
                    }
                }
            }

            predicates.put(uri, newPredicate);
            return newPredicate;
        }
    }
}
