/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.mptstore.query.component.BasicNodePattern;
import org.nsdl.mptstore.query.component.BasicTriplePattern;
import org.nsdl.mptstore.query.component.GraphPattern;
import org.nsdl.mptstore.query.component.GraphQuery;
import org.nsdl.mptstore.query.component.NodePattern;
import org.nsdl.mptstore.rdf.Node;
import org.nsdl.mptstore.rdf.ObjectNode;
import org.nsdl.mptstore.rdf.PredicateNode;
import org.nsdl.mptstore.rdf.SubjectNode;
import org.nsdl.mptstore.util.NTriplesUtil;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.impl.fedora.ri.PreparedQuery;
import org.nsdl.repository.model.Datastream;

import fedora.common.Constants;
import fedora.common.PID;

/**
 * Reads datastreams from Fedora using lazy evaluation when possible
 * 
 * @author birkland
 * 
 */
public class LazyAPIADatastreamReader implements DatastreamReader {

    private final FedoraAccess repo;
    private static final HashMap<FedoraAccess, PreparedQuery> dsIDQuery = new HashMap<FedoraAccess, PreparedQuery>();
    private Set<String> m_datastreamIDs;
    private final PID pid;

    public LazyAPIADatastreamReader(FedoraAccess repo, PID pid) {
        this.pid = pid;
        this.repo = repo;
    }

    public Datastream readDatastream(String dsID) {
        if (getDatastreamIDs().contains(dsID)) {
            return new DeferredAPIADatastream(dsID, pid, repo);
        } else {
            return null;
        }
    }

    public Iterable<Datastream> readDatastreams() {
        ArrayList<Datastream> datastreams = new ArrayList<Datastream>();
        for (String dsID : getDatastreamIDs()) {
            datastreams.add(new DeferredAPIADatastream(dsID, pid, repo));
        }
        return datastreams;
    }

    /** Query the RI to find of the current object's datastream IDs */
    private Set<String> getDatastreamIDs() {

        if (m_datastreamIDs == null) {
            m_datastreamIDs = new HashSet<String>();
            PreparedQuery p = dsIDQuery.get(repo);
            if (p == null) {
                synchronized (dsIDQuery) {
                    dsIDQuery.put(repo, getDSQuery());
                    p = dsIDQuery.get(repo);
                }
            }

            try {
                QueryResults results = repo.getRI().execute(p,
                        Arrays.asList("<" + pid.toURI() + ">"));
                while (results.hasNext()) {
                    List<Node> result = results.next();

                    String ds = result.get(0).getValue().replace(
                            pid.toURI() + "/", "");

                    if (result.get(1).toString().contains("/*/" + ds + ">")) {
                        m_datastreamIDs.add(ds);
                    }
                }
                results.close();
            } catch (QueryException e) {
                throw new RuntimeException(e);
            }
        }
        return m_datastreamIDs;
    }

    /** Prepare an MPTStore query for returning all datastreamIDs */
    private PreparedQuery getDSQuery() {
        GraphQuery g = new GraphQuery();
        GraphPattern p = new GraphPattern();
        try {
            NodePattern<SubjectNode> subject = new BasicNodePattern<SubjectNode>(
                    NTriplesUtil.parseSubject(PreparedQuery.PLACEHOLDER));
            NodePattern<PredicateNode> pred = new BasicNodePattern<PredicateNode>(
                    NTriplesUtil.parsePredicate("<"
                            + Constants.VIEW.DISSEMINATES + ">"));
            NodePattern<ObjectNode> object = new BasicNodePattern<ObjectNode>(
                    "$datastream");

            p.addTriplePattern(new BasicTriplePattern(subject, pred, object));

            subject = new BasicNodePattern<SubjectNode>("$datastream");
            pred = new BasicNodePattern<PredicateNode>(NTriplesUtil
                    .parsePredicate("<" + Constants.VIEW.DISSEMINATION_TYPE
                            + ">"));
            object = new BasicNodePattern<ObjectNode>("$dissType");
            p.addTriplePattern(new BasicTriplePattern(subject, pred, object));

            subject = new BasicNodePattern<SubjectNode>("$datastream");
            pred = new BasicNodePattern<PredicateNode>(NTriplesUtil
                    .parsePredicate("<" + Constants.MODEL.STATE.uri + ">"));
            object = new BasicNodePattern<ObjectNode>("<"
                    + Constants.MODEL.ACTIVE.uri + ">");

            p.addTriplePattern(new BasicTriplePattern(subject, pred, object));

            g.addRequired(p);

            return repo.getRI().prepare(g,
                    Arrays.asList("$datastream", "$dissType"));
        } catch (Exception e) {
            throw new RepositoryException(e);
        }
    }

    /*
     * N.B. Here's a non-lazy way to do it for (repo.server.types.gen.Datastream
     * d : repo.getAPI() .getDatastreams(pid.toString(), null,
     * Datastream.State.Active.toString())) {
     */
}
