/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.impl.fedora.ri.RIQueryEngine;
import org.nsdl.repository.impl.fedora.ri.RIRelsReader;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.oai.OAIConstants;
import org.nsdl.repository.util.ReadOnlyObject;
import org.nsdl.util.DateUtility;
import org.nsdl.util.DateUtilityException;

import fedora.common.PID;

/**
 * NDR object that dynamically loads content from Fedora upon request
 * <p>
 * Creates an NDR object representation that is bound to a specific object in a
 * Fedora repository. Information is loaded from the only upon request, and is
 * subsequently cached for rapid access. Consequently, the life of these objects
 * should be relatively short lived: once accessed, it is unaware of any changes
 * to the underlying data. Thus: do not cache. Not thread safe
 * </p>
 * 
 * @author birkland
 */
public class LazyFedoraBackedObject extends ReadOnlyObject implements
        FedoraNDRObject {

    private static final Property.Name lastModified = new Property.Name(
            fedora.common.Constants.VIEW.uri, "lastModifiedDate", "fedora-view");

    private static final Property.Name created = new Property.Name(
            fedora.common.Constants.MODEL.uri, "createdDate", "fedora-model");

    private final FedoraNDRAccess repo;

    private final PID pid;

    private final String handle;

    private Type<? extends NDRObjectInfo> type;

    private final RelsReader m_relReader;

    private final DatastreamReader m_dsReader;

    private static final Logger logger = Logger
            .getLogger(LazyFedoraBackedObject.class.getName());

    /**
     * Create a Lazy, dynamic NDR object
     * 
     * @param identifier
     *            Handle, PID, or Fedora URI of an NDR object
     * @param repository
     *            Fedora NDR repository
     */
    public LazyFedoraBackedObject(String identifier, FedoraNDRAccess repository)
            throws ExistenceException {
        this.repo = repository;

        if (identifier.startsWith(Constants.FEDORA_INFO_URI)) {
            this.pid = PID.getInstance(identifier);
            this.handle = getHandle(pid);
        } else if (identifier.contains(":")) {
            /* Hack: encoded slash handling is a bit ugly */
            this.pid = PID.getInstance(identifier.replace("/", "%2F"));
            this.handle = getHandle(pid);
        } else {
            this.handle = identifier;

            /* Get the pid from handle */
            try {
                this.pid = FedoraUtils.getPidFromHandle(repo.getRepository(),
                        handle);
            } catch (ParseException e) {
                throw new ExistenceException("Malformed handle " + handle, e);
            }
        }
        if (this.pid == null || !repository.getRepository().objectExists(pid)) {
            throw new ExistenceException("No object with handle " + identifier);
        }
        m_relReader = new RIRelsReader(this, repo);
        m_dsReader = new LazyAPIADatastreamReader(repo.getRepository(), pid);
    }

    public LazyFedoraBackedObject(PID pid, FedoraNDRAccess ndr) {
        this.repo = ndr;
        this.pid = pid;
        this.handle = getHandle(pid);

        if (this.pid == null || !ndr.getRepository().objectExists(pid)) {
            throw new ExistenceException("No object with handle "
                    + pid.toString());
        }
        m_relReader = new RIRelsReader(this, repo);
        m_dsReader = new LazyAPIADatastreamReader(repo.getRepository(), pid);
    }

    public Relationship getRelationship(Relationship.Name predicate)
            throws AccessException {
        Set<Relationship> rels = getRelationships(predicate);

        if (rels.size() > 1) {
            throw new AccessException("More than one relationship named "
                    + predicate.toString() + " was found in " + handle());
        } else if (rels.size() == 0) {
            return null;
        }

        /* We know at this point that there is exactly one relationship here */
        for (Relationship r : rels) {
            return r;
        }

        return null;
    }

    public Property getProperty(Property.Name predicate) {
        Set<Property> props = getProperties(predicate);

        if (props.size() > 1) {
            throw new RuntimeException("More than one property named "
                    + predicate.toString() + " was found in " + handle());
        } else if (props.size() == 0) {
            return null;
        }

        /* We know at this point that there is exactly one relationship here */
        for (Property r : props) {
            return r;
        }

        return null;
    }

    public Set<Relationship> getRelationships() {
        return new SetReader<Relationship>(m_relReader.readRelationships());
    }

    public Set<Property> getProperties() {
        return new SetReader<Property>(m_relReader.readProperties());
    }

    public Set<Relationship> getRelationships(Relationship.Name predicate) {
        return new SetReader<Relationship>(m_relReader
                .readRelationships(predicate));
    }

    public Set<Property> getProperties(Property.Name predicate) {
        return new SetReader<Property>(m_relReader.readProperties(predicate));
    }

    public Type<? extends NDRObject> type() {
        if (this.type != null) {
            return type;
        }

        logger.trace("BEGIN polling for type");
        Set<String> types = new HashSet<String>();
        for (Property type : getProperties(Components.OBJECT_TYPE)) {
            types.add(type.value());
        }

        if (types.contains(Constants.OBJECT_TYPE_RESOURCE)) {
            this.type = Type.Resource;
        }

        if (types.contains(Constants.OBJECT_TYPE_METADATA_PROVIDER)) {
            this.type = Type.MetadataProvider;
        }

        if (types.contains(Constants.OBJECT_TYPE_METADATA)) {
            this.type = Type.Metadata;
        }

        if (types.contains(Constants.OBJECT_TYPE_AGGREGATOR)) {
            this.type = Type.Aggregator;
        }

        if (types.contains(Constants.OBJECT_TYPE_AGENT)) {
            this.type = Type.Agent;
        }

        if (this.type == null) {
            /* XXX Hack, for now */
            if (!OAIConstants.XSL_OBJECT_PID.equals(pid.toString())) {
                throw new RuntimeException(
                        "Could not deduce object type from RELS-EXT for "
                                + this.pid);
            }
        }

        logger.trace("END polling type for " + pid);
        return this.type;
    }

    public String handle() {
        return this.handle;
    }

    public PID pid() {
        return this.pid;
    }

    public Set<Datastream> getDatastreams() {
        return new SetReader<Datastream>(m_dsReader.readDatastreams());
    }

    public Datastream getDatastream(Datastream.Name dsID) {
        return getDatastream(dsID.getID());
    }

    public Datastream getDatastream(String dsID) {
        return m_dsReader.readDatastream(dsID);
    }

    public Date getLastModifiedDate() {
        try {
            return DateUtility.getDateFromISO8601Milisecond(getProperty(
                    lastModified).value());
        } catch (DateUtilityException e) {
            throw new RuntimeException(e);
        }
    }

    public Date getCreatedDate() {
        try {
            return DateUtility
                    .getDateFromISO8601Milisecond(getProperty(created).value());
        } catch (DateUtilityException e) {
            throw new RuntimeException(e);
        }
    }

    public NDRObject.State state() {
        QueryResults results;
        try {
            results = repo.getRepository().getRI().execute(
                    "<" + pid.toURI() + "> <"
                            + fedora.common.Constants.MODEL.STATE.uri + "> *",
                    RIQueryEngine.Language.SPO);
        } catch (QueryException e) {
            throw new RuntimeException("Could not determine state", e);
        }

        try {
            if (results.hasNext()) {
                NDRObject.State state = State.getInstanceFromURI(results.next()
                        .get(2).getValue());
                return state;
            } else {
                throw new RepositoryException("<" + pid.toURI() + "> <"
                        + fedora.common.Constants.MODEL.STATE.uri + "> *");
            }
        } finally {
            results.close();
        }

    }

    private String getHandle(PID pid) throws ExistenceException {
        String handle = FedoraUtils.getHandleFromPID(repo.getRepository(), pid);
        if (handle == null) {
            handle = pid.toString();
        }
        return handle;
    }

    public FedoraDatastream getFedoraDatastream(String id) {
        return repo.getDatastreamFactory()
                .toFedoraDatastream(getDatastream(id));
    }

    public void setPID(PID pid) {
        throw new UnsupportedOperationException("Object is read-only");
    }

    private static class SetReader<T> implements Set<T> {
        private final Iterable<T> m_src;

        public SetReader(Iterable<T> src) {
            m_src = src;
        }

        public Iterator<T> iterator() {
            return m_src.iterator();
        }

        @SuppressWarnings("unchecked")
        public boolean removeAll(Collection c) {
            return false;
        }

        @SuppressWarnings("unchecked")
        public boolean retainAll(Collection c) {
            return false;
        }

        public boolean add(Object arg0) {
            return false;
        }

        @SuppressWarnings("unchecked")
        public boolean addAll(Collection arg0) {
            return false;
        }

        public void clear() {
            throw new UnsupportedOperationException("Read-Only set");
        }

        public boolean contains(Object obj) {
            for (T member : m_src) {
                if (member.equals(obj)) {
                    return true;
                }
            }

            return false;
        }

        @SuppressWarnings("unchecked")
        public boolean containsAll(Collection coll) {
            candidates: for (Object candidate : coll) {
                for (T member : m_src) {
                    if (member.equals(candidate)) {
                        continue candidates;
                    }
                }
                return false;
            }
            return true;
        }

        public boolean isEmpty() {
            return m_src.iterator().hasNext();
        }

        public boolean remove(Object arg0) {
            return false;
        }

        public int size() {
            int count = 0;
            for (Iterator<T> i = m_src.iterator(); i.hasNext(); i.next()) {
                count++;
            }
            return count;
        }

        public Object[] toArray() {
            ArrayList<Object> array = new ArrayList<Object>();
            for (T member : m_src) {
                array.add(member);
            }
            return array.toArray();
        }

        public <E> E[] toArray(E[] arg0) {
            ArrayList<Object> array = new ArrayList<Object>();
            for (T member : m_src) {
                array.add(member);
            }
            return array.toArray(arg0);
        }
    }
}
