/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.impl.MutationListener;
import org.nsdl.repository.model.types.Type;

import fedora.common.Models;
import fedora.common.PID;

/**
 * Allows valid modifications to NDR objects via fedora API-M
 * <p>
 * Adding/deleting/modifying object components causes no immediate modifications
 * to the object state in the NDR, but instead queues actions for execution upon
 * commit. Upon commit()
 * <ul>
 * <li>Adds, deletes, modifies any {@link org.nsdl.repository.model.Datastream
 * Datastream} or {@link org.nsdl.repository.impl.fedora.Disseminator
 * Disseminator} components defined</li>
 * <li>Modifies the RELS-EXT datastream based on
 * {@link org.nsdl.repository.model.Property Property} or
 * {@link org.nsdl.repository.model.Relationship Relationship} components.</li>
 * <li>Changes the object state, if required</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 */
public class MutableAPIMFedoraObject extends FedoraMutableObjectBase {

    private static final Logger logger = Logger
            .getLogger(MutableAPIMFedoraObject.class.getName());

    private final FedoraNDRAccess ndr;

    private boolean isNew = false;

    private LazyFedoraBackedObject origObject;

    private Map<String, Datastream> replacementDatastreams = new HashMap<String, Datastream>();

    /**
     * Creates an Fedora-bound {@link NDRObject} instance
     * <p>
     * If <b>create</b> is true, will create an object with no initial contents,
     * otherwise will poll the NDR for the given
     * {@link org.nsdl.repository.model.NDRObjectInfo} for initial content.
     * </p>
     * 
     * @param ndr
     *            Access to Fedora NDR repository where object will be bound
     * @param objectInfo
     *            Identifies the object to be bound
     */
    public MutableAPIMFedoraObject(FedoraNDRAccess ndr, NDRObjectInfo objectInfo)
            throws ExistenceException {
        this.ndr = ndr;

        /* This is not a new object, so populate with content */
        logger.trace("Creating from NDRObjectInfo");

        if (objectInfo == null) {
            throw new ExistenceException(
                    "Object identity is null.  Cannot open");
        }

        initFromID(ndr, objectInfo.handle());

        /* Just being paranoid... */
        if (objectInfo.type() != null
                && !objectInfo.type().equals(origObject.type())) {
            throw new IllegalArgumentException("Error opening object "
                    + objectInfo.handle() + " Specified type is "
                    + objectInfo.type() + ", but actual type is "
                    + origObject.type());
        }
    }

    /**
     * Create a mutable Fedora object given a type and handle.
     * 
     * @param ndr
     *            Access to Fedora NDR repository where object will be bound
     * @param type
     *            {@link org.nsdl.repository.model.types.Type} of object
     * @param handle
     *            Handle of object
     */
    public MutableAPIMFedoraObject(FedoraNDRAccess ndr,
            Type<? extends NDRObject> type, String handle) {
        this.ndr = ndr;

        /* This is not a new object, so populate with content */
        logger.trace("Creating from handle and type");

        origObject = new LazyFedoraBackedObject(handle, ndr);

        init(this);
        getDatastreamTracker().setListener(new DatastreamPropagator());
    }

    public MutableAPIMFedoraObject(FedoraNDRAccess ndr, PID pid) {
        this.ndr = ndr;
        initFromID(ndr, pid.toString());
    }

    public void initFromID(FedoraNDRAccess ndr, String id) {
        origObject = new LazyFedoraBackedObject(id, ndr);

        init(origObject);

        getDatastreamTracker().setListener(new DatastreamPropagator());
    }

    public Date getLastModifiedDate() {

        return new Date(origObject.getLastModifiedDate().getTime());
    }

    public Date getCreatedDate() {
        return new Date(origObject.getCreatedDate().getTime());
    }

    public void commit() {

        boolean isModified = false;
        logger.trace("Committing changes");

        /* Remove any deleted datastreams */
        for (Datastream removed : getDatastreamTracker().getDeleted()) {
            isModified = true;
            logger.debug("removing datastream " + removed.getName());
            FedoraUtils.deleteDatastream(ndr.getRepository(), pid(), ndr
                    .getDatastreamFactory().toFedoraDatastream(removed),
                    "Datastream removed by NSDL API");
        }

        /* Modify any modified (replaced) datastreams */
        for (Datastream modified : replacementDatastreams.values()) {
            isModified = true;
            logger.debug("modifying datastream " + modified.getName());
            FedoraUtils
                    .modifyDatastream(ndr.getRepository(), pid(), ndr
                            .getDatastreamFactory().toFedoraDatastream(
                                    modified,
                                    origObject.getFedoraDatastream(
                                            modified.getName().getID())
                                            .controlGroup()),
                            "Datastream modified by NSDL API");
        }

        /* Add any new datastreams */
        for (Datastream added : getDatastreamTracker().getAdded()) {
            isModified = true;
            logger.debug("adding datastream " + added.getName());
            FedoraUtils.addDatastream(ndr.getRepository(), pid(), ndr
                    .getDatastreamFactory().toFedoraDatastream(added),
                    "Datastream added by NSDL API");
        }

        /* Modify RELS-EXT */
        /* TODO: don't do this here.. treat RELS-EXT just as another datastream? */
        if (getRelationshipTracker().hasModified()
                || getPropertyTracker().hasModified()) {
            isModified = true;
            logger.trace("Modifying RELS-EXT");
            RelsExtBuilder relsExt = new RelsExtBuilder(pid(), ndr);
            relsExt.addProperties(getProperties());
            relsExt.addRelationships(getRelationships());

            /* TODO: fix this so it's not so hard coded! */
            relsExt.addModel(Models.FEDORA_OBJECT_3_0.getURI());
            if (FoxmlObjectBuilder.m_modelMap.containsKey(type())) {
                relsExt.addModel(FoxmlObjectBuilder.m_modelMap.get(type()));
            }
            try {
                ndr.getRepository().getAPI().modifyDatastreamByValue(
                        pid().toString(), "RELS-EXT", new String[0],
                        "Relationships", true, "text/xml", "unknown",
                        relsExt.toString(), "A",
                        "RELS-EXT modified by NDR API", true);
            } catch (Exception e) {
                logger.warn("Could not modify RELS-EXT");
                throw new RepositoryException("Could not modify RELS-EXT", e);
            }
        }

        /*
         * Change object state, if state change is desired, or just touch object
         * if no other changes have been made
         */
        if ((getPreviousState() != null || !isModified) && !isNew) {

            String message;
            if (!isModified) {
                logger.info("No changes, but touching object anyway");
                message = "Touched by NSDL API";
            } else {
                message = "State modified by NSDL API";
            }
            try {
                logger.info("Modifying state of " + pid() + " to " + state());
                ndr.getRepository().getAPI().modifyObject(pid().toString(),
                        state().getFoxmlName(), type().toString(), message);
            } catch (Exception e) {
                logger.warn("Could not modify object state", e);
                throw new RepositoryException("Could not modify object state",
                        e);
            }
        }

        logger.trace("Success");
    }

    private class DatastreamPropagator implements MutationListener<Datastream> {
        public void notifyAdd(Datastream addedElement) {
        }

        public void notifyRemove(Datastream removedElement) {
        }

        public void notifyUpdate(Datastream newElement) {
            logger.trace("Notified of replaced datastream: "
                    + newElement.getName().getID());
            replacementDatastreams
                    .put(newElement.getName().getID(), newElement);
        }
    }

    public FedoraDatastream getFedoraDatastream(String id) {
        return ndr.getDatastreamFactory().toFedoraDatastream(getDatastream(id));
    }
}
