/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora;

import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.impl.fedora.ri.RIQueryEngine;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.content.CommonModel;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import fedora.common.PID;

/**
 * Reads properties and relationships from FOXML RDF
 * <p>
 * Will generate a set of relationships as parsed from the XML RDF as found in
 * FOXML objects
 * </p>
 */

public class RDFRelsReader implements RelsReader {

    private static final Logger logger = Logger.getLogger(RDFRelsReader.class);

    private final FedoraNDRAccess repository;

    private final NDRObject containingObject;

    private final Set<Property> properties = new HashSet<Property>();

    private final Set<Relationship> relationships = new HashSet<Relationship>();

    private static final String RDF_NS = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";

    /**
     * Read relationships from RDF within a byte stream
     * 
     * @param object
     *            The NDR object that is the subject of all the relationships
     * @param fedora
     *            The fedora repository in which the related objects reside
     * 
     */
    public RDFRelsReader(FedoraNDRObject object, FedoraNDRAccess fedora) {
        this.repository = fedora;
        this.containingObject = object;

        try {
            logger.trace("BEGIN parsing RDF rels:");

            parseRDFRelationships(NDRXMLutil.getDocument(object
                    .getFedoraDatastream("RELS-EXT").getContent(), true, false));
            logger.trace("END parsing RDF rels");
        } catch (Exception e) {
            throw new RepositoryException("Could not get relationships", e);
        }
    }

    public Iterable<Relationship> readRelationships() {
        return relationships;
    }

    public Iterable<Relationship> readRelationships(Relationship.Name type) {
        HashSet<Relationship> filteredRels = new HashSet<Relationship>();
        for (Relationship r : readRelationships()) {
            if (r.predicate().equals(type)) {
                filteredRels.add(r);
            }
        }
        return filteredRels;
    }

    public Iterable<Property> readProperties() {
        return properties;
    }

    public Iterable<Property> readProperties(Property.Name type) {
        HashSet<Property> filteredProperties = new HashSet<Property>();
        for (Property p : readProperties()) {
            if (p.predicate().equals(type)) {
                filteredProperties.add(p);
            }
        }
        return filteredProperties;
    }

    private void parseRDFRelationships(Node container) {
        Node rdf = NDRXMLutil.getNodeByName(container, RDF_NS, "Description");
        if (rdf == null) {
            throw new IllegalArgumentException(
                    "Could not find any RDF in input");
        }

        NodeList rels = rdf.getChildNodes();

        /* Would probably like to make this threadsafe at some point.. */
        for (int i = 0; i < rels.getLength(); i++) {
            Node child = rels.item(i);
            if (child.getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }

            String rdfResource = NDRXMLutil.getAttributeValue((Element) child,
                    RDF_NS, "resource");
            String rdfLiteral = child.getTextContent();

            if (rdfResource != null
                    && rdfResource.startsWith(Constants.FEDORA_INFO_URI)) {

                if (isNDRObject(repository.getRepository().getRI(), PID
                        .getInstance(rdfResource))) {

                    Relationship.Name name = null;
                    if (child.getPrefix() != null) {
                        name = new Relationship.Name(child.getNamespaceURI(),
                                child.getLocalName(), child.getPrefix());
                    } else {
                        name = new Relationship.Name(child.getNamespaceURI(),
                                child.getLocalName());
                    }

                    this.relationships.add(new Relationship(containingObject,
                            name, repository.getObject(PID
                                    .getInstance(rdfResource))));
                }
            } else if (rdfLiteral != null) {
                if (child.getPrefix() != null) {
                    this.properties.add(new Property(containingObject,
                            new Property.Name(child.getNamespaceURI(), child
                                    .getLocalName(), child.getPrefix()),
                            rdfLiteral));
                } else {
                    this.properties.add(new Property(containingObject,
                            new Property.Name(child.getNamespaceURI(), child
                                    .getLocalName()), rdfLiteral));
                }
            } else {
                throw new RuntimeException(
                        "Unexpected contents in object rels.  resource: "
                                + rdfResource + " literal:" + rdfLiteral);
            }
        }
    }

    public static boolean isNDRObject(RIQueryEngine ri, PID pid) {
        QueryResults results = null;
        try {
            results = ri.execute("<" + pid.toURI() + "> <"
                    + CommonModel.Components.OBJECT_TYPE.getURI() + "> *",
                    RIQueryEngine.Language.SPO);
            if (results.hasNext()) {
                return true;
            } else {
                return false;
            }
        } catch (QueryException e) {
            throw new RuntimeException(e);
        } finally {
            if (results != null) {
                results.close();
            }
        }
    }
}
