/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.io.InputStream;
import java.net.URL;

import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.model.Datastream;

/**
 * Represents a datastream that is a redirect to another URL.
 * <p>
 * Perhaps most commonly used for resource content in an overlay model, this
 * datastream may be specified solely in terms of an external URL. As one would
 * perhaps expect, {@link #getContent() getContents()} will return an
 * {@linkplain InputStream InputStream} containing the binary content as it
 * exists from the external source, though underneath an HTTP connection is made
 * to actually retrieve the contents, as it does not exist in the repository
 * itself.
 * </p>
 * 
 * @author birkland
 * @since 1.0
 */
public class RedirectDatastream implements FedoraDatastream {

    private URL url;
    private final Datastream.Name dsID;
    private String mimeType = "text/html";
    private String label;
    /**
     * Initialize a redirect datastream given a URL.
     * 
     * @param dsId
     *            Datastream name.
     * @param url
     *            URL source for the redirected content.
     * @param mimeType
     *            MIME type of the destination content. Use null if not known.
     */
    public RedirectDatastream(Datastream.Name dsId, URL url, String fileName,
            String mimeType) {
        this.url = url;
        if (mimeType != null) {
            this.mimeType = mimeType;
        }
        this.dsID = dsId;
        label = fileName;
    }

    public ControlGroup controlGroup() {
        return ControlGroup.Redirect;
    }

    public String[] getAltIDs() {
        return null;
    }

    public void setContent(InputStream content) {
        throw new UnsupportedOperationException("Not implemented");
    }

    public InputStream getContent() throws AccessException {
        throw new UnsupportedOperationException("Not implemented");
    }

    public InputStream getContent(String version) throws AccessException {
        try {
            return url.openStream();
        } catch (Throwable t) {
            throw new AccessException("Error reading datastream content", t);
        }
    }

    public String getFormatURI() {
        return null;
    }

    public Datastream.Name getName() {
        return this.dsID;
    }

    public void setLabel(String label) {
        this.label = label;
    }

    public String getLabel() {
        return label;
    }

    public String getMimeType() {
        return mimeType;
    }

    public void setMimeType(String mime) {
        this.mimeType = mime;
    }

    public URL getURL() {
        return url;
    }

    public void setURL(URL url) {
        this.url = url;
    }

    public long getSize() {
        return 0;
    }

    public State getState() {
        return State.Active;
    }

    public boolean isVersionable() {
        return true;
    }

    public boolean isCanonical() {
        return false;
    }
}
