/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map.Entry;

import javax.xml.namespace.QName;

import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;

import fedora.common.Constants;
import fedora.common.PID;

import static org.nsdl.repository.util.Encoder.xmlEncode;

/**
 * Builds RELS-EXT datastream content;
 * <p>
 * Given NDR properties or relationships, or explicit predicates and values,
 * builds an RDF document suitable for inclusion in a Fedora RELS-EXT
 * datastream..
 * </p>
 * 
 * @author birkland
 * 
 */
public class RelsExtBuilder implements DOPartWriter {
    private final URI pidURI;

    /* Namespace declarations */
    private final HashMap<String, String> nsDecls = new HashMap<String, String>();

    /* Writers for each triple */
    private final HashSet<RDFTripleWriter> rels = new HashSet<RDFTripleWriter>();

    private final FedoraNDRAccess ndr;

    private static final QName HAS_MODEL = new QName(
            Constants.MODEL.HAS_MODEL.namespace.uri,
            Constants.MODEL.HAS_MODEL.localName, Constants.MODEL.prefix);

    private int conflictSuffix = 1;

    public RelsExtBuilder(PID pid, FedoraNDRAccess ndr) {

        this.ndr = ndr;

        try {
            this.pidURI = new URI(pid.toURI());
        } catch (URISyntaxException e) {
            throw new RuntimeException("PANIC: pid URI corrupt:" + pid.toURI());
        }

        nsDecls.put("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
    }

    /**
     * Add NDR properties to a RELS-EXT
     * 
     * @param props
     *            Collection of properties to add
     */
    public void addProperties(Collection<? extends Property> props) {
        for (Property p : props) {
            addNamespace(p.predicate());
            rels.add(new RDFTripleWriter(pidURI, p.predicate(), p.value()));
        }
    }

    /**
     * Add NDR relationships to a RELS-EXT
     * 
     * @param rels
     *            Collection of relationships to add
     */
    public void addRelationships(Collection<? extends Relationship> rels) {
        for (Relationship r : rels) {

            addNamespace(r.predicate());
            try {
                this.rels.add(new RDFTripleWriter(pidURI, r.predicate(),
                        new URI(ndr.getObject(r.related()).pid().toURI())));
            } catch (Exception e) {
                /* Should never happen */
                throw new IllegalStateException(
                        "PANIC: PID uri did not parse!", e);
            }
        }
    }

    public void addModel(URI modelURI) {
        addNamespace(HAS_MODEL);
        this.rels.add(new RDFTripleWriter(pidURI, HAS_MODEL, modelURI));
    }

    /**
     * Add a literal value the RELS-EXT
     * 
     * @param rel
     *            Name of the predicate
     * @param value
     *            Literal value
     */
    public void addLiteral(QName rel, String value) {
        addNamespace(rel);
        rels.add(new RDFTripleWriter(pidURI, rel, value));
    }

    /**
     * Add a relationship to the RELS-EXT
     * 
     * @param rel
     *            Name of the predicate
     * @param value
     *            UTI of the object
     */
    public void addRelationship(QName rel, URI value) {
        addNamespace(rel);
        rels.add(new RDFTripleWriter(pidURI, rel, value));
    }

    public String toString() {
        StringBuilder out = new StringBuilder();

        out.append("   <rdf:RDF ");
        out.append(namespaceDeclarations());
        out.append(">\n");

        out.append("    <rdf:Description rdf:about=\"");
        out.append(pidURI);
        out.append("\">\n");

        for (RDFTripleWriter w : rels) {
            out.append("     ");
            out.append(w.value());
            out.append("\n");
        }

        out.append("    </rdf:Description>\n");
        out.append("   </rdf:RDF>\n");

        return out.toString();
    }

    public String getRepresentation() {
        StringBuilder rep = new StringBuilder();
        rep.append("<foxml:datastream CONTROL_GROUP=\"X\" ID=\"RELS-EXT\" "
                + "STATE=\"A\" VERSIONABLE=\"true\">\n");
        rep.append(" <foxml:datastreamVersion ID=\"RELS-EXT.0\" "
                + "LABEL=\"Relationships\" "
                + "FORMAT_URI=\"info:fedora/fedora-system:FedoraRELSExt-1.0\" "
                + "MIMETYPE=\"application/rdf+xml\">\n");
        rep.append("  <foxml:xmlContent>\n");
        rep.append(this.toString());
        rep.append("  </foxml:xmlContent>\n");
        rep.append(" </foxml:datastreamVersion>\n");
        rep.append("</foxml:datastream>\n");

        return rep.toString();
    }

    /* TODO: handle namespace prefix conflicts */
    private String namespaceDeclarations() {
        StringBuilder s = new StringBuilder();
        for (Entry<String, String> ns : nsDecls.entrySet()) {
            s.append("xmlns:");
            s.append(ns.getKey());
            s.append("=\"");
            s.append(ns.getValue());
            s.append("\" ");
        }

        return s.toString();
    }

    private void addNamespace(QName ns) {
        if (ns.getPrefix() == null || ns.getPrefix().equals("")) {
            return;
        }

        if (nsDecls.containsKey(ns.getPrefix())) {
            if (!nsDecls.get(ns.getPrefix()).equals(ns.getNamespaceURI())) {
                /* Conflict! */
                nsDecls.put(ns.getPrefix() + "_" + conflictSuffix, ns
                        .getNamespaceURI());
            }
        } else {
            nsDecls.put(ns.getPrefix(), ns.getNamespaceURI());
        }
    }

    private static class RDFTripleWriter {
        private final String serializedValue;

        public RDFTripleWriter(URI subject, QName predicate, String value) {
            String prefix = "";
            String xmlns = "";
            StringBuilder b = new StringBuilder();
            if (predicate.getPrefix() != null) {
                prefix = predicate.getPrefix() + ":";
            } else {
                xmlns = " xmlns=\"" + predicate.getNamespaceURI() + "\"";
            }

            b.append('<');
            b.append(prefix);
            b.append(predicate.getLocalPart());
            b.append(xmlns);
            b.append('>');

            b.append(xmlEncode(value));

            b.append("</");
            b.append(prefix);
            b.append(predicate.getLocalPart());
            b.append('>');

            this.serializedValue = b.toString();

        }

        public RDFTripleWriter(URI subject, QName predicate, URI object) {
            String prefix = "";
            StringBuilder b = new StringBuilder();
            if (predicate.getPrefix() != null) {
                prefix = predicate.getPrefix() + ":";
            }
            b.append('<');
            b.append(prefix);
            b.append(predicate.getLocalPart());
            b.append(' ');
            b.append("rdf:resource=\"");
            b.append(xmlEncode(object.toString()));
            b.append("\" />");

            this.serializedValue = b.toString();
        }

        public String toString() {
            return serializedValue;
        }

        public CharSequence value() {
            return serializedValue;
        }

        public boolean equals(Object o) {
            if (o == null)
                return false;
            return serializedValue.equals(o.toString());
        }

        public int hashCode() {
            return (serializedValue.hashCode());
        }
    }
}
