/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.util.StreamUtil;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Datastream containing XML content.
 * <p>
 * Implements a datastream that is optimized for handling XML content.
 * </p>
 * 
 * @author birkland
 * @since 1.0
 */
public class XMLDatastream implements FedoraDatastream {

    private static final Logger logger = Logger.getLogger(XMLDatastream.class);

    private final Datastream.Name dsID;
    private String label;
    private String mimetype = "application/xml";
    private byte[] xmlBytes;
    private ControlGroup controlGroup = ControlGroup.InlineXML;

    /**
     * Create an XML datastream with the given content and mimetype.
     * <p>
     * By default, will create a managed datastream which contains XML content
     * with MIME type <tt>application/xml</tt>. A different MIME type may be
     * supplied in the constructor, and a method exists for changing the storage
     * to inline XML, if desired.
     * </p>
     * 
     * @param dsID
     *            datastream ID
     * @param content
     *            Node containing XML content
     * @param mimetype
     *            MIME type of the XML content. Defaults to application/xml if
     *            null
     */
    public XMLDatastream(Datastream.Name dsID, Node content, String fileName,
            String mimetype) {
        logger.warn("BOGUS: XMLDatastream should not be instantiated directly");
        this.dsID = dsID;
        if (mimetype != null) {
            this.mimetype = mimetype;
        }

        label = fileName;

        try {
            if (content == null) {
                /* If no metadata content is defined, no bytes for you */
                xmlBytes = new byte[0];
            } else {
                /* Note: UTF-8 is assumed here */
                if (Document.class.isAssignableFrom(content.getClass())) {
                    content = ((Document) content).getDocumentElement();
                }
                xmlBytes = NDRXMLutil.serializeDOMElement((Element) content)
                        .replaceAll("<\\?.+\\?>\n", "").getBytes("UTF-8");
            }
        } catch (Exception e) {
            throw new RuntimeException("Could not serialize xml content", e);
        }
    }

    public XMLDatastream(Datastream.Name dsID, InputStream content,
            String filename, String mimetype) {
        this.dsID = dsID;
        this.mimetype = mimetype;
        label = filename;
        try {
            xmlBytes = new String(StreamUtil.streamToByteArray(content),
                    "UTF-8").replaceAll("<\\?.+\\?>\n", "").getBytes("UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException(
                    "This platform doesn't support UTF-8.  Gah!!");
        }
    }

    /**
     * Set the Fedora storage control group.
     * <p>
     * XML may be managed externally in Fedora, or inline. The default is
     * managed, but this method may be used to change it to inline.
     * </p>
     * 
     * @param g
     *            Control group. Valid values are Inline or Managed.
     */
    public void setControlGroup(ControlGroup g) {
        if (g.equals(ControlGroup.External) || g.equals(ControlGroup.Redirect)) {
            throw new IllegalArgumentException(
                    "XML content datastreams may only be inline or managed");
        }
        controlGroup = g;
    }

    /** {@inheritDoc} */
    public ControlGroup controlGroup() {
        return controlGroup;
    }

    public String[] getAltIDs() {
        return new String[0];
    }

    public InputStream getContent() throws AccessException {
        return new ByteArrayInputStream(xmlBytes);
    }

    public InputStream getContent(String version) throws AccessException {
        return new ByteArrayInputStream(xmlBytes);
    }

    public void setContent(InputStream input) {
        /*
         * XXX: We're putting this into an in-memory byte array, so we hope it's
         * not big.
         */

        ByteArrayOutputStream container = new ByteArrayOutputStream();
        int read;
        byte[] buf = new byte[4096];

        try {
            while ((read = input.read(buf)) != -1) {
                container.write(buf, 0, read);
            }
        } catch (IOException e) {
            throw new AccessException("Could not set datastream content", e);
        }

        this.xmlBytes = container.toByteArray();
    }

    public String getFormatURI() {
        return null;
    }

    public Datastream.Name getName() {
        return this.dsID;
    }

    public void setLabel(String label) {
        this.label = label;
    }

    public String getLabel() {
        return label;
    }

    public String getMimeType() {
        return mimetype;
    }

    public void setMimeType(String mimetype) {
        this.mimetype = mimetype;
    }

    /**
     * Get the URL for this datastream.
     * <p>
     * XMLDatastream does not necessarily exist in a repository yet, so does not
     * have a URL;
     * </p>
     * 
     * @return null
     */
    public URL getURL() {
        return null;
    }

    public void setURL(URL url) {
        throw new UnsupportedOperationException(
                "Cannot set the URL for inline XML datastreams");
    }

    public long getSize() {
        return xmlBytes.length;
    }

    public State getState() {
        return State.Active;
    }

    public boolean isVersionable() {
        return true;
    }

    public boolean isCanonical() {
        return true;
    }
}
