/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora.ri;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.List;
import java.util.Properties;

import javax.sql.DataSource;

import org.apache.commons.dbcp.BasicDataSource;
import org.apache.commons.dbcp.BasicDataSourceFactory;
import org.apache.log4j.Logger;
import org.nsdl.mptstore.core.BasicTableManager;
import org.nsdl.mptstore.core.DDLGenerator;
import org.nsdl.mptstore.core.TableManager;
import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.mptstore.query.SQLUnionQueryResults;
import org.nsdl.mptstore.query.component.GraphQuery;
import org.nsdl.mptstore.query.lang.QueryCompiler;
import org.nsdl.mptstore.query.lang.spo.SPOQueryCompiler;
import org.nsdl.mptstore.query.provider.GraphQuerySQLProvider;
import org.nsdl.mptstore.query.provider.SQLProvider;
import org.nsdl.mptstore.rdf.Node;
import org.nsdl.repository.util.Configuration;

/**
 * Queries an MPTStore ResourceIndex directly using JDBC.
 * 
 * Since there is no SPARQL (or equvalent) interface to ResourceIndex queries
 * using MPTStore, queries must be sent directly as SQL calls to the underlying
 * RDBMS. This class is responsible for maintaining connections to the RDBMS,
 * and formulating the appropriate queries underneath.
 * 
 * <h2>JDBC configuration Parameters</h2>
 * <p>
 * <dl>
 * <dt><code>param.fedora.ri.mptstore.db.jdbcURL</code></dt>
 * <dd>JDBC URL</dd>
 * 
 * <dt><code>param.fedora.ri.mptstore.db.user</code></dt>
 * <dd>Database user</dd>
 * 
 * <dt><code>param.fedora.ri.mptstore.db.passwd</code></dt>
 * <dd>User password</dd>
 * 
 * <dt><code>param.fedora.ri.mptstore.db.driverClassName</code></dt>
 * <dd>JDBC driver class name</dd>
 * 
 * <dt><code>param.fedora.ri.mptstore.map</code></dt>
 * <dd>Name of the MPTStore mapping table. Default is <code>tmap</code></dd>
 * 
 * <dt><code>param.fedora.ri.mptstore.prefix</code></dt>
 * <dd>MPTstore table mapping prefix. Default is <code>p</code></dd>
 * 
 * <dt><code>param.fedora.ri.mptstore.DDLGenerator</code></dt>
 * <dd>MPTStore DDL Generator</dd>
 * </dl>
 * </p>
 * 
 * 
 * <h2>DBCP Connection pooling Parameters</h2>
 * <p>
 * <dl>
 * <dt>
 * <code>repository.impl.fedora.ri.MPTQueryEngine.dbcp.poolPreparedStatements</code>
 * </dt>
 * <dd>Whether to pool prepared statements. Default is <code>true</code></dd>
 * 
 * <dt><code>repository.impl.fedora.ri.MPTQueryEngine.dbcp.initialSize</code></dt>
 * <dd>Initial connection pool size. Default is <code>0</code></dd>
 * 
 * <dt><code>repository.impl.fedora.ri.MPTQueryEngine.dbcp.maxActive</code></dt>
 * <dd>Maximum number of open connections. Default is <code>-1</code>
 * (unbounded)</dd>
 * 
 * <dt><code>repository.impl.fedora.ri.MPTQueryEngine.dbcp.minIdle</code></dt>
 * <dd>Minimum number of idle connections. Default is <code>0</code></dd>
 * 
 * <dt><code>repository.impl.fedora.ri.MPTQueryEngine.dbcp.maxIdle</code></dt>
 * <dd>Maximum number of idle connections. Default is <code>-1</code>(unbounded)
 * </dd>
 * 
 * </dl>
 * </p>
 * 
 * @author birkland
 * @since 1.0
 */
public class MPTQueryEngine implements RIQueryEngine, Configuration.Listener {

    private TableManager tableManager = null;

    private DataSource dataSource = null;

    private static final Logger logger = Logger.getLogger(MPTQueryEngine.class);

    /** JDBC URL */
    public static final String PROP_JDBC_URI = "param.fedora.ri.mptstore.db.jdbcURL";

    /** User of RI database */
    public static final String PROP_DB_USER = "param.fedora.ri.mptstore.db.user";

    /** Password of RI database */
    public static final String PROP_DB_PASSWD = "param.fedora.ri.mptstore.db.passwd";

    /** Classname of JDBC driver */
    public static final String PROP_DB_DRIVER_CLASSNAME = "param.fedora.ri.mptstore.db.driverClassName";

    /** name of map table in mptstore */
    public static final String PROP_MAP = "param.fedora.ri.mptstore.map";

    public static final String PROP_MAP_DEFAULT = "tMap";

    /** prefix used to create new tables */
    public static final String PROP_PREFIX = "param.fedora.ri.mptstore.prefix";

    public static final String PROP_PREFIX_DEFAULT = "p";

    /** link between mptstore and the database implementation */
    public static final String PROP_DDL_GENERATOR_CLASSNAME = "param.fedora.ri.mptstore.DDLGenerator";

    public static final String PROP_DBCP_POOL_STATEMENTS = "repository.impl.fedora.ri.MPTQueryEngine.dbcp.poolPreparedStatements";

    public static final String PROP_DBCP_POOL_STATEMENTS_DEFAULT = "true";

    public static final String PROP_DBCP_INITIAL = "repository.impl.fedora.ri.MPTQueryEngine.dbcp.initialSize";

    public static final String PROP_DBCP_INITIAL_DEFAULT = "0";

    public static final String PROP_DBCP_MAX_ACTIVE = "repository.impl.fedora.ri.MPTQueryEngine.dbcp.maxActive";

    public static final String PROP_DBCP_MAX_ACTIVE_DEFAULT = "-1";

    public static final String PROP_DBCP_MAX_IDLE = "repository.impl.fedora.ri.MPTQueryEngine.dbcp.maxIdle";

    public static final String PROP_DBCP_MAX_IDLE_DEFAULT = "-1";

    public static final String PROP_DBCP_MIN_IDLE = "repository.impl.fedora.ri.MPTQueryEngine.dbcp.minIdle";

    public static final String PROP_DBCP_MIN_IDLE_DEFAULT = "0";

    public static final String PROP_DBCP_MIN_EVICTABLE_IDLE_TIME_MILLIS = "repository.impl.fedora.ri.MPTQueryEngine.dbcp.minEvictableIdleTimeMillis";

    public static final String PROP_DBCP_MIN_EVICTABLE_IDLE_TIME_MILLIS_DEFAULT = "600000";

    public static final String PROP_DBCP_TIME_BETWEEN_EVICTION_RUNS_MILLIS = "repository.impl.fedora.ri.MPTQueryEngine.dbcp.timeBetweenEvictionRunsMillis";

    public static final String PROP_DBCP_TIME_BETWEEN_EVICTION_RUNS_MILLIS_DEFAULT = "300000";

    public static final String PROP_DBCP_NUM_TESTS_PER_EVICTION = "repository.impl.fedora.ri.numTestsPerEviction";

    public static final String PROP_DBCP_NUM_TESTS_PER_EVICTION_DEFAULT = "-1";

    /**
     * No-argument constructor.
     * 
     * It is necessary to initialize via {@link #updateConfig(Configuration)}
     * before this class can be used in any sense.
     * 
     */
    public MPTQueryEngine() {
    }

    public QueryResults execute(GraphQuery graphQ, List<String> targets)
            throws QueryException {
        GraphQuerySQLProvider sqlGen = new GraphQuerySQLProvider(tableManager,
                graphQ, true);
        sqlGen.setTargets(targets);

        try {
            Connection connect = dataSource.getConnection();
            connect.setAutoCommit(false);
            return new ResultWrapper(new SQLUnionQueryResults(connect, sqlGen,
                    100, true));
        } catch (SQLException e) {
            throw new RuntimeException("Could not connect to data source", e);
        }
    }

    public QueryResults execute(PreparedQuery q, List<String> params)
            throws QueryException {
        PreparedSQLUnionQueryResults prepared = new PreparedSQLUnionQueryResults(
                q, 100, true);
        try {
            Connection connect = dataSource.getConnection();
            connect.setAutoCommit(false);
            prepared.execute(connect, params);
        } catch (SQLException e) {
            throw new QueryException("Could not connect to data source");
        }
        return new ResultWrapper(prepared);
    }

    public QueryResults execute(String query, Language lang)
            throws QueryException {
        switch (lang) {
        case SPO:
            return new ResultWrapper(execSPO(query));
        }
        throw new IllegalArgumentException("The language " + lang
                + " is not supported");
    }

    public PreparedQuery prepare(GraphQuery q, List<String> targets)
            throws QueryException {
        GraphQuerySQLProvider sqlGen = new GraphQuerySQLProvider(tableManager,
                q, true);
        sqlGen.setTargets(targets);

        return new PreparedQuery(sqlGen, PreparedQuery.PLACEHOLDER);
    }

    public void updateConfig(Configuration props) {
        /* Initialize the DDL generator */
        DDLGenerator generator;
        try {
            generator = (DDLGenerator) Class.forName(
                    props.getProperty(PROP_DDL_GENERATOR_CLASSNAME))
                    .newInstance();
        } catch (Exception e) {
            throw new RuntimeException("Could not initialize DDL generator", e);
        }

        /* Initialize database connection */
        BasicDataSource source;
        Properties dbParams = new Properties();

        dbParams.setProperty("url", props.getProperty(PROP_JDBC_URI));
        dbParams.setProperty("username", props.getProperty(PROP_DB_USER));
        dbParams.setProperty("password", props.getProperty(PROP_DB_PASSWD));
        dbParams.setProperty("driverClassName", props
                .getProperty(PROP_DB_DRIVER_CLASSNAME));

        dbParams.setProperty("initialSize", props.getProperty(
                PROP_DBCP_INITIAL, PROP_DBCP_INITIAL_DEFAULT));
        dbParams.setProperty("maxActive", props.getProperty(
                PROP_DBCP_MAX_ACTIVE, PROP_DBCP_MAX_ACTIVE_DEFAULT));

        dbParams.setProperty("minIdle", props.getProperty(PROP_DBCP_MIN_IDLE,
                PROP_DBCP_MIN_IDLE_DEFAULT));
        dbParams.setProperty("maxIdle", props.getProperty(PROP_DBCP_MAX_IDLE,
                PROP_DBCP_MAX_IDLE_DEFAULT));

        dbParams.setProperty("poolPreparedStatements", props.getProperty(
                PROP_DBCP_POOL_STATEMENTS, PROP_DBCP_POOL_STATEMENTS_DEFAULT));

        dbParams.setProperty("minEvictableIdleTimeMillis", props.getProperty(
                PROP_DBCP_MIN_EVICTABLE_IDLE_TIME_MILLIS,
                PROP_DBCP_MIN_EVICTABLE_IDLE_TIME_MILLIS_DEFAULT));

        dbParams.setProperty("timeBetweenEvictionRunsMillis", props
                .getProperty(PROP_DBCP_TIME_BETWEEN_EVICTION_RUNS_MILLIS,
                        PROP_DBCP_TIME_BETWEEN_EVICTION_RUNS_MILLIS_DEFAULT));

        dbParams.setProperty("numTestsPerEviction", props.getProperty(
                PROP_DBCP_NUM_TESTS_PER_EVICTION,
                PROP_DBCP_NUM_TESTS_PER_EVICTION_DEFAULT));

        try {
            Class.forName(props.getProperty(PROP_DB_DRIVER_CLASSNAME));
            source = (BasicDataSource) BasicDataSourceFactory
                    .createDataSource(dbParams);
        } catch (Exception e) {
            throw new RuntimeException(
                    "Could not establish database connection", e);
        }
        this.dataSource = source;

        /* Finally, create the table manager */
        BasicTableManager manager;
        String mapTable = props.getProperty(PROP_MAP, PROP_MAP_DEFAULT);
        String prefix = props.getProperty(PROP_PREFIX, PROP_PREFIX_DEFAULT);
        try {
            manager = new BasicTableManager(source, generator, mapTable, prefix);
        } catch (SQLException e) {
            throw new RuntimeException("Could not initialize table mapper", e);
        }
        tableManager = manager;
    }

    private QueryResults execSPO(String query) throws QueryException {
        QueryCompiler compiler = new SPOQueryCompiler(tableManager, true);
        SQLProvider compiled = compiler.compile(query);
        Connection connection = null;
        try {
            connection = dataSource.getConnection();
            connection.setAutoCommit(false);
        } catch (SQLException e) {
            throw new RuntimeException("Could not connect to data source");
        }

        return new SQLUnionQueryResults(connection, compiled, 100, true);
    }

    private class ResultWrapper implements QueryResults {
        private final QueryResults delegate;

        private ResultWrapper(QueryResults toWrap) {
            this.delegate = toWrap;
        }

        public List<String> getTargets() {
            return delegate.getTargets();
        }

        public boolean hasNext() {
            return delegate.hasNext();
        }

        public void remove() {
            delegate.remove();
        }

        public List<Node> next() {
            return delegate.next();
        }

        public void close() {
            try {
                delegate.close();
            } catch (Throwable e) {
                if (logger.isDebugEnabled()) {
                    logger.debug("Trouble closing connection, but continuing",
                            e);
                } else {
                    logger.warn("Trouble closing RI connection.  Ignoring");
                }
            }
        }
    }
}
