/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora.ri;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Queue;

import org.apache.log4j.Logger;
import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.mptstore.query.RuntimeQueryException;
import org.nsdl.mptstore.query.provider.SQLProvider;
import org.nsdl.mptstore.rdf.Node;
import org.nsdl.mptstore.util.DBUtil;
import org.nsdl.mptstore.util.NTriplesUtil;

public class PreparedSQLUnionQueryResults implements QueryResults {

    public static final String PLACEHOLDER = "<info:mptstore/placeholder>";

    private static final Logger logger = Logger
            .getLogger(PreparedSQLUnionQueryResults.class);

    /** The master list of statements to execute for every query */
    private final List<String> statements = new ArrayList<String>();

    private Connection currentConnection;

    private List<String> params;

    /**
     * This will contain pending statements to execute for a given request. It
     * is reset upon execute(), and depleted with various next() calls
     */
    private Queue<String> currentQueries = new LinkedList<String>();

    private ResultSet currentResults;

    private PreparedStatement currentStatement;

    private List<Node> nextTuple;

    private final int fetchSize;

    private final int tupleSize;

    private final List<String> targets;

    private boolean _closed;

    private final boolean autoRelease;

    public PreparedSQLUnionQueryResults(final SQLProvider sqlProvider,
            final String varName, final int fetchSize,
            final boolean autoReleaseConection) throws QueryException {

        /* Populate the master statement list */
        for (String rawSQL : sqlProvider.getSQL()) {
            logger.trace("Given raw SQL " + rawSQL);
            String sql = rawSQL.replaceAll("'" + varName + "'", "?");
            logger.trace("Converted to " + sql);

            statements.add(sql);
        }

        this.fetchSize = fetchSize;
        this.tupleSize = sqlProvider.getTargets().size();
        this.targets = sqlProvider.getTargets();
        this.autoRelease = autoReleaseConection;
    }

    public PreparedSQLUnionQueryResults(final SQLProvider sqlProvider,
            final int fetchSize, final boolean autoReleaseConection)
            throws QueryException {

        /* Populate the master statement list */
        for (String rawSQL : sqlProvider.getSQL()) {
            statements.add(rawSQL);
        }

        this.fetchSize = fetchSize;
        this.tupleSize = sqlProvider.getTargets().size();
        this.targets = sqlProvider.getTargets();
        this.autoRelease = autoReleaseConection;
    }

    public void execute(Connection c, List<String> values) {
        close();
        open();
        currentQueries = new LinkedList<String>(statements);
        currentConnection = c;
        params = values;
        try {
            startNextQuery();
            readNextTuple();
        } catch (Exception e) {
            throw new RuntimeException("Could not start query", e);
        }
    }

    public List<String> getTargets() {
        return targets;
    }

    public boolean hasNext() {
        return nextTuple != null;
    }

    public List<Node> next() throws RuntimeQueryException {
        if (!hasNext()) {
            throw new NoSuchElementException();
        } else {
            List<Node> thisTuple = nextTuple;
            try {
                readNextTuple();
                return thisTuple;
            } catch (QueryException e) {
                close(); // proactively close on error
                logger.error(e);
                throw new RuntimeQueryException(e);
            }
        }
    }

    public void remove() throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }

    public void close() {
        nextTuple = null;
        if (!_closed) {
            if (currentResults != null) {
                try {
                    currentResults.close();
                } catch (Exception e) {
                    logger.warn("Error closing result set", e);
                }
            }
            if (currentStatement != null) {
                try {
                    currentStatement.close();
                } catch (Exception e) {
                    logger.warn("Error closing statement", e);
                }
            }
            if (autoRelease && currentConnection != null) {
                DBUtil.release(currentConnection);
            }
            _closed = true;
        }
        currentResults = null;
        currentStatement = null;
        currentConnection = null;
        currentQueries = null;
    }

    private void open() {
        _closed = false;
    }

    private void startNextQuery() throws SQLException {
        if (!currentQueries.isEmpty()) {
            logger.trace("Executing query part..");
            if (currentResults != null) {
                currentResults.close();
                currentStatement.close();
            }
            currentStatement = currentConnection
                    .prepareStatement(currentQueries.remove());

            currentStatement.setFetchSize(fetchSize);
            currentStatement.setFetchDirection(ResultSet.FETCH_FORWARD);

            /* populate with given params */
            for (int i = 0; i < params.size(); i++) {
                currentStatement.setString(i + 1, params.get(i));
                logger.trace("Using parameter " + params.get(i));
            }

            currentResults = currentStatement.executeQuery();

        } else { /* Done! */
            logger.trace("Query is done");
            close();
            currentResults = null;
        }
    }

    /**
     * Set nextTuple to the next tuple.
     * 
     * If there are no more tuples, proactively close and set
     * <code>nextTuple</code> to <code>null</code>.
     * 
     * @throws QueryException
     *             if an unexpected error occurs.
     */
    private void readNextTuple() throws QueryException {

        try {
            while (currentResults != null && !currentResults.next()) {
                startNextQuery();
            }

            if (currentResults == null) {
                nextTuple = null;
            } else {
                nextTuple = new ArrayList<Node>(tupleSize);
                for (int i = 1; i <= tupleSize; i++) {
                    String nodeString = DBUtil.getLongString(currentResults, i);
                    try {
                        if (nodeString != null) {
                            nextTuple.add(NTriplesUtil.parseNode(nodeString));
                        } else {
                            nextTuple.add(null);
                        }
                    } catch (ParseException e) {
                        throw new QueryException("Error parsing RDF node ("
                                + nodeString + ") from database: "
                                + e.getMessage(), e);
                    }
                }
            }
        } catch (SQLException e) {
            throw new QueryException("Error querying database", e);
        }
    }
}
