/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.impl.fedora.ri;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.mptstore.query.component.BasicNodePattern;
import org.nsdl.mptstore.query.component.BasicTriplePattern;
import org.nsdl.mptstore.query.component.GraphPattern;
import org.nsdl.mptstore.query.component.GraphQuery;
import org.nsdl.mptstore.query.component.NodePattern;
import org.nsdl.mptstore.rdf.ObjectNode;
import org.nsdl.mptstore.rdf.PredicateNode;
import org.nsdl.mptstore.rdf.SubjectNode;
import org.nsdl.mptstore.util.NTriplesUtil;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.impl.fedora.FedoraNDRAccess;
import org.nsdl.repository.impl.fedora.FedoraNDRObject;
import org.nsdl.repository.impl.fedora.RDFRelsReader;
import org.nsdl.repository.impl.fedora.RelsReader;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;

import fedora.common.PID;

/** Reads properties and relationships from the ResourceIndex */
public class RIRelsReader implements RelsReader {

    private static final Logger logger = Logger.getLogger(RIRelsReader.class);

    private final FedoraNDRAccess repo;

    private final FedoraNDRObject currentObject;

    private Set<Relationship> relationships = new HashSet<Relationship>();

    private Set<Property> properties = new HashSet<Property>();

    private HashMap<QName, PreparedQuery> pollQueries = new HashMap<QName, PreparedQuery>();

    public RIRelsReader(NDRObjectInfo object, FedoraNDRAccess fedora) {

        this.currentObject = fedora.getObject(object);
        this.repo = fedora;
    }

    public Iterable<Property> readProperties() {
        getRels();
        return properties;
    }

    public Iterable<Relationship> readRelationships() {
        getRels();
        return relationships;
    }

    public Iterable<Property> readProperties(Property.Name propertyType) {
        ArrayList<Property> properties = new ArrayList<Property>();
        for (String value : pollRels(propertyType)) {
            if (!value.startsWith(Constants.FEDORA_INFO_URI)) {
                properties
                        .add(new Property(currentObject, propertyType, value));
            }
        }
        return properties;
    }

    public Iterable<Relationship> readRelationships(
            Relationship.Name relationshipType) {
        ArrayList<Relationship> relationships = new ArrayList<Relationship>();
        for (String relatedPID : pollRels(relationshipType)) {
            if (relatedPID.startsWith(Constants.FEDORA_INFO_URI)
                    && RDFRelsReader.isNDRObject(repo.getRepository().getRI(),
                            PID.getInstance(relatedPID))) {
                relationships.add(new Relationship(currentObject,
                        relationshipType, repo.getObject(PID
                                .getInstance(relatedPID))));
            }
        }
        return relationships;
    }

    /* Find all rels with a given predicate */
    private Set<String> pollRels(QName predicate) {
        Set<String> polled = new HashSet<String>();
        QueryResults results;

        if (!pollQueries.containsKey(predicate)) {
            NodePattern<SubjectNode> subject;
            NodePattern<PredicateNode> pred;
            NodePattern<ObjectNode> object;

            GraphQuery g = new GraphQuery();
            GraphPattern p = new GraphPattern();
//            logger.debug("Creating new prepared query for " + predicate);
            try {
                subject = new BasicNodePattern<SubjectNode>(NTriplesUtil
                        .parseSubject(PreparedQuery.PLACEHOLDER));
                pred = new BasicNodePattern<PredicateNode>(NTriplesUtil
                        .parsePredicate("<" + predicate.getNamespaceURI()
                                + predicate.getLocalPart() + ">"));
                object = new BasicNodePattern<ObjectNode>("$related");

                p
                        .addTriplePattern(new BasicTriplePattern(subject, pred,
                                object));

                g.addRequired(p);

                pollQueries.put(predicate, repo.getRepository().getRI()
                        .prepare(g, Arrays.asList("$related")));
            } catch (Exception e) {
                throw new RuntimeException("Could not build poll query");
            }
        }

        try {
            ArrayList<String> params = new ArrayList<String>();
            params.add("<" + currentObject.pid().toURI() + ">");
            results = repo.getRepository().getRI().execute(
                    pollQueries.get(predicate), params);
        } catch (Exception e) {
            throw new RepositoryException("Could not poll for rels ", e);
        }

        while (results.hasNext()) {
            polled.add(results.next().get(0).getValue());
        }

        results.close();
        return polled;
    }

    /* We kinda cheat here and read from RELS-EXT instead of the RI */
    private void getRels() {
        if (relationships.size() > 0 || properties.size() > 0)
            return;
        try {
            logger.trace("BEGIN reading rels of " + currentObject.handle());
            RelsReader reader = new RDFRelsReader(currentObject, repo);
            for (Relationship r : reader.readRelationships()) {
                this.relationships.add(r);
            }

            for (Property p : reader.readProperties()) {
                this.properties.add(p);
            }
            logger.trace("SUCCESS reading rels of " + currentObject.handle());
        } catch (Exception e) {
            throw new RepositoryException("Could not read relationships of "
                    + currentObject.handle(), e);
        }
    }
}
