/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora.ri;

import java.text.ParseException;

import org.apache.log4j.Logger;
import org.nsdl.mptstore.query.component.BasicNodePattern;
import org.nsdl.mptstore.query.component.BasicTriplePattern;
import org.nsdl.mptstore.query.component.NodePattern;
import org.nsdl.mptstore.query.component.TriplePattern;
import org.nsdl.mptstore.rdf.ObjectNode;
import org.nsdl.mptstore.rdf.PredicateNode;
import org.nsdl.mptstore.rdf.SubjectNode;
import org.nsdl.mptstore.util.NTriplesUtil;

/**
 * Collection of various utility functions for sealing with triples or triple
 * constructs.
 * 
 * @author birkland
 * @since 1.0
 */
public class TripleUtils {
    private static final Logger logger = Logger.getLogger(TripleUtils.class);

    /** Cannot be constructed */
    private TripleUtils() {

    }

    /**
     * Form a TriplePattern from its constituent components.
     * 
     * All inputs <em>should</em> be in NTriples format, but if not, will
     * attempt to make sense of them and escape them as appropriate.
     * 
     * @param s
     *            Subject URI.
     * @param p
     *            Predicate URI.
     * @param o
     *            Object URI or Literal
     * @return resulting TriplePattern
     * @throws ParseException
     *             in cases where the subject, predicate, or object cannot be
     *             successfully parsed.
     */
    public static final TriplePattern getPattern(String s, String p, String o)
            throws ParseException {
        NodePattern<SubjectNode> subject;
        NodePattern<PredicateNode> predicate;
        NodePattern<ObjectNode> object;

        if (s.startsWith("$")) {
            subject = new BasicNodePattern<SubjectNode>(s);
        } else {
            subject = new BasicNodePattern<SubjectNode>(NTriplesUtil
                    .parseSubject(nTripleEncodeURI(s)));
        }

        /* Right now, we don't consider predicates as variables... */
        predicate = new BasicNodePattern<PredicateNode>(NTriplesUtil
                .parsePredicate(nTripleEncodeURI(p)));

        /*
         * If the object is not surrounded by a "<>" or '""', assume that it is
         * destined to be a literal and surround it with quotes
         */
        if (o.startsWith("$")) {
            object = new BasicNodePattern<ObjectNode>(o);
        } else if (!o.startsWith("\"")
                && !(o.startsWith("<") && o.endsWith(">"))) {
            object = new BasicNodePattern<ObjectNode>(NTriplesUtil
                    .parseObject(nTripleEncodeLiteral(o)));
        } else {
            object = new BasicNodePattern<ObjectNode>(NTriplesUtil
                    .parseObject(o));
        }

        TriplePattern resultPattern = new BasicTriplePattern(subject,
                predicate, object);
        logger.debug("Using query pattern: " + resultPattern);
        return resultPattern;
    }

    /**
     * Puts a string URI into NTriple format.
     * 
     * Essentially, just surrounds the string with "&lt; and &gt; characters if
     * it isn't already.
     * 
     * @param uri
     *            URI to encode
     * @return encoded URI.
     */
    public static final String nTripleEncodeURI(String uri) {
        if (!uri.startsWith("<")) {
            return "<" + uri + ">";
        } else {
            return uri;
        }
    }

    /**
     * Puts a string literal into NTriple format.
     * 
     * Surrounds the literal in quotes, and escapes any unicode or whitespace
     * (tab, newline, etc) characters.
     * 
     * @param literal
     *            The literal to encode.
     * @return encoded literal
     */
    public static final String nTripleEncodeLiteral(String literal) {
        return "\"" + NTriplesUtil.escapeLiteralValue(literal) + "\"";
    }
}
