/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model;

import java.io.InputStream;
import java.net.URL;

import org.nsdl.repository.access.AccessException;

/**
 * Represents a stream of data in an NDR object.
 * <p>
 * As it stands now, properties of an NDR datastream are identical to that of a
 * Fedora datastream. That is likely to change should we decide to abstract it
 * away to its essence.
 * </p>
 * 
 * @author birkland
 * @since 1.0
 * 
 */
public interface Datastream {

    /** Datastream identifier */
    public Datastream.Name getName();

    /** Datastream mimetype */
    public String getMimeType();

    /** Datastream size, in bytes. */
    public long getSize();

    /**
     * Gets the location of a datastream content.
     * <p>
     * Datastream content does not necessarily have to reside in the datastream
     * itself - the datastream may be a proxy for content elsewhere on the web.
     * If this is the case, then <code>getURL()</code> will return the URL at
     * which the content actually resides. Otherwise, it will return the access
     * URL of the datastream in the NCore repository in which it resides.
     * </p>
     * 
     * <p>
     * This method may return null for newly created datastreams that have not
     * been written/comitted to a repository;
     * </p>
     * 
     * @return URL pointing to the content location.
     */
    public URL getURL();

    /**
     * Sets the URL source for datastream content
     * 
     * When a redirect URL is set, the content of the datastream is set to
     * whatever is accessible at the given address (implemented as a redirect).
     * Any contents that may have been previouslt set with
     * {@link #setContent(InputStream)} previously will be replaced.
     * 
     * @param redirect
     *            location of datastream content.
     */
    public void setURL(URL redirect);

    /**
     * Get the current contents of the datastream.
     * <p>
     * May be null if the datastream is a redirect. Redirect URLs should be read
     * by getRedirectURL();
     * </p>
     * 
     * @return InputStream containing datastream content.
     * @throws AccessException
     *             if there is some problem accessing datastream content.
     */
    public InputStream getContent() throws AccessException;

    /**
     * Get the contents of a specific datasteream version.
     * <p>
     * May be null if the datastream is a redirect. Redirect URLs should be read
     * by getRedirectURL();
     * </p>
     * 
     * @param version
     *            Representaion of datastream version.
     * @return InputStream containing datastream content.
     * @throws AccessException
     *             if there is some problem accessing datastream content.
     */
    public InputStream getContent(String version) throws AccessException;

    /**
     * Sets the binary content contained within the resource.
     * 
     * Any existing content will be replaced with the binary data present in the
     * given <code>InputStream</code>. If the resource previously had a redirect
     * URL, that URL will be made <code>null</code>.
     * 
     * @param content
     *            Content to reside in the resource.
     */
    public void setContent(InputStream content);

    /**
     * Sets the MIME type of the resource.
     * 
     * @param mime
     *            Name of the MIME type (e.g. <tt>text/xml</tt>).
     */
    public void setMimeType(String mime);

    /**
     * Gets a file name associated with this datastream content, if any.
     * 
     * @return String containing the content's suggested file name, null if it
     *         has no suggested name.
     */
    public String getLabel();

    /**
     * Set the suggested file name associated with this datastream content.
     * 
     * @param name
     *            Suggested file name
     */
    public void setLabel(String name);

    /**
     * Determines if the datastream contains the canonical version of the data.
     * <p>
     * For a canonical datastream:
     * <ul>
     * <li>The datastream itstlf is the primary source of the data</li>
     * <li>Content is modified/viewed by {@link #getContent()} and
     * {@link #setContent(InputStream)}</li>
     * <li>The URL may be viewed, but not modified</li>
     * </ul>
     * 
     * <p>
     * For non-canonical datastream:
     * <ul>
     * <li>The primary content itself lives at some external URL</li>
     * <li>Content may not be modified by {@linkm #setContent(InputStream)}</li>
     * <li>Content may still be viewed by {@link #getContent()}. Doing this is
     * identical to reading the content from the URL itself.</li>
     * <li>The URL may be changed by {@link #setURL(URL)}</li>
     * </ul>
     * </p>
     * 
     * @return True if the datastream is canonical.
     */
    public boolean isCanonical();

    public static class Name implements Component.Name<Datastream> {
        private final String id;
        private final int hash;
        
        public Name(String name) {
            id = name;
            hash = this.id.hashCode();
        }

        @Override
        public int hashCode() {
            return hash;
        }

        @Override
        public boolean equals(Object o) {
            if (!(o == null) && o instanceof Name) {
                return (((Name) o).toString()).equals(id);
            } else {
                return false;
            }
        }

        @Override
        public String toString() {
            return id;
        }

        public String getID() {
            return id;
        }
    }
}
