/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model;

import java.util.Collection;
import java.util.Date;

import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.error.RepositoryException;

import fedora.common.Constants;

/**
 * Represents the contents of an NDR object.
 * <p>
 * An NDR object is modelled as having three types of content:
 * <ul>
 * <li>{@link org.nsdl.repository.model.Property} - named literal values
 * <li>{@link org.nsdl.repository.model.Datastream} - named containers for
 * digital content
 * <li>{@link org.nsdl.repository.model.Relationship} - named relationships to
 * other NDR objects
 * </ul>
 * </p>
 * <p>
 * Properties and Relationships are named by a URI, but in the NSDL model are
 * addressed by a qualified name where namespace + localvalue = URI.
 * Datastreams, on the other hand, are accessed bt a unique String identifier.
 * There may be more than one property or relationship for a given name, but
 * datastream IDs are unique.
 * </p>
 * 
 * @author birkland
 * @since 1.0
 */
public interface NDRObject extends NDRObjectInfo {

    /** State of an NDR object. */
    public enum State {

        /** Object exists and is accessible in the NDR. */
        Active("Active", "A", Constants.MODEL.ACTIVE.uri),

        /**
         * Object exists in the NDR, but is not discoverable.
         * <p>
         * Unless specifically addressed by handle, or searched for by state, an
         * Inactive object will be invisible to NDR requests.
         * </p>
         */
        Inactive("Inactive", "I", Constants.MODEL.INACTIVE.uri),

        /**
         * Object once existed in the NDR, has been removed, and is now no
         * longer discoverable.
         * <p>
         * Unless specifically addressed by handle, or searched for by state, a
         * Deleted object will be invisible to NDR requests.
         * </p>
         */
        Deleted("Deleted", "D", Constants.MODEL.DELETED.uri);

        private final String uri;
        private final String foxmlName;
        private final String humanName;

        private State(String humanName, String foxmlName, String uri) {
            this.foxmlName = foxmlName;
            this.uri = uri;
            this.humanName = humanName;
        }

        /**
         * Get the URI of the state.
         * <p>
         * Object state is represented as a URI in parsed triple form. Searches
         * against the RI, therefore, must represent object state as one of
         * these URI values.
         * </p>
         * 
         * @return String containing State URI.
         */
        public String getURI() {
            return uri;
        }

        /** Get the state name as represented in FOXML. */
        public String getFoxmlName() {
            return foxmlName;
        }

        /** Get the state as a human readable name. */
        public String toString() {
            return humanName;
        }

        public static State getInstanceFromURI(String nameURI) {
            for (State s : State.values()) {
                if (s.getURI().equals(nameURI)) {
                    return s;
                }
            }

            throw new IllegalArgumentException("No state exists with uri '"
                    + nameURI + "'");
        }

        public static State getInstanceFromHumanName(String name) {
            for (State s : State.values()) {
                if (s.humanName.equals(name)) {
                    return s;
                }
            }

            throw new IllegalArgumentException("No state exists with name '"
                    + name + "'");
        }
    }

    /**
     * Get the current state of the object.
     * 
     * @return The current state.
     */
    public NDRObject.State state();
    
    /**
     * Get the creation date of the object.
     * 
     * @return The creation date.
     */
    public Date getCreatedDate();

    /**
     * Get the last known modification date of the object.
     * 
     * @return The modification date.
     */
    public Date getLastModifiedDate();
    
    /**
     * Retrieve a specific property from an object.
     * <p>
     * Retrieves a single property from an object with a given predicate. If
     * there are multiple properties with that predicate, will throw an access
     * exception.
     * </p>
     * 
     * @param predicate
     *            Qualified name of the predicate URI.
     * @return The single matching property.
     * @throws {@link org.nsdl.repository.access.AccessException}
     *             if there is more than one property with that predicate.
     */
    public Property getProperty(Property.Name predicate) throws AccessException;

    /**
     * Retrieve all properties contained in the object.
     * 
     * @return Set of all properties.
     */
    public Collection<Property> getProperties();

    /**
     * Retrieve all properties with a given predicate.
     * 
     * @param predicate
     *            Qualified name representing predicate URI.
     * @return Set of matching properties.
     */
    public Collection<Property> getProperties(Property.Name predicate);
    
    /**
     * Retrieve a specific relationship from an object.
     * <p>
     * Retrieves a single relationship from an object with a given predicate. If
     * there are multiple relationships with that predicate, will throw an
     * access exception.
     * </p>
     * 
     * @param predicate
     *            Qualified name of the predicate URI.
     * @return The single matching relationship.
     * @throws {@link org.nsdl.repository.access.AccessException}
     *             if there is more than one relationsiup with that predicate.
     */
    public Relationship getRelationship(Relationship.Name predicate)
            throws AccessException;

    /**
     * Retrieve all object relationships.
     * 
     * @return Set of all relationships in the object.
     */
    public Collection<Relationship> getRelationships();

    /**
     * Retrieve all relationships with a given predicate.
     * 
     * @param predicate
     *            Qualified name representing predicate URI.
     * @return Set of matching relationships.
     */
    public Collection<Relationship> getRelationships(Relationship.Name predicate);
    
    /**
     * Set the object's state.
     * 
     * @param s
     *            New state of the object.
     */
    public void setState(NDRObject.State s);

    /**
     * Add a single property to an object.
     * <p>
     * Properties (predicate + value pairs) that already exist in the object
     * will be ignored.
     * </p>
     * 
     * @param predicate
     *            Name of the property to add.
     * @param value
     *            Its literal value.
     */
    public void addProperty(Property.Name predicate, String value);

    /**
     * Add all given properties to the object.
     * <p>
     * Properties (predicate + value pairs) that already exist in the object
     * will be ignored.
     * </p>
     * 
     * @param toAdd
     *            Properties to add.
     */
    public void addProperties(Collection<Property> toAdd);

    /**
     * Remove a specific property from the object, if it exists.
     * 
     * @param name
     *            Name of the property to remove.
     * @param value
     *            Value of the property to be removed.
     */
    public void removeProperty(Property.Name name, String value);

    /**
     * Remove all the given properties.
     * <p>
     * For each property to remove, will only remove properties with matching
     * predicates and values. If there is no match (i.e. the property does not
     * exist in the object), it is skipped.
     * </p>
     * 
     * @param toRemove
     *            Properties to remove.
     */
    public void removeProperties(Collection<Property> toRemove);

    

    /**
     * Add a single Relationship to an object.
     * <p>
     * Relationships (subject + predicate + object) that already exist in the
     * object will be ignored. For adding a single relationship, only a
     * predicate and object is supplied, the subject is assumed to be the
     * {@linkplain NDRObject object} itself.
     * </p>
     * 
     * @param predicate
     *            Name of the relationship to add.
     * @param object
     *            The related object.
     */
    public void addRelationship(Relationship.Name predicate,
            NDRObjectInfo object);

    /**
     * Add all given Relationships to the object.
     * <p>
     * Relationships (subject + predicate + object) that already exist in the
     * object will be ignored.
     * </p>
     * 
     * @param toAdd
     *            Relationships to add.
     */
    public void addRelationships(Collection<Relationship> toAdd);

    /**
     * Remove a specific relationship from an object.
     * 
     * @param name
     *            Name of the relationship to remove
     * @param relatedObject
     *            Related object in the relationship to remove.
     */
    public void removeRelationship(Relationship.Name name,
            NDRObjectInfo relatedObject);

    /**
     * Remove all the given relationships.
     * <p>
     * For each relationship to remove, will only remove relationships with
     * matching predicates and objects. If there is no match (i.e. the
     * relationship does not exist in the object), it is skipped.
     * </p>
     * 
     * @param toRemove
     *            Relationships to remove.
     */
    public void removeRelationships(Collection<Relationship> toRemove);

    /**
     * Create and add a new datastream to an object.
     * 
     * @param name
     *            Name of the datastream to create and add.
     * @return New, empty added datastream.
     */
    public Datastream addDatastream(Datastream.Name name);

    /**
     * Add all the given datastreams to an object.
     * <p>
     * If there already exists a datastream in the object with the given ID, an
     * exception will be raised.
     * </p>
     * 
     * @param toAdd
     *            Datastreams to add.
     * @throws IllegalStateException
     *             There already exists a datastream in the object with a
     *             matching ID.
     */
    public void addDatastreams(Collection<Datastream> toAdd);

    /**
     * Remove the datastream with the given name, if it exists.
     * 
     * @param name
     *            Name of datastream to remove.
     */
    public void removeDatastream(Datastream.Name name);

    /**
     * Remove the given datastreams from an object
     * <p>
     * Datastreams with matching IDs will be removed from the object.
     * </p>
     * 
     * @param toRemove
     */
    public void removeDatastreams(Collection<Datastream> toRemove);
    
    /**
     * Retrieve a datastream with a particular datastream ID.
     * 
     * @param id
     *            The datastream identifier.
     * @return The specific datastream.
     */
    public Datastream getDatastream(org.nsdl.repository.model.Datastream.Name id);
    
    
    /**
     * Retrieve all object datastreams
     * 
     * @return Set of all datastreams in the object.
     */
    public Collection<Datastream> getDatastreams();
    
    
    /**
     * Set the object's handle.
     * 
     * @param handle
     *            New handle of the object.
     */
    public void setHandle(String handle);

    /**
     * Commit the object's changes to the repository.
     * 
     * @throws RepositoryException
     */
    public void commit() throws AccessException;
}
