/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;

/**
 * Represents a named literal value associated with an object.
 * <p>
 * While there may be domain-specific constraints on the content or form of the
 * literal value, the only basic requirement of an NDR property is that it is
 * not null.
 * </p>
 * 
 * @author birkland
 * @since 1.0
 */
public class Property implements Component {
    private final NDRObjectInfo subject;
    private final Name predicate;
    private String value;
    private int hash;

    /**
     * Construct a fully defined property.
     * 
     * @param subject
     *            Identity of the object that contains the property.
     * @param predicate
     *            Property name.
     * @param value
     *            Property value.
     */
    public Property(NDRObjectInfo subject, Component.Name<Property> predicate,
            String value) {
        if (predicate == null) {
            throw new NullPointerException("Undefined value in property "
                    + predicate + " " + value);
        } else if (!(predicate instanceof Name)) {
            throw new IllegalArgumentException("Expected name to be "
                    + Name.class.getName() + ", but instead got a "
                    + predicate.getClass().getName());
        }

        this.subject = subject;
        this.predicate = (Name) predicate;
        this.value = value;
        this.hash = hash();
    }

    private int hash() {
        return (predicate.getNamespaceURI() + predicate.getLocalPart() + " " + value)
                .hashCode();
    }

    /** {@inheritDoc} */
    public Type getType() {
        return Component.Type.Property;
    }

    /** {@inheritDoc} */
    public NDRObjectInfo subject() {
        return this.subject;
    }

    /**
     * The name of the property.
     * 
     * @return Qualified property name.
     */
    public Property.Name predicate() {
        return this.predicate;
    }

    /**
     * The value of the property.
     * 
     * @return String containing the property literal value.
     */
    public String value() {
        return this.value;
    }

    public void setValue(String value) {
        this.value = value;
        this.hash = hash();
    }

    /**
     * Text representation of a property.
     * 
     * Prints out the subject handle, predicate name, and literal value.
     * 
     * @return Text string of the property.
     */
    @Override
    public String toString() {
        String sub = subject() != null ? subject().handle() : null;
        String val = value() != null ? value().toString() : null;
        return "[" + sub + " " + predicate().toString() + " " + val + "]";
    }

    /**
     * Equivalence of NDR properties.
     * <p>
     * Properties in the NDR are considered equal if the predicates (i.e.
     * property name) and values are equal. Note that equality of the subject is
     * <em>not</em> considered.
     * </p>
     * 
     * @param o
     *            Object to compare the current Property with.
     * @return true if the properties are equal.
     */
    @Override
    public boolean equals(Object o) {
        if (o == null)
            return false;
        if (!(o instanceof Property))
            return false;

        Property comp = (Property) o;

        if (comp.value() == null) return false;
        return comp.predicate().equals(predicate()) && comp.value().equals(value());
    }

    /**
     * Hashed value of this relationship.
     * 
     * As equality for properties is only determined by the predicate and
     * literal value, the hash value follows suit. All equivalent properties
     * will have equivalent hash values.
     * 
     * @return int hash value.
     */
    @Override
    public int hashCode() {
        return hash;
    }

    /**
     * Qualified name of an NDR object Property.
     * 
     * @author birkland
     * @since 1.1
     */
    public static class Name extends QName implements Component.Name<Property> {
        public static final long serialVersionUID = 1;

        /**
         * Create a fully qualified property name.
         * <p>
         * All properties within the NDR must exist within an explicit
         * namespace. The prefix is primarily useful for serializing the
         * property name in formats such as XML or SPARQL.
         * </p>
         * 
         * @param namespaceURI
         *            URI string of the namespace.
         * @param localPart
         *            String containing a local name of the property.
         * @param prefix
         *            Preferred prefix (should be short and simple) to use when
         *            serializing the property name.
         */
        public Name(String namespaceURI, String localPart, String prefix) {
            super(namespaceURI, localPart, prefix);
        }

        public Name(String namespaceURI, String localPart) {
            super(namespaceURI, localPart);
        }

        /**
         * Get the URI value for this Property name.
         * 
         * Every qualified property name corresponds to a unique URI, which is
         * simply a concatenation of the namespace and local name of the given
         * property. This URI may be used to represent the property in RDF, for
         * example.
         * 
         * @return the URI value.
         */
        public String getURI() {
            return this.getNamespaceURI() + this.getLocalPart();
        }
    }

    /** Pattern for matching properties based on subject and predicate types. */
    public static class Pattern implements Component.Pattern {
        private final org.nsdl.repository.model.types.Type<? extends NDRObject> subject;
        private final Property.Name predicate;

        private static final Logger logger = Logger.getLogger(Pattern.class
                .getName());

        /**
         * Get a pattern that matches the given component.
         * 
         * @param c
         *            The component to find a pattern for.
         * @return The Pattern that matches the component. May be null if the
         *         component is not a Property.
         */
        public static Pattern getPattern(Component c) {
            logger.trace("Getting pattern for component..\n");
            if (Property.class.isInstance(c)) {
                logger.trace("success");
                return new Pattern(c.subject().type(), ((Property) c)
                        .predicate());
            } else {
                logger.warn("getPattern(Component c) not applicable");
                return null;
            }
        }

        /**
         * Create a pattern described in term object type and property name.
         * 
         * @param subjectType
         *            The type of object containing the property.
         * @param predicate
         *            Qualified name of the property.
         */
        public Pattern(
                org.nsdl.repository.model.types.Type<? extends NDRObject> subjectType,
                Property.Name predicate) {

            if (subjectType == null || predicate == null) {
                throw new IllegalArgumentException("Cannot instantiate "
                        + "pattern with null arguments");
            }

            this.subject = subjectType;
            this.predicate = predicate;

        }

        /**
         * The type of object that matches this pattern.
         * 
         * @return Matching object type.
         */
        public org.nsdl.repository.model.types.Type<? extends NDRObject> subject() {
            return this.subject;
        }

        /**
         * Qualified name of the property
         * 
         * @return Qualified name.
         */
        public Property.Name predicate() {
            return this.predicate;
        }

        /** {@inheritDoc} */
        public Component.Type getMatchingType() {
            return Component.Type.Property;
        }

        /** {@inheritDoc} */
        public boolean matches(Component c) {

            logger.trace("Checking match");
            if (c == null) {
                logger.trace("null, so no match");
                return false;
            }

            if (!(c instanceof Property)) {
                logger.trace("Incorrect type.  No match");
                return false;
            }

            boolean subjectsMatch = false;
            if (this.subject == null) {
                subjectsMatch = true;
            } else {
                subjectsMatch = this.subject.equals(((Property) c).subject()
                        .type());
            }

            boolean predicatesMatch = false;
            if (this.predicate == null) {
                predicatesMatch = true;
            } else {
                predicatesMatch = this.predicate.equals(((Property) c)
                        .predicate());
            }

            logger.trace(predicatesMatch && subjectsMatch);
            return predicatesMatch && subjectsMatch;
        }

        @Override
        public boolean equals(Object p) {
            if (p == null)
                return false;
            if (Property.Pattern.class.isAssignableFrom(p.getClass())) {
                boolean subjectsMatch = false;
                if (this.subject == null) {
                    subjectsMatch = ((Pattern) p).subject() == null;
                } else {
                    subjectsMatch = this.subject
                            .equals(((Pattern) p).subject());
                }

                boolean predicatesMatch = false;
                if (this.predicate == null) {
                    predicatesMatch = ((Pattern) p).predicate() == null;
                } else {
                    predicatesMatch = this.predicate.equals(((Pattern) p)
                            .predicate());
                }
                return subjectsMatch && predicatesMatch;
            }
            return false;
        }

        @Override
        public String toString() {
            return "Subject Type: " + this.subject.toString() + "\n"
                    + "Predicate: " + this.predicate.toString() + "\n";
        }

        @Override
        /* Ehh, probably not the most efficient, but should work */
        public int hashCode() {
            return (this.subject.toString() + this.predicate.toString())
                    .hashCode();
        }
    }
}
