/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;

/**
 * Represents a named relationship with another NDR object.
 * <p>
 * Relationships essentially represent labelled vertices of a graph of NDR
 * objects. Various models may have their own semantics defined for specific
 * relationships. On a basic structural level, though, they are composed of a
 * referring object (subject), a referred object (object), and a label
 * (predicate).
 * </p>
 * 
 * @author birkland
 * @since 1.0
 */
public class Relationship implements Component {
    private final NDRObjectInfo subject;

    private final Name predicate;

    private NDRObjectInfo object;

    private int hash = 0;

    /**
     * Create a relationship between two NDR objects.
     * 
     * @param subject
     *            Referring object in a directional relationship.
     * @param predicate
     *            Relationship name or label.
     * @param object
     *            Referred object in a directional relationship.
     */
    public Relationship(NDRObjectInfo subject,
            Component.Name<Relationship> predicate, NDRObjectInfo object) {
        if (subject == null || predicate == null) {
            throw new NullPointerException("Null pointer in relationship "
                    + subject + " " + predicate + " " + object);
        } else if (!(predicate instanceof Name)) {
            throw new IllegalArgumentException("Expected name to be "
                    + Name.class.getName() + ", but instead got a "
                    + predicate.getClass().getName());
        }

        this.subject = subject;
        this.predicate = (Name) predicate;
        this.object = object;
        this.hash = hash();

    }

    private int hash() {

        if (hash == 0) {
            /* If this is initialization, just use local vars */
            if (object == null) {
                return (subject + " " + predicate).hashCode();
            } else {
                return (predicate.toString() + " " + (object == null ? "null"
                        : object.handle())).hashCode();
            }
        } else {
            /* Otherwise, use methods.. who knows if we're extended */
            if (related() == null) {
                return (subject() + " " + predicate()).hashCode();
            } else {
                return (predicate.toString() + " " + (related() == null ? "null"
                        : related().handle())).hashCode();
            }
        }
    }

    /**
     * Referring object in the relationship.
     * 
     * @return Identity of the referring object.
     */
    public NDRObjectInfo subject() {
        return this.subject;
    }

    /**
     * The name or label of the relationship.
     * 
     * @return Qualified name of the relationship.
     */
    public Relationship.Name predicate() {
        return this.predicate;
    }

    /**
     * The referred object in the relationship
     * 
     * @return Identity of the referred (target) object.
     */
    public NDRObjectInfo related() {
        return this.object;
    }

    public void setRelated(NDRObjectInfo relatedTo) {
        this.object = relatedTo;
        this.hash = hash();
    }

    /** {@inheritDoc} */
    public Component.Type getType() {
        return Component.Type.Relationship;
    }

    /**
     * Text representation of a relationship.
     * 
     * Prints out the subject handle, predicate name, and object handle for the
     * current relationship.
     * 
     * @return Text string of relationship.
     */
    public String toString() {
        String sub = subject() != null ? subject().handle() : null;
        String obj = related() != null ? related().handle() : null;
        return ("[" + sub + " " + predicate() + " " + obj + "]");
    }

    /**
     * Determine equality of two relationships.
     * <p>
     * Relationships are considered equal if their predicates and their subject
     * handles match exactly. Note: equivalence does <em>not</em> require that
     * subject handles be equal!. So, for the for the purpose of equivalence as
     * defined here, X-R-Y is equivalent to Z-R-Y For objects X, Z, and Y, and
     * relationship name R.
     * </p>
     * 
     * @param o
     *            Object to compare relationship with.
     * @return true if the relationships are equivalent.
     */
    public boolean equals(Object o) {
        if (o == null)
            return false;
        if (!(o instanceof Relationship))
            return false;

        /* If we're comparing to ourself, then yes, we're equal */
        if (this == o) {
            return true;
        }

        Relationship comp = (Relationship) o;

        /*
         * Ugg.. digging a grave perhaps? The following rules apply:
         */
        // A B C == Y B C
        // A B C != A B D
        // A B null == A B null
        // null B C == null B C
        if (comp.related() == null && related() == null) {

            /* A B null == A B null */
            if (comp.subject() == null || subject() == null
                    || comp.subject.handle() == null
                    || subject().handle() == null) {
                return false;
            } else {
                return (comp.subject().handle().equals(subject().handle()) && comp
                        .predicate().equals(predicate()));
            }
        } else if (comp.related() == null || related() == null
                || comp.related().handle() == null) {
            /* A B C != A B null */
            return false;
        } else {
            /* A B C = Y B C */
            return comp.predicate().equals(predicate())
                    && comp.related().handle().equals(related().handle());
        }
    }

    /**
     * Hashed value of this relationship.
     * 
     * As equality for relationships is only determined by the predicate and
     * object, the hash value follows suit. All equivalent relationships will
     * have equivalent hash values.
     * 
     * @return int hash value.
     */
    public int hashCode() {
        return hash;
    }

    public static enum Direction {
        IN, OUT
    }

    /**
     * Qualified name of an NDR object Relationship.
     * 
     * @author birkland
     * @since 1.1
     */
    public static class Name extends QName implements
            Component.Name<Relationship> {
        public static final long serialVersionUID = 1;

        /**
         * Create a fully qualified relationship name.
         * <p>
         * All relationships within the NDR must exist within an explicit
         * namespace. The prefix is primarily useful for serializing the
         * relationship name in formats such as XML or SPARQL.
         * </p>
         * 
         * @param namespaceURI
         *            URI string of the namespace.
         * @param localPart
         *            String containing a local name of the relationship.
         * @param prefix
         *            Preferred prefix (should be short and simple) to use when
         *            serializing the relationship name.
         */
        public Name(String namespaceURI, String localPart, String prefix) {
            super(namespaceURI, localPart, prefix);
        }

        public Name(String namespaceURI, String localPart) {
            super(namespaceURI, localPart);
        }

        /**
         * Get the URI value for this Relationship name.
         * 
         * Every qualified relationship name corresponds to a unique URI, which
         * is simply a concatenation of the namespace and local name of the
         * given Relationship. This URI may be used to represent the
         * relationship in RDF, for example.
         * 
         * @return the URI value.
         */
        public String getURI() {
            return this.getNamespaceURI() + this.getLocalPart();
        }
    }

    /**
     * Matches relationships based on predicate and subject, object types.
     * <p>
     * The RelationshipPattern matches a given relationship based on Subject
     * type, predicate, and Object type. There are no provisions for partial
     * matches or "wildcards".
     * </p>
     * 
     * @author birkland
     * 
     */
    public static class Pattern implements Component.Pattern {
        private final org.nsdl.repository.model.types.Type<? extends NDRObject> subject;
        private final Relationship.Name predicate;
        private final org.nsdl.repository.model.types.Type<? extends NDRObject> object;

        private static final Logger logger = Logger.getLogger(Pattern.class
                .getName());

        public static Pattern getPattern(Component c) {
            if (Relationship.class.isInstance(c) == false) {
                logger.warn("getPattern to " + c.getClass().getName()
                        + " not applicable");
                return null;
            }

            return new Pattern(c.subject().type(), ((Relationship) c)
                    .predicate(), ((Relationship) c).related().type());
        }

        /** {@inheritDoc} */
        public Component.Type getMatchingType() {
            return Component.Type.Relationship;
        }

        /**
         * Create a pattern based on the related object types and the
         * relationship name.
         * 
         * @param subject
         *            Type of referring object in matching relationships.
         * @param predicate
         *            Name of matching relationship type.
         * @param object
         *            Type of referred object in matching relationships.
         */
        public Pattern(
                org.nsdl.repository.model.types.Type<? extends NDRObject> subject,
                Relationship.Name predicate,
                org.nsdl.repository.model.types.Type<? extends NDRObject> object) {

            this.subject = subject;
            this.predicate = predicate;
            this.object = object;
        }

        /**
         * Type of object matching the referring object.
         * 
         * @return Matching referring object type.
         */
        public org.nsdl.repository.model.types.Type<? extends NDRObject> subject() {
            return this.subject;
        }

        /**
         * Name of the matching relationship.
         * 
         * @return Qualified relationship name.
         */
        public Relationship.Name predicate() {
            return this.predicate;
        }

        /**
         * Type of object matching the referred object.
         * 
         * @return Matching referred object type.
         */
        public org.nsdl.repository.model.types.Type<? extends NDRObject> object() {
            return this.object;
        }

        /** {@inheritDoc} */
        public boolean matches(Component c) {

            if (c == null) {
                logger.trace("null, so no match");
                return false;
            }
            if (!(c instanceof Relationship)) {
                return false;
            }
            logger.trace("Checking match of " + getPattern(c) + " to " + this);

            boolean subjectsMatch = false;
            if (this.subject == null) {
                subjectsMatch = true;
            } else {
                subjectsMatch = this.subject.equals(((Relationship) c)
                        .subject().type());
            }

            boolean predicatesMatch = false;
            if (this.predicate == null) {
                predicatesMatch = true;
            } else {
                predicatesMatch = this.predicate.equals(((Relationship) c)
                        .predicate());
            }

            boolean objectsMatch = false;
            if (this.object == null) {
                objectsMatch = true;
            } else {
                objectsMatch = this.object.equals(((Relationship) c).related()
                        .type());
            }

            logger.trace(subjectsMatch && predicatesMatch && objectsMatch);
            return (subjectsMatch && predicatesMatch && objectsMatch);
        }

        @Override
        public boolean equals(Object p) {

            if (p == null)
                return false;
            if (Relationship.Pattern.class.isAssignableFrom(p.getClass())) {
                boolean subjectsMatch = false;
                if (this.subject == null) {
                    subjectsMatch = ((Pattern) p).subject() == null;
                } else {
                    subjectsMatch = this.subject
                            .equals(((Pattern) p).subject());
                }

                boolean predicatesMatch = false;
                if (this.predicate == null) {
                    predicatesMatch = ((Pattern) p).predicate() == null;
                } else {
                    predicatesMatch = this.predicate.equals(((Pattern) p)
                            .predicate());
                }

                boolean objectsMatch = false;
                if (this.object == null) {
                    objectsMatch = ((Pattern) p).object() == null;
                } else {
                    objectsMatch = this.object.equals(((Pattern) p).object());
                }

                return subjectsMatch && predicatesMatch && objectsMatch;
            }
            return false;
        }

        @Override
        /* Ehh, probably not the most efficient, but should work */
        public int hashCode() {
            return (this.subject.toString() + this.predicate.toString() + this.object
                    .toString()).hashCode();
        }

        @Override
        public String toString() {
            return "[" + this.subject + ", " + this.predicate + ", "
                    + this.object + "]";
        }
    }
}
