/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.util.NDRObjectTemplate;

import static org.nsdl.repository.model.content.ModelUtil.*;
import static org.nsdl.repository.model.content.Scope.*;

/**
 * Agent-specific content model.
 * <p>
 * Defines the components specific to an Agent, and enforces the following
 * rules:
 * <ul>
 * <li>An agent must have exactly one identifier property, which must be unique
 * among all active agents. This identifier must be one of the three distinct
 * identifier properties enumerated in
 * {@link org.nsdl.repository.model.content.AgentModel.Identifiers}</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class AgentModel implements ContentModel {

    /** Namespace for all components in the common model */
    public static final String NS = CommonModel.NS;

    /** Default prefix for all components in common model */
    public static final String PREFIX = CommonModel.PREFIX;

    private static final Logger logger = Logger.getLogger(AgentModel.class);

    private NDRAccess ndr;

    /**
     * Components Agents have in the base model.
     * 
     * @author birkland
     * @since 1.1
     */
    public static interface Components extends CommonModel.Components {

        /**
         * Host-based identifier.
         * 
         * @see org.nsdl.repository.model.content.AgentModel.Identifiers#HOST
         */
        public static final Property.Name IDENTIFIER_HOST =
                new Property.Name(NS, Constants.PROPERTY_HAS_RESOURCE + "HOST",
                        PREFIX);

        /**
         * User-based identifier.
         * 
         * @see org.nsdl.repository.model.content.AgentModel.Identifiers#USER
         */
        public static final Property.Name IDENTIFIER_USER =
                new Property.Name(NS, Constants.PROPERTY_HAS_RESOURCE + "USER",
                        PREFIX);

        /**
         * Other identifier.
         * 
         * @see org.nsdl.repository.model.content.AgentModel.Identifiers#OTHER
         */
        @Shared
        public static final Property.Name IDENTIFIER_OTHER =
                new Property.Name(NS,
                        Constants.PROPERTY_HAS_RESOURCE + "OTHER", PREFIX);
    }

    /**
     * Agent properties that serve as Identifiers.
     * 
     * @author birkland
     * @since 1.1
     */
    public static interface Identifiers extends Definitions {

        /**
         * Host-based identifier.
         * <p>
         * Identifier that has a value based on a host name or domain name. For
         * example, <code>dlese.org</code>, <code>java.sun.com</code>.
         * This should be used for agents that represent organizations or
         * entities.
         * </p>
         */
        public static final Property.Name HOST = Components.IDENTIFIER_HOST;

        /**
         * User-based identifier.
         * <p>
         * Represents an Agent that is a single human user. There are no
         * restrictions on form, but typically, <code>USER</code> identifier
         * values are based on an individual's username in an application or
         * identity management system.
         * </p>
         */
        public static final Property.Name USER = Components.IDENTIFIER_USER;

        /**
         * Other identifier.
         * <p>
         * Can be used to identify agents by some other criteria other than user
         * name or host name. In general, it is preferable to use one of
         * <code>USER</code> or <code>HOST</code> if it makes sense to do
         * so.
         * </p>
         */
        @Shared
        public static final Property.Name OTHER = Components.IDENTIFIER_OTHER;
    }

    /* Cache these here since they never change */
    private static final Set<Property.Name> identifiers =
            getPropertyNames(Identifiers.class, All);

    private static final Set<Property.Name> properties =
            getPropertyNames(Components.class, All);

    private static final Set<Property.Name> exclusiveProperties =
            getPropertyNames(Components.class, Exclusive);

    private static final Set<Relationship.Name> exclusiveRelationships =
            getRelationshipNames(Components.class, Exclusive);

    private static final Set<Relationship.Name> relationships =
            getRelationshipNames(Components.class, All);

    public void init(NDRAccess ndr) {
        this.ndr = ndr;
    }

    public void checkCompliance(NDRObject o) throws ContentModelException {

        logger.trace("Checking base Agent compliance");
        if (!Type.Agent.equals(o.type())) {
            checkExclusiveComponents(o);

            logger.trace("Not an Agent.  Done checking coompliance");
            return;
        }

        checkProtectedNS(o);

        assureExactlyOneAgentIdentifier(o);
        assureUniqueIdentifier(o);
        logger.trace("Done checking Agent compliance");
    }

    private void assureExactlyOneAgentIdentifier(NDRObject o)
            throws ContentModelException {

        HashSet<Property> encounteredIDs = new HashSet<Property>();

        for (Property p : o.getProperties()) {
            if (identifiers.contains(p.predicate())) {
                encounteredIDs.add(p);
            }
        }

        if (encounteredIDs.size() == 0) {
            throw new ContentModelException("No identifiers have been provided");
        } else if (encounteredIDs.size() > 1) {
            throw new ContentModelException("Too many identifiers provided!  "
                    + "An agent must have exactly one");
        }

        Property.Name idProperty = encounteredIDs.iterator().next().predicate();

        if (!identifiers.contains(idProperty)) {
            throw new ContentModelException(idProperty
                    + " is not a valid agent identifier property");
        }
    }

    private void assureUniqueIdentifier(NDRObject o)
            throws ContentModelException {

        if (!State.Active.equals(o.state())) {
            return;
        }

        Property id = null;
        for (Property p : o.getProperties()) {
            if (identifiers.contains(p.predicate())) {
                id = p;
                break;
            }
        }

        NDRObjectTemplate template = new NDRObjectTemplate(Type.Agent);
        template.setState(State.Active);
        template.addProperty(id.predicate(), id.value());

        /* N.B. Assumes handle equality */
        Results<NDRObjectInfo> matches =
                ndr.getObjectFinder().matchObject(template);
        try {
            for (NDRObjectInfo match : matches) {
                if (!match.handle().equals(o.handle())) {
                    throw new ContentModelException("Agent identifier "
                            + id.predicate() + " '" + id.value()
                            + "' is being used by " + match.handle());
                }
            }
        } finally {
            matches.close();
        }
    }

    private static void checkProtectedNS(NDRObject o)
            throws ContentModelException {

        checkNSProperties(NS, properties, o);
        checkNSRelationships(NS, relationships, o);
    }

    private static void checkExclusiveComponents(NDRObject o)
            throws ContentModelException {
        checkExclusiveProperties(o, exclusiveProperties);
        checkExclusiveRelationships(o, exclusiveRelationships);
    }

}
