/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.content;

import java.util.Collection;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.NDRObject.State;

import static org.nsdl.repository.model.NDRObject.State.Active;
import static org.nsdl.repository.model.content.ModelUtil.*;
import static org.nsdl.repository.model.content.Scope.*;

/**
 * Aggregator-specific content model.
 * <p>
 * Defines components specific to Aggregators, and enforces the following rules:
 * <ul>
 * <li>Active aggregators must be
 * {@link AggregatorModel.Components#AGGREGATOR_FOR AGGREGATOR_FOR}
 * <em>exactly</em> one active agent.</li>
 * <li>Active Aggregators must have <em>at most</em> one
 * {@link AggregatorModel.Components#ASSOCIATED_WITH ASSOCIATED_WITH}
 * relationship to an active Resource</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class AggregatorModel implements ContentModel {

    private static final Logger logger = Logger
            .getLogger(AggregatorModel.class);

    /** Namespace for all components in the common model. */
    public static final String NS = CommonModel.NS;

    /** Default prefix for all components in common model. */
    public static final String PREFIX = CommonModel.PREFIX;

    private NDRAccess ndr;

    /**
     * Components Aggregators have in the base model.
     * 
     * @author birkland
     * @since 1.1
     */
    public static interface Components extends CommonModel.Components {

        /**
         * Relates an Aggregator to its owning Agent.
         * <p>
         * Every Aggregator is "owned" by exactly one Agent. In that sense, the
         * Agent that "owns" a particular Aggregator is responsible for its
         * content. Thus, the provenance of every membership in a given
         * Aggregator can be attributed to exactly one Agent.
         * </p>
         */
        public static final Relationship.Name AGGREGATOR_FOR = new Relationship.Name(
                NS, "aggregatorFor", PREFIX);

        /**
         * Associates an aggregation with a representative Resource.
         * <p>
         * If an Aggregator is associated with a resource, that means that the
         * Resource's content provides some concrete representation of that
         * aggregation. For example, suppose somebody is creating a wiki page
         * about clouds, and has chosen some cloud resources from the NSDL to
         * discuss on the page. If that wiki page were a resource in the NDR,
         * then it would be possible to construct an aggregator containing those
         * resources and associate that group with the wiki page in the
         * repository. Thus, the group or cloud resources would be associated
         * with a wiki page that describes them as a group.
         * </p>
         */
        public static final Relationship.Name ASSOCIATED_WITH = new Relationship.Name(
                NS, "associatedWith", PREFIX);

        public static final Datastream.Name SERVICE_DESCRIPTION = new Datastream.Name(
                "serviceDescription");
    }

    /* Cache these here since they never change */
    private static final Set<Property.Name> properties = getPropertyNames(
            Components.class, All);
    static {
        /*
         * XXX OAI model properties are in the same namespace, so we have to
         * consider them here
         */
        properties.addAll(getPropertyNames(OAIModel.Set.Components.class, All));
    }

    private static final Set<Property.Name> exclusiveProperties = getPropertyNames(
            Components.class, Exclusive);

    private static final Set<Relationship.Name> exclusiveRelationships = getRelationshipNames(
            Components.class, Exclusive);

    private static final Set<Relationship.Name> relationships = getRelationshipNames(
            Components.class, All);

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        this.ndr = ndr;
    }

    /** {@inheritDoc} */
    public void checkCompliance(NDRObject o) throws ContentModelException {

        logger.trace("Checking base Aggregator compliance");

        /* Only consider Aggregators */
        if (!Type.Aggregator.equals(o.type())) {

            checkExclusiveComponents(o);

            if (Type.Agent.equals(o.type())) {
                logger.trace("Checking aggregator dependencies of agent");
                assureAgentDependenciesSatisfied(o);
            } else if (Type.Resource.equals(o.type())) {
                logger.trace("Checking aggregator dependencies of resource");
                assureResourceDependenciesSatisfied(o);
            }
            logger.trace("Not an Aggregator.  Done checking coompliance");

            return;
        }

        checkProtectedNS(o);

        assureExactlyOneActiveAgent(o);
        assureZeroOrOneActiveAssociations(o);

        logger.trace("Done checking compliance of Aggregator");
    }

    private void assureExactlyOneActiveAgent(NDRObject o)
            throws ContentModelException {
        Collection<Relationship> aggRels = o
                .getRelationships(Components.AGGREGATOR_FOR);

        if (aggRels.size() != 1) {
            throw new ContentModelException(aggRels.size()
                    + " aggregator relationships found, there "
                    + "should be exactly 1");
        }

        NDRObject agent = null;
        try {
            agent = ndr.getObject(aggRels.iterator().next().related());
        } catch (ExistenceException e) {
            throw new ContentModelException("No Agent for aggregator "
                    + o.handle(), e);
        }

        if (!Type.Agent.equals(agent.type())) {
            throw new ContentModelException(
                    "Aggregator must be aggregator for " + "an agent.  "
                            + agent.handle() + " is a " + agent.type());
        }

        /* If we are active, the related agent must be active */
        if (Active.equals(o.state()) && !Active.equals(agent.state())) {
            throw new ContentModelException(
                    "The aggregator's agent must be active");
        }
    }

    /**
     * There most be at most on associatedWith relationship
     * <p>
     * If there exists one associatedWith relationship, then the object at the
     * other end must be a resource and it must be active.
     * </p>
     */
    private void assureZeroOrOneActiveAssociations(NDRObject o)
            throws ContentModelException {
        Collection<Relationship> awRels = o
                .getRelationships(Components.ASSOCIATED_WITH);

        if (awRels.size() > 1) {
            throw new ContentModelException("There must be at most one "
                    + "associatedWith relationship for an aggregator. "
                    + "this one has " + awRels.size());
        }

        if (awRels.size() == 1) {
            NDRObject associated = null;
            try {
                associated = ndr.getObject(awRels.iterator().next().related());
            } catch (ExistenceException e) {
                throw new ContentModelException("Associated resource "
                        + awRels.iterator().next().related().handle()
                        + " does not exist inthe repository", e);
            }

            if (!Type.Resource.equals(associated.type())) {
                throw new ContentModelException("Aggregators may only"
                        + " be associated with resources."
                        + associated.handle() + " is a " + associated.type());
            }

            if (Active.equals(o.state()) && !Active.equals(associated.state())) {
                throw new ContentModelException("Associated resource "
                        + associated.handle() + " is not active");
            }
        }
    }

    /**
     * Make sure that a non-active agent has no active Aggregators attached to
     * it
     * 
     * @param o
     *            NDRObject that is supposed to be an agent
     */
    private void assureAgentDependenciesSatisfied(NDRObject o)
            throws ContentModelException {

        if (!o.state().equals(State.Active)) {

            Results<Relationship> aggregatorForRels = ndr.getObjectFinder()
                    .getRelatedTo(o, Components.AGGREGATOR_FOR);
            try {
                for (Relationship aggregatorFor : aggregatorForRels) {

                    NDRObject aggregator = ndr.getObject(aggregatorFor
                            .subject());

                    if (aggregator.state().equals(State.Active)) {
                        throw new ContentModelException(
                                "Unresolved dependency with Agent's Aggregator "
                                        + aggregator.handle());
                    }
                }
            } finally {
                aggregatorForRels.close();
            }
        }
    }

    /**
     * Make sure that a non-active Resource has no active Aggregators
     * associating with it
     * 
     * @param o
     *            NDRObject that is supposed to be a resource
     */
    private void assureResourceDependenciesSatisfied(NDRObject o)
            throws ContentModelException {

        if (!o.state().equals(State.Active)) {

            Results<Relationship> associatedWithRels = ndr.getObjectFinder()
                    .getRelatedTo(o, Components.ASSOCIATED_WITH);

            try {
                for (Relationship associatedWith : associatedWithRels) {

                    NDRObject aggregator = ndr.getObject(associatedWith
                            .subject());

                    if (aggregator.state().equals(State.Active)) {
                        throw new ContentModelException(
                                "Unresolved dependency with associated Aggregator "
                                        + aggregator.handle());
                    }
                }
            } finally {
                associatedWithRels.close();
            }
        }
    }

    private static void checkProtectedNS(NDRObject o)
            throws ContentModelException {

        checkNSProperties(NS, properties, o);
        checkNSRelationships(NS, relationships, o);
    }

    private static void checkExclusiveComponents(NDRObject o)
            throws ContentModelException {
        checkExclusiveProperties(o, exclusiveProperties);
        checkExclusiveRelationships(o, exclusiveRelationships);
    }
}
