/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.content;

import java.util.Collection;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.NDRObject.State;

import static org.nsdl.repository.model.content.ModelUtil.*;
import static org.nsdl.repository.model.content.Scope.*;

/**
 * Authorization-specific content model.
 * <p>
 * Defines the components that are specific to authorization in the repository
 * and enforces the following rules:
 * <ul>
 * <li>Every Aggregator or MetadataProvider must have an
 * {@link AuthorizationModel.Grouping.Components#AUTHORIZED AUTHORIZED}
 * relationship indicating who has permission to alter its contents.</li>
 * <li>The object of the
 * {@link AuthorizationModel.Grouping.Components#AUTHORIZED AUTHORIZED}
 * relationship must be an active Aggregator or Agent.</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class AuthorizationModel implements ContentModel {

    /** Authorization namespace. */
    public static final String NS = "http://ns.nsdl.org/ndr/auth#";

    /** Default authorization prefix */
    public static final String PREFIX = "auth";

    /* Models that comprise this model */
    private AuthorizationModel.Agent agentModel = new AuthorizationModel.Agent();
    private AuthorizationModel.Grouping groupingModel = new AuthorizationModel.Grouping();

    private static final Logger logger = Logger
            .getLogger(AuthorizationModel.class);

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        agentModel.init(ndr);
        groupingModel.init(ndr);
    }

    /**
     * Authorization model as applied to groupings (Aggregators,
     * MetadataProviders).
     * <p>
     * Aggregators and MetadataProviders must specify either an Agent or an
     * Aggregator of agents as the sole entity (or entities) allowed to modify
     * its contents. This model defines that
     * {@link AuthorizationModel.Grouping.Components#AUTHORIZED authorization}
     * specification relationship, and assures that it points to the correct
     * type of object.
     * </p>
     * 
     * @author birkland
     * @since 1.1
     * 
     */
    public static class Grouping implements ContentModel {
        private NDRAccess ndr;

        /* Cache these here since they never change */
        private final Set<Property.Name> exclusiveProperties = getPropertyNames(
                Components.class, Exclusive);

        private final Set<Relationship.Name> exclusiveRelationships = getRelationshipNames(
                Components.class, Exclusive);

        /**
         * Properties and relationships unique to grouping objects in the
         * Authorization model.
         * <p>
         * Grouping objects are Aggregators or MetadataProviders. We define an
         * 'authorized to change' relationship for these objects.
         * </p>
         * 
         * @author birkland
         * @since 1.1
         */
        public static interface Components extends Definitions {

            /**
             * Relationship to the Agent or group of Agents can modify
             * Aggregator contents.
             * <p>
             * This relationship relates a Grouping object (i.e. an Aggregator
             * or MetadataProvider) to an Agent or group of Agents. It indicates
             * that the specified Agent (or group of Agents) are the sole
             * entities in the repository that have permission to modify the
             * given Aggregator or MetadataProvider's contents.
             * </p>
             */
            public static final Relationship.Name AUTHORIZED = new Relationship.Name(
                    NS, "authorizedToChange", PREFIX);

        }

        /** {@inheritDoc} */
        public void init(NDRAccess ndr) {
            this.ndr = ndr;
        }

        /** {@inheritDoc} */
        public void checkCompliance(NDRObject o) {

            if (!(Type.Aggregator.equals(o.type()) || Type.MetadataProvider
                    .equals(o.type()))) {

                /*
                 * if called the 'usual' way, through
                 * AgentModel.checkCompliance(), we will never get here
                 */
                logger.trace("Not an Aggregator or MetadataProvider");

                /*
                 * Make sure that no Grouping-specific components are found on
                 * non-Grouping objects.
                 */
                checkExclusiveProperties(o, exclusiveProperties);
                checkExclusiveRelationships(o, exclusiveRelationships);
            } else {

                /*
                 * For objects thatare groupings, make sure that they have an
                 * authorization relationship to the right kind of object
                 * (Aggregator or Agent, in this case)
                 */
                assureProperAuthorizedObject(o);
            }
        }

        /**
         * Authorized entity should be an active Agent or Aggregator.
         * 
         * @param o
         *            Object that has authorization relationship to an entity.
         * @throws ContentModelException
         */
        private void assureProperAuthorizedObject(NDRObject o)
                throws ContentModelException {

            Collection<Relationship> authRels = o
                    .getRelationships(Components.AUTHORIZED);

            if (authRels.isEmpty()) {
                throw new ContentModelException(
                        "Authorization (authorizedToChange relationship) "
                                + "is not defined");
            }

            for (Relationship authorization : authRels) {

                NDRObject authorized = ndr.getObject(authorization.related());

                if (!(Type.Agent.equals(authorized.type()) || Type.Aggregator
                        .equals(authorized.type()))) {
                    throw new ContentModelException(
                            "Only an Agent or an Aggregator "
                                    + "may be given authorization for this object.  "
                                    + authorized.type() + " is neither.");
                }

                if (State.Active.equals(o.state())
                        && !State.Active.equals(authorized.state())) {
                    throw new ContentModelException("Authorized object "
                            + authorized.handle() + " must be active");
                }
            }
        }
    }

    /**
     * Authorization model as applied to Agents.
     * <p>
     * Currently, only the <code>publicKey</code> datastream is of interest to
     * Agents in the authorization model. It is optional. This sub-model merely
     * enforces that <code>publicKey</code>, if present, is only on an Agent and
     * no other object type.
     * </p>
     * 
     * @author birkland
     * @since 1.1
     */
    public static class Agent implements ContentModel {

        /** {@inheritDoc} */
        public void init(NDRAccess ndr) {
            /* not needed */
        }

        /**
         * Checks compliance with respect to Agent sub-model.
         * 
         * @param o
         *            NDRObject to be checked for compliance with Agent
         *            sub-model;
         */
        public void checkCompliance(NDRObject o) throws ContentModelException {

            /*
             * All we do here is make sure that if an object is not an agent, it
             * doesn't have a public key datastream.
             */
            if (!o.type().equals(Type.Agent)) {
                if (o.getDatastream(Components.PUBLIC_KEY) != null) {
                    throw new ContentModelException(o.type()
                            + " types should not have a "
                            + Components.PUBLIC_KEY + " datastream");
                }
            }
        }

        /**
         * Components specific to Agents in the Authorization model.
         * 
         * @author birkland
         * @since 1.1
         */
        public interface Components extends Definitions {
            /**
             * Public key datastream.
             * <p>
             * The public key datastream contains a serialized version of an
             * Agent's public key. This key is used in the authentication
             * protocol suppprted by the API, version 1.0. It is optional (not
             * all agents need or want to log in and make updates to the
             * repository), but if present, it must be an Agent.
             * </p>
             */
            public static final Datastream.Name PUBLIC_KEY = new Datastream.Name(
                    "PublicKey");
        }
    }

    /** {@inheritDoc} */
    public void checkCompliance(NDRObject o) throws ContentModelException {

        logger.trace("Checking authorization model compliance");

        if (Type.Aggregator.equals(o.type())
                || Type.MetadataProvider.equals(o.type())) {

            groupingModel.checkCompliance(o);
        } else if (Type.Agent.equals(o.type())) {
            agentModel.checkCompliance(o);
        } else {

            checkNSProperties(NS, null, o);
            checkNSRelationships(NS, null, o);
        }
        logger.trace("done checking authorization model compliance");
    }
}
