/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.NDRObject.State;

import static org.nsdl.repository.model.content.ModelUtil.*;
import static org.nsdl.repository.model.content.CommonModel.Components.MEMBER_OF;
import static org.nsdl.repository.model.content.MetadataModel.Components.METADATA_FOR;
import static org.nsdl.repository.model.content.MetadataModel.Components.METADATA_PROVIDED_BY;

/**
 * Models a Collection within the repository.
 * <p>
 * Collections are defined as a set of Resources and Metadata grouped by a
 * single Agent such that all Metadata in a collection describes a Resource in
 * the same collection. Thus, a Collection is composed of the Resources,
 * Metadata, an Aggregator, at least one MetadataProvider, and an Agent. This
 * model defines the relationships unique to this construct and enforces the
 * following rules for Active objects:
 * <ul>
 * <li>Every collection must have exactly one Aggregator</li>
 * <li>Any MetadataProviders that have an
 * {@link CollectionsModel.Components#AGGREGATED_BY AGGREGATED_BY} relationship
 * to that Aggregator are part of definition of the collection.</li>
 * <li>Any Metadata objects provided collection MetadataProviders are in the
 * collection.</li>
 * <li>Each Metadata object in the collection must describe a Resource that is
 * in the collection</li>
 * <li>Every Resource in the collection must be
 * {@link CommonModel.Components#MEMBER_OF MEMBER_OF} the collection Aggregator.</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class CollectionsModel implements ContentModel {

    private static final Logger logger =
            Logger.getLogger(CollectionsModel.class);

    /**
     * Collections namespace (currently the Common prefix).
     * <p>
     * Currently, the collections model uses the same prefix as the common
     * model. This may be changed.
     * </p>
     */
    public static final String NS = CommonModel.NS;

    /** Default collections prefix */
    public static final String PREFIX = CommonModel.PREFIX;

    /**
     * Collection model-specific components.
     * 
     * @author birkland
     * @since 1.1
     */
    public static interface Components extends Definitions {

        /**
         * Makes a MetadataProvider part of a collection definition.
         * <p>
         * Any MetadataProviders that have the
         * {@linkplain CollectionsModel.Components#AGGREGATED_BY AGGREGATED_BY}
         * relationshyp to a given Aggregator comprise that collection's
         * definition. Any Metadata provided by these providers is part of the
         * collection.
         * </p>
         */
        public static final Relationship.Name AGGREGATED_BY =
                new Relationship.Name(NS, "aggregatedBy", PREFIX);

        /**
         * Collection naming authority. Basically, this is an identifier for a
         * collection. it is here for legacy reasons, and shall either be
         * removed or formally introduced. Unless you know what it is, ignore
         * it!
         */
        public static final Property.Name COLLECTION_NA =
                new Property.Name(NS, "collectionNA", PREFIX);
    }

    private NDRAccess ndr;

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        this.ndr = ndr;
    }

    private static Set<Relationship.Name> exclusiveRels =
            ModelUtil.getRelationshipNames(Components.class, Scope.Exclusive);

    /** {@inheritDoc} */
    public void checkCompliance(NDRObject o) throws ContentModelException {
        if (Type.Metadata.equals(o.type()) && State.Active.equals(o.state())) {
            assureMetadataDescribesAggregatedResource(o);
            checkExclusiveRelationships(o, exclusiveRels);
        } else if ((Type.Resource.equals(o.type()) || Type.Aggregator.equals(o
                .type()))
                && State.Active.equals(o.state())) {
            assureDeaggregatedResourceHasNoRequiredMetadata(o);
            checkExclusiveRelationships(o, exclusiveRels);
        } else if (Type.MetadataProvider.equals(o.type())
                && State.Active.equals(o.state())) {
            assureAggregatedByIsCorrect(o);
        } else if (Type.MetadataProvider.equals(o.type())) {
            /* Deleted MPs won't have anything checked */
        } else {
            checkExclusiveRelationships(o, exclusiveRels);
        }
    }

    public void assureAggregatedByIsCorrect(NDRObject o)
            throws ContentModelException {

        Relationship aggregatedBy = null;
        try {
            aggregatedBy = o.getRelationship(Components.AGGREGATED_BY);
        } catch (AccessException e) {
            throw new ContentModelException(
                    "too many AggregatedBy relationships");
        }

        if (aggregatedBy == null) {
            return;
        }

        NDRObject aggregator = ndr.getObject(aggregatedBy.related());

        if (State.Active.equals(aggregator.state())
                && !Type.Aggregator.equals(aggregator.type())) {
            throw new ContentModelException(
                    "aggregatedBy relationship must point to an Aggregator");
        }
    }

    private void assureMetadataDescribesAggregatedResource(NDRObject o)
            throws ContentModelException {

        NDRObject mp = null;
        try {
            mp =
                    ndr.getObject(o.getRelationship(METADATA_PROVIDED_BY)
                            .related());
        } catch (ExistenceException e) {
            throw new ContentModelException(
                    "MetadataProvider must exist in the repository", e);
        }
        Relationship aggBy = mp.getRelationship(Components.AGGREGATED_BY);

        if (aggBy == null) {
            return;
        }

        NDRObject aggregator = ndr.getObject(aggBy.related());

        boolean aggregatedObjectFound = false;

        for (Relationship m4 : o.getRelationships(METADATA_FOR)) {

            NDRObject describedObject = null;

            try {
                describedObject = ndr.getObject(m4.related());
            } catch (ExistenceException e) {
                throw new ContentModelException(
                        "The objects that Metadata describe must exist!");
            }

            logger.trace("MetadataProvider " + mp.handle()
                    + " specifies aggregatedBy.  "
                    + "Checking if the described object " + " "
                    + describedObject.handle() + " is in it..");

            if (isInAggregation(describedObject, aggregator)) {
                aggregatedObjectFound = true;
                break;
            }
        }

        if (!aggregatedObjectFound) {
            String handle = "";
            if (o.handle() != null)
                handle = o.handle() + " ";
            throw new ContentModelException("Metadata " + handle
                    + "does not describe any object "
                    + "contained within Aggregator " + aggregator.handle()
                    + ".  The 'aggregatedBy' relationship "
                    + "in MetadataProvider " + mp.handle()
                    + " implies that it must");
        }

    }

    private boolean isInAggregation(NDRObject resource, NDRObject aggregation)
            throws ContentModelException {

        logger.trace("Determining if " + resource.handle()
                + " is aggregated by " + aggregation.handle());

        boolean encountered = false;

        /* NB: assumes handle equality */
        for (Relationship r : resource.getRelationships(MEMBER_OF)) {
            if (r.related().handle().equals(aggregation.handle())) {
                encountered = true;
            }
        }

        if (encountered) {
            logger.trace("Yes, " + resource.handle()
                    + " has been found in aggregator");

        } else {
            logger.trace(resource.handle() + " not found in aggregation");
        }
        return encountered;
    }

    /**
     * Assure that modifying this object does not violate aggregatedBy contract.
     * <p>
     * If a resource (or aggregator, etc) is modified AND it is described by
     * some Metadata objects whose Metadata's Provider specifies an
     * {@linkplain CollectionsModel.Components#AGGREGATED_BY AGGREGATED_BY}
     * contract, then make sure that the resource is still in the required
     * aggregations.
     * </p>
     * 
     * @param resource
     *            Member of an aggregation that might be part an aggregatedBy
     *            contract.
     * @throws ContentModelException
     */
    private void assureDeaggregatedResourceHasNoRequiredMetadata(
            NDRObject resource) throws ContentModelException {

        Set<String> aggregatorHandles = new HashSet<String>();
        for (Relationship memberOf : resource.getRelationships(MEMBER_OF)) {
            aggregatorHandles.add(memberOf.related().handle());
        }

        try {
            ndr.getObject(resource);
        } catch (ExistenceException e) {
            /*
             * If the resource doesn't exist in the NDR, then there can't
             * possibly be any metadata
             */
            return;
        }

        Results<Relationship> metadataForRels =
                ndr.getObjectFinder().getRelatedTo(resource, METADATA_FOR);
        try {
            for (Relationship metadataFor : metadataForRels) {

                NDRObject metadata = ndr.getObject(metadataFor.subject());

                if (!metadata.state().equals(State.Active)) {
                    /* We don't care about non-active metadata */
                    continue;
                }

                NDRObject provider =
                        ndr.getObject(metadata.getRelationship(
                                METADATA_PROVIDED_BY).related());

                Relationship aggregatedBy =
                        provider.getRelationship(Components.AGGREGATED_BY);
                if (aggregatedBy != null) {
                    String aggregator = aggregatedBy.related().handle();
                    if (!aggregatorHandles.contains(aggregator)) {
                        throw new ContentModelException(
                                "By the aggregatedBy contract specified in "
                                        + provider.handle() + ", "
                                        + resource.type() + " "
                                        + resource.handle()
                                        + " must be a member of Aggregator "
                                        + aggregator);
                    }
                }
            }
        } finally {
            metadataForRels.close();
        }
    }
}
