/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.content;

import java.util.Collection;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;

/**
 * Minimal content model applicable to all objects.
 * <p>
 * The <code>hasHandle</code> and <code>objectType</code> properties, as well as
 * the <code>memberOf</code> relationship are common to all objects. This class
 * defines these components and enforces the following behaviour:
 * <ul>
 * <li>All objects must have exactly one handle, which must be unique</li>
 * <li>Objects must have a valid objectType that matches those defined in
 * {@link org.nsdl.repository.model.types.Type}.</li>
 * <li>TODO: For active objects that declare a <code>memberOf</code>
 * relationship to another object, that object must be an active Aggregator.</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class CommonModel implements ContentModel {

    /** Namespace for all components in the common model */
    public static final String NS = Constants.REPOS_NAMESPACE;

    /** Default prefix for all components in common model */
    public static final String PREFIX = "nsdl";

    /**
     * Components that are common to all NDR objects.
     * 
     * @author birkland
     * @since 1.1
     */
    public static interface Components extends Definitions {
        
        public static final Property.Name NSDL_STATE = new Property.Name(NS,
                "state", PREFIX);

        /**
         * Handle of an NDR object.
         * <p>
         * Uniquely identifies NDR objects in a repository. Every NDR object
         * must have exactly one handle.
         * </p>
         */
        public static final Property.Name HAS_HANDLE = new Property.Name(NS,
                "hasHandle", PREFIX);

        /**
         * Type of an NDR object.
         * <p>
         * Every NDR object must have a type that exists amongst the enumeration
         * in {@link org.nsdl.repository.model.types.Type}.
         * </p>
         */
        public static final Property.Name OBJECT_TYPE = new Property.Name(NS,
                "objectType", PREFIX);

        /**
         * Indicates membership in an Aggregation.
         * <p>
         * NDR objects declare that they are a member of an aggregation through
         * this relationship.
         * </p>
         */
        public static final Relationship.Name MEMBER_OF = new Relationship.Name(
                NS, "memberOf", PREFIX);

        /**
         * Simple metadata describing the object
         * <p>
         * Must be unqualified Dublin Core. Provides a quick summary of the
         * object. It is typically auto-generated upon creation unless
         * specified.
         * </p>
         */
        public static final Datastream.Name DC = new Datastream.Name("DC");
    }

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        /* Not needed */
    }

    /** {@inheritDoc} */
    public void checkCompliance(NDRObject o) throws ContentModelException {
        assureExactlyOneHandle(o);
        assureProperType(o);
    }

    private void assureExactlyOneHandle(NDRObject o)
            throws ContentModelException {
        Collection<Property> handles = o.getProperties(Components.HAS_HANDLE);
        if (handles.size() != 1) {
            throw new ContentModelException(handles.size() + " handles found. "
                    + "There should be exactly 1");
        }

        String propertyHandle = handles.iterator().next().value();
        if (!propertyHandle.equals(o.handle())) {
            throw new ContentModelException("Object's handle property ("
                    + propertyHandle + ") does not match its assigned handle "
                    + o.handle());
        }
    }

    private void assureProperType(NDRObject o) throws ContentModelException {
        Collection<Property> objectTypes = o
                .getProperties(Components.OBJECT_TYPE);

        if (objectTypes.size() < 1) {
            throw new ContentModelException(
                    "No objectType property found for this object!");
        }

        /* Make sure that the type is known */
        for (Property p : objectTypes) {
            try {
                Type.getInstance(p.value());
            } catch (IllegalArgumentException e) {
                throw new ContentModelException(e.getMessage());
            }
        }
    }
}
