/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.model.NDRObject;

/**
 * Defines the structure and referential requirements of objects in the
 * repository.
 * <p>
 * A Content Model defines the
 * {@link org.nsdl.repository.model.Property Properties},
 * {@link org.nsdl.repository.model.Relationship Relationships}, and
 * {@link org.nsdl.repository.model.Datastream Datastreams} that may be found in
 * an object, as well as any specific characteristics they may have (e.g. a
 * particular property must be unique, a particular relationship must point to
 * an object that fulfulls certain requirements, etc). This information is
 * primarily manifested in the {@link ContentModel#checkCompliance(NDRObject)}
 * method, whereby an object is determined to comply or not comply with the
 * given model.
 * </p>
 * 
 * <p>
 * Besides content model enforcement, there are a number of conventions for
 * implementing classes that aid in defining the model for developers in code.
 * <ul>
 * <li>
 * <p>
 * Content may contain inner {@link Definitions Definition} interfaces listing
 * names of components in that model.
 * </p>
 * <p>
 * Content models may contain inner interfaces that define component names in
 * that model as static final fields. The convention used here is that content
 * models contain an interface named <code>Components</code>.. If there are
 * groups of these components that have some special meaning or commonality
 * (such as {@link ResourceModel.Identifiers Identifier} properties of
 * resources), these may exist as separately named interfaces. Models that are
 * composed as an overlay of other models may "import" components from the other
 * models by extending their Components interfaces into its own.
 * </p>
 * </li>
 * <li>
 * <p>
 * The Properties and Relationships of unrelated content models are in different
 * namespaces,
 * </p>
 * <p>
 * Namespaces allow orthogonality between content models. In order to overlay
 * independant content models upon repository object, it is important that these
 * models not conflict (i.e. define the same relationship or property, but
 * perhaps have different semantics). If each content model confines any object
 * components it defines to its own separate namespace, then orthogonality can
 * be guaranteed. Indeed, the core object content models will not allow any
 * objects that contain any properties or relationships in their namespace that
 * are not defined by these models. In general, individual content models should
 * offer the same protection for their own namespaces. A cocnvention used here
 * is that content models contain static final fields <code>NS</code> and
 * <code>PREFIX</code> to advertise their namespaces.
 * </p>
 * </li>
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public interface ContentModel {

    /**
     * Determine if an object complies with the content model
     * <p>
     * Performs any necessary checks to determine if an object complies with any
     * rules or constraints present in the model. If the object does
     * <em>not</em> comply with the model, a {@link ContentModelException}
     * will be thrown, which may possibly contain an explanation of the failure.
     * </p>
     * 
     * @param o
     *            NDRObject that is to be checked
     * 
     * @throws ContentModelException
     *             The object cannot be determined to comply with the model
     */
    public void checkCompliance(NDRObject o) throws ContentModelException;

    /**
     * Initialize a content model in a particular repository
     * 
     * @param ndr
     *            The repository in which to enforce the content model
     */
    public void init(NDRAccess ndr);
}
