/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.content;

import java.util.Collection;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.types.Metadata;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.NDRObjectTemplate;

import static org.nsdl.repository.model.NDRObject.State.Active;
import static org.nsdl.repository.model.content.ModelUtil.*;
import static org.nsdl.repository.model.content.Scope.*;

/**
 * Metadata-specific model constraints.
 * <p>
 * Defines the components specific to an Agent, and enforces the following for
 * Metadata objects:
 * <ul>
 * <li>Must contain at least one metadataFor relationship, and each one points
 * to a living resource or aggregator</li>
 * <li>May contain at most one {@link MetadataModel.Components#UNIQUE_ID
 * UNIQUE_ID} property, which must be unique for a given MetadataProvider</li>
 * <li>Must have exactly one
 * {@link MetadataModel.Components#METADATA_PROVIDED_BY METADATA_PROVIDED_BY}
 * relationship to an active Metadataprovider.</li>
 * <li>Must contain at least one {@link MetadataModel.Components#METADATA_FOR
 * METADATA_FOR} relationship to an active object.
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class MetadataModel implements ContentModel {

    private static final Logger logger = Logger.getLogger(MetadataModel.class);

    public static final String NS = CommonModel.NS;
    public static final String PREFIX = CommonModel.PREFIX;

    private NDRAccess ndr;

    /**
     * Components in the core Metadata model.
     * 
     * @author birkland
     * @since 1.1
     */
    public static interface Components extends CommonModel.Components {

        /**
         * Unique identifier.
         * <p>
         * Used to address or differentiate Metadata objects that are provided
         * by the same MetadataProvider. It is optional, and inteded mostly for
         * the convenience of the provider.
         * </p>
         */
        public static final Property.Name UNIQUE_ID = new Property.Name(NS,
                "uniqueID", PREFIX);

        /** Relates Metadata to the object it describes. */
        public static final Relationship.Name METADATA_FOR = new Relationship.Name(
                NS, "metadataFor", PREFIX);

        /**
         * The provider of a Metadata object.
         * <p>
         * Every Metadata object must be provided by exactly only one
         * MetadataProvider. This relationship is used for determining the
         * provenance of every Metadata statement in the repository.
         * </p>
         */
        public static final Relationship.Name METADATA_PROVIDED_BY = new Relationship.Name(
                NS, "metadataProvidedBy", PREFIX);

        /**
         * Indicates that the objects described by this Metadata reference other
         * objects.
         * <p>
         * Objects that reference others will be connected by an arc in the
         * graph through a Metadata object. The Metadata object will have a
         * {@link MetadataModel.Components#METADATA_FOR METADATA_FOR}
         * relationship to the referring object, and a
         * {@linkplain MetadataModel.Components#REFERENCES REFERENCES}
         * relationship to the referred object.
         * </p>
         * 
         * <p>
         * This relationship is essentially a reified
         * <code>dct:references</code> as found in Dublin Core. It is not truly
         * part of the Metadata model proper, so it may find another home.
         * </p>
         */
        /* XXX Probably not good to have this here */
        public static final Relationship.Name REFERENCES = new Relationship.Name(
                NS, "annotates", PREFIX);

        /**
         * Datastream containing nsdl_dc metadata.
         */
        public static final Datastream.Name NSDL_DC = new Datastream.Name(
                "format_nsdl_dc");

        /**
         * Datastream containing oai_dc metadata.
         */
        public static final Datastream.Name FORMAT_OAI_DC = new Datastream.Name(
                "format_oai_dc");
    }

    /* Cache these here since they never change */
    private static final Set<Property.Name> properties = getPropertyNames(
            Components.class, All);
    static {
        properties
                .addAll(getPropertyNames(OAIModel.Item.Components.class, All));
    }

    private static final Set<Property.Name> exclusiveProperties = getPropertyNames(
            Components.class, Exclusive);

    private static final Set<Relationship.Name> exclusiveRelationships = getRelationshipNames(
            Components.class, Exclusive);

    private static final Set<Relationship.Name> relationships = getRelationshipNames(
            Components.class, All);

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        this.ndr = ndr;
    }

    /** {@inheritDoc} */
    public void checkCompliance(NDRObject o) throws ContentModelException {

        logger.trace("Checking base Metadata compliance");

        if (!o.type().equals(Type.Metadata)) {
            checkExclusiveComponents(o);

            if (Type.MetadataProvider.equals(o.type())) {

                assureProviderDependenciesSatisfied(o);
            } else {
                assureDescribedObjectDependenciesSatisfied(o);
            }
            logger.trace("Not Metadata.  Done checking coompliance");
            return;
        }

        checkProtectedNS(o);

        assureMetadataForIsResourceOrAggregator(o);

        assureAtMostOneUniqueId(o);
        assureUniqueUniqueId(o);

        assureExactlyOneActiveMetadataProvider(o);

        logger.trace("Done checking compliance of Metadata");
    }

    private void assureMetadataForIsResourceOrAggregator(NDRObject o)
            throws ContentModelException {
        Collection<Relationship> metadataForRels = o
                .getRelationships(Components.METADATA_FOR);

        if (metadataForRels.size() < 1) {
            throw new ContentModelException("There must be at least one "
                    + "metadataFor relationship present");
        }

        for (Relationship r : metadataForRels) {

            /*
             * Retrieve it from the repository to make sure it actually exists
             * (as opposed to being an uncommitted object, etc)
             */
            NDRObject described = null;
            try {
                described = ndr.getObject(r.related());
            } catch (ExistenceException e) {
                throw new ContentModelException("metadataFor object "
                        + r.related().handle()
                        + " does not exist in the repository");
            }

            if (!(described.type().equals(Type.Resource) || described.type()
                    .equals(Type.Aggregator))) {
                throw new ContentModelException("Metadata must only describe "
                        + "Resources or Aggregators, but " + described.handle()
                        + " is a " + described.type());
            }
        }
    }

    private static void assureAtMostOneUniqueId(NDRObject o)
            throws ContentModelException {
        Collection<Property> itemIds = o.getProperties(Components.UNIQUE_ID);

        if (itemIds.size() > 1) {
            throw new ContentModelException("Cannot have more than one "
                    + "uniqueID property");
        }
    }

    private void assureUniqueUniqueId(NDRObject o) throws ContentModelException {

        /* If we're not active, anything is OK */
        if (!o.state().equals(State.Active)) {
            return;
        }

        /* If we don't have a unique ID, we're OK */
        if (o.getProperty(Metadata.UNIQUE_ID) == null) {
            return;
        }

        /* Find all objects with same uniqueID and metadataProvidedBy */
        NDRObjectTemplate mdTemplate = new NDRObjectTemplate(Type.Metadata);
        mdTemplate.setState(State.Active);
        mdTemplate.addProperties(o.getProperties(Components.UNIQUE_ID));
        mdTemplate.addRelationships(o
                .getRelationships(Components.METADATA_PROVIDED_BY));

        Results<NDRObjectInfo> matches = ndr.getObjectFinder().matchObject(
                mdTemplate);
        try {
            for (NDRObjectInfo match : matches) {
                /*
                 * If we have a match thatisn't this object, then we have a
                 * duplicate uniqueID out there
                 */
                if (!match.handle().equals(o.handle())) {
                    throw new ContentModelException("The given uniqueID "
                            + o.getProperty(Components.UNIQUE_ID).value()
                            + " is already " + " in use by object "
                            + match.handle());
                }
            }
        } finally {
            matches.close();
        }
    }

    private void assureExactlyOneActiveMetadataProvider(NDRObject o)
            throws ContentModelException {
        Collection<Relationship> mpBy = o
                .getRelationships(Components.METADATA_PROVIDED_BY);

        /* Make sure there is exactly one metadataProvidedBy relationship */
        if (mpBy.size() != 1) {
            throw new ContentModelException("There must be exactly one "
                    + "matadataProvidedBy relationsip for this Metadata "
                    + "object, but this one has " + mpBy.size());
        }

        NDRObject provider;
        try {
            provider = ndr.getObject(o.getRelationship(
                    Components.METADATA_PROVIDED_BY).related());

        } catch (ExistenceException e) {
            throw new ContentModelException("MetadataProvider "
                    + o.getRelationships(Components.METADATA_PROVIDED_BY)
                            .iterator().next().related().handle()
                    + " does not exist in the repository");
        }

        /* Make sure that it points to a MetadataProvider */
        if (!provider.type().equals(Type.MetadataProvider)) {
            throw new ContentModelException(
                    "Metadata must be provided by a Metadataprovider.");
        }

        /* Make sure that the MetadataProvider is active if we are */
        if (o.state().equals(Active) && !provider.state().equals(Active)) {
            throw new ContentModelException("MetadataProvider "
                    + provider.handle() + " is not active");
        }
    }

    /**
     * Make sure that a non-active Resource has no active Metadata objects
     * describing it.
     * 
     * @param o
     *            NDRObject that is supposed to be a Resource/Metadata/Agent
     */
    private void assureDescribedObjectDependenciesSatisfied(NDRObject o)
            throws ContentModelException {

        if (!o.state().equals(State.Active)) {
            /*
             * Since onlyactive objects are found, the presence of any results
             * means that there is a content model violation
             */

            Results<Relationship> metadataForRels = ndr.getObjectFinder()
                    .getRelatedTo(o, Components.METADATA_FOR);
            try {
                for (Relationship metadataFor : metadataForRels) {

                    NDRObject metadata = ndr.getObject(metadataFor.subject());

                    if (metadata.state().equals(Active)) {
                        throw new ContentModelException(
                                "Unresolved dependency with describing metadata "
                                        + metadata.handle());
                    }
                }
            } finally {
                metadataForRels.close();
            }
        }
    }

    /**
     * Make sure that a non-active MetadataProvider has no active Metadata
     * provided by it
     * 
     * @param o
     *            NDRObject that is supposed to be a MetadataProvider
     */
    private void assureProviderDependenciesSatisfied(NDRObject o)
            throws ContentModelException {

        if (!o.state().equals(State.Active)) {
            /*
             * Since onlyactive objects are found, the presence of any results
             * means that there is a content model violation
             */

            Results<Relationship> mpByRels = ndr.getObjectFinder()
                    .getRelatedTo(o, Components.METADATA_PROVIDED_BY);
            try {
                for (Relationship mpBy : mpByRels) {
                    NDRObject metdata = ndr.getObject(mpBy.subject());

                    if (metdata.state().equals(Active)) {
                        throw new ContentModelException(
                                "Unresolved dependency with provided Metadata  "
                                        + metdata.handle());
                    }
                }
            } finally {
                mpByRels.close();
            }
        }
    }

    private static void checkProtectedNS(NDRObject o)
            throws ContentModelException {

        checkNSProperties(NS, properties, o);
        checkNSRelationships(NS, relationships, o);
    }

    private static void checkExclusiveComponents(NDRObject o)
            throws ContentModelException {
        checkExclusiveProperties(o, exclusiveProperties);
        checkExclusiveRelationships(o, exclusiveRelationships);

    }
}
