/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.content;

import java.util.Collection;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.NDRObject.State.Active;
import static org.nsdl.repository.model.content.ModelUtil.*;
import static org.nsdl.repository.model.content.Scope.*;

/**
 * MetadataProvider-specific content model.
 * <p>
 * Defines components specific to MetadataProviders, and enforces the following
 * rule:
 * <ul>
 * <li>MetadataProviders must have exactly one
 * {@link MetadataProviderModel.Components#METADTA_PROVIDER_FOR
 * METADATA_PROVIDER_FOR} relationship to an active Agent.</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class MetadataProviderModel implements ContentModel {

    private static final Logger logger = Logger
            .getLogger(MetadataProviderModel.class);

    /** Namespace for all components in the common model */
    public static final String NS = CommonModel.NS;

    /** Default prefix for all components in common model */
    public static final String PREFIX = CommonModel.PREFIX;

    private NDRAccess ndr;

    /**
     * Components MetadataProvider objects have in the base model.
     * 
     * @author birkland
     * @since 1.1
     */
    public static interface Components extends CommonModel.Components {

        /**
         * Relates a MetadataProvider to its owning Agent.
         * <p>
         * Every Metadataprovider is "owned" by exactly one Agent. In that
         * sense, the Agent that "owns" a particular MetadataProvider is
         * responsible for its content. Thus, the provenance of every metadata
         * object can be attributed to exactly one Agent.
         * </p>
         */
        public static final Relationship.Name METADTA_PROVIDER_FOR = new Relationship.Name(
                NS, "metadataProviderFor", PREFIX);

        public static final Datastream.Name SERVICE_DESCRIPTION = new Datastream.Name(
                "serviceDescription");
    }

    /* Cache these here since they never change */
    private static final Set<Property.Name> properties = getPropertyNames(
            Components.class, All);

    /*
     * XXX: Treat OAI properties as its own, since theey are unfortunately in
     * the same namespace.
     */
    static {
        properties.addAll(getPropertyNames(OAIModel.Set.Components.class, All));
    }

    private static final Set<Property.Name> exclusiveProperties = getPropertyNames(
            Components.class, Exclusive);

    private static final Set<Relationship.Name> exclusiveRelationships = getRelationshipNames(
            Components.class, Exclusive);

    private static final Set<Relationship.Name> relationships = getRelationshipNames(
            Components.class, All);

    static {
        relationships.addAll(getRelationshipNames(
                CollectionsModel.Components.class, All));
    }

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        this.ndr = ndr;
    }

    /** {@inheritDoc} */
    public void checkCompliance(NDRObject o) throws ContentModelException {

        logger.trace("Checking base MetadataProvider compliance");

        if (!Type.MetadataProvider.equals(o.type())) {
            checkExclusiveComponents(o);

            if (Type.Agent.equals(o.type())) {
                logger.trace("Checking MetadataProvider dependencies of Agent");
                assureAgentDependenciesSatisfied(o);
            }
            logger.trace("Not a MetadataProvider.  Done checking coompliance");
            return;
        }

        checkProtectedNS(o);

        assureExactlyOneAgent(o);

        logger.trace("Done checking compliance of MetadataProvider");

    }

    private void assureExactlyOneAgent(NDRObject o)
            throws ContentModelException {

        Collection<Relationship> mp4Rels = o
                .getRelationships(Components.METADTA_PROVIDER_FOR);

        if (mp4Rels.size() != 1) {
            throw new ContentModelException(mp4Rels.size()
                    + " metadataProviderFor relationships found, there "
                    + "should be exactly 1");
        }

        NDRObject agent = null;
        try {
            agent = ndr.getObject(mp4Rels.iterator().next().related());
        } catch (ExistenceException e) {
            throw new ContentModelException("Agent "
                    + mp4Rels.iterator().next().related().handle()
                    + " does not exist in the repository");
        }

        if (!Type.Agent.equals(agent.type())) {
            throw new ContentModelException("MetadataProvider must be for "
                    + "an agent.  " + agent.handle() + " is a " + agent.type());
        }

        /* If we are active, the related agent must be active */
        if (Active.equals(o.state()) && !Active.equals(agent.state())) {
            throw new ContentModelException(
                    "The MetadataProvider's agent must be active");
        }
    }

    /**
     * Make sure that a non-active agent has no active Aggregators attached to
     * it
     * 
     * @param o
     *            NDRObject that is supposed to be an agent
     */
    private void assureAgentDependenciesSatisfied(NDRObject o)
            throws ContentModelException {

        if (!o.state().equals(State.Active)) {

            Results<Relationship> mp4Rels = ndr.getObjectFinder().getRelatedTo(
                    o, Components.METADTA_PROVIDER_FOR);

            try {
                for (Relationship mp4 : mp4Rels) {

                    NDRObject provider = ndr.getObject(mp4.subject());

                    if (provider.state().equals(State.Active)) {
                        throw new ContentModelException(
                                "Unresolved dependency with Agent's MetadataProvider "
                                        + mp4.subject().handle());
                    }
                }
            } finally {
                mp4Rels.close();
            }
        }
    }

    private static void checkProtectedNS(NDRObject o)
            throws ContentModelException {

        checkNSProperties(NS, properties, o);
        checkNSRelationships(NS, relationships, o);
    }

    private static void checkExclusiveComponents(NDRObject o)
            throws ContentModelException {
        checkExclusiveProperties(o, exclusiveProperties);
        checkExclusiveRelationships(o, exclusiveRelationships);
    }
}
