/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.content;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import javax.xml.namespace.QName;

import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;

import fedora.common.Constants;

/**
 * Utility class for implementing or interacting with content models.
 * 
 * @author birkland
 * @since 1.1
 */
public final class ModelUtil {

    private static final Set<Property.Name> READ_ONLY_PROPS = new HashSet<Property.Name>();

    static {
        READ_ONLY_PROPS.add(CommonModel.Components.OBJECT_TYPE);
        READ_ONLY_PROPS.add(CommonModel.Components.HAS_HANDLE);
        READ_ONLY_PROPS.add(new Property.Name(Constants.VIEW.uri,
                Constants.VIEW.LAST_MODIFIED_DATE.localName, "fedora-view"));
        READ_ONLY_PROPS.add(new Property.Name(Constants.VIEW.uri,
                "createdDate", "fedora-view"));
    }

    /** No need to instantiate */
    private ModelUtil() {
    }

    /**
     * Check that all properties in a given namespace are valid.
     * <p>
     * Given a set of valid properties in a given namespace, look for properties
     * in an object that are in that namespace, but are <em>not</em> among the
     * list of valid properties.
     * </p>
     * 
     * @param namespace
     *            The namespace to check
     * @param validProperties
     *            All valid properties in that namespace
     * @param o
     *            The object to check
     * @throws ContentModelException
     */
    protected static void checkNSProperties(String namespace,
            Collection<Property.Name> validProperties, NDRObject o)
            throws ContentModelException {

        if (validProperties == null) {
            validProperties = new ArrayList<Property.Name>(0);
        }

        for (Property p : o.getProperties()) {
            if (namespace.equals(p.predicate().getNamespaceURI())) {
                if (!validProperties.contains(p.predicate())) {
                    throw new ContentModelException("Illegal property "
                            + p.predicate().getLocalPart() + " in namespace "
                            + namespace);
                }
            }
        }
    }

    /**
     * Check that all relationships in a given namespace are valid.
     * <p>
     * Given a set of valid relationships in a given namespace, look for
     * relationships in an object that are in that namespace, but are
     * <em>not</em> among the list of valid relationships.
     * </p>
     * 
     * @param namespace
     *            The namespace to check
     * @param validRelationships
     *            All valid relationships in that namespace
     * @param o
     *            The object to check
     * @throws ContentModelException
     */
    protected static void checkNSRelationships(String namespace,
            Collection<Relationship.Name> validRelationships, NDRObject o)
            throws ContentModelException {

        if (validRelationships == null) {
            validRelationships = new ArrayList<Relationship.Name>(0);
        }

        for (Relationship r : o.getRelationships()) {
            if (namespace.equals(r.predicate().getNamespaceURI())) {
                if (!validRelationships.contains(r.predicate())) {
                    throw new ContentModelException("Illegal relationship "
                            + r.predicate().getLocalPart() + " in namespace "
                            + namespace);
                }
            }
        }
    }

    /**
     * Make sure than an object does not have any of the given properties.
     * 
     * @param o
     *            The object to check
     * @param exclusiveProperties
     *            The properties it shouldn't have
     * @throws ContentModelException
     *             Thrown of the object has property it shouldn't
     */
    protected static void checkExclusiveProperties(NDRObject o,
            Collection<Property.Name> exclusiveProperties)
            throws ContentModelException {

        for (Property p : o.getProperties()) {
            if (exclusiveProperties.contains(p.predicate())) {
                throw new ContentModelException("Property " + p.predicate()
                        + "should not be in objects of type" + o.type());
            }
        }
    }

    /**
     * Make sure than an object does not have any of the given relationships.
     * 
     * @param o
     *            The object to check
     * @param exclusiveRelationships
     *            The relationships it shouldn't have
     * @throws ContentModelException
     *             Thrown of the object has a relationship it shouldn't
     */
    protected static void checkExclusiveRelationships(NDRObject o,
            Collection<Relationship.Name> exclusiveRelationships)
            throws ContentModelException {

        for (Relationship r : o.getRelationships()) {
            if (exclusiveRelationships.contains(r.predicate())) {
                throw new ContentModelException("relationship " + r.predicate()
                        + "should not be in objects of type " + o.type());
            }
        }
    }

    /**
     * Get all the properties defined as public final fields.
     * <p>
     * Given a class that may contain a number of public final fields defining
     * components in a model, determine which of them are properties and return
     * them represented as a set.
     * </p>
     * 
     * @param def
     *            Class containing property definitions.
     * @param scope
     *            Desired exclusivity of these properties.
     * @return Set containing all compliant property names.
     */
    public static Set<Property.Name> getPropertyNames(
            Class<? extends Definitions> def, Scope scope) {
        Set<Property.Name> names = new HashSet<Property.Name>();
        for (QName rel : getRels(def, Property.Name.class, scope)) {
            names.add(new Property.Name(rel.getNamespaceURI(), rel
                    .getLocalPart(), rel.getPrefix()));
        }
        return names;
    }

    /**
     * Get all the relationships defined as public final fields.
     * <p>
     * Given a class that may contain a number of public final fields defining
     * components in a model, determine which of them are relationships and
     * return them represented as a set.
     * </p>
     * 
     * @param def
     *            Class containing property definitions.
     * @param scope
     *            Desired exclusivity of these properties.
     * @return Set containing all compliant property names.
     */
    public static Set<Relationship.Name> getRelationshipNames(
            Class<? extends Definitions> def, Scope scope) {
        Set<Relationship.Name> names = new HashSet<Relationship.Name>();
        for (QName rel : getRels(def, Relationship.Name.class, scope)) {
            names.add(new Relationship.Name(rel.getNamespaceURI(), rel
                    .getLocalPart(), rel.getPrefix()));
        }
        return names;
    }

    public static boolean isReadOnlyProperty(Property.Name name) {
        return READ_ONLY_PROPS.contains(name);
    }

    private static Set<QName> getRels(Class<?> cm,
            Class<? extends QName> desiredType, Scope scope) {
        Set<QName> values = new HashSet<QName>();

        for (Field f : cm.getFields()) {

            if (desiredType.isAssignableFrom(f.getType())
                    && Modifier.isPublic(f.getModifiers())) {

                try {
                    switch (scope) {
                    case All:
                        values.add((QName) f.get(cm));
                        break;
                    case Exclusive:
                        if (!f.isAnnotationPresent(Shared.class)
                                && f.getDeclaringClass().equals(cm))
                            values.add((QName) f.get(cm));
                        break;
                    case Shared:
                        if (f.isAnnotationPresent(Shared.class)
                                || !f.getDeclaringClass().equals(cm))
                            values.add((QName) f.get(cm));
                        break;
                    }
                } catch (IllegalAccessException e) {
                    /* Should never happen */
                    throw new RuntimeException("Error accessing model", e);
                } catch (ClassCastException e) {
                    throw new RuntimeException("Annotated field '"
                            + f.getName() + "' must be a QName type", e);
                }
            }
        }

        return values;
    }
}
