/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.content;

import java.util.Arrays;
import java.util.List;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.model.NDRObject;

/**
 * Complete content model as defined by the NSDL.
 * <p>
 * This essentially 'concatenates' all individual common, cor, and specialized e
 * content models so that all models may conveniently be checked in one place.
 * This super-model is composed of:
 * <ul>
 * <li>{@link CommonModel}</li>
 * <li>{@link AgentModel}</li>
 * <li>{@link AggregatorModel}</li>
 * <li>{@link MetadataModel}</li>
 * <li>{@link MetadataProviderModel}</li>
 * <li>{@link ResourceModel}</li>
 * </ul>
 * <ul>
 * <li>{@link AuthorizationModel}</li>
 * <li>{@link CollectionsModel}</li>
 * <li>{@link OAIModel}</li>
 * </ul>
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class NSDLContentModel implements ContentModel {

    private static final Logger logger = Logger
            .getLogger(NSDLContentModel.class);

    private static final List<ContentModel> models = Arrays.asList(
            new CommonModel(), new AgentModel(), new AggregatorModel(),
            new MetadataModel(), new MetadataProviderModel(),
            new ResourceModel(), new AuthorizationModel());

    public void checkCompliance(NDRObject o) throws ContentModelException {
        logger.trace("BEGIN checking NSDL compliance of " + o.handle());

        for (ContentModel model : models) {
            model.checkCompliance(o);
        }

        logger.trace("SUCCESS checking NSDL compliance of " + o.handle());
    }

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        for (ContentModel m : models) {
            m.init(ndr);
        }
    }
}
