/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import java.util.Arrays;
import java.util.Collection;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.NDRObjectTemplate;

import static org.nsdl.repository.model.NDRObject.State.Active;
import static org.nsdl.repository.model.content.AggregatorModel.Components.AGGREGATOR_FOR;
import static org.nsdl.repository.model.content.ModelUtil.*;
import static org.nsdl.repository.model.content.MetadataProviderModel.Components.METADTA_PROVIDER_FOR;
import static org.nsdl.repository.model.content.Scope.Exclusive;

/**
 * Framework for disseminating object metadata via OAI.
 * <p>
 * The OAI model defines a set of properties used to annotate objects so that
 * they may be disseminated via OAI. This is a compound model in that objects
 * relevant to OAI dissemination fall under the {@link OAIModel.Set} or
 * {@link OAIModel.Item} models, which correspond to OAI items and OAI Sets,
 * respectively. Using {@link OAIModel#checkCompliance(NDRObject)} of the
 * top-level {@linkplain OAIModel} will check the compliance with respect to
 * both sub-models, which should be the most common use case.
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class OAIModel implements ContentModel {

    private static final Logger logger = Logger.getLogger(OAIModel.class);

    /**
     * OAI component namespace.
     * <p>
     * For historical reasons, most of the components in the OAI model are
     * actually under the common model namespace.
     * </p>
     */
    public static final String NS = "http://ns.nsdl.org/ndr/oai#";

    /** Default OAI component prefix. */
    public static final String PREFIX = "oai";

    private static ContentModel oaiItemModel = new OAIModel.Item();
    private static ContentModel oaiProviderModel = new OAIModel.Set();

    /**
     * Components that are common all OAIModel objects.
     * 
     * @author birkland
     * @since 1.1
     */
    public static interface Common extends Definitions {

        /**
         * Declares the visibility of an OAI set.
         * <p>
         * If a {@link OAIModel.Set Set} object is
         * {@link OAIModel.Visibility#PUBLIC PUBLIC}, all its items are allowed
         * to be served out via OAI. If
         * {@link OAIModel.Visibility#PRIVATE PRIVATE}, then none are.
         * </p>
         * 
         * @see OAIModel.Visibility
         */
        public static final Property.Name VISIBILITY =
                new Property.Name(NS, "visibility", PREFIX);
    }

    /**
     * Legal values of the {@link OAIModel.Common#VISIBILITY} property.
     * 
     * @author birkland
     * @since 1.1
     */
    public static enum Visibility {

        /**
         * All members of a {@link OAIModel.Set Set} may be served out via OAI.
         */
        PUBLIC("public"),

        /** No members of a {@link OAIModel.Set Set} may be served out via OAI. */
        PRIVATE("private"),

        /** Some members of a {@link OAIModel.Set Set} may be served out via OAI */
        PROTECTED("protected");

        private String stringValue;

        private Visibility(String value) {
            this.stringValue = value;
        }

        public String toString() {
            return stringValue;
        }
    }

    /**
     * Objects that can be served out as OAI Items.
     * <p>
     * As it stands now, <code>Metadata</code> objects may be served out via
     * OAI. That is to say, their Metadata datastream payload is served out in
     * OAI. It is intended that Resources also be served out as a
     * Resource-centric OAI feed, but there is no infrastructure to back that up
     * at this time. For OAI items, the following rules are enforced:
     * <ul>
     * <li> The {@link OAIModel.Item.Components#ITEM_ID ITEM_ID} value, for
     * active objects, must be unique repository-wide</li>
     * </ul>
     * </p>
     * 
     * @author birkland
     * @since 1.1
     */
    public static class Item implements ContentModel, OAIModel.Common {
        private NDRAccess ndr;

        /**
         * Components unique to OAI Items.
         * 
         * @author birkland
         * @since 1.1
         */
        public interface Components extends OAIModel.Common {

            /** OAI Identifier of an object. */
            public static final Property.Name ITEM_ID =
                    new Property.Name(CommonModel.NS, "itemId",
                            CommonModel.PREFIX);
        }

        /** {@inheritDoc} */
        public void checkCompliance(NDRObject o) {
            if (Type.Metadata.equals(o.type())) {
                assureAtMostOneItemId(o);
                assureUniqueItemId(o, ndr);
                checkExclusiveProperties(o, exclusiveGroupingProperties);
            } else {
                checkUnrelatedObjectComponents(o);
            }
        }

        /** {@inheritDoc} */
        public void init(NDRAccess ndr) {
            this.ndr = ndr;
        }
    }

    /**
     * Objects that represent OAI Sets.
     * <p>
     * <code>Metadataproviders</code> and <code>Aggregators</code> may
     * represent sets, for Metadata and Resource-centric OAI disseminations,
     * respectively. This model enforces the following rules for either object:
     * <ul>
     * <li> They may use any {@link OAIModel.Set.Components#SET_NAME SET_NAME}</li>
     * <li> The value of {@link OAIModel.Set.Components#SET_SPEC SET_SPEC} must
     * be unique across agents. That is to say, any number of
     * <code>Metadataproviders</code> or <code>Aggregators</code> owned by
     * the same agent may have the same
     * {@linkplain OAIModel.Set.Components#SET_SPEC SET_SPEC} value, but two
     * such objects owned by different <code>Agents</code> may not.</li>
     * </ul>
     * </p>
     * 
     * @author birkland
     * @since 1.1
     */
    public static class Set implements ContentModel, OAIModel.Common {
        private NDRAccess ndr;

        /**
         * Components unique to OAI Set objects.
         * 
         * @author birkland
         * @since 1.1
         */
        public static interface Components extends OAIModel.Common {

            /**
             * OAI <code>setSpec</code> represented by the object with this.
             * property.
             * <p>
             * Any Resource or Metadata {@link OAIModel.Item Items} that are a
             * member of or provided by the object with this property will
             * appear in an OAI set with given <code>setSpec</code>.
             * </p>
             */
            public static final Property.Name SET_SPEC =
                    new Property.Name(CommonModel.NS, "setSpec",
                            CommonModel.PREFIX);

            /**
             * OAI <code>setName</code> represented by the object with this.
             * property.
             */
            public static final Property.Name SET_NAME =
                    new Property.Name(CommonModel.NS, "setName",
                            CommonModel.PREFIX);
        }

        /** {@inheritDoc} */
        public void checkCompliance(NDRObject o) {
            if (Type.Aggregator.equals(o.type())) {
                assureSetSpecUniquenessForAgent(o, Type.Aggregator,
                        AGGREGATOR_FOR, ndr);
                checkExclusiveProperties(o, exclusiveMetadataProperties);
            } else if (Type.MetadataProvider.equals(o.type())) {
                assureSetSpecUniquenessForAgent(o, Type.MetadataProvider,
                        METADTA_PROVIDER_FOR, ndr);
                checkExclusiveProperties(o, exclusiveMetadataProperties);
            } else {
                checkUnrelatedObjectComponents(o);
            }
        }

        /** {@inheritDoc} */
        public void init(NDRAccess ndr) {
            this.ndr = ndr;
        }
    }

    private static final java.util.Set<Property.Name> exclusiveGroupingProperties =
            getPropertyNames(OAIModel.Set.Components.class, Exclusive);
    private static final java.util.Set<Property.Name> exclusiveMetadataProperties =
            getPropertyNames(OAIModel.Item.Components.class, Exclusive);

    /** {@inheritDoc} */
    public void checkCompliance(NDRObject o) throws ContentModelException {

        logger.trace("Checking OAI Model Compliance..");
        if (Type.Aggregator.equals(o.type())
                || Type.MetadataProvider.equals(o.type())) {

            oaiProviderModel.checkCompliance(o);
        } else if (Type.Metadata.equals(o.type())) {
            oaiItemModel.checkCompliance(o);
        } else {
            checkExclusiveProperties(o, exclusiveMetadataProperties);
            checkExclusiveProperties(o, exclusiveGroupingProperties);
        }
        logger.trace("done checking compliance");
    }

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        oaiItemModel.init(ndr);
        oaiProviderModel.init(ndr);
    }

    private static void checkUnrelatedObjectComponents(NDRObject o)
            throws ContentModelException {
        checkExclusiveProperties(o, exclusiveMetadataProperties);
        checkExclusiveProperties(o, exclusiveGroupingProperties);
    }

    /**
     * Make sure that no other Agent's Aggregator/MP object uses same setSpec
     * <p>
     * Will search over all Agents or MPs with the same setSpec. If there is a
     * match that is active and owned by a different agent, a
     * ContentModelException will be raised.
     * </p>
     * 
     * @param o
     *            object being checked
     * 
     * @throws ContentModelException
     */
    private static void assureSetSpecUniquenessForAgent(NDRObject o,
            Type<? extends NDRObject> type, Relationship.Name agentRel,
            NDRAccess ndr) throws ContentModelException {

        Property setSpec;
        try {
            setSpec = o.getProperty(Set.Components.SET_SPEC);
        } catch (AccessException e) {
            throw new ContentModelException("Too many setSpec values");
        }

        if (setSpec == null) {
            logger.trace("no setSpec defined");
            return;
        }
        NDRObjectTemplate template = new NDRObjectTemplate(type);

        template.addProperty(Set.Components.SET_SPEC, setSpec.value());

        NDRObject thisAgent = null;

        try {
            Relationship rel = o.getRelationship(agentRel);
            if (rel == null) {
                throw new ContentModelException(o.type() + " " + o.handle()
                        + "does not have an agent!");
            }
            thisAgent = ndr.getObject(rel.related());
        } catch (ExistenceException e) {
            throw new ContentModelException("Agent "
                    + o.getRelationships(agentRel).iterator().next().related()
                            .handle() + " does not exist in the repository!", e);
        } catch (AccessException e) {
            throw new ContentModelException(
                    "Too many agents for " + o.handle(), e);
        }

        Results<NDRObjectInfo> matches =
                ndr.getObjectFinder().matchObject(template);

        try {
            for (NDRObjectInfo match : matches) {
                NDRObject matchingObject = ndr.getObject(match);

                /*
                 * No point in checking if we've found ourself! N.B. Assumes
                 * handle equality
                 */
                if (match.handle().equals(o.handle()))
                    continue;

                try {
                    NDRObject matchingObjectAgent =
                            ndr.getObject(matchingObject.getRelationship(
                                    agentRel).related());

                    /* N.B. Assumes handle equality */
                    if (!matchingObjectAgent.handle()
                            .equals(thisAgent.handle())
                            && (matchingObject.state().equals(Active) && o
                                    .state().equals(Active))) {
                        throw new ContentModelException("Another agent ("
                                + matchingObjectAgent.handle() + ") is "
                                + "using the setSpec '" + setSpec.value()
                                + "' in " + matchingObject.type() + " "
                                + matchingObject.handle()
                                + ".  Please use another");

                    }
                } catch (AccessException e) {
                    /* Should never happen */
                    throw new ContentModelException(
                            "Could not check setSpec compliance", e);
                }
            }
        } finally {
            matches.close();
        }
    }

    private static void assureAtMostOneItemId(NDRObject o)
            throws ContentModelException {
        Collection<Property> itemIds = o.getProperties(Item.Components.ITEM_ID);

        if (itemIds.size() > 1) {
            throw new ContentModelException("Cannot have more than one "
                    + "itemId property");
        }
    }

    private static void assureUniqueItemId(NDRObject o, NDRAccess ndr)
            throws ContentModelException {
        Property itemIdProperty = null;

        try {
            itemIdProperty = o.getProperty(Item.Components.ITEM_ID);
        } catch (AccessException e) {
            throw new ContentModelException("too mant itemId properties in "
                    + o.handle());
        }

        if (itemIdProperty != null) {
            NDRObjectTemplate template = new NDRObjectTemplate(Type.Metadata);
            template.addProperties(Arrays.asList(itemIdProperty));

            Results<NDRObjectInfo> matches =
                    ndr.getObjectFinder().matchObject(template);
            try {
                for (NDRObjectInfo match : matches) {
                    /*
                     * If we have a match that *isn't* this object, then we have
                     * a duplicate itemId out there
                     */
                    if (!match.handle().equals(o.handle())) {
                        throw new ContentModelException("The given itemId "
                                + itemIdProperty.value() + " is already "
                                + " in use by object " + match.handle());
                    }
                }
            } finally {
                matches.close();
            }
        }
    }
}
