/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.content;

import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.NDRObjectTemplate;

import static org.nsdl.repository.model.content.ModelUtil.*;
import static org.nsdl.repository.model.content.Scope.*;

/**
 * Resource-specific content model.
 * <p>
 * Defines components that are specific to a <code>Resource</code> and enforces
 * the following rules:
 * <ul>
 * <li>A resource must have exactly one {@link ResourceModel.Identifiers
 * Identifier}</li>
 * <li>A resource's Identifier must be globally unique repository-wide</li>
 * </ul>
 * 
 * @author birkland
 * @since 1.1
 */
public class ResourceModel implements ContentModel {

    public static final String NS = CommonModel.NS;
    public static final String PREFIX = CommonModel.PREFIX;

    private static final Logger logger = Logger.getLogger(ResourceModel.class);

    private NDRAccess ndr;

    /**
     * Components in the core Resource object model.
     * 
     * @author birkland
     * @since 1.1
     */
    public static interface Components extends CommonModel.Components {

        /**
         * URL-based Identifier.
         * 
         * @see ResourceModel.Identifiers#URL
         */
        public static final Property.Name IDENTIFIER_URL = new Property.Name(
                NS, Constants.PROPERTY_HAS_RESOURCE + "URL", PREFIX);

        /**
         * Other identifier.
         * 
         * @see ResourceModel.Identifiers#OTHER
         */
        @Shared
        public static final Property.Name IDENTIFIER_OTHER = new Property.Name(
                NS, Constants.PROPERTY_HAS_RESOURCE + "OTHER", PREFIX);

        /** Datastream that contains resource content */
        public static final Datastream.Name CONTENT = new Datastream.Name(
                "content");

    }

    /**
     * Resource properties that serve as identifiers.
     * 
     * @author birkland
     * @since 1.1
     */
    public static interface Identifiers extends Definitions {

        /**
         * URL-based identifier
         * <p>
         * Identifier that has a URL as its value. This is the most common
         * identifier. Since resource identifiers are expected to be unique
         * repository-wide, the repository may choose to store one canonical
         * version of a number of technically equivalent URLs.
         * </p>
         * 
         * @see org.nsdl.repository.api.request.filters.URINormalizationFilter
         */
        public static final Property.Name URL = Components.IDENTIFIER_URL;

        /**
         * Other identifier.
         * <p>
         * Can be used to identify resources by some other criteria other than a
         * URL.
         * </p>
         */
        @Shared
        public static final Property.Name OTHER = Components.IDENTIFIER_OTHER;
    }

    private static final Set<Property.Name> identifiers = getPropertyNames(
            Identifiers.class, All);

    /* Cache these here since they never change */
    private static final Set<Property.Name> properties = getPropertyNames(
            Components.class, All);

    private static final Set<Property.Name> exclusiveProperties = getPropertyNames(
            Components.class, Exclusive);

    private static final Set<Relationship.Name> exclusiveRelationships = getRelationshipNames(
            Components.class, Exclusive);

    private static final Set<Relationship.Name> relationships = getRelationshipNames(
            Components.class, All);

    /** {@inheritDoc} */
    public void init(NDRAccess ndr) {
        this.ndr = ndr;
    }

    /** {@inheritDoc} */
    public void checkCompliance(NDRObject o) throws ContentModelException {

        logger.trace("Checking base Resource compliance");

        if (!Type.Resource.equals(o.type())) {
            checkExclusiveComponents(o);

            logger.trace("Not a Resource.  Done checking coompliance");
            return;
        }

        checkProtectedNS(o);

        assureExactlyOneIdentifier(o);
        assureUniqueIdentifier(o);

        logger.trace("Done checking Resource compliance");
    }

    private void assureExactlyOneIdentifier(NDRObject o)
            throws ContentModelException {

        HashSet<Property> encounteredIDs = new HashSet<Property>();

        for (Property p : o.getProperties()) {
            if (identifiers.contains(p.predicate())) {
                encounteredIDs.add(p);
            }
        }

        if (encounteredIDs.size() == 0) {
            throw new ContentModelException("No identifiers have been provided");
        } else if (encounteredIDs.size() > 1) {
            throw new ContentModelException("Too many identifiers provided!  "
                    + "A resource must have exactly one");
        }
    }

    private void assureUniqueIdentifier(NDRObject o)
            throws ContentModelException {

        if (!State.Active.equals(o.state())) {
            return;
        }

        Property id = null;
        for (Property p : o.getProperties()) {
            if (identifiers.contains(p.predicate())) {
                id = p;
                break;
            }
        }

        NDRObjectTemplate template = new NDRObjectTemplate(Type.Resource);
        template.setState(State.Active);
        template.addProperty(id.predicate(), id.value());

        /* N.B. Assumes handle equality */

        Results<NDRObjectInfo> matches = ndr.getObjectFinder().matchObject(
                template);

        try {
            for (NDRObjectInfo match : matches) {
                if (!match.handle().equals(o.handle())) {
                    throw new ContentModelException("Resource identifier "
                            + id.predicate() + " '" + id.value()
                            + "' is being used by " + match.handle());
                }
            }
        } finally {
            matches.close();
        }
    }

    private static void checkProtectedNS(NDRObject o)
            throws ContentModelException {

        checkNSProperties(NS, properties, o);
        checkNSRelationships(NS, relationships, o);
    }

    private static void checkExclusiveComponents(NDRObject o)
            throws ContentModelException {
        checkExclusiveProperties(o, exclusiveProperties);
        checkExclusiveRelationships(o, exclusiveRelationships);
    }

}
