/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.NDRObject;

/**
 * Basic implementation component tracker implementation.
 * <p>
 * Primarily useful in mutable objects, this is a utility class that contains
 * sets of object components, allowing mutations (adding, deleting, replacing)
 * to its elements and tracking these changes.
 * </p>
 * 
 * @author birkland
 * @since 1.1
 * 
 * @param <T>
 *            The type of {@link org.nsdl.repository.model.Component} being
 *            tracked.
 */
abstract class BasicComponentTracker<T> implements ComponentTracker<T> {

    /** Raw components that pre-existed at the point of starting tracking */
    private final NDRObject m_baseline;

    /** Raw components that have been added since tracking */
    private final Set<T> m_added = new HashSet<T>();

    /** Raw components that have been removed from the baseline */
    private final Set<T> m_removed = new HashSet<T>();

    /**
     * Raw current versions of components that had existed in the baseline, but
     * have subsequently been changed
     */
    private final Set<T> m_updated = new HashSet<T>();

    /** Old versions of objects in the baseline that have been updated */
    private final Set<T> m_obsolete = new HashSet<T>();

    /** Pool of current components that have been wrapped by an UpdateWrapper */
    private final Set<T> m_wrapped = new HashSet<T>();

    private MutationListener<T> m_listener;

    private WrapperListener<T> m_wrapperListener = new WrapperListenerImpl();

    public BasicComponentTracker(NDRObject initialState,
            Class<T> trackedComponent) {
        m_baseline = initialState;
    }

    public void setListener(MutationListener<T> listener) {
        m_listener = listener;
    }

    public void add(T component) {
        if (!containsIn(m_baseline, component)
                && !containsIn(m_added, component)
                && !containsIn(m_updated, component)) {
            notifyAdd(component);
            m_added.add(component);
        }
    }

    public void remove(T component) {
        if (containsIn(m_added, component)) {
            notifyRemove(component);
            removeFrom(m_added, component);
        } else if (containsIn(m_updated, component)) {
            notifyRemove(component);
            removeFrom(m_updated, component);
        } else if (containsIn(m_removed, component)) {
            /* nothing */
        } else if (containsIn(m_baseline, component)) {
            notifyRemove(component);
            m_removed.add(component);
        } else {
            throw new RuntimeException(
                    "Component not found in added, updated, "
                            + "removed, or baseline queues" + component);
        }
    }

    public Collection<T> get(Component.Name<T> name) {
        Set<T> retrieved = new HashSet<T>();

        retrieved.addAll(retrieveFrom(getCurrent(), name));

        return retrieved;
    }

    public Collection<T> getAdded() {
        List<T> added = new ArrayList<T>();

        for (T a : m_added) {
            added.add(wrap(a));
        }
        return added;
    }

    public Collection<T> getCurrent() {
        Set<T> current = new HashSet<T>();

        /* Any added components are current */
        current.addAll(getAdded());

        /* Any updated components are current */
        current.addAll(getUpdated());

        /* Any baseline components that are NOT obsolete or removed are current */
        for (T b : retrieveAllFrom(m_baseline)) {
            if (!containsIn(m_removed, b) && !containsIn(m_obsolete, b)) {
                current.add(wrap(b));
            }
        }

        return current;
    }

    public Collection<T> getDeleted() {
        List<T> removed = new ArrayList<T>();

        for (T d : m_removed) {
            removed.add(unwrap(d));
        }

        return removed;
    }

    public Collection<T> getUpdated() {
        List<T> updated = new ArrayList<T>();

        for (T u : m_updated) {
            updated.add(wrap(u));
        }

        return updated;
    }

    public boolean hasModified() {
        return !m_updated.isEmpty() || !m_removed.isEmpty()
                || !m_added.isEmpty();
    }

    private T wrap(T component) {
        if (!containsIn(m_wrapped, component)) {
            WrappedComponent<T> u = wrapComponent(component);
            u.setUpdateListener(m_wrapperListener);
            m_wrapped.add(u.getWrappedInstance());
        }

        return retrieveFrom(m_wrapped, component);
    }

    @SuppressWarnings("unchecked")
    private T unwrap(T component) {
        if (containsIn(m_wrapped, component)) {
            return ((WrappedComponent<T>) retrieveFrom(m_wrapped, component))
                    .getRawInstance();
        } else {
            return component;
        }
    }

    interface WrappedComponent<T> {

        public T getWrappedInstance();

        public T getRawInstance();

        public void setUpdateListener(WrapperListener<T> l);
    }

    interface WrapperListener<T> {
        public void registerUpdate(T previous, T current);
    }

    class WrapperListenerImpl implements WrapperListener<T> {
        public void registerUpdate(T previous, T current) {

            if ((containsIn(m_baseline, previous))
                    && !containsIn(m_obsolete, previous)) {
                m_updated.add(current);
                m_obsolete.add(previous);
            }

            /* Update caches in case of hashCode update */
            updateIn(m_added, previous, current);
            updateIn(m_updated, previous, current);

            if (m_listener != null) {
                m_listener.notifyUpdate(current);
            }
        }
    }

    /*
     * If a component changes, it hash code may. If so, update the given
     * container to contain the latest version of the component (removing the
     * old one with the old hash value)
     */
    private void updateIn(Collection<T> container, T previous, T current) {

        T found = searchIn(container, current);
        if (found != null && !containsIn(container, current)) {
            container.remove(found);
            container.add(current);
        }
    }

    abstract boolean containsIn(NDRObject object, T component);

    abstract boolean containsIn(Collection<T> container, T component);

    abstract T searchIn(Collection<T> container, T component);

    abstract boolean removeFrom(Collection<T> container, T component);

    abstract T retrieveFrom(Collection<T> container, T component);

    abstract Collection<T> retrieveFrom(Collection<T> container,
            Component.Name<T> name);

    abstract Collection<T> retrieveAllFrom(NDRObject o);

    abstract WrappedComponent<T> wrapComponent(T component);

    private void notifyAdd(T added) {
        if (m_listener != null) {
            m_listener.notifyAdd(added);
        }
    }

    private void notifyRemove(T removed) {
        if (m_listener != null) {
            m_listener.notifyRemove(removed);
        }
    }
}
