/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.impl;

import java.util.Collection;

/**
 * Tracks changes (adds, deletes) that occur to an
 * {@link org.nsdl.repository.model.NDRObject object's} components.
 * <p>
 * It is sometimes useful to track changes to a mutable object throughout their
 * update life cycle. This class allows access to a summary of changes (i.e.
 * which components have been added, removed, or replaces) as well as the
 * ability to attach a {@link MutationListener listener} to the tracker in order
 * to get realtime callbacks for each individual change event.
 * </p>
 * 
 * @author birkland
 * @since 1.0
 * 
 * @param <T>
 *            The type of {@link org.nsdl.repository.model.Component} being
 *            tracked.
 */
public interface ComponentTracker<T> {

    /**
     * Determine if there have been any changes to the tracked components.
     * 
     * @return true if modified.
     */
    public boolean hasModified();

    /**
     * Get all components that have been added since commit().
     * 
     * @return Set of components added
     */
    public Collection<T> getAdded();

    /**
     * Get all components that were present at the last commit and have
     * subsequently been deleted.
     * 
     * @return Set of all deleted components.
     */
    public Collection<T> getDeleted();

    /**
     * Get all components existed, but have been updated or otherwise modified,
     * since last commit.
     * 
     * @return Set of original components that have been subsequently replaced.
     */
    public Collection<T> getUpdated();
    
    public Collection<T> getCurrent();

    /**
     * Attach a listener that will pass along notifications of each add, delete,
     * or replace operation.
     * 
     * @param listener
     *            MutationListener to be called whenever an event occurs to the
     *            tracked components.
     */
    public void setListener(MutationListener<T> listener);
}
