/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.impl;

import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;

public class DatastreamTracker extends BasicComponentTracker<Datastream> {

    public DatastreamTracker() {
        super(new EmptyObject(), Datastream.class);
    }

    public DatastreamTracker(NDRObject base) {
        super(base, Datastream.class);
    }

    @Override
    boolean containsIn(NDRObject object, Datastream component) {
        return component != null
                && object.getDatastream(component.getName()) != null;
    }

    @Override
    boolean containsIn(Collection<Datastream> container, Datastream component) {
        return searchIn(container, component) != null;
    }

    @Override
    Datastream searchIn(Collection<Datastream> container, Datastream component) {
        if (component != null) {
            for (Datastream d : container) {
                if (component.getName().equals(d.getName())) {
                    return d;
                }
            }
        }

        return null;
    }

    @Override
    boolean removeFrom(Collection<Datastream> container, Datastream component) {

        Datastream toRemove = null;
        if (component != null) {
            for (Datastream d : container) {
                if (component.getName().equals(d.getName())) {
                    toRemove = d;
                    break;
                }
            }
        }

        if (toRemove != null) {
            return container.remove(toRemove);
        } else {
            return false;
        }
    }

    @Override
    Collection<Datastream> retrieveFrom(Collection<Datastream> container,
            Component.Name<Datastream> name) {
        List<Datastream> datastreams = new ArrayList<Datastream>();

        for (Datastream d : container) {
            if (name.equals(d.getName())) {
                datastreams.add(d);
                break;
            }
        }

        return datastreams;
    }

    @Override
    Datastream retrieveFrom(Collection<Datastream> container,
            Datastream component) {
        for (Datastream d : container) {
            if (component.getName().equals(d.getName())) {
                return d;
            }
        }

        return null;
    }

    @Override
    Collection<Datastream> retrieveAllFrom(NDRObject o) {
        return o.getDatastreams();
    }

    @Override
    WrappedComponent<Datastream> wrapComponent(Datastream component) {
        return new TrackedDatastream(component);
    }

    private class TrackedDatastream implements Datastream,
            WrappedComponent<Datastream> {

        private final Datastream m_delegate;
        private final BasicDatastream m_previous;
        public WrapperListener<Datastream> listener;

        public TrackedDatastream(Datastream delegate) {
            m_delegate = delegate;

            m_previous = new BasicDatastream(delegate.getName());
            m_previous.setURL(m_delegate.getURL());
            m_previous.setMimeType(m_delegate.getMimeType());
            m_previous.setLabel(m_delegate.getLabel());
        }

        public void setUpdateListener(WrapperListener<Datastream> l) {
            this.listener = l;
        }

        public Datastream getWrappedInstance() {
            return this;
        }

        public Datastream getRawInstance() {
            return m_delegate;
        }

        public Datastream.Name getName() {
            return m_delegate.getName();
        }

        public String getMimeType() {
            return m_delegate.getMimeType();
        }

        public String getLabel() {
            return m_delegate.getLabel();
        }

        public long getSize() {
            return m_delegate.getSize();
        }

        public URL getURL() {
            return m_delegate.getURL();
        }

        public void setURL(URL redirect) {
            m_previous.setURL(m_delegate.getURL());
            m_delegate.setURL(redirect);
            if (listener != null) {
                listener.registerUpdate(m_previous, this);
            }
        }

        public InputStream getContent() throws AccessException {
            return m_delegate.getContent();
        }

        public InputStream getContent(String version) throws AccessException {
            return m_delegate.getContent(version);
        }

        public void setContent(InputStream content) {
            m_delegate.setContent(content);
            if (listener != null) {
                listener.registerUpdate(m_previous, this);
            }
        }

        public void setMimeType(String mime) {
            m_previous.setMimeType(m_delegate.getMimeType());
            m_delegate.setMimeType(mime);
            if (listener != null) {
                listener.registerUpdate(m_previous, this);
            }
        }

        public boolean isCanonical() {
            return m_delegate.isCanonical();
        }

        public void setLabel(String name) {
            m_previous.setLabel(m_delegate.getLabel());
            m_delegate.setLabel(name);
            if (listener != null) {
                listener.registerUpdate(m_previous, this);
            }
        }

    }
}
