/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.impl;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.net.URL;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.util.StreamUtil;

/**
 * Wrapper that provides mutation functionality to read-only datastreams.
 * 
 * @author birkland
 * @since 1.3
 */
class MutableDatastreamDecorator implements Datastream {
    private static final Logger logger = Logger
            .getLogger(MutableDatastreamDecorator.class);
    private final Datastream m_delegate;

    private byte[] m_content;
    private URL m_url;
    private String m_mime;
    private String m_fileName;

    public MutableDatastreamDecorator(Datastream delegate) {
        m_delegate = delegate;
    }

    public InputStream getContent() throws AccessException {
        if (!isCanonical()) {
            try {
                return getURL().openStream();
            } catch (Exception e) {
                throw new AccessException("Could not read remote content from "
                        + getURL().toString(), e);
            }
        }

        if (m_content == null) {
            pollContent();
        }

        return new ByteArrayInputStream(m_content);
    }

    private void pollContent() {
        m_content = StreamUtil.streamToByteArray(m_delegate.getContent());
    }

    public InputStream getContent(String version) throws AccessException {
        /* Past versions should be static */
        return m_delegate.getContent(version);
    }

    public String getMimeType() {
        if (m_mime == null) {
            m_mime = m_delegate.getMimeType();
        }
        return m_mime;
    }

    public Name getName() {
        return m_delegate.getName();
    }

    public long getSize() {

        if (m_content != null) {
            return m_content.length;
        } else if (m_url != null) {
            /* XXX Expensive way to do this */
            try {
                return StreamUtil.streamToByteArray(m_url.openStream()).length;
            } catch (Exception e) {
                logger.warn("Could not determine size of " + m_url.toString());
                if (logger.isDebugEnabled()) {
                    logger.warn("Could not determine size of "
                            + m_url.toString(), e);
                } else {
                    logger.warn("Could not determine size of "
                            + m_url.toString());
                }
                return 0;
            }
        } else {
            return m_delegate.getSize();
        }
    }

    public URL getURL() {
        if (m_url == null) {
            m_url = m_delegate.getURL();
        }

        return m_url;
    }

    public void setContent(InputStream content) {
        if (isCanonical()) {
            m_content = StreamUtil.streamToByteArray(content);
        } else {
            throw new AccessException(
                    "Cannot change content of non-canonical datastream '"
                            + getName().getID() + "'");
        }
    }

    public void setMimeType(String mime) {
        m_mime = mime;
    }

    public void setURL(URL redirect) {
        if (!isCanonical()) {
            m_url = redirect;
        } else {
            throw new AccessException(
                    "Cannot change URL of canonical datastream '"
                            + getName().getID() + "'");
        }
    }

    public boolean isCanonical() {
        return m_delegate.isCanonical();
    }

    public void setLabel(String name) {
        m_fileName = name;
    }

    public String getLabel() {
        if (m_fileName == null) {
            m_fileName = m_delegate.getLabel();
        }
        return m_fileName;
    }
}
