/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.impl;

import java.util.Collection;
import java.util.HashSet;

import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.util.NDRObjectTemplate;

/**
 * Provides basic boilerplate mutable object functionality
 * <p>
 * Takes care of retrieving, modifying, and tracking changes to an NDR object.
 * Specifically, this class implements all <tt>get</tt>, <tt>add</tt>,
 * <tt>remove</tt>, and <tt>replace</tt> operations on the object components,
 * and has protected methods that allow concrete subclasses to access the
 * ultimate change statistics as well as listeners for individual add, remove,
 * or get, or replace operation
 * </p>
 * 
 * @author birkland
 * 
 */
public abstract class MutableObjectBase implements NDRObject {

    private BasicComponentTracker<Property> properties;

    private BasicComponentTracker<Relationship> relationships;

    private BasicComponentTracker<Datastream> datastreams;

    private Type<? extends NDRObject> type;

    private NDRObject.State state;

    private NDRObject.State originalState;

    private String handle;

    private boolean initialized = false;

    /**
     * Initialize a new object
     * <p>
     * Creates a new new object from scratch and populates it with the basic
     * information provided (e.g. type, handle, pid)
     * </p>
     * 
     * @param type
     *            The new object's type
     * @param handle
     *            The handle of the new object
     */
    protected final void initNew(Type<? extends NDRObject> type, String handle) {
        initialized = true;
        this.properties = new PropertyTracker();
        this.relationships = new RelationshipTracker();
        this.datastreams = new DatastreamTracker();
        this.type = type;
        this.state = NDRObject.State.Inactive;
        setState(NDRObject.State.Active);

        this.handle = handle;
        for (Property p : getDefiningProperties()) {
            properties.add(p);
        }
    }

    /**
     * Initialize from an existing object
     * <p>
     * Initialized this mutable object with the contents of an existing object.
     * For the purposes of tracking changes, the object's initial contents and
     * state are treats as the baseline, and any subsequent changes are tracked.
     * </p>
     * 
     * @param initial
     *            The initial state and contents of the object
     */
    protected final void init(NDRObject initial) {
        if (initial == this) {
            throw new RuntimeException("Cant initialize with itself!");
        }
        this.properties = new PropertyTracker(initial);
        this.relationships = new RelationshipTracker(initial);
        this.datastreams = new DatastreamTracker(MutableDatastreamSource
                .wrap(initial));

        this.type = initial.type();

        this.state = initial.state();
        this.handle = initial.handle();
        initialized = true;
    }

    protected final void resetTrackers() {
        originalState = null;
        NDRObject baseline = new NDRObjectTemplate(type);
        baseline.addProperties(getProperties());
        baseline.addRelationships(getRelationships());
        baseline.addDatastreams(getDatastreams());

        if (state != null) {
            baseline.setState(state);
        }

        baseline.setHandle(handle);
        init(baseline);
    }

    /** Get all the Property modifications */
    protected final ComponentTracker<Property> getPropertyTracker() {
        checkInit();
        return properties;
    }

    /** Get all the Relationship modifications */
    protected final ComponentTracker<Relationship> getRelationshipTracker() {
        checkInit();
        return relationships;
    }

    /** Get all the Datastream modifications */
    protected final ComponentTracker<Datastream> getDatastreamTracker() {
        checkInit();
        return datastreams;
    }

    /** Returns the previous state if it has been changes, null otherwise */
    protected final NDRObject.State getPreviousState() {
        checkInit();
        return originalState;
    }

    ComponentFactory<Property> getPropertyGenerator() {
        return new ComponentFactory<Property>() {
            public Property makeComponent(Component.Name<Property> name) {
                return new Property(MutableObjectBase.this, name, null);
            }
        };
    }

    ComponentFactory<Relationship> getRelationshipGenerator() {
        return new ComponentFactory<Relationship>() {
            public Relationship makeComponent(Component.Name<Relationship> name) {
                return new Relationship(MutableObjectBase.this, name, null);
            }
        };
    }

    ComponentFactory<Datastream> getDatastreamGenerator() {
        return new ComponentFactory<Datastream>() {
            public Datastream makeComponent(Component.Name<Datastream> name) {
                return new BasicDatastream(name);
            }
        };
    }

    public final NDRObject.State state() {
        checkInit();
        return this.state;
    }

    public final void setState(NDRObject.State state) {
        checkInit();

        if (!this.state.equals(state)) {

            if (state == null) {
                throw new NullPointerException(
                        "Attempt to give an object null state!");
            }
            if (getPreviousState() == null) {
                originalState = state();
            }
            this.state = state;
        }
    }

    public final void addDatastreams(Collection<Datastream> d) {
        checkInit();

        for (Datastream ds : d) {
            this.datastreams.add(ds);
        }
    }

    public final void addProperties(Collection<Property> toAdd) {
        checkInit();
        for (Property p : toAdd) {
            this.properties.add(p);
        }
    }

    public final void addProperty(Property.Name predicate, String value) {
        checkInit();
        Property toAdd = getPropertyGenerator().makeComponent(predicate);
        toAdd.setValue(value);
        this.properties.add(toAdd);
    }

    public final void addRelationships(Collection<Relationship> toAdd) {
        checkInit();
        for (Relationship r : toAdd) {
            this.relationships.add(r);
        }
    }

    public final void addRelationship(Relationship.Name predicate,
            NDRObjectInfo object) {
        checkInit();
        Relationship toAdd = getRelationshipGenerator()
                .makeComponent(predicate);
        toAdd.setRelated(object);
        this.relationships.add(toAdd);
    }

    public final void removeDatastreams(Collection<Datastream> toRemove) {
        checkInit();
        for (Datastream d : toRemove) {
            this.datastreams.remove(d);
        }
    }

    public final void removeProperties(Collection<Property> toRemove) {
        checkInit();
        for (Property p : toRemove) {
            this.properties.remove(p);
        }
    }

    public final void removeRelationships(Collection<Relationship> toRemove) {
        checkInit();
        for (Relationship r : toRemove) {
            this.relationships.remove(r);
        }
    }

    public final Datastream getDatastream(Datastream.Name id) {
        checkInit();
        Collection<Datastream> ds = datastreams.get(id);

        int size = ds.size();
        if (size == 1) {
            return ds.iterator().next();
        } else if (size == 0) {
            return null;
        } else {
            throw new AccessException("More than one datastream " + id.getID()
                    + " found in " + handle());
        }
    }

    public final Collection<Datastream> getDatastreams() {
        checkInit();
        return datastreams.getCurrent();
    }

    public final Collection<Property> getProperties() {
        checkInit();
        return properties.getCurrent();
    }

    public final Collection<Property> getProperties(Property.Name predicate) {
        checkInit();
        return properties.get(predicate);
    }

    public final Property getProperty(Property.Name predicate)
            throws AccessException {
        checkInit();
        Collection<Property> props = getProperties(predicate);

        int size = props.size();
        if (size > 1) {
            throw new AccessException("More than one property named "
                    + predicate.toString() + " was found in " + handle());
        } else if (size == 0) {
            return null;
        }

        /* We know at this point that there is exactly one relationship here */
        for (Property r : props) {
            return r;
        }

        return null;
    }

    public final Relationship getRelationship(Relationship.Name predicate)
            throws AccessException {
        checkInit();
        Collection<Relationship> rels = getRelationships(predicate);

        int size = rels.size();
        if (size > 1) {
            throw new AccessException("More than one relationship named "
                    + predicate.toString() + " was found in " + handle());
        } else if (size == 0) {
            return null;
        }

        /* We know at this point that there is exactly one relationship here */
        for (Relationship r : rels) {
            return r;
        }

        return null;
    }

    public final Collection<Relationship> getRelationships() {
        checkInit();
        return relationships.getCurrent();
    }

    public final Collection<Relationship> getRelationships(
            Relationship.Name predicate) {
        checkInit();
        return relationships.get(predicate);
    }

    public final void removeProperty(Property.Name name, String value) {
        checkInit();
        properties.remove(new Property(this, name, value));
    }

    public final void removeRelationship(Relationship.Name name,
            NDRObjectInfo object) {
        checkInit();
        relationships.remove(new Relationship(this, name, object));
    }

    public final void removeDatastream(Datastream.Name name) {
        checkInit();
        datastreams.remove(getDatastream(name));
    }

    public final Datastream addDatastream(Datastream.Name name) {
        checkInit();
        Datastream toAdd = getDatastreamGenerator().makeComponent(name);
        datastreams.add(toAdd);
        return toAdd;
    }

    public final String handle() {
        checkInit();
        return handle;
    }

    public final void setHandle(String handle) {
        checkInit();
        if (handle != null) {
            Property hasHandle = this.getProperty(Components.HAS_HANDLE);

            if (hasHandle != null) {
                hasHandle.setValue(handle);
            } else {
                this.addProperty(Components.HAS_HANDLE, handle);
            }
        } else {
            this.removeProperties(this.getProperties(Components.HAS_HANDLE));
        }

        this.handle = handle;
    }

    public final Type<? extends NDRObject> type() {
        checkInit();
        return type;
    }

    /**
     * Determine if this object has been properly initialized
     * 
     * @throws IllegalStateException
     *             if the object has not been in itialized
     */
    private void checkInit() {
        if (!initialized) {
            throw new IllegalStateException("This object is not initialized. "
                    + "Use init()");
        }
    }

    /** Generate the basic defining object properties (handle type) */
    private Collection<Property> getDefiningProperties() {
        Collection<Property> defining = new HashSet<Property>();

        if (this.type != null) {
            defining.add(new Property(this, Components.OBJECT_TYPE, this.type
                    .toString()));
        }

        if (this.handle != null) {
            defining
                    .add(new Property(this, Components.HAS_HANDLE, this.handle));
        }

        return defining;
    }

}
