/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.impl;

import java.util.Collection;
import java.util.Date;

import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;

/**
 * Forms the basis for wrappers around mutable NDR objects
 * <p>
 * A MutableObjectWrapper simply wraps access to all methods of a MutableObject
 * delegate. Parameters are passed and returned from the delegate unaltered. It
 * is intended to be used as a basis for wrapper classes that need to override
 * just a few methods of {@link NDRObject}
 * </p>
 * 
 * @author birkland
 * 
 */
public abstract class MutableObjectWrapper implements NDRObject {
    public final NDRObject objectDelegate;

    protected boolean isModified = false;

    public MutableObjectWrapper(NDRObject wrappedObject) {
        this.objectDelegate = wrappedObject;
    }

    public Type<? extends NDRObject> type() {
        return objectDelegate.type();
    }

    public String handle() {
        return objectDelegate.handle();
    }

    public Date getCreatedDate() {
        return objectDelegate.getCreatedDate();
    }

    public Date getLastModifiedDate() {
        return objectDelegate.getLastModifiedDate();
    };

    public NDRObject.State state() {
        return objectDelegate.state();
    }

    public Property getProperty(Property.Name predicate) throws AccessException {
        return objectDelegate.getProperty(predicate);
    }

    public Collection<Property> getProperties() {
        return objectDelegate.getProperties();
    }

    public Collection<Property> getProperties(Property.Name predicate) {
        return objectDelegate.getProperties(predicate);
    }

    public Relationship getRelationship(Relationship.Name predicate)
            throws AccessException {
        return objectDelegate.getRelationship(predicate);
    }

    public Collection<Relationship> getRelationships() {
        return objectDelegate.getRelationships();
    }

    public Collection<Relationship> getRelationships(Relationship.Name predicate) {
        return objectDelegate.getRelationships(predicate);
    }

    public Datastream getDatastream(Datastream.Name id) {
        return objectDelegate.getDatastream(id);
    }

    public Collection<Datastream> getDatastreams() {
        return objectDelegate.getDatastreams();
    }

    public void addProperties(Collection<Property> p) {
        isModified = true;
        objectDelegate.addProperties(p);
    }

    public void addProperty(Property.Name predicate, String value) {
        isModified = true;
        objectDelegate.addProperty(predicate, value);
    }

    public void removeProperties(Collection<Property> p) {
        isModified = true;
        objectDelegate.removeProperties(p);
    }

    public void removeProperty(Property.Name name, String value) {
        isModified = true;
        objectDelegate.removeProperty(name, value);
    }

    public void addRelationships(Collection<Relationship> r) {
        isModified = true;
        objectDelegate.addRelationships(r);
    }

    public void addRelationship(Relationship.Name predicate,
            NDRObjectInfo object) {
        isModified = true;
        objectDelegate.addRelationship(predicate, object);
    }

    public final Datastream addDatastream(Datastream.Name name) {
        isModified = true;
        return objectDelegate.addDatastream(name);
    }

    public final void removeDatastream(Datastream.Name name) {
        isModified = true;
        objectDelegate.removeDatastream(name);
    }

    public void removeRelationships(Collection<Relationship> r) {
        isModified = true;
        objectDelegate.removeRelationships(r);
    }

    public void removeRelationship(Relationship.Name name, NDRObjectInfo object) {
        isModified = true;
        objectDelegate.removeRelationship(name, object);
    }

    public void addDatastreams(Collection<Datastream> d) {
        isModified = true;
        objectDelegate.addDatastreams(d);
    }

    public void removeDatastreams(Collection<Datastream> d) {
        isModified = true;
        objectDelegate.removeDatastreams(d);
    }

    public void setState(NDRObject.State s) {
        isModified = true;
        objectDelegate.setState(s);
    }

    public void setHandle(String handle) {
        isModified = true;
        objectDelegate.setHandle(handle);
    }

    public void commit() {
        objectDelegate.commit();
        isModified = false;
    }
}
