/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.impl;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;

public class PropertyTracker extends BasicComponentTracker<Property> {

    public PropertyTracker() {
        super(new EmptyObject(), Property.class);
    }

    public PropertyTracker(NDRObject base) {
        super(base, Property.class);
    }

    @Override
    boolean containsIn(NDRObject object, Property component) {

        if (component != null) {
            return containsIn(object.getProperties(component.predicate()),
                    component);
        }
        return false;
    }

    /*
     * FIXME: Brute force. simple container.contains(component) doesn't
     * necessarily work because properties are mutable, and the HashCodes of two
     * otherwise equal relationships may not be equal. This is a workaround, but
     * the overall cause should be re-evaluated
     */
    @Override
    boolean containsIn(Collection<Property> container, Property component) {
        for (Property p : container) {
            if (p.equals(component)) {
                return true;
            }
        }
        return false;
    }

    @Override
    Property searchIn(Collection<Property> container, Property component) {
        for (Property p : container) {
            if (p.equals(component))
                return p;
        }
        return null;
    }

    @Override
    boolean removeFrom(Collection<Property> container, Property component) {

        Property toRemove = retrieveFrom(container, component);
        if (toRemove != null) {
            container.remove(toRemove);
            return true;
        } else {
            return false;
        }
    }

    @Override
    Collection<Property> retrieveFrom(Collection<Property> container,
            Component.Name<Property> name) {

        Set<Property> properties = new HashSet<Property>();
        for (Property p : container) {
            if (p != null && p.predicate().equals(name)) {
                properties.add(p);
            }
        }

        return properties;
    }

    @Override
    Property retrieveFrom(Collection<Property> container, Property component) {

        for (Property p : container) {
            if (p.equals(component)) {
                return p;
            }
        }

        return null;
    }

    @Override
    Collection<Property> retrieveAllFrom(NDRObject o) {
        return o.getProperties();
    }

    @Override
    WrappedComponent<Property> wrapComponent(final Property component) {

        return new TrackedProperty(component);
    }

    private static class TrackedProperty extends Property implements
            WrappedComponent<Property> {

        private final Property m_delegate;
        public String m_previous;
        public WrapperListener<Property> m_listener;

        public TrackedProperty(Property delegate) {
            super(delegate.subject(), delegate.predicate(), delegate.value());
            m_delegate = delegate;
        }

        public void setUpdateListener(WrapperListener<Property> l) {
            this.m_listener = l;
        }

        @Override
        public String value() {
            return m_delegate.value();
        }

        @Override
        public void setValue(String newValue) {

            if (m_previous == null || !m_previous.equals(newValue)) {

                if (m_listener != null) {
                    m_listener.registerUpdate(new Property(subject(),
                            predicate(), value()), new Property(subject(),
                            predicate(), newValue));
                }

                m_previous = value();
                m_delegate.setValue(newValue);
            }
        }

        @Override
        public int hashCode() {
            return m_delegate.hashCode();
        }

        @Override
        public boolean equals(Object o) {
            return m_delegate.equals(o);
        }

        @Override
        public String toString() {
            return m_delegate.toString();
        }

        public Property getWrappedInstance() {
            return this;
        }

        public Property getRawInstance() {
            return m_delegate;
        }
    }
}
