/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.impl;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.nsdl.repository.model.Component;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Relationship;

public class RelationshipTracker extends BasicComponentTracker<Relationship> {

    public RelationshipTracker() {
        super(new EmptyObject(), Relationship.class);
    }

    public RelationshipTracker(NDRObject base) {
        super(base, Relationship.class);
    }

    @Override
    boolean containsIn(NDRObject object, Relationship component) {

        if (component != null) {
            return containsIn(object.getRelationships(component.predicate()),
                    component);
        }
        return false;
    }

    @Override
    boolean containsIn(Collection<Relationship> container,
            Relationship component) {
        /*
         * FIXME: Brute force. simple container.contains(component) doesn't
         * necessarily work because relationships are mutable, and the HashCodes
         * of two otherwise equal relationships may not be equal. This is a
         * workaround, but the overall cause should be re-evaluated
         */
        for (Relationship r : container) {
            if (r.equals(component))
                return true;
        }
        return false;
    }

    @Override
    Relationship searchIn(Collection<Relationship> container,
            Relationship component) {
        for (Relationship r : container) {
            if (r.equals(component))
                return r;
        }
        return null;
    }

    @Override
    boolean removeFrom(Collection<Relationship> container,
            Relationship component) {

        Relationship toRemove = retrieveFrom(container, component);
        if (toRemove != null) {
            container.remove(toRemove);
            return true;
        } else {
            return false;
        }
    }

    @Override
    Collection<Relationship> retrieveFrom(Collection<Relationship> container,
            Component.Name<Relationship> name) {

        Set<Relationship> rels = new HashSet<Relationship>();
        for (Relationship r : container) {
            if (r.predicate().equals(name)) {
                rels.add(r);
            }
        }

        return rels;
    }

    @Override
    Relationship retrieveFrom(Collection<Relationship> container,
            Relationship component) {
        for (Relationship r : container) {
            if (r.equals(component)) {
                return r;
            }
        }

        return null;
    }

    @Override
    Collection<Relationship> retrieveAllFrom(NDRObject o) {
        return o.getRelationships();
    }

    @Override
    WrappedComponent<Relationship> wrapComponent(Relationship component) {
        return new TrackedRelationship(component);
    }

    private class TrackedRelationship extends Relationship implements
            WrappedComponent<Relationship> {

        private NDRObjectInfo m_previous;
        private final Relationship m_delegate;
        private WrapperListener<Relationship> listener;

        public TrackedRelationship(Relationship delegate) {
            super(delegate.subject(), delegate.predicate(), delegate.related());
            m_delegate = delegate;
            m_previous = delegate.related();
        }

        public void setUpdateListener(WrapperListener<Relationship> l) {
            this.listener = l;
        }

        @Override
        public NDRObjectInfo related() {
            return m_delegate.related();
        }

        @Override
        public void setRelated(NDRObjectInfo relatedTo) {
            if (!m_previous.handle().equals(relatedTo.handle())) {

                if (listener != null) {
                    listener.registerUpdate(new Relationship(subject(),
                            predicate(), related()), new Relationship(
                            subject(), predicate(), relatedTo));
                }

                m_delegate.setRelated(relatedTo);
                m_previous = relatedTo;
            }
        }

        @Override
        public int hashCode() {
            return m_delegate.hashCode();
        }

        @Override
        public boolean equals(Object o) {
            return m_delegate.equals(o);
        }

        public Relationship getWrappedInstance() {
            return this;
        }

        public Relationship getRawInstance() {
            return m_delegate;
        }
    }
}
