/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.types;

import java.util.HashSet;
import java.util.Set;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.util.NDRObjectTemplate;
import org.nsdl.repository.util.StreamUtil;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * User, institution, or other entity in the NCore model.
 * 
 * @author birkland
 * @since 1.2
 */
class GenericAgentImpl extends ConcreteObjectBase implements Agent {

    private final NDRAccess m_ndr;

    private final String DC_NAMESPACE = "http://purl.org/dc/elements/1.1/";
    private final String DC_TITLE = "title";
    private final String EMPTY_CONTENT = "<oai_dc:dc xmlns:oai_dc=\"http://www.openarchives.org/OAI/2.0/oai_dc/\" "
            + "xmlns:dc=\"http://purl.org/dc/elements/1.1/\">\n"
            + "  <dc:type>Agent</dc:type>\n"
            + "  <dc:title>Agent</dc:title>\n"
            + "</oai_dc:dc>";

    public GenericAgentImpl(NDRObject identity, NDRAccess repository)
            throws ExistenceException {
        super(identity, repository);
        m_ndr = repository;
        if (!Type.Agent.equals(type())) {
            throw new ExistenceException(identity.handle() + "is not an Agent");
        }
    }

    public Set<Aggregator> getAggregators() {
        Set<Aggregator> aggregators = new HashSet<Aggregator>();

        NDRObjectTemplate template = new NDRObjectTemplate(Type.Aggregator);
        template.addRelationship(Aggregator.AGGREGATOR_FOR, this);
        Results<NDRObjectInfo> results = m_ndr.getObjectFinder().matchObject(
                template);

        try {
            for (NDRObjectInfo identity : results) {
                aggregators.add(m_ndr.getObject(identity, Type.Aggregator));
            }
        } finally {
            if (results != null) {
                results.close();
            }
        }

        return aggregators;
    }

    public String getTitle() {
        return getTitleElement().getTextContent();
    }

    private Element getTitleElement() {
        Datastream dc = getDatastream(DC);
        if (dc == null) {
            dc = addDatastream(DC);
            dc.setContent(StreamUtil.stringToStream(EMPTY_CONTENT));
        }
        try {
            NodeList titles = NDRXMLutil.getDocument(dc.getContent(), true,
                    false).getElementsByTagNameNS(DC_NAMESPACE, DC_TITLE);

            if (titles.getLength() != 1) {
                throw new RuntimeException(
                        "Wrong number of titles in Agent's DC stream ("
                                + titles.getLength() + ")");
            }

            return (Element) titles.item(0);
        } catch (Exception e) {
            throw new RuntimeException("Could not parse Agent's DC", e);
        }
    }

    public Set<MetadataProvider> getMetadataProviders() {
        Set<MetadataProvider> aggregators = new HashSet<MetadataProvider>();

        NDRObjectTemplate template = new NDRObjectTemplate(
                Type.MetadataProvider);
        template.addRelationship(MetadataProvider.METADTA_PROVIDER_FOR, this);
        Results<NDRObjectInfo> results = m_ndr.getObjectFinder().matchObject(
                template);

        try {
            for (NDRObjectInfo identity : results) {
                aggregators.add(m_ndr
                        .getObject(identity, Type.MetadataProvider));
            }
        } finally {
            if (results != null) {
                results.close();
            }
        }

        return aggregators;
    }

    public Aggregator newAggregator() {
        Aggregator agg = m_ndr.newObject(Type.Aggregator);
        agg.addRelationship(Aggregator.AGGREGATOR_FOR, this);
        return agg;
    }

    public MetadataProvider newMetadataProvider() {
        MetadataProvider mp = m_ndr.newObject(Type.MetadataProvider);
        mp.addRelationship(MetadataProvider.METADTA_PROVIDER_FOR, this);
        return mp;
    }

    public void setTitle(String title) {
        Element titleElement = getTitleElement();
        titleElement.setTextContent(title);

        try {
            getDatastream(DC).setContent(
                    StreamUtil.stringToStream(NDRXMLutil
                            .serializeDOMElement(titleElement
                                    .getOwnerDocument().getDocumentElement())));
        } catch (Exception e) {
            throw new RuntimeException("Could not serialize Agent DC stream", e);
        }
    }
}
