/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.types;

import java.util.HashSet;
import java.util.Set;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;

/**
 * Branded grouping of NCore objects.
 * 
 * @author birkland
 * @since 1.2
 */
class GenericAggregatorImpl extends GroupingBase implements Aggregator {

    private final NDRAccess ndr;

    private final Set<NDRObject> newMembers = new HashSet<NDRObject>();

    public GenericAggregatorImpl(NDRObject delegate, NDRAccess repository) {
        super(delegate, repository);

        if (!Type.Aggregator.equals(type())) {
            throw new ExistenceException(delegate.handle()
                    + " is not an Aggregator");
        }

        this.ndr = repository;
    }

    /**
     * {@linkplain org.nsdl.repository.model.content.AggregatorModel.Components#ASSOCIATED_WITH
     * Associate} a resource with the given Aggregator.
     * 
     * If there is a Resource already associated with this Aggregator, it will
     * be unassociated and replaced with the given Resource.
     * 
     * @param resource
     *            Resource to associate.
     */
    public void setAssociatedResource(Resource resource) {
        removeRelationships(getRelationships(ASSOCIATED_WITH));

        if (resource != null) {
            addRelationship(ASSOCIATED_WITH, resource);
        }
    }

    /**
     * Get the resource associated with this Aggregator, if present.
     * 
     * @return Associated resource, null if there are none.
     */
    public Resource getAssociatedResource() {
        Relationship associated = getRelationship(ASSOCIATED_WITH);
        if (associated != null) {
            return ndr.getObject(associated.related(), Type.Resource);
        } else {
            return null;
        }
    }

    /**
     * Add an object as a member of this Aggregator.
     * 
     * Adding a member through this method will not be reflected in the
     * repository until {@link #commit()} is called.
     * 
     * @param toAdd
     *            Identity of the object to add.
     */
    public final void addMember(NDRObjectInfo toAdd) {
        NDRObject member = ndr.getObject(toAdd);
        member.addRelationship(membershipRel(), this);
        synchronized (this) {
            newMembers.add(member);
        }
    }

    /** {@inheritDoc} */
    public <T extends NDRObject> Results<T> getMembers(final Type<T> type) {
        return super.getMembers(type);
    }

    /** {@inheritDoc} */
    public <T extends NDRObject> long countMembers(final Type<T> type) {
        return super.countMembers(type);
    }

    public final void commit() {
        synchronized (this) {
            for (NDRObject member : newMembers) {
                member.commit();
            }
            newMembers.clear();
        }

        super.commit();
    }

    @Override
    Relationship.Name agentRel() {
        return AGGREGATOR_FOR;
    }

    @Override
    Relationship.Name membershipRel() {
        return MEMBER_OF;
    }
}
