/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.types;

/*
 * -----------------------------------------------------------------------------
 * 
 * <p><b>License and Copyright: </b>The contents of this file are subject to
 * the Educational Community License (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License
 * at <a href="http://www.opensource.org/licenses/ecl1.txt">
 * http://www.opensource.org/licenses/ecl1.txt.</a></p>
 * 
 * <p>Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License.</p>
 * 
 * <p>The entire file consists of original code. Copyright &copy; 2004-2007
 * Cornell University. All rights reserved.</p>
 * 
 * -----------------------------------------------------------------------------
 */

import java.util.Iterator;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.content.MetadataModel;

/**
 * Branded source of {@linkplain Metadata} in the NCore model.
 * 
 * @author birkland
 * @since 1.2
 */
class GenericMetadataProviderImpl extends GroupingBase implements
        MetadataProvider {

    private final NDRAccess ndr;

    public GenericMetadataProviderImpl(NDRObject delegate, NDRAccess repository)
            throws ExistenceException {
        super(delegate, repository);

        if (!Type.MetadataProvider.equals(type())) {
            throw new ExistenceException(delegate.handle()
                    + " is not a MetadataProvider");
        }

        this.ndr = repository;
    }

    /**
     * Create new Metadata that describes a given object.
     * 
     * Will pre-populate the Metadata object with the correct relationships to
     * this provider and the described object (Resource or Aggregator).
     * 
     * The new Metadata object have an empty payload (i.e. it will have no
     * default DublinCore content) .
     * 
     * @param describedObject
     *            Identity of the object that will be described by the given
     *            Metadata object.
     * 
     * @return Minimally populated Metadata object.
     */
    public Metadata newMetadata(NDRObjectInfo describedObject) {
        Metadata metadata = ndr.newObject(Type.Metadata);
        metadata.addRelationship(Metadata.METADATA_PROVIDED_BY, this);
        metadata.addRelationship(Metadata.METADATA_FOR, describedObject);

        return metadata;
    }

    public Results<Metadata> getMembers() {

        /* We know MetadataProviders only have Metadata members */
        final Results<? extends NDRObject> objectResults = super
                .getMembers(Type.Metadata);

        return new Results<Metadata>() {

            public Iterator<Metadata> iterator() {
                return new Iterator<Metadata>() {

                    final Iterator<? extends NDRObject> iteratorDelegate = objectResults
                            .iterator();

                    public Metadata next() {
                        return (ndr.getObject(iteratorDelegate.next(),
                                Type.Metadata));
                    }

                    public boolean hasNext() {
                        return iteratorDelegate.hasNext();
                    }

                    public void remove() {
                        iteratorDelegate.remove();
                    }
                };
            }

            public void close() {
                objectResults.close();
            }

        };
    }

    public Grouping.Branding getBranding() {
        throw new UnsupportedOperationException();
    }

    @Override
    Relationship.Name agentRel() {
        return METADTA_PROVIDER_FOR;
    }

    @Override
    Relationship.Name membershipRel() {
        return MetadataModel.Components.METADATA_PROVIDED_BY;
    }

}
