/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.types;

import java.io.InputStream;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.MimeTypeFinder;

/**
 * Web resource in the NCore model.
 * 
 * @author birkland
 * @since 1.2
 */
class GenericResourceImpl extends ConcreteObjectBase implements Resource {

    private final NDRAccess ndr;

    public GenericResourceImpl(NDRObject identity, NDRAccess repository)
            throws ExistenceException {
        super(identity, repository);
        if (!Type.Resource.equals(type())) {
            throw new ExistenceException(identity.handle()
                    + " is not a Resource");
        }

        this.ndr = repository;
    }

    /**
     * Get all Metadata that describes this Resource.
     * 
     * @return Set of all active describing Metadata.
     */
    public Set<Metadata> getMetadata() {
        Set<Metadata> metadata = new HashSet<Metadata>();

        TypeFactory factory = new TypeFactory(ndr);

        Results<Relationship> metadataRels = ndr.getObjectFinder()
                .getRelatedTo(this, Metadata.METADATA_FOR);

        try {
            for (Relationship r : metadataRels) {
                metadata.add(factory.instantiate(ndr.getObject(r.subject()),
                        Type.Metadata));
            }
        } finally {
            metadataRels.close();
        }

        return metadata;
    }

    public Set<Metadata> getMetadata(NDRObjectInfo provider) {
        return null;
    }

    public void setContent(InputStream data, String... mime_or_label) {
        Datastream content = getDatastream(CONTENT);
        if (content == null) {
            content = addDatastream(CONTENT);
        }

        content.setContent(data);

        if (mime_or_label != null) {
            for (String mol : mime_or_label) {
                if (MimeTypeFinder.isMimeType(mol)) {
                    content.setMimeType(mol);
                } else {
                    content.setLabel(mol);
                }
            }
        }
    }

    public void setContentURL(URL url, String... mime_or_label) {
        Datastream content = getDatastream(CONTENT);
        if (content == null) {
            content = addDatastream(CONTENT);
        }

        if (mime_or_label != null) {
            for (String mol : mime_or_label) {
                if (MimeTypeFinder.isMimeType(mol)) {
                    content.setMimeType(mol);
                } else {
                    content.setLabel(mol);
                }
            }
        }
        content.setURL(url);
    }

    public String getMimeType() {
        Datastream content = getDatastream(CONTENT);

        if (content != null) {
            return content.getMimeType();
        } else {
            return null;
        }
    }

    /**
     * Get the current resource content.
     * 
     * @return Content as a binary stream.
     */
    public InputStream getContent() {
        Datastream content = getDatastream(CONTENT);

        if (content != null) {
            return content.getContent();
        }

        return null;
    }

    /**
     * Get a URL that resolves to the resource's content.
     * 
     * @return Most specific URL that resolves to resource content.
     */
    public URL getContentURL() {
        Datastream content = getDatastream(CONTENT);

        if (content != null) {
            return content.getURL();
        }

        return null;
    }

    public Set<Aggregator> getAssociated() {
        throw new UnsupportedOperationException();
    }
}
