/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.types;

import java.net.URL;

import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.NDRObject;

public interface Grouping extends ConcreteObject {

    /**
     * Get the Agent that owns the given grouping.
     * 
     * Every Aggregator or MetadataProvider is "owned" by exactly one Agent. In
     * that sense, the Agent that "owns" a particular
     * Aggregator/Metadataprovider is responsible for its content. Thus, the
     * provenance of every membership in a given Aggregator/MetadataProvider can
     * be attributed to exactly one Agent. This method returns that Agent.
     * 
     * @return The owning Agent.
     */
    public Agent getAgent();

    /**
     * Get all members of the current object.
     * 
     * @return Results containing all members.
     */
    public Results<? extends NDRObject> getMembers();

    /**
     * Get all members that are of the given type.
     * 
     * This will return only those members that are of the supplied type. If the
     * given type is null, then all members will be returned, which is the same
     * effect as calling {@link #getMembers()}
     * 
     * @param <T>
     *            Concrete member type.
     * @param type
     *            Type of member to be retrieved.
     */
    public <T extends NDRObject> Results<T> getMembers(final Type<T> type);
    /**
     * Count all active member objects of the given type.
     * 
     * If null, the behaviour of this method is identical to
     * {@link #countMembers()}, counting every member, regardless of type.
     * 
     * @param <T>
     *            Concrete member type.
     * @param type
     *            Type of object to consider in count.
     * @return Number of active objects that are of the given type.
     */
    <T extends NDRObject> long countMembers(Type<T> type);

    /**
     * Count all the active member objects.
     * 
     * @return Number of members.
     */
    long countMembers();

    /**
     * Return the current branding.
     * 
     * @return branding.
     */
    public Branding getBranding();

    interface Branding {
        /**
         * Use the image at the given URL as the brannd image.
         * 
         * @param location
         *            location of the brand image as a web resource.
         * @param brandTitle Title of the brand image.
         * @param width Requested display width of image (px).
         * @param height Requested display height of image (px).
         */
        public void setBrandImage(URL location, String brandTitle, int width, int height);

        /**
         * Get a URL that resolves to the brand image.
         * 
         * @return Brand image URL.
         */
        public URL getBrandImageURL();
        
        public int getBrandImageHeight();
        
        public int getBrandImageWidth();
        
        public String getBrandImageTitle();

        /**
         * Get the brand title.
         * 
         * @return Brand title.
         */
        public String getTitle();

        /**
         * Set the brand title.
         * 
         * @param title
         *            Brand title.
         */
        public void setTitle(String title);

        /**
         * Set the description of his branded group of objects.
         * 
         * @param desc
         *            String describing the branded group.
         */
        public void setDescription(String desc);

        /**
         * Get the description of his branded group of objects.
         * 
         * @return String describing the branded group.
         */
        public String getDescription();
    }
}
