/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.types;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;

import javax.xml.namespace.QName;

import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.NDRObjectTemplate;
import org.nsdl.repository.util.ResultsCombiner;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import static org.nsdl.repository.model.content.AggregatorModel.Components.SERVICE_DESCRIPTION;

abstract class GroupingBase extends ConcreteObjectBase implements Grouping {

    private final NDRAccess ndr;

    GroupingBase(NDRObject o, NDRAccess ndr) {
        super(o, ndr);
        this.ndr = ndr;
    }

    /**
     * Get the Agent that owns the given grouping.
     * 
     * Every Aggregator or MetadataProvider is "owned" by exactly one Agent. In
     * that sense, the Agent that "owns" a particular
     * Aggregator/Metadataprovider is responsible for its content. Thus, the
     * provenance of every membership in a given Aggregator/MetadataProvider can
     * be attributed to exactly one Agent. This method returns that Agent.
     * 
     * @return The owning Agent.
     */
    public Agent getAgent() {
        return new GenericAgentImpl(ndr.getObject(getRelationship(agentRel())
                .related()), ndr);
    }

    /**
     * Get all members of the current object.
     * 
     * @return Results containing all members.
     */
    public Results<? extends NDRObject> getMembers() {

        ArrayList<Results<? extends NDRObject>> results = new ArrayList<Results<? extends NDRObject>>();
        for (Type<? extends NDRObject> t : Type.values()) {
            results.add(getMembers(t));
        }
        return new ResultsCombiner<NDRObject>(results);
    }

    public <T extends NDRObject> Results<T> getMembers(final Type<T> type) {

        final NDRObjectTemplate template = new NDRObjectTemplate(type);
        template.setState(State.Active);
        template.addRelationship(membershipRel(), this);

        return new Results<T>() {

            final Results<NDRObjectInfo> resultsDelegate = ndr
                    .getObjectFinder().matchObject(template);

            public Iterator<T> iterator() {
                return new Iterator<T>() {

                    final Iterator<NDRObjectInfo> iteratorDelegate = resultsDelegate
                            .iterator();

                    public T next() {
                        return ndr.getObject(iteratorDelegate.next(), type);
                    }

                    public boolean hasNext() {
                        return iteratorDelegate.hasNext();
                    }

                    public void remove() {
                        iteratorDelegate.remove();
                    }
                };
            }

            public void close() {
                resultsDelegate.close();
            }

        };
    }

    /**
     * Count all active member objects of the given type.
     * 
     * If null, the behaviour of this method is identical to
     * {@link #countMembers()}, counting every member, regardless of type.
     * 
     * @param <T>
     *            Concrete member type.
     * @param type
     *            Type of object to consider in count.
     * @return Number of active objects that are of the given type.
     */
    public <T extends NDRObject> long countMembers(Type<T> type) {
        long count = 0;
        Results<? extends NDRObject> members = getMembers();
        try {
            for (Iterator<? extends NDRObject> i = members.iterator(); i
                    .hasNext(); i.next()) {
                count++;
            }
        } finally {
            members.close();
        }

        return count;
    }

    /**
     * Count all the active member objects.
     * 
     * @return Number of members.
     */
    public long countMembers() {
        return countMembers((Type<? extends NDRObject>) null);
    }

    /**
     * Return the current branding.
     * 
     * @return branding.
     */
    public Grouping.Branding getBranding() {
        Datastream serviceDescription = getDatastream(SERVICE_DESCRIPTION);
        if (serviceDescription == null) {
            serviceDescription = this.addDatastream(SERVICE_DESCRIPTION);
        }

        return new ServiceDescriptionBranding(serviceDescription);
    }

    private class ServiceDescriptionBranding implements Grouping.Branding {

        private final Datastream m_source;
        private final Document m_serviceDescription;
        private final QName TITLE = new QName(
                "http://purl.org/dc/elements/1.1/", "title", "dc");
        private final QName DESCRIPTION = new QName(
                "http://purl.org/dc/elements/1.1/", "description", "dc");
        private final QName IMAGE_URL = new QName(null, "brandURL");
        private final QName IMAGE_HEIGHT = new QName(null, "height");
        private final QName IMAGE_WIDTH = new QName(null, "width");
        private final QName IMAGE_TITLE = new QName(null, "title");

        private static final String EMPTY_SD_CONTENT = "<serviceDescription xmlns:dc=\"http://purl.org/dc/elements/1.1/\" >\n"
                + "  <dc:title />\n"
                + "  <dc:description />\n"
                + "  <image>\n"
                + "    <brandURL />\n"
                + "    <title />\n"
                + "    <width />\n"
                + "    <height />\n" + "  </image>\n" + "</serviceDescription>";

        public ServiceDescriptionBranding(Datastream src) {
            m_source = src;

            try {

                m_serviceDescription = NDRXMLutil.getDocument(EMPTY_SD_CONTENT,
                        true, false);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }

            InputStream content = src.getContent();

            /* If we have content, only copy over what we know/care about */
            if (content != null) {
                try {
                    Document original = NDRXMLutil.getDocument(
                            src.getContent(), true, false);

                    copyIfNotNull(TITLE, original);
                    copyIfNotNull(DESCRIPTION, original);
                    copyIfNotNull(IMAGE_TITLE, original);
                    copyIfNotNull(IMAGE_HEIGHT, original);
                    copyIfNotNull(IMAGE_WIDTH, original);
                    copyIfNotNull(IMAGE_URL, original);
                } catch (Exception e) {
                    throw new RuntimeException(
                            "Could not read ServiceDescription from "
                                    + handle());
                }
            }

            sync();
        }

        public void setBrandImage(URL location, String imageTitle, int width,
                int height) {

            getElementByName(IMAGE_URL).setTextContent(location.toString());
            getElementByName(IMAGE_TITLE).setTextContent(imageTitle);
            getElementByName(IMAGE_WIDTH).setTextContent(
                    Integer.toString(width));
            getElementByName(IMAGE_HEIGHT).setTextContent(
                    Integer.toString(height));

            sync();
        }

        public URL getBrandImageURL() {

            String val = getElementByName(IMAGE_URL).getTextContent();

            if ("".equals(val)) {
                return null;
            } else {
                try {
                    return new URL(val);
                } catch (Exception e) {
                    throw new RuntimeException(e);
                }
            }
        }

        public int getBrandImageHeight() {
            String val = getElementByName(IMAGE_HEIGHT).getTextContent();
            if ("".equals(val)) {
                return 0;
            } else {
                return new Integer(val);
            }
        }

        public int getBrandImageWidth() {
            String val = getElementByName(IMAGE_WIDTH).getTextContent();
            if ("".equals(val)) {
                return 0;
            } else {
                return new Integer(val);
            }
        }

        public String getBrandImageTitle() {
            return getElementByName(IMAGE_TITLE).getTextContent();
        }

        public String getTitle() {
            return getElementByName(TITLE).getTextContent();
        }

        public void setTitle(String title) {
            getElementByName(TITLE).setTextContent(title);
            sync();
        }

        public void setDescription(String desc) {
            getElementByName(DESCRIPTION).setTextContent(desc);
            sync();
        }

        public String getDescription() {
            return getElementByName(DESCRIPTION).getTextContent();
        }

        private Element getElementByName(QName name) {
            return getElementByName(name, m_serviceDescription);
        }

        private Element getElementByName(QName name, Document source) {
            NodeList elements = source.getElementsByTagNameNS(name
                    .getNamespaceURI(), name.getLocalPart());

            if (elements.getLength() == 1) {
                return (Element) elements.item(0);
            } else {
                return null;
            }
        }

        private void copyIfNotNull(QName name, Document source) {
            Element element = getElementByName(name, source);

            if (element != null) {
                String value = element.getTextContent();
                getElementByName(name).setTextContent(value);
            }
        }

        private void sync() {

            try {
                m_source.setContent(new ByteArrayInputStream(NDRXMLutil
                        .serializeDOMdoc(m_serviceDescription)
                        .getBytes("UTF-8")));
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    abstract Relationship.Name membershipRel();

    abstract Relationship.Name agentRel();

}
