/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.types;

import java.io.InputStream;
import java.net.URL;
import java.util.Set;

import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.content.ResourceModel;

/**
 * Web resource in the NCore model.
 * 
 * @author birkland
 * @since 1.2
 */
public interface Resource extends ConcreteObject, ResourceModel.Components {

    /**
     * Get all aggregations that are associated with this resource
     * 
     * @return Set containing all associated Aggregators. Empry if there are
     *         none.
     */
    public Set<Aggregator> getAssociated();

    /**
     * Get all Metadata that describes this Resource.
     * 
     * @return Set of all active describing Metadata.
     */
    public Set<Metadata> getMetadata();

    /**
     * Get all Metadata that describes this Resource sourced from a particular
     * MetadataProvider
     * 
     * @param provider
     *            Identity of the MetadataProvider from which desired Metadata
     *            should be from.
     * @return Set of all compliant metadata.
     */
    public Set<Metadata> getMetadata(NDRObjectInfo provider);

    /**
     * Set the content of this Resource as a stream of provided data.
     * 
     * @param data
     *            Content to contain within the Resource.
     * @param mime_and_or_label
     *            Strings representing mimetype and label (desired file name).
     *            Neither are required, and order is arbitrary.
     */
    public void setContent(InputStream data, String... mime_and_or_label);

    /**
     * Set this resource's content as an external URL.
     * 
     * @param url
     * @param mime_and_or_label
     *            Strings representing mimetype and label (desired file name).
     *            Neither are required, and order is arbitrary.
     */
    public void setContentURL(URL url, String... mime_and_or_label);

    /**
     * Get the current resource content.
     * 
     * @return Content as a binary stream.
     */
    public InputStream getContent();

    /** Get the MIME content type of the resource content. */
    public String getMimeType();

    /**
     * Get a URL that resolves to the resource's content.
     * 
     * @return Most specific URL that resolves to resource content.
     */
    public URL getContentURL();
}
