/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.types;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import org.nsdl.repository.model.NDRObject;

/**
 * Core object types in the NSDL model.
 * 
 * @author birkland
 * @since 1.1
 */
public class Type<T extends NDRObject> {

    public static final Type<? extends Agent> Agent = new Type<GenericAgentImpl>(
            GenericAgentImpl.class, "Agent");
    public static final Type<? extends Aggregator> Aggregator = new Type<GenericAggregatorImpl>(
            GenericAggregatorImpl.class, "Aggregator");
    public static final Type<? extends Metadata> Metadata = new Type<GenericMetadataImpl>(
            GenericMetadataImpl.class, "Metadata");
    public static final Type<? extends MetadataProvider> MetadataProvider = new Type<GenericMetadataProviderImpl>(
            GenericMetadataProviderImpl.class, "MetadataProvider");
    public static final Type<? extends Resource> Resource = new Type<GenericResourceImpl>(
            GenericResourceImpl.class, "Resource");

    private static final HashMap<String, Type<NDRObject>> types = buildTypeSet();

    @SuppressWarnings("unchecked")
    private static HashMap<String, Type<NDRObject>> buildTypeSet() {
        HashMap<String, Type<NDRObject>> t = new HashMap<String, Type<NDRObject>>();

        for (Field f : Type.class.getFields()) {
            if (Type.class.isAssignableFrom(f.getType())) {
                try {
                    Type value = (Type) f.get(Type.class);
                    t.put(value.toString(), value);
                } catch (Exception e) {
                    /* Should never happen */
                    throw new RuntimeException(e);
                }
            }
        }

        return t;
    }

    private final Class<T> m_concreteClass;
    private final String m_name;
    private final int m_hashCode;

    private Type(Class<T> classImpl, String name) {
        m_concreteClass = classImpl;
        m_name = name;
        m_hashCode = m_concreteClass.hashCode();
    }

    public static final Type<? extends NDRObject> getInstance(String name) {

        if (name != null) {
            for (String typeName : types.keySet()) {
                if (typeName.toLowerCase().equals(name.toLowerCase())) {
                    return types.get(typeName);
                }
            }
        }

        throw new UnknownTypeException("'" + name
                + "' is not a known object type");
    }

    public static class UnknownTypeException extends IllegalArgumentException {
        static final long serialVersionUID = 1;

        public UnknownTypeException(String message) {
            super(message);
        }
    }

    Class<T> getConcreteClass() {
        return m_concreteClass;
    }

    public String toString() {
        return m_name;
    }

    public static Set<Type<NDRObject>> values() {
        return new HashSet<Type<NDRObject>>(types.values());
    }

    @SuppressWarnings("unchecked")
    public boolean equals(Object o) {
        if (o == null || !(o instanceof Type))
            return false;
        return (((Type) o).m_concreteClass == m_concreteClass);

    }

    public int hashCode() {
        return m_hashCode;
    }
}
