/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.types;

import java.lang.reflect.Constructor;

import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.types.Type;

/**
 * Creates instances of specific NDRObject types (Aggregator, Metadata, etc).
 * 
 * This class is primarily useful in implementing
 * {@link org.nsdl.repository.access.NDRAccess NDRAccess} instances.
 * 
 * @author birkland
 * @since 1.2
 */
public class TypeFactory {

    private final NDRAccess ndr;

    /**
     * Instantiate a factory that builds objects in a specific repository.
     * 
     * @param ndr
     *            Repository from which object instances will be built.
     */
    public TypeFactory(NDRAccess ndr) {
        this.ndr = ndr;
    }

    /**
     * Given a generic NDRObject, instantiate it as a specific concrete type.
     * 
     * @param <T>
     *            Concrete object type.
     * @param object
     *            Generic NDR object instance to turn into a concrete type.
     * @param type
     *            Type of object to create.
     * @return Instance of concrete NDRObject type.
     * @throws AccessException
     *             if the object cannot be instantiated as the desired type.
     */
    public <T extends NDRObject> T instantiate(NDRObject object, Type<T> type)
            throws AccessException {
        if (type.getConcreteClass().isAssignableFrom(object.getClass())) {
            return type.getConcreteClass().cast(object);
        } else {
            try {
                Constructor<T> c =
                        type.getConcreteClass().getConstructor(NDRObject.class,
                                NDRAccess.class);
                return c.newInstance(object, ndr);
            } catch (ExistenceException e) {
                throw e;
            } catch (Throwable t) {
                System.err.println("Error creating: " + type);
                throw new AccessException("Could not open object "
                        + object.handle(), t);
            }
        }
    }
}
