/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.oai;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.content.AgentModel;
import org.nsdl.repository.model.content.AggregatorModel;
import org.nsdl.repository.model.content.CollectionsModel;
import org.nsdl.repository.model.content.MetadataModel;
import org.nsdl.repository.model.content.MetadataProviderModel;
import org.nsdl.repository.util.NDRObjectTemplate;

public class CollectionContext {

    private static final Logger logger = Logger
            .getLogger(CollectionContext.class);

    private static final Relationship.Name COLLECTION_METADATA_FOR = new Relationship.Name(
            "http://ncs.nsdl.org", "collectionMetadataFor", "ncs");
    private final NDRObject baseObject;
    private final NDRAccess ndr;
    private final boolean isCollectionMD;
    private NDRObjectInfo describedCollectionAggregator;

    public CollectionContext(NDRObject base, NDRAccess ndr) {
        this.baseObject = base;
        this.ndr = ndr;
        this.isCollectionMD = determineIfCollectionMD(base);
    }

    public boolean isCollectionMetadata() {
        return isCollectionMD;
    }

    public NDRObjectInfo getDescribedCollectionAggregator() {
        return describedCollectionAggregator;
    }

    public NDRObject getCollectionMetadata() {
        if (isCollectionMD) {
            return baseObject;
        } else {
            NDRObject provider = ndr.getObject(baseObject.getRelationship(
                    MetadataModel.Components.METADATA_PROVIDED_BY).related());

            /*
             * If there NCS metadata describing this, then we can take the easy
             * path...
             */
            NDRObject ncsCollect = getNCSMetadata(provider);
            if (ncsCollect != null) {
                return ncsCollect;
            }

            /* If no NCS metadata, start digging... */

            Relationship aggBy = provider
                    .getRelationship(CollectionsModel.Components.AGGREGATED_BY);
            if (aggBy == null)
                return null;

            NDRObjectTemplate mdTemplate = new NDRObjectTemplate(Type.Metadata);
            mdTemplate.addRelationship(MetadataModel.Components.METADATA_FOR,
                    aggBy.related());

            Results<NDRObjectInfo> candidates = ndr.getObjectFinder()
                    .matchObject(mdTemplate);

            try {
                for (NDRObjectInfo candidate : candidates) {
                    NDRObject collectionMD = ndr.getObject(candidate);
                    CollectionContext context = new CollectionContext(
                            collectionMD, ndr);
                    if (context.isCollectionMD) {
                        return collectionMD;
                    }
                }
            } finally {
                candidates.close();
            }
        }
        return null;
    }

    private NDRObject getNCSMetadata(NDRObject provider) {
        NDRObjectTemplate template = new NDRObjectTemplate(Type.Metadata);
        template.addRelationship(COLLECTION_METADATA_FOR, provider);

        Results<NDRObjectInfo> matches = ndr.getObjectFinder().matchObject(
                template);
        try {
            for (NDRObjectInfo match : matches) {
                return ndr.getObject(match);
            }
        } finally {
            matches.close();
        }

        return null;
    }

    /**
     * Determine if a piece of metadata is collection metadata.
     * 
     * Metadata is collection metadata if all the following applies:
     * 
     * <ul>
     * <li>The metadata describes an aggregator that is member of NSDL
     * collections</li>
     * </li> The metadata's provider's agent is a member of the collection
     * administrators. </li>
     */

    private boolean determineIfCollectionMD(NDRObject metadata) {

        logger.trace(" Determining if " + baseObject.handle()
                + " is collection metadata..");
        if (determineDescribedCollectionAggregator(metadata) == null) {
            logger.trace(baseObject.handle()
                    + " does not describe a collection aggregator");
            return false;
        }

        NDRObject provider = ndr.getObject(metadata.getRelationship(
                MetadataModel.Components.METADATA_PROVIDED_BY).related());

        NDRObjectTemplate blessedAgent = new NDRObjectTemplate(Type.Agent);
        blessedAgent.setHandle(provider.getRelationship(
                MetadataProviderModel.Components.METADTA_PROVIDER_FOR)
                .related().handle());
        blessedAgent.addRelationship(AgentModel.Components.MEMBER_OF,
                OAITransformer.COLLECITON_ADMIN);

        Results<NDRObjectInfo> foundAgents = ndr.getObjectFinder().matchObject(
                blessedAgent);
        try {
            return foundAgents.iterator().hasNext();
        } finally {
            foundAgents.close();
        }
    }

    private NDRObjectInfo determineDescribedCollectionAggregator(
            NDRObject metadata) {

        NDRObjectTemplate collectionAggTemplate = new NDRObjectTemplate(
                Type.Aggregator);
        collectionAggTemplate.addRelationship(
                AggregatorModel.Components.MEMBER_OF,
                OAIConstants.COLLETION_AGG);
        collectionAggTemplate.addExternalRelationship(metadata,
                MetadataModel.Components.METADATA_FOR);

        logger.trace("Finding described collection aggregators..");
        Results<NDRObjectInfo> foundAggregators = ndr.getObjectFinder()
                .matchObject(collectionAggTemplate);

        try {
            this.describedCollectionAggregator = foundAggregators.iterator()
                    .next();
            if (this.describedCollectionAggregator != null) {
                logger.trace("Found collection aggregator "
                        + this.describedCollectionAggregator.handle());
            } else {
                logger.trace("Did not find described collection aggregator");
            }
            return describedCollectionAggregator;
        } finally {
            foundAggregators.close();
        }
    }
}
