/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.oai;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.content.MetadataModel;
import org.nsdl.repository.model.content.OAIModel;
import org.nsdl.repository.model.types.Metadata;
import org.nsdl.repository.util.StreamUtil;
import org.nsdl.util.DateUtility;

import fedora.common.Constants;

/**
 * @author Edwin Shin
 */
public class OAIMetadata implements OAIConstants, Constants {

    private static final Logger logger = Logger.getLogger(OAIMetadata.class);
    private final NDRAccess ndr;
    private final NDRObject baseMetadataObject;
    private final OAITransformer transformer;
    private CollectionContext context;

    /**
     * If true, this will only allow nsdl_dc, oai_dc, and nsdl_links in the
     * nsdl_search metadata format
     */
    private boolean HACK = true;

    public OAIMetadata(NDRObjectInfo object, OAITransformer transformer,
            NDRAccess ndr) throws IOException {
        this.ndr = ndr;
        baseMetadataObject = ndr.getObject(object);
        this.transformer = transformer;
    }

    public NDRObject getBaseObject() {
        return baseMetadataObject;
    }

    private static void copy(final InputStream input, final OutputStream output)
            throws IOException {
        final byte[] buffer = new byte[4096];
        int n = 0;
        while (-1 != (n = input.read(buffer))) {
            output.write(buffer, 0, n);
        }
    }

    private static String getMetadataFormat(String datastreamID) {
        if (datastreamID.startsWith(MD_DATASTREAM_PREFIX)) {
            datastreamID = datastreamID
                    .substring(MD_DATASTREAM_PREFIX.length());
        }
        if (datastreamID.endsWith(MD_INFO_DATASTREAM_SUFFIX)) {
            datastreamID = datastreamID.substring(0, datastreamID.length()
                    - MD_INFO_DATASTREAM_SUFFIX.length());
        }
        return datastreamID;
    }

    public InputStream getMetadataInfo(String metadataFormat)
            throws IOException {

        // 9/16/05 we no longer have an _info redirect for all calculated
        // formats
        // The calculated formats use the nsdl_dc info stream.
        String lstrFormat = MD_NSDL_DC;

        // Native formats still have their own info though...
        if (metadataFormat.startsWith("native"))
            lstrFormat = metadataFormat;

        Datastream d = baseMetadataObject.getDatastream(new Datastream.Name(
                MD_DATASTREAM_PREFIX + getMetadataFormat(lstrFormat)
                        + MD_INFO_DATASTREAM_SUFFIX));

        if (d == null) {
            return new ByteArrayInputStream(new byte[0]);
        } else {
            return d.getContent();
        }
    }

    public void writeOAIAbout(String metadataFormat, OutputStream out) {
        transformer.getAbout(this, metadataFormat, out);
    }

    /**
     * @return An array of Metadata - in the order they must appear in the
     *         output.
     */
    private List<MetadataFormatInstance> getAllNonAggregateMetadata() {
        List<MetadataFormatInstance> md = new ArrayList<MetadataFormatInstance>();

        // Need to place the mdformats in the set in a specific order...
        // ...because the schema for the aggregates calls for them
        // to be ordered. :-(
        // These local vars will hold the ones we care about.
        InputStream nsdl_dc = null;
        InputStream oai_dc = null;
        InputStream nsdl_links = null;

        // This set will hold all others.
        Set<MetadataFormatInstance> otherFormats = new HashSet<MetadataFormatInstance>();

        try {
            for (Datastream d : baseMetadataObject.getDatastreams()) {
                // Get all datastream names
                String dsName = d.getName().getID();

                if (dsName.startsWith(MD_DATASTREAM_PREFIX)
                        && !dsName.endsWith(MD_INFO_DATASTREAM_SUFFIX)) {

                    // Check for stream name and...
                    // This gets the format we're looking at
                    String format = getMetadataFormat(dsName);

                    // ...get the metadata stream with that name
                    // Check for known format & load them
                    if (format.equalsIgnoreCase(MD_NSDL_DC)) {
                        // When we encounter this format, we know to load
                        // oai_dc & (for now) nsdl_links.
                        nsdl_dc = getMetadata(MD_NSDL_DC);
                        oai_dc = getMetadata(MD_OAI_DC);
                        nsdl_links = getMetadata(MD_NSDL_LINKS);
                    } else if (!HACK) {
                        // something else, add it to the set of others.
                        otherFormats.add(new MetadataFormatInstance(
                                getMetadata(format), format));
                    }
                }
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

        // Now, set up the set to return in the correct order
        // load array with defaults (even if null)
        md.add(0, new MetadataFormatInstance(nsdl_dc, MD_NSDL_DC));
        md.add(1, new MetadataFormatInstance(oai_dc, MD_OAI_DC));
        md.add(2, new MetadataFormatInstance(nsdl_links, MD_NSDL_LINKS));

        // Load the others last...
        md.addAll(otherFormats);

        return md;
    }

    public void writeMetadata(String format, OutputStream out) {
        InputStream metadataContent = null;
        try {
            metadataContent = getMetadata(format);

            if (metadataContent == null) {
                throw new AccessException("Metadata in format " + format
                        + " does not exist in " + baseMetadataObject.handle());
            }
            copy(metadataContent, out);
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            if (metadataContent != null) {
                try {
                    metadataContent.close();
                } catch (IOException e) {
                    /* probably should log... */
                }
            }
        }
    }

    public CollectionContext getCollectionContext() {
        if (context == null) {
            context = new CollectionContext(baseMetadataObject, ndr);
        }

        return context;
    }

    public InputStream getMetadata(String format) throws IOException {

        if (format.equalsIgnoreCase(MD_OAI_DC)) {
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            Datastream nsdl_dc = baseMetadataObject
                    .getDatastream(Metadata.NSDL_DC);

            if (nsdl_dc == null) {
                Datastream oai_dc = baseMetadataObject
                        .getDatastream(MetadataModel.Components.FORMAT_OAI_DC);
                if (oai_dc == null) {
                    return null;
                } else {
                    return oai_dc.getContent();
                }
            }
            transformer.getOAIDC(nsdl_dc.getContent(), out);

            return new ByteArrayInputStream(out.toByteArray());
        } else if (format.equalsIgnoreCase(MD_NSDL_ALL)) {

            StringBuilder sb = new StringBuilder();
            sb
                    .append("<?xml version=\"1.0\" encoding=\"UTF-8\"?><nsdl_all:nsdl_all xmlns:nsdl_all=\"http://ns.nsdl.org/nsdl_all_v1.02/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://ns.nsdl.org/nsdl_all_v1.02/ http://ns.nsdl.org/schemas/nsdl_all/nsdl_all_v1.02.xsd\" schemaVersion=\"1.02.000\">");

            for (MetadataFormatInstance metadata : getAllNonAggregateMetadata()) {
                if (metadata.data != null) {
                    sb.append(getDatestamp(metadata, MD_NSDL_ALL));
                    sb.append(StreamUtil.streamToString(metadata.data)
                            .replaceFirst("\\s*<\\?xml.*?\\?>\\s*", ""));
                }
            }
            sb.append("</nsdl_all:nsdl_all>");
            return new ByteArrayInputStream(sb.toString().getBytes("UTF-8"));
        } else if (format.equalsIgnoreCase(MD_NSDL_SEARCH)) {

            StringBuffer sb = new StringBuffer();
            sb
                    .append("<?xml version=\"1.0\" encoding=\"UTF-8\"?><nsdl_search:nsdl_search xmlns:nsdl_search=\"http://ns.nsdl.org/nsdl_search_v1.02/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://ns.nsdl.org/nsdl_search_v1.02/ http://ns.nsdl.org/schemas/nsdl_search/nsdl_search_v1.02.xsd\" schemaVersion=\"1.02.000\">");
            StringBuffer sbPublic = new StringBuffer();
            sbPublic.append("<nsdl_search:public>");
            StringBuffer sbRestricted = new StringBuffer();
            sbRestricted.append("<nsdl_search:restricted>");

            for (MetadataFormatInstance metadata : getAllNonAggregateMetadata()) {
                if (metadata.data != null) {
                    sbPublic.append(getDatestamp(metadata, MD_NSDL_SEARCH));
                    sbPublic.append(StreamUtil.streamToString(metadata.data)
                            .replaceFirst("\\s*<\\?xml.*?\\?>\\s*", ""));
                }
            }

            sbPublic.append("</nsdl_search:public>");
            sbRestricted.append("</nsdl_search:restricted>");
            sb.append(sbPublic);
            sb.append("</nsdl_search:nsdl_search>");
            return new ByteArrayInputStream(sb.toString().getBytes("UTF-8"));
        } else if (format.equalsIgnoreCase(MD_NSDL_LINKS)) {

            String datestamp = DateUtility
                    .getISO8601MilisecondFromDate(baseMetadataObject
                            .getLastModifiedDate());
            String link = "";

            NDRObject collectionMetatata = getCollectionContext()
                    .getCollectionMetadata();

            if (collectionMetatata != null) {
                Property linkProperty = collectionMetatata
                        .getProperty(OAIModel.Item.Components.ITEM_ID);

                if (linkProperty != null) {
                    link = linkProperty.value();
                } else {
                    logger.warn("Found collection metadata ("
                            + collectionMetatata.handle() + ") for item "
                            + baseMetadataObject.handle()
                            + ", but it has no oai identifier");
                }
            } else {
                try {
                    InfoStream info = new InfoStream(baseMetadataObject
                            .getDatastream(
                                    new Datastream.Name("format_nsdl_dc_info"))
                            .getContent());
                    link = info.getLink();

                } catch (Exception e) {
                    logger
                            .warn("No collection metadata or info stream found for "
                                    + baseMetadataObject.handle());
                }
                logger.info("No collection metadata found for "
                        + baseMetadataObject.handle() + ".  Used info "
                        + "stream to find link value of '" + link + "'");
            }

            StringBuffer sb = new StringBuffer();
            sb
                    .append("<nsdl_links:nsdl_links xmlns:nsdl_links=\"http://ns.nsdl.org/nsdl_links_v1.00\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://ns.nsdl.org/nsdl_links_v1.00 http://ns.nsdl.org/schemas/nsdl_links/nsdl_links_v1.00.xsd\" schemaVersion=\"1.00.001\">");
            sb
                    .append("<nsdl_links:link linkType=\"primaryCollection\" datestamp=\"");
            sb.append(datestamp);
            sb.append("\" linkSource=\"");
            sb.append(link);
            sb.append("\">");
            sb.append(link);
            sb.append("</nsdl_links:link>");
            sb.append("</nsdl_links:nsdl_links>");
            return new ByteArrayInputStream(sb.toString().getBytes("UTF-8"));
        } else {
            Datastream d = baseMetadataObject
                    .getDatastream(new Datastream.Name(MD_DATASTREAM_PREFIX
                            + getMetadataFormat(format)));
            if (d == null) {
                return null;
            } else {
                return d.getContent();
            }
        }
    }

    private String getDatestamp(MetadataFormatInstance data, String parentFormat) {
        String format = getMetadataFormat(parentFormat);
        String prefix = getMetadataFormat(data.format);
        return "<"
                + format
                + ":datestamp mdPrefix=\""
                + prefix
                + "\">"
                + DateUtility.getISO8601MilisecondFromDate(baseMetadataObject
                        .getLastModifiedDate()) + "</" + format + ":datestamp>";
    }

    private class MetadataFormatInstance {
        private final InputStream data;
        private final String format;

        public MetadataFormatInstance(InputStream data, String format) {
            this.data = data;
            this.format = format;
        }
    }
}
