/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.oai;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.Iterator;

import javax.xml.transform.Result;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import fedora.common.Constants;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.content.CollectionsModel;
import org.nsdl.repository.model.content.MetadataModel;
import org.nsdl.repository.model.content.MetadataProviderModel;
import org.nsdl.repository.model.content.OAIModel;
import org.nsdl.repository.model.content.ResourceModel;
import org.nsdl.repository.model.types.Aggregator;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.Encoder;
import org.nsdl.util.DateUtility;

public class OAITransformer implements OAIConstants, Constants {

    private static final Logger logger = Logger.getLogger(OAITransformer.class);

    private TransformerFactory tFactory;
    private HashMap<String, Templates> templates;
    private final NDRAccess ndr;

    public OAITransformer(String xslID, NDRAccess ndr) {
        this.ndr = ndr;

        tFactory = TransformerFactory.newInstance();

        templates = new HashMap<String, Templates>();

        NDRObject xslObject = ndr.getObject(OAIConstants.XSL_OBJECT_PID);

        InputStream is_datastream = null;

        try {
            for (Datastream d : xslObject.getDatastreams()) {
                String dsName = d.getName().getID();
                if (dsName.startsWith(XSL_DATASTREAM_PREFIX)) {
                    logger.debug(" OAITransformer:  loading xsl from " + xslID
                            + " : " + dsName);
                    is_datastream = d.getContent();
                    System.out.println(" OAITransformer:  client.get returned");

                    templates.put(dsName, tFactory
                            .newTemplates(new StreamSource(is_datastream)));
                    is_datastream.close();
                }
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        } finally {
            if (is_datastream != null) {
                try {
                    is_datastream.close();
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
        }
    }

    public void getOAIDC(InputStream nsdl_dc, OutputStream out) {
        Result result = new StreamResult(out);
        try {
            Transformer transformer = getTransformer(XSL_DATASTREAM_PREFIX
                    + MD_OAI_DC);
            transformer.transform(new StreamSource(nsdl_dc), result);
        } catch (Exception e) {
            throw new RuntimeException(e);
        } finally {
            try {
                if (nsdl_dc != null)
                    nsdl_dc.close();
            } catch (IOException e1) {
                logger.warn("Could not close DC stream", e1);
            }
        }
    }

    public void getAbout(OAIMetadata metadata, String metadataFormat,
            OutputStream out) {

        // 9/16/05 All calculated formats use nsdl_dc info now...
        String lstrMetadataFormat = MD_NSDL_DC;

        // ...but natives still have their own...
        if (metadataFormat.startsWith("native"))
            lstrMetadataFormat = metadataFormat;

        Result result = new StreamResult(out);
        InputStream svc_is = null;
        InputStream info_is = null;

        try {
            StreamSource source = null;
            Transformer transformer = getTransformer(XSL_ABOUTS);

            InfoStream info = getInfo(metadata.getBaseObject(),
                    lstrMetadataFormat);

            if (metadata.getCollectionContext().isCollectionMetadata()) {
                info.setCategory("collection");
                info.setLink(metadata.getBaseObject().getProperty(
                        OAIModel.Item.Components.ITEM_ID).value());
            } else {
                info.setCategory("item");
                NDRObject collectionMD = metadata.getCollectionContext()
                        .getCollectionMetadata();
                if (collectionMD != null) {
                    Property parentID = collectionMD
                            .getProperty(OAIModel.Item.Components.ITEM_ID);
                    if (parentID != null) {
                        info.setLink(parentID.value());
                    }
                }
            }

            svc_is = getServiceDescription(metadata);
            info_is = info.getContents();

            // Put info & MDP service description together as a single
            // string.
            source = new StreamSource(new StringReader(getAbouts(info_is,
                    svc_is)));

            // Set the date parameter for the xsl.
            transformer.setParameter("firstCreated", DateUtility
                    .getISO8601MilisecondFromDate(metadata.getBaseObject()
                            .getCreatedDate()));

            StringBuilder resources = new StringBuilder();
            for (Relationship r : metadata.getBaseObject().getRelationships(
                    MetadataModel.Components.METADATA_FOR)) {
                NDRObject resource = ndr.getObject(r.related());
                Property url = resource
                        .getProperty(ResourceModel.Components.IDENTIFIER_URL);
                if (resource.type().equals(Type.Resource) && url != null) {
                    resources.append("<resource handle=\"");
                    resources.append(resource.handle());
                    resources.append("\">");
                    resources.append(Encoder.xmlEncode(url.value()));
                    resources.append("</resource>\n");
                }
            }

            transformer.setParameter("resources", resources.toString());

            // Call to xform
            transformer.transform(source, result);

        } catch (TransformerConfigurationException e) {
            throw new RuntimeException(e);
        } catch (TransformerException e) {
            throw new RuntimeException(e);
        } finally {
            try {
                if (info_is != null)
                    info_is.close();
                if (svc_is != null)
                    svc_is.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    /**
     * Merges the info stream from the Metadata Object and the
     * ServiceDescription stream from the metadataProvider. Produces a merged
     * string.
     * 
     * @param info
     * @param service
     * @return
     */
    private String getAbouts(InputStream info, InputStream service) {
        StringWriter sw = new StringWriter();

        // info stream from metadata item
        InputStreamReader infoReader = null;

        // serviceDescription stream from MD provider for the item
        InputStreamReader serviceReader = null;
        try {
            infoReader = new InputStreamReader(info, "UTF-8");
            serviceReader = new InputStreamReader(service, "UTF-8");

            char[] buffer = new char[4096];
            int n = 0;
            while ((n = infoReader.read(buffer)) != -1) {
                sw.write(buffer, 0, n);
            }
            while ((n = serviceReader.read(buffer)) != -1) {
                sw.write(buffer, 0, n);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            try {
                info.close();
                service.close();
                infoReader.close();
                serviceReader.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }

        // the datastreams didn't previously include the xml processing
        // instructions
        // so the following regex is just a hack to strip the PIs for now
        return "<abouts>"
                + sw.toString().replaceAll("\\s*<\\?xml.*?\\?>\\s*", "")
                + "</abouts>";
    }

    private Transformer getTransformer(String key)
            throws TransformerConfigurationException {
        if (!key.startsWith(XSL_DATASTREAM_PREFIX)) {
            key = key.concat(XSL_DATASTREAM_PREFIX);
        }

        if (!templates.containsKey(key)) {
            throw new RuntimeException("Cannnot find xsl transform " + key);
        }
        return (templates.get(key)).newTransformer();
    }

    private InfoStream getInfo(NDRObject metadata, String format) {
        Datastream about = metadata.getDatastream(getFormatInfoDSName(format));
        if (about == null) {
            return new InfoStream();
        } else {
            return new InfoStream(about.getContent());
        }

    }

    private static Datastream.Name getFormatDSName(String identifier) {
        if (!identifier.startsWith(OAIConstants.MD_DATASTREAM_PREFIX)) {
            identifier = OAIConstants.MD_DATASTREAM_PREFIX + identifier;
        }

        return new Datastream.Name(identifier);
    }

    private static Datastream.Name getFormatInfoDSName(String identifier) {
        return new Datastream.Name(getFormatDSName(identifier)
                + OAIConstants.MD_INFO_DATASTREAM_SUFFIX);
    }

    private InputStream getServiceDescription(OAIMetadata metadata) {

        CollectionContext context = metadata.getCollectionContext();
        if (context.isCollectionMetadata()) {
            NDRObjectInfo aggInfo = context.getDescribedCollectionAggregator();
            if (aggInfo == null) {
                return null;
            } else {
                NDRObject agg = ndr.getObject(aggInfo);
                Results<Relationship> collectionMPResults = ndr
                        .getObjectFinder().getRelatedTo(agg,
                                CollectionsModel.Components.AGGREGATED_BY);
                try {
                    Iterator<Relationship> rel = collectionMPResults.iterator();
                    if (!rel.hasNext()) {
                        logger
                                .warn("Expected to find a metadataProvider "
                                        + "with an aggregatedBy relationship to "
                                        + agg.handle()
                                        + ".  Using ServiceDecription from aggregator...");

                        Datastream sd = agg
                                .getDatastream(Aggregator.SERVICE_DESCRIPTION);
                        if (sd == null) {
                            return null;
                        } else {
                            return sd.getContent();
                        }
                    }

                    Relationship toMP = rel.next();
                    Datastream sd = ndr
                            .getObject(toMP.subject())
                            .getDatastream(
                                    MetadataProviderModel.Components.SERVICE_DESCRIPTION);
                    if (sd == null) {
                        return null;
                    } else {
                        return sd.getContent();
                    }
                } finally {
                    collectionMPResults.close();
                }
            }
        } else {
            /* Just return the serviceDescription of its metadataProvider */
            NDRObject mdp = ndr.getObject(metadata.getBaseObject()
                    .getRelationship(
                            MetadataModel.Components.METADATA_PROVIDED_BY)
                    .related());
            Datastream sd = mdp
                    .getDatastream(MetadataProviderModel.Components.SERVICE_DESCRIPTION);
            if (sd == null) {
                logger.info("No ServiceDescription in provider " + mdp.handle()
                        + ", using fake defaults");
                return new ServiceDescriptionStream().getContents();
            } else {
                return new ServiceDescriptionStream(sd.getContent())
                        .getContents();
            }
        }
    }
}
