/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.tools;

import java.util.concurrent.Executor;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.NDRAccessFactory;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.content.ContentModel;
import org.nsdl.repository.model.content.ContentModelException;
import org.nsdl.repository.model.content.NSDLContentModel;
import org.nsdl.repository.util.BoundedBlockingPoolExecutor;
import org.nsdl.repository.util.Configuration;
import org.nsdl.repository.util.NDRObjectTemplate;

public class ContentModelValidator {

    private static final String PROP_CONFIG_DIR = "repository.config";
    private static final String PROP_THREADS =
            "repository.tools.validator.threads";
    private static final Logger logger =
            Logger.getLogger(ContentModelValidator.class);

    public static void main(String[] args) throws Exception {

        Configuration config = new Configuration(PROP_CONFIG_DIR);
        NDRAccess ndr = NDRAccessFactory.getAccess(config);

        /*
         * A multi-threaded Executor will run content model validation in
         * parallel, with the given number of worker threads.
         */
        String numThreads = config.getProperty(PROP_THREADS, "4");
        Executor executor =
                BoundedBlockingPoolExecutor.getInstance(Integer
                        .parseInt(numThreads));

        /* TODO: specify models via properties */
        ContentModel model = new NSDLContentModel();
        model.init(ndr);

        for (Type<? extends NDRObject> type : Type.values()) {

            NDRObjectTemplate template = new NDRObjectTemplate(type);
            template.setState(State.Active);

            logger.info("Working on objects of type " + type);

            for (NDRObjectInfo found : ndr.getObjectFinder().matchObject(
                    template)) {
                executor.execute(new ContentModelCheck(found, model, ndr));
            }
        }
    }

    private static class ContentModelCheck implements Runnable {

        private final NDRObjectInfo identity;
        private final ContentModel model;
        private final NDRAccess ndr;

        public ContentModelCheck(NDRObjectInfo toCheck, ContentModel m,
                NDRAccess n) {
            this.identity = toCheck;
            this.model = m;
            this.ndr = n;
        }

        public void run() {
            logger.info("Checking " + identity.handle());
            NDRObject toTest = ndr.getObject(identity);

            try {
                model.checkCompliance(toTest);
            } catch (ContentModelException e) {
                logger.error(toTest.handle() + " failed: " + e.getMessage());
            } catch (Throwable t) {
                logger
                        .error("Could not process record " + identity.handle(),
                                t);
            }
        }
    }
}
