/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.tools;

import java.util.ArrayList;
import java.util.Date;

import org.apache.log4j.Logger;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.Results;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Resource;
import org.nsdl.repository.model.types.Type;

/** Merges a set of equivalent resources into one */
public class ResourceMerger {

    private static final Logger LOG = Logger.getLogger(ResourceMerger.class);

    private final NDRAccess m_ndr;

    public ResourceMerger(NDRAccess ndr) {
        m_ndr = ndr;
    }

    /** Merge all resources into the oldest. */
    public Resource mergeAll(Resource... resources) {
        Resource oldestResource = null;
        Date oldestDate = new Date();

        /* Go through all resources and find the oldest */
        for (Resource r : resources) {
            if (r.getCreatedDate().compareTo(oldestDate) < 0) {

                oldestResource = r;
                oldestDate = r.getCreatedDate();
            }
        }

        if (oldestResource != null) {
            mergeTo(oldestResource, resources);
        }
        return oldestResource;
    }

    /** Convenience memthod for merging resources from an iterable stream */
    public Resource mergeAll(Iterable<NDRObjectInfo> resources) {
        ArrayList<Resource> rList = new ArrayList<Resource>();

        for (NDRObjectInfo o : resources) {
            rList.add(m_ndr.getObject(o, Type.Resource));
        }

        return mergeAll(rList.toArray(new Resource[0]));
    }

    /** Merge all resources into the given canonical resource */
    public void mergeTo(Resource canonical, Resource... mergeCandidates) {
        for (Resource toMerge : mergeCandidates) {

            /* Don't merge the canonical resource into itself */
            if (toMerge.handle().equals(canonical.handle())) {
                continue;
            }

            /* First, copy over internal components */
            canonical.addRelationships(toMerge.getRelationships());

            /*
             * XXX: For now, just drop the 'hasHandle' property. At some point,
             * we may want to do something different, (especially if we want
             * handles to be persistent for life), but that would imply objects
             * having multiple handles, etc. etc. Certainly something not to
             * deal with right now...
             */
            for (Property p : toMerge.getProperties()) {
                if (!p.predicate().equals(Resource.HAS_HANDLE)) {
                    canonical.addProperty(p.predicate(), p.value());
                }
            }

            /*
             * TODO: Right now, we ignore all datastreams. In reality, we
             * probably want to merge in all NON CONTENT datastreams. Merging
             * resources implies that they have equivalent content, so there's
             * no point in merging the content datastream
             */

            /* Now, external components */
            Results<Relationship> externalRels = m_ndr.getObjectFinder()
                    .getRelatedTo(toMerge);

            try {
                for (Relationship r : externalRels) {
                    changeRelationshipObject(canonical, r);
                }
            } finally {
                externalRels.close();
            }

            m_ndr.purgeObject(toMerge);
            LOG.info("Purged resource " + toMerge.handle());
        }

        LOG.info("Committed canonical resource " + canonical.handle());
        canonical.commit();
    }

    private void changeRelationshipObject(NDRObject destinationObject,
            Relationship relationship) {
        NDRObject relationshipOrigin = m_ndr.getObject(relationship.subject());

        relationshipOrigin.removeRelationship(relationship.predicate(),
                relationship.related());
        relationshipOrigin.addRelationship(relationship.predicate(),
                destinationObject);

        relationshipOrigin.commit();
        LOG.info("Changed " + relationshipOrigin.handle() + "->"
                + relationship.predicate().getLocalPart() + " from "
                + relationship.related() + " to " + destinationObject.handle());
    }
}
