/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Properties;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * Represents a persistent, mutable set of configuration properties.
 * <p>
 * Can read configuration from {@link java.util.Properties} text files, as well
 * as {@link System} properties. System property values always override those
 * loaded from configuration properties files. Classes may subscribe to the
 * current configuration (by implementing {@link Listener}) in order to be
 * notified of changes to its properties upon reload();
 * </p>
 * 
 * @author birkland
 * @since 1.1
 */
public class Configuration {

    private Map<String, String> m_localProperties = new HashMap<String, String>();

    private final List<Listener> listeners = new ArrayList<Listener>();

    private final File configLocation;

    /**
     * Create a configuration based solely on System properties.
     * <p>
     * With no specified file, this class will essentially act like a proxy to
     * System properties, with added {@link #reset() reload()} capabilities.
     * </p>
     */
    public Configuration() {
        configLocation = null;
    }

    /**
     * Create a configuration with a specified configuration file or directory.
     * <p>
     * The given {@link File File} may be either a directory or normal file. If
     * a directory, configuration will be loaded from every file therein.
     * Otherwise, if a normal file, configuration will be loaded from that file
     * alone.
     * </p>
     * 
     * @param location
     *            File or directory where configuration resides.
     */
    public Configuration(File location) {
        this.configLocation = location;
        load();
    }

    /**
     * Create a configuration using the file location specified in a system
     * property.
     * <p>
     * Will look for a file or directory at the location specified in the given
     * system property. If the location is a single file, will load
     * configuration from that one file. If a directory, will try to load every
     * file in that directory.
     * </p>
     * 
     * @param fileLocationProperty
     *            System property whose value specifies file or directory where
     *            configuration resides.
     */
    public Configuration(String fileLocationProperty) {
        String location = System.getProperty(fileLocationProperty);
        if (location != null) {
            configLocation = new File(location);
        } else {
            configLocation = null;
        }

        load();
    }

    /**
     * Returns a set view of the keys contained in this configuration. The set is backed by the configuration map, 
     * so changes to the map are reflected in the set, and vice-versa. See Map's keySet() for more information.
     * </p>
     * 
     * @return Set of keys
     */
    public Set<String> keySet() {
        return m_localProperties.keySet();
    }

    /**
     * Determine if a given property is defined.
     * <p>
     * Search file-defined properties and System-defined properties for the
     * presence of the given key.
     * </p>
     * 
     * @param key
     *            Name of property.
     * @return true if the property has been defined, false otherwise.
     */
    public boolean containsKey(String key) {
        return m_localProperties.containsKey(key)
                || System.getProperties().containsKey(key);
    }

    /**
     * Get a specific property.
     * <p>
     * If a System property has been defined for the given property, return it.
     * Otherwise, return any value that is defined in any specified
     * configuration files.
     * </p>
     * 
     * @param key
     *            Name of the property to get.
     * @return String value of the property, null if undefined.
     */
    public String getProperty(String key) {
        return System.getProperty(key, m_localProperties.get(key));
    }

    /**
     * Get a specific property, returning a default value if it is not assigned.
     * 
     * @param key
     *            Name of property to get.
     * @param defaultIfNotSet
     *            Default value to return if the property value is undefined
     *            otherwise.
     * @return String value of property.
     */
    public String getProperty(String key, String defaultIfNotSet) {
        String property = getProperty(key);
        return property != null ? property : defaultIfNotSet;
    }

    /**
     * Return all values of properties that whose name starts with the given
     * string.
     * <p>
     * Will return values in the order based upon an ASCII sort of their
     * property keys. For example, if a properties file contains the following:
     * 
     * <pre>
     * config.property.1 = one
     * config.property.3 = three
     * config.property.2 = two
     * </pre>
     * 
     * then <code>getProperties("config.property")</code> would return the list
     * <code>["one", "two", "three"]</code>.
     * <p>
     * 
     * @param startsWith
     *            String that all matching property names are to start with.
     * @return List of configuration property values, returned in the prescribed
     *         order.
     */
    public List<String> getProperties(String startsWith) {
        SortedMap<String, String> matches = getPropertyMap(startsWith);
        ArrayList<String> values = new ArrayList<String>();
        for (String match : matches.values()) {
            values.add(match);
        }

        return values;
    }

    /**
     * Returns a map of property key/values whose key starts with the given
     * string.
     * 
     * @param startsWith
     *            String that all matching property names are to start with.
     * @return Map of configuration property keys and values.
     */
    public SortedMap<String, String> getPropertyMap(String startsWith) {
        SortedMap<String, String> matches = new TreeMap<String, String>();

        Properties systemProps = System.getProperties();
        for (Object key : systemProps.keySet()) {
            if (((String) key).startsWith(startsWith)) {
                matches
                        .put((String) key, systemProps
                                .getProperty((String) key));
            }
        }

        for (String key : m_localProperties.keySet()) {
            if (key.startsWith(startsWith) && !matches.containsKey(key)) {
                matches.put(key, m_localProperties.get(key));
            }
        }

        return matches;
    }

    public void putProperties(Properties props) {
    	for (Enumeration<?> keys = props.propertyNames(); keys.hasMoreElements();) {
    		String key = (String) keys.nextElement();
    		m_localProperties.put(key, props.getProperty(key)); 
    	}
    }

    /**
     * Add a single property.
     * <p>
     * This property will not be known to any config-enabled modules until
     * {@link #update()} is called. If this configuration is {@link reset()},
     * any properties added individually through this method will be lost.
     * </p>
     * 
     * @param propName
     *            Name of the property.
     * @param value
     *            Value of the property.
     */
    public void putProperty(String propName, String value) {
        m_localProperties.put(propName, value);
    }

    /**
     * Add a listener to be notified upon configuration reload.
     * 
     * @param listener
     *            Listener to notify when configuration is reloaded.
     */
    public void addListener(Listener listener) {
        listeners.add(listener);
    }

    private void load() {
        if (configLocation == null)
            return;

        if (configLocation.isDirectory()) {
            for (File configFile : configLocation.listFiles()) {
                putProperties(configFile);
            }
        } else if (configLocation.isFile()) {
            putProperties(configLocation);
        } else {
            throw new RuntimeException(
                    "Do not know how to read config location");
        }
    }

    public void putProperties(File file) {

        if (file.isDirectory()) {
            for (File configFile : file.listFiles()) {
                if (configFile.getName().endsWith(".properties")) {
                    putProperties(configFile);
                }
            }
        } else {

            Properties reader = new Properties();
            FileInputStream stream = null;
            try {
                stream = new FileInputStream(file);
                reader.load(stream);

                for (Object key : reader.keySet()) {
                    m_localProperties.put((String) key, reader
                            .getProperty((String) key));

                }
            } catch (IOException e) {
                throw new RuntimeException(
                        "Error reading configuration properties file", e);
            } finally {
                if (stream != null) {
                    try {
                        stream.close();
                    } catch (IOException e) {
                    }
                }
            }
        }
    }

    /**
     * Reload the configuration from any relevant configuration files, and
     * notify listeners of change.
     */
    public void reset() {
        m_localProperties = new HashMap<String, String>();
        load();
        update();
    }

    /** Update listeners with current configuration */
    public void update() {
        for (Listener l : listeners) {
            l.updateConfig(this);
        }
    }

    /**
     * An Object that can listen for configuration changes.
     * <p>
     * Upon notification of configuration update, an implementing class will do
     * the appropriate actions to reflect the configuration change, if
     * applicable.
     * </p>
     * 
     * @author birkland
     * @since 1.1
     */
    public interface Listener {

        /**
         * Notify listener of configuration change.
         * 
         * @param config
         *            Updated configuration parameters.
         */
        public void updateConfig(Configuration config);
    }
}
