/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.util;

import org.nsdl.repository.access.AccessException;

/**
 * Factory for building and instantiating configurable classes.
 * 
 * Given a class specified in a configuration property, string, or Class object,
 * will create a new instance using a no-argument constructor and, if the class
 * implements {@link Configuration.Listener}, will bootstrap the class with the
 * provided configuration.
 * 
 * @author birkland
 * @since 1.1
 * @param <T>
 *            Type of object produced by a given factory instance. Usually an
 *            interface or superclass of the instances produced by the factory.
 */
public class ConfigurationFactory<T> {

    private final Class<T> type;

    private ConfigurationFactory(Class<T> clazz) {
        this.type = clazz;
    }

    /**
     * Create a factory that builds objects of the given class.
     * 
     * @param clazz
     *            Class of object this factory creates.
     * @return Factory instance.
     */
    public static <T> ConfigurationFactory<T> getInstance(Class<T> clazz) {
        return new ConfigurationFactory<T>(clazz);
    }

    /**
     * Build an instance of the class specified in the given property.
     * 
     * @param property
     *            Name of the property whose value contains a class name.
     * @param config
     *            Configuration used to initialize the new object.
     * @return Fully configured object instance.
     */
    public T buildFromProperty(String property, Configuration config) {
        return buildFromClassName(config.getProperty(property), config);
    }

    /**
     * Build an instance of the class specified in the given string.
     * 
     * @param implClassNname
     *            Fully qualified name of the class to create.
     * @param config
     *            Configuration used to initialize the new object.
     * @return Fully configured object instance.
     */
    public T buildFromClassName(String implClassNname, Configuration config) {
        try {
            Class<? extends T> implClass =
                    Class.forName(implClassNname).asSubclass(type);

            return buildFromClass(implClass, config);
        } catch (ClassNotFoundException e) {
            throw new AccessException("Could not create instance of "
                    + implClassNname + ": class not found", e);
        }
    }

    /**
     * Build an instance from the given class.
     * 
     * @param implClass
     *            Class from which to create a new instance.
     * @param config
     *            Configuration used to initialize the new object.
     * @return Fully configured object instance.
     */
    public T buildFromClass(Class<? extends T> implClass, Configuration config) {
        T impl;
        try {
            impl = implClass.newInstance();

            if (impl instanceof Configuration.Listener) {
                ((Configuration.Listener) impl).updateConfig(config);
            }

        } catch (Exception e) {
            throw new AccessException("Could not create instance of "
                    + implClass.getName(), e);
        }

        return impl;
    }
}
