/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashMap;
import java.util.Map;

import org.nsdl.repository.model.Datastream;

import eu.medsea.util.MimeException;
import eu.medsea.util.MimeUtil;

/**
 * Utility class for determining characteristics of content (such as MIME type).
 * 
 * @author birkland
 * @since 1.3
 */
public class MimeTypeFinder {

    private static final String OCTET_STREAM = "application/octet-stream";
    /** New OOXML mime types, not yet in mimeUtil */
    private static final Map<String, String> OOXML_MIMES = new HashMap<String, String>();

    static {
        OOXML_MIMES.put(".docm",
                "application/vnd.ms-word.document.macroEnabled.12");
        OOXML_MIMES
                .put(".docx",
                        "application/vnd.openxmlformats-officedocument.wordprocessingml.document");
        OOXML_MIMES.put(".dotm",
                "application/vnd.ms-word.template.macroEnabled.12");
        OOXML_MIMES
                .put(".dotx",
                        "application/vnd.openxmlformats-officedocument.wordprocessingml.template");
        OOXML_MIMES.put(".potm",
                "application/vnd.ms-powerpoint.template.macroEnabled.12");
        OOXML_MIMES
                .put(".potx",
                        "application/vnd.openxmlformats-officedocument.presentationml.template");
        OOXML_MIMES.put(".ppam",
                "application/vnd.ms-powerpoint.addin.macroEnabled.12");
        OOXML_MIMES.put(".ppsm",
                "application/vnd.ms-powerpoint.slideshow.macroEnabled.12");
        OOXML_MIMES
                .put(".ppsx",
                        "application/vnd.openxmlformats-officedocument.presentationml.slideshow");
        OOXML_MIMES.put(".pptm",
                "application/vnd.ms-powerpoint.presentation.macroEnabled.12");
        OOXML_MIMES
                .put(".pptx",
                        "application/vnd.openxmlformats-officedocument.presentationml.presentation");
        OOXML_MIMES.put(".xlam",
                "application/vnd.ms-excel.addin.macroEnabled.12");
        OOXML_MIMES.put(".xlsb",
                "application/vnd.ms-excel.sheet.binary.macroEnabled.12");
        OOXML_MIMES.put(".xlsm",
                "application/vnd.ms-excel.sheet.macroEnabled.12");
        OOXML_MIMES
                .put(".xlsx",
                        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet");
        OOXML_MIMES.put(".xltm",
                "application/vnd.ms-excel.template.macroEnabled.12");
        OOXML_MIMES
                .put(".xltx",
                        "application/vnd.openxmlformats-officedocument.spreadsheetml.template");
    }
    /** 3 second timeout for analyzing remote content */
    public static final int TIMEOUT = 3000;

    public static String findMimeType(Datastream d) {
        if (d.isCanonical()) {
            return findMimeType(d.getLabel(), d.getContent());
        } else {
            return findMimeType(d.getURL());
        }
    }

    public static String findMimeType(URL url) {
        if (url != null) {
            try {
                URLConnection uc = url.openConnection();
                uc.setConnectTimeout(TIMEOUT);
                String mime = uc.getContentType();
                if (mime == null) {
                    return findMimeType(url.getFile(), uc.getInputStream());
                } else {
                    if (mime.indexOf(';') > 0) {
                        return mime.subSequence(0, mime.indexOf(';'))
                                .toString();
                    } else {
                        return mime;
                    }
                }
            } catch (Exception e) {
                /* Oh well.. */
            }
        }

        return null;
    }

    public static String findMimeType(File file) {
        FileInputStream content = null;
        try {
            content = new FileInputStream(file);
        } catch (Exception e) {
            /* Oh well */
        }
        return findMimeType(file.getName(), content);
    }

    public static String findMimeType(String filename, InputStream content) {
        String mime = null;

        /* First, try by filename */
        if (filename != null) {
            mime = MimeUtil.getMimeType(new File(filename));
        }

        /*
         * If that doesn't work, try stripping out extraneous dots (seems to be
         * funny bug in MimeUtil)
         */
        if (OCTET_STREAM.equals(mime) && filename.contains(".")) {
            mime = MimeUtil.getMimeType(new File(stripDots(filename)));
        }

        /* If that don't work, analyze content */
        if (content != null && (mime == null || OCTET_STREAM.equals(mime))) {
            try {
                mime = MimeUtil.getMimeType(content);
            } catch (MimeException e) {
                /*
                 * FIXME: MimeUtil requires IS to support mark anr reset. Is
                 * there any way around that? Too few InputStreams actuallly
                 * suport that.
                 */
            }
        }

        /* Lastly, try some common ones that aren't in MimeUtil */
        if (mime == null && filename != null || OCTET_STREAM.equals(mime)) {
            if (filename.contains(".")) {
                String alternate = OOXML_MIMES.get(filename.substring(filename
                        .lastIndexOf('.')));

                if (alternate != null) {
                    return alternate;
                } else {
                    return sanitize(mime);
                }
            }
        } else {
            return sanitize(mime);
        }

        /* Fail! */
        return OCTET_STREAM;
    }

    public static boolean isMimeType(String candidate) {
        if (MimeUtil.isMimeTypeKnown(candidate)) {
            return true;
        } else {
            return candidate.matches("^[^/]+/[^/]+$");
        }
    }

    private static String sanitize(String input) {
        if (input.indexOf(',') > 0) {
            return input.split(",")[0];
        } else {
            return input;
        }
    }

    private static String stripDots(String input) {
        StringBuilder newName = new StringBuilder();
        String[] parts = input.split("\\.");
        for (int i = 0; i < parts.length - 1; i++) {
            newName.append(parts[i]);
        }

        newName.append('.');
        newName.append(parts[parts.length - 1]);
        return newName.toString();
    }
}
