/*
 * -----------------------------------------------------------------------------
 * 
 * <p><b>License and Copyright: </b>The contents of this file are subject to
 * the Educational Community License (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License
 * at <a href="http://www.opensource.org/licenses/ecl1.txt">
 * http://www.opensource.org/licenses/ecl1.txt.</a></p>
 * 
 * <p>Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under the
 * License.</p>
 * 
 * <p>The entire file consists of original code. Copyright &copy; 2006-2007
 * Cornell University. All rights reserved.</p>
 * 
 * -----------------------------------------------------------------------------
 */

package org.nsdl.repository.util;

import org.apache.log4j.Logger;
import java.util.*;
import javax.servlet.http.HttpServletRequest;

/**
 * <h3>NSDL OnRamp/OnFire Project</h3>
 * <hr width="150px" align="left">
 * 
 * <p><b>Title:</b> RequestParameters.java</p>
 * <dl>
 *   <dt><b>Description:</b></dt>
 *     <dd>Define structure to process parameters from the HTTP_REQUEST.</dd>
 * </dl>
 *
 * @version NDR 2.0
 * @author  E.L. Rayle (elrayle@cs.cornell.edu)
 * @since   2007_11_29
 */
/*
 *----------------------------------------------------------
 * History:
 *   2007_11_29 elr Created as part of OnFire project
 *   2010_05_19 elr Copied to NDR repository project
 *----------------------------------------------------------
 * Notes:
 *----------------------------------------------------------
 */

public class RequestParameters
{
	private static final long serialVersionUID = 1L;

	/** Full name of this class. */
	static String classname = RequestParameters.class.getName();
	static int lastdot = classname.lastIndexOf(".");
	static String shortname = classname.substring(lastdot+1);
	static String packagename = classname.substring(0, lastdot);

	private static final Logger logger = Logger.getLogger(classname);
	
    /*--------------------*/
    /* Constants          */
    /*--------------------*/

    /*--------------------*/
    /* Private Variables  */
    /*--------------------*/
	// parameter map for parameters passed in on the request
	Map requestParameters = null;
	
	
    /*-------------------------------------------------------------------------*/
    /** 
     * Constructor to create a new instance of this class.
     */
    /*-------------------------------------------------------------------------*/
	public RequestParameters(){}

    /*#########################################################################*/
    /*                                Methods                                  */
    /*#########################################################################*/

    /** Set request parameters. @param request - request holding the parameters */
    public void setRequestParameters(HttpServletRequest request){ this.requestParameters = request.getParameterMap();; }

    /** Set request parameters. @param requestParameters - map of request parameters */
    public void setRequestParameters(Map requestParameters){ this.requestParameters = requestParameters; }

	/*-------------------------------------------------------------------------*/
    /**
     * Get the value of a parameter that was passed in on the request.
     * 
     * @param parmName - name of the parameter
     *
     * @return value of the parameter
     */
    /*-------------------------------------------------------------------------*/
    public Map getParameters()
    {
    	return(this.requestParameters);
    }

	/*-------------------------------------------------------------------------*/
    /**
     * Get the value of a parameter that was passed in on the request.
     * 
     * @param parmName - name of the parameter
     *
     * @return value of the parameter
     */
    /*-------------------------------------------------------------------------*/
    public String getParameter(
    	String parmName )
    {
    	return(getParameter(parmName,null));
    }

    /*-------------------------------------------------------------------------*/
    /**
     * Get the value of a parameter that was passed in on the request.
     * 
     * @param parmName - name of the parameter
     * @param defaultValue - value to use if the parameter is not present
     *
     * @return value of the parameter
     */
    /*-------------------------------------------------------------------------*/
    public String getParameter(
    	String parmName,
    	String defaultValue )
    {
    	if( this.requestParameters == null )
    		// Return null to distinguish between the entire table 
    		// missing versus just the parameter missing?
    		return null;
    		
    	String parmValue = getParameter(requestParameters,parmName,null);
    	if( parmValue == null ) parmValue = defaultValue;
    	return(parmValue);
    }

    /*-------------------------------------------------------------------------*/
    /** 
     * Get a parameter off the request parameter map.
     *
     * @param  parameterMap - map of parameters
     * @param  parameterName - name of the parameter 
     * @param  defaultValue - value to return if the parameter does not exists 
     *
     * @return Value of the parameter.  Returns the defaultValue if the parameter is not in the map.
     */
    /*-------------------------------------------------------------------------*/
	static private String getParameter(
		Map parameterMap,
		String parameterName,
		String defaultValue )
    {
		String[] parmArray = (String[])parameterMap.get( parameterName );
    	
    	String parmValue = null;
    	if( parmArray != null && parmArray.length > 0 )
    		parmValue = parmArray[0];

    	if( parmValue == null )
    		parmValue = defaultValue;
		
    	return parmValue;
    }

    /*-------------------------------------------------------------------------*/
    /** 
     * Get a parameter off the request parameter map.
     *
     * @param  HttpRequest - map of parameters
     * @param  parameterName - name of the parameter 
     * @param  defaultValue - value to return if the parameter does not exists 
     *
     * @return Value of the parameter.  Returns the defaultValue if the parameter is not in the map.
     */
    /*-------------------------------------------------------------------------*/
	static public String getParameter(
		HttpServletRequest request,
		String parameterName,
		String defaultValue )
    {
		Map parameterMap = request.getParameterMap();
		return getParameter(parameterMap,parameterName,defaultValue);
    }
}