/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
package org.nsdl.repository.util;

import java.util.Collection;
import java.util.Iterator;
import java.util.NoSuchElementException;

import org.nsdl.repository.access.Results;

/**
 * Combines several {@link Results} into a single results object.
 * <p>
 * A collection of results objects are given at time of construction, and in
 * operation, the combiner will pull results from each one until empty.
 * Normally, the given constituent results are drained and close()d once empty.
 * If close() is called on the ResultsCombiner before all constituent result
 * sets have been processed, it will close() the current and all remaining
 * {@link Results} objects.
 * </p>
 * 
 * @author birkland
 * 
 * @since 1.3
 */
public class ResultsCombiner<T> implements Results<T> {

    private final CombinedResultsIterator m_iterator;

    public ResultsCombiner(final Collection<Results<? extends T>> toCombine) {
        m_iterator = new CombinedResultsIterator(toCombine.iterator());
    }

    public ResultsCombiner(final Results<? extends T>... toCombine) {

        m_iterator = new CombinedResultsIterator(
                new Iterator<Results<? extends T>>() {

                    int i = 0;

                    public boolean hasNext() {
                        return i < toCombine.length;
                    }

                    public Results<? extends T> next() {
                        return toCombine[i++];
                    }

                    public void remove() {
                        throw new UnsupportedOperationException();
                    }
                });
    }

    public void close() {
        m_iterator.close();
    }

    public Iterator<T> iterator() {
        return m_iterator;
    }

    private class CombinedResultsIterator implements Iterator<T> {

        private final Iterator<Results<? extends T>> m_resultSets;
        private Results<? extends T> m_currentResults;
        private Iterator<? extends T> m_currentResultsIterator;

        public CombinedResultsIterator(
                Iterator<Results<? extends T>> resultSetsIterator) {
            m_resultSets = resultSetsIterator;

            if (m_resultSets.hasNext()) {
                m_currentResults = m_resultSets.next();
                m_currentResultsIterator = m_currentResults.iterator();
            }
        }

        public boolean hasNext() {
            if (m_currentResults == null) {
                return false;
            }

            while (!m_currentResultsIterator.hasNext()) {
                m_currentResults.close();
                if (!m_resultSets.hasNext()) {
                    return false;
                } else {
                    m_currentResults = m_resultSets.next();
                    m_currentResultsIterator = m_currentResults.iterator();
                }
            }
            return true;
        }

        public T next() {
            if (hasNext()) {
                return m_currentResultsIterator.next();
            } else {
                throw new NoSuchElementException("Result iterator is empty!");
            }
        }

        public void remove() {
            throw new UnsupportedOperationException(
                    "Cannot remove from cocmbined result sets");
        }

        public void close() {
            m_currentResults.close();
            while (m_resultSets.hasNext()) {
                m_resultSets.next().close();
            }
        }
    }
}
