/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.util;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;

import org.apache.log4j.Logger;

/**
 * Static convenience methods for converting streams to other useful forms.
 * 
 * @author birkland
 * 
 */
public class StreamUtil {
    private static final Logger logger = Logger.getLogger(StreamUtil.class);
    /**
     * Create a string from an InputStream.
     * <p>
     * Will attempt to create a string from the bytes in the InputStream.
     * Assumes that the stream uses a utf-8 character encoding.
     * </p>
     * 
     * @param stream
     *            InputStream to convert.
     * @return assembled String.
     */
    public static String streamToString(InputStream stream) {
    	return streamToString(stream,4096);
    }
    public static String streamToString(InputStream stream, int sizeLimit) {
//    	logger.debug("*** Entering streamToString(InputStream,sizeLimit="+sizeLimit+")");
    	if( sizeLimit > 0 ){
//        	logger.debug("read bytes up to sizeLimit ("+sizeLimit+")");
    		// read bytes upto sizeLimit
            StringBuilder out = new StringBuilder();
            byte[] b = new byte[sizeLimit];
            try {
                for (int n; (n = stream.read(b)) != -1;) {
                    out.append(new String(b, 0, n, "UTF-8"));
                }
                return out.toString();
            } catch (Exception e) {
                throw new RuntimeException("Could not read string!", e);
            }
    	}
    	else
    	{
//    		logger.debug("read all bytes");
    		// read all bytes
    		StringBuilder contents = new StringBuilder();
            String line;
            BufferedReader reader = new BufferedReader(
                    new InputStreamReader(stream));
            try {
                while ((line = reader.readLine()) != null) {
                    contents.append(line + "\n");
                }
            } catch (IOException e) {
                throw new RuntimeException(e);
            } finally {
                try {
                    reader.close();
                } catch (IOException e) {
//                    logger.warn("Could not close reader (not that anyone cares..)");
                }
            }
            return contents.toString();
    	}
    }

    /**
     * Creates an array of bytes from an InputStream.
     * <p>
     * Watch out! Will create a byte array containing all bytes from the
     * InputStream, without regard to size.
     * </p>
     * 
     * @param stream
     *            InputStream to convert
     * @return (possibly huge) byte array.
     */
    public static byte[] streamToByteArray(InputStream stream) {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        int num_read = 0;
        byte[] buf = new byte[1024];
        try {
            while ((num_read = stream.read(buf)) != -1) {
                out.write(buf, 0, num_read);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        return out.toByteArray();
    }

    public static InputStream stringToStream(String input) {
        try {
            return new ByteArrayInputStream(input.getBytes("UTF-8"));
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException("Could not encode string", e);
        }
    }
}
