/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.util;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Date;

import org.apache.log4j.AppenderSkeleton;
import org.apache.log4j.spi.LoggingEvent;

/**
 * Appender that logs messages executing in a single thread if its execution
 * time exceeds a set threshold.
 * <p>
 * Useful for debugging/logging single-thread requests that take a long time,
 * this appender keeps all logging messages produced by a thread between
 * specified start/finish markers in an in-memory buffer, writing these messages
 * to disk in its own file if the thread execution time exceeds the given
 * threshold.
 * 
 * For each thread that exceeds the threshold execution time between markers,
 * all log messages written between the start/finish markers will be written to
 * a file. The file will be named <code>XXXX-YYYYYYY.log</code> where
 * <code>XXXX</code> is the execution time, and <code>YYYYYYY</code> is the time
 * the appender started recording the log events (at the start marker.
 * </p>
 * 
 * <p>
 * <h2>Usage</h2>
 * 
 * <pre>
 * ThreadExecutionTimeAppender.start(); // Start marker 
 * try {
 *     // do processing and logging
 * } finally {
 *     ThreadExecutionTimeAppender.finish(); // End marker
 * }
 * </pre>
 * 
 * </p>
 * 
 * <p>
 * <h2>Log4j Configuration Params</h2>
 * <dl>
 * <dt><code>ExecutionTimeThreshold</code></dt>
 * <dd>Minimum time (in miliseconds) that a thread must execute between
 * start/finish markers in order for the log to be written to disk. Threads that
 * execute faster than the threshold will not be logged at all.</dd>
 * 
 * <dt><code>Directory</code></dt>
 * <dd>Directory into which log files will be written.</dd>
 * </dl>
 * </p>
 * 
 * @author birkland
 * 
 */
public class ThreadExecutionTimeAppender extends AppenderSkeleton {

    private static int threshold;

    private static String logDir;

    private static ThreadLocal<TimeThresholdLogWriter> logWriter = new ThreadLocal<TimeThresholdLogWriter>() {
        @Override
        public TimeThresholdLogWriter initialValue() {
            return new TimeThresholdLogWriter();
        }
    };

    public static void start() {
        logWriter.get().start();
    }

    public static void finish() {
        logWriter.get().finish();
        logWriter.remove();
    }

    public boolean requiresLayout() {
        return true;
    }

    public void append(LoggingEvent event) {
        logWriter.get().appendLog(layout.format(event));
    }

    public void close() {
        /* Do nothing */
    }

    public static void setExecutionTimeThreshold(int miliseconds) {
        threshold = miliseconds;
    }

    public int getExecutionTimeThreshold() {
        return threshold;
    }

    public static void setDirectory(String dir) {
        logDir = dir;
    }

    public String getDirectory() {
        return logDir;
    }

    private static class TimeThresholdLogWriter {
        private final StringBuilder buffer = new StringBuilder();
        private long start = 0;
        private long finish = 0;
        private boolean listening = false;

        public void start() {
            start = new Date().getTime();
            listening = true;
        }

        public void appendLog(String logMessage) {
            if (listening) {
                buffer.append(logMessage);
            }
        }

        public void finish() {
            try {
                if (listening && logDir != null) {
                    finish = new Date().getTime();
                    writeLog();
                }
            } catch (IOException e) {
                throw new RuntimeException("Could not write log", e);
            } finally {
                listening = false;
            }
        }

        private void writeLog() throws IOException {
            int executionTime = (int) (finish - start);
            if (executionTime > threshold) {
                FileWriter out = null;
                try {
                    out = new FileWriter(logDir + File.separator
                            + pad(executionTime) + "-" + start + ".log");

                    out.append(buffer);
                } finally {
                    if (out != null)
                        out.close();
                }
            }
        }

        private static String pad(int value) {
            StringBuilder padded = new StringBuilder();
            String val = Integer.toString(value);

            for (int i = val.length(); i < 8; i++) {
                padded.append('0');
            }

            padded.append(val);

            return padded.toString();
        }
    }
}
