/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.util;

import java.text.*;
import java.util.*;

/**
 * utility methods for working with dates in different formats
 * 
 * @author Naomi Dushay
 **/
public abstract class DateUtility {
    private static String errmsg;

    private static final String ISO8601PlusDateFormat = "yyyy-MM-dd'T'HH:mm:ss";
    private static final String ISO8601PlusDashedDateFormat = "yyyy-MM-dd'T'HH-mm-ss'Z'";
    private static final String ISO8601SecondsUTCFormat = "yyyy-MM-dd'T'HH:mm:ss'Z'";
    private static final String ISO8601DateFormat = "yyyy-MM-dd";
    private static final String ISO8601MilisecondsUTCFormat = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'";
    private static final String ISO8601CentisecondsUTCFormat = "yyyy-MM-dd'T'HH:mm:ss.SS'Z'";
    private static final String ISO8601DecisecondsUTCFormat = "yyyy-MM-dd'T'HH:mm:ss.S'Z'";

    /** @return the current local time in ISO8601 Plus format (not UTC) */
    public static String getTimestamp() {
        SimpleDateFormat df = new SimpleDateFormat(ISO8601PlusDateFormat);
        Date currDate = Calendar.getInstance().getTime();
        return (df.format(currDate));
    }

    public static String getTimestamp(String lstrFormat) {
        SimpleDateFormat df = new SimpleDateFormat(lstrFormat);
        Date currDate = Calendar.getInstance().getTime();
        return (df.format(currDate));
    }

    /**
     * @return a String containing the current local time in ISO8601 Plus UTC
     *         Zed ("Complete date PLUS hours, minutes and seconds variant", as
     *         specified by the OAI protocol document) with dashes substituted
     *         for colons
     **/
    public static String getCurrTimeInDashedISO8601Plus() {
        SimpleDateFormat sdf = new SimpleDateFormat(ISO8601PlusDashedDateFormat);
        TimeZone utc = TimeZone.getTimeZone("GMT:00");
        sdf.setTimeZone(utc);
        Calendar calendar = Calendar.getInstance();
        Date currDate = calendar.getTime();
        return sdf.format(currDate);
    }

    /**
     * @return a String containing the current local time in ISO8601 Plus UTC
     *         Zed ("Complete date PLUS hours, minutes and seconds variant", as
     *         specified by the OAI protocol document)
     **/
    public static String getCurrTimeInISO8601SecondsUTC() {
        SimpleDateFormat sdf = new SimpleDateFormat(ISO8601SecondsUTCFormat);
        TimeZone utc = TimeZone.getTimeZone("GMT:00");
        sdf.setTimeZone(utc);
        Calendar calendar = Calendar.getInstance();
        Date currDate = calendar.getTime();
        return sdf.format(currDate);
    }

    /**
     * @return a String containing the current local time in ISO8601 Plus
     *         ("Complete date PLUS hours, minutes and seconds variant", as
     *         specified by the OAI protocol document)
     **/
    public static String getCurrTimeInISO8601Plus() {
        // Format the current time.
        SimpleDateFormat df = new SimpleDateFormat(ISO8601PlusDateFormat);
        Calendar calendar = Calendar.getInstance();
        Date currDate = calendar.getTime();
        String dateString = df.format(currDate);

        // deal with time zone offset
        int tzOffset = calendar.get(java.util.Calendar.ZONE_OFFSET);

        // check for daylight savings time
        TimeZone tz = calendar.getTimeZone();
        if (tz.inDaylightTime(currDate))
            tzOffset = tzOffset + calendar.get(java.util.Calendar.DST_OFFSET);

        // format the time zone offset
        String tzOffsetStr = new String();
        if (tzOffset == 0)
            tzOffsetStr = "Z";
        else {
            NumberFormat nf = new DecimalFormat("00");

            tzOffset = tzOffset / 60000;

            // ahead or behind UTC (Coordinated Universal Time)?
            String os = "+";
            if (tzOffset < 0)
                os = "-";

            int offsetHours = Math.abs(tzOffset / 60);
            tzOffsetStr = os + nf.format(offsetHours);

            int offsetMinutes = Math.abs(tzOffset % 60);
            tzOffsetStr = tzOffsetStr + ":" + nf.format(offsetMinutes);
        }

        String retString = dateString + tzOffsetStr;
        return (retString);
    }

    /**
     * @return a String containing the current local time in ISO8601 Plus UTC
     *         Zed ("Complete date PLUS hours, minutes and seconds variant", as
     *         specified by the OAI protocol document)
     **/
    public static String getISO8601SecondsUTCFromDate(Date inDate) {
        SimpleDateFormat sdf = new SimpleDateFormat(ISO8601SecondsUTCFormat);
        TimeZone utc = TimeZone.getTimeZone("GMT:00");
        sdf.setTimeZone(utc);
        return sdf.format(inDate);
    }

    /**
     * given a date as a string in the format specified,
     * 
     * @return a String in ISO8601 Plus
     *         ("Complete date PLUS hours, minutes and seconds variant", as
     *         specified by the OAI protocol document)
     **/
    public static String getISO8601PlusFromDate(String incomingDateStr,
            String incomingDateFormatStr) throws DateUtilityException {

        Date incomingDate = convertStringToDate(incomingDateStr,
                incomingDateFormatStr);

        SimpleDateFormat ISO8601DF = new SimpleDateFormat(ISO8601PlusDateFormat);

        String dateString = ISO8601DF.format(incomingDate);

        // deal with time zone offset
        Calendar calendar = Calendar.getInstance();
        int tzOffset = calendar.get(java.util.Calendar.ZONE_OFFSET);

        // check for daylight savings time
        TimeZone tz = calendar.getTimeZone();
        if (tz.inDaylightTime(incomingDate))
            tzOffset = tzOffset + calendar.get(java.util.Calendar.DST_OFFSET);

        // format the time zone offset
        String tzOffsetStr = new String();
        if (tzOffset == 0)
            tzOffsetStr = "Z";
        else {
            NumberFormat nf = new DecimalFormat("00");

            tzOffset = tzOffset / 60000;

            // ahead or behind UTC (Coordinated Universal Time)?
            String os = "+";
            if (tzOffset < 0)
                os = "-";

            int offsetHours = Math.abs(tzOffset / 60);
            tzOffsetStr = os + nf.format(offsetHours);

            int offsetMinutes = Math.abs(tzOffset % 60);
            tzOffsetStr = tzOffsetStr + ":" + nf.format(offsetMinutes);
        }

        String retString = dateString + tzOffsetStr;
        return (retString);
    }

    /**
     * given a date as a string in the format specified,
     * 
     * @return a String in ISO8601 (no hours ...)
     **/
    public static String getISO8601FromDate(String incomingDateStr,
            String incomingDateFormatStr) throws DateUtilityException {
        Date incomingDate = convertStringToDate(incomingDateStr,
                incomingDateFormatStr);

        SimpleDateFormat ISO8601DF = new SimpleDateFormat(ISO8601DateFormat);

        String retString = ISO8601DF.format(incomingDate);

        return (retString);
    }

    public static String getISO8601MilisecondFromDate(Date input) {
        SimpleDateFormat formatter = new SimpleDateFormat(
                ISO8601MilisecondsUTCFormat);
        return formatter.format(input);
    }

    /**
     * given a date as a String in ISO8601 Plus
     * ("Complete date PLUS hours, minutes and seconds variant", as specified by
     * the OAI protocol document)
     * 
     * @return a string in specified format
     **/
    public static String getDateFromISO8601Plus(String incomingDateStr,
            String outgoingDateFormatStr) throws DateUtilityException {

        Date incomingDate = convertStringToDate(incomingDateStr,
                ISO8601PlusDateFormat);

        SimpleDateFormat outgoingDF = new SimpleDateFormat(
                outgoingDateFormatStr);

        String retString = outgoingDF.format(incomingDate);

        return (retString);
    }

    /**
     * given a date as a String in ISO8601 (no hours ...)
     * 
     * @return a string in specified format
     **/
    public static String getDateFromISO8601(String incomingDateStr,
            String outgoingDateFormatStr) throws DateUtilityException {
        Date incomingDate = convertStringToDate(incomingDateStr,
                ISO8601DateFormat);

        SimpleDateFormat outgoingDF = new SimpleDateFormat(
                outgoingDateFormatStr);

        String retString = outgoingDF.format(incomingDate);

        return (retString);
    }

    /**
     * given a date as a String in ISO8601 Plus
     * ("Complete date PLUS hours, minutes and seconds variant", as specified by
     * the OAI protocol document)
     * 
     * @return a java Date object initialized appropriately
     **/
    public static Date getDateFromISO8601Plus(String incomingDateStr)
            throws DateUtilityException {
        return (convertStringToDate(incomingDateStr, ISO8601PlusDateFormat));
    }

    public static Date getDateFromISO8601Milisecond(String incoming)
            throws DateUtilityException {

        /*
         * Ick! Fedora uses varying significant digits, and SimpleDateFormat is
         * inflexible in that regard. So we need to deal with separate deci,
         * centi, mili, and plain second cases:
         * :10Z
         * :10.5Z
         * :10.51Z
         * :10.512Z
         */
        if (incoming.matches(".+?\\.[0-9][0-9]Z")) {
            /* 00.00Z */
            return convertStringToDate(incoming, ISO8601CentisecondsUTCFormat);
        } else if (incoming.matches(".+?\\.[0-9]Z")) {
            /* 00.0Z */
            return convertStringToDate(incoming, ISO8601DecisecondsUTCFormat);
        } else if (incoming.matches(".+?\\:\\d\\dZ$")){
            /* 00Z */
            return convertStringToDate(incoming, ISO8601SecondsUTCFormat);
        } else {
            /* 00.000Z */
            return convertStringToDate(incoming, ISO8601MilisecondsUTCFormat);
        }
    }

    /**
     * determine if the dateStr is in ISO8601 Plus UTC
     * ("Complete date PLUS hours, minutes and seconds variant", as specified by
     * the OAI protocol document)
     * 
     * @param dateStr
     * @return true if string is a date in the right format, false o.w.
     */
    public static boolean isISO8601SecondsUTCFormat(String dateStr) {
        try {
            convertStringToDate(dateStr, ISO8601SecondsUTCFormat);
            return true;
        } catch (DateUtilityException e) {
            return false;
        }
    }

    /**
     * @param elapsedTime
     *            - the number of seconds of elapsed time, as a double
     * @return a string in the form " x hours y minutes z seconds"
     */
    public static String getFormattedSeconds(double elapsedTime) {
        int seconds = (int) elapsedTime % 60;
        int minutes = (int) (elapsedTime / 60) % 60;
        int hours = (int) elapsedTime / (60 * 60);

        String msg = ((hours > 0) ? hours + " hours " : "")
                + ((minutes > 0) ? minutes + " minutes " : "")
                + ((seconds > 0) ? seconds + " seconds " : "");
        if (msg.length() == 0)
            msg = " less than 1 second";
        return msg;
    }

    /**
     * given a date as a String
     * 
     * @return a Date object in specified format
     **/
    private static Date convertStringToDate(String dateStr, String dateFormatStr)
            throws DateUtilityException {
        SimpleDateFormat theDateFormat = new SimpleDateFormat(dateFormatStr);
        theDateFormat.setLenient(true); // DON'T require strict adherence to
                                        // format

        Date theDate = null;
        try {
            theDate = theDateFormat.parse(dateStr);
        } catch (ParseException pe) {
            errmsg = "error parsing date: " + dateStr + " -- "
                    + pe.getMessage();
            throw new DateUtilityException(errmsg, pe);
        }

        return (theDate);
    }

} // End class DateUtility
