/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.util.security;

import java.security.KeyException;
import java.security.KeyManagementException;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;

/** Utility class for gen. key pairs with XML public and PEM private keys
 * 
 * @author birkland
 *
 */
public class AsymmetricKeyPair {
	private final KeyPair keyPair;
	private final KeyTranslator pemTranslator;
	private final KeyTranslator xmlTranslator;
    
    private static final String NSDL_AUTH_NAMESPACE = 
        "http://ns.nsdl.org/ndr/auth#";
    private static final String NSDL_AUTH_SCHEMA = 
        "http://ns.nsdl.org/schemas/ndr/ndrAuth_v1.0.0.xsd";
	
	public AsymmetricKeyPair(String algorithm, String keyName) 
			throws KeyManagementException {
		KeyPairGenerator keygen;
		try {
			keygen = KeyPairGenerator.getInstance(algorithm);
		} catch (NoSuchAlgorithmException e) {
			throw new KeyManagementException("Could not initialize key generator", e);
		}
		this.keyPair = keygen.generateKeyPair();
		this.pemTranslator = KeyTranslatorFactory.getInstance(algorithm, "PEM");
		this.xmlTranslator = KeyTranslatorFactory.getInstance(algorithm, "W3C");
	}
	
	public String getPublicKey() throws KeyException {
		return xmlTranslator.encodePublicKey(keyPair.getPublic());
	}
	
	public String getPrivateKey() throws KeyException {
		return pemTranslator.encodePrivateKey(keyPair.getPrivate());
	}
	
	public static void main(String[] args) {
		AsymmetricKeyPair pair;
        final String preamble = " xmlns:auth=\"" + NSDL_AUTH_NAMESPACE + "\" " + 
        "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"" +
        NSDL_AUTH_NAMESPACE + " " + NSDL_AUTH_SCHEMA + "\">\n";
		try {
			pair = new AsymmetricKeyPair("RSA", "Authorization: nsdl-1.0");
            System.out.print("<auth:PublicKey " + preamble +
                    pair.getPublicKey() + 
                    "</auth:PublicKey>\n\n");
			System.out.print(pair.getPrivateKey() + "\n");
		} catch (KeyException e) {
			System.err.print("Could not gey key!\n");
		}
	}
}
