/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.util.security;

import java.security.KeyManagementException;


/** Factory class for producing key translators
 * 
 * @see org.nsdl.util.security.KeyTranslatorFactory
 * 			.getInstance(String algorithm, String format)
 * @author birkland
 *
 */
public class KeyTranslatorFactory {
	
	/** Get a translator that will read and write a key to a specific format
	 * 
	 * Currently, translators exist that will transform RSA keys into
	 * W3C XML and PEM formats.  Therefore, the only acceptable
	 * value for &lt;algorithm&gt; is 'W3C', while both 
	 * 'PEM' and 'W3C' are accepted for &lt;format&gt;
	 * 
	 * @param algorithm Key algorithm. 
	 * @param format Key serialization format.
	 * @return KeyTranslator
	 * @throws KeyManagementException
	 */
	public static KeyTranslator getInstance(String algorithm, String format) 
		throws KeyManagementException {
		
		/* Right now, we only handle RSA keys */
		if (!algorithm.equalsIgnoreCase("RSA")) {
			throw new KeyManagementException
				("No key translator exists for algorithm " + algorithm);
		}
		
		/* Valid formats are PEM and W3C */
		if (format.equalsIgnoreCase("W3C")) {
			return new RSAtoXMLKeyTranslator();
		} else if (format.equalsIgnoreCase("PEM")) {
			return new RSAtoPEMKeyTranslator();
		} else {
			throw new KeyManagementException
				("Unknown key translation format " + format);
		}
	}
}
