/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.util.security;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.security.Key;
import java.security.KeyException;
import java.security.KeyFactory;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.spec.EncodedKeySpec;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;

import org.apache.axis.encoding.Base64;

/** Serializes RSA keys into PEM format.
 * 
 * Specifically, it uses PKCS#8 encoding private keys, 
 * X.509 for public.  These values are then ASN.1 and base64 encoded.
 * 
 * @author birkland
 *
 */
public class RSAtoPEMKeyTranslator implements KeyTranslator {
	
	public String encodePublicKey(PublicKey key)
			throws KeyException {
		return encodeRSAKey(key, KeyType.PUBLIC);
	}

	public String encodePrivateKey(PrivateKey key)
			throws KeyException {
		return encodeRSAKey(key, KeyType.PRIVATE);
	}

	public String encodePrivateKey(PrivateKey key,
			String passphrase) throws KeyException {
		/* TODO passphrase wrapping */
		return encodeRSAKey(key, KeyType.PRIVATE);
	}

	public PublicKey decodePublicKey(String source)
			throws KeyException {
		return (PublicKey) decodeRSAKey(source, KeyType.PUBLIC);
	}

	public PrivateKey decodePrivateKey(String source)
			throws KeyException {
		return (PrivateKey) decodeRSAKey(source, KeyType.PRIVATE);
	}

	public PrivateKey decodePrivateKey(String source,
			String passphrase) throws KeyException {
		/* TODO: passphrase wrapping */
		return (PrivateKey) decodeRSAKey(source, KeyType.PRIVATE);
	}
	
	private String encodeRSAKey (Key key, KeyType visibility) 
			throws KeyException {
		
		String keyString = null;
		EncodedKeySpec spec = null;
		try {
			KeyFactory factory = KeyFactory.getInstance(key.getAlgorithm());
			if (KeyType.PUBLIC.equals(visibility)) {
				spec = factory.getKeySpec(key, X509EncodedKeySpec.class);	
			} else if (KeyType.PRIVATE.equals(visibility)) {
				spec = factory.getKeySpec(key, PKCS8EncodedKeySpec.class);
			} else {
				throw new IllegalArgumentException
					("Bad key visibility: Expect either public or private");
			}
			
			keyString = Base64.encode(spec.getEncoded());

		} catch (NoSuchAlgorithmException e) {
			
		} catch (InvalidKeySpecException e) {
			
		}
		return 
		"-----BEGIN " + visibility.toString() + " KEY-----\n" +
		KeyStringFormatter.format(keyString, 0, 65) +
		"-----END "+ visibility.toString() + " KEY-----\n";
	}
	
	private Key decodeRSAKey(String encoded, KeyType visibility) 
			throws KeyManagementException{
		
		byte[] binaryContent = null;
		KeySpec spec;
		
		/* Read in the key */
		String line;
		StringBuilder base64Content = new StringBuilder();
		BufferedReader reader = new BufferedReader(new StringReader(encoded));
		try {
			while ((line = reader.readLine()) != null) {
				if (line.startsWith("-")) continue;
				base64Content.append(line);
			}
		} catch (IOException e) {
			throw new KeyManagementException("Could not read key", e);
		}
		
		/* Decode the key */
		binaryContent = Base64.decode(base64Content.toString());
		try {
			KeyFactory factory = KeyFactory.getInstance("RSA");
			if (visibility.equals(KeyType.PUBLIC)) {
				spec = new X509EncodedKeySpec(binaryContent);
				return factory.generatePublic(spec);
			} else if (visibility.equals(KeyType.PRIVATE)) {
				spec = new PKCS8EncodedKeySpec(binaryContent);
				return factory.generatePrivate(spec);
			} else {
				throw new KeyManagementException("Unknown key type");
			}
		} catch (NoSuchAlgorithmException e) {
			throw new KeyManagementException (e);
		} catch (InvalidKeySpecException e) {
			throw new KeyManagementException (e);
		}
	}
}
