/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.util.security;

import java.math.BigInteger;
import java.security.Key;
import java.security.KeyException;
import java.security.KeyFactory;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.security.spec.RSAPrivateKeySpec;
import java.security.spec.RSAPublicKeySpec;

import org.apache.axis.encoding.Base64;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/** Translates RSA keys to W3C XML format
 * 
 * See schema at:
 * http://www.w3.org/TR/xmldsig-core/xmldsig-core-schema.xsd
 * 
 * @author birkland
 *
 */
public class RSAtoXMLKeyTranslator implements KeyTranslator {

	private static final String KEY_NAMESPACE = 
			"http://www.w3.org/2000/09/xmldsig#";
	private static final String KEY_SCHEMA_LOCATION = 
			"http://www.w3.org/TR/xmldsig-core/xmldsig-core-schema.xsd";
	
	public String encodePublicKey(PublicKey key) 
			throws KeyException {
		return encodeRSAKey(key, KeyType.PUBLIC);
	}
	public String encodePrivateKey(PrivateKey key) 
			throws KeyException {
		return encodeRSAKey(key, KeyType.PRIVATE);
	}
	
	public String encodePrivateKey(PrivateKey key,  String passphrase) 
			throws KeyException {
		/* TODO: passphrase wrapping */
		return encodeRSAKey(key, KeyType.PRIVATE);
	}
	
	public PublicKey decodePublicKey(String source) 
			throws KeyException {
		return (PublicKey) decodeRSAKey(source, KeyType.PUBLIC);
	}
	
	public PrivateKey decodePrivateKey(String source) 
			throws KeyException {
		return (PrivateKey) decodeRSAKey(source, KeyType.PRIVATE);
	}
	
	public PrivateKey decodePrivateKey(String source, String passphrase) 
			throws KeyException {
		/* TODO: passphrase wraping */
		return (PrivateKey) decodeRSAKey(source, KeyType.PRIVATE);
	}
	
	private String encodeRSAKey(Key key, KeyType visibility) 
			throws KeyManagementException {
		String modulus;
		String exponent;
		try {
			KeyFactory factory = KeyFactory.getInstance(key.getAlgorithm());
			
			/* 
			 * toByteArray[] is big endian, so just base64 encoding it is OK as far 
			 * as the XML spec goes
			 */
			if (visibility.equals(KeyType.PUBLIC)) {
				RSAPublicKeySpec pubSpec = 
					factory.getKeySpec(key, RSAPublicKeySpec.class);
				modulus = 
					Base64.encode(pubSpec.getModulus().toByteArray());
				exponent = 
					Base64.encode(pubSpec.getPublicExponent().toByteArray());
			} else {
				RSAPrivateKeySpec privSpec = 
					factory.getKeySpec(key, RSAPrivateKeySpec.class);
				modulus = 
					Base64.encode(privSpec.getModulus().toByteArray());
				exponent = 
					Base64.encode(privSpec.getPrivateExponent().toByteArray());
			}
		} catch (InvalidKeySpecException e) {
			throw new KeyManagementException (e);
		} catch (NoSuchAlgorithmException e) {
			throw new KeyManagementException(e);
		}
		return 						  		
				" <KeyInfo " + 
					"xmlns=\"" + KEY_NAMESPACE + "\" " + 
					"xmlns:xsi = \"http://www.w3.org/2001/XMLSchema-instance\" " +
					"xsi:schemaLocation=\"" + KEY_NAMESPACE + " " + KEY_SCHEMA_LOCATION + "\">\n" +
				"	<KeyName> RSA Key </KeyName>\n" +
				"	<KeyValue>\n" +
				"		<RSAKeyValue>\n" +
				"			<Modulus>\n" + 
								KeyStringFormatter.format(modulus, 4) + 
				"			</Modulus>\n" +
				"			<Exponent>\n" + 
								KeyStringFormatter.format(exponent, 4) + 
				"			</Exponent>\n" +
				"		</RSAKeyValue>\n" +
				"	</KeyValue>\n" +
				" </KeyInfo>\n";
	}
	
	
	private Key decodeRSAKey(String xmlDataStream, KeyType visibility) 
			throws KeyManagementException {
		BigInteger exponent = null;
		BigInteger modulus = null;
		KeySpec spec;
		try {
			Document keyContainer =  NDRXMLutil.getDocument(xmlDataStream, true, false);
			
			NodeList keyContainerContents;
			
			if (keyContainer.getDocumentElement().getNamespaceURI().equals(KEY_NAMESPACE) &&
					keyContainer.getDocumentElement().getLocalName().equals("KeyInfo")) {
				
				/* If the XML key representation has KeyInfo as the document root.. */
				keyContainerContents = keyContainer.getChildNodes();
			} else {
				
				/* Otherwise, use the parent element */
				keyContainerContents = NDRXMLutil.getNodeByName(
						keyContainer.getDocumentElement(), KEY_NAMESPACE, "KeyInfo")
							.getParentNode().getChildNodes();
			}
			
			for (int i=0; i < keyContainerContents.getLength(); i++) {
				if (keyContainerContents.item(i).getNodeType() == Node.ELEMENT_NODE) {
					Element keyInfoElement = (Element) keyContainerContents.item(i);
						
					Element valueContainer = 
						NDRXMLutil.getChildElement(keyInfoElement, KEY_NAMESPACE, "KeyValue");
						
					Element value = 
						NDRXMLutil.getChildElement(valueContainer, KEY_NAMESPACE, "RSAKeyValue");
						
					String encodedModulus = 
						NDRXMLutil.getElementValue(
								NDRXMLutil.getChildElement(value, KEY_NAMESPACE, "Modulus"));
					modulus = new BigInteger(Base64.decode(encodedModulus));
						
					String encodedExponent = 
						NDRXMLutil.getElementValue(
								NDRXMLutil.getChildElement(value, KEY_NAMESPACE, "Exponent"));
					exponent = new BigInteger(Base64.decode(encodedExponent));
				}
			}
		} catch (Exception e) { /* XMLUtil only throws Exception classed Exceptions */
			throw new KeyManagementException("Could not read key from data stream", e);
		}
		/* If we don't have the requested key, return null */
		if (exponent == null && modulus == null) {return null;}
		
		try {
			KeyFactory factory = KeyFactory.getInstance("RSA");
			if (visibility.equals(KeyType.PUBLIC)) {
				spec = new RSAPublicKeySpec(modulus, exponent);
				return factory.generatePublic(spec);
			} else if (visibility.equals(KeyType.PRIVATE)) {
				spec = new RSAPrivateKeySpec(modulus, exponent);
				return factory.generatePrivate(spec);
			} else {
				throw new KeyManagementException("Unknown key type");
			}
		} catch (NoSuchAlgorithmException e) {
			throw new KeyManagementException (e);
		} catch (InvalidKeySpecException e) {
			throw new KeyManagementException (e);
		}
	}
}
