/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.util.xml;

import org.apache.log4j.Logger;
import org.apache.xerces.parsers.DOMParser;
import org.apache.xml.serialize.*;

import org.xml.sax.*;
import org.xml.sax.helpers.*;

import org.w3c.dom.*;
import org.w3c.dom.traversal.*;

import java.io.*;
import java.util.*;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

/**
 * provide XML encode and decode methods and XML parser feature strings
 * 
 */
public abstract class NDRXMLutil
{
  private static final Logger logger = Logger.getLogger(NDRXMLutil.class);


	//  private static final String ME = "NDRXMLutil";

  public static final String XSI_URI = "http://www.w3.org/2001/XMLSchema-instance";

  public static final String SCHEMA_NAMESPACE = "http://www.w3.org/2001/XMLSchema";

  /** SAX parser class name. */
  public static final String SAX_PARSER_CLASSNAME = "org.apache.xerces.parsers.SAXParser";

  // feature ids

  /** Namespaces feature id (http://xml.org/sax/features/namespaces). */
  public static final String NAMESPACES_FEATURE_ID = "http://xml.org/sax/features/namespaces";

  /**
   * SAX only!! Namespace-prefixes feature id
   * (http://xml.org/sax/features/namespace-prefixes).
   */
  public static final String NAMESPACE_PREFIXES_FEATURE_ID = "http://xml.org/sax/features/namespace-prefixes";

  /** Validation feature id (http://xml.org/sax/features/validation). */
  public static final String VALIDATION_FEATURE_ID = "http://xml.org/sax/features/validation";

  /**
   * Lexical handler property id
   * (http://xml.org/sax/properties/lexical-handler).
   */
  public static final String LEX_HANDLER_PROPERTY_ID = "http://xml.org/sax/properties/lexical-handler";

  /**
   * Dynamic validation feature id
   * (http://apache.org/xml/features/validation/dynamic). If true, only validate
   * if grammar (Schema) is provided.
   */
  public static final String DYNAMIC_VALIDATION_FEATURE_ID = "http://apache.org/xml/features/validation/dynamic";

  /**
   * Schema validation feature id
   * (http://apache.org/xml/features/validation/schema).
   */
  public static final String SCHEMA_VALIDATION_FEATURE_ID = "http://apache.org/xml/features/validation/schema";

  /**
   * Schema full checking feature id
   * (http://apache.org/xml/features/validation/schema-full-checking). If true,
   * check schemas in addition to instance docs.
   */
  public static final String SCHEMA_FULL_CHECKING_FEATURE_ID = "http://apache.org/xml/features/validation/schema-full-checking";

  /**
   * lex handler char refs feature id
   * (http://apache.org/xml/features/scanner/notify-char-refs).
   */
  public static final String NOTIFY_CHAR_REFS_FEATURE_ID = "http://apache.org/xml/features/scanner/notify-char-refs";

  /**
   * lex handler builtin refs feature id
   * (http://apache.org/xml/features/scanner/notify-builtin-refs).
   */
  public static final String NOTIFY_BUILTIN_REFS_ID = "http://apache.org/xml/features/scanner/notify-builtin-refs";

  /**
   * initialize a SAX parser (XMLReader)
   * 
   * @param validate -
   *          true if (deep) schema validation is to be turned on
   * @return the XMLReader instantiated with features set
   */
  public static XMLReader prepareSAXParser(XMLReader parser, boolean validate)
      throws SAXException
  {
    parser = XMLReaderFactory.createXMLReader(SAX_PARSER_CLASSNAME);
    parser = setSAXParserFeatures(parser, validate);
    return parser;
  }

  /**
   * set features for a SAX parser (XMLReader)
   * 
   * @param validate -
   *          true if (deep) schema validation is to be turned on
   * @return the XMLReader instantiated with features set
   */
  public static XMLReader setSAXParserFeatures(XMLReader parser,
      boolean validate) throws SAXException
  {
    // must use namespaces
    parser.setFeature(NAMESPACES_FEATURE_ID, true);
    // must use namespace prefixes
    parser.setFeature(NAMESPACE_PREFIXES_FEATURE_ID, true);
    parser.setFeature(VALIDATION_FEATURE_ID, validate);

    // the following are Apache Xerces features
    parser.setFeature(SCHEMA_VALIDATION_FEATURE_ID, validate);
    parser.setFeature(SCHEMA_FULL_CHECKING_FEATURE_ID, validate);
    parser.setFeature(DYNAMIC_VALIDATION_FEATURE_ID, validate);
    return parser;
  }

  /**
   * set features for a DOMParser, Xerces flavor
   * 
   * @param parser -
   *          features will be set for this DOMParser
   * @param validate -
   *          true if (deep) schema validation is to be turned on
   * @return the DOMParser with features set
   */
  public static DOMParser setDOMParserFeatures(DOMParser parser,
      boolean validate) throws SAXException
  {
    // must use namespaces
    parser.setFeature(NAMESPACES_FEATURE_ID, true);
    /*
     * this is a sax only feature
     * parser.setFeature(NAMESPACE_PREFIXES_FEATURE_ID, true);
     */
    parser.setFeature(VALIDATION_FEATURE_ID, validate);

    // the following are Apache Xerces features
    parser.setFeature(SCHEMA_VALIDATION_FEATURE_ID, validate);
    parser.setFeature(SCHEMA_FULL_CHECKING_FEATURE_ID, validate);
    parser.setFeature(DYNAMIC_VALIDATION_FEATURE_ID, validate);
    return parser;
  }

  /**
   * given a SAXParseException, create a message more suitable for humans
   * 
   * @param errorType -
   *          is it a warning, error or fatal error
   * @param ex
   *          the SAXParseException we're fiddlin' with
   * @return a string form of the message (more) suitable for human consumption
   */
  public static String niceMessage(String errorType, SAXParseException ex)
  {
    String systemID = ex.getSystemId();
    if (systemID != null)
    {
      int index = systemID.lastIndexOf(java.io.File.separator);
      if (index != -1) systemID = systemID.substring(index + 1);
    }

    String msg = "Parsing/Validation " + errorType + ":"
        + (systemID != null ? "" : systemID) + ":" + ex.getLineNumber() + ":"
        + ex.getColumnNumber() + ": " + ex.getMessage();
    return (msg);
  }

  /**
   * Return the first element in the NodeList
   * 
   * @param elementNodeList -
   *          a DOM NodeList containing the element(s) from which a text value
   *          will be extracted
   * @return the first element in the NodeList. Returns null if the
   *         elementNodeList has no elements
   */
  public static Element getFirstElement(NodeList elementNodeList)
  {
    if (elementNodeList == null) return null;

    int numElements = elementNodeList.getLength();
    if (numElements == 0)
      return null;
    else
    {
      for (int i = 0; i < numElements; i++)
      {
        Node node = elementNodeList.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE) return (Element) node;
      }
    }
    return null; // if we get here, there's no value available
  }

  /**
   * Return the child element specified by the params
   * 
   * @param element -
   *          a DOM Element containing elements from which a child element will
   *          be extracted
   * @param namespaceURI -
   *          the namespace URI for the desired child element, or null if no
   *          namespace is used
   * @param localname -
   *          the local (unqualified) name of the desired child element
   * @return the matching child element in the NodeList. Returns null if the
   *         elementNodeList has no matching elements.
   * @throws ClassCastException
   *           if elementNodeList has non-Element Nodes.
   */
  public static Element getChildElement(Element element, String namespaceURI,
      String localname)
  {
    if (element == null || localname == null) return null;

    NodeList childNodes = element.getChildNodes();
    int numChildren = childNodes.getLength();
    if (numChildren == 0)
      return null;
    else
    {
      for (int i = 0; i < numChildren; i++)
      {
        Node node = childNodes.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE)
        {
          if (namespaceURI == null)
          {
            if (node.getNodeName().equals(localname)) return (Element) node;
          }
          else if (node.getNamespaceURI().equals(namespaceURI))
              if (node.getLocalName().equals(localname)) return (Element) node;
        }
      }
    }
    return null; // if we get here, there's no value available
  }

  /**
   * Find the first element child of the passed parentElement
   * 
   * @param parentElement -
   *          a DOM Element
   * @return the FIRST element child of the parent elements. Returns null if the
   *         parent element is null or if there is no child element
   */
  public static Element getFirstChildElement(Element parentElement)
  {
    if (parentElement == null) return null;

    NodeList childList = parentElement.getChildNodes();
    for (int j = 0; j < childList.getLength(); j++)
    {
      Node child = childList.item(j);
      if (child.getNodeType() == Node.ELEMENT_NODE) return (Element) child;
    }
    return null;
  }

  /**
   * Find the first element child of the elements in the NodeList and return
   * that element
   * 
   * @param elementNodeList -
   *          a DOM NodeList containing element(s)
   * @return the FIRST element child of the elements in the NodeList. Returns
   *         null if the elementNodeList has no elements or if there is no child
   *         element for any elements in the list
   * @throws ClassCastException
   *           if elementNodeList has non-Element Nodes.
   */
  public static Element getFirstChildElement(NodeList elementNodeList)
  {
    if (elementNodeList == null) return null;

    int numElements = elementNodeList.getLength();
    if (numElements == 0)
      return null;
    else
    {
      for (int i = 0; i < numElements; i++)
      {
        // NOTE: this can cause a ClassCastException
        Element listElementNode = (Element) elementNodeList.item(i);

        NodeList childList = listElementNode.getChildNodes();
        for (int j = 0; j < childList.getLength(); j++)
        {
          Node child = childList.item(j);
          if (child.getNodeType() == Node.ELEMENT_NODE) return (Element) child;
        }
      }
    }
    return null;
  }

  /**
   * Find the last element child of the elements in the NodeList and return
   * that element
   * 
   * @param elementNodeList -
   *          a DOM NodeList containing element(s)
   * @return the LAST element child of the elements in the NodeList. Returns
   *         null if the elementNodeList has no elements or if there is no child
   *         element for any elements in the list
   * @throws ClassCastException
   *           if elementNodeList has non-Element Nodes.
   */
  public static Element getLastChildElement(NodeList elementNodeList)
  {
    if (elementNodeList == null) return null;

    int numElements = elementNodeList.getLength();
    
    if (numElements == 0)
      return null;
    else
    {
      Node lnodANode;
      
      for (int i = numElements; i > 0 ; i--)
      {             
        lnodANode = elementNodeList.item(i-1);                
        
        if (lnodANode!=null && lnodANode.getNodeType() == Element.ELEMENT_NODE)
        {
          return (Element) lnodANode;
        }        
      }
    }
    return null;
  }
  
  /**
   * return a list of values for the specified element name.
   * 
   * @param elementNodeList -
   *          a DOM NodeList containing the elements from which text values will
   *          be extracted
   * @return an ArrayList containing a value for each element in the elementList
   */
  public static ArrayList<String> getElementValues(NodeList elementNodeList)
  {
    if (elementNodeList == null) return null;

    int numElements = elementNodeList.getLength();
    ArrayList<String> elementValueList = new ArrayList<String>(numElements);
    for (int i = 0; i < numElements; i++)
    {
      // NOTE: this can cause a ClassCastException
      Element e = (Element) elementNodeList.item(i);
      elementValueList.add(NDRXMLutil.getElementValue(e));
    }
    return elementValueList;
  }

  /**
   * return a list all child elements
   * 
   * @param element -
   *          a DOM Element containing elements from which a child element will
   *          be extracted
   * @return an ArrayList containing the matching child elements
   */
  public static ArrayList<Element> getChildElements(Element element)
  {
    ArrayList<Element> childElementList = new ArrayList<Element>();

    NodeList childNodes = element.getChildNodes();
    int numChildren = childNodes.getLength();
    if (numChildren == 0)
      return null;
    else
    {
      for (int i = 0; i < numChildren; i++)
      {
        Node node = childNodes.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE)
        	childElementList.add((Element)node);
      }
    }
    return childElementList;
  }

  /**
   * return a list of child elements specified by the params
   * 
   * @param element -
   *          a DOM Element containing elements from which a child element will
   *          be extracted
   * @param namespaceURI -
   *          the namespace URI for the desired child element, or null if no
   *          namespace is used
   * @param localname -
   *          the local (unqualified) name of the desired child element
   * @return an ArrayList containing the matching child elements
   */
  public static ArrayList<Element> getChildElements(Element element, String namespaceURI,
	      String localname)
  {
    ArrayList<Element> childElementList = new ArrayList<Element>();

    NodeList childNodes = element.getChildNodes();
    int numChildren = childNodes.getLength();
    if (numChildren == 0)
      return null;
    else
    {
      for (int i = 0; i < numChildren; i++)
      {
	      Node node = childNodes.item(i);
	      if (node.getNodeType() == Node.ELEMENT_NODE)
	      {
	        if (namespaceURI == null)
	        {
	          if (node.getNodeName().equals(localname))
	          	childElementList.add((Element)node);
	        }
	        else if (node.getNamespaceURI().equals(namespaceURI))
	            if (node.getLocalName().equals(localname))
		          	childElementList.add((Element)node);
	      }
      }
    }
    return childElementList;
  }

  /**
   * return the first child element that matches the specified xpath
   * 
   * @param element -
   *          a DOM Element containing elements from which a child element will
   *          be extracted
   * @param xpath -
   *          fully qualified xpath (NOTE: wildcards are NOT allowed)
   * @return null if any part of xpath cannot be found; otherwise, returns the element for the end of the xpath
   */
  public static Element getElementAtXPath(Element startingElement, String xpath)
  {
  	String[] xpathParts = xpath.split("/");
	
	// find parent of final node in xpath
	Element currentElement = startingElement;
	for( int x = 0; x < xpathParts.length; ++x )
	{
		String nodeName = xpathParts[x];
		String[] nodeNameParts = nodeName.split(":");

		String localName = null;
		String namespaceURI = null;
		if( nodeNameParts.length == 1 ) 
			localName    = nodeNameParts[0];
		else 
		{
    		namespaceURI = nodeNameParts[0];
    		localName    = nodeNameParts[1];
		}
		
		currentElement = NDRXMLutil.getChildElement( currentElement, namespaceURI, localName);
		
		if( currentElement == null ) 
			// couldn't find a node on the xpath
			return null;
	}
		
	return currentElement;
  }
  
  /**
   * return a list of child elements specified by the xpath
   * 
   * @param element -
   *          a DOM Element containing elements from which a child element will
   *          be extracted
   * @param xpath -
   *          fully qualified xpath (NOTE: wildcards are NOT allowed)
   * @return an ArrayList containing matching child elements
   */
  public static ArrayList<Element> getElementsAtXPath(Element startingElement, String xpath)
  {
  	String[] xpathParts = xpath.split("/");

	// find parent of final node in xpath
	Element prevElement = startingElement;
	for( int x = 0; x < xpathParts.length - 1; ++x )
	{
		String nodeName = xpathParts[x];
		prevElement = NDRXMLutil.getChildElement( prevElement, null, nodeName);
		
		if( prevElement == null ){ 
			// couldn't find a node on the xpath
			return null;
		}
	}

	// use the last part of the xpath and get it's children's values
	String nodeName = xpathParts[xpathParts.length-1];
	ArrayList<Element> elements = NDRXMLutil.getChildElements(prevElement, null, nodeName);
	
	return elements;
  }
  
  /**
   * return a list of values for child elements specified by the xpath
   * 
   * @param element -
   *          a DOM Element containing elements from which a child element will
   *          be extracted
   * @param xpath -
   *          fully qualified xpath (NOTE: wildcards are NOT allowed)
   * @return an ArrayList containing a value for each matching child element
   */
  public static ArrayList<String> getElementValuesAtXPath(Element startingElement, String xpath)
  {
  	String[] xpathParts = xpath.split("/");

	// find parent of final node in xpath
	Element prevElement = startingElement;
	for( int x = 0; x < xpathParts.length - 1; ++x )
	{
		String nodeName = xpathParts[x];
		prevElement = NDRXMLutil.getChildElement( prevElement, null, nodeName);
		
		if( prevElement == null ){ 
			// couldn't find a node on the xpath
			return null;
		}
	}

	// use the last part of the xpath and get it's children's values
	String nodeName = xpathParts[xpathParts.length-1];
	ArrayList<String> values = NDRXMLutil.getChildElementValues(prevElement, null, nodeName);
	
	return values;
  }
  
  /**
   * Find the first element in the NodeList that has a value (has a text child,
   * in DOM parlance), and return that value.
   * 
   * @param elementNodeList -
   *          a DOM NodeList containing the element(s) from which a text value
   *          will be extracted
   * @return the value of the FIRST text child of the FIRST element in the
   *         NodeList that has a value. Returns an empty string if there is no
   *         text child, null if the elementNodeList has no elements
   * @throws ClassCastException
   *           if elementNodeList has non-Element Nodes.
   */
  public static String getFirstElementValue(NodeList elementNodeList)
  {
    if (elementNodeList == null) return null;

    int numElements = elementNodeList.getLength();
    if (numElements == 0)
      return null;
    else
    {
      for (int i = 0; i < numElements; i++)
      {
        // NOTE: this can cause a ClassCastException
        Element elementNode = (Element) elementNodeList.item(i);
        String textChildStr = getElementValue(elementNode);
        if (textChildStr != null) return textChildStr;
      }
    }
    return null; // if we get here, there's no value available
  }

  /**
   * Return the concatenated value of the all text children of the element.
   * 
   * @param element -
   *          a DOM Element
   * @return the concatenated value of the all text children of the element.
   *         Returns null if there is no text child
   */
  public static String getElementValue(Element element)
  {
    if (element == null) return null;

    boolean gotValue = false;
    String textChildStr = new String();

    NodeList childList = element.getChildNodes();
    for (int i = 0; i < childList.getLength(); i++)
    {
      Node child = childList.item(i);
      if (child.getNodeType() == Node.TEXT_NODE)
      {
        textChildStr = textChildStr + child.getNodeValue();
        gotValue = true;
      }
    }

    if (gotValue)
      return textChildStr;
    else
      return null;
  }

  /**
   * Replace the element's value with the passed value, (if the element already
   * had a value), or add new value
   * 
   * @param element -
   *          a DOM Element
   * @param newValueStr -
   *          the new value for the element
   * @return the element with a single text child containing the new value.
   */
  public static Element setElementValue(Element element, String newValueStr)
  {
    if (element == null) return null;

    NodeList childList = element.getChildNodes();
    boolean foundOne = false;
    for (int j = 0; j < childList.getLength(); j++)
    {
      Node child = childList.item(j);
      if (child.getNodeType() == Node.TEXT_NODE)
      {
        if (!foundOne)
        {
          child.setNodeValue(newValueStr);
          foundOne = true;
        }
        else
          // only want one text child after this method completes
          // 9-10-04 the next node in the list is found with nextSibling(), so
          // removing
          //  a node means subsequent nodes can't be found, even though they're
          // there.
          //          element.removeChild(child);
          child.setNodeValue(null);
      }
    }
    
    if (!foundOne) {
    	Text newValue = element.getOwnerDocument().createTextNode(newValueStr);
    	element.appendChild(newValue);
    }
    return element;
  }

  /**
   * @param elementNode
   *          a DOM Element
   * @param attrNamespaceURI
   *          the namespace URI for the desired attribute, or null if no
   *          namespace is used
   * @param attrLocalName
   *          the local (unqualified) name of the desired attribute
   * @return a string containing the value of the attribute, or null if there is
   *         no such attribute.
   */
  public static String getAttributeValue(Element elementNode,
      String attrNamespaceURI, String attrLocalName)
  {
    if (elementNode == null) return null;

    NamedNodeMap nnm = elementNode.getAttributes();
    Node attrNode = null;
    if (attrNamespaceURI == null)
      attrNode = nnm.getNamedItem(attrLocalName);
    else
      attrNode = nnm.getNamedItemNS(attrNamespaceURI, attrLocalName);

    if (attrNode != null)
      return attrNode.getNodeValue();
    else
      return null;
  }

  /**
   * returns the schemaLocation for the indicated namespaceURI if it is in scope
   * for the indicated Element.
   * 
   * @param element -
   *          a DOM Element
   * @param namespaceURI -
   *          the URI of the namespace for which we are seeking a schema
   *          location, or null if we are to find the schema location for no
   *          namespace
   * @return the schema location as a String (expected value is a URL), or null
   *         if there is no such schema location provided for this element.
   */
  public static String getSchemaLocation(Element element, String namespaceURI)
  {
    String schemaLoc = null;
    if (namespaceURI != null)
    {
      String schemaLocAttrValue = NDRXMLutil.getAttributeValue(element, XSI_URI,
          "schemaLocation");
      if (schemaLocAttrValue != null)
      {
        StringTokenizer slst = new StringTokenizer(schemaLocAttrValue);
        while (slst.hasMoreTokens())
        {
          String tok = slst.nextToken();
          if (tok.equals(namespaceURI))
          {
            schemaLoc = slst.nextToken();
            break;
          }
        }
      }
    }
    else
      // if no namespace, schema should be in xsi:noNamespaceSchemaLocation
      schemaLoc = NDRXMLutil.getAttributeValue(element, XSI_URI,
          "noNamespaceSchemaLocation");

    return (schemaLoc);
  }

  /**
   * write a serialized version of the passed document to the passed file will
   * NOT overwrite file if it already exists
   * 
   * @param doc
   *          the DOM document to be serialized
   * @param absFileName
   *          the full (absolute) file name for the serialized document
   * @return true if the file was successfully written, false otherwise
   */
  public static boolean serializeDOMdoc(Document doc, String absFileName)
      throws IOException
  {
    return (serializeDOMdoc(doc, absFileName, false));
  }

  /**
   * write a serialized version of the passed document to the passed file
   * 
   * @param doc
   *          the DOM document to be serialized
   * @param absFileName
   *          the full (absolute) file name for the serialized document
   * @param overwrite -
   *          true if file is to be overwritten, false if not
   * @return true if the file was successfully written, false otherwise
   */
  public static boolean serializeDOMdoc(Document doc, String absFileName,
      boolean overwrite) throws IOException
  {
    File outFile = new File(absFileName);

    if (!outFile.createNewFile()) // file already exists
        if (!overwrite) return false;

    OutputFormat formatter = new OutputFormat(doc); // FIXME - is doc useful
    // here?
    formatter.setIndent(2); // set indenting to be 2 spaces instead of 4
    formatter.setPreserveSpace(true);

    FileOutputStream fos = new FileOutputStream(outFile);
    OutputStreamWriter utf8Writer = new OutputStreamWriter(fos, "UTF-8");
    XMLSerializer serializer = new XMLSerializer(utf8Writer, formatter);
    serializer.serialize(doc);

    utf8Writer.close();
    fos.close();
    return true;
  }

  /**
   * write a serialized version of the passed document to a string. Note that
   * character encodings may not be correct.
   * 
   * @param doc
   *          the DOM document to be serialized
   * @return a String containing the DOM document
   */
  public static String serializeDOMdoc(Document doc) throws IOException
  {
    StringWriter sout = new StringWriter();

    OutputFormat formatter = new OutputFormat(doc); // FIXME - is doc useful
    // here?
    formatter.setIndenting(true);
    formatter.setLineWidth(72);
    formatter.setIndent(2); // set indenting to be 2 spaces instead of 4
    //formatter.setPreserveSpace(true);
    formatter.setEncoding("UTF-8");  // is the default

    XMLSerializer serializer = new XMLSerializer(sout, formatter);
    serializer.serialize(doc);
    return sout.toString();
  }

  /**
   * write a serialized version of the passed element to a string. Note that
   * character encodings may not be correct.
   * 
   * @param node
   *          the Element to be serialized
   * @return a String containing the DOM document
   */
  public static String serializeDOMElement(Element node) throws IOException
  {
    StringWriter sout = new StringWriter();

    OutputFormat formatter = new OutputFormat();
    formatter.setIndenting(true);
    formatter.setLineWidth(72);
    formatter.setIndent(2); // set indenting to be 2 spaces instead of 4
    //formatter.setPreserveSpace(true);
    formatter.setEncoding("UTF-8");  // is the default

    XMLSerializer serializer = new XMLSerializer(sout, formatter);
    serializer.serialize(node);
    return sout.toString().replaceAll("\\s*<\\?xml.*?\\?>\\s*", "");
  }

  /**
   * unescapes an xml encoded string adapted from
   * http://kelp.enhydra.org/software/cvs/cvsweb.cgi/Projects/EnhydraOrg/toolsTech/kXML/src/org/kxml/Xml.java?rev=1.1&content-type=text/x-cvsweb-markup
   */
  public static String XMLdecode(String cooked)
  {
    StringBuffer result = new StringBuffer();
    int i0 = 0;
    while (true)
    {
      // i1 is index of "&"
      int i1 = cooked.indexOf('&', i0);
      if (i1 == -1) break; // if there are no ampersands, we're done.

      result.append(cooked.substring(i0, i1));

      // i2 is index of ";" after "&"
      int i2 = cooked.indexOf(';', i1);
      if (i2 == -1)
      {
        // there's no semicolon after this ampersand
        //   so it should have been an XML parsing error
        //   but we'll leave it alone here!
        result.append("&");
        // orig code was:
        //        result.append ("&amp;");
        i0 = i1 + 1;
        break;
      }

      // decode is what's between "&" and the next ";"
      String decode = cooked.substring(i1 + 1, i2);
      if (decode.startsWith("#x"))
        result.append((char) Integer.parseInt(decode.substring(2), 16));
      else if (decode.startsWith("#"))
        result.append((char) Integer.parseInt(decode.substring(1)));
      else if (decode.equals("lt"))
        result.append('<');
      else if (decode.equals("gt"))
        result.append('>');
      else if (decode.equals("quot"))
        result.append('"');
      else if (decode.equals("amp"))
        result.append('&');
      else if (decode.equals("apos"))
        result.append('\'');
      else
        throw new RuntimeException("illegal character entity: &" + decode + ";");

      i0 = i2 + 1;
    }
    result.append(cooked.substring(i0));

    return result.toString();
  }

  /**
   * XML encodes a string. Assumes all character codes greater than 127 are to
   * be encoded
   * 
   * @param raw -
   *          the string to be XML encoded
   */
  public static String XMLencode(String raw)
  {
    return XMLencode(raw, true);
  }

  /**
   * XML encodes a string adapted from
   * http://kelp.enhydra.org/software/cvs/cvsweb.cgi/Projects/EnhydraOrg/toolsTech/kXML/src/org/kxml/Xml.java?rev=1.1&content-type=text/x-cvsweb-markup
   * 
   * @param raw -
   *          the string to be XML encoded
   * @param encodeAbove127 -
   *          true if all character codes greater than 127 are to be encoded
   */
  public static String XMLencode(String raw, boolean encodeAbove127)
  {
    int len = raw.length();

    StringBuffer cooked = new StringBuffer(raw.length());

    for (int i = 0; i < len; i++)
    {
      char c = raw.charAt(i);

      switch (c)
      {
        case '<':
          cooked.append("&lt;");
          break;
        case '>':
          cooked.append("&gt;");
          break;
        case '&':
          cooked.append("&amp;");
          break;
        case '\'':
          cooked.append("&apos;");
          break;
        case '"':
          cooked.append("&quot;");
          break;
        default:
          if (c >= 128 && encodeAbove127)
            cooked.append("&#" + ((int) c) + ";");
          else
            cooked.append(c);
      }
    }
    return cooked.toString();
  }


  
  /** Get node by Qualified name
   * 
   * @param startNode Node to search in
   * @param name Qualified name of a child element to find
   * @return matching Node
   */
  public static Node getNodeByName(Node startNode, QName name) {
	  DocumentTraversal trav;
	  NodeIterator iterator;
	  Node child = null;
	  
      /* XXX It's dangerous to assume that this cast is legit */
	  trav = (DocumentTraversal) startNode.getOwnerDocument();
      if (trav == null) {
          trav = (DocumentTraversal) startNode;
      }
	  
	  iterator = trav.createNodeIterator(startNode, NodeFilter.SHOW_ELEMENT, null, false);
	 
	  while((child = iterator.nextNode()) != null) {
	      QName childName = new QName(child.getNamespaceURI(), child.getLocalName());
	      
	      if (name.equals(childName)) {
			  return child;
		  }
	  }
	  
	  return null;
  }
  
  
  public static Node getNodeByName(Collection<Node> nodes, QName name) {

  	//  fails for added elements
	  for (Node item : nodes) 
	  {
      if (name.getNamespaceURI().equals(item.getNamespaceURI()) &&
      		name.getLocalPart().equals(item.getLocalName()))
      {
      	return item;
      }
	  }
	  
	  return null;
  }
  
  
  /** Get node by qualified name
   * 
   * @param startNode Node to search in
   * @param localName Local name of element to find
   * @param namespaceURI Namespace of element to find
   * @return matching Node
   */
  public static Node getNodeByName(Node startNode, String namespaceURI, String localName) {
	  return getNodeByName(startNode, new QName(namespaceURI, localName));
  }
  
  public static Node getNodeByName(Collection<Node> nodes, String namespaceURI, String localName) {
	  return getNodeByName(nodes, new QName(namespaceURI, localName));
  }

  /**
   * Removes first element node with the specified name
   */
  public static void removeNodeByName(Node lnodStartNode, String ElementName)
  {
    DocumentTraversal lobjTrav;
    NodeIterator lobjIterator;
    Node lnodChild = null;

    lobjTrav = (DocumentTraversal) lnodStartNode.getOwnerDocument();

    lobjIterator = lobjTrav.createNodeIterator(lnodStartNode,
        NodeFilter.SHOW_ELEMENT, null, false);

    while ((lnodChild = lobjIterator.nextNode()) != null)
    {
      if (lnodChild.getNodeName().equals(ElementName))
      {
        lnodChild.getParentNode().removeChild(lnodChild);
        break;
      }
    }
  }
  
  
 
  /** Returns the first element with the specified qualified name and value
   * 
   * This the namespace-aware version getNodeByNameAndValue()
   * 
   * @param startNode The node that contains the child elements to search through
   * @param name The qualified name of the element to find
   * @param value String value of the element to match
   * @return The first matching Node
   */
  public static Node getNodeByNameAndValue(Node startNode, QName name, String value) {
	  DocumentTraversal trav;
	  NodeIterator iterator;
	  Node child = null;
	
	  trav = (DocumentTraversal) startNode.getOwnerDocument();
	
	  iterator = trav.createNodeIterator(startNode,
	      NodeFilter.SHOW_ELEMENT, null, false);
	
	  while ((child = iterator.nextNode()) != null)
	  {
	    if (child.getLocalName().equals(name.getLocalPart()) &&
	  		  child.getNamespaceURI().equals(name.getNamespaceURI()) &&
	        getElementValue((Element) child).equalsIgnoreCase(value)
	        ) break;
	  }
	
	  return child;
  }
  
  /**
   * returns the first element node with the specified attribute name & value
   */
  public static Node getNodeByAttribute(Node lnodStartNode,
      String AttributeName, String AttributeValue)
  {
    DocumentTraversal lobjTrav;
    NodeIterator lobjIterator;
    Node lnodChild = null;
    String lstrVal;

    lobjTrav = (DocumentTraversal) lnodStartNode.getOwnerDocument();

    lobjIterator = lobjTrav.createNodeIterator(lnodStartNode,
        NodeFilter.SHOW_ALL, null, false);

    while ((lnodChild = lobjIterator.nextNode()) != null)
    {
      //  Check of there are any attributes for this node - only way to tell.
      if (lnodChild.getNodeType() == Node.ELEMENT_NODE)
      {
        lstrVal = getAttributeValue((Element) lnodChild, null, AttributeName);
        if (lstrVal != null && lstrVal.equalsIgnoreCase(AttributeValue)) break;
      }
    }

    return lnodChild;
  }

  /**
   * Creates a Document using the JaxP factory methods and the input string.
   * NamespaceAware and Validating are booleans to set these for the parse.
   */
  public static Document getDocument(String lstrInputXML,
      boolean NamespaceAware, boolean Validating) throws Exception
  {
      return getDocument(
              new ByteArrayInputStream(
                      lstrInputXML.getBytes("UTF-8")), 
                      NamespaceAware, 
                      Validating);
  }

  /**
   * Creates a Document using the JaxP factory methods and the input string.
   * NamespaceAware and Validating are booleans to set these for the parse.
   */
  public static Document getDocument(InputStream stream,
      boolean NamespaceAware, boolean Validating) throws Exception
  {
    DocumentBuilderFactory factory = null;
    DocumentBuilder builder = null;
    Document ldocDoc = null;

    try
    {
      // use jaxp methods to load a local DOM thingy
        factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(NamespaceAware);
        factory.setValidating(Validating);

        builder = factory.newDocumentBuilder();

      ldocDoc = builder.parse(stream);
    }
    catch (Exception e)
    {
      throw new Exception("Error constructing DOM from XML string."
          + e.getClass().getName() + " : " + e.getMessage());
    }

    return ldocDoc;
  }
  
  /**
   * Returns the input XML string as a node in the document space of the 
   * document parameter.
   */
  public static Node getStringAsNode(String lstrInputXML,
      Document ldocParent) throws Exception
  {
 
    DocumentBuilderFactory factory = null;
    DocumentBuilder builder = null;

    Document ldocDoc = null;    
    Node lnodTemp = null;

    try
    {
      // use jaxp methods to load a local DOM thingy
      if (ldocDoc == null)
      {
        factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        factory.setValidating(false);

        builder = factory.newDocumentBuilder();
      }

      //  Build a new document from the string
      Document ldocTemp = builder.parse(new ByteArrayInputStream(
          lstrInputXML.getBytes("UTF-8")));

     // Document ldocTemp = builder.parse(new InputSource(new StringReader(lstrInputXML)));

      //  Turn it into a node in the parent document space.
      lnodTemp = ldocParent.importNode(ldocTemp.getDocumentElement(),true);
      
    }
    catch (Exception e)
    {
      throw new Exception(
          "Error constructing Node from XML string. "
              + e.getClass().getName() + " : " + e.getMessage());
    }

    return lnodTemp;
  }
  
  public static void declareNamespace(Element targetNode, String namespaceURI, String prefix) 
    throws DOMException {
      declareNamespace(targetNode, namespaceURI, prefix, false);
  }

  public static void declareNamespace(Element targetNode, String namespaceURI, String prefix, boolean force) 
  		throws DOMException
  {
	  
	  if (targetNode.lookupNamespaceURI(prefix) == null || force) {
		  /* if the given prefix is unused, then we can use it */
		  
		  Attr newNamespace = targetNode.getOwnerDocument().createAttribute("xmlns:" + prefix);
		  newNamespace.setNodeValue(namespaceURI);
		  targetNode.setAttributeNode(newNamespace);
	  } else if (targetNode.lookupNamespaceURI(prefix).equals(namespaceURI)) {
		  /* Our prefix has already been declared, so we need not do anything*/
		  return;
	  } else {
		  /* The given prefix is being used for a different namespace.  */
		  
		  /* XXX Right now, we just throw an exception.  Perhaps a better thing to do 
		   * would be to see if there are any child elements that are in the existing
		   * prefix's namespace.  If not, then we can probably safely re-declare it
		   */
		  
		  throw new DOMException(DOMException.NAMESPACE_ERR, "Another namespace is already in " +
				  "use for prefix " + prefix);
	  }
  }

  /**
   * Get the value of an element that is a child of another element in an xml document.
   *
   * @param element -
   *          a DOM Element containing elements from which a child element will
   *          be extracted
   * @param namespaceURI -
   *          the namespace URI for the desired child element, or null if no
   *          namespace is used
   * @param localname -
   *          the local (unqualified) name of the desired child element
   * 
   * @return value of the child element
   */
  public static String getChildElementValue(Element element, String namespaceURI,
	      String localname)
  {
  	Element childElement = getChildElement(element,namespaceURI,localname);
  	if( childElement == null ) return null;

  	String childElementValue = getElementValue(childElement);
  	return childElementValue;
  }

  /**
   * return a list of values for all child elements specified by the params
   * 
   * @param element -
   *          a DOM Element containing elements from which a child element will
   *          be extracted
   * @param namespaceURI -
   *          the namespace URI for the desired child element, or null if no
   *          namespace is used
   * @param localname -
   *          the local (unqualified) name of the desired child element
   * @return an ArrayList containing a value for each matching child element
   */
  public static ArrayList<String> getChildElementValues(Element element, String namespaceURI,
	      String localname)
  {
    if (element == null || localname == null) return null;

    ArrayList<String> elementValueList = new ArrayList<String>();

    NodeList childNodes = element.getChildNodes();
    int numChildren = childNodes.getLength();
    if (numChildren == 0)
    {
    	return null;
    } 
    else
    {
      for (int i = 0; i < numChildren; i++)
      {
        Node node = childNodes.item(i);
        if (node.getNodeType() == Node.ELEMENT_NODE)
        {
          if (namespaceURI == null)
          {
            if (localname.equals(node.getNodeName()))
            	elementValueList.add(NDRXMLutil.getElementValue((Element)node));
          }
          else if (namespaceURI.equals(node.getNamespaceURI())){
              if (localname.equals(node.getLocalName())) 
              	  elementValueList.add(NDRXMLutil.getElementValue((Element)node));
          }
        }
      }
    }
    return elementValueList; 
  }
}
