/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.util.xml;

import java.io.StringReader;
import java.util.HashSet;
import javax.xml.stream.*;
import static javax.xml.stream.XMLStreamConstants.*;

public class XMLStreamComparator {
	public static boolean equals(XMLStreamReader a, XMLStreamReader b)  throws XMLStreamException {
		do {
			getRelevant(a);
			getRelevant(b);
			if (a.getEventType() != b.getEventType()) {
				return false;
			}
			
			switch (a.getEventType()) {
				case START_ELEMENT: {
					if (!compareElement(a, b)) {return false;}
					if (!compareAttributes(a, b)) {return false;}
					if (!compareNS(a, b)) {return false;}
				} break;
				case CHARACTERS: {
					if (!compareText(a, b)) {return false;}
				} break;
				case ENTITY_REFERENCE: {
					if (!compareText(a, b)) {return false;}
				} break;
				case CDATA: {
					if (!compareText(a, b)) {return false;}
				} break;
				case END_ELEMENT: {
					if (!compareElement(a, b)) {return false;}
				} break;
			}
		} while (a.getEventType() != END_DOCUMENT);
		return true;
	}
	
	private static boolean compareElement(XMLStreamReader a, XMLStreamReader b)
			throws XMLStreamException {
		if (!a.getName().equals(b.getName())) {return false;}
		return true;
	}
	
	private static boolean compareAttributes(XMLStreamReader a, XMLStreamReader b)
		throws XMLStreamException {
		if (a.getAttributeCount() != b.getAttributeCount()) {return false;}
		
		for (int i = 0; i < a.getAttributeCount(); i++) {
			if (!a.getAttributeName(i).equals(b.getAttributeName(i))) {return false;}
			if (!a.getAttributeValue(i).equals(b.getAttributeValue(i))) {return false;}
		}
		return true;
	}
	
	private static boolean compareNS(XMLStreamReader a, XMLStreamReader b) 
		throws XMLStreamException {
		if (a.getNamespaceCount() != b.getNamespaceCount()) {return false;}
		
		HashSet<String> aNS = new HashSet<String>();
		HashSet<String> bNS = new HashSet<String>();
		for (int i = 0; i < a.getNamespaceCount(); i++) {
			aNS.add(a.getNamespaceURI(i));
			bNS.add(b.getNamespaceURI(i));
		}
		if (!aNS.containsAll(bNS)) {return false;}
		return true;
	}
	
	private static boolean compareText(XMLStreamReader a, XMLStreamReader b) 
		throws XMLStreamException {
		
		/* XXX There may be a slight vulnerability here in that the parser is free to 
		 * choose how much text to return.  We are assuming that it returns the same chunk
		 * size for both files, but there's no guarantee of that... 
		 */
		
		if (!a.getText().equals(b.getText())) {return false;}
		return true;
	}
	
	private static int getRelevant(XMLStreamReader r) throws XMLStreamException {
		while(true) {
			int event = r.next();
			if (event == START_DOCUMENT || event == END_DOCUMENT ||
				event == START_ELEMENT || event == END_ELEMENT ||
				event == CHARACTERS || event == ENTITY_REFERENCE || event == CDATA) {
				
				if (event != CHARACTERS) {
					return event;
				} else {
					if (r.getText().trim().length() > 0) {
						return event;
					}
				}
			}
		}
	}
	
	/* Just as an example.. */
	public static void main(String[] args) throws XMLStreamException {
		final String xml1 = 
			"<test><one>whitespace test</one><two this=\"that\"></two></test>";
		final String xml2 = "   <test>    \n" +
				"<one>whitespace test</one>\n\n" +
				"<two \n this=\"that\"></two> \n </test>";
		
		final String xml3 = "<test><one>Whitespace Test</one><two this=\"that\"></two>" +
			"</test>";
		
		/* Setting up the factory is expensive, so keep it persistent and share, if
		 * possible
		 */
		XMLInputFactory i = XMLInputFactory.newInstance();
		
		System.out.print("xml1\n------\n" + xml1 + "\n-----\n\n");
		System.out.print("xml2\n------\n" + xml2 + "\n-----\n\n");
		System.out.print("xml3\n------\n" + xml3 + "\n-----\n\n");
		
		/* Creating XMLStreamReaders from strings */
		XMLStreamReader xmlReader1 = i.createXMLStreamReader(new StringReader(xml1));
		XMLStreamReader xmlReader2 = i.createXMLStreamReader(new StringReader(xml2));
		if (equals(xmlReader1, xmlReader2)) {
			System.out.print("xml 1 and 2 are equal\n");
		}
		
		/* Note: Streams can only be read/processed once, hence we need to create a new
		 * stream for xml1
		 */
		xmlReader1 = i.createXMLStreamReader(new StringReader(xml1));
		XMLStreamReader xmlReader3 = i.createXMLStreamReader(new StringReader(xml3));
		if (!equals(xmlReader1, xmlReader3)) {
			System.out.print("xml 2 and 3 are not\n");
		}
		System.out.print("Done!\n");
 	}
}
