/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access.filters;

import java.util.Arrays;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;

import static junit.framework.Assert.*;
import static org.nsdl.repository.model.content.AggregatorModel.Components.AGGREGATOR_FOR;
import static org.nsdl.repository.model.content.AuthorizationModel.Grouping.Components.AUTHORIZED;
import static org.nsdl.repository.model.content.MetadataProviderModel.Components.METADTA_PROVIDER_FOR;

public class DefaultAuthorizationFilterTest {

    private static NDRAccess ndr;
    private static NDRObject agent;
    private static DefaultAuthorizationFilter filter;
    private static FilterContext trueContext;

    @BeforeClass
    public static void setupNDR() {
        ndr = MockFedoraNDRAccess.getInstance();
        agent = ndr.newObject(Type.Agent);
        filter = new DefaultAuthorizationFilter();
        filter.init(ndr);

        MockFilterContext cxt =
                new MockFilterContext(FilterState.BEFORE_COMMIT);
        trueContext = cxt;
    }

    /** filter should add auth to aggregator without auth */
    @Test
    public void addAuthorizationAggregatorTest() throws Exception {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        aggregator.addRelationships(Arrays.asList(new Relationship(aggregator,
                AGGREGATOR_FOR, agent)));

        filter.doFilter(aggregator, trueContext);

        Relationship auth = aggregator.getRelationship(AUTHORIZED);

        assertNotNull("Authorization relationship not set", auth);
        assertEquals(
                "Agent should be the target of authorization relationship",
                agent, auth.related());
    }

    /** filter should add auth to MP without auth */
    @Test
    public void addAuthorizationMPTest() throws Exception {
        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.addRelationships(Arrays.asList(new Relationship(mp,
                METADTA_PROVIDER_FOR, agent)));

        filter.doFilter(mp, trueContext);

        Relationship auth = mp.getRelationship(AUTHORIZED);

        assertNotNull("Authorization relationship not set", auth);
        assertEquals(
                "Agent should be the target of authorization relationship",
                agent, auth.related());
    }

    /** filter should ignore other objects */
    @Test
    public void noAddAuthorizationOtherObjectsTest() throws Exception {
        NDRObject resource = ndr.newObject(Type.Resource);
        NDRObject agent = ndr.newObject(Type.Agent);
        NDRObject metadata = ndr.newObject(Type.Metadata);

        filter.doFilter(resource, trueContext);
        filter.doFilter(agent, trueContext);
        filter.doFilter(metadata, trueContext);

        assertNull(resource.getRelationship(AUTHORIZED));
        assertNull(agent.getRelationship(AUTHORIZED));
        assertNull(metadata.getRelationship(AUTHORIZED));
    }

    @Test
    public void noAddAuthorizationExistingAggregatorTest() throws Exception {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        NDRObject otherAgent = ndr.newObject(Type.Agent);
        aggregator.addRelationships(Arrays.asList(new Relationship(aggregator,
                AGGREGATOR_FOR, agent), new Relationship(aggregator,
                AUTHORIZED, otherAgent)));

        filter.doFilter(aggregator, trueContext);

        Relationship auth = aggregator.getRelationship(AUTHORIZED);

        assertNotNull("Authorization relationship was unset!", auth);
        assertNotSame(
                "Agent of existing auth relationship should not be changed",
                agent, auth.related());
    }

    @Test
    public void noAddAuthorizationExistingMPTest() throws Exception {
        NDRObject mp = ndr.newObject(Type.Aggregator);
        NDRObject otherAgent = ndr.newObject(Type.Agent);
        mp.addRelationships(Arrays.asList(new Relationship(mp,
                METADTA_PROVIDER_FOR, agent), new Relationship(mp, AUTHORIZED,
                otherAgent)));

        filter.doFilter(mp, trueContext);

        Relationship auth = mp.getRelationship(AUTHORIZED);

        assertNotNull("Authorization relationship was unset!", auth);
        assertNotSame(
                "Agent of existing auth relationship should not be changed",
                agent, auth.related());
    }
}
