/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access.filters;

import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.access.filters.HandleManagementFilter.HandleFactory;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;

import static junit.framework.Assert.*;

public class HandleManagementFilterTest {

    private static NDRAccess ndr;
    private static HandleManagementFilter filter;
    private HandleProbeInterface probe;

    @BeforeClass
    public static void setUp() {
        filter = new HandleManagementFilter();
        ndr = MockFedoraNDRAccess.getInstance();
    }

    @Before
    public void initProbe() {
        probe = new HandleProbeInterface();
        filter.setHandleFactory(probe);
    }

    /** Make sure handle is assigned to new objects */
    @Test
    public void assignNewHandleTest() throws AccessException {
        MockFilterContext context = new MockFilterContext(
                FilterState.NEW_OBJECT);

        NDRObject o = ndr.newObject(Type.Resource);

        filter.doFilter(o, context);

        assertTrue("Filter did not assign a handle to a new object",
                probe.assigned);
        assertNotNull("Object does not have a hasHandle property!", o
                .getProperty(Components.HAS_HANDLE));
        assertEquals("Filter did not really change the object's handle",
                HandleProbeInterface.HANDLE_VALUE, o.handle());
        assertEquals("Filter did not change the object's handle property",
                HandleProbeInterface.HANDLE_VALUE, o.getProperty(
                        Components.HAS_HANDLE).value());
    }

    /**
     * Make sure that the filter does not assign a handle for any state other
     * than new object
     */
    @Test
    public void doNotAssignHandleWhenNotWantedTest() {
        MockFilterContext context = new MockFilterContext(
                FilterState.NEW_OBJECT);

        NDRObject o = ndr.newObject(Type.Resource);

        for (FilterState s : FilterState.values()) {
            if (!s.equals(FilterState.NEW_OBJECT)) {
                context.setState(s);
                filter.doFilter(o, context);

                assertFalse("Filter should not have assigned a handle here!",
                        probe.assigned);
            }
        }
    }

    /**
     * Assure that after COMMIT, a subsequent rollback does not delete the
     * handle
     */
    @Test
    public void commitHandleDurabilityTest() {
        MockFilterContext context = new MockFilterContext(
                FilterState.NEW_OBJECT);

        NDRObject o = ndr.newObject(Type.Resource);

        /* Simulate the initial creation */
        filter.doFilter(o, context);
        assertTrue("Filter did not assign a handle to a new object",
                probe.assigned);
        assertFalse("Filter deleted handle on initial creation", probe.deleted);

        /* Simulate a successful commit */
        context.setState(FilterState.AFTER_COMMIT);
        filter.doFilter(o, context);
        assertFalse("Filter deleted handle on commit", probe.deleted);

        /* Do a subsequent rollback */
        context.setState(FilterState.ROLLBACK);
        filter.doFilter(o, context);

        assertFalse(
                "Filter should not have deleted handle upon ROLLBACK after a COMMIT!",
                probe.deleted);
    }

    /**
     * Make sure an assignment followed by a rollback creates then deletes a
     * handle from the handle server interface
     */
    @Test
    public void rollbackAssignmentTest() {
        MockFilterContext context = new MockFilterContext(
                FilterState.NEW_OBJECT);

        NDRObject o = ndr.newObject(Type.Resource);

        filter.doFilter(o, context);
        assertFalse("Filter deleted the handle prematurely", probe.deleted);

        context.setState(FilterState.ROLLBACK);
        filter.doFilter(o, context);
        assertTrue("Filter did not delete the handle upon rollback",
                probe.deleted);
    }

    /**
     * Make sure that a rollback by itself (i.e. not preceeded by a 'new
     * object') does not delete the handle
     */
    @Test
    public void rollbackIsolationTest() {
        MockFilterContext context = new MockFilterContext(FilterState.ROLLBACK);
        NDRObject o = ndr.newObject(Type.Resource);

        filter.doFilter(o, context);
        assertFalse(
                "Filter should not have deleted the handle from an isolated rollback",
                probe.deleted);
    }

    /** Purge should remove a handle */
    @Test
    public void purgeTest() {

        MockFilterContext context = new MockFilterContext(FilterState.PURGE);

        NDRObject o = ndr.newObject(Type.Resource);

        filter.doFilter(o, context);
        assertTrue("Filter should have deleted handle upon purge",
                probe.deleted);
    }

    private class HandleProbeInterface implements HandleFactory {
        public boolean assigned = false;
        public boolean deleted = false;
        public static final String HANDLE_VALUE = "assigned";

        public String getHandle(NDRObject o) {
            assigned = true;
            return HANDLE_VALUE;
        }

        public void deleteHandle(String handle) {
            deleted = true;
        }
    }
}
