/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access.filters;

import java.util.Date;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.content.OAIModel.Visibility;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.MetadataModel.Components.METADATA_PROVIDED_BY;
import static org.nsdl.repository.model.content.OAIModel.Set.Components.VISIBILITY;

import static junit.framework.Assert.*;

public class OAIVisibilityFilterTest {
    private static NDRAccess ndr;
    private static OAIVisibilityFilter filter;

    @BeforeClass
    public static void setUp() {
        filter = new OAIVisibilityFilter();
        ndr = MockFedoraNDRAccess.getInstance();
        filter.init(ndr);
    }

    @Test
    public void fromNullpropagationTest() throws Exception {

        NDRObject provider = ndr.newObject(Type.MetadataProvider);
        provider.commit();

        NDRObject metadata1 = ndr.newObject(Type.Metadata);
        metadata1.addRelationship(METADATA_PROVIDED_BY, provider);
        metadata1.commit();
        Date m1Create = metadata1.getCreatedDate();

        NDRObject metadata2 = ndr.newObject(Type.Metadata);
        metadata2.addRelationship(METADATA_PROVIDED_BY, provider);
        metadata2.commit();
        Date m2Create = metadata2.getCreatedDate();

        Thread.sleep(100); /* Just to make sure mod date is different */

        /* Register filter as if we just opened this object */
        MockFilterContext providerContext = new MockFilterContext(
                FilterState.OPEN);
        filter.doFilter(provider, providerContext);

        provider.addProperty(VISIBILITY, Visibility.PROTECTED.toString());

        /* Triger after commit filter actions. Should touch metadata */
        providerContext.setState(FilterState.AFTER_COMMIT);
        filter.doFilter(provider, providerContext);

        /* XXX nasty way to wait for OAIVisibility filter to finish */
        Thread.currentThread().join(200);

        metadata1 = ndr.getObject(metadata1);
        metadata2 = ndr.getObject(metadata2);
        assertNotSame("Filter did not modify First Metadata "
                + metadata1.handle() + ", created: " + m1Create.getTime()
                + ", modified: " + metadata1.getLastModifiedDate().getTime(),
                m1Create, metadata1.getLastModifiedDate());
        assertNotSame("Filter did not modify Second Metadata "
                + metadata2.handle() + ", created: " + m2Create.getTime()
                + ", modified: " + metadata2.getLastModifiedDate().getTime(),
                m2Create, metadata2.getLastModifiedDate());

    }

    @Test
    public void defaultVisibilityChange() {
        NDRObject provider = ndr.newObject(Type.MetadataProvider);
        provider.commit();

        NDRObject metadata1 = ndr.newObject(Type.Metadata);
        metadata1.addRelationship(METADATA_PROVIDED_BY, provider);
        metadata1.commit();
        Date m1Create = metadata1.getCreatedDate();

        /* Register filter as if we just opened this object */
        MockFilterContext providerContext = new MockFilterContext(
                FilterState.OPEN);
        filter.doFilter(provider, providerContext);

        provider.addProperty(VISIBILITY, Visibility.PUBLIC.toString());

        /* Triger after commit filter actions. Should touch metadata */
        providerContext.setState(FilterState.AFTER_COMMIT);
        filter.doFilter(provider, providerContext);

        assertEquals(m1Create.getTime(), metadata1.getLastModifiedDate()
                .getTime());
    }
}
