/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.access.filters;

import java.io.ByteArrayInputStream;
import java.net.URL;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.types.Resource;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.Configuration;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertNotNull;
import static junit.framework.Assert.assertTrue;

public class ResourceContentFilterTest {
    private static NDRAccess ndr;
    private static ResourceContentFilter filter;
    private static final String TEST_BASE_URL = "http://example.org/ResourceContentFilterTest/";

    private static final String EXAMPLE_URL = "http://example.org";

    @BeforeClass
    public static void setupNDR() {
        ndr = MockFedoraNDRAccess.getInstance();
        filter = new ResourceContentFilter();
        filter.init(ndr);
        Configuration config = new Configuration();
        config.putProperty("param.repository.baseURL", TEST_BASE_URL);
        config.update();
        filter.updateConfig(config);
    }

    /*
     * New object with an identifier but NO content stream should result in a
     * new content stream containing a redirect datastream pointing to the
     * identifier URI.
     */
    @Test
    public void newObjectWithIdentifierTest() {
        MockFilterContext cxt = new MockFilterContext(FilterState.BEFORE_COMMIT);
        NDRObject object = ndr.newObject(Type.Resource);
        object.addProperty(Resource.IDENTIFIER_URL, EXAMPLE_URL);

        filter.doFilter(object, cxt);

        assertEquals(EXAMPLE_URL, object.getProperty(Resource.IDENTIFIER_URL)
                .value());
        assertNotNull(object.getDatastream(Resource.CONTENT));
        assertEquals(EXAMPLE_URL, object.getDatastream(Resource.CONTENT)
                .getURL().toString());
    }

    /*
     * New object WITHOUT an identifier but WITH a content stream should be
     * given a URI identifier that contains the URL to the datastream.
     */
    @Test
    public void newObjectWithContentTest() {
        MockFilterContext cxt = new MockFilterContext(FilterState.BEFORE_COMMIT);
        NDRObject object = ndr.newObject(Type.Resource);
        Datastream content = object.addDatastream(Resource.CONTENT);
        content.setContent(new ByteArrayInputStream("Content".getBytes()));

        filter.doFilter(object, cxt);

        assertNotNull(object.getProperty(Resource.IDENTIFIER_URL));
        assertTrue(object.getProperty(Resource.IDENTIFIER_URL).value().equals(
                TEST_BASE_URL + "get/" + object.handle() + "/content"));
    }

    /*
     * Changing the identifier should just change the identifier - content
     * datastream should stay the same
     */
    @Test
    public void changeIdentifierTest() throws Exception {
        final String CHANGED_URL = EXAMPLE_URL + 2;
        MockFilterContext cxt = new MockFilterContext(FilterState.OPEN);
        NDRObject object = ndr.newObject(Type.Resource);
        object.addProperty(Resource.IDENTIFIER_URL, EXAMPLE_URL);
        object.addDatastream(Resource.CONTENT).setURL(new URL(EXAMPLE_URL));

        filter.doFilter(object, cxt);

        object.getProperty(Resource.IDENTIFIER_URL).setValue(CHANGED_URL);

        cxt.setState(FilterState.BEFORE_COMMIT);
        filter.doFilter(object, cxt);

        assertEquals(CHANGED_URL, object.getProperty(Resource.IDENTIFIER_URL)
                .value());
        assertEquals(EXAMPLE_URL, object.getDatastream(Resource.CONTENT)
                .getURL().toString());
    }

    /*
     * Changing the content should just change the content - identifier does not
     * change
     */
    @Test
    public void changeContentTest() {
        MockFilterContext cxt = new MockFilterContext(FilterState.BEFORE_COMMIT);
        NDRObject object = ndr.newObject(Type.Resource);
        Datastream content = object.addDatastream(Resource.CONTENT);
        content.setContent(new ByteArrayInputStream("Content".getBytes()));

        filter.doFilter(object, cxt);

        String initialIdentifier = object.getProperty(Resource.IDENTIFIER_URL)
                .value();

        cxt = new MockFilterContext(FilterState.OPEN);
        filter.doFilter(object, cxt);

        object.getDatastream(Resource.CONTENT).setContent(
                new ByteArrayInputStream("Content_2".getBytes()));

        cxt.setState(FilterState.BEFORE_COMMIT);
        filter.doFilter(object, cxt);

        /* Identifier should not change */
        assertEquals(initialIdentifier, object.getProperty(
                Resource.IDENTIFIER_URL).value());
    }

    /* A non-related change should not affect the identifier or content streams */
    @Test
    public void changeNothingTest() throws Exception {
        MockFilterContext cxt = new MockFilterContext(FilterState.OPEN);
        NDRObject object = ndr.newObject(Type.Resource);
        object.addProperty(Resource.IDENTIFIER_URL, EXAMPLE_URL);
        object.addDatastream(Resource.CONTENT).setURL(new URL(EXAMPLE_URL));

        /* Meaningless change */
        object.addRelationship(Resource.MEMBER_OF, object);

        filter.doFilter(object, cxt);

        /* Identifier should not change */
        assertEquals(EXAMPLE_URL, object.getProperty(Resource.IDENTIFIER_URL)
                .value());

        /* Content should not change */
        assertEquals(EXAMPLE_URL, object.getDatastream(Resource.CONTENT)
                .getURL().toString());
    }

    /*
     * Conflicts between the identifier and content stream are assumed to be
     * intentional. Allow them to happen.
     */
    @Test
    public void newObjectConflictingIdentifierTest() throws Exception {
        final String CONFLICTING_URL = EXAMPLE_URL + "Conflicting";
        MockFilterContext cxt = new MockFilterContext(FilterState.BEFORE_COMMIT);
        NDRObject object = ndr.newObject(Type.Resource);
        object.addProperty(Resource.IDENTIFIER_URL, EXAMPLE_URL);
        object.addDatastream(Resource.CONTENT).setURL(new URL(CONFLICTING_URL));

        filter.doFilter(object, cxt);

        assertEquals(EXAMPLE_URL, object.getProperty(Resource.IDENTIFIER_URL)
                .value());
        assertNotNull(object.getDatastream(Resource.CONTENT));
        assertEquals(CONFLICTING_URL, object.getDatastream(Resource.CONTENT)
                .getURL().toString());
    }

    /*
     * If a new object with a redirect datastream is created, use the redirected
     * URL as the identifier.
     */
    @Test
    public void newObjectRedirectContentUnspecifiedIdentifierTest()
            throws Exception {
        MockFilterContext cxt = new MockFilterContext(FilterState.BEFORE_COMMIT);
        NDRObject object = ndr.newObject(Type.Resource);
        object.addDatastream(Resource.CONTENT).setURL(new URL(EXAMPLE_URL));

        filter.doFilter(object, cxt);

        assertNotNull(object.getProperties(Resource.IDENTIFIER_URL));
        assertEquals(EXAMPLE_URL, object.getProperty(Resource.IDENTIFIER_URL)
                .value());
    }
}
