/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.access.filters;

import java.util.Arrays;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.impl.fedora.XMLDatastream;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.types.Type;
import org.nsdl.util.DateUtility;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import static junit.framework.Assert.*;
import static org.nsdl.repository.model.content.MetadataModel.Components.DC;

public class TouchDCFilterTest {

    private static final String DC_URI = "http://purl.org/dc/elements/1.1/";

    private static NDRAccess ndr;
    private static TouchDCFilter filter;
    private static FilterContext trueContext;

    @BeforeClass
    public static void setupNDR() {
        ndr = MockFedoraNDRAccess.getInstance();

        filter = new TouchDCFilter();
        filter.init(ndr);

        MockFilterContext cxt = new MockFilterContext(FilterState.BEFORE_COMMIT);
        trueContext = cxt;
    }

    /* Make sure that we add a DC ds to an object that doesn't have it */
    @Test
    public void newObjectTest() throws Exception {
        NDRObject newObject = ndr.newObject(Type.Metadata);
        filter.doFilter(newObject, trueContext);

        assertNotNull("DC datastream should have been created", newObject
                .getDatastream(DC));
    }

    @Test
    public void addDateTest() throws Exception {
        NDRObject o = ndr.newObject(Type.Metadata);

        /* build datastream without a dc:date */
        Datastream dc = createDCDatastream(null);

        o.addDatastreams(Arrays.asList(dc));

        filter.doFilter(o, trueContext);

        assertNotNull("Date should not be null", getDateContent(o));
    }

    @Test
    public void modifyDateTest() throws Exception {
        NDRObject o = ndr.newObject(Type.Metadata);

        /* build datastream without a dc:date */
        Datastream dc = createDCDatastream("DATE");

        o.addDatastreams(Arrays.asList(dc));

        filter.doFilter(o, trueContext);

        assertNotSame("dc:date should be different", "DATE", getDateContent(o));
    }

    private static String getDateContent(NDRObject o) throws Exception {
        Datastream dc = o.getDatastream(DC);
        Document dcDoc = NDRXMLutil.getDocument(dc.getContent(), true, false);
        Node existingDate = NDRXMLutil.getNodeByName(dcDoc, DC_URI, "date");
        if (existingDate == null) {
            System.err.println("existing date is null");
            return null;
        }

        return existingDate.getTextContent();
    }

    private static Datastream createDCDatastream(String date) throws Exception {
        String description = "<oai_dc:dc xmlns:oai_dc=\"http://www.openarchives.org/OAI/2.0/oai_dc/\" "
                + "xmlns:dc=\"http://purl.org/dc/elements/1.1/\">\n"
                + " <dc:title>NSDL Repository Metadata</dc:title>\n";

        if (date != null) {
            description += " <dc:date>"
                    + DateUtility.getCurrTimeInDashedISO8601Plus()
                    + "</dc:date>";
        }
        description += "</oai_dc:dc>";
        Node dsContent = (Node) NDRXMLutil
                .getDocument(description, true, false).getDocumentElement();
        
        return new XMLDatastream(DC, dsContent, "DC", "text/xml");
    }
}
