/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth;

import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.NoSuchAlgorithmException;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;

import static junit.framework.Assert.*;

public class HttpAuthHeaderTest {
    private PrivateKey privateKey;
    private PublicKey publicKey;
    private static final String REQUEST = "GET /test/Test";
    private static final String AGENT = "2200/200605092004123T";
    private ThreadLocal<DateFormat> df = new ThreadLocal<DateFormat>() {
        protected DateFormat initialValue() {
            return new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss Z");
        }
    };

    @Before
    public void setUp() throws NoSuchAlgorithmException {
        if (this.privateKey == null || this.publicKey == null) {
            KeyPairGenerator generator = KeyPairGenerator.getInstance("RSA");
            KeyPair pair = generator.generateKeyPair();
            this.privateKey = pair.getPrivate();
            this.publicKey = pair.getPublic();
        }
    }

    /** Test to see if basic authentication (signing/validating) works */
    @Test
    public void testAuthentication() throws AuthenticationException {
        Date now = new Date();
        String dateHeader = "Date: " + df.get().format(now) + "\n";

        String stringHeader = "accept-language: en-us,en;q=0.5\n"
                + "accept-encoding: gzip,deflate\n"
                + "accept-charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7\n"
                + "keep-alive: 300\n" + "x-nsdl-test: test\n" + dateHeader;
        HttpHeader header = HttpAuthHeader.sign(new HttpHeader(stringHeader),
                REQUEST, AGENT, privateKey);
        HttpAuthHeader auth = new HttpAuthHeader(header, REQUEST);

        assertTrue(auth.isSigned());
        assertTrue(auth.getAgent().equals(AGENT));
        assertTrue(auth.isValid(publicKey));
    }

    /** Test to see if out-of-bounds times are rejected */
    @Test
    public void testDateRejection() throws AuthenticationException {
        /* Create a time that was 15 minutes ago */
        Date now = new Date(new Date().getTime() - (15 * 1000 * 60));
        String dateHeader = "Date: " + df.get().format(now) + "\n";

        String stringHeader = "accept-language: en-us,en;q=0.5\n"
                + "accept-encoding: gzip,deflate\n"
                + "accept-charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7\n"
                + "keep-alive: 300\n" + "x-nsdl-test: test\n" + dateHeader;
        HttpHeader header = HttpAuthHeader.sign(new HttpHeader(stringHeader),
                REQUEST, AGENT, privateKey);
        HttpAuthHeader auth = new HttpAuthHeader(header, REQUEST);
        assertFalse(auth.isValid(publicKey));
    }

    /** test to see if headers without a date are rejected */
    @Test
    public void testNonExistantDateRejection() {
        String stringHeader = "accept-language: en-us,en;q=0.5\n"
                + "accept-encoding: gzip,deflate\n"
                + "accept-charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7\n"
                + "keep-alive: 300\n" + "x-nsdl-test: test\n";

        try {
            HttpAuthHeader.sign(new HttpHeader(stringHeader), REQUEST, AGENT,
                    privateKey);
            fail("Should have rejected a header without a date");
        } catch (AuthenticationException e) {
            /* expected */
        }
    }

    /** Test to assure that the x-nsdl-date is obeyed if it conflicts with Date */
    @Test
    public void testMultipleDates() throws AuthenticationException {
        /* Create a time that was 15 minutes ago */
        Date ago = new Date(new Date().getTime() - (15 * 1000 * 60));

        /* Create the current time */
        Date now = new Date();
        String dateHeader = "Date: " + df.get().format(ago) + "\n"
                + "x-nsdl-date: " + df.get().format(now) + "\n";

        String stringHeader = "accept-language: en-us,en;q=0.5\n"
                + "accept-encoding: gzip,deflate\n"
                + "accept-charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7\n"
                + "keep-alive: 300\n" + "x-nsdl-test: test\n" + dateHeader;
        HttpHeader header = HttpAuthHeader.sign(new HttpHeader(stringHeader),
                REQUEST, AGENT, privateKey);
        HttpAuthHeader auth = new HttpAuthHeader(header, REQUEST);
        assertTrue(auth.isValid(publicKey));
    }

    /**
     * Test to assure that canonical header generation is stable with multiple
     * x-nsdl items
     */
    @Test
    public void testMultipleNSDL() throws AuthenticationException {
        Date now = new Date();
        String dateHeader = "Date: " + df.get().format(now) + "\n";

        String stringHeader = "accept-language: en-us,en;q=0.5\n"
                + "accept-encoding: gzip,deflate\n"
                + "accept-charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7\n"
                + "keep-alive: 300\n" + "x-nsdl-XYZZY: magic\n"
                + "x-nsdl-test: test-another\n" + "x-nsdl-test: test\n"
                + dateHeader;
        HttpHeader header = HttpAuthHeader.sign(new HttpHeader(stringHeader),
                REQUEST, AGENT, privateKey);
        HttpAuthHeader auth = new HttpAuthHeader(header, REQUEST);

        assertTrue(auth.isSigned());
        assertTrue(auth.getAgent().equals(AGENT));
        assertTrue(auth.isValid(publicKey));
    }

    /**
     * Test to assure that if the canonical header parts are modified, the
     * header is rejected
     */
    @Test
    public void testAlteredHeader() throws AuthenticationException {
        Date now = new Date();
        String dateHeader = "Date: " + df.get().format(now) + "\n";

        String stringHeader = "accept-language: en-us,en;q=0.5\n"
                + "accept-encoding: gzip,deflate\n"
                + "accept-charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7\n"
                + "keep-alive: 300\n" + "x-nsdl-test: test\n" + dateHeader;
        HttpHeader header = HttpAuthHeader.sign(new HttpHeader(stringHeader),
                REQUEST, AGENT, privateKey);
        String headerString = header.toString().replaceAll(" test", " poison");
        HttpAuthHeader auth = new HttpAuthHeader(new HttpHeader(headerString),
                REQUEST);
        assertTrue(auth.isSigned());
        assertFalse(auth.isValid(publicKey));

    }

    @Test
    @Ignore
    public void testTime() throws AuthenticationException {
        long NUM_TESTS = 100;
        Date now = new Date();
        String dateHeader = "Date: " + df.get().format(now) + "\n";

        String stringHeader = "accept-language: en-us,en;q=0.5\n"
                + "accept-encoding: gzip,deflate\n"
                + "accept-charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7\n"
                + "keep-alive: 300\n" + "x-nsdl-test: test\n" + dateHeader;
        HttpHeader origHead = new HttpHeader(stringHeader);
        Date start = new Date();
        for (int i = 0; i < NUM_TESTS; i++) {
            HttpAuthHeader.generateSignature(origHead, REQUEST, AGENT,
                    privateKey);
        }
        System.err.print((new Date().getTime() - start.getTime()) / NUM_TESTS
                + "ms per sig\n");

        HttpHeader header = HttpAuthHeader.sign(new HttpHeader(stringHeader),
                REQUEST, AGENT, privateKey);
        HttpAuthHeader auth = new HttpAuthHeader(header, REQUEST);

        start = new Date();
        for (int i = 0; i < 100; i++) {
            auth.isValid(publicKey);
        }
        System.err.print((new Date().getTime() - start.getTime()) / NUM_TESTS
                + "ms per validation\n");
    }

    /** Test to see if basic authentication (signing/validating) works */
    @Test
    public void testBadRequest() throws AuthenticationException {
        Date now = new Date();
        String dateHeader = "Date: " + df.get().format(now) + "\n";

        String stringHeader = "accept-language: en-us,en;q=0.5\n"
                + "accept-encoding: gzip,deflate\n"
                + "accept-charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7\n"
                + "keep-alive: 300\n" + "x-nsdl-test: test\n" + dateHeader;
        HttpHeader header = HttpAuthHeader.sign(new HttpHeader(stringHeader),
                REQUEST + "BAD", AGENT, privateKey);
        HttpAuthHeader auth = new HttpAuthHeader(header, REQUEST);

        assertTrue(auth.isSigned());
        assertTrue(auth.getAgent().equals(AGENT));
        assertFalse(auth.isValid(publicKey));
    }
}
