/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth.rules;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.api.auth.MockAgentAuthInfo;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.NDRObjectTemplate;
import org.nsdl.repository.util.StaticObjectInfo;

import static junit.framework.Assert.assertFalse;
import static junit.framework.Assert.assertTrue;
import static org.nsdl.repository.model.content.AggregatorModel.Components.ASSOCIATED_WITH;

public class AssociationRelationshipsTest {

    private static Relationship relationship;

    private static MockFedoraNDRAccess ndr;

    @BeforeClass
    public static void init() {
        ndr = MockFedoraNDRAccess.getInstance();

        NDRObject testSubject = ndr.newObject(Type.Aggregator);
        testSubject.commit();

        NDRObject testObject = ndr.newObject(Type.Resource);
        testObject.commit();

        relationship = new Relationship(testSubject, ASSOCIATED_WITH,
                testObject);
    }

    private ComponentRule rule = new AssociationRelationships(ndr);

    @Test
    public void testAcceptance() {

        MockAgentAuthInfo authAgent = new MockAgentAuthInfo(
                new NDRObjectTemplate(Type.Agent, "Authorized"), true);

        authAgent.setAuthorizedFor(relationship.subject().handle());

        assertTrue(rule.checkCompliance(relationship, null, authAgent));
    }

    @Test
    public void testNonMatchingRelationshipname() {
        Relationship nonMatchingRelationship = new Relationship(
                new StaticObjectInfo("TestSubject", Type.Aggregator),
                new Relationship.Name(Constants.REPOS_NAMESPACE + "X",
                        "associatedWith", "X"), new NDRObjectTemplate(
                        Type.Resource, "TestObject"));

        MockAgentAuthInfo authAgent = new MockAgentAuthInfo(
                new NDRObjectTemplate(Type.Agent, "Authorized"), true);

        authAgent.setAuthorizedFor("TestSubject");

        assertTrue(this.rule.checkCompliance(nonMatchingRelationship, null,
                authAgent));

        Relationship.Pattern pattern = Relationship.Pattern
                .getPattern(nonMatchingRelationship);
        assertFalse(this.rule.matchingPatterns().contains(pattern));
    }

    @Test
    public void testNonMatchingRelationship() {

        Relationship nonMatchingRelationship = new Relationship(
                new StaticObjectInfo("TestSubject", Type.Aggregator),
                new Relationship.Name("ns", "not_an_associatedWith", "no"),
                new NDRObjectTemplate(Type.Resource, "TestObject"));

        MockAgentAuthInfo authAgent = new MockAgentAuthInfo(
                new NDRObjectTemplate(Type.Agent, "Authorized"), true);

        authAgent.setAuthorizedFor("TestSubject");

        /** Should be true because the relationship does not match */
        assertTrue(this.rule.checkCompliance(nonMatchingRelationship, null,
                authAgent));

        Relationship.Pattern pattern = Relationship.Pattern
                .getPattern(nonMatchingRelationship);
        assertFalse(this.rule.matchingPatterns().contains(pattern));
    }

    @Test
    public void testMatchingPatterns() {
        Relationship.Pattern pattern = Relationship.Pattern
                .getPattern(relationship);

        assertTrue(this.rule.matchingPatterns().contains(pattern));
    }

    @Test
    public void testUnauthorizedAgent() {
        MockAgentAuthInfo authAgent = new MockAgentAuthInfo(
                new NDRObjectTemplate(Type.Agent, "Authorized"), true);

        authAgent.setAuthorizedFor("NotThisTestSubject");

        assertFalse(rule.checkCompliance(relationship, null, authAgent));
    }
}
