/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.auth.rules;

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.junit.BeforeClass;
import org.junit.Test;

import org.nsdl.repository.api.auth.MockAgentAuthInfo;
import org.nsdl.repository.api.request.MockRequest;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.NDRObjectTemplate;

import static junit.framework.Assert.*;
import static org.nsdl.repository.model.content.MetadataModel.Components.METADATA_FOR;
import static org.nsdl.repository.model.content.MetadataModel.Components.METADATA_PROVIDED_BY;

public class MetadataProvisionRelationshipsTest {

    private static ComponentRule rule;

    private static MockFedoraNDRAccess ndr;

    private static MockAgentAuthInfo agentAuthorization;

    @BeforeClass
    public static void init() {
        ndr = MockFedoraNDRAccess.getInstance();

        rule = new MetadataProvisionRelationships(ndr);

        /* Create the authorized agent */
        NDRObject authAgent = ndr.newObject(Type.Agent, "2200/Agent");
        authAgent.commit();

        /* Create the metadataProvider object */
        NDRObject metadataProviderObject = ndr.newObject(Type.MetadataProvider,
                "2200/Provider");
        metadataProviderObject.commit();

        /* Create the metadata object */
        NDRObject metadataObject = ndr
                .newObject(Type.Metadata, "2200/Metadata");
        metadataProvidedByRelationship = new Relationship(metadataObject,
                METADATA_PROVIDED_BY, metadataProviderObject);
        metadataObject.addRelationships(Arrays
                .asList(metadataProvidedByRelationship));
        metadataObject.commit();

        /* Create agent authorization */
        agentAuthorization = new MockAgentAuthInfo(authAgent, true);
        agentAuthorization.setAuthorizedFor(metadataProviderObject.handle());

        metadataForRelationships = buildMetadataFor(metadataObject);
    }

    private static List<Relationship> metadataForRelationships;

    private static Relationship metadataProvidedByRelationship;

    private static List<Relationship> buildMetadataFor(NDRObject metadataObj) {
        return Arrays
                .asList(new Relationship[] {

                        new Relationship(metadataObj, METADATA_FOR,
                                new NDRObjectTemplate(Type.Resource,
                                        "ResourceHandle")),

                        new Relationship(metadataObj, METADATA_FOR,
                                new NDRObjectTemplate(Type.Aggregator,
                                        "AggregatorHandle")),

                        new Relationship(
                                metadataObj,
                                METADATA_FOR,
                                new NDRObjectTemplate(Type.Agent, "AgentHandle")), });
    }

    @Test
    public void testMetadataForAcceptance() {

        Iterator<Relationship> i = metadataForRelationships.iterator();
        while (i.hasNext()) {
            assertTrue(rule.checkCompliance(i.next(), null, agentAuthorization));
        }
    }

    @Test
    public void testMetadataForMatching() {
        for (Relationship r : metadataForRelationships) {
            assertTrue(rule.matchingPatterns().contains(
                    Relationship.Pattern.getPattern(r)));
        }

        assertEquals(rule.matchingPatterns().size(), metadataForRelationships
                .size() + 1);
    }

    @Test
    public void testMetadataprovidedByAcceptanceCreate() {

        MockRequest request = new MockRequest();
        request.setType(Request.Type.Create);

        assertTrue(rule.checkCompliance(metadataProvidedByRelationship,
                request, agentAuthorization));
    }

    @Test
    public void testMetadataprovidedByAcceptanceUpdate() {

        MockRequest request = new MockRequest();
        request.setType(Request.Type.Update);

        assertTrue(rule.checkCompliance(metadataProvidedByRelationship,
                request, agentAuthorization));
    }

    @Test
    public void testMetadataprovidedByAcceptanceDelete() {

        MockRequest request = new MockRequest();
        request.setType(Request.Type.Delete);

        assertTrue(rule.checkCompliance(metadataProvidedByRelationship,
                request, agentAuthorization));
    }

    @Test
    public void testMetadataProvidedByMatching() {
        assertTrue(rule.matchingPatterns()
                .contains(
                        Relationship.Pattern
                                .getPattern(metadataProvidedByRelationship)));
        assertEquals(rule.matchingPatterns().size(), metadataForRelationships
                .size() + 1);
    }

    /* Should probably test modification of existing Matadata objects */
    @Test
    public void testMetadataProvidedByDenialUpdate() {
        agentAuthorization.setAuthorizedFor("OtherMetadataProviderObject");

        MockRequest request = new MockRequest();
        request.setType(Request.Type.Update);

        assertFalse(rule.checkCompliance(metadataProvidedByRelationship,
                request, agentAuthorization));
    }

    @Test
    public void testMetadataProvidedByDenialCreate() {
        agentAuthorization.setAuthorizedFor("OtherMetadataProviderObject");

        MockRequest request = new MockRequest();
        request.setType(Request.Type.Create);

        assertFalse(rule.checkCompliance(metadataProvidedByRelationship,
                request, agentAuthorization));
    }

    @Test
    public void testMetadataProvidedByDenialDelete() {
        agentAuthorization.setAuthorizedFor("OtherMetadataProviderObject");

        MockRequest request = new MockRequest();
        request.setType(Request.Type.Delete);

        assertFalse(rule.checkCompliance(metadataProvidedByRelationship,
                request, agentAuthorization));
    }
}
