/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.request;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.net.URL;
import java.util.Arrays;
import java.util.Set;

import javax.servlet.ServletContext;
import javax.servlet.ServletInputStream;

import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.multipart.ByteArrayPartSource;
import org.apache.commons.httpclient.methods.multipart.FilePart;
import org.apache.commons.httpclient.methods.multipart.MultipartRequestEntity;
import org.apache.commons.httpclient.methods.multipart.Part;
import org.apache.commons.httpclient.methods.multipart.PartSource;
import org.junit.Before;
import org.junit.Test;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.api.request.XMLRequest;
import org.nsdl.repository.api.request.Request.Commands;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.StreamUtil;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Document;

import servletunit.HttpServletRequestSimulator;
import servletunit.ServletContextSimulator;

import junit.framework.JUnit4TestAdapter;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

public class XMLRequestTest {

    private static MockFedoraNDRAccess ndr = MockFedoraNDRAccess.getInstance();
    private ContentSimulator request;

    public static String pidNS = "XMLRequestTest:";
    private static String URL_BASE = "http://ndrtest.nsdl.org/api/";
    private static String INPUT_XML = "inputXML";
    private static String DS_CONTENT = "datastreamContent";

    private static String BINARY_STREAM_PLACEHOLDER = "<!-- bs -->";
    private static String BINARY_DS_NAME = "binary_content";
    private static String BINARY_DS_MIME = "application/test";

    private static final byte[] BINARY_CONTENT = { -62, -75, 3, 0, 0, 12, 127 };

    private static String basicQuery = "<inputXML xmlns=\"http://ns.nsdl.org/ndr/request_v1.00/\" "
            + "xmlns:nsdl=\"http://ns.nsdl.org/api/relationships#\" xmlns:test_prop=\"TEST\">\n"
            + "   <resource>\n"
            + "      <properties>\n"
            + "         <identifier type=\"URL\">http://nsdl.org</identifier>\n"
            + "      </properties>\n"
            + "      <relationships>\n"
            + "         <memberOf>2200/Test1</memberOf>\n"
            + "         <nsdl:memberOf>2200/Test2</nsdl:memberOf>\n"
            + "      </relationships>\n"
            + "      <data>\n"
            + "       <add>\n"
            + "       <testXMLStream MIMETYPE=\"test/xml\">\n"
            + "        <data>\n"
            + "          XML_STREAM!\n"
            + "        </data>\n"
            + "       </testXMLStream>\n"
            + "       <testRedirectStream ID=\"redir\" MIMETYPE=\"application/html+xml\">\n"
            + "        http://nsdl.org\n" + "       </testRedirectStream>\n"
            + BINARY_STREAM_PLACEHOLDER
            + "       <format type=\"test_dc\">\n"
            + "        <meta>\n"
            + "         <nsdl_dc>\n"
            + "          META_TEST\n"
            + "         </nsdl_dc>\n"
            + "        </meta>\n"
            + "        <info>\n"
            + "         <test>\n"
            + "          INFO_TEST\n"
            + "         </test>\n"
            + "        </info>\n"
            + "       </format>\n"
            + "       <format type=\"test_null\">\n"
            + "        <meta/>\n"
            + "       </format>\n"
            + "       <format type=\"test_prop\">\n"
            + "        <meta>\n"
            + "         <test_prop:blah>test</test_prop:blah>\n"
            + "        </meta>\n"
            + "       </format>"
            + "       <test_prop:test_prop nsdl:foo=\"BLAH\">\n"
            + "        <test_prop:test>test</test_prop:test>\n"
            + "       </test_prop:test_prop>\n"
            + "       </add>\n"
            + "      </data>\n" + "   </resource>\n" + "</inputXML>";

    @Before
    public void setUp() {
        request = new ContentSimulator(new ServletContextSimulator());
    }

    public static junit.framework.Test suite() {
        return new JUnit4TestAdapter(XMLRequestTest.class);
    }

    /**
     * Test if XMLRequest correctly determines the acted-upon object from
     * request parameters
     */
    @Test
    public void testExistingObjectFromRequest() throws Exception {
        String desiredHandle = "2200/resourceHandle";
        String pathInfo = "modifyResource/" + desiredHandle;

        /* Set up the request parameters */
        request.setRequestURL(URL_BASE + pathInfo);
        String[] query = { basicQuery };
        request.setParameterValue(INPUT_XML, query);
        request.setPathInfo(pathInfo);

        NDRObject newObject = ndr.newObject(Type.Resource, desiredHandle);
        newObject.commit();

        XMLRequest testRequest = new XMLRequest(request);

        assertEquals(testRequest.getObjectInfo().handle(), desiredHandle);
    }

    /**
     * For a request adding a new object, assure that we determine the type
     * correctly from inputXML
     */
    @Test
    public void testNewObjectType() throws Exception {
        String pathInfo = "addResource";

        /* Set up the request parameters */
        request.setRequestURL(URL_BASE + pathInfo);
        String[] query = { basicQuery };
        request.setParameterValue(INPUT_XML, query);
        request.setPathInfo(pathInfo);

        XMLRequest testRequest = new XMLRequest(request);

        assertEquals("Expected XMLRequest to recognize type.", Type.Resource,
                testRequest.getObjectInfo().type());

    }

    /**
     * Anything in the "request" namespace should be translated to the "nsdl
     * relationships" namespace
     */
    @Test
    public void testRequestTranslation() throws Exception {
        String pathInfo = "addResource";

        /* Set up the request parameters */
        request.setRequestURL(URL_BASE + pathInfo);
        String[] query = { basicQuery };
        request.setParameterValue(INPUT_XML, query);
        request.setPathInfo(pathInfo);

        ((NDRObject) ndr.newObject(Type.Resource, "2200/Test1")).commit();
        ((NDRObject) ndr.newObject(Type.Resource, "2200/Test2")).commit();

        XMLRequest testRequest = new XMLRequest(request);

        /* Test the two MemberOf relationships */
        Set<Relationship> addRels = testRequest
                .getRelationships(Request.Commands.ADD);
        assertEquals("Did not find added relationships", 2, addRels.size());

        for (Relationship r : addRels) {
            assertEquals(r.predicate().getNamespaceURI(),
                    Constants.REPOS_NAMESPACE);
        }

        /* Test the identifier */
        Set<Property> addProperties = testRequest
                .getProperties(Request.Commands.ADD);
        assertEquals(addProperties.size(), 1);

        for (Property p : addProperties) {
            assertEquals(p.predicate().getNamespaceURI(),
                    Constants.REPOS_NAMESPACE);
            assertEquals(p.predicate().getLocalPart(), "hasResourceURL");
        }
    }

    /** Makes sure an explicit mimetype is accepted */
    @Test
    public void testMimetype() throws Exception {
        Request testRequest = getPopulatedRequest();

        Set<Datastream> datastreams = testRequest.getDatastreams(Commands.ADD);

        String mimetype = null;
        for (Datastream d : datastreams) {
            if (d.getName().getID().equals("testXMLStream")) {
                mimetype = d.getMimeType();
            }
        }

        assertEquals(mimetype, "test/xml");

    }

    @Test
    public void testOverridenIDName() throws Exception {
        Request testRequest = getPopulatedRequest();

        Set<Datastream> datastreams = testRequest.getDatastreams(Commands.ADD);

        boolean found = false;
        for (Datastream d : datastreams) {
            if (d.getName().getID().equals("redir")) {
                found = true;
            }
        }
        assertTrue(found);
    }

    @Test
    public void testRedirectDetection() throws Exception {
        Request testRequest = getPopulatedRequest();

        Set<Datastream> datastreams = testRequest.getDatastreams(Commands.ADD);

        Datastream redir = null;
        for (Datastream d : datastreams) {
            if (d.getName().getID().equals("redir")) {
                redir = d;
            }
        }

        assertEquals(redir.getURL().toString(), new URL("http://nsdl.org")
                .toString());
    }

    @Test
    public void testMetadataSpecialCase() throws Exception {
        Request testRequest = getPopulatedRequest();
        Set<Datastream> datastreams = testRequest.getDatastreams(Commands.ADD);
        Datastream meta = null;
        Datastream info = null;

        for (Datastream d : datastreams) {
            if (d.getName().getID().equals("format_test_dc")) {
                meta = d;
            } else if (d.getName().getID().equals("format_test_dc_info")) {
                info = d;
            }
        }

        assertNotNull("Metadata was found to be null", meta);
        assertNotNull("Metadata info was found to be null", info);
    }

    /**
     * Make sue we have no problem with non-defiend metadata.
     * <p>
     * What we want is a datastream defined.. but with no content
     * </p>
     * 
     * @throws Exception
     */
    @Test
    public void testNullMetadata() throws Exception {
        Request testRequest = getPopulatedRequest();
        Set<Datastream> datastreams = testRequest.getDatastreams(Commands.ADD);
        Datastream meta = null;
        Datastream info = null;

        for (Datastream d : datastreams) {
            if (d.getName().getID().equals("format_test_null")) {
                meta = d;
            } else if (d.getName().getID().equals("format_test_null_info")) {
                info = d;
            }
        }

        /* metadata container present */
        assertNotNull(meta);

        /* No info defined.. */
        assertNull(info);

        byte[] content = new byte[8];
        int numRead = meta.getContent().read(content);

        /* make sure there is no datastream content.. */
        assertEquals(numRead, -1);
    }

    @Test
    public void testNSPropigation() throws Exception {
        Request testRequest = getPopulatedRequest();
        Set<Datastream> datastreams = testRequest.getDatastreams(Commands.ADD);
        Datastream test_prop = null;
        Datastream format_test_prop = null;

        for (Datastream d : datastreams) {
            if (d.getName().getID().equals("test_prop")) {
                test_prop = d;
            } else if (d.getName().getID().equals("format_test_prop")) {
                format_test_prop = d;
            }
        }

        assertNotNull(test_prop);
        assertNotNull(format_test_prop);

        Document test_prop_doc = NDRXMLutil.getDocument(test_prop.getContent(),
                true, false);
        Document format_test_prop_doc = NDRXMLutil.getDocument(format_test_prop
                .getContent(), true, false);

        /* See if namespace was preserved for element */
        assertTrue(test_prop_doc.getDocumentElement().getNamespaceURI().equals(
                "TEST"));

        /* Try for an attribute */
        assertNotNull(test_prop_doc.getDocumentElement().getAttributeNS(
                "http://ns.nsdl.org/api/relationships#", "foo"));

        /* See if element namespace was preserved in "special case" format DS */
        assertTrue(format_test_prop_doc.getDocumentElement().getNamespaceURI()
                .equals("TEST"));
    }

    @Test
    public void binaryContentTest() throws Exception {
        Request testRequest = getBinaryPopulatedRequest();
        Set<Datastream> datastreams = testRequest.getDatastreams(Commands.ADD);
        byte[] readBytes;
        Datastream bds = null;
        for (Datastream d : datastreams) {
            if (d.getName().getID().equals(BINARY_DS_NAME)) {
                bds = d;
            }
        }

        /* Make sure we get the datastream */
        assertNotNull(bds);

        /* Make sure we can detect the mime type from the multipart declaration */
        assertEquals(BINARY_DS_MIME, bds.getMimeType());

        readBytes = StreamUtil.streamToByteArray(bds.getContent());

        /* Make sure the content is the same */
        assertTrue(Arrays.equals(readBytes, BINARY_CONTENT));

    }

    private Request getPopulatedRequest() throws Exception {
        String pathInfo = "addMetadata/";

        /* Set up the request parameters */
        request.setRequestURL(URL_BASE + pathInfo);
        String[] query = { basicQuery };
        request.setParameterValue(INPUT_XML, query);
        request.setPathInfo(pathInfo);
        XMLRequest testRequest = new XMLRequest(request);
        return testRequest;
    }

    private Request getBinaryPopulatedRequest() throws Exception {
        Part[] parts = new Part[2];
        String pathInfo = "addMetadata/";

        /* Set up the request parameters */
        request.setRequestURL(URL_BASE + pathInfo);
        String query = basicQuery.replace(BINARY_STREAM_PLACEHOLDER, "<"
                + BINARY_DS_NAME + " />\n");

        // parts[0] = new StringPart(INPUT_XML, query, "UTF-8");
        parts[0] = new FilePart(INPUT_XML, new ByteArrayPartSource(INPUT_XML,
                query.getBytes("UTF-8")));
        parts[1] = new BinaryPart(BINARY_DS_NAME, new ByteArrayPartSource(
                DS_CONTENT, BINARY_CONTENT), BINARY_DS_MIME);
        MultipartRequestEntity e = new MultipartRequestEntity(parts,
                new PostMethod().getParams());
        request.setContentType(e.getContentType());
        request.setPathInfo(pathInfo);

        ByteArrayOutputStream o = new ByteArrayOutputStream();
        e.writeRequest(o);
        request.setContent(o.toString("UTF-8"));
        XMLRequest testRequest = new XMLRequest(request);
        return testRequest;
    }

    private static class BinaryPart extends FilePart {
        public BinaryPart(String filename, PartSource source, String mime) {
            super(filename, source, mime, null);
        }

        public String getCharSet() {
            return null;
        }
    }

    private static class ContentSimulator extends HttpServletRequestSimulator {
        private String m_content = "";

        public ContentSimulator(ServletContext cxt) {
            super(cxt);
        }

        public void setContent(String content) {
            m_content = content;
        }

        public ServletInputStream getInputStream() {
            return new StringServletInputStream(m_content);
        }
    }

    /* The servlet spec is really icky about this. */
    private static class StringServletInputStream extends ServletInputStream {
        private final ByteArrayInputStream m_src;

        public StringServletInputStream(String content) {
            try {
                m_src = new ByteArrayInputStream(content.getBytes("UTF-8"));
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }

        public int read() {
            return m_src.read();
        }
    }
}
