/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.request.filters;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.api.request.MockRequest;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.api.request.Request.Commands;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.content.ModelUtil;
import org.nsdl.repository.model.content.ResourceModel;
import org.nsdl.repository.model.content.Scope;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.StaticObjectInfo;

import static junit.framework.Assert.*;
import static org.nsdl.repository.model.content.ResourceModel.Components.IDENTIFIER_URL;

public class AddResourceFilterTest {
    private static NDRAccess ndr;
    private static AddResourceFilter filter;

    private static Map<Property.Name, NDRObject> existingResources = new HashMap<Property.Name, NDRObject>();

    private static Collection<Property.Name> identifiers;

    @BeforeClass
    public static void setupNDR() throws Exception {
        ndr = MockFedoraNDRAccess.getInstance();

        identifiers = ModelUtil.getPropertyNames(
                ResourceModel.Identifiers.class, Scope.All);

        for (Property.Name idproperty : identifiers) {
            NDRObject resource = ndr.newObject(Type.Resource);

            resource.addProperties(Arrays.asList(new Property(resource,
                    idproperty, idproperty.getURI())));

            resource.commit();
            existingResources.put(idproperty, resource);
        }

        filter = new AddResourceFilter();
        filter.init(ndr);
    }

    @Test
    public void addRedundantResourceTest() {
        for (Property.Name idProperty : identifiers) {

            MockRequest request = new MockRequest();
            request.setType(Request.Type.Create);

            NDRObjectInfo toAdd = new StaticObjectInfo(Type.Resource);

            request.setObjectInfo(toAdd);

            request.putProperty(Commands.ADD, new Property(toAdd, idProperty,
                    idProperty.getURI()));

            filter.process(request);

            assertEquals("Filter changed the number of request elements!", 1,
                    request.getProperties(Commands.ADD).size());
            assertEquals("Filter did not point request to redundant resource!",
                    existingResources.get(idProperty).handle(), request
                            .getObjectInfo().handle());
        }
    }

    @Test
    public void addNonRedundantResourceTest() {
        for (Property.Name idProperty : identifiers) {

            MockRequest request = new MockRequest();
            request.setType(Request.Type.Create);

            NDRObjectInfo toAdd = new StaticObjectInfo(Type.Resource);

            String identifier = idProperty.getURI() + "#does_not_exist";

            request.setObjectInfo(toAdd);

            request.putProperty(Commands.ADD, new Property(toAdd, idProperty,
                    identifier));

            filter.process(request);

            assertEquals("Filter changed the number of request elements!", 1,
                    request.getProperties(Commands.ADD).size());
            assertNotSame(
                    "Filter should not have changed the requested resource!",
                    existingResources.get(idProperty).handle(), request
                            .getObjectInfo().handle());
        }
    }

    @Test
    public void createOnlyTest() {
        MockRequest request = new MockRequest();
        request.setType(Request.Type.Create);

        NDRObjectInfo toAdd = new StaticObjectInfo(Type.Resource);

        request.setObjectInfo(toAdd);

        request.putProperty(Commands.ADD, new Property(toAdd, IDENTIFIER_URL,
                IDENTIFIER_URL.getURI()));

        for (Request.Type rType : Arrays.asList(Request.Type.Delete,
                Request.Type.Read, Request.Type.Update)) {
            request.setType(rType);
            filter.process(request);

            assertNotSame(
                    "Filter should not have changed the requested resource!",
                    existingResources.get(IDENTIFIER_URL).handle(), request
                            .getObjectInfo().handle());
        }

        request.setType(Request.Type.Create);
        filter.process(request);
        assertEquals("Filter did not point request to redundant resource!",
                existingResources.get(IDENTIFIER_URL).handle(), request
                        .getObjectInfo().handle());

    }

    /* We should do something more elegant if there is no identifier */
    @Ignore
    @Test
    public void noIdentifierBarfingTest() {
        MockRequest request = new MockRequest();
        request.setType(Request.Type.Create);

        NDRObjectInfo toAdd = new StaticObjectInfo(Type.Resource);

        request.setObjectInfo(toAdd);

        filter.process(request);
    }
}
