/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.request.filters;

import java.util.Arrays;
import java.util.Set;

import org.junit.Before;
import org.junit.Test;
import org.nsdl.repository.api.request.MockRequest;
import org.nsdl.repository.api.request.RequestFilter;
import org.nsdl.repository.api.request.Request.Commands;
import org.nsdl.repository.api.request.Request.Type;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.impl.fedora.XMLDatastream;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.util.StaticObjectInfo;
import org.nsdl.util.xml.NDRXMLutil;

import static junit.framework.Assert.*;
import static org.nsdl.repository.model.types.Type.Metadata;
import static org.nsdl.repository.model.types.Type.Resource;
import static org.nsdl.repository.model.content.ResourceModel.Components.IDENTIFIER_URL;
import static org.nsdl.repository.model.content.MetadataModel.Components.NSDL_DC;
import static org.nsdl.repository.model.content.MetadataModel.Components.REFERENCES;

public class MetadataReferenceFilterTest {

    private MockFedoraNDRAccess ndr;
    private RequestFilter filter;
    private MockRequest request;
    private NDRObjectInfo requestObject;
    private NDRObject existingResource;
    private static final String EXISTING_URL = "http://nsdl.org";

    @Before
    public void setup() {
        ndr = MockFedoraNDRAccess.getInstance();
        filter = new MetadataReferenceFilter();
        filter.init(ndr);

        /* add the existing resource */
        existingResource = ndr.newObject(Resource);
        existingResource.addProperty(IDENTIFIER_URL, EXISTING_URL);
        existingResource.commit();

        /* create a default request */
        request = new MockRequest();
        requestObject = new StaticObjectInfo(Metadata);
        request.setType(Type.Create);
        request.setObject(requestObject);
    }

    /** Add references relationships to new metadata */
    @Test
    public void newMetadataReferencesTest() throws Exception {
        String metatata = "<nsdl_dc xmlns:dct=\"http://purl.org/dc/terms/\">"
                + " <dct:references>" + EXISTING_URL + "</dct:references>\n"
                + " <dct:references>http://nonexisting</dct:references>\n"
                + " <dct:isReferencedBy>nobody</dct:isReferencedBy>"
                + "</nsdl_dc>";
        Datastream metadata = new XMLDatastream(NSDL_DC, NDRXMLutil
                .getDocument(metatata, true, false), "DC", "application/xml");
        request.putDatastream(Commands.ADD, metadata);
        filter.process(request);

        Set<Relationship> addedRels = request.getRelationships(Commands.ADD);

        assertEquals("Wrong number of added relationships", 1, addedRels.size());
        Relationship added = addedRels.iterator().next();

        assertEquals(REFERENCES, added.predicate());
        assertEquals(existingResource.handle(), added.related().handle());
    }

    /**
     * If a references relationship exists in the object, but it is *not implied
     * by the metadata, remove it.
     * 
     * @throws Exception
     */
    @Test
    public void existingMetadataRemoveTest() throws Exception {
        NDRObject existingMD = ndr.newObject(Metadata);
        existingMD.addRelationship(REFERENCES, existingResource);
        existingMD.commit();
        request.setObjectInfo(existingMD);
        request.setType(Type.Update);

        String metatata = "<nsdl_dc xmlns:dct=\"http://purl.org/dc/terms/\">"
                + " <dct:references>http://nonexisting</dct:references>\n"
                + " <dct:isReferencedBy>nobody</dct:isReferencedBy>"
                + "</nsdl_dc>";
        Datastream metadata = new XMLDatastream(NSDL_DC, NDRXMLutil
                .getDocument(metatata, true, false), "DC", "application/xml");
        request.putDatastream(Commands.REPLACE, metadata);

        filter.process(request);

        Set<Relationship> toBeDeleted = request
                .getRelationships(Commands.DELETE);

        assertEquals(0, request.getRelationships(Commands.ADD).size());
        assertEquals(0, request.getRelationships(Commands.REPLACE).size());
        assertEquals(1, toBeDeleted.size());

        Relationship deletedRel = toBeDeleted.iterator().next();

        assertEquals(REFERENCES, deletedRel.predicate());
        assertEquals(existingResource.handle(), deletedRel.related().handle());
    }

    /** If no metadata is specified in the request, do nothing */
    @Test
    public void existingMetadataNeutralTest() {
        NDRObject existingMD = ndr.newObject(Metadata);
        existingMD.addRelationship(REFERENCES, existingResource);
        existingMD.commit();
        request.setObjectInfo(existingMD);

        filter.process(request);

        assertEquals(0, request.getRelationships(Commands.ADD).size());
        assertEquals(0, request.getRelationships(Commands.REPLACE).size());
        assertEquals(0, request.getRelationships(Commands.DELETE).size());
    }

    /** Make sure that we can add relationships to md that already has others */
    @Test
    public void existingMetadataAddTest() throws Exception {

        /* Add the existing metadata object */
        NDRObject existingMD = ndr.newObject(Metadata);
        existingMD.addRelationship(REFERENCES, existingResource);
        existingMD.commit();
        request.setObjectInfo(existingMD);

        String NEW_RESOURCE = "http://new.resource";
        String metatata = "<nsdl_dc xmlns:dct=\"http://purl.org/dc/terms/\">"
                + " <dct:references>" + EXISTING_URL + "</dct:references>\n"
                + " <dct:references>" + NEW_RESOURCE + "</dct:references>\n"
                + " <dct:isReferencedBy>nobody</dct:isReferencedBy>"
                + "</nsdl_dc>";
        Datastream metadata = new XMLDatastream(NSDL_DC, NDRXMLutil
                .getDocument(metatata, true, false), "DC", "application/xml");
        request.putDatastream(Commands.ADD, metadata);
        request.setType(Type.Update);

        /* Add a resources that matches the new one in the metadata */
        NDRObject newResource = ndr.newObject(Resource);
        newResource.addProperties(Arrays.asList(new Property(newResource,
                IDENTIFIER_URL, NEW_RESOURCE)));
        newResource.commit();

        filter.process(request);

        Set<Relationship> addedRels = request.getRelationships(Commands.ADD);

        assertEquals(1, addedRels.size());
        Relationship added = addedRels.iterator().next();

        assertEquals(REFERENCES, added.predicate());
        assertEquals(newResource.handle(), added.related().handle());
    }
}
