/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.util;

import static junit.framework.Assert.*;
import junit.framework.JUnit4TestAdapter;

import org.junit.After;
import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.api.Constants;
import org.nsdl.repository.error.RepositoryException;
import org.nsdl.repository.impl.fedora.FedoraUtils;
import org.nsdl.repository.impl.fedora.MockFedoraAccess;
import org.nsdl.repository.impl.fedora.risearch.MPTTestQueryEngine;
import org.nsdl.repository.model.types.Agent;
import org.nsdl.repository.model.types.Type;

import fedora.common.PID;

/**
 * Units tests covering PID-Handle translation
 * 
 * @author birkland
 * 
 */
public class FedoraUtilsPidTranslationTest {
    public static final String pidNS = "FedoraUtilsPidTranslationTest:";

    private static String HAS_HANDLE = Constants.REPOS_NAMESPACE + "hasHandle";
    private static String OBJECT_TYPE = Constants.REPOS_NAMESPACE
            + "objectType";
    private static PID matchingPID = PID.getInstance(pidNS + "matching");
    private static PID secondPID = PID.getInstance(pidNS + "second");
    private static String matchingHandle = "2200/FedoraUtilsPidTranslationTest.match";
    private static String secondHandle = "2200/FedoraUtilsPidTranslationTest.second";

    private static MockFedoraAccess fedora = new MockFedoraAccess();
    private static MPTTestQueryEngine ri;

    /** We clear the RI after every test */
    @After
    public void tearDown() {
        ri.reset();
    }

    @BeforeClass
    public static void initRI() {
        ri = MPTTestQueryEngine.getInstance();
        fedora.setRI(ri);
    }

    /** Set up the test RI for further manipulation.. */
    public static junit.framework.Test suite() {
        return new JUnit4TestAdapter(FedoraUtilsPidTranslationTest.class);
    }

    /** Tests finding a PID given a handle */
    @Test
    public void testPIDFromExistingHandle() throws Exception {

        /* Put in a matching and non-matching handle */
        TripleTesterUtils.useTripleLiteral(matchingPID.toURI(), HAS_HANDLE,
                matchingHandle, ri);
        TripleTesterUtils.useTripleLiteral(secondPID.toURI(), HAS_HANDLE,
                secondHandle, ri);

        PID foundPID = FedoraUtils.getPidFromHandle(fedora, matchingHandle);

        /* We should have found the matching PID */
        assertEquals(matchingPID.toString(), foundPID.toString());
    }

    /** Tests finding a PID given a handle and type */
    @Test
    public void testPIDFromExistingHandleType() throws Exception {

        Type<? extends Agent> type = Type.Agent;

        /* put in a matching PID, non-matchingPID, and make them both Agents */
        TripleTesterUtils.useTripleLiteral(matchingPID.toURI(), HAS_HANDLE,
                matchingHandle, ri);
        TripleTesterUtils.useTripleLiteral(secondPID.toURI(), HAS_HANDLE,
                secondHandle, ri);
        TripleTesterUtils.useTripleLiteral(matchingPID.toURI(), OBJECT_TYPE,
                type.toString(), ri);
        TripleTesterUtils.useTripleLiteral(secondPID.toURI(), OBJECT_TYPE, type
                .toString(), ri);

        PID foundPID = FedoraUtils.getPIDFromHandle(fedora, matchingHandle,
                type);

        /* We should have found the matching PID */
        assertEquals(matchingPID.toString(), foundPID.toString());

        /* We should not find it if the type is wrong.. */
        foundPID = FedoraUtils.getPIDFromHandle(fedora, matchingHandle,
                Type.Aggregator);
        assertNull(foundPID);

    }

    /**
     * Assure that we will raise exception if we encounter more than one PID
     * with a particular handle (should never happen under normal circumstances)
     */
    @Test
    public void testDuplicateHandleRejection() throws Exception {

        /* Add two different PIDS that have the same handle (a no-no!) */
        TripleTesterUtils.useTripleLiteral(matchingPID.toURI(), HAS_HANDLE,
                matchingHandle, ri);
        TripleTesterUtils.useTripleLiteral(secondPID.toURI(), HAS_HANDLE,
                matchingHandle, ri);

        try {
            FedoraUtils.getPidFromHandle(fedora, matchingHandle);
            fail("Did not catch duplicate handle");
        } catch (RepositoryException e) {
            /* Good. We want an exception here */
            assertTrue(true);
        }
    }

    /** Tests getting a handle from a PID */
    @Test
    public void testHandlefromPID() throws Exception {

        /* Put in a matching and non-matching handle */
        TripleTesterUtils.useTripleLiteral(matchingPID.toURI(), HAS_HANDLE,
                matchingHandle, ri);
        TripleTesterUtils.useTripleLiteral(secondPID.toURI(), HAS_HANDLE,
                secondHandle, ri);

        String foundHandle = FedoraUtils.getHandleFromPID(fedora, matchingPID);

        assertEquals(foundHandle, matchingHandle);
    }
}
