/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.api.util;

import java.util.List;

import javax.xml.namespace.QName;

import org.nsdl.repository.api.Constants;
import org.nsdl.repository.api.request.Request;
import org.nsdl.repository.api.request.RequestUtils;
import org.nsdl.util.xml.NDRXMLutil;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import junit.framework.TestCase;

/**
 * Doesn't cover evey possible combination, but these tests should cover most
 * types of cases encountered
 * 
 * @author birkland
 * 
 */
public class RequestUtilsTest extends TestCase {
    private final Element xml;

    public RequestUtilsTest() throws Exception {
        this.xml =
                NDRXMLutil.getDocument(xmlSource, true, false)
                        .getDocumentElement();
    }

    public void testNonDefaultAddProperties() {
        List<Node> properties =
                RequestUtils.getInputProperties(xml, Request.Commands.ADD,
                        false);

        /* there are two elements in the <add> block */
        assertEquals(2, properties.size());

        /*
         * We wrote the test XML so that all correct answers will be in the
         * addNS namespace
         */
        for (Node item : properties) {
            assertEquals(item.getNamespaceURI(), addNS);
        }
    }

    public void testDefaultAddProperties() {
        List<Node> properties =
                RequestUtils
                        .getInputProperties(xml, Request.Commands.ADD, true);

        /*
         * there are two elements in the add block, and two that aren't in any
         * command
         */
        assertEquals(4, properties.size());

        int flatCount = 0; /* properties not in a command */
        int addCount = 0; /* properties in an add command */
        for (Node item : properties) {
            if (item.getNamespaceURI().equals(addNS)) {
                addCount++;
            }
            if (item.getNamespaceURI().equals(flatNS)) {
                flatCount++;
            }
        }

        assertEquals(2, flatCount);
        assertEquals(2, addCount);
    }

    public void testAllProperties() {
        List<Node> properties =
                RequestUtils.getInputProperties(xml, null, true);

        /*
         * there are 2 elements in the add block, 0 in modify, 1 in delete, and
         * 2 that aren't in any command
         */
        assertEquals(5, properties.size());

        int flatCount = 0; /* properties not in a command */
        int addCount = 0; /* properties in an add command */
        int deleteCount = 0;/* properties in the delete command */
        for (Node item : properties) {
            if (item.getNamespaceURI().equals(addNS)) {
                addCount++;
            }
            if (item.getNamespaceURI().equals(flatNS)) {
                flatCount++;
            }
            if (item.getNamespaceURI().equals(deleteNS)) {
                deleteCount++;
            }
        }

        assertEquals(2, flatCount);
        assertEquals(2, addCount);
        assertEquals(1, deleteCount);
    }

    public void testModifyPropertiesEmpty() {
        List<Node> properties =
                RequestUtils.getInputProperties(xml, Request.Commands.REPLACE,
                        false);

        /* There are none */
        assertEquals(0, properties.size());
    }

    public void testFlatDefaultModifyDatastreams() {
        List<Node> datastreams =
                RequestUtils.getInputDatastreams(xml, Request.Commands.REPLACE,
                        true);

        /* There is one flat element that is included by default */

        assertEquals(1, datastreams.size());

        /* Make sure the element has its child attached */
        assertTrue(NDRXMLutil.getChildElement(((Element) datastreams.get(0)),
                flatNS, "d1a") != null);
    }

    public void testFlatNonDefaultModifyDatastreams() {
        List<Node> datastreams =
                RequestUtils.getInputDatastreams(xml, Request.Commands.REPLACE,
                        false);

        /* There are no datastreams in a modify command */
        assertEquals(0, datastreams.size());
    }

    public void testDefaultAddDatastreams() {
        List<Node> datastreams =
                RequestUtils.getInputDatastreams(xml, Request.Commands.ADD,
                        true);

        /*
         * There is one flat element that is included by default, and one 'add'
         * element
         */

        assertEquals(2, datastreams.size());

        /* Make sure we have the correct ones */
        for (Node item : datastreams) {
            if (item.getNamespaceURI().equals(flatNS)) {
                assertTrue(NDRXMLutil.getChildElement(((Element) item), flatNS,
                        "d1a") != null);
            } else {
                assertTrue(NDRXMLutil.getChildElement(((Element) item), addNS,
                        "d1a") != null);
            }
        }
    }

    public void testFlatDefaultDeleteRelationships() {
        List<Node> relationships =
                RequestUtils.getInputRelationships(xml,
                        Request.Commands.DELETE, true);

        /* There are two flat elements */
        assertEquals(2, relationships.size());

        /*
         * We wrote the test XML so that all correct answers will be in the
         * flatNS namespace
         */
        for (Node item : relationships) {
            assertEquals(item.getNamespaceURI(), flatNS);
        }
    }

    public void testBadComponent() {
        QName component = new QName("blah", "blah");
        List<Node> nothing =
                RequestUtils.getInputComponents(xml, component,
                        Request.Commands.ADD, true);
        assertEquals(0, nothing.size());
    }

    private static final String addNS = "http://ns.nsdl.org/test/add/";
    private static final String modifyNS = "http://ns.nsdl.org/test/modify/";
    private static final String deleteNS = "http://ns.nsdl.org/test/delete/";
    private static final String flatNS = "http://ns.nsdl.org/test/flat/";
    private static final String xmlSource =
            "<inputXML xmlns=\"" + Constants.REQUEST_NAMESPACE + "\" "
                    + " xmlns:flat=\"" + flatNS + "\"" + " xmlns:add=\""
                    + addNS + "\"" + " xmlns:modify=\"" + modifyNS + "\""
                    + " xmlns:delete=\"" + deleteNS + "\">\n"
                    + "	<properties>\n" + "		<flat:p1>value_p1</flat:p1>\n"
                    + "		<add>\n" + "			<add:p1>value_p1</add:p1>\n"
                    + "			<add:p2>value_p2</add:p2>\n" + "		</add>\n"
                    + "		<flat:p2>value_p2</flat:p2>\n" + "		<modify>\n"
                    + "		</modify>\n" + "		<delete>\n"
                    + "			<delete:p1>value_p1</delete:p1>\n" + "		</delete>\n"
                    + "	</properties>\n" + "	<data>\n" + "		<add>\n"
                    + "			<add:d1>\n" + "				<add:d1a>content</add:d1a>\n"
                    + "			</add:d1>\n" + "		</add>\n" + "		<flat:d1>\n"
                    + "			<flat:d1a>content</flat:d1a>\n" + "		</flat:d1>"
                    + "		<delete>\n" + "			<delete:d1>value_d1</delete:d1>\n"
                    + "		</delete>\n" + "	</data>\n" + "	<relationships>\n"
                    + "		<flat:r1>value_r1</flat:r1>\n" + "		<add>\n"
                    + "			<add:r1>value_p1</add:r1>\n"
                    + "			<add:r2>value_p2</add:r2>\n" + "		</add>\n"
                    + "		<flat:r2>value_r2</flat:r2>\n" + "		<modify>\n"
                    + "			<modify:r1>value_r1</modify:r1>\n"
                    + "			<modify:r2>value_r2</modify:r2>\n" + "		</modify>\n"
                    + "	</relationships>\n" + "</inputXML>";
}
