/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.text.ParseException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;

import org.nsdl.mptstore.rdf.Triple;
import org.nsdl.mptstore.util.NTriplesUtil;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.NDRObject.State;

import fedora.common.Constants;
import fedora.common.PID;

import static org.nsdl.repository.impl.fedora.ri.TripleUtils.nTripleEncodeLiteral;
import static org.nsdl.repository.impl.fedora.ri.TripleUtils.nTripleEncodeURI;

/**
 * Generates triples that are parsed from an parts of an {@link NDRObject}
 * <p>
 * Since this is for the purpose of supporting tests, not all the triples that
 * are usually generated for each object actully are - only those that are
 * relevant for tests. As the tests smature, the number of triple types that
 * this class does not generate will decrease.
 * </p>
 * 
 * @author birkland
 * 
 */
public class ComponentTripleGenerator {

    /**
     * Get triples that correspond to object proprties
     * 
     * @param properties
     *            The properties to convert to triples
     * @return Iterator containing all triples
     */
    public static Iterator<Triple> getPropertyTriples(
            Iterable<Property> properties, FedoraNDRAccess ndr) {
        return new PropertyTripleIterator(properties, ndr);
    }

    /**
     * Get triples that correspond to relationships
     * 
     * @param rels
     *            The relationships to convert to triples
     * @return Iterator containing all triples
     */
    public static Iterator<Triple> getRelationshipTriples(
            Iterable<Relationship> rels, FedoraNDRAccess ndr) {
        return new RelationshipTripleIterator(rels, ndr);
    }

    /**
     * Get triples that correspond to datastreams
     * <p>
     * Actual parsing not implemented yet
     * </p>
     */
    public static Iterator<Triple> getDatastreamTriples(Iterable<Datastream> ds) {
        return new DatastreamTripleIterator(ds);
    }

    /**
     * Get triples that are due to object-level properties
     * <p>
     * Actual parsing not implemented yet
     * </p>
     */
    public static Iterator<Triple> getObjectLevelTriples(FedoraNDRObject o) {
        return new ObjectTripleIterator(o);
    }

    public static Iterator<Triple> getStateTriples(FedoraNDRObject o,
            NDRObject.State s) {
        return new ObjectTripleIterator(o, s);
    }

    private static class PropertyTripleIterator implements Iterator<Triple> {

        private final Iterator<Property> propIterator;
        private final FedoraNDRAccess ndr;

        public PropertyTripleIterator(Iterable<Property> props,
                FedoraNDRAccess fedora) {
            this.propIterator = props.iterator();
            this.ndr = fedora;
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }

        public boolean hasNext() {
            return propIterator.hasNext();
        }

        public Triple next() {
            Property p = propIterator.next();
            if (p == null) {
                return null;
            }
            try {
                Triple t = new Triple(NTriplesUtil
                        .parseSubject(nTripleEncodeURI(getSubjectPid(
                                p.subject(), ndr).toURI())), NTriplesUtil
                        .parsePredicate(nTripleEncodeURI(p.predicate()
                                .getNamespaceURI()
                                + p.predicate().getLocalPart())), NTriplesUtil
                        .parseObject(nTripleEncodeLiteral(p.value())));
                return t;
            } catch (Exception e) {
                throw new RuntimeException("P SUBJECT" + p.subject(), e);
            }
        }
    }

    private static class RelationshipTripleIterator implements Iterator<Triple> {

        private final Iterator<Relationship> relIterator;
        private final FedoraNDRAccess ndr;

        public RelationshipTripleIterator(Iterable<Relationship> rels,
                FedoraNDRAccess ndr) {
            this.relIterator = rels.iterator();
            this.ndr = ndr;
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }

        public boolean hasNext() {
            return relIterator.hasNext();
        }

        public Triple next() {
            Relationship r = relIterator.next();
            if (r == null) {
                return null;
            }
            try {
                Triple t = new Triple(NTriplesUtil
                        .parseSubject(nTripleEncodeURI(getSubjectPid(
                                r.subject(), ndr).toURI())), NTriplesUtil
                        .parsePredicate(nTripleEncodeURI(r.predicate()
                                .getNamespaceURI()
                                + r.predicate().getLocalPart())), NTriplesUtil
                        .parseObject(nTripleEncodeURI(ndr
                                .getObject(r.related()).pid().toURI())));
                return t;
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
    }

    private static class DatastreamTripleIterator implements Iterator<Triple> {

        public DatastreamTripleIterator(Iterable<Datastream> d) {
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }

        public boolean hasNext() {
            return false;
        }

        public Triple next() {
            return null;
        }
    }

    private static class ObjectTripleIterator implements Iterator<Triple> {

        private final Queue<Triple> buffer = new LinkedList<Triple>();

        public ObjectTripleIterator(FedoraNDRObject o, State s) {
            try {
                /* Right now we just parse state */
                buffer
                        .add(new Triple(
                                NTriplesUtil.parseSubject(nTripleEncodeURI(o
                                        .pid().toURI())),
                                NTriplesUtil
                                        .parsePredicate(nTripleEncodeURI(Constants.MODEL.STATE.uri)),
                                NTriplesUtil.parseObject(nTripleEncodeURI(s
                                        .getURI()))));
            } catch (ParseException e) {
                throw new RuntimeException(e);
            }
        }

        public ObjectTripleIterator(FedoraNDRObject o) {
            try {
                /* Right now we just parse state */
                buffer
                        .add(new Triple(
                                NTriplesUtil.parseSubject(nTripleEncodeURI(o
                                        .pid().toURI())),
                                NTriplesUtil
                                        .parsePredicate(nTripleEncodeURI(Constants.MODEL.STATE.uri)),
                                NTriplesUtil.parseObject(nTripleEncodeURI(o
                                        .state().getURI()))));
            } catch (ParseException e) {
                throw new RuntimeException(e);
            }
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }

        public boolean hasNext() {
            return !buffer.isEmpty();
        }

        public Triple next() {
            Triple t = buffer.remove();
            return t;
        }
    }

    private static PID getSubjectPid(NDRObjectInfo o, FedoraNDRAccess ndr) {
        PID subj = null;
        if (o instanceof MockFedoraNDRAccess.MockMutableObject) {
            subj = ((MockFedoraNDRAccess.MockMutableObject) o).pid();
        } else {
            subj = ndr.getObject(o).pid();
        }

        return subj;
    }
}
