/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.util.Iterator;

import org.junit.After;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.impl.fedora.FedoraNDRObjectFinder;
import org.nsdl.repository.impl.fedora.risearch.MPTTestQueryEngine;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.content.MetadataModel;
import org.nsdl.repository.model.content.OAIModel;
import org.nsdl.repository.model.content.ResourceModel;
import org.nsdl.repository.model.types.Type;

import junit.framework.JUnit4TestAdapter;
import static junit.framework.Assert.*;

public class FedoraNDRObjectFinderTest {

    private static MockFedoraAccess fedora = new MockFedoraAccess();

    private static MockFedoraNDRAccess ndr = new MockFedoraNDRAccess(fedora);

    private static FedoraNDRObjectFinder finder = new FedoraNDRObjectFinder(ndr);

    private NDRObject testMetadata;

    /** We clear the RI after every test */
    @After
    public void resetRepositoryContents() {
        fedora.getRI().reset();
        ndr = new MockFedoraNDRAccess(fedora);
    }

    @Before
    public void populate() {
        NDRObject r = ndr.newObject(Type.Resource);
        r.commit();

        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.commit();

        NDRObject m = ndr.newObject(Type.Metadata);
        m.addProperty(OAIModel.Item.Components.ITEM_ID, "XYZ");

        m.addRelationship(MetadataModel.Components.METADATA_FOR, r);
        m.addRelationship(MetadataModel.Components.METADATA_PROVIDED_BY, mp);

        m.commit();

        this.testMetadata = m;

        finder = new FedoraNDRObjectFinder(ndr);
    }

    @BeforeClass
    public static void initRI() {
        fedora.setRI(MPTTestQueryEngine.getInstance());
        ndr = new MockFedoraNDRAccess(fedora);
    }

    /** Set up the test RI for further manipulation.. */
    public static junit.framework.Test suite() {
        return new JUnit4TestAdapter(FedoraNDRObjectFinderTest.class);
    }

    /** Find an object that we know exists */
    @Test
    public void testSimpleFindURL() {
        String url = "http://to.find";
        NDRObject toFind = ndr.newObject(Type.Resource);
        toFind.addProperty(ResourceModel.Components.IDENTIFIER_URL, url);

        toFind.commit();

        assertNotNull(finder.findResource(
                ResourceModel.Components.IDENTIFIER_URL, url));
    }

    /** Try to find an object that does not exist */
    @Test
    public void testNoFindURL() {
        String url = "http://to.find";
        NDRObject toFind = ndr.newObject(Type.Resource);
        toFind.addProperty(ResourceModel.Components.IDENTIFIER_URL, url);

        toFind.commit();

        assertNull(finder.findResource(ResourceModel.Components.IDENTIFIER_URL,
                "http://do.not.find"));
    }

    @Test
    public void testMatchObject() {

        Iterator<NDRObjectInfo> results = finder.matchObject(testMetadata)
                .iterator();
        assertTrue(results.hasNext());
    }

    @Test
    public void matchUTF8() {
        NDRObject m = ndr.newObject(Type.Metadata);
        m.addProperty(MetadataModel.Components.UNIQUE_ID, "dấu móc☞");
        m.commit();

        Iterator<NDRObjectInfo> results = finder.matchObject(m).iterator();
        assertTrue("failed UTF-8 match", results.hasNext());
    }

    @Test
    public void matchQuotes() {
        NDRObject m = ndr.newObject(Type.Metadata);
        m.addProperty(MetadataModel.Components.UNIQUE_ID, "123\"45\"678");
        m.commit();

        Iterator<NDRObjectInfo> results = finder.matchObject(m).iterator();
        assertTrue("failed quotation match", results.hasNext());
    }
}
