/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.nsdl.repository.impl.fedora.FedoraNDRObjectFinder;
import org.nsdl.repository.impl.fedora.risearch.MPTTestQueryEngine;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.content.ContentModelException;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.model.types.TypeFactory;
import org.nsdl.repository.util.StaticObjectInfo;

import fedora.common.PID;

/**
 * Simple memory-based NDR instance
 * <p>
 * Insert docs here
 * </p>
 * 
 * @author birkland
 * 
 */
public class MockFedoraNDRAccess implements FedoraNDRAccess {

    private final MockFedoraAccess fedora;
    private final FedoraNDRObjectFinder finder;
    private final FedoraNDRAccess ndr;
    private int handleSequence;
    private final TypeFactory factory = new TypeFactory(this);

    private final Map<String, MockMutableObject> objects = new HashMap<String, MockMutableObject>();

    public MockFedoraNDRAccess(MockFedoraAccess fedora) {
        this.fedora = fedora;
        this.finder = new FedoraNDRObjectFinder(this);
        this.ndr = this;
        handleSequence = 1;
    }

    public MockMutableObject newObject(Type<? extends NDRObject> type) {
        MockMutableObject m = new MockMutableObject(type, "2200/"
                + ++handleSequence);
        return m;
    }

    public <T extends NDRObject> T newObject(Type<T> type) {
        MockMutableObject m = new MockMutableObject(type, "2200/"
                + ++handleSequence);
        return factory.instantiate(m, type);
    }

    public <T extends FedoraNDRObject> T newObject(Type<T> type) {
        MockMutableObject m = new MockMutableObject(type, "2200/"
                + ++handleSequence);
        return factory.instantiate(m, type);
    }

    public MockMutableObject newObject(NDRObjectInfo info) {
        return newObject(info.type(), info.handle());
    }

    public MockMutableObject newObject(Type<? extends NDRObject> type,
            String handle) {
        MockMutableObject m = new MockMutableObject(type, handle);
        return m;
    }

    public void deleteObject(NDRObjectInfo object) throws ExistenceException {
        MockMutableObject m = openObject(object);
        m.setState(NDRObject.State.Deleted);
        m.commit();
    }

    public MockMutableObject getObject(NDRObjectInfo identity)
            throws ExistenceException {
        return this.openObject(identity);
    }

    public <T extends NDRObject> T getObject(NDRObjectInfo identity,
            Type<T> type) {
        NDRObject fetched = getObject(identity);
        if (fetched != null) {
            return factory.instantiate(getObject(identity), type);
        } else {
            return null;
        }
    }

    public <T extends NDRObject> T getObject(String identity, Type<T> type) {
        NDRObject fetched = getObject(identity);
        if (fetched != null) {
            return factory.instantiate(getObject(identity), type);
        } else {
            return null;
        }
    }

    public MockMutableObject getObject(PID pid) {
        String handle = pid.toString().replace("test:", "").replace("%2F", "/");
        return getObject(new StaticObjectInfo(handle));
    }

    public MockMutableObject getObject(String identifier) {
        return getObject(new StaticObjectInfo(identifier));
    }

    public MockMutableObject openObject(NDRObjectInfo object)
            throws ExistenceException {
        if (!objects.containsKey(object.handle())) {
            String handles = "";
            for (String handle : objects.keySet()) {
                handles += " " + handle;
            }
            throw new ExistenceException("No object " + object.handle() + " "
                    + object.getClass() + ", handles are " + handles);
        } else {
            return new MockMutableObject(objects.get(object.handle()));
        }
    }

    public <T extends NDRObject> T openObject(NDRObjectInfo object, Type<T> type) {
        return factory.instantiate(openObject(object), type);
    }

    public <T extends FedoraNDRObject> T openObject(NDRObjectInfo object,
            Type<T> type) {
        return factory.instantiate(openObject(object), type);
    }

    public void purgeObject(NDRObjectInfo object) throws ExistenceException,
            ContentModelException {
        MockMutableObject m = openObject(object);
        m.purge();
    }

    public FedoraNDRObjectFinder getObjectFinder() {
        return finder;
    }

    public FedoraAccess getRepository() {
        return this.fedora;
    }

    public boolean isReadOnly() {
        return false;
    }

    class MockMutableObject extends FedoraMutableObjectBase implements
            FedoraNDRObject {

        private boolean isNew;
        private Date created;
        private Date lastModified;

        public MockMutableObject(Type<? extends NDRObject> type, String handle) {
            initNew(type, handle, PID.getInstance("test:"
                    + handle.replace("/", "%2F")));
            this.isNew = true;
        }

        public MockMutableObject(FedoraNDRObject o) {
            init(o);
            this.created = o.getCreatedDate();
            this.lastModified = o.getLastModifiedDate();
            this.isNew = false;
        }

        public FedoraDatastream getFedoraDatastream(String name) {
            throw new UnsupportedOperationException();
        }

        public Date getLastModifiedDate() {
            return lastModified;
        }

        public Date getCreatedDate() {
            return created;
        }

        public void commit() {

            MockMutableObject existing = null;
            if (isNew) {
                if (objects.containsKey(handle())) {
                    throw new IllegalStateException("An object already exists "
                            + "with handle " + handle());
                }
                objects.put(handle(), new MockMutableObject(this));

                existing = objects.get(handle());
                Date newCreated = new Date();

                this.created = newCreated;
                this.lastModified = newCreated;
                existing.created = newCreated;
                existing.lastModified = newCreated;
                isNew = false;
            } else {
                existing = objects.get(handle());
                Date newModified = new Date();

                this.lastModified = newModified;
                existing.lastModified = newModified;

            }

            existing.removeProperties(getPropertyTracker().getDeleted());
            existing.removeDatastreams(getDatastreamTracker().getDeleted());
            existing.removeRelationships(getRelationshipTracker().getDeleted());

            existing.addProperties(getPropertyTracker().getAdded());
            existing.addDatastreams(getDatastreamTracker().getAdded());
            existing.addRelationships(getRelationshipTracker().getAdded());

            existing.setState(state());
            existing.resetTrackers();

            /* Delete old component-level triples */
            fedora.getRI().remove(
                    ComponentTripleGenerator.getPropertyTriples(
                            getPropertyTracker().getDeleted(), ndr));
            fedora.getRI().remove(
                    ComponentTripleGenerator.getRelationshipTriples(
                            getRelationshipTracker().getDeleted(), ndr));
            fedora.getRI().remove(
                    ComponentTripleGenerator
                            .getDatastreamTriples(getDatastreamTracker()
                                    .getDeleted()));

            /* Add the new component-level triples */
            fedora.getRI().populate(
                    ComponentTripleGenerator.getPropertyTriples(
                            getPropertyTracker().getAdded(), ndr));
            fedora.getRI().populate(
                    ComponentTripleGenerator.getRelationshipTriples(
                            getRelationshipTracker().getAdded(), ndr));
            fedora.getRI().populate(
                    ComponentTripleGenerator
                            .getDatastreamTriples(getDatastreamTracker()
                                    .getAdded()));

            /* Deal with state, if necessary */

            if (getPreviousState() != null) {
                fedora.getRI().remove(
                        ComponentTripleGenerator.getStateTriples(this,
                                getPreviousState()));

                fedora.getRI()
                        .populate(
                                ComponentTripleGenerator.getStateTriples(this,
                                        state()));
            }

            resetTrackers();
        }

        public void purge() {
            fedora.getRI().remove(
                    ComponentTripleGenerator.getObjectLevelTriples(this));
            fedora.getRI().remove(
                    ComponentTripleGenerator.getPropertyTriples(
                            getProperties(), ndr));
            fedora.getRI().remove(
                    ComponentTripleGenerator.getRelationshipTriples(
                            getRelationships(), ndr));
            fedora.getRI().remove(
                    ComponentTripleGenerator
                            .getDatastreamTriples(getDatastreams()));
            objects.remove(this.handle());
        }
    }

    public static MockFedoraNDRAccess getInstance() {
        MockFedoraAccess fedora = new MockFedoraAccess();
        fedora.setRI(MPTTestQueryEngine.getInstance());
        return new MockFedoraNDRAccess(fedora);
    }

    public FedoraDatastreamFactory getDatastreamFactory() {
        return new FedoraDatastreamFactory();
    }
}
