/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.impl.fedora.risearch;

import org.apache.commons.dbcp.BasicDataSource;
import org.apache.commons.dbcp.BasicDataSourceFactory;
import org.nsdl.mptstore.core.BasicTableManager;
import org.nsdl.mptstore.core.DDLGenerator;
import org.nsdl.mptstore.core.GenericDatabaseAdaptor;
import org.nsdl.mptstore.core.TableManager;
import org.nsdl.mptstore.query.QueryException;
import org.nsdl.mptstore.query.QueryResults;
import org.nsdl.mptstore.query.SQLUnionQueryResults;
import org.nsdl.mptstore.query.component.GraphQuery;
import org.nsdl.mptstore.query.lang.QueryCompiler;
import org.nsdl.mptstore.query.lang.spo.SPOQueryCompiler;
import org.nsdl.mptstore.query.provider.GraphQuerySQLProvider;
import org.nsdl.mptstore.query.provider.SQLProvider;
import org.nsdl.mptstore.rdf.Triple;
import org.nsdl.repository.impl.fedora.ri.PreparedQuery;
import org.nsdl.repository.impl.fedora.ri.PreparedSQLUnionQueryResults;

import javax.sql.DataSource;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

public class MPTTestQueryEngine implements TestRIQueryEngine {
	public static final String PROP_JDBC_URI = "test.ri.mpt.db.jdbcURL";
	public static final String PROP_DB_USER = "test.ri.mpt.db.user";
	public static final String PROP_DB_PASSWD = "test.ri.mpt.db.passwd";
	public static final String PROP_DB_DRIVER_CLASSNAME = "test.ri.mpt.db.driverClassName";
	/** name of map table in mptstore */
	public static final String PROP_MAP = "test.ri.mpt.map";
	/** prefix used to create new tables */
	public static final String PROP_PREFIX = "test.ri.mpt.prefix";
	/** link between mptstore and the database implementation */
	public static final String PROP_DDL_GENERATOR_CLASSNAME = "test.ri.mpt.DDLGenerator";

	private TableManager tableManager = null;
	private GenericDatabaseAdaptor adaptor = null;
	private DataSource dataSource = null;
	private boolean backslashIsEscape = false;

	// ----------------------------------------------------- RIQueryEngine
	// methods

	/** execute a graphQuery */
	public QueryResults execute(GraphQuery graphQ, List<String> targets)
			throws QueryException {

		GraphQuerySQLProvider sqlGen = new GraphQuerySQLProvider(tableManager,
				graphQ, backslashIsEscape);
		sqlGen.setTargets(targets);

		try {
			return new SQLUnionQueryResults(dataSource.getConnection(), sqlGen,
					100, true);
		} catch (SQLException e) {
			throw new RuntimeException("Could not connect to data source", e);
		}
	}

	public QueryResults execute(PreparedQuery q, List<String> params)
			throws QueryException {
		PreparedSQLUnionQueryResults prepared = new PreparedSQLUnionQueryResults(
				q, 100, true);
		try {
			prepared.execute(dataSource.getConnection(), params);
		} catch (SQLException e) {
			throw new QueryException("Could not connect to data source");
		}
		return prepared;
	}

	public QueryResults execute(String query, Language lang)
			throws QueryException {
		switch (lang) {
		case SPO:
			return execSPO(query);
		}
		throw new IllegalArgumentException("The language " + lang
				+ " is not supported");
	}

	public PreparedQuery prepare(GraphQuery q, List<String> targets) throws QueryException {
        GraphQuerySQLProvider sqlGen = new GraphQuerySQLProvider(tableManager,
                q, backslashIsEscape);
            sqlGen.setTargets(targets);
            
            return new PreparedQuery(sqlGen, PreparedQuery.PLACEHOLDER);
    }
	/**
	 * initialize a query engine according to the properties initializes
	 * tableManager and dataSource
	 */
	public void init(Properties props) {
		/* Initialize the DDL generator */
		DDLGenerator generator;
		try {
			generator = (DDLGenerator) Class.forName(
					props.getProperty(PROP_DDL_GENERATOR_CLASSNAME))
					.newInstance();
		} catch (Exception e) {
			throw new RuntimeException("Could not initialize DDL generator", e);
		}

		/* Initialize database connection */
		BasicDataSource source;
		Properties dbParams = new Properties();

		dbParams.setProperty("url", props.getProperty(PROP_JDBC_URI));
		dbParams.setProperty("username", props.getProperty(PROP_DB_USER));
		dbParams.setProperty("password", props.getProperty(PROP_DB_PASSWD));
		dbParams.setProperty("driverClassName", props
				.getProperty(PROP_DB_DRIVER_CLASSNAME));
		try {
			Class.forName(props.getProperty(PROP_DB_DRIVER_CLASSNAME));
			source = (BasicDataSource) BasicDataSourceFactory
					.createDataSource(dbParams);
		} catch (Exception e) {
			throw new RuntimeException(
					"Could not establish database connection", e);
		}
		this.dataSource = source;

		/* Finally, create the table manager */
		BasicTableManager manager;
		String mapTable = props.getProperty(PROP_MAP);
		String prefix = props.getProperty(PROP_PREFIX);
		try {
			manager = new BasicTableManager(source, generator, mapTable, prefix);
		} catch (SQLException e) {
			throw new RuntimeException("Could not initialize table mapper", e);
		}
		tableManager = manager;

		adaptor = new GenericDatabaseAdaptor(manager, false);

	}

	// ------------------------------------------------- TestRIQueryEngine
	// methods

	/** add all triples in set to triple store */
	public void populate(Iterator<Triple> triples) {
		doAction(triples, TripleAction.add);
	}

	/** Removes these triples from the triple store */
	public void remove(Iterator<Triple> triples) {
		doAction(triples, TripleAction.delete);
	}

	private void doAction(Iterator<Triple> triples, TripleAction action) {
		Connection c;
		try {
			c = dataSource.getConnection();
		} catch (SQLException e) {
			throw new RuntimeException(
					"Could not get connection to datasttore", e);
		}

		try {
			c.setAutoCommit(false);
			switch (action) {
			case delete:
				adaptor.deleteTriples(c, triples);
				break;
			case add:
				adaptor.addTriples(c, triples);
				break;
			}
		} catch (Exception e) {
			try {
				c.rollback();
			} catch (Exception e2) {
			}
			throw new RuntimeException("Could not " + action + "  triples", e);
		} finally {
			try {
				c.commit();
			} catch (Exception e) {
				throw new RuntimeException("Could not " + action + " triples",
						e);
			}

			try {
				c.close();
			} catch (Exception e) {
			}
		}
	}

	/** remove all triples from triple store */
	public void reset() {
		Connection c = null;
		try {
			c = dataSource.getConnection();
			c.setAutoCommit(true);
			adaptor.deleteAllTriples(c);
		} catch (Exception e) {
			throw new RuntimeException("Could not clear triples!", e);
		} finally {
			try {
				c.close();
			} catch (Exception e) {
			}
		}
	}
	
	private QueryResults execSPO(String query) throws QueryException {
        QueryCompiler compiler = new SPOQueryCompiler(tableManager, true);
        SQLProvider compiled = compiler.compile(query);
        Connection connection = null;
        try {
            connection = dataSource.getConnection();
            connection.setAutoCommit(false);
        } catch (SQLException e) {
            throw new RuntimeException("Could not connect to data source");
        }
        
        return new SQLUnionQueryResults(connection, compiled, 100, true);
        
    }
	
	public static MPTTestQueryEngine getInstance() {
		MPTTestQueryEngine ri = new MPTTestQueryEngine();
		Properties props = new Properties();
		props.setProperty(MPTTestQueryEngine.PROP_DB_DRIVER_CLASSNAME,
				"org.h2.Driver");
		props.setProperty(MPTTestQueryEngine.PROP_DB_USER, "test");
		props.setProperty(MPTTestQueryEngine.PROP_DB_PASSWD, "test");
		props.setProperty(MPTTestQueryEngine.PROP_DDL_GENERATOR_CLASSNAME,
				"org.nsdl.mptstore.impl.h2.H2DDLGenerator");
		props.setProperty(MPTTestQueryEngine.PROP_JDBC_URI,
				"jdbc:h2:mem:XMLRequestTest");
		props.setProperty(MPTTestQueryEngine.PROP_MAP, "tmap");
		props.setProperty(MPTTestQueryEngine.PROP_PREFIX, "t");
		ri.init(props);
		ri.reset();
		return ri;
	}

	private enum TripleAction {
		add, delete
	}

}
