/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import java.util.Arrays;

import junit.framework.Assert;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.AgentModel.Components.*;

public class AgentModelTest {

    private static NDRAccess ndr;
    private static AgentModel model;

    @BeforeClass
    public static void setUp() {
        ndr = MockFedoraNDRAccess.getInstance();
        model = new AgentModel();
        model.init(ndr);
    }

    @Test
    public void basicAcceptanceTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);

        Property identifier = new Property(agent, IDENTIFIER_HOST, "host");

        agent.addProperties(Arrays.asList(identifier));

        model.checkCompliance(agent);
    }

    @Test(expected = ContentModelException.class)
    public void exclusiveIdentifierTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Resource);

        Property identifier = new Property(agent, IDENTIFIER_HOST, "host");

        agent.addProperties(Arrays.asList(identifier));

        model.checkCompliance(agent);
    }

    @Test
    public void passUnrelatedobjectTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.addProperty(AgentModel.Components.IDENTIFIER_HOST,
                "passUnrelatedObjectsTest:host");
        agent.commit();

        NDRObject agg = ndr.newObject(Type.Aggregator);
        agg.addRelationship(AggregatorModel.Components.AGGREGATOR_FOR, agent);
        agg.addProperty(OAIModel.Set.Components.SET_SPEC,
                "passUnrelatedObjectsTest:aggSpec");
        agg.commit();

        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(ResourceModel.Components.IDENTIFIER_URL,
                "http://test");
        resource.addRelationship(Components.MEMBER_OF, agg);
        resource.commit();

        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.addRelationship(
                MetadataProviderModel.Components.METADTA_PROVIDER_FOR, agent);
        mp.addRelationship(AuthorizationModel.Grouping.Components.AUTHORIZED,
                agent);
        mp.addProperty(OAIModel.Set.Components.SET_SPEC,
                "passUnrelatedObjectsTest:mpSpec");
        mp.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addRelationship(MetadataModel.Components.METADATA_FOR,
                resource);
        metadata.addRelationship(MetadataModel.Components.METADATA_PROVIDED_BY,
                mp);
        metadata.addProperty(MetadataModel.Components.UNIQUE_ID,
                "passUnrelatedObjectsTest:uniqueId");
        metadata.commit();

        model.checkCompliance(agent);
        model.checkCompliance(agg);
        model.checkCompliance(resource);
        model.checkCompliance(metadata);
        model.checkCompliance(mp);
    }

    @Test
    public void passSharedPropertyTest() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);

        Property identifier = new Property(resource, IDENTIFIER_OTHER, "host");

        resource.addProperties(Arrays.asList(identifier));

        model.checkCompliance(resource);
    }

    @Test(expected = ContentModelException.class)
    public void incompleteObjectRejectionTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);

        model.checkCompliance(agent);
    }

    @Test(expected = ContentModelException.class)
    public void tooManyIdentifierTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);

        Property identifier = new Property(agent, IDENTIFIER_HOST, "host");
        Property user = new Property(agent, IDENTIFIER_USER, "user");

        agent.addProperties(Arrays.asList(identifier, user));

        model.checkCompliance(agent);
    }

    @Test(expected = ContentModelException.class)
    public void wrongIdentifierTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);

        Property identifier = new Property(agent,
                ResourceModel.Components.IDENTIFIER_URL, "host");

        agent.addProperties(Arrays.asList(identifier));

        model.checkCompliance(agent);
    }

    @Test(expected = ContentModelException.class)
    public void identifierUniquenessTest() throws ContentModelException {
        String id = "info:test/identifierUniquenessTest";

        NDRObject existingAgent = ndr.newObject(Type.Agent);
        existingAgent.addProperty(IDENTIFIER_HOST, id);
        existingAgent.commit();

        NDRObject newAgent = ndr.newObject(Type.Agent);
        newAgent.addProperty(IDENTIFIER_HOST, id + 2);

        try {
            model.checkCompliance(existingAgent);
            model.checkCompliance(newAgent);
        } catch (ContentModelException e) {
            Assert.fail("Should not fail yet!");
        }

        newAgent.removeProperties(newAgent.getProperties(IDENTIFIER_HOST));

        newAgent.addProperty(IDENTIFIER_HOST, id);

        /* Should FAIL */
        model.checkCompliance(newAgent);
    }

    @Test
    public void identifierUniquenessInactiveTest() throws ContentModelException {
        String id = "info:test/identifierUniquenessInactiveTest";
        NDRObject existingAgent = ndr.newObject(Type.Agent);
        existingAgent.setState(State.Inactive);
        existingAgent.addProperty(IDENTIFIER_HOST, id);

        existingAgent.commit();

        NDRObject newAgent = ndr.newObject(Type.Agent);
        newAgent.setState(State.Active);
        newAgent.addProperty(IDENTIFIER_HOST, id);

        /* Should be OK since existing agent is inactive */
        model.checkCompliance(existingAgent);
        model.checkCompliance(newAgent);

        newAgent.setState(State.Inactive);
        model.checkCompliance(newAgent);

        /* Should be OK since new agent is inactive */
        existingAgent.setState(State.Active);
        existingAgent.commit();
        model.checkCompliance(newAgent);
    }

    @Test
    public void unrelatedElementsTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);

        agent.addProperty(IDENTIFIER_HOST, "host");
        agent.addProperty(new Property.Name("info:test/", "nonexistant", "no"),
                "value");
        agent.addRelationship(new Relationship.Name("info:test/",
                "nonexistant_relationship", "no"), agent);
        model.checkCompliance(agent);
    }
}
