/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import java.util.Arrays;

import junit.framework.Assert;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.AggregatorModel.Components.*;

public class AggregatorModelTest {
    private static NDRAccess ndr;
    private static AggregatorModel model;
    private static NDRObject agent;

    @BeforeClass
    public static void setUp() throws Exception {
        ndr = MockFedoraNDRAccess.getInstance();

        /* Initialize the model */
        model = new AggregatorModel();
        model.init(ndr);

        agent = ndr.newObject(Type.Agent);
        agent.commit();
    }

    @Test
    public void basicAcceptanceTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        Relationship aggregatorFor = new Relationship(aggregator,
                AGGREGATOR_FOR, agent);

        aggregator.addRelationships(Arrays.asList(aggregatorFor));

        model.checkCompliance(aggregator);
    }

    @Test
    public void passUnrelatedObjectTest() throws ContentModelException {
        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);

        Relationship mp4 = new Relationship(metadataProvider,
                MetadataProviderModel.Components.METADTA_PROVIDER_FOR, agent);

        metadataProvider.addRelationships(Arrays.asList(mp4));

        model.checkCompliance(metadataProvider);
    }

    @Test(expected = ContentModelException.class)
    public void exclusiveRelationshipsTest() throws ContentModelException {

        NDRObject mp = ndr.newObject(Type.MetadataProvider);

        Relationship mp4 = new Relationship(mp, AGGREGATOR_FOR, agent);

        mp.addRelationships(Arrays.asList(mp4));

        model.checkCompliance(mp);
    }

    @Test(expected = ContentModelException.class)
    public void noAgentsTest() throws ContentModelException {
        NDRObject metadataProvider = ndr.newObject(Type.Aggregator);

        model.checkCompliance(metadataProvider);
    }

    @Test(expected = ContentModelException.class)
    public void wrongAgentStateTest() throws ContentModelException {

        NDRObject myAgent = ndr.newObject(Type.Agent);
        myAgent.setState(State.Inactive);
        myAgent.commit();

        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        Relationship aggregatorFor = new Relationship(aggregator,
                AGGREGATOR_FOR, myAgent);

        aggregator.addRelationships(Arrays.asList(aggregatorFor));

        aggregator.setState(State.Inactive);

        try {
            model.checkCompliance(aggregator);
        } catch (ContentModelException e) {
            Assert.fail("Inactive aggregator for inactive agent should be OK");
        }

        aggregator.setState(State.Active);

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void tooManyAgentsTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        NDRObject agent2 = ndr.newObject(Type.Agent);

        Relationship aggregatorFor = new Relationship(aggregator,
                AGGREGATOR_FOR, agent);

        Relationship aggregatorFor2 = new Relationship(aggregator,
                AGGREGATOR_FOR, agent2);

        aggregator.addRelationships(Arrays
                .asList(aggregatorFor, aggregatorFor2));

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void nonexistantAgentTest() throws ContentModelException {
        NDRObject nonCommittedAgent = ndr.newObject(Type.Agent);
        /* Intentionally not committing */

        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        Relationship aggregatorFor = new Relationship(aggregator,
                AGGREGATOR_FOR, nonCommittedAgent);

        aggregator.addRelationships(Arrays.asList(aggregatorFor));

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void wrongAgentTypeTest() throws ContentModelException {
        NDRObject notAgent = ndr.newObject(Type.Resource);

        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        Relationship aggregatorFor = new Relationship(aggregator,
                AGGREGATOR_FOR, notAgent);

        aggregator.addRelationships(Arrays.asList(aggregatorFor));

        model.checkCompliance(aggregator);
    }

    /**
     * Deleting an agent that has an Aggregator AGGREGATOR_FOR it should be
     * disallowed
     */
    @Test(expected = ContentModelException.class)
    public void agentDeletingTest() throws ContentModelException {
        NDRObject thisAgent = ndr.newObject(Type.Agent);
        thisAgent.commit();

        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        aggregator.addRelationship(AGGREGATOR_FOR, thisAgent);
        aggregator.commit();

        /* Should be OK at this point, of course */
        try {
            model.checkCompliance(agent);
            model.checkCompliance(aggregator);
        } catch (ContentModelException e) {
            throw new Error("Active agent and Aggregator should not fail!", e);
        }

        thisAgent.setState(State.Deleted);
        aggregator.setState(State.Inactive);
        aggregator.commit();

        try {
            model.checkCompliance(thisAgent);
        } catch (ContentModelException e) {
            throw new Error("Inactive agent and Aggregator should not fail!", e);
        }

        aggregator.setState(State.Active);
        aggregator.commit();

        /* Should fail, since an aggregator references it */
        model.checkCompliance(thisAgent);
    }

    @Test
    public void associatedResourceAcceptanceTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.commit();

        Relationship aggregatorFor = new Relationship(aggregator,
                AGGREGATOR_FOR, agent);
        Relationship associatedWith = new Relationship(aggregator,
                ASSOCIATED_WITH, resource);

        aggregator.addRelationships(Arrays
                .asList(aggregatorFor, associatedWith));

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void nonexistantAssociatedResourceTest()
            throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        NDRObject resource = ndr.newObject(Type.Resource);
        /* Intentionally not committing resource */

        Relationship aggregatorFor = new Relationship(aggregator,
                AGGREGATOR_FOR, agent);
        Relationship associatedWith = new Relationship(aggregator,
                ASSOCIATED_WITH, resource);

        aggregator.addRelationships(Arrays
                .asList(aggregatorFor, associatedWith));

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void wrongAssociatedResourceStateTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        NDRObject resource = ndr.newObject(Type.Resource);
        resource.setState(State.Inactive);
        resource.commit();

        Relationship aggregatorFor = new Relationship(aggregator,
                AGGREGATOR_FOR, agent);
        Relationship associatedWith = new Relationship(aggregator,
                ASSOCIATED_WITH, resource);

        aggregator.addRelationships(Arrays
                .asList(aggregatorFor, associatedWith));

        try {
            aggregator.setState(State.Inactive);
            model.checkCompliance(aggregator);
        } catch (ContentModelException e) {
            Assert.fail("Inactive associated resource with "
                    + "inactive aggregator should be OK");
        }

        aggregator.setState(State.Active);

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void wrongAssociatedObjectTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        NDRObject md = ndr.newObject(Type.Metadata);

        Relationship aggregatorFor = new Relationship(aggregator,
                AGGREGATOR_FOR, agent);
        Relationship associatedWith = new Relationship(aggregator,
                ASSOCIATED_WITH, md);

        aggregator.addRelationships(Arrays
                .asList(aggregatorFor, associatedWith));

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void tooManyAssociatedResourcesTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        NDRObject resource = ndr.newObject(Type.Resource);
        NDRObject resource2 = ndr.newObject(Type.Resource);

        aggregator.addRelationship(AGGREGATOR_FOR, agent);
        aggregator.addRelationship(ASSOCIATED_WITH, resource);
        aggregator.addRelationship(ASSOCIATED_WITH, resource2);

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void associatedResourceDeletingTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        NDRObject resource = ndr.newObject(Type.Resource);
        resource.commit();

        aggregator.addRelationship(AGGREGATOR_FOR, agent);
        aggregator.addRelationship(ASSOCIATED_WITH, resource);
        aggregator.commit();

        try {
            model.checkCompliance(aggregator);
            model.checkCompliance(resource);
        } catch (ContentModelException e) {
            throw new Error("Active associated Resource and Aggregator "
                    + "should not fail!!", e);
        }

        resource.setState(State.Deleted);
        aggregator.setState(State.Inactive);
        aggregator.commit();

        try {
            model.checkCompliance(resource);
        } catch (ContentModelException e) {
            throw new Error("Inactive associated Resource and Aggregator "
                    + "should not fail!", e);
        }

        aggregator.setState(State.Active);
        aggregator.commit();

        model.checkCompliance(resource);
    }

    /** Just don't fail when we encounter OAI properties */
    @Test
    public void oaiPropertiesTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        aggregator.addProperty(OAIModel.Set.Components.SET_SPEC, "spec");
        aggregator.addRelationship(AGGREGATOR_FOR, agent);

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void wrongPropertyTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        aggregator.addProperty(MetadataModel.Components.UNIQUE_ID, "no");
        aggregator.addRelationship(AGGREGATOR_FOR, agent);

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void wrongRelationshipTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        aggregator.addRelationship(AGGREGATOR_FOR, agent);
        aggregator.addRelationship(
                MetadataProviderModel.Components.METADTA_PROVIDER_FOR, agent);

        model.checkCompliance(aggregator);
    }

    /** Different namespace elements are OK */
    @Test
    public void unrelatedElementsTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);

        aggregator.addProperty(new Property.Name("info:test/", "nonexistant",
                "no"), "123");
        aggregator.addRelationship(AGGREGATOR_FOR, agent);
        aggregator.addRelationship(
                AuthorizationModel.Grouping.Components.AUTHORIZED, agent);

        model.checkCompliance(aggregator);
    }
}
