/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import java.util.Arrays;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.AuthorizationModel.Grouping.Components.*;

public class AuthorizationModelTest {

    private static NDRAccess ndr;
    private static AuthorizationModel model;
    private static NDRObject authAgent;
    private static NDRObject authAggregator;

    @BeforeClass
    public static void setUp() throws ContentModelException {
        ndr = MockFedoraNDRAccess.getInstance();
        model = new AuthorizationModel();
        model.init(ndr);

        authAgent = ndr.newObject(Type.Agent);
        authAgent.commit();

        authAggregator = ndr.newObject(Type.Aggregator);
        authAggregator.commit();
    }

    @Test
    public void basicAcceptanceTestMP() throws ContentModelException {
        NDRObject mp = ndr.newObject(Type.MetadataProvider);

        Relationship auth = new Relationship(mp, AUTHORIZED, authAgent);

        mp.addRelationships(Arrays.asList(auth));

        model.checkCompliance(mp);
    }

    @Test
    public void basicAcceptanceTestAgg() throws ContentModelException {
        NDRObject agg = ndr.newObject(Type.Aggregator);

        Relationship auth = new Relationship(agg, AUTHORIZED, authAgent);

        agg.addRelationships(Arrays.asList(auth));

        model.checkCompliance(agg);
    }

    @Test
    public void authAggregatorAcceptanceTest() throws ContentModelException {
        NDRObject agg = ndr.newObject(Type.Aggregator);

        Relationship auth = new Relationship(agg, AUTHORIZED, authAggregator);

        agg.addRelationships(Arrays.asList(auth));

        model.checkCompliance(agg);
    }

    @Test
    public void passUnrelatedObjectsTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);

        Property uniqueID = new Property(metadata,
                MetadataModel.Components.UNIQUE_ID, "123");
        Relationship m4 = new Relationship(metadata,
                MetadataModel.Components.METADATA_FOR, authAggregator);

        metadata.addProperties(Arrays.asList(uniqueID));
        metadata.addRelationships(Arrays.asList(m4));

        model.checkCompliance(metadata);
    }

    @Test(expected = ContentModelException.class)
    public void inactiveAgentTest() throws ContentModelException {

        NDRObject inactiveAgent = ndr.newObject(Type.Agent);
        inactiveAgent.setState(State.Inactive);
        inactiveAgent.commit();

        NDRObject agg = ndr.newObject(Type.Aggregator);

        Relationship auth = new Relationship(agg, AUTHORIZED, inactiveAgent);

        agg.addRelationships(Arrays.asList(auth));

        model.checkCompliance(agg);
    }

    @Test(expected = ContentModelException.class)
    public void inactiveAggregatorTest() throws ContentModelException {

        NDRObject inactiveAggregator = ndr.newObject(Type.Aggregator);
        inactiveAggregator.setState(State.Inactive);
        inactiveAggregator.commit();

        NDRObject agg = ndr.newObject(Type.Aggregator);

        Relationship auth = new Relationship(agg, AUTHORIZED,
                inactiveAggregator);

        agg.addRelationships(Arrays.asList(auth));

        model.checkCompliance(agg);
    }

    @Test(expected = ContentModelException.class)
    public void wrongObjectTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.commit();

        NDRObject mp = ndr.newObject(Type.MetadataProvider);

        Relationship auth = new Relationship(mp, AUTHORIZED, metadata);

        mp.addRelationships(Arrays.asList(auth));

        model.checkCompliance(mp);
    }

    @Test
    public void multipleAgentsTest() throws ContentModelException {
        NDRObject anotherAgent = ndr.newObject(Type.Agent);
        anotherAgent.commit();

        NDRObject mp = ndr.newObject(Type.MetadataProvider);

        Relationship auth = new Relationship(mp, AUTHORIZED, authAgent);
        Relationship auth2 = new Relationship(mp, AUTHORIZED, anotherAgent);

        mp.addRelationships(Arrays.asList(auth, auth2));

        /* Should pass fine */
        model.checkCompliance(mp);
    }

    @Test
    public void multipleAggregatorsTest() throws ContentModelException {

        NDRObject anotherAgg = ndr.newObject(Type.Aggregator);
        anotherAgg.commit();

        NDRObject agg = ndr.newObject(Type.Aggregator);

        Relationship auth = new Relationship(agg, AUTHORIZED, authAggregator);
        Relationship auth2 = new Relationship(agg, AUTHORIZED, anotherAgg);

        agg.addRelationships(Arrays.asList(auth, auth2));

        /* should work fine */
        model.checkCompliance(agg);
    }

    @Test(expected = ContentModelException.class)
    public void exclusiveRelationshipsTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);

        Property uniqueID = new Property(metadata,
                MetadataModel.Components.UNIQUE_ID, "123");
        Relationship m4 = new Relationship(metadata,
                MetadataModel.Components.METADATA_FOR, authAggregator);
        Relationship auth = new Relationship(metadata, AUTHORIZED, authAgent);

        metadata.addProperties(Arrays.asList(uniqueID));
        metadata.addRelationships(Arrays.asList(m4, auth));

        model.checkCompliance(metadata);
    }

    @Test(expected = ContentModelException.class)
    public void otherObjectsNamespaceViolationTest()
            throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);

        Property uniqueID = new Property(metadata,
                MetadataModel.Components.UNIQUE_ID, "123");
        Relationship m4 = new Relationship(metadata,
                MetadataModel.Components.METADATA_FOR, authAggregator);
        Relationship auth = new Relationship(metadata, new Relationship.Name(
                AuthorizationModel.NS, "doesNotExist", "auth"), authAgent);

        metadata.addProperties(Arrays.asList(uniqueID));
        metadata.addRelationships(Arrays.asList(m4, auth));

        model.checkCompliance(metadata);
    }

    @Test(expected = ContentModelException.class)
    public void wrongRelationshipsTest() throws ContentModelException {
        NDRObject mp = ndr.newObject(Type.MetadataProvider);

        Relationship auth = new Relationship(mp, new Relationship.Name(
                AuthorizationModel.NS, "doesNotExist", "auth"), authAgent);

        mp.addRelationships(Arrays.asList(auth));

        model.checkCompliance(mp);
    }

    @Test
    public void unrelatedComponentsTest() throws ContentModelException {
        NDRObject mp = ndr.newObject(Type.MetadataProvider);

        Property setSpec = new Property(mp, OAIModel.Set.Components.SET_SPEC,
                "setSpec");
        Property otherProp = new Property(mp, new Property.Name(
                "http://does_not_exist", "prop", "no"), "value");

        Relationship auth = new Relationship(mp, AUTHORIZED, authAgent);
        Relationship mp4 = new Relationship(mp,
                MetadataProviderModel.Components.METADTA_PROVIDER_FOR,
                authAgent);
        Relationship other = new Relationship(mp, new Relationship.Name(
                "http://does_not_exist", "rel", "no"), authAgent);

        mp.addProperties(Arrays.asList(setSpec, otherProp));
        mp.addRelationships(Arrays.asList(auth, mp4, other));

        model.checkCompliance(mp);
    }
}
