/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.CollectionsModel.Components.*;
import static org.nsdl.repository.model.content.MetadataModel.Components.*;

public class CollectionsModelTest {

    private static NDRAccess ndr;
    private static CollectionsModel model;
    private static NDRObject collectionMP;
    private static NDRObject collectionAggregator;
    private static NDRObject resourceinCollection;

    @BeforeClass
    public static void setUp() throws ContentModelException {
        ndr = MockFedoraNDRAccess.getInstance();
        model = new CollectionsModel();
        model.init(ndr);

        NDRObject agg = ndr.newObject(Type.Aggregator);
        agg.commit();
        collectionAggregator = agg;

        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.addRelationship(AGGREGATED_BY, collectionAggregator);
        mp.commit();
        collectionMP = mp;

        NDRObject res = ndr.newObject(Type.Resource);
        res.addRelationship(Components.MEMBER_OF, collectionAggregator);
        res.commit();
        resourceinCollection = res;
    }

    /** Accept collection metadata and resource */
    @Test
    public void basicAcceptanceTest() throws ContentModelException {
        /* The resource already defined should be OK alone */
        model.checkCompliance(resourceinCollection);

        /* ..as should the Metadata Provider */
        model.checkCompliance(collectionMP);

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addRelationship(METADATA_FOR, resourceinCollection);
        metadata.addRelationship(METADATA_PROVIDED_BY, collectionMP);

        model.checkCompliance(metadata);
    }

    /**
     * Removing a resource with no colection md from a collection agg. should be
     * OK
     */
    @Test
    public void basicRemoveResourceAcceptanceTest()
            throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addRelationship(Components.MEMBER_OF, collectionAggregator);
        resource.commit();

        /* should be OK here by default */
        model.checkCompliance(resource);

        resource.removeRelationships(resource
                .getRelationships(Components.MEMBER_OF));

        /* should still be OK */
        model.checkCompliance(resource);
    }

    /** Pass unrelated objects */
    @Test
    public void passUnrelatedObjectsTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.addProperty(AgentModel.Components.IDENTIFIER_HOST, "test.test");
        agent.commit();

        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.addRelationship(
                MetadataProviderModel.Components.METADTA_PROVIDER_FOR, agent);
        mp.addRelationship(AuthorizationModel.Grouping.Components.AUTHORIZED,
                agent);
        mp.addProperty(OAIModel.Set.Components.SET_SPEC, "spec");
        mp.commit();

        NDRObject agg = ndr.newObject(Type.Aggregator);
        agg.addRelationship(AggregatorModel.Components.AGGREGATOR_FOR, agent);
        agg.addProperty(OAIModel.Set.Components.SET_SPEC, "spec");
        agg.commit();

        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(ResourceModel.Components.IDENTIFIER_URL,
                "http://test");
        resource.addRelationship(Components.MEMBER_OF, agg);
        resource.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addRelationship(METADATA_FOR, resource);
        metadata.addRelationship(METADATA_PROVIDED_BY, mp);
        metadata.addProperty(MetadataModel.Components.UNIQUE_ID, "123");
        metadata.commit();

        model.checkCompliance(agent);
        model.checkCompliance(agg);
        model.checkCompliance(mp);
        model.checkCompliance(metadata);
        model.checkCompliance(resource);
    }

    /** Should fail for 'aggregatedBy' pointing to a non-aggregator */
    @Test(expected = ContentModelException.class)
    public void wrongObjectTest() throws ContentModelException {
        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.addRelationship(AGGREGATED_BY, resourceinCollection);

        model.checkCompliance(mp);
    }

    /** Should fail for an 'aggregatedBy' relationship on a non-MP */
    @Test(expected = ContentModelException.class)
    public void exclusiveRelationshipTest() throws ContentModelException {
        NDRObject agg = ndr.newObject(Type.Aggregator);
        agg.addRelationship(AGGREGATED_BY, collectionAggregator);

        model.checkCompliance(agg);
    }

    /** Should pass through unrelated components in other namespaces OK */
    @Test
    public void unrelatedComponentsTest() throws ContentModelException {

    }

    /** adding resource without any MD should be OK */
    @Test
    public void nonDescribedResourceTest() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addRelationship(Components.MEMBER_OF, collectionAggregator);

        model.checkCompliance(resource);
    }

    /** Remove MD and resource from collection */
    @Test
    public void successfullyRemoveResourceTest() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addRelationship(Components.MEMBER_OF, collectionAggregator);
        resource.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addRelationship(METADATA_FOR, resource);
        metadata.addRelationship(METADATA_PROVIDED_BY, collectionMP);
        metadata.commit();

        /* deleting the metadata should be OK */
        ndr.deleteObject(metadata);
        model.checkCompliance(metadata);
        model.checkCompliance(resource);

        /* ..then removing the memberOf of the resource should be OK */
        resource.removeRelationships(resource
                .getRelationships(Components.MEMBER_OF));
        model.checkCompliance(resource);

    }

    /**
     * MD in a collection that does *not* describe a collection resource is not
     * allowed
     */
    @Test(expected = ContentModelException.class)
    public void collectionMetadataDescribesNonCollectionResourceTest()
            throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addRelationship(METADATA_FOR, resource);
        metadata.addRelationship(METADATA_PROVIDED_BY, collectionMP);

        /* fail! described resource is not in collection */
        model.checkCompliance(metadata);
    }

    /**
     * Should not be able to remove a resource from a collection if it is still
     * described by metadata
     */
    @Test(expected = ContentModelException.class)
    public void removeResourceFromCollectionWithoutRemovingMDTest()
            throws ContentModelException {

        NDRObject resource = ndr.getObject(resourceinCollection);

        /* Assure that the resource passes the model in its native state */
        try {
            model.checkCompliance(resource);
        } catch (ContentModelException e) {
            Assert.fail("Initial resource did not pass model!");
        }

        /* Create the metadata */
        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addRelationship(METADATA_FOR, resource);
        metadata.addRelationship(METADATA_PROVIDED_BY, collectionMP);
        metadata.commit();

        /* both should still pass at this point! */
        try {
            model.checkCompliance(resource);
            model.checkCompliance(metadata);
        } catch (ContentModelException e) {
            Assert.fail("Added md violated model!");
        }

        resource.removeRelationships(resource
                .getRelationships(Components.MEMBER_OF));

        /* Should fail! */
        model.checkCompliance(resource);

    }
}
