/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import java.util.Set;

import junit.framework.Assert;

import org.junit.Test;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;

public class ContentUtilTest {

    private static final String NS = "ns";
    private static final String PREFIX = "px";

    public static interface ComponentBase extends Definitions {
        public Property.Name PB1 = new Property.Name(NS, "pb1", PREFIX);

        public Relationship.Name RB1 =
                new Relationship.Name("NS", "rb1", PREFIX);
    }

    public static interface Components extends ComponentBase {

        public Property.Name P1 = new Property.Name(NS, "p1", PREFIX);

        @Shared
        public Property.Name P2 = new Property.Name(NS, "p2", PREFIX);

        public Property.Name P3 = new Property.Name(NS, "p3", PREFIX);

        public Relationship.Name R1 = new Relationship.Name(NS, "r1", PREFIX);

        @Shared
        public Relationship.Name R2 = new Relationship.Name(NS, "r2", PREFIX);

        public Relationship.Name R3 = new Relationship.Name(NS, "r3", PREFIX);
    }

    @Test
    public void simpleGetPropertyNamesTest() {
        Set<Property.Name> p =
                ModelUtil.getPropertyNames(ComponentBase.class, Scope.All);

        Assert.assertEquals("Did not fetch the right number of properties!", 1,
                p.size());

        Assert.assertTrue(
                "fetched property names don't contain the correct value!",
                p.contains(ComponentBase.PB1));
    }

    @Test
    public void simpleGetRelationshipNamesTest() {
        Set<Relationship.Name> r =
                ModelUtil
                        .getRelationshipNames(ComponentBase.class, Scope.All);

        Assert.assertEquals("Did not fetch the right number of relationships!",
                1, r.size());

        Assert.assertTrue(
                "fetched relationship names don't contain the correct value!",
                r.contains(ComponentBase.RB1));
    }

    @Test
    public void exclusiveGetPropertyNamesTest() {
        Set<Property.Name> p =
                ModelUtil.getPropertyNames(Components.class, Scope.Exclusive);

        Assert.assertEquals("Did not fetch the right number of exclusive properties!", 2,
                p.size());

        Assert.assertTrue(
                "fetched exclusive property names don't contain the correct value!",
                p.contains(Components.P1));
        
        Assert.assertTrue(
                "fetched exclusive property names don't contain the correct value!",
                p.contains(Components.P3));
    }

    @Test
    public void exclusiveGetRelationshipNamesTest() {
        Set<Relationship.Name> r =
                ModelUtil
                        .getRelationshipNames(Components.class, Scope.Exclusive);

        Assert.assertEquals("Did not fetch the right number of relationships!",
                2, r.size());

        Assert.assertTrue(
                "fetched exclusive relationship names don't contain the correct value!",
                r.contains(Components.R1));
        Assert.assertTrue(
                "fetched exclusive relationship names don't contain the correct value!",
                r.contains(Components.R3));
    }
    
    @Test
    public void sharedGetPropertyNamesTest() {
        Set<Property.Name> p =
                ModelUtil.getPropertyNames(Components.class, Scope.Shared);

        Assert.assertEquals("Did not fetch the right number of shared properties!", 2,
                p.size());

        Assert.assertTrue(
                "fetched shared property names don't contain the correct value!",
                p.contains(Components.P2));
        
        Assert.assertTrue(
                "fetched shared property names don't contain the correct value!",
                p.contains(ComponentBase.PB1));
    }

    @Test
    public void sharedGetRelationshipNamesTest() {
        Set<Relationship.Name> r =
                ModelUtil
                        .getRelationshipNames(Components.class, Scope.Shared);

        Assert.assertEquals("Did not fetch the right number of relationships!",
                2, r.size());

        Assert.assertTrue(
                "fetched exclusive relationship names don't contain the correct value!",
                r.contains(Components.R2));
        Assert.assertTrue(
                "fetched exclusive relationship names don't contain the correct value!",
                r.contains(ComponentBase.RB1));
    }
}
