/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.MetadataModel.Components.*;

public class MetadataModelTest {
    private static NDRAccess ndr;
    private static MetadataModel model;
    private static NDRObject existingMetadataProvider;
    private static NDRObject existingResource;
    private static NDRObject existingAggregator;

    @BeforeClass
    public static void setUp() throws Exception {
        ndr = MockFedoraNDRAccess.getInstance();

        /* Initialize the model */
        model = new MetadataModel();
        model.init(ndr);

        existingMetadataProvider = ndr.newObject(Type.MetadataProvider);
        existingMetadataProvider.commit();

        existingResource = ndr.newObject(Type.Resource);
        existingResource.commit();

        existingAggregator = ndr.newObject(Type.Aggregator);
        existingAggregator.commit();
    }

    /** A simple Metadata case that should pass easily */
    @Test
    public void basicAcceptanceTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        model.checkCompliance(metadata);
    }

    /** MetadataFor aggregators is OK */
    @Test
    public void aggregatorAcceptanceTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingAggregator);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        model.checkCompliance(metadata);
    }

    /** Other well-formed objects should pass through */
    @Test
    public void passUnrelatedObjectsTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.addProperty(AgentModel.Components.IDENTIFIER_HOST, "test.test");
        agent.commit();

        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.addRelationship(
                MetadataProviderModel.Components.METADTA_PROVIDER_FOR, agent);
        mp.addRelationship(AuthorizationModel.Grouping.Components.AUTHORIZED,
                agent);
        mp.addProperty(OAIModel.Set.Components.SET_SPEC, "spec");
        mp.commit();

        NDRObject agg = ndr.newObject(Type.Aggregator);
        agg.addRelationship(AggregatorModel.Components.AGGREGATOR_FOR, agent);
        agg.addProperty(OAIModel.Set.Components.SET_SPEC, "spec");
        agg.commit();

        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(ResourceModel.Components.IDENTIFIER_URL,
                "http://test");
        resource.addRelationship(Components.MEMBER_OF, agg);
        resource.commit();

        model.checkCompliance(agent);
        model.checkCompliance(agg);
        model.checkCompliance(mp);
        model.checkCompliance(resource);
    }

    /** No other objects can have Metadata's properties */
    @Test(expected = ContentModelException.class)
    public void exclusivePropertiesTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        aggregator.addProperty(UNIQUE_ID, "unique");

        model.checkCompliance(aggregator);
    }

    /** No other objects can have Metadata's relationships */
    @Test(expected = ContentModelException.class)
    public void exclusiveRelationshipsTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        aggregator.addRelationship(METADATA_FOR, existingResource);

        model.checkCompliance(aggregator);
    }

    /** Non-Metadata properties in the NSDL namespace should be banned */
    @Test(expected = ContentModelException.class)
    public void wrongPropertyTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        /* Just check that we haven't screwed up at this point */
        try {
            model.checkCompliance(metadata);
        } catch (ContentModelException e) {
            Assert.fail("Test should not have failed yet");
        }

        metadata.addProperty(new Property.Name(MetadataModel.NS,
                "doesNotExist", MetadataModel.PREFIX), "value");

        model.checkCompliance(metadata);
    }

    /** non-Metadata relationships in the NSDL namespace should be banned */
    @Test(expected = ContentModelException.class)
    public void wrongRelationshipTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        /* Just check that we haven't screwed up at this point */
        try {
            model.checkCompliance(metadata);
        } catch (ContentModelException e) {
            Assert.fail("Test should not have failed yet");
        }

        metadata.addRelationship(new Relationship.Name(MetadataModel.NS,
                "doesNotExist", MetadataModel.PREFIX), existingResource);

        model.checkCompliance(metadata);
    }

    /** Elements from other namespaces are OK */
    @Test
    public void unrelatedElementsTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        /* Just check that we haven't screwed up at this point */
        try {
            model.checkCompliance(metadata);
        } catch (ContentModelException e) {
            Assert.fail("Test should not have failed yet");
        }

        metadata.addRelationship(new Relationship.Name("http://new_ns", "rel",
                "new"), existingResource);
        metadata.addProperty(new Property.Name("http://new_ns", "prop", "new"),
                "value");

        model.checkCompliance(metadata);
    }

    /** Metadata for the wrong type should fail */
    @Test(expected = ContentModelException.class)
    public void wrongMetadataForTypeTest() throws ContentModelException {
        NDRObject nonResource = ndr.newObject(Type.Metadata);
        nonResource.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, nonResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        model.checkCompliance(metadata);
    }

    /** MetadataFor object must exist in repository */
    @Test(expected = ContentModelException.class)
    public void nonexistantMetadataForTest() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        /* Intentionally not adding */

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, resource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        model.checkCompliance(metadata);
    }

    @Test(expected = ContentModelException.class)
    public void deletingMetadataForTest() throws ContentModelException {

        NDRObject myResource = ndr.newObject(Type.Resource);
        myResource.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "deletingMetadataForTest:unique");
        metadata.addRelationship(METADATA_FOR, myResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);
        metadata.commit();

        try {
            model.checkCompliance(metadata);
            model.checkCompliance(myResource);
        } catch (ContentModelException e) {
            throw new Error(
                    "Sould not have failed with Active resource and MD", e);
        }

        metadata.setState(State.Deleted);
        metadata.commit();
        myResource.setState(State.Inactive);

        try {
            model.checkCompliance(myResource);
        } catch (ContentModelException e) {
            throw new Error(
                    "Should not have failed with Inactive resoource and MD");
        }

        metadata.setState(State.Active);
        metadata.commit();

        model.checkCompliance(myResource);

    }

    /** Describing a mix of Resources and Aggregators is OK */
    @Test
    public void describesMultipleHeterogeneousObjectsTest()
            throws ContentModelException {
        NDRObject newResource = ndr.newObject(Type.Resource);
        newResource.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata.addRelationship(METADATA_FOR, newResource);
        metadata.addRelationship(METADATA_FOR, existingAggregator);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        model.checkCompliance(metadata);
    }

    /** OAI elements (itemId) should be OK */
    @Test
    public void oaiElementsAcceptanceTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");

        metadata.addProperty(OAIModel.Item.Components.ITEM_ID, "itemId");

        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        model.checkCompliance(metadata);
    }

    @Test(expected = ContentModelException.class)
    public void nonexistantProviderTest() throws ContentModelException {
        NDRObject provider = ndr.newObject(Type.MetadataProvider);
        /* intentionally not adding */

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata.addRelationship(METADATA_PROVIDED_BY, provider);

        model.checkCompliance(metadata);
    }

    /** Enforce that there must be at least one metadataFor relationship */
    @Test(expected = ContentModelException.class)
    public void missingProviderTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);

        metadata.addProperty(UNIQUE_ID, "unique");
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        model.checkCompliance(metadata);
    }

    /**
     * Enforce that 'metadataprovidedBy' points to a MetadataProvider
     */
    @Test(expected = ContentModelException.class)
    public void wrongProviderTypeTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata.addRelationship(METADATA_PROVIDED_BY, existingAggregator);

        model.checkCompliance(metadata);
    }

    /** Active Metadata to an inactive provider is bad */
    @Test(expected = ContentModelException.class)
    public void inactiveProviderTest() throws ContentModelException {

        NDRObject provider = ndr.newObject(Type.MetadataProvider);
        provider.setState(State.Inactive);
        provider.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata.addRelationship(METADATA_PROVIDED_BY, provider);

        try {
            metadata.setState(State.Inactive);

            /* Should pass */
            model.checkCompliance(metadata);

        } catch (ContentModelException e) {
            Assert.fail("Inactive metadata in inactive provider failed test");
        } finally {
            metadata.setState(State.Active);
        }

        /* Should fail */
        model.checkCompliance(metadata);
    }

    @Test(expected = ContentModelException.class)
    public void deletingProviderTest() throws ContentModelException {
        NDRObject myProvider = ndr.newObject(Type.MetadataProvider);
        myProvider.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata.addRelationship(METADATA_PROVIDED_BY, myProvider);
        metadata.commit();

        try {
            model.checkCompliance(metadata);
            model.checkCompliance(myProvider);
        } catch (ContentModelException e) {
            throw new Error(
                    "Should not have failed yet with both objects Active!", e);
        }

        myProvider.setState(State.Deleted);
        metadata.setState(State.Inactive);
        metadata.commit();

        /* Both not active should be OK */
        try {
            model.checkCompliance(myProvider);
        } catch (ContentModelException e) {
            throw new Error(
                    "Should not have failed wiith both objects Inactive", e);
        }

        metadata.setState(State.Active);
        metadata.commit();

        /* Should FAIL */
        model.checkCompliance(myProvider);
    }

    /** No uniqueID is OK */
    public void missingUniqueIDTest() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);

        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        model.checkCompliance(metadata);
    }

    /** Another Metadata object having same UniqueID should cause failure */
    @Test(expected = ContentModelException.class)
    public void nonUniqueUniqueIDTest() throws ContentModelException {
        String ID = "nonUniqueUniqueIDTest";

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, ID);
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);
        metadata.commit();

        try {
            model.checkCompliance(metadata);
        } catch (ContentModelException e) {
            Assert.fail("Should not have failed at this point!");
        }

        NDRObject metadata2 = ndr.newObject(Type.Metadata);
        metadata2.addProperty(UNIQUE_ID, ID);
        metadata2.addRelationship(METADATA_FOR, existingResource);
        metadata2.addRelationship(METADATA_PROVIDED_BY,
                existingMetadataProvider);

        model.checkCompliance(metadata2);
    }

    /* Duplicated inactive uniqueIDs are OK */
    @Test
    public void inactiveUniqueIDTest() throws ContentModelException {
        String ID = "inactiveUniqueIDTest";

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, ID);
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);
        metadata.commit();

        try {
            model.checkCompliance(metadata);
        } catch (ContentModelException e) {
            throw new Error("Should noot have failed at this point!", e);
        }

        NDRObject metadata2 = ndr.newObject(Type.Metadata);
        metadata2.addProperty(UNIQUE_ID, ID);
        metadata2.addRelationship(METADATA_FOR, existingResource);
        metadata2.addRelationship(METADATA_PROVIDED_BY,
                existingMetadataProvider);

        metadata2.setState(State.Inactive);
        model.checkCompliance(metadata2);

        metadata.setState(State.Inactive);
        metadata.commit();
        model.checkCompliance(metadata2);

        metadata2.setState(State.Active);
        model.checkCompliance(metadata2);
    }

    /** Multiple unique uniqueIDs should be OK */
    @Test
    public void differentUniqueIDAcceptanceTest() throws ContentModelException {
        String ID = "differentUniqueIDAcceptanceTest";

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, ID);
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        try {
            model.checkCompliance(metadata);
            metadata.commit();
        } catch (ContentModelException e) {
            Assert.fail("Should not have failed at this point!");
        }

        NDRObject metadata2 = ndr.newObject(Type.Metadata);
        metadata2.addProperty(UNIQUE_ID, ID + "different");
        metadata2.addRelationship(METADATA_FOR, existingResource);
        metadata2.addRelationship(METADATA_PROVIDED_BY,
                existingMetadataProvider);

        /* should pass */
        model.checkCompliance(metadata2);
    }

    /** Another metadataProvider having same uniqueID should be OK */
    @Test
    public void globalUniqueIDIsolationTest() throws ContentModelException {
        NDRObject differentProvider = ndr.newObject(Type.MetadataProvider);
        differentProvider.commit();

        String ID = "globalUniqueIDIsolationTest";

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, ID);
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);

        try {
            model.checkCompliance(metadata);
            metadata.commit();
        } catch (ContentModelException e) {
            Assert.fail("Should not have failed at this point!");
        }

        NDRObject metadata2 = ndr.newObject(Type.Metadata);
        metadata2.addProperty(UNIQUE_ID, ID);
        metadata2.addRelationship(METADATA_FOR, existingResource);
        metadata2.addRelationship(METADATA_PROVIDED_BY, differentProvider);

        model.checkCompliance(metadata2);
    }

    /** Should be exactly one MetadataProvider */
    @Test(expected = ContentModelException.class)
    public void exactlyOneProviderTest() throws ContentModelException {

        NDRObject otherProvider = ndr.newObject(Type.MetadataProvider);
        otherProvider.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addProperty(UNIQUE_ID, "unique");
        metadata.addRelationship(METADATA_FOR, existingResource);
        metadata
                .addRelationship(METADATA_PROVIDED_BY, existingMetadataProvider);
        metadata.addRelationship(METADATA_PROVIDED_BY, otherProvider);

        try {
            model.checkCompliance(metadata);
            Assert.fail("Should have failed on too many MetadataProviders");
        } catch (ContentModelException e) {
            /* expected */
        }

        metadata.removeRelationships(metadata
                .getRelationships(METADATA_PROVIDED_BY));

        /* fail */
        model.checkCompliance(metadata);
    }
}
