/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import junit.framework.Assert;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.MetadataProviderModel.Components.*;

public class MetadataProviderModelTest {

    private static NDRAccess ndr;
    private static NDRObjectInfo existingAgent;
    private static NDRObjectInfo existingAggregator;

    private static MetadataProviderModel model;

    @BeforeClass
    public static void setUp() throws Exception {
        ndr = MockFedoraNDRAccess.getInstance();

        NDRObject agent = ndr.newObject(Type.Agent);
        agent.commit();
        existingAgent = agent;

        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        aggregator.commit();
        existingAggregator = aggregator;

        model = new MetadataProviderModel();
        model.init(ndr);
    }

    @Test
    public void basicAcceptanceTest() throws ContentModelException {
        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);
        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, ndr
                .getObject(existingAgent));

        model.checkCompliance(metadataProvider);
    }

    @Test
    public void oaiPropertiesAcceptanceTest() throws ContentModelException {
        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);
        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, ndr
                .getObject(existingAgent));

        metadataProvider.addProperty(OAIModel.Set.Components.SET_SPEC, "spec");

        model.checkCompliance(metadataProvider);
    }

    @Test
    public void collectionRelationshipsAcceptanceTest()
            throws ContentModelException {
        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);
        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, ndr
                .getObject(existingAgent));

        metadataProvider.addRelationship(
                CollectionsModel.Components.AGGREGATED_BY, ndr
                        .getObject(existingAggregator));

        model.checkCompliance(metadataProvider);
    }

    @Test(expected = ContentModelException.class)
    public void exclusiveRelationshipsTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        aggregator.addRelationship(METADTA_PROVIDER_FOR, ndr
                .getObject(existingAgent));

        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void wrongRelationshipsTest() throws ContentModelException {
        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);
        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, ndr
                .getObject(existingAgent));

        metadataProvider.addRelationship(
                AggregatorModel.Components.ASSOCIATED_WITH, ndr
                        .getObject(existingAgent));
        model.checkCompliance(metadataProvider);
    }

    @Test(expected = ContentModelException.class)
    public void wrongPropertiesTest() throws ContentModelException {
        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);
        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, ndr
                .getObject(existingAgent));

        metadataProvider.addProperty(MetadataModel.Components.UNIQUE_ID,
                "unique");

        model.checkCompliance(metadataProvider);
    }

    @Test
    public void passUnrelatedObjectsTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.addProperty(AgentModel.Components.IDENTIFIER_HOST, "test.test");
        agent.commit();

        NDRObject agg = ndr.newObject(Type.Aggregator);
        agg.addRelationship(AggregatorModel.Components.AGGREGATOR_FOR, agent);
        agg.addProperty(OAIModel.Set.Components.SET_SPEC, "spec");
        agg.commit();

        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(ResourceModel.Components.IDENTIFIER_URL,
                "http://test");
        resource.addRelationship(Components.MEMBER_OF, agg);
        resource.commit();

        NDRObject provider = ndr.newObject(Type.MetadataProvider);
        provider.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addRelationship(MetadataModel.Components.METADATA_FOR,
                resource);
        metadata.addRelationship(MetadataModel.Components.METADATA_PROVIDED_BY,
                provider);
        metadata.addProperty(MetadataModel.Components.UNIQUE_ID, "123");
        metadata.commit();

        model.checkCompliance(agent);
        model.checkCompliance(agg);
        model.checkCompliance(resource);
        model.checkCompliance(metadata);
    }

    @Test(expected = ContentModelException.class)
    public void wrongAgentTypeTest() throws ContentModelException {
        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);
        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, ndr
                .getObject(existingAggregator));

        model.checkCompliance(metadataProvider);
    }

    @Test(expected = ContentModelException.class)
    public void noMetadataProviderGivenTest() throws ContentModelException {
        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);

        model.checkCompliance(metadataProvider);
    }

    @Test(expected = ContentModelException.class)
    public void nonexistiantAgentTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        /* intentionally not adding */

        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);
        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, agent);

        model.checkCompliance(metadataProvider);
    }

    @Test(expected = ContentModelException.class)
    public void tooManyAgentsTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.commit();

        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);
        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, agent);

        /* should be OK here */
        try {
            model.checkCompliance(metadataProvider);
        } catch (ContentModelException e) {
            Assert.fail("Should not have failed yet!");
        }

        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, ndr
                .getObject(existingAgent));

        model.checkCompliance(metadataProvider);
    }

    @Test(expected = ContentModelException.class)
    public void agentDeletingTest() throws ContentModelException {
        NDRObject thisAgent = ndr.newObject(Type.Agent);
        thisAgent.commit();

        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);
        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, thisAgent);
        metadataProvider.commit();

        /* Should be OK at this point, of course */
        try {
            model.checkCompliance(thisAgent);
            model.checkCompliance(metadataProvider);
        } catch (ContentModelException e) {
            Assert.fail("Should not have failed yet!");
        }

        thisAgent.setState(State.Deleted);
        metadataProvider.setState(State.Inactive);
        metadataProvider.commit();

        try {
            model.checkCompliance(thisAgent);
        } catch (ContentModelException e) {
            throw new Error("Inactive agent and Aggregator should not fail!", e);
        }

        metadataProvider.setState(State.Active);
        metadataProvider.commit();

        /* Should fail, since a MetadataProvider references it */
        model.checkCompliance(thisAgent);
    }

    @Test(expected = ContentModelException.class)
    public void inactiveMetadataProviderTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.setState(State.Inactive);
        agent.commit();

        NDRObject metadataProvider = ndr.newObject(Type.MetadataProvider);
        metadataProvider.addRelationship(METADTA_PROVIDER_FOR, agent);

        try {
            metadataProvider.setState(State.Inactive);
            /* should be OK if both are inactive */
            model.checkCompliance(metadataProvider);
        } catch (ContentModelException e) {
            Assert.fail("MP and Agent both inactive should be OK");
        } finally {
            metadataProvider.setState(State.Active);
        }

        model.checkCompliance(metadataProvider);
    }
}
