/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import junit.framework.Assert;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.AggregatorModel.Components.*;
import static org.nsdl.repository.model.content.MetadataProviderModel.Components.*;
import static org.nsdl.repository.model.content.OAIModel.Set.Components.*;
import static org.nsdl.repository.model.content.OAIModel.Item.Components.*;

public class OAIModelTest {

    private static NDRAccess ndr;
    private static OAIModel model;

    @BeforeClass
    public static void setup() {
        ndr = MockFedoraNDRAccess.getInstance();
        model = new OAIModel();

        model.init(ndr);

    }

    @Test
    public void basicAcceptanceTestMetadataProvider()
            throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.commit();

        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.addProperty(SET_SPEC, "spec");
        mp.addProperty(SET_NAME, "name");
        mp.addRelationship(METADTA_PROVIDER_FOR, agent);

        model.checkCompliance(mp);
    }

    @Test
    public void basicAcceptanceTestAggregator() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.commit();

        NDRObject agg = ndr.newObject(Type.Aggregator);
        agg.addProperty(SET_SPEC, "spec");
        agg.addProperty(SET_NAME, "name");
        agg.addRelationship(AGGREGATOR_FOR, agent);

        model.checkCompliance(agg);
    }

    @Test
    public void basicAcceptanceTestMetadata() throws ContentModelException {
        NDRObject metadata = ndr.newObject(Type.Metadata);

        metadata.addProperty(ITEM_ID, "id");

        model.checkCompliance(metadata);
    }

    @Test(expected = ContentModelException.class)
    public void exclusivePropertiesTestItemid() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(ITEM_ID, "item");

        model.checkCompliance(resource);
    }

    @Test(expected = ContentModelException.class)
    public void exclusivePropertiesTestSetName() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(SET_NAME, "name");

        model.checkCompliance(resource);
    }

    @Test(expected = ContentModelException.class)
    public void exclusivePropertiesTestSetSpec() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(SET_SPEC, "spec");

        model.checkCompliance(resource);
    }

    @Test
    public void passUnrelatedObjectsTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.addProperty(AgentModel.Components.IDENTIFIER_HOST,
                "passUnrelatedObjectsTest:host");
        agent.commit();

        NDRObject agg = ndr.newObject(Type.Aggregator);
        agg.addRelationship(AggregatorModel.Components.AGGREGATOR_FOR, agent);
        agg.addProperty(SET_SPEC, "passUnrelatedObjectsTest:aggSpec");
        agg.commit();

        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(ResourceModel.Components.IDENTIFIER_URL,
                "http://test");
        resource.addRelationship(Components.MEMBER_OF, agg);
        resource.commit();

        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.addRelationship(METADTA_PROVIDER_FOR, agent);
        mp.addRelationship(AuthorizationModel.Grouping.Components.AUTHORIZED,
                agent);
        mp.addProperty(SET_SPEC, "passUnrelatedObjectsTest:mpSpec");
        mp.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addRelationship(MetadataModel.Components.METADATA_FOR,
                resource);
        metadata.addRelationship(MetadataModel.Components.METADATA_PROVIDED_BY,
                mp);
        metadata.addProperty(MetadataModel.Components.UNIQUE_ID,
                "passUnrelatedObjectsTest:uniqueId");
        metadata.commit();

        model.checkCompliance(agent);
        model.checkCompliance(agg);
        model.checkCompliance(resource);
        model.checkCompliance(metadata);
        model.checkCompliance(mp);
    }

    /** ItemId must be globally unique */
    @Test(expected = ContentModelException.class)
    public void uniqueItemIdTest() throws ContentModelException {
        String ID = "uniqueItemIdTest:itemId";
        NDRObject metadata1 = ndr.newObject(Type.Metadata);
        metadata1.addProperty(ITEM_ID, ID);
        metadata1.commit();

        try {
            model.checkCompliance(metadata1);
        } catch (ContentModelException e) {
            Assert.fail("Initial metadata with unique id should be OK");
        }

        NDRObject metadata2 = ndr.newObject(Type.Metadata);
        metadata2.addProperty(ITEM_ID, ID + "unique");

        try {
            model.checkCompliance(metadata2);
        } catch (ContentModelException e) {
            Assert.fail("Second metdata with uniqueID failed");
        }

        metadata2.removeProperties(metadata2.getProperties(ITEM_ID));
        metadata2.addProperty(ITEM_ID, ID);

        model.checkCompliance(metadata2);
    }

    /** ItemIds must only be unique amongst active objects */
    @Test
    public void uniqueInactiveItemIdTest() throws ContentModelException {
        String ID = "uniqueInactiveItemIdTest:itemId";
        NDRObject metadata1 = ndr.newObject(Type.Metadata);
        metadata1.setState(State.Inactive);
        metadata1.addProperty(ITEM_ID, ID);
        metadata1.commit();

        NDRObject metadata2 = ndr.newObject(Type.Metadata);
        metadata2.addProperty(ITEM_ID, ID + "unique");

        /* should work OK, since metadata1 is inactive */
        model.checkCompliance(metadata1);
        model.checkCompliance(metadata2);
    }

    /** Allow multiple MPs with the same setSpec for a given agent */
    @Test
    public void multipleSetSpecSameAgentMetadataProviderTest()
            throws ContentModelException {
        String setSpec = "multipleSetSpecSameAgentMetadataProviderTest:spec";
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.commit();

        NDRObject metadataProvider1 = ndr.newObject(Type.MetadataProvider);
        metadataProvider1.addProperty(SET_SPEC, setSpec);
        metadataProvider1.addRelationship(METADTA_PROVIDER_FOR, agent);
        metadataProvider1.commit();

        NDRObject metadataProvider2 = ndr.newObject(Type.MetadataProvider);
        metadataProvider2.addProperty(SET_SPEC, setSpec + 2);
        metadataProvider2.addRelationship(METADTA_PROVIDER_FOR, agent);

        /* Both should work, since setSpecs are different */
        model.checkCompliance(metadataProvider1);
        model.checkCompliance(metadataProvider2);

        metadataProvider2.removeProperties(metadataProvider2
                .getProperties(SET_SPEC));
        metadataProvider2.addProperty(SET_SPEC, setSpec);

        /* Should pass as well */
        model.checkCompliance(metadataProvider2);

    }

    /** Allow multiple Aggregators with the same setSpec for a given agent */
    @Test
    public void multipleSetSpecSameAgentAggregatorTest()
            throws ContentModelException {
        String setSpec = "multipleSetSpecSameAgentAggregatorTest:spec";

        NDRObject agent = ndr.newObject(Type.Agent);
        agent.commit();

        NDRObject aggregator1 = ndr.newObject(Type.Aggregator);
        aggregator1.addProperty(SET_SPEC, setSpec);
        aggregator1.addRelationship(AGGREGATOR_FOR, agent);
        aggregator1.commit();

        NDRObject aggregator2 = ndr.newObject(Type.Aggregator);
        aggregator2.addProperty(SET_SPEC, setSpec + 2);
        aggregator2.addRelationship(AGGREGATOR_FOR, agent);

        /* Both should work, since setSpecs are different */
        model.checkCompliance(aggregator1);
        model.checkCompliance(aggregator2);

        aggregator2.removeProperties(aggregator2.getProperties(SET_SPEC));
        aggregator2.addProperty(SET_SPEC, setSpec);

        /* Should pass */
        model.checkCompliance(aggregator2);

    }

    /** Aggregators/MPs on different objects can't have the same setSpec */
    @Test(expected = ContentModelException.class)
    public void uniqueSetSpecDifferentAgentMetadataProviderTest()
            throws ContentModelException {
        String setSpec = "uniqueSetSpecDifferentAgentMetadataProviderTest:spec";

        NDRObject agent1 = ndr.newObject(Type.Agent);
        agent1.commit();

        NDRObject agent2 = ndr.newObject(Type.Agent);
        agent2.commit();

        NDRObject metadataProvider1 = ndr.newObject(Type.MetadataProvider);
        metadataProvider1.addProperty(SET_SPEC, setSpec);
        metadataProvider1.addRelationship(METADTA_PROVIDER_FOR, agent1);
        metadataProvider1.commit();

        NDRObject metadataProvider2 = ndr.newObject(Type.MetadataProvider);
        metadataProvider2.addProperty(SET_SPEC, setSpec + 2);
        metadataProvider2.addRelationship(METADTA_PROVIDER_FOR, agent2);

        /* Both should work, since setSpecs are different */
        try {
            model.checkCompliance(metadataProvider1);
            model.checkCompliance(metadataProvider2);
        } catch (ContentModelException e) {
            Assert.fail("Should not have failed yet");
        }

        metadataProvider2.removeProperties(metadataProvider2
                .getProperties(SET_SPEC));
        metadataProvider2.addProperty(SET_SPEC, setSpec);

        /* Should FAIL */
        model.checkCompliance(metadataProvider2);
    }

    @Test(expected = ContentModelException.class)
    public void uniqueSetSpecDifferentAgentAggregatorTest()
            throws ContentModelException {
        String setSpec = "uniqueSetSpecDifferentAgentAggregatorTest:spec";

        NDRObject agent1 = ndr.newObject(Type.Agent);
        agent1.commit();

        NDRObject agent2 = ndr.newObject(Type.Agent);
        agent2.commit();

        NDRObject aggregator1 = ndr.newObject(Type.Aggregator);
        aggregator1.addProperty(SET_SPEC, setSpec);
        aggregator1.addRelationship(AGGREGATOR_FOR, agent1);
        aggregator1.commit();

        NDRObject aggregator2 = ndr.newObject(Type.Aggregator);
        aggregator2.addProperty(SET_SPEC, setSpec + 2);
        aggregator2.addRelationship(AGGREGATOR_FOR, agent2);

        /* Both should work, since setSpecs are different */
        try {
            model.checkCompliance(aggregator1);
            model.checkCompliance(aggregator2);
        } catch (ContentModelException e) {
            Assert.fail("Should not have failed yet");
        }

        aggregator2.removeProperties(aggregator2.getProperties(SET_SPEC));
        aggregator2.addProperty(SET_SPEC, setSpec);

        /* Should FAIL */
        model.checkCompliance(aggregator2);
    }

    /** An MP can have the same setSpec as an inactive/deleted object */
    @Test
    public void uniqueSetSpecInactiveMetadataProviderTest()
            throws ContentModelException {
        String setSpec = "uniqueSetSpecInactiveAgentMetadataProviderTest:spec";

        NDRObject agent1 = ndr.newObject(Type.Agent);
        agent1.commit();

        NDRObject agent2 = ndr.newObject(Type.Agent);
        agent2.commit();

        NDRObject metadataProvider1 = ndr.newObject(Type.MetadataProvider);
        metadataProvider1.addProperty(SET_SPEC, setSpec);
        metadataProvider1.addRelationship(METADTA_PROVIDER_FOR, agent1);
        metadataProvider1.commit();

        NDRObject metadataProvider2 = ndr.newObject(Type.MetadataProvider);
        metadataProvider2.addProperty(SET_SPEC, setSpec);
        metadataProvider2.addRelationship(METADTA_PROVIDER_FOR, agent2);

        /* With MetadataProvider 1 inactive, model should pass for both */
        metadataProvider1.setState(State.Inactive);
        metadataProvider1.commit();

        model.checkCompliance(metadataProvider1);
        model.checkCompliance(metadataProvider2);

        /* With both inactive, model should pass for both */
        metadataProvider2.setState(State.Inactive);

        model.checkCompliance(metadataProvider1);
        model.checkCompliance(metadataProvider2);

        /*
         * With MetadataProvider 2 inactive, MetadataProvider 1 active, should
         * be OK
         */
        metadataProvider1.setState(State.Active);
        metadataProvider1.commit();

        model.checkCompliance(metadataProvider1);
        model.checkCompliance(metadataProvider2);
    }

    @Test
    public void uniqueSetSpecInactiveAggregatorTest()
            throws ContentModelException {
        String setSpec = "uniqueSetSpecInactiveAggregatorTest:spec";

        NDRObject agent1 = ndr.newObject(Type.Agent);
        agent1.commit();

        NDRObject agent2 = ndr.newObject(Type.Agent);
        agent2.commit();

        NDRObject aggregator1 = ndr.newObject(Type.Aggregator);
        aggregator1.addProperty(SET_SPEC, setSpec);
        aggregator1.addRelationship(AGGREGATOR_FOR, agent1);
        aggregator1.commit();

        NDRObject aggregator2 = ndr.newObject(Type.Aggregator);
        aggregator2.addProperty(SET_SPEC, setSpec + 2);
        aggregator2.addRelationship(AGGREGATOR_FOR, agent2);

        /* With Aggregator 1 inactive, model should pass for both */
        aggregator1.setState(State.Inactive);
        aggregator1.commit();

        model.checkCompliance(aggregator1);
        model.checkCompliance(aggregator2);

        /* With both inactive, model should pass for both */
        aggregator2.setState(State.Inactive);

        model.checkCompliance(aggregator1);
        model.checkCompliance(aggregator2);

        /*
         * With Aggregator 2 inactive, Aggregator 1 active, should be OK
         */
        aggregator1.setState(State.Active);
        aggregator2.commit();

        model.checkCompliance(aggregator1);
        model.checkCompliance(aggregator2);
    }

    /**
     * Should break down gracefully (ContentModelException) if agent not in
     * repository
     */
    @Test(expected = ContentModelException.class)
    public void uniqueSetSpecAgentNoExistMetadataProviderTest()
            throws ContentModelException {

        NDRObject agent = ndr.newObject(Type.Agent);
        /* Intentionally not committing */

        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.addProperty(SET_SPEC, "spec");
        mp.addProperty(SET_NAME, "name");
        mp.addRelationship(METADTA_PROVIDER_FOR, agent);

        model.checkCompliance(mp);
    }

    /**
     * Should break down gracefully (ContentModelException) if agent not in
     * repository
     */
    @Test(expected = ContentModelException.class)
    public void uniqueSetSpecAgentNoExistAggregatorTest()
            throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        /* Intentionally not committing */

        NDRObject agg = ndr.newObject(Type.Aggregator);
        agg.addProperty(SET_SPEC, "spec");
        agg.addProperty(SET_NAME, "name");
        agg.addRelationship(AGGREGATOR_FOR, agent);

        model.checkCompliance(agg);
    }
}
