/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.content;

import junit.framework.Assert;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.access.NDRAccess;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;

import static org.nsdl.repository.model.content.ResourceModel.Components.*;

public class ResourceModelTest {

    private static NDRAccess ndr;
    private static ResourceModel model;

    @BeforeClass
    public static void setup() {
        ndr = MockFedoraNDRAccess.getInstance();

        model = new ResourceModel();
        model.init(ndr);
    }

    @Test
    public void basicAcceptanceTest() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(IDENTIFIER_URL, "info:/test/basicAcceptanceTest");

        /* check committed and uncommitted acceptance */

        model.checkCompliance(resource);

        resource.commit();

        model.checkCompliance(resource);
    }

    @Test(expected = ContentModelException.class)
    public void exclusivePropertiesTest() throws ContentModelException {
        NDRObject aggregator = ndr.newObject(Type.Aggregator);
        aggregator.addProperty(IDENTIFIER_URL,
                "info:/test/exclusivePropertiesTest");
        model.checkCompliance(aggregator);
    }

    @Test(expected = ContentModelException.class)
    public void wrongPropertiesTest() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(MetadataModel.Components.UNIQUE_ID, "id");

        model.checkCompliance(resource);
    }

    @Test
    public void passUnrelatedObjectsTest() throws ContentModelException {
        NDRObject agent = ndr.newObject(Type.Agent);
        agent.addProperty(AgentModel.Components.IDENTIFIER_HOST,
                "passUnrelatedObjectsTest:host");
        agent.commit();

        NDRObject agg = ndr.newObject(Type.Aggregator);
        agg.addRelationship(AggregatorModel.Components.AGGREGATOR_FOR, agent);
        agg.addProperty(OAIModel.Set.Components.SET_SPEC,
                "passUnrelatedObjectsTest:aggSpec");
        agg.commit();

        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(ResourceModel.Components.IDENTIFIER_URL,
                "http://test");
        resource.addRelationship(Components.MEMBER_OF, agg);
        resource.commit();

        NDRObject mp = ndr.newObject(Type.MetadataProvider);
        mp.addRelationship(
                MetadataProviderModel.Components.METADTA_PROVIDER_FOR, agent);
        mp.addRelationship(AuthorizationModel.Grouping.Components.AUTHORIZED,
                agent);
        mp.addProperty(OAIModel.Set.Components.SET_SPEC,
                "passUnrelatedObjectsTest:mpSpec");
        mp.commit();

        NDRObject metadata = ndr.newObject(Type.Metadata);
        metadata.addRelationship(MetadataModel.Components.METADATA_FOR,
                resource);
        metadata.addRelationship(MetadataModel.Components.METADATA_PROVIDED_BY,
                mp);
        metadata.addProperty(MetadataModel.Components.UNIQUE_ID,
                "passUnrelatedObjectsTest:uniqueId");
        metadata.commit();

        model.checkCompliance(agent);
        model.checkCompliance(agg);
        model.checkCompliance(resource);
        model.checkCompliance(metadata);
        model.checkCompliance(mp);
    }

    @Test(expected = ContentModelException.class)
    public void identifierUniquenessTest() throws ContentModelException {
        String id = "info:test/identifierUniquenessTest";

        NDRObject existingResource = ndr.newObject(Type.Resource);
        existingResource.addProperty(IDENTIFIER_OTHER, id);
        existingResource.commit();

        NDRObject newResource = ndr.newObject(Type.Resource);
        newResource.addProperty(IDENTIFIER_OTHER, id + 2);

        try {
            model.checkCompliance(newResource);
        } catch (ContentModelException e) {
            Assert.fail("Should not have failed yet");
        }

        newResource.removeProperties(newResource
                .getProperties(IDENTIFIER_OTHER));
        newResource.addProperty(IDENTIFIER_OTHER, id);

        /* Should fail */
        model.checkCompliance(newResource);
    }

    /** Inactive resources can share the same identifier */
    @Test
    public void identifierUniquenessInactiveTest() throws ContentModelException {
        String id = "info:test/identifierUniquenessInactiveTest";

        NDRObject existingResource = ndr.newObject(Type.Resource);
        existingResource.addProperty(IDENTIFIER_OTHER, id);
        existingResource.commit();

        NDRObject newResource = ndr.newObject(Type.Resource);
        newResource.addProperty(IDENTIFIER_OTHER, id);

        existingResource.setState(State.Inactive);
        existingResource.commit();

        model.checkCompliance(existingResource);
        model.checkCompliance(newResource);

        newResource.setState(State.Inactive);
        model.checkCompliance(newResource);

        existingResource.setState(State.Active);
        existingResource.commit();

        model.checkCompliance(newResource);
    }

    @Test(expected = ContentModelException.class)
    public void missingIdentifierTest() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        model.checkCompliance(resource);
    }

    @Test(expected = ContentModelException.class)
    public void tooManyIdentifiersTest() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(IDENTIFIER_URL, "http://tooManyIdentifiersTest");
        resource.addProperty(IDENTIFIER_URL, "http://tooManyIdentifiersTest2");
        model.checkCompliance(resource);
    }

    @Test(expected = ContentModelException.class)
    public void tooManyIdentifiersHeterogeneousTest()
            throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource.addProperty(IDENTIFIER_URL, "http://tooManyIdentifiersTest");
        resource
                .addProperty(IDENTIFIER_OTHER, "http://tooManyIdentifiersTest2");
        model.checkCompliance(resource);
    }

    @Test
    public void unrelatedElementsTest() throws ContentModelException {
        NDRObject resource = ndr.newObject(Type.Resource);
        resource
                .addProperty(IDENTIFIER_URL, "info:/test/unrelatedElementsTest");
        resource.addProperty(new Property.Name("info:test:/new_property",
                "prop", "no"), "value");
        resource.addRelationship(new Relationship.Name(
                "info:test/new_relationship", "rel", "no"), resource);
    }
}
