/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.impl;

import java.io.ByteArrayInputStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.junit.Test;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.util.NDRObjectTemplate;
import org.nsdl.repository.util.StreamUtil;

import static junit.framework.Assert.*;

public class DatastreamTrackerTest {

    private static final Datastream.Name n1 = new Datastream.Name("test_1");
    private static final Datastream.Name n2 = new Datastream.Name("test_2");
    private static final Datastream.Name n3 = new Datastream.Name("test_3");
    private static final Datastream.Name n4 = new Datastream.Name("test_4");

    private static final Datastream d1 = new BasicDatastream(n1);
    private static final Datastream d2 = new BasicDatastream(n2);
    private static final Datastream d3 = new BasicDatastream(n3);

    private static final String d1_content = "DS1";
    private static final String d2_content = "DS2";
    private static final String d3_content = "DS3";

    static {
        d1.setContent(new ByteArrayInputStream(d1_content.getBytes()));
        d2.setContent(new ByteArrayInputStream(d2_content.getBytes()));
        d3.setContent(new ByteArrayInputStream(d3_content.getBytes()));
    }

    private static Collection<Datastream> datastreamSample = Arrays.asList(d1,
            d2, d3);

    @Test
    public void testSimpleAdd() {
        BasicComponentTracker<Datastream> tracker = getPopulatedTracker();
        String CONTENT = "DS4";
        Datastream added = new BasicDatastream(n4);
        added.setContent(new ByteArrayInputStream(CONTENT.getBytes()));
        tracker.add(added);

        assertTrue(tracker.hasModified());
        assertEquals(added.getName(), tracker.getAdded().iterator().next()
                .getName());
    }

    @Test
    public void testSimpleDelete() {
        BasicComponentTracker<Datastream> tracker = getPopulatedTracker();

        tracker.remove(d2);
        tracker.remove(d3);

        Set<Datastream.Name> current = new HashSet<Datastream.Name>();
        Set<Datastream.Name> deleted = new HashSet<Datastream.Name>();

        for (Datastream d : tracker.getCurrent()) {
            current.add(d.getName());
        }

        for (Datastream d : tracker.getDeleted()) {
            deleted.add(d.getName());
        }

        assertTrue(tracker.hasModified());
        assertEquals(1, tracker.getCurrent().size());
        assertEquals(2, tracker.getDeleted().size());
        assertFalse(current.contains(n2));
        assertFalse(current.contains(n3));
        assertTrue(deleted.contains(n2));
        assertTrue(deleted.contains(n3));
    }

    @Test
    public void testSimpleReplace() {
        BasicComponentTracker<Datastream> tracker = getPopulatedTracker();

        final String CONTENT = "replaceValue";

        Datastream d = tracker.get(n1).iterator().next();
        d.setContent(new ByteArrayInputStream(CONTENT.getBytes()));

        assertTrue("Tracker did not detect update", tracker.hasModified());
        assertEquals(tracker.getCurrent().size(), datastreamSample.size());
        assertEquals(1, tracker.getUpdated().size());
        assertEquals(n1, tracker.getUpdated().iterator().next().getName());
        assertEquals(CONTENT, StreamUtil.streamToString(tracker.getUpdated()
                .iterator().next().getContent()));
    }

    @Test
    public void testDuplicateEqualitySkip() {
        BasicComponentTracker<Datastream> tracker = getPopulatedTracker();

        tracker.add(d1);

        assertFalse(tracker.hasModified());
        assertEquals(tracker.getAdded().size(), 0);
    }

    public static DatastreamTracker getPopulatedTracker() {
        NDRObject base = new NDRObjectTemplate(null);
        base.addDatastreams(datastreamSample);
        return new DatastreamTracker(base);
    }
}
