/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */

package org.nsdl.repository.model.impl;

import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;

import org.junit.Test;
import org.nsdl.repository.access.AccessException;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.impl.fedora.RedirectDatastream;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.NDRObject.State;
import org.nsdl.repository.model.content.CommonModel.Components;
import org.nsdl.repository.model.types.Type;
import org.nsdl.repository.util.NDRObjectTemplate;

import static junit.framework.Assert.*;
import static org.nsdl.repository.model.content.CommonModel.Components.HAS_HANDLE;
import static org.nsdl.repository.model.content.ResourceModel.Components.IDENTIFIER_URL;

public class MutableObjectBaseTest {

    private static Datastream.Name REDIR = new Datastream.Name("REDIR");

    private static final int NEW_PROPERTY_COUNT = 2;

    /* Assure that a new object gets initialized correctly */
    @Test
    public void newInitTest() throws AccessException {
        MutableObjectProbe probe = new MutableObjectProbe();
        probe.initNew(Type.Agent, "newInit");

        Property objectType = probe.getProperty(Components.OBJECT_TYPE);
        Property handle = probe.getProperty(HAS_HANDLE);

        /* We should objectType and handle properties created. */
        assertEquals("Object was created with wrong object type", "Agent",
                objectType.value());
        assertEquals("Object was created with wrong handle", "newInit", handle
                .value());

        /* The property tracker should identify these properties as new */
        assertEquals(probe.getPropertyTracker().getAdded().size(),
                NEW_PROPERTY_COUNT);

        /* We should have an active state */
        assertEquals(probe.state(), State.Active);

    }

    /* Assure that we can write to and read back changes to a new object */
    @Test
    public void addCollectionsOfComponentsTest() throws Exception {
        MutableObjectProbe probe = new MutableObjectProbe();
        probe.initNew(Type.Agent, "newRW");

        Property newProperty = new Property(probe, IDENTIFIER_URL,
                "propertyValue");
        Relationship newRelationship = new Relationship(probe,
                Components.MEMBER_OF, new NDRObjectTemplate(Type.Aggregator,
                        "aggregator"));
        Datastream newDatastream = new RedirectDatastream(REDIR, new URL(
                "http://nsdl.org"), "DC", "text/html");

        probe.addProperties(Arrays.asList(newProperty));
        probe.addRelationships(Arrays.asList(newRelationship));
        probe.addDatastreams(Arrays.asList(newDatastream));

        /*
         * we should be able to read back what we added, and identify them as
         * adds
         */
        assertEquals(probe.getProperty(IDENTIFIER_URL), newProperty);
        assertEquals(probe.getRelationship(Components.MEMBER_OF),
                newRelationship);
        assertNotNull(probe.getDatastream(REDIR));

        assertTrue(probe.getPropertyTracker().getAdded().contains(newProperty));
        assertTrue(probe.getRelationshipTracker().getAdded().contains(
                newRelationship));

    }

    /* Assure that we can write to and read back changes to a new object */
    @Test
    public void addIndividualComponentsTest() throws Exception {
        MutableObjectProbe probe = new MutableObjectProbe();
        probe.initNew(Type.Agent, "newRW");

        Property newProperty = new Property(probe, IDENTIFIER_URL,
                "propertyValue");
        Relationship newRelationship = new Relationship(probe,
                Components.MEMBER_OF, new NDRObjectTemplate(Type.Aggregator,
                        "aggregator"));

        probe.addProperty(newProperty.predicate(), newProperty.value());
        probe.addRelationship(newRelationship.predicate(), newRelationship
                .related());
        Datastream d = probe.addDatastream(REDIR);
        d.setURL(new URL("http://nsdl.org"));
        d.setMimeType("text/html");

        /*
         * we should be able to read back what we added, and identify them as
         * adds
         */
        assertEquals("propertyValue", probe.getProperty(IDENTIFIER_URL).value());
        assertEquals("aggregator", probe.getRelationship(Components.MEMBER_OF)
                .related().handle());
        assertNotNull(probe.getDatastream(REDIR));

        assertTrue(probe.getPropertyTracker().getAdded().contains(newProperty));
        assertTrue(probe.getRelationshipTracker().getAdded().contains(
                newRelationship));

    }

    @Test
    public void removeNewComponentsTest() throws Exception {
        MutableObjectProbe probe = new MutableObjectProbe();
        probe.initNew(Type.Agent, "newRW");

        Property newProperty = new Property(probe, IDENTIFIER_URL,
                "propertyValue");
        Relationship newRelationship = new Relationship(probe,
                Components.MEMBER_OF, new NDRObjectTemplate(Type.Aggregator,
                        "aggregator"));
        Datastream newDatastream = new RedirectDatastream(REDIR, new URL(
                "http://nsdl.org"), "REDIR", "text/html");

        /* Add components to object */
        probe.addProperties(Arrays.asList(newProperty));
        probe.addRelationships(Arrays.asList(newRelationship));
        probe.addDatastreams(Arrays.asList(newDatastream));

        /* Delete components */
        probe.removeProperties(Arrays.asList(newProperty));
        probe.removeRelationships(Arrays.asList(newRelationship));
        probe.removeDatastreams(Arrays.asList(newDatastream));

        /* Make sure we don't get back deleted components! */
        assertNull(probe.getProperty(IDENTIFIER_URL));
        assertNull(probe.getRelationship(Components.MEMBER_OF));
        assertNull(probe.getDatastream(REDIR));

        /* Make sure these are not registered as adds any more */
        assertFalse(probe.getPropertyTracker().getAdded().contains(newProperty));
        assertFalse(probe.getRelationshipTracker().getAdded().contains(
                newRelationship));

        /* Make sure that these are NOT registered as deletes */
        assertFalse(probe.getPropertyTracker().getDeleted().contains(
                newProperty));
        assertFalse(probe.getRelationshipTracker().getDeleted().contains(
                newRelationship));
    }

    @Test
    public void removeExistingComponentsTest() throws Exception {
        MutableObjectProbe initial = new MutableObjectProbe();
        initial.initNew(Type.Agent, "newRW");

        Property newProperty = new Property(initial, IDENTIFIER_URL,
                "propertyValue");
        Relationship newRelationship = new Relationship(initial,
                Components.MEMBER_OF, new NDRObjectTemplate(Type.Aggregator,
                        "aggregator"));
        Datastream newDatastream = new RedirectDatastream(REDIR, new URL(
                "http://nsdl.org"), "REDIR", "text/html");

        /* Add components to object */
        initial.addProperties(Arrays.asList(newProperty));
        initial.addRelationships(Arrays.asList(newRelationship));
        initial.addDatastreams(Arrays.asList(newDatastream));

        MutableObjectProbe probe = new MutableObjectProbe();
        probe.init(initial);

        /* Delete components */
        probe.removeProperties(Arrays.asList(newProperty));
        probe.removeRelationships(Arrays.asList(newRelationship));
        probe.removeDatastreams(Arrays.asList(newDatastream));

        /* Make sure we don't get back deleted components! */
        assertNull(probe.getProperty(IDENTIFIER_URL));
        assertNull(probe.getRelationship(Components.MEMBER_OF));
        assertNull(probe.getDatastream(REDIR));

        /* Make sure that these are registered as deletes */
        assertTrue(probe.getPropertyTracker().getDeleted()
                .contains(newProperty));
        assertTrue(probe.getRelationshipTracker().getDeleted().contains(
                newRelationship));
    }

    /* Assure that we can read and write changes to a wrapped existing object */
    @Test
    public void existingInitTest() throws Exception {

        /* Establish an "existing" object */
        NDRObject existing = MockFedoraNDRAccess.getInstance().newObject(
                Type.Aggregator);
        Property existingProperty = new Property(existing, IDENTIFIER_URL,
                "propertyValue");
        Relationship existingRelationship = new Relationship(existing,
                Components.MEMBER_OF, new NDRObjectTemplate(Type.Aggregator,
                        "existingAggregator"));
        Datastream existingDatastream = new RedirectDatastream(REDIR, new URL(
                "http://nsdl.org"), "REDIR", "text/html");

        existing.addProperties(Arrays.asList(existingProperty));
        existing.addRelationships(Arrays.asList(existingRelationship));
        existing.addDatastreams(Arrays.asList(existingDatastream));

        MutableObjectProbe probe = new MutableObjectProbe();
        probe.init(existing);

        /* Assure that object is populated correctly */
        assertEquals(probe.getProperty(IDENTIFIER_URL), existingProperty);
        assertEquals(probe.getRelationship(Components.MEMBER_OF),
                existingRelationship);
        assertNotNull(probe.getDatastream(REDIR));

        assertFalse(probe.getPropertyTracker().getAdded().contains(
                existingProperty));
        assertFalse(probe.getRelationshipTracker().getAdded().contains(
                existingRelationship));
    }

    @Test(expected = AccessException.class)
    public void accessDuplicatePropertyTest() throws AccessException {
        MutableObjectProbe probe = new MutableObjectProbe();
        probe.initNew(Type.Agent, "newInit");

        Property dupeObjectType = new Property(probe, Components.OBJECT_TYPE,
                "dupe1");

        probe.addProperties(Arrays.asList(dupeObjectType));

        probe.getProperty(Components.OBJECT_TYPE);
    }

    @Test(expected = AccessException.class)
    public void accessDuplicateRelationshipTest() throws AccessException {
        MutableObjectProbe probe = new MutableObjectProbe();
        probe.initNew(Type.Agent, "newInit");

        MutableObjectProbe o1 = new MutableObjectProbe();
        o1.initNew(Type.Aggregator, "o1");
        MutableObjectProbe o2 = new MutableObjectProbe();
        o2.initNew(Type.Aggregator, "o2");

        Relationship mo1 = new Relationship(probe, Components.MEMBER_OF, o1);
        Relationship mo2 = new Relationship(probe, Components.MEMBER_OF, o2);

        probe.addRelationships(Arrays.asList(mo1, mo2));

        probe.getRelationship(Components.MEMBER_OF);
    }

    @Test
    public void listenerTest() {
        MutableObjectProbe probe = new MutableObjectProbe();
        probe.initNew(Type.Agent, "testAgent");
        ListenerProbe<Property> listener = new ListenerProbe<Property>();
        probe.getPropertyTracker().setListener(listener);

        probe.addProperty(IDENTIFIER_URL, "URL");

        assertEquals("Listener did not detect add.", 1, listener.added.size());

        probe.getProperty(IDENTIFIER_URL).setValue("URL2");

        assertEquals("Listener did not detect update.", 1, listener.updated
                .size());
        assertEquals("Listener did not report right version of object", "URL2",
                listener.updated.get(0).value());

        probe.removeProperty(IDENTIFIER_URL, "URL2");
        assertEquals("Listener did not detect delete.", 1, listener.removed
                .size());
        assertEquals("Listener did not report right version of object", "URL2",
                listener.removed.get(0).value());

    }

    private class ListenerProbe<T> implements MutationListener<T> {
        private ArrayList<T> added = new ArrayList<T>();
        private ArrayList<T> removed = new ArrayList<T>();
        private ArrayList<T> updated = new ArrayList<T>();

        public void notifyAdd(T addedElement) {
            added.add(addedElement);
        }

        public void notifyRemove(T removedElement) {
            removed.add(removedElement);
        }

        public void notifyUpdate(T updated) {
            this.updated.add(updated);
        }
    }

    private class MutableObjectProbe extends MutableObjectBase {

        public Date getCreatedDate() {
            return new Date();
        }

        public Date getLastModifiedDate() {
            return new Date();
        }

        public void commit() {

        }

    }

}
