/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.impl;

import java.util.Arrays;
import java.util.Collection;

import org.junit.Test;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.util.NDRObjectTemplate;
import org.nsdl.repository.util.StaticObjectInfo;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertFalse;
import static junit.framework.Assert.assertTrue;

public class RelationshipTrackerTest {
    private static final Relationship.Name q1 = new Relationship.Name(
            "info/test#1", "test_1", "t");
    private static final Relationship.Name q2 = new Relationship.Name(
            "info/test#2", "test_2", "t");
    private static final Relationship.Name q3 = new Relationship.Name(
            "info/test#3", "test_3", "t");

    private static final NDRObjectInfo o1 = new StaticObjectInfo("Test1");
    private static final NDRObjectInfo o2 = new StaticObjectInfo("Test2");
    private static final NDRObjectInfo o3 = new StaticObjectInfo("Test3");

    private static final Relationship rs1 = new Relationship(o1, q1, o1);
    private static final Relationship rs2 = new Relationship(o1, q2, o2);
    private static final Relationship rs3 = new Relationship(o1, q2, o3);
    private static Collection<Relationship> relationshipSample = Arrays.asList(
            rs1, rs2, rs3);

    @Test
    public void testSimpleAdd() {
        BasicComponentTracker<Relationship> tracker = getPopulatedTracker();

        Relationship added = new Relationship(o1, q3, new StaticObjectInfo(
                "TestSimpleAdd"));
        tracker.add(added);

        assertTrue(tracker.hasModified());
        assertEquals(added, tracker.getAdded().iterator().next());
    }

    @Test
    public void testSimpleDelete() {
        BasicComponentTracker<Relationship> tracker = getPopulatedTracker();

        tracker.remove(rs2);
        tracker.remove(rs3);

        assertTrue(tracker.hasModified());
        assertEquals(1, tracker.getCurrent().size());
        assertEquals(2, tracker.getDeleted().size());
        assertFalse(tracker.getCurrent().contains(rs2));
        assertFalse(tracker.getCurrent().contains(rs3));
        assertTrue(tracker.getDeleted().contains(rs2));
        assertTrue(tracker.getDeleted().contains(rs3));
    }

    @Test
    public void testSimpleReplace() {
        BasicComponentTracker<Relationship> tracker = getPopulatedTracker();

        Relationship p = tracker.get(q1).iterator().next();
        p.setRelated(new StaticObjectInfo("replaceValue"));

        assertTrue("Tracker did not detect update", tracker.hasModified());
        assertEquals(tracker.getCurrent().size(), relationshipSample.size());

        Collection<Relationship> updated = tracker.getUpdated();

        assertTrue(tracker.getUpdated().contains(rs1));
        updated.clear();
    }

    @Test
    public void testDuplicateEqualitySkip() {
        BasicComponentTracker<Relationship> tracker = getPopulatedTracker();

        tracker.add(rs1);

        assertFalse(tracker.hasModified());
        assertEquals(tracker.getAdded().size(), 0);
    }

    public static RelationshipTracker getPopulatedTracker() {
        NDRObject base = new NDRObjectTemplate(null);
        base.addRelationships(relationshipSample);
        return new RelationshipTracker(base);
    }

}
