/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.model.types;

import java.io.ByteArrayInputStream;
import java.net.URL;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.Datastream;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.types.Grouping.Branding;

import static junit.framework.Assert.assertEquals;

public class GroupingTests {

    private static MockFedoraNDRAccess ndr;

    @BeforeClass
    public static void setupNDR() {
        ndr = MockFedoraNDRAccess.getInstance();
    }

    @Test
    public void newBrandingTest() throws Exception {
        Aggregator agg = ndr.newObject(Type.Aggregator);
        Branding brand = agg.getBranding();

        /* just make sure we don't die by accessing an uninitialized branding */
        brand.getBrandImageHeight();
        brand.getBrandImageWidth();
        brand.getBrandImageTitle();
        brand.getBrandImageURL();
        brand.getDescription();
        brand.getTitle();

        /* Set values */
        brand.setBrandImage(new URL("http://example.org"), "example title", 12,
                10);
        brand.setDescription("This is a description");
        brand.setTitle("This is a title");

        /* Get them back */
        assertEquals("URL is incorrect", "http://example.org", brand
                .getBrandImageURL().toString());
        assertEquals("Width is incorrect", 12, brand.getBrandImageWidth());
        assertEquals("Height is incorrect", 10, brand.getBrandImageHeight());
        assertEquals("Image title is incorrect", "example title", brand
                .getBrandImageTitle());
        assertEquals("Title is incorrect", "This is a title", brand.getTitle());
        assertEquals("Description is incorrect", "This is a description", brand
                .getDescription());
    }

    @Test
    public void partialContentTest() throws Exception {
        NDRObject newAgg = ndr.newObject(Type.Aggregator);
        Datastream serviceDescription = newAgg
                .addDatastream(Aggregator.SERVICE_DESCRIPTION);
        serviceDescription
                .setContent(new ByteArrayInputStream(
                        ("<serviceDescription xmlns:dc=\"http://purl.org/dc/elements/1.1/\">\n"
                                + "<dc:title>TITLE</dc:title>"
                                + "<dc:description>DESCRIPTION</dc:description>"
                                + "</serviceDescription>").getBytes()));
        newAgg.commit();

        Aggregator asAgg = ndr.getObject(newAgg, Type.Aggregator);
        assertEquals("TITLE", asAgg.getBranding().getTitle());
        assertEquals("DESCRIPTION", asAgg.getBranding().getDescription());

        /* Make sure we don't throw an exception here */
        asAgg.getBranding().setBrandImage(new URL("http://nsdl.org/nsdl.jpg"),
                "title", 32, 32);
    }

    @Test
    public void persistentBrandingTest() throws Exception {
        Aggregator agg = ndr.newObject(Type.Aggregator);
        Branding brand = agg.getBranding();

        /* Set values */
        brand.setBrandImage(new URL("http://example.org"), "example title", 12,
                10);
        brand.setDescription("This is a description");
        brand.setTitle("This is a title");

        agg.commit();

        brand = ndr.getObject(agg, Type.Aggregator).getBranding();

        /* Get them back */
        assertEquals("URL is incorrect", "http://example.org", brand
                .getBrandImageURL().toString());
        assertEquals("Width is incorrect", 12, brand.getBrandImageWidth());
        assertEquals("Height is incorrect", 10, brand.getBrandImageHeight());
        assertEquals("Image title is incorrect", "example title", brand
                .getBrandImageTitle());
        assertEquals("Title is incorrect", "This is a title", brand.getTitle());
        assertEquals("Description is incorrect", "This is a description", brand
                .getDescription());
    }
}
