/*
 * ----------------------------------------------------------------------------
 *
 * License and Copyright:
 *
 * The contents of this file are subject to the Educational Community
 * License v1.0  (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at
 * http://www.opensource.org/licenses/ecl1.php.
 *
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2002-2009 by Cornell University All rights reserved.
 *
 * ----------------------------------------------------------------------------
 */
 
package org.nsdl.repository.tools;

import junit.framework.Assert;

import org.junit.BeforeClass;
import org.junit.Test;
import org.nsdl.repository.impl.fedora.MockFedoraNDRAccess;
import org.nsdl.repository.model.ExistenceException;
import org.nsdl.repository.model.NDRObject;
import org.nsdl.repository.model.NDRObjectInfo;
import org.nsdl.repository.model.Property;
import org.nsdl.repository.model.Relationship;
import org.nsdl.repository.model.types.Aggregator;
import org.nsdl.repository.model.types.Metadata;
import org.nsdl.repository.model.types.MetadataProvider;
import org.nsdl.repository.model.types.Resource;
import org.nsdl.repository.model.types.Type;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertFalse;
import static junit.framework.Assert.assertTrue;

public class ResourceMergerTest {

    private static MockFedoraNDRAccess ndr;
    private static ResourceMerger merger;

    @BeforeClass
    public static void init() {
        ndr = MockFedoraNDRAccess.getInstance();
        merger = new ResourceMerger(ndr);
    }

    /** Makes sure all relationships are merged */
    @Test
    public void mergedInternalRelationshipsTest() {
        final String URL = "http://mergedRelationshipTest";
        Aggregator aShared = ndr.newObject(Type.Aggregator);
        Aggregator a1 = ndr.newObject(Type.Aggregator);
        Aggregator a2 = ndr.newObject(Type.Aggregator);
        aShared.commit();
        a1.commit();
        a2.commit();

        Resource r1 = ndr.newObject(Type.Resource);
        r1.addProperty(Resource.IDENTIFIER_URL, URL);
        r1.addRelationship(Aggregator.MEMBER_OF, aShared);
        r1.addRelationship(Aggregator.MEMBER_OF, a1);
        r1.commit();

        Resource r2 = ndr.newObject(Type.Resource);
        r2.addProperty(Resource.IDENTIFIER_URL, URL);
        r2.addRelationship(Aggregator.MEMBER_OF, aShared);
        r2.addRelationship(Aggregator.MEMBER_OF, a2);
        r2.commit();

        merger.mergeTo(r1, r2);

        assertEquals("Wrong number of merged relationships!", 3, r1
                .getRelationships().size());
        assertTrue(relationshipExists(r1, Resource.MEMBER_OF, a1));
        assertTrue(relationshipExists(r1, Resource.MEMBER_OF, a2));
        assertTrue(relationshipExists(r1, Resource.MEMBER_OF, aShared));
    }

    /** Makes sure all properties are merged */
    @Test
    public void mergedPropertiesTest() {
        final String URL = "http://mergedPropertiesTest";
        Resource r1 = ndr.newObject(Type.Resource);
        r1.addProperty(Resource.IDENTIFIER_URL, URL);
        r1.addProperty(Metadata.UNIQUE_ID, "r1");
        r1.commit();

        Resource r2 = ndr.newObject(Type.Resource);
        r2.addProperty(Resource.IDENTIFIER_URL, URL);
        r2.addProperty(Metadata.UNIQUE_ID, "r2");
        r2.commit();

        merger.mergeTo(r1, r2);
        assertEquals("Wrong number of merged properties", 5, r1.getProperties()
                .size());
        assertTrue(propertyExists(r1, Resource.OBJECT_TYPE, "Resource"));
        assertTrue(propertyExists(r1, Resource.HAS_HANDLE, r1.handle()));
        assertTrue(propertyExists(r1, Resource.IDENTIFIER_URL, URL));
        assertTrue(propertyExists(r1, Metadata.UNIQUE_ID, "r1"));
        assertTrue(propertyExists(r1, Metadata.UNIQUE_ID, "r2"));

    }

    /** makes sure now-obsolete resources are purged */
    @Test
    public void removeMergedResourceTest() {
        Resource r1 = ndr.newObject(Type.Resource);
        r1.commit();

        Resource r2 = ndr.newObject(Type.Resource);
        r2.commit();

        Resource r3 = ndr.newObject(Type.Resource);
        r3.commit();

        merger.mergeTo(r3, r2, r1);

        /* Shouldn't fail */
        ndr.getObject(r3);

        try {
            ndr.getObject(r2);
            Assert.fail("Now obsolete resource should be purged!");
        } catch (ExistenceException e) {
            /* Expected */
        }

        try {
            ndr.getObject(r1);
            Assert.fail("Now obsolete resource should be purged!");
        } catch (ExistenceException e) {
            /* Expected */
        }
    }

    /** Makes sure all external relationships point to single resource */
    @Test
    public void externalRelationshipChangeTest() {
        Resource r1 = ndr.newObject(Type.Resource);
        r1.commit();

        Resource r2 = ndr.newObject(Type.Resource);
        r2.commit();

        Metadata m1 = ndr.newObject(Type.Metadata);
        m1.addRelationship(Metadata.METADATA_FOR, r1);
        m1.commit();

        MetadataProvider mp2 = ndr.newObject(Type.MetadataProvider);
        mp2.commit();

        Metadata m2 = ndr.newObject(Type.Metadata);
        m2.addRelationship(Metadata.METADATA_FOR, r2);
        m2.addRelationship(Metadata.METADATA_PROVIDED_BY, mp2);
        m2.commit();

        merger.mergeTo(r1, r2);

        /* re-fresh metadata from repository */
        m1 = ndr.getObject(m1, Type.Metadata);
        m2 = ndr.getObject(m2, Type.Metadata);

        /* Make sure we didn't change r1's relationships */
        assertTrue(relationshipExists(m1, Metadata.METADATA_FOR, r1));
        assertFalse(relationshipExists(m1, Metadata.METADATA_FOR, r2));

        /* Make sure we changed m2 relationships pointing to r2 */
        assertTrue(relationshipExists(m2, Metadata.METADATA_FOR, r1));
        assertFalse(relationshipExists(m2, Metadata.METADATA_FOR, r2));

        /* Should not have affected an unrelated relationship */
        assertTrue(relationshipExists(m2, Metadata.METADATA_PROVIDED_BY, mp2));
    }

    /** With multiple resources and 'mergeAll', merge into the oldest */
    @Test
    public void oldestResourceTest() {
        Resource r1 = ndr.newObject(Type.Resource);
        r1.commit();

        Resource r2 = ndr.newObject(Type.Resource);
        r2.commit();

        Resource r3 = ndr.newObject(Type.Resource);
        r3.commit();

        Resource merged = merger.mergeAll(r1, r2, r3);

        assertTrue(merged.handle().equals(r1.handle()));
    }

    private boolean relationshipExists(NDRObject o, Relationship.Name name,
            NDRObjectInfo related) {
        for (Relationship r : o.getRelationships(name)) {
            if (r.related().handle().equals(related.handle())) {
                return true;
            }
        }
        return false;
    }

    private boolean propertyExists(NDRObject o, Property.Name name, String value) {
        for (Property p : o.getProperties(name)) {
            if (p.value().equals(value)) {
                return true;
            }
        }
        return false;
    }
}
